/* cnic.c: Broadcom CNIC core network driver.
 *
 * Copyright (c) 2006-2007 Broadcom Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation.
 *
 * Written by: John(Zongxi) Chen (zongxi@broadcom.com)
 * Modified and maintained by: Michael Chan <mchan@broadcom.com>
 */

#include <linux/version.h>
#if (LINUX_VERSION_CODE < 0x020612)
#include <linux/config.h>
#endif

#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/list.h>
#include <linux/slab.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/netdevice.h>
#include <linux/inetdevice.h>
#include <linux/in.h>
#include <linux/dma-mapping.h>
#include <asm/byteorder.h>
#include <linux/delay.h>
#include <linux/ethtool.h>
#include <linux/if_vlan.h>
#define BCM_VLAN 1
#include <net/ip.h>
#include <net/tcp.h>
#include <net/arp.h>
#include <net/neighbour.h>
#include <net/route.h>
#include <net/ipv6.h>
#include <net/ip6_route.h>
#include <net/addrconf.h>
#include <net/netevent.h>

#ifdef __VMKLNX__
#include "iscsi_net.h"
#endif
#include "cnic_drv.h"
#include "bnx2.h"
#include "cnic_if.h"
#include "cnic.h"
#include "cnic_cm.h"

#define DRV_MODULE_NAME 	"cnic"
#define PFX DRV_MODULE_NAME ": "

static char version[] __devinitdata =
	"Broadcom NetXtreme II CNIC Driver " DRV_MODULE_NAME " v" CNIC_MODULE_VERSION " (" CNIC_MODULE_RELDATE ")\n";

MODULE_AUTHOR("John(Zongxi) Chen <zongxic@broadcom.com>");
MODULE_DESCRIPTION("Broadcom NetXtreme II CNIC Driver");
MODULE_LICENSE("GPL");
MODULE_VERSION(CNIC_MODULE_VERSION);

static LIST_HEAD(cnic_dev_list);
static DEFINE_RWLOCK(cnic_dev_lock);
static DEFINE_MUTEX(cnic_lock);

struct cnic_ulp_ops *cnic_ulp_tbl[MAX_CNIC_ULP_TYPE];

static int cnic_service_bnx2(void *, void *);
static void cnic_service_stop(void *);
static void cnic_service_start(void *);

static struct cnic_ops cnic_bnx2_ops = {
	.cnic_owner	= THIS_MODULE,
	.cnic_handler	= cnic_service_bnx2,
	.cnic_stop	= cnic_service_stop,
	.cnic_start	= cnic_service_start,
};

static inline void cnic_hold(struct cnic_dev *dev)
{
	atomic_inc(&dev->ref_count);
}

static inline void cnic_put(struct cnic_dev *dev)
{
	atomic_dec(&dev->ref_count);
}

static void cnic_bnx2_ctx_wr(struct cnic_dev *dev, u32 cid_addr, u32 off,
			     u32 val)
{
	struct cnic_local *cp = dev->cnic_priv;
	struct cnic_eth_dev *ethdev = cp->ethdev;

	ethdev->drv_ctx_wr(dev->netdev, cid_addr, off, val);
}

static void cnic_bnx2_reg_wr_ind(struct cnic_dev *dev, u32 off, u32 val)
{
	struct cnic_local *cp = dev->cnic_priv;
	struct cnic_eth_dev *ethdev = cp->ethdev;

	ethdev->drv_reg_wr(dev->netdev, off, val);
}

static u32 cnic_bnx2_reg_rd_ind(struct cnic_dev *dev, u32 off)
{
	struct cnic_local *cp = dev->cnic_priv;
	struct cnic_eth_dev *ethdev = cp->ethdev;

	return (ethdev->drv_reg_rd(dev->netdev, off));
}

int cnic_register_driver(int ulp_type, struct cnic_ulp_ops *ulp_ops)
{
	struct cnic_dev *dev;

	if (ulp_type >= MAX_CNIC_ULP_TYPE) {
		printk(KERN_ERR PFX "cnic_register_driver: Bad type %d\n",
		       ulp_type);
		return -EINVAL;
	}
	mutex_lock(&cnic_lock);
	if (cnic_ulp_tbl[ulp_type]) {
		printk(KERN_ERR PFX "cnic_register_driver: Type %d has already "
				    "been registered\n", ulp_type);
		mutex_unlock(&cnic_lock);
		return -EBUSY;
	}

#ifndef __VMKLNX__
	read_lock(&cnic_dev_lock);
#endif
	list_for_each_entry(dev, &cnic_dev_list, list) {
		struct cnic_local *cp = dev->cnic_priv;

		clear_bit(ULP_F_INIT, &cp->ulp_flags[ulp_type]);
	}
#ifndef __VMKLNX__
	read_unlock(&cnic_dev_lock);
#endif

	rcu_assign_pointer(cnic_ulp_tbl[ulp_type], ulp_ops);
	mutex_unlock(&cnic_lock);

	/* Prevent race conditions with netdev_event */
	rtnl_lock();
#ifndef __VMKLNX__
	read_lock(&cnic_dev_lock);
#endif
	list_for_each_entry(dev, &cnic_dev_list, list) {
		struct cnic_local *cp = dev->cnic_priv;

		if (!test_and_set_bit(ULP_F_INIT, &cp->ulp_flags[ulp_type]))
			ulp_ops->cnic_init(dev);
	}
#ifndef __VMKLNX__
	read_unlock(&cnic_dev_lock);
#endif
	rtnl_unlock();

	return 0;
}

int cnic_unregister_driver(int ulp_type)
{
	struct cnic_dev *dev;

	if (ulp_type >= MAX_CNIC_ULP_TYPE) {
		printk(KERN_ERR PFX "cnic_unregister_driver: Bad type %d\n",
		       ulp_type);
		return -EINVAL;
	}
	mutex_lock(&cnic_lock);
	if (!cnic_ulp_tbl[ulp_type]) {
		printk(KERN_ERR PFX "cnic_unregister_driver: Type %d has not "
				    "been registered\n", ulp_type);
		goto out_unlock;
	}
#ifndef __VMKLNX__
	read_lock(&cnic_dev_lock);
#endif
	list_for_each_entry(dev, &cnic_dev_list, list) {
		struct cnic_local *cp = dev->cnic_priv;
		
		if (rcu_dereference(cp->ulp_ops[ulp_type])) {
			printk(KERN_ERR PFX "cnic_unregister_driver: Type %d "
			       "still has devices registered\n", ulp_type);
			read_unlock(&cnic_dev_lock);
			goto out_unlock;
		}
	}
#ifndef __VMKLNX__
	read_unlock(&cnic_dev_lock);
#endif

	rcu_assign_pointer(cnic_ulp_tbl[ulp_type], NULL);

	mutex_unlock(&cnic_lock);
#ifndef __VMKLNX__
	synchronize_rcu();
#endif
	return 0;

out_unlock:
	mutex_unlock(&cnic_lock);
	return -EINVAL;
}

EXPORT_SYMBOL(cnic_register_driver);
EXPORT_SYMBOL(cnic_unregister_driver);

static int cnic_start_hw(struct cnic_dev *);
static void cnic_stop_hw(struct cnic_dev *);

static int cnic_register_device(struct cnic_dev *dev, int ulp_type,
				void *ulp_ctx)
{
	struct cnic_local *cp = dev->cnic_priv;
	struct cnic_ulp_ops *ulp_ops;

	if (ulp_type >= MAX_CNIC_ULP_TYPE) {
		printk(KERN_ERR PFX "cnic_register_device: Bad type %d\n",
		       ulp_type);
		return -EINVAL;
	}
	mutex_lock(&cnic_lock);
	if (cnic_ulp_tbl[ulp_type] == NULL) {
		printk(KERN_ERR PFX "cnic_register_device: Driver with type %d "
				    "has not been registered\n", ulp_type);
		mutex_unlock(&cnic_lock);
		return -EAGAIN;
	}
	if (rcu_dereference(cp->ulp_ops[ulp_type])) {
		printk(KERN_ERR PFX "cnic_register_device: Type %d has already "
		       "been registered to this device\n", ulp_type);
		mutex_unlock(&cnic_lock);
		return -EBUSY;
	}
	if (!try_module_get(cnic_ulp_tbl[ulp_type]->owner)) {
		mutex_unlock(&cnic_lock);
		return -EBUSY;
	}

	clear_bit(ULP_F_START, &cp->ulp_flags[ulp_type]);
	cp->ulp_handle[ulp_type] = ulp_ctx;
	ulp_ops = cnic_ulp_tbl[ulp_type];
	rcu_assign_pointer(cp->ulp_ops[ulp_type], ulp_ops);
	cnic_hold(dev);
	if (!dev->use_count) {
		if (!test_bit(CNIC_F_IF_GOING_DOWN, &dev->flags)) {
			if (dev->netdev->flags & IFF_UP)
				set_bit(CNIC_F_IF_UP, &dev->flags);
		}
	}
	dev->use_count++;

	if (dev->use_count == 1) {
		if (test_bit(CNIC_F_IF_UP, &dev->flags))
			cnic_start_hw(dev);
	}

	if (test_bit(CNIC_F_CNIC_UP, &dev->flags))
		if (!test_and_set_bit(ULP_F_START, &cp->ulp_flags[ulp_type]))
			ulp_ops->cnic_start(cp->ulp_handle[ulp_type]);

	mutex_unlock(&cnic_lock);

	return 0;

}

static int cnic_unregister_device(struct cnic_dev *dev, int ulp_type)
{
	struct cnic_local *cp = dev->cnic_priv;

	if (ulp_type >= MAX_CNIC_ULP_TYPE) {
		printk(KERN_ERR PFX "cnic_unregister_device: Bad type %d\n",
		       ulp_type);
		return -EINVAL;
	}
	mutex_lock(&cnic_lock);
	if (rcu_dereference(cp->ulp_ops[ulp_type])) {
		dev->use_count--;
		module_put(cp->ulp_ops[ulp_type]->owner);
		rcu_assign_pointer(cp->ulp_ops[ulp_type], NULL);
		if (dev->use_count == 0) {
			cnic_stop_hw(dev);
#ifdef __VMKLNX__
			cp->vmknic_valid = 0;
#endif
		}
		cnic_put(dev);
	} else {
		printk(KERN_ERR PFX "cnic_unregister_device: device not "
		       "registered to this ulp type %d\n", ulp_type);
		mutex_unlock(&cnic_lock);
		return -EINVAL;
	}
	mutex_unlock(&cnic_lock);

#ifndef __VMKLNX__
	synchronize_rcu();
#endif

	return 0;
}

static void cnic_free_dma(struct cnic_dev *dev, struct cnic_dma *dma)
{
	int i;

	if (!dma->pg_arr)
		return;

	for (i = 0; i < dma->num_pages; i++) {
		if (dma->pg_arr[i]) {
			pci_free_consistent(dev->pcidev, BCM_PAGE_SIZE,
					    dma->pg_arr[i], dma->pg_map_arr[i]);
			dma->pg_arr[i] = NULL;
		}
	}
	if (dma->pgtbl) {
		pci_free_consistent(dev->pcidev, dma->pgtbl_size,
				    dma->pgtbl, dma->pgtbl_map);
		dma->pgtbl = NULL;
	}
	kfree(dma->pg_arr);
	dma->pg_arr = NULL;
	dma->num_pages = 0;
}

static void cnic_setup_page_tbl(struct cnic_dev *dev, struct cnic_dma *dma)
{
	int i;
	u32 *page_table = dma->pgtbl;

	for (i = 0; i < dma->num_pages; i++) {
		/* Each entry needs to be in big endian format. */
		*page_table = (u32) ((u64) dma->pg_map_arr[i] >> 32);
		page_table++;
		*page_table = (u32) dma->pg_map_arr[i];
		page_table++;
	}
}

static int cnic_alloc_dma(struct cnic_dev *dev, struct cnic_dma *dma,
			  int pages, int use_pg_tbl)
{
	int i, size;
	struct cnic_local *cp = dev->cnic_priv;

	size = pages * (sizeof(void *) + sizeof(dma_addr_t));
	dma->pg_arr = kzalloc(size, GFP_ATOMIC);
	if (dma->pg_arr == NULL)
		return -ENOMEM;

	dma->pg_map_arr = (dma_addr_t *) (dma->pg_arr + pages);
	dma->num_pages = pages;

	for (i = 0; i < pages; i++) {
		dma->pg_arr[i] = pci_alloc_consistent(dev->pcidev,
						      BCM_PAGE_SIZE,
						      &dma->pg_map_arr[i]);
		if (dma->pg_arr[i] == NULL)
			goto error;
	}
	if (!use_pg_tbl)
		return 0;

	dma->pgtbl_size = ((pages * 8) + BCM_PAGE_SIZE - 1) &
			  (BCM_PAGE_SIZE - 1);
	dma->pgtbl = pci_alloc_consistent(dev->pcidev, dma->pgtbl_size,
					  &dma->pgtbl_map);
	if (dma->pgtbl == NULL)
		goto error;

	cp->setup_pgtbl(dev, dma);

	return 0;

error:
	cnic_free_dma(dev, dma);
	return -ENOMEM;
}

static void cnic_free_resc(struct cnic_dev *dev)
{
	struct cnic_local *cp = dev->cnic_priv;
	int i;

	for (i = 0; i < cp->ctx_blks; i++) {
		if (cp->ctx_arr[i].ctx) {
			pci_free_consistent(dev->pcidev, cp->ctx_blk_size,
					    cp->ctx_arr[i].ctx,
					    cp->ctx_arr[i].mapping);
			cp->ctx_arr[i].ctx = NULL;
		}
	}
	kfree(cp->ctx_arr);
	cp->ctx_arr = NULL;
	cp->ctx_blks = 0;

	cnic_free_dma(dev, &cp->kwq_info);
	cnic_free_dma(dev, &cp->kcq_info);
	return;
}

static int cnic_alloc_context(struct cnic_dev *dev)
{
	struct cnic_local *cp = dev->cnic_priv;

	if (CHIP_NUM(cp) == CHIP_NUM_5709) {
		int i, k, arr_size;

		cp->ctx_blk_size = BCM_PAGE_SIZE;
		cp->cids_per_blk = BCM_PAGE_SIZE / 128;
		arr_size = BNX2_MAX_CID / cp->cids_per_blk *
			   sizeof(struct cnic_ctx);
		cp->ctx_arr = kmalloc(arr_size, GFP_KERNEL);
		if (cp->ctx_arr == NULL)
			return -ENOMEM;

		memset(cp->ctx_arr, 0, arr_size);

		k = 0;
		for (i = 0; i < 2; i++) {
			u32 j, reg, off, lo, hi;

			if (i == 0)
				off = BNX2_PG_CTX_MAP;
			else
				off = BNX2_ISCSI_CTX_MAP;

			reg = cnic_bnx2_reg_rd_ind(dev, off);
			lo = reg >> 16;
			hi = reg & 0xffff;
			for (j = lo; j < hi; j += cp->cids_per_blk, k++)
				cp->ctx_arr[k].cid = j;
		}

		cp->ctx_blks = k;

		for (i = 0; i < cp->ctx_blks; i++) {
			cp->ctx_arr[i].ctx =
				pci_alloc_consistent(dev->pcidev, BCM_PAGE_SIZE,
						     &cp->ctx_arr[i].mapping);
			if (cp->ctx_arr[i].ctx == NULL)
				return -ENOMEM;
		}
	}
	return 0;
}

static int cnic_alloc_bnx2_resc(struct cnic_dev *dev)
{
	struct cnic_local *cp = dev->cnic_priv;
	int ret;

	ret = cnic_alloc_dma(dev, &cp->kwq_info, KWQ_PAGE_CNT, 1);
	if (ret)
		goto error;
	cp->kwq = (struct kwqe **) cp->kwq_info.pg_arr;

	ret = cnic_alloc_dma(dev, &cp->kcq_info, KCQ_PAGE_CNT, 1);
	if (ret)
		goto error;
	cp->kcq = (struct kcqe **) cp->kcq_info.pg_arr;

	ret = cnic_alloc_context(dev);
	if (ret)
		goto error;

	return 0;
	
error:
	cnic_free_resc(dev);
	return ret;
}

static inline u32 cnic_kwq_avail(struct cnic_local *cp)
{
	return (cp->max_kwq_idx -
		((cp->kwq_prod_idx - cp->kwq_con_idx) & cp->max_kwq_idx));
}

static int cnic_submit_bnx2_kwqes(struct cnic_dev *dev, struct kwqe *wqes[],
				  u32 num_wqes)
{
	struct cnic_local *cp = dev->cnic_priv;
	struct kwqe *prod_qe;
	u16 prod, sw_prod, i;

	if (!test_bit(CNIC_F_CNIC_UP, &dev->flags))
		return -EAGAIN;		/* bnx2 is down */

	spin_lock_bh(&cp->cnic_ulp_lock);
	if (num_wqes > cnic_kwq_avail(cp) &&
	    !(cp->cnic_local_flags & CNIC_LCL_FL_KWQ_INIT)) {
		spin_unlock_bh(&cp->cnic_ulp_lock);
		return -EAGAIN;
	}

	cp->cnic_local_flags &= ~CNIC_LCL_FL_KWQ_INIT;

	prod = cp->kwq_prod_idx;
	sw_prod = prod & MAX_KWQ_IDX;
	for (i = 0; i < num_wqes; i++) {
		prod_qe = &cp->kwq[KWQ_PG(sw_prod)][KWQ_IDX(sw_prod)];
		memcpy(prod_qe, wqes[i], sizeof(struct kwqe));
		prod++;
		sw_prod = prod & MAX_KWQ_IDX;
	}
	cp->kwq_prod_idx = prod;

	CNIC_WR16(dev, cp->kwq_io_addr, cp->kwq_prod_idx);

	spin_unlock_bh(&cp->cnic_ulp_lock);
	return 0;
}

static void service_kcqes(struct cnic_dev *dev, int num_cqes)
{
	struct cnic_local *cp = dev->cnic_priv;
	int i, j;

	i = 0;
	j = 1;
	while (num_cqes) {
		struct cnic_ulp_ops *ulp_ops;
		int ulp_type;
		u32 kcqe_op_flag = cp->completed_kcq[i]->kcqe_op_flag;
		u32 kcqe_layer = kcqe_op_flag & KCQE_FLAGS_LAYER_MASK;

		while (j < num_cqes) {
			u32 next_op = cp->completed_kcq[i + j]->kcqe_op_flag;

			if ((next_op & KCQE_FLAGS_LAYER_MASK) != kcqe_layer)
				break;
			j++;
		}

		if (kcqe_layer == KCQE_FLAGS_LAYER_MASK_L5_RDMA)
			ulp_type = CNIC_ULP_RDMA;
		else if (kcqe_layer == KCQE_FLAGS_LAYER_MASK_L5_ISCSI)
			ulp_type = CNIC_ULP_ISCSI;
		else if (kcqe_layer == KCQE_FLAGS_LAYER_MASK_L4)
			ulp_type = CNIC_ULP_L4;
		else {
			printk(KERN_ERR PFX "%s: Unknown type of KCQE(0x%x)\n",
			       dev->netdev->name, kcqe_op_flag);
			goto end;
		}

		rcu_read_lock();
		ulp_ops = rcu_dereference(cp->ulp_ops[ulp_type]);
		if (likely(ulp_ops)) {
			ulp_ops->indicate_kcqes(cp->ulp_handle[ulp_type],
						  cp->completed_kcq + i, j);
		}
		rcu_read_unlock();
end:
		num_cqes -= j;
		i += j;
		j = 1;
	}
	return;
}

static int cnic_get_kcqes(struct cnic_dev *dev, u16 hw_prod, u16 sw_prod)
{
	struct cnic_local *cp = dev->cnic_priv;
	u16 i, max;
	struct kcqe *kcqe;
	int kcqe_cnt = 0;

	i = sw_prod & MAX_KCQ_IDX;
	max = hw_prod & MAX_KCQ_IDX;
	while ((i != max) && (kcqe_cnt < MAX_COMPLETED_KCQE)) {
		cp->completed_kcq[kcqe_cnt++] = &cp->kcq[KCQ_PG(i)][KCQ_IDX(i)];
		i = (i + 1) & MAX_KCQ_IDX;
	}

	kcqe = cp->completed_kcq[kcqe_cnt - 1];
	while (kcqe->kcqe_op_flag & KCQE_FLAGS_NEXT) {
		kcqe_cnt--;
		if (kcqe_cnt == 0)
			goto done;
		kcqe = cp->completed_kcq[kcqe_cnt - 1];
	}
done:
	return kcqe_cnt;
}

static int cnic_service_bnx2(void *data, void *status_blk)
{
	struct cnic_dev *dev = data;
	struct status_block *sblk = status_blk;
	struct cnic_local *cp = dev->cnic_priv;
	u32 status_idx = sblk->status_idx;
	u16 hw_prod, sw_prod;
	int kcqe_cnt;

	if (unlikely(!test_bit(CNIC_F_CNIC_UP, &dev->flags)))
		return status_idx;

	cp->kwq_con_idx = sblk->status_rx_quick_consumer_index15;

	hw_prod = sblk->status_completion_producer_index;
	sw_prod = cp->kcq_prod_idx;
	while (sw_prod != hw_prod) {
		kcqe_cnt = cnic_get_kcqes(dev, hw_prod, sw_prod);
		if (kcqe_cnt == 0)
			goto done;
		sw_prod += kcqe_cnt;

		service_kcqes(dev, kcqe_cnt);

		/* Tell compiler that status_blk fields can change. */
		barrier();
		if (status_idx != sblk->status_idx) {
			status_idx = sblk->status_idx;
			cp->kwq_con_idx =
				sblk->status_rx_quick_consumer_index15;
			hw_prod = sblk->status_completion_producer_index;
		} else
			break;
	}

done:
	CNIC_WR16(dev, cp->kcq_io_addr, sw_prod);

	cp->kcq_prod_idx = sw_prod;
	return status_idx;
}

static void cnic_service_bnx2_msix(unsigned long data)
{
	struct cnic_dev *dev = (struct cnic_dev *) data;
	struct cnic_local *cp = dev->cnic_priv;
	struct status_block_msix *status_blk = cp->bnx2_status_blk;
	u32 status_idx = status_blk->status_idx;
	u16 hw_prod, sw_prod;
	int kcqe_cnt;

	cp->kwq_con_idx = status_blk->status_cmd_consumer_index;

	hw_prod = status_blk->status_completion_producer_index;
	sw_prod = cp->kcq_prod_idx;
	while (sw_prod != hw_prod) {
		kcqe_cnt = cnic_get_kcqes(dev, hw_prod, sw_prod);
		if (kcqe_cnt == 0)
			goto done;
		sw_prod += kcqe_cnt;

		service_kcqes(dev, kcqe_cnt);

		/* Tell compiler that status_blk fields can change. */
		barrier();
		if (status_idx != status_blk->status_idx) {
			status_idx = status_blk->status_idx;
			cp->kwq_con_idx = status_blk->status_cmd_consumer_index;
			hw_prod = status_blk->status_completion_producer_index;
		} else
			break;
	}

done:
	CNIC_WR16(dev, cp->kcq_io_addr, sw_prod);

	cp->kcq_prod_idx = sw_prod;
	cp->last_status_idx = status_idx;
	CNIC_WR(dev, BNX2_PCICFG_INT_ACK_CMD, BNX2_CNIC_INT_NUM |
		BNX2_PCICFG_INT_ACK_CMD_INDEX_VALID | cp->last_status_idx);
}

#if (LINUX_VERSION_CODE >= 0x20613)
static irqreturn_t cnic_irq(int irq, void *dev_instance)
#else
static irqreturn_t cnic_irq(int irq, void *dev_instance, struct pt_regs *regs)
#endif
{
	struct cnic_dev *dev = dev_instance;
	struct cnic_local *cp = dev->cnic_priv;
	u16 prod = cp->kcq_prod_idx & MAX_KCQ_IDX;

	if (cp->ack_int)
		cp->ack_int(dev);

	prefetch(cp->status_blk);
	prefetch(&cp->kcq[KCQ_PG(prod)][KCQ_IDX(prod)]);

	if (likely(test_bit(CNIC_F_CNIC_UP, &dev->flags)))
		tasklet_schedule(&cp->cnic_irq_task);

	return IRQ_HANDLED;
}

static void cnic_ulp_stop(struct cnic_dev *dev)
{
	struct cnic_local *cp = dev->cnic_priv;
	int if_type;

	rcu_read_lock();
	for (if_type = 0; if_type < MAX_CNIC_ULP_TYPE; if_type++) {
		struct cnic_ulp_ops *ulp_ops;

		ulp_ops = rcu_dereference(cp->ulp_ops[if_type]);
		if (!ulp_ops)
			continue;

		if (test_and_clear_bit(ULP_F_START, &cp->ulp_flags[if_type]))
			ulp_ops->cnic_stop(cp->ulp_handle[if_type]);
	}
	rcu_read_unlock();
}

static void cnic_ulp_start(struct cnic_dev *dev)
{
	struct cnic_local *cp = dev->cnic_priv;
	int if_type;

	rcu_read_lock();
	for (if_type = 0; if_type < MAX_CNIC_ULP_TYPE; if_type++) {
		struct cnic_ulp_ops *ulp_ops;

		ulp_ops = rcu_dereference(cp->ulp_ops[if_type]);
		if (!ulp_ops)
			continue;

		if (!test_and_set_bit(ULP_F_START, &cp->ulp_flags[if_type]))
			ulp_ops->cnic_start(cp->ulp_handle[if_type]);
	}
	rcu_read_unlock();
}

static void cnic_service_start(void *data)
{
	struct cnic_dev *dev = data;

	cnic_hold(dev);
	mutex_lock(&cnic_lock);
	set_bit(CNIC_F_IF_UP, &dev->flags);
	if (dev->use_count) {
		if (!cnic_start_hw(dev))
			cnic_ulp_start(dev);
	}
	mutex_unlock(&cnic_lock);
	cnic_put(dev);
}

static void cnic_service_stop(void *data)
{
	struct cnic_dev *dev = data;

	cnic_hold(dev);
	mutex_lock(&cnic_lock);
	clear_bit(CNIC_F_IF_UP, &dev->flags);
	cnic_ulp_stop(dev);
	cnic_stop_hw(dev);
	mutex_unlock(&cnic_lock);
	cnic_put(dev);
}

static void cnic_ulp_init(struct cnic_dev *dev)
{
	int i;
	struct cnic_local *cp = dev->cnic_priv;

	rcu_read_lock();
	for (i = 0; i < MAX_CNIC_ULP_TYPE_EXT; i++) {
		struct cnic_ulp_ops *ulp_ops;

		ulp_ops = rcu_dereference(cnic_ulp_tbl[i]);
		if (!ulp_ops || !try_module_get(ulp_ops->owner))
			continue;

		if (!test_and_set_bit(ULP_F_INIT, &cp->ulp_flags[i]))
			ulp_ops->cnic_init(dev);

		module_put(ulp_ops->owner);
	}
	rcu_read_unlock();
}

static void cnic_ulp_exit(struct cnic_dev *dev)
{
	int i;
	struct cnic_local *cp = dev->cnic_priv;

	rcu_read_lock();
	for (i = 0; i < MAX_CNIC_ULP_TYPE_EXT; i++) {
		struct cnic_ulp_ops *ulp_ops;

		ulp_ops = rcu_dereference(cnic_ulp_tbl[i]);
		if (!ulp_ops || !try_module_get(ulp_ops->owner))
			continue;

		if (test_and_clear_bit(ULP_F_INIT, &cp->ulp_flags[i]))
			ulp_ops->cnic_exit(dev);

		module_put(ulp_ops->owner);
	}
	rcu_read_unlock();
}

static int cnic_queue_work(struct cnic_local *cp, u32 work_type, void *data)
{
	struct cnic_work_node *node;
	int bytes = sizeof(u32 *);

	spin_lock_bh(&cp->wr_lock);

	node = &cp->cnic_work_ring[cp->cnic_wr_prod];
	node->work_type = work_type;
	if (work_type == WORK_TYPE_KCQE)
		bytes = sizeof(struct kcqe);
	if (work_type == WORK_TYPE_REDIRECT)
		bytes = sizeof(struct cnic_redirect_entry);
	memcpy(&node->work_data, data, bytes);
	cp->cnic_wr_prod++;
	cp->cnic_wr_prod &= WORK_RING_SIZE_MASK;

	spin_unlock_bh(&cp->wr_lock);
	return 0;
}

static int cnic_cm_offload_pg(struct cnic_sock *csk)
{
	struct cnic_dev *dev = csk->dev;
	struct l4_kwq_offload_pg *l4kwqe;
	struct kwqe *wqes[1];
#ifdef __VMKLNX__
	struct cnic_local *cp = dev->cnic_priv;
	struct iscsiVmknicInfo *nic_info = &cp->vmknic_info;
#else
	struct neighbour *neigh = csk->dst->neighbour;
	struct net_device *netdev = neigh->dev;
#endif

	l4kwqe = (struct l4_kwq_offload_pg *) &csk->kwqe1;
	memset(l4kwqe, 0, sizeof(*l4kwqe));
	wqes[0] = (struct kwqe *) l4kwqe;

	l4kwqe->op_code = L4_KWQE_OPCODE_VALUE_OFFLOAD_PG;
	l4kwqe->flags =
		L4_LAYER_CODE << L4_KWQ_OFFLOAD_PG_LAYER_CODE_SHIFT;
	l4kwqe->l2hdr_nbytes = ETH_HLEN;
#ifdef __VMKLNX__
	l4kwqe->da0 = csk->ha[0];
	l4kwqe->da1 = csk->ha[1];
	l4kwqe->da2 = csk->ha[2];
	l4kwqe->da3 = csk->ha[3];
	l4kwqe->da4 = csk->ha[4];
	l4kwqe->da5 = csk->ha[5];

	l4kwqe->sa0 = nic_info->srcMACAddr[0];
	l4kwqe->sa1 = nic_info->srcMACAddr[1];
	l4kwqe->sa2 = nic_info->srcMACAddr[2];
	l4kwqe->sa3 = nic_info->srcMACAddr[3];
	l4kwqe->sa4 = nic_info->srcMACAddr[4];
	l4kwqe->sa5 = nic_info->srcMACAddr[5];
#else
	l4kwqe->da0 = neigh->ha[0];
	l4kwqe->da1 = neigh->ha[1];
	l4kwqe->da2 = neigh->ha[2];
	l4kwqe->da3 = neigh->ha[3];
	l4kwqe->da4 = neigh->ha[4];
	l4kwqe->da5 = neigh->ha[5];

	l4kwqe->sa0 = netdev->dev_addr[0];
	l4kwqe->sa1 = netdev->dev_addr[1];
	l4kwqe->sa2 = netdev->dev_addr[2];
	l4kwqe->sa3 = netdev->dev_addr[3];
	l4kwqe->sa4 = netdev->dev_addr[4];
	l4kwqe->sa5 = netdev->dev_addr[5];
#endif

	l4kwqe->etype = ETH_P_IP;
	l4kwqe->ipid_count = DEF_IPID_COUNT;
	l4kwqe->host_opaque = csk->l5_cid;

	if (csk->vlan_id) {
		l4kwqe->pg_flags |= L4_KWQ_OFFLOAD_PG_VLAN_TAGGING;
		l4kwqe->vlan_tag = csk->vlan_id;
		l4kwqe->l2hdr_nbytes += 4;
	}

	return (dev->submit_kwqes(dev, wqes, 1));
}

#ifndef __VMKLNX__
static int cnic_cm_update_pg(struct cnic_sock *csk)
{
	struct cnic_dev *dev = csk->dev;
	struct l4_kwq_update_pg *l4kwqe;
	struct kwqe *wqes[1];
	struct neighbour *neigh = csk->dst->neighbour;

	l4kwqe = (struct l4_kwq_update_pg *) &csk->kwqe1;
	memset(l4kwqe, 0, sizeof(*l4kwqe));
	wqes[0] = (struct kwqe *) l4kwqe;

	l4kwqe->opcode = L4_KWQE_OPCODE_VALUE_UPDATE_PG;
	l4kwqe->flags =
		L4_LAYER_CODE << L4_KWQ_UPDATE_PG_LAYER_CODE_SHIFT;
	l4kwqe->pg_cid = csk->pg_cid;
	l4kwqe->da0 = neigh->ha[0];
	l4kwqe->da1 = neigh->ha[1];
	l4kwqe->da2 = neigh->ha[2];
	l4kwqe->da3 = neigh->ha[3];
	l4kwqe->da4 = neigh->ha[4];
	l4kwqe->da5 = neigh->ha[5];

	l4kwqe->pg_host_opaque = 0;
	l4kwqe->pg_valids = L4_KWQ_UPDATE_PG_VALIDS_DA;

	return (dev->submit_kwqes(dev, wqes, 1));
}
#endif

static int cnic_cm_upload_pg(struct cnic_sock *csk)
{
	struct cnic_dev *dev = csk->dev;
	struct l4_kwq_upload *l4kwqe;
	struct kwqe *wqes[1];

	l4kwqe = (struct l4_kwq_upload *) &csk->kwqe1;
	memset(l4kwqe, 0, sizeof(*l4kwqe));
	wqes[0] = (struct kwqe *) l4kwqe;

	l4kwqe->opcode = L4_KWQE_OPCODE_VALUE_UPLOAD_PG;
	l4kwqe->flags =
		L4_LAYER_CODE << L4_KWQ_UPLOAD_LAYER_CODE_SHIFT;
	l4kwqe->cid = csk->pg_cid;

	return (dev->submit_kwqes(dev, wqes, 1));
}

static void cnic_redirect(struct cnic_local *cp, struct dst_entry *new,
			  struct dst_entry *old)
{
	int i, found = 0;

	spin_lock_bh(&cp->cm_lock);
	for (i = 0; i < MAX_CM_SK_TBL_SZ; i++) {
		struct cnic_sock *csk;

		csk = &cp->csk_tbl[i];
		if (test_bit(SK_F_INUSE, &csk->flags) && csk->dst == old)
			found = 1;
	}
	spin_unlock_bh(&cp->cm_lock);

	if (found) {
		struct cnic_redirect_entry cnic_redir;

		dst_hold(new);
		cnic_redir.old_dst = old;
		cnic_redir.new_dst = new;
		cnic_queue_work(cp, WORK_TYPE_REDIRECT, &cnic_redir);
		tasklet_schedule(&cp->cnic_task);
	}
}

static void cnic_update_neigh(struct cnic_local *cp, struct neighbour *neigh)
{
	int i, found = 0;

	spin_lock_bh(&cp->cm_lock);
	for (i = 0; i < MAX_CM_SK_TBL_SZ; i++) {
		struct cnic_sock *csk;

		csk = &cp->csk_tbl[i];
		if (test_bit(SK_F_INUSE, &csk->flags) && csk->dst) {
			if (csk->dst->neighbour == neigh) {
				found = 1;
				break;
			}
		}
	}
	spin_unlock_bh(&cp->cm_lock);

	if (!found)
		return;

	neigh_hold(neigh);

	cnic_queue_work(cp, WORK_TYPE_NEIGH_UPDATE, &neigh);
	tasklet_schedule(&cp->cnic_task);
}

static int cnic_net_callback(struct notifier_block *this, unsigned long event,
	void *ptr)
{
	struct cnic_local *cp = container_of(this, struct cnic_local, cm_nb);

	if (event == NETEVENT_NEIGH_UPDATE) {
		struct neighbour *neigh = ptr;

		cnic_update_neigh(cp, neigh);

	} else if (event == NETEVENT_REDIRECT) {
		struct netevent_redirect *netevent = ptr;
		struct dst_entry *old_dst = netevent->old;
		struct dst_entry *new_dst = netevent->new;

		cnic_redirect(cp, new_dst, old_dst);
	}
	return 0;
}

static int cnic_ok_to_connect(struct cnic_sock *csk)
{
	if (test_bit(SK_F_INUSE, &csk->flags) &&
	    !test_bit(SK_F_OFFLD_PENDING, &csk->flags) &&
	    !test_bit(SK_F_OFFLD_COMPLETE, &csk->flags) &&
	    test_bit(SK_F_PG_OFFLD_COMPLETE, &csk->flags))
		return 1;
	return 0;
}

static int cnic_cm_conn_req(struct cnic_sock *csk)
{
	struct cnic_dev *dev = csk->dev;
	struct l4_kwq_connect_req1 *l4kwqe1;
	struct l4_kwq_connect_req2 *l4kwqe2;
	struct l4_kwq_connect_req3 *l4kwqe3;
	struct kwqe *wqes[3];
	u8 tcp_flags = 0;
	int num_wqes = 2;

	l4kwqe1 = (struct l4_kwq_connect_req1 *) &csk->kwqe1;
	l4kwqe2 = (struct l4_kwq_connect_req2 *) &csk->kwqe2;
	l4kwqe3 = (struct l4_kwq_connect_req3 *) &csk->kwqe3;
	memset(l4kwqe1, 0, sizeof(*l4kwqe1));
	memset(l4kwqe2, 0, sizeof(*l4kwqe2));
	memset(l4kwqe3, 0, sizeof(*l4kwqe3));

	l4kwqe3->op_code = L4_KWQE_OPCODE_VALUE_CONNECT3;
	l4kwqe3->flags =
		L4_LAYER_CODE << L4_KWQ_CONNECT_REQ3_LAYER_CODE_SHIFT;
	l4kwqe3->ka_timeout = csk->ka_timeout;
	l4kwqe3->ka_interval = csk->ka_interval;
	l4kwqe3->ka_max_probe_count = csk->ka_max_probe_count;
	l4kwqe3->tos = csk->tos;
	l4kwqe3->ttl = csk->ttl;
	l4kwqe3->snd_seq_scale = csk->snd_seq_scale;
#ifdef __VMKLNX__
	l4kwqe3->pmtu = 1500;
#else
	l4kwqe3->pmtu = dst_mtu(csk->dst);
#endif
	l4kwqe3->rcv_buf = csk->rcv_buf;
	l4kwqe3->snd_buf = csk->snd_buf;
	l4kwqe3->seed = csk->seed;

	wqes[0] = (struct kwqe *) l4kwqe1;
	if (test_bit(SK_F_IPV6, &csk->flags)) {
		wqes[1] = (struct kwqe *) l4kwqe2;
		wqes[2] = (struct kwqe *) l4kwqe3;
		num_wqes = 3;

		l4kwqe1->conn_flags = L4_KWQ_CONNECT_REQ1_IP_V6;
		l4kwqe2->op_code = L4_KWQE_OPCODE_VALUE_CONNECT2;
		l4kwqe2->flags =
			L4_KWQ_CONNECT_REQ2_LINKED_WITH_NEXT |
			L4_LAYER_CODE << L4_KWQ_CONNECT_REQ2_LAYER_CODE_SHIFT;
		l4kwqe2->src_ip_v6_2 = be32_to_cpu(csk->src_ip[1]);
		l4kwqe2->src_ip_v6_3 = be32_to_cpu(csk->src_ip[2]);
		l4kwqe2->src_ip_v6_4 = be32_to_cpu(csk->src_ip[3]);
		l4kwqe2->dst_ip_v6_2 = be32_to_cpu(csk->dst_ip[1]);
		l4kwqe2->dst_ip_v6_3 = be32_to_cpu(csk->dst_ip[2]);
		l4kwqe2->dst_ip_v6_4 = be32_to_cpu(csk->dst_ip[3]);
		l4kwqe3->mss = l4kwqe3->pmtu - sizeof(struct ipv6hdr) -
			       sizeof(struct tcphdr);
	} else {
		wqes[1] = (struct kwqe *) l4kwqe3;
		l4kwqe3->mss = l4kwqe3->pmtu - sizeof(struct iphdr) -
			       sizeof(struct tcphdr);
	}

	l4kwqe1->op_code = L4_KWQE_OPCODE_VALUE_CONNECT1;
	l4kwqe1->flags =
		(L4_LAYER_CODE << L4_KWQ_CONNECT_REQ1_LAYER_CODE_SHIFT) |
		 L4_KWQ_CONNECT_REQ3_LINKED_WITH_NEXT;
	l4kwqe1->cid = csk->cid;
	l4kwqe1->pg_cid = csk->pg_cid;
	l4kwqe1->src_ip = be32_to_cpu(csk->src_ip[0]);
	l4kwqe1->dst_ip = be32_to_cpu(csk->dst_ip[0]);
	l4kwqe1->src_port = be16_to_cpu(csk->src_port);
	l4kwqe1->dst_port = be16_to_cpu(csk->dst_port);
	if (test_bit(SK_TCP_NO_DELAY_ACK, &csk->tcp_flags))
		tcp_flags |= L4_KWQ_CONNECT_REQ1_NO_DELAY_ACK;
	if (test_bit(SK_TCP_KEEP_ALIVE, &csk->tcp_flags))
		tcp_flags |= L4_KWQ_CONNECT_REQ1_KEEP_ALIVE;
	if (test_bit(SK_TCP_NAGLE, &csk->tcp_flags))
		tcp_flags |= L4_KWQ_CONNECT_REQ1_NAGLE_ENABLE;
	if (test_bit(SK_TCP_TIMESTAMP, &csk->tcp_flags))
		tcp_flags |= L4_KWQ_CONNECT_REQ1_TIME_STAMP;
	if (test_bit(SK_TCP_SACK, &csk->tcp_flags))
		tcp_flags |= L4_KWQ_CONNECT_REQ1_SACK;
	if (test_bit(SK_TCP_SEG_SCALING, &csk->tcp_flags))
		tcp_flags |= L4_KWQ_CONNECT_REQ1_SEG_SCALING;
	
	l4kwqe1->tcp_flags = tcp_flags;

	return (dev->submit_kwqes(dev, wqes, num_wqes));
}

static int cnic_cm_close_req(struct cnic_sock *csk)
{
	struct cnic_dev *dev = csk->dev;
	struct l4_kwq_close_req *l4kwqe;
	struct kwqe *wqes[1];

	l4kwqe = (struct l4_kwq_close_req *) &csk->kwqe2;
	memset(l4kwqe, 0, sizeof(*l4kwqe));
	wqes[0] = (struct kwqe *) l4kwqe;

	l4kwqe->op_code = L4_KWQE_OPCODE_VALUE_CLOSE;
	l4kwqe->flags = L4_LAYER_CODE << L4_KWQ_CLOSE_REQ_LAYER_CODE_SHIFT;
	l4kwqe->cid = csk->cid;

	return (dev->submit_kwqes(dev, wqes, 1));
}

static int cnic_cm_abort_req(struct cnic_sock *csk)
{
	struct cnic_dev *dev = csk->dev;
	struct l4_kwq_reset_req *l4kwqe;
	struct kwqe *wqes[1];

	l4kwqe = (struct l4_kwq_reset_req *) &csk->kwqe2;
	memset(l4kwqe, 0, sizeof(*l4kwqe));
	wqes[0] = (struct kwqe *) l4kwqe;

	l4kwqe->op_code = L4_KWQE_OPCODE_VALUE_RESET;
	l4kwqe->flags = L4_LAYER_CODE << L4_KWQ_RESET_REQ_LAYER_CODE_SHIFT;
	l4kwqe->cid = csk->cid;

	return (dev->submit_kwqes(dev, wqes, 1));
}

static int cnic_cm_create(struct cnic_dev *dev, int ulp_type, u32 cid,
			  u32 l5_cid, struct cnic_sock **csk, void *context)
{
	struct cnic_local *cp = dev->cnic_priv;
	struct cnic_sock *csk1;

	if (l5_cid >= MAX_CM_SK_TBL_SZ)
		return -EINVAL;
		
	csk1 = &cp->csk_tbl[l5_cid];
	if (test_and_set_bit(SK_F_INUSE, &csk1->flags))
		return -EINVAL;

	csk1->dev = dev;
	csk1->cid = cid;
	csk1->l5_cid = l5_cid;
	csk1->ulp_type = ulp_type;
	csk1->context = context;

	csk1->ka_timeout = DEF_KA_TIMEOUT;
	csk1->ka_interval = DEF_KA_INTERVAL;
	csk1->ka_max_probe_count = DEF_KA_MAX_PROBE_COUNT;
	csk1->tos = DEF_TOS;
	csk1->ttl = DEF_TTL;
	csk1->snd_seq_scale = DEF_SND_SEQ_SCALE;
	csk1->rcv_buf = DEF_RCV_BUF;
	csk1->snd_buf = DEF_SND_BUF;
	csk1->seed = DEF_SEED;

	*csk = csk1;

	return 0;
}

static int cnic_cm_destroy(struct cnic_sock *csk)
{
	struct cnic_local *cp = csk->dev->cnic_priv;

	if (!test_bit(SK_F_INUSE, &csk->flags))
		return -EINVAL;

	spin_lock_bh(&cp->cm_lock);
#ifndef __VMKLNX__
	if (csk->dst) {
		if (csk->dst->neighbour)
			neigh_release(csk->dst->neighbour);
		dst_release(csk->dst);
		csk->dst = NULL;
	}
#endif
	csk->flags = 0;
	spin_unlock_bh(&cp->cm_lock);
	return 0;
}

static inline u16 cnic_get_vlan(struct net_device *dev,
				struct net_device **vlan_dev)
{
#ifndef __VMKLNX__
	if (dev->priv_flags & IFF_802_1Q_VLAN) {
#ifdef VLAN_DEV_INFO
		*vlan_dev = VLAN_DEV_INFO(dev)->real_dev;
		return VLAN_DEV_INFO(dev)->vlan_id;
#else
		*vlan_dev = vlan_dev_info(dev)->real_dev;
		return vlan_dev_info(dev)->vlan_id;
#endif
	}
#endif
	*vlan_dev = dev;
	return 0;
}

#ifndef __VMKLNX__
static int cnic_get_v4_route(struct sockaddr_in *dst_addr,
			     struct sockaddr_in *src_addr,
			     struct dst_entry **dst)
{
	struct flowi fl;
	int err;
	struct rtable *rt;

	memset(&fl, 0, sizeof(fl));
	fl.nl_u.ip4_u.daddr = dst_addr->sin_addr.s_addr;
	if (src_addr)
		fl.nl_u.ip4_u.saddr = src_addr->sin_addr.s_addr;

#if (LINUX_VERSION_CODE >= 0x020619)
	err = ip_route_output_key(&init_net, &rt, &fl);
#else
	err = ip_route_output_key(&rt, &fl);
#endif
	if (!err)
		*dst = &rt->u.dst;
	return err;
}

static int cnic_get_v6_route(struct sockaddr_in6 *dst_addr,
			     struct sockaddr_in6 *src_addr,
			     struct dst_entry **dst)
{
#if defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE)
	struct flowi fl;

	memset(&fl, 0, sizeof(fl));
	ipv6_addr_copy(&fl.fl6_dst, &dst_addr->sin6_addr);
	if (ipv6_addr_type(&fl.fl6_dst) & IPV6_ADDR_LINKLOCAL)
		fl.oif = dst_addr->sin6_scope_id;

	if (src_addr)
		ipv6_addr_copy(&fl.fl6_src, &src_addr->sin6_addr);

	*dst = ip6_route_output(NULL, &fl);
	if (*dst)
		return 0;
#endif

	return -ENETUNREACH;
}
#endif /*__VMKLNX__*/

#ifdef __VMKLNX__
static struct cnic_dev *cnic_cm_select_dev(uint8_t *vmknic,
					   struct sockaddr_in *dst_addr,
					   int ulp_type)
{
	struct iscsiVmknicInfo nic_info;
	struct cnic_dev *dev;
	int err, found = 0;

	nic_info.vmknic = vmknic;
	nic_info.srcIPAddr.su_sin.sin_family = AF_INET;
	err = vmk_iscsi_vmknic_lookup(&nic_info);
	if (err)
		return NULL;

#ifndef __VMKLNX__
	read_lock(&cnic_dev_lock);
#endif
	list_for_each_entry(dev, &cnic_dev_list, list) {
		struct cnic_local *cp = dev->cnic_priv;

		if (cp->vmknic_valid && !strcmp(cp->vmknic, vmknic)) {
			found = 1;
			break;
		}
		if (dev->netdev) {
			if (dev->pcidev->bus->number == nic_info.pciBus &&
			    dev->pcidev->devfn == nic_info.pciFunc) {

				found = 1;
				strcpy(cp->vmknic, vmknic);
				memcpy(&cp->vmknic_info, &nic_info,
				       sizeof(nic_info));
				cp->vmknic_valid = 1;
				break;
			}
		}
	}
#ifndef __VMKLNX__
	read_unlock(&cnic_dev_lock);
#endif

	if (!found)
		dev = NULL;
	return dev;
}

#else
static struct cnic_dev *cnic_cm_select_dev(struct sockaddr_in *dst_addr,
					   int ulp_type)
{
	struct cnic_dev *dev = NULL;
	struct dst_entry *dst;
	struct net_device *netdev = NULL;
	int err = -ENETUNREACH, found = 0;

	if (dst_addr->sin_family == AF_INET)
		err = cnic_get_v4_route(dst_addr, NULL, &dst);
	else if (dst_addr->sin_family == AF_INET6) {
		struct sockaddr_in6 *dst_addr6 =
			(struct sockaddr_in6 *) dst_addr;

		err = cnic_get_v6_route(dst_addr6, NULL, &dst);
	} else
		return NULL;

	if (err)
		return NULL;

	if (!dst->dev)
		goto done;

	cnic_get_vlan(dst->dev, &netdev);

#ifndef __VMKLNX__
	read_lock(&cnic_dev_lock);
#endif
	list_for_each_entry(dev, &cnic_dev_list, list) {
		if (netdev == dev->netdev) {
			found = 1;
			break;
		}
	}
#ifndef __VMKLNX__
	read_unlock(&cnic_dev_lock);
#endif

done:
	dst_release(dst);
	if (!found)
		dev = NULL;
	return dev;
}
#endif /*__VMKLNX__*/

#ifndef __VMKLNX__
static int cnic_resolve_addr(struct cnic_sock *csk)
{
	struct neighbour *neigh = csk->dst->neighbour;
	int err = 0;

	if (neigh->nud_state & NUD_VALID) {
		set_bit(SK_F_PG_OFFLD_PENDING, &csk->flags);
		err = cnic_cm_offload_pg(csk);
		goto done;
	}
	neigh_event_send(neigh, NULL);
done:
	return err;
}

static int cnic_get_route(struct cnic_sock *csk, struct cnic_sockaddr *saddr)
{
	struct cnic_dev *dev = csk->dev;
	int is_v6, err;
	struct dst_entry *dst;
	struct net_device *realdev;

	if (saddr->local.v6.sin6_family == AF_INET6 &&
	    saddr->remote.v6.sin6_family == AF_INET6)
		is_v6 = 1;
	else if (saddr->local.v4.sin_family == AF_INET &&
		 saddr->remote.v4.sin_family == AF_INET)
		is_v6 = 0;
	else
		return -EINVAL;

	clear_bit(SK_F_IPV6, &csk->flags);

	if (is_v6) {
		set_bit(SK_F_IPV6, &csk->flags);
		err = cnic_get_v6_route(&saddr->remote.v6,
					&saddr->local.v6, &dst);
		if (err)
			return err;

		if (!dst || dst->error || !dst->dev)
			goto err_out;

		ipv6_get_saddr(dst, &saddr->remote.v6.sin6_addr,
			       &saddr->local.v6.sin6_addr);

		memcpy(&csk->src_ip[0], &saddr->local.v6.sin6_addr,
		       sizeof(struct in6_addr));
		memcpy(&csk->dst_ip[0], &saddr->remote.v6.sin6_addr,
		       sizeof(struct in6_addr));
		csk->src_port = saddr->local.v6.sin6_port;
		csk->dst_port = saddr->remote.v6.sin6_port;

	} else {
		err = cnic_get_v4_route(&saddr->remote.v4, &saddr->local.v4,
					&dst);
		if (err)
			return err;

		if (!dst || dst->error || !dst->dev)
			goto err_out;

		csk->dst_ip[0] = saddr->remote.v4.sin_addr.s_addr;
		csk->src_ip[0] = saddr->local.v4.sin_addr.s_addr;
		csk->src_port = saddr->local.v4.sin_port;
		csk->dst_port = saddr->remote.v4.sin_port;

		if (csk->src_ip[0] == 0) {
			csk->src_ip[0] =
				inet_select_addr(dst->dev, csk->dst_ip[0],
						 RT_SCOPE_LINK);
		}
	}

	csk->vlan_id = cnic_get_vlan(dst->dev, &realdev);
	if (realdev != dev->netdev)
		goto err_out;

	csk->dst = dst;
	return 0;

err_out:
	dst_release(dst);
	return -ENETUNREACH;
}
#else

static int cnic_get_route(struct cnic_sock *csk, struct cnic_sockaddr *saddr)
{
	struct cnic_dev *dev = csk->dev;
	struct cnic_local *cp = dev->cnic_priv;
	struct iscsiVmknicInfo *nic_info = &cp->vmknic_info;
	struct iscsiArpInfo arp_info;
	int err, retry = 0;

	if (saddr->local.v4.sin_family != AF_INET ||
	    saddr->remote.v4.sin_family != AF_INET6)
		return -EINVAL;

	arp_info.portID = nic_info->portID;
	memcpy(&arp_info.srcIPAddr, &nic_info->srcIPAddr, sizeof(su));
	memcpy(&arp_info.dstIPAddr, &saddr->remote, sizeof(su));
	memcpy(&arp_info.srcMACAddr, nic_info->srcMACAddr, ETH_ADDR_LENGTH);

	while (retry++ < 3) {
		err = vmk_iscsi_arp_lookup(&arp_info);
		if (!err)
			break;
		msleep(1000);
	}

	if (err)
		return err;

	clear_bit(SK_F_IPV6, &csk->flags);

	memcpy(csk->ha, arp_info.dstMACAddr, 6);
	csk->dst_ip[0] = saddr->remote.v4.sin_addr.s_addr;
	csk->src_ip[0] = nic_info->srcIPAddr.su_sin.sin_addr.s_addr;
	csk->src_port = saddr->local.v4.sin_port;
	csk->dst_port = saddr->remote.v4.sin_port;
	csk->vlan_id = 0;

	set_bit(SK_F_PG_OFFLD_PENDING, &csk->flags);
	err = cnic_cm_offload_pg(csk);

	return 0;
}
#endif /*__VMKLNX__*/

static int cnic_cm_connect(struct cnic_sock *csk, struct cnic_sockaddr *saddr)
{
#ifndef __VMKLNX__
	struct neighbour *neigh;
#endif
	int err = 0;

	if (!test_bit(SK_F_INUSE, &csk->flags))
		return -EINVAL;

	err = cnic_get_route(csk, saddr);
	if (err)
		return err;

#ifndef __VMKLNX__
	neigh = csk->dst->neighbour;
	if (!neigh)
		goto err_out;

	neigh_hold(neigh);

	err = cnic_resolve_addr(csk);
	if (!err)
		return 0;

	neigh_release(neigh);

err_out:
	dst_release(csk->dst);
	csk->dst = NULL;
	return err;
#endif
}

static int cnic_cm_abort(struct cnic_sock *csk)
{
	if (!test_bit(SK_F_INUSE, &csk->flags))
		return -EINVAL;

	return (cnic_cm_abort_req(csk));
}

static int cnic_cm_close(struct cnic_sock *csk)
{
	if (!test_bit(SK_F_INUSE, &csk->flags))
		return -EINVAL;

	return (cnic_cm_close_req(csk));
}

static void cnic_cm_upcall(struct cnic_local *cp, struct cnic_sock *csk,
			   u8 opcode)
{
	struct cnic_ulp_ops *ulp_ops;
	int ulp_type = csk->ulp_type;

	rcu_read_lock();
	ulp_ops = rcu_dereference(cp->ulp_ops[ulp_type]);
	if (ulp_ops) {
		if (opcode == L4_KCQE_OPCODE_VALUE_CONNECT_COMPLETE)
			ulp_ops->cm_connect_complete(csk);
		else if (opcode == L4_KCQE_OPCODE_VALUE_CLOSE_COMP)
			ulp_ops->cm_close_complete(csk);
		else if (opcode == L4_KCQE_OPCODE_VALUE_RESET_RECEIVED)
			ulp_ops->cm_remote_abort(csk);
		else if (opcode == L4_KCQE_OPCODE_VALUE_RESET_COMP)
			ulp_ops->cm_abort_complete(csk);
		else if (opcode == L4_KCQE_OPCODE_VALUE_CLOSE_RECEIVED)
			ulp_ops->cm_remote_close(csk);
	}
	rcu_read_unlock();
}

static int cnic_cm_set_pg(struct cnic_sock *csk)
{
#ifndef __VMKLNX__
	struct neighbour *neigh = csk->dst->neighbour;
	int valid = neigh->nud_state & NUD_VALID;

	if (!valid) {
		if (!test_bit(SK_F_PG_OFFLD_PENDING, &csk->flags) &&
		    !test_bit(SK_F_PG_OFFLD_COMPLETE, &csk->flags))
			return -ETIMEDOUT;
	} else if (test_bit(SK_F_PG_OFFLD_COMPLETE, &csk->flags))
		cnic_cm_update_pg(csk);
	else if (!test_and_set_bit(SK_F_PG_OFFLD_PENDING, &csk->flags))
		cnic_cm_offload_pg(csk);
#endif
	return 0;
}

static void cnic_cm_process_neigh(struct cnic_dev *dev, struct neighbour *neigh)
{
#ifndef __VMKLNX__
	struct cnic_local *cp = dev->cnic_priv;
	int i;

	for (i = 0; i < MAX_CM_SK_TBL_SZ; i++) {
		struct cnic_sock *csk;
		int abort = 0;

		csk = &cp->csk_tbl[i];
		spin_lock_bh(&cp->cm_lock);
		if (test_bit(SK_F_INUSE, &csk->flags) && csk->dst &&
		    csk->dst->neighbour == neigh) {
			if (cnic_cm_set_pg(csk))
				abort = 1;
		}
		spin_unlock_bh(&cp->cm_lock);
		if (abort)
			cnic_cm_upcall(cp, csk,
				L4_KCQE_OPCODE_VALUE_CONNECT_COMPLETE);
	}
	neigh_release(neigh);
#endif
}

static void cnic_cm_process_redirect(struct cnic_dev *dev,
				     struct cnic_redirect_entry *redir)
{
#ifndef __VMKLNX__
	struct cnic_local *cp = dev->cnic_priv;
	int i;

	for (i = 0; i < MAX_CM_SK_TBL_SZ; i++) {
		struct cnic_sock *csk;
		int abort = 0;

		spin_lock_bh(&cp->cm_lock);
		csk = &cp->csk_tbl[i];
		if (test_bit(SK_F_INUSE, &csk->flags) &&
		    csk->dst == redir->old_dst) {
			csk->dst = redir->new_dst;
			dst_hold(csk->dst);
			neigh_hold(csk->dst->neighbour);
			if (redir->old_dst->neighbour);
				neigh_release(redir->old_dst->neighbour);
			dst_release(redir->old_dst);
			if (cnic_cm_set_pg(csk))
				abort = 1;
		}
		spin_unlock_bh(&cp->cm_lock);
		if (abort)
			cnic_cm_upcall(cp, csk,
				L4_KCQE_OPCODE_VALUE_CONNECT_COMPLETE);
	}

	dst_release(redir->new_dst);
#endif
}

static void cnic_close_conn(struct cnic_sock *csk)
{
	clear_bit(SK_F_OFFLD_COMPLETE, &csk->flags);
	clear_bit(SK_F_OFFLD_PENDING, &csk->flags);
	cnic_cm_upload_pg(csk);
	clear_bit(SK_F_PG_OFFLD_COMPLETE, &csk->flags);
}

static void cnic_cm_process_offld_pg(struct cnic_dev *dev, struct l4_kcq *kcqe)
{
	struct cnic_local *cp = dev->cnic_priv;
	u32 l5_cid = kcqe->pg_host_opaque;
	struct cnic_sock *csk = &cp->csk_tbl[l5_cid];

	if (!test_bit(SK_F_INUSE, &csk->flags))
		return;

	csk->pg_cid = kcqe->pg_cid;
	set_bit(SK_F_PG_OFFLD_COMPLETE, &csk->flags);
	clear_bit(SK_F_PG_OFFLD_PENDING, &csk->flags);
	if (cnic_ok_to_connect(csk)) {
		set_bit(SK_F_OFFLD_PENDING, &csk->flags);
		cnic_cm_conn_req(csk);
	}
}

static void cnic_cm_process_kcqe(struct cnic_dev *dev, struct kcqe *kcqe)
{
	struct cnic_local *cp = dev->cnic_priv;
	struct l4_kcq *l4kcqe = (struct l4_kcq *) kcqe;
	u8 opcode = l4kcqe->op_code;
	u32 l5_cid;
	struct cnic_sock *csk;

	if (opcode == L4_KCQE_OPCODE_VALUE_OFFLOAD_PG) {
		cnic_cm_process_offld_pg(dev, l4kcqe);
		return;
	}

	l5_cid = l4kcqe->conn_id;
	if (l5_cid >= MAX_CM_SK_TBL_SZ)
		return;

	csk = &cp->csk_tbl[l5_cid];

	if (!test_bit(SK_F_INUSE, &csk->flags))
		return;

	if (opcode == L4_KCQE_OPCODE_VALUE_CONNECT_COMPLETE) {

		clear_bit(SK_F_OFFLD_PENDING, &csk->flags);
		if (l4kcqe->status == 0)
			set_bit(SK_F_OFFLD_COMPLETE, &csk->flags);
		cnic_cm_upcall(cp, csk, opcode);

	} else if (opcode == L4_KCQE_OPCODE_VALUE_CLOSE_COMP ||
		   opcode == L4_KCQE_OPCODE_VALUE_RESET_COMP ||
		   opcode == L4_KCQE_OPCODE_VALUE_RESET_RECEIVED) {

		cnic_close_conn(csk);
		cnic_cm_upcall(cp, csk, opcode);

	} else if (opcode == L4_KCQE_OPCODE_VALUE_CLOSE_RECEIVED) {
		cnic_cm_upcall(cp, csk, opcode);
	}
}

static void cnic_cm_indicate_kcqe(void *data, struct kcqe *kcqe[], u32 num_cqe)
{
	struct cnic_dev *dev = data;
	int i;
	struct cnic_local *cp = dev->cnic_priv;

	for (i = 0; i < num_cqe; i++)
		cnic_queue_work(cp, WORK_TYPE_KCQE, kcqe[i]);

	tasklet_schedule(&cp->cnic_task);
}

static void cnic_cm_indicate_event(void *data, unsigned long event)
{
}

static void cnic_cm_dummy(void *data)
{
}

static struct cnic_ulp_ops cm_ulp_ops = {
	.cnic_start		= cnic_cm_dummy,
	.cnic_stop		= cnic_cm_dummy,
	.indicate_kcqes		= cnic_cm_indicate_kcqe,
	.indicate_netevent	= cnic_cm_indicate_event,
	.indicate_inetevent	= cnic_cm_indicate_event,
};

static void cnic_task(unsigned long data)
{
	struct cnic_local *cp = (struct cnic_local *) data;
	struct cnic_dev *dev = cp->dev;
	u32 cons = cp->cnic_wr_cons;
	u32 prod = cp->cnic_wr_prod;

	while (cons != prod) {
		struct cnic_work_node *node;

		node = &cp->cnic_work_ring[cons];
		if (node->work_type == WORK_TYPE_KCQE)
			cnic_cm_process_kcqe(dev, &node->work_data.kcqe);
		else if (node->work_type == WORK_TYPE_NEIGH_UPDATE)
			cnic_cm_process_neigh(dev, node->work_data.neigh);
		else if (node->work_type == WORK_TYPE_REDIRECT)
			cnic_cm_process_redirect(dev,
				&node->work_data.cnic_redir);
		cons++;
		cons &= WORK_RING_SIZE_MASK;
	}
	cp->cnic_wr_cons = cons;
}

static void cnic_free_dev(struct cnic_dev *dev)
{
	int i = 0;

	while ((atomic_read(&dev->ref_count) != 0) && i < 10) {
		msleep(100);
		i++;
	}
	if (atomic_read(&dev->ref_count) != 0)
		printk(KERN_ERR PFX "%s: Failed waiting for ref count to go"
				    " to zero.\n", dev->netdev->name);

	printk(KERN_INFO PFX "Removed CNIC device: %s\n", dev->netdev->name);
	dev_put(dev->netdev);
	kfree(dev);
}

static void cnic_cm_free_mem(struct cnic_dev *dev)
{
	struct cnic_local *cp = dev->cnic_priv;

	kfree(cp->csk_tbl);
	cp->csk_tbl = NULL;
}

static int cnic_cm_alloc_mem(struct cnic_dev *dev)
{
	struct cnic_local *cp = dev->cnic_priv;

	cp->csk_tbl = kmalloc(sizeof(struct cnic_sock) * MAX_CM_SK_TBL_SZ,
			      GFP_KERNEL);
	if (!cp->csk_tbl)
		return -ENOMEM;
	memset(cp->csk_tbl, 0, sizeof(struct cnic_sock) * MAX_CM_SK_TBL_SZ);
	return 0;
}


static void cnic_cm_stop_bnx2_hw(struct cnic_dev *dev)
{
}

static int cnic_cm_init_bnx2_hw(struct cnic_dev *dev)
{
	u32 seed;

#if (LINUX_VERSION_CODE >= 0x020612)
	get_random_bytes(&seed, 4);
#else
	seed = 0x12345678;
#endif
	cnic_bnx2_ctx_wr(dev, 45, 0, seed);
	return 0;
}

static int cnic_cm_open(struct cnic_dev *dev)
{
	struct cnic_local *cp = dev->cnic_priv;
	int err;

	err = cnic_cm_alloc_mem(dev);
	if (err)
		return err;

	err = cp->start_cm(dev);

	if (err)
		goto err_out;

	spin_lock_init(&cp->cm_lock);
	spin_lock_init(&cp->wr_lock);

	tasklet_init(&cp->cnic_task, &cnic_task, (unsigned long) cp);

	cp->cm_nb.notifier_call = cnic_net_callback;
	register_netevent_notifier(&cp->cm_nb);

	dev->cm_create = cnic_cm_create;
	dev->cm_destroy = cnic_cm_destroy;
	dev->cm_connect = cnic_cm_connect;
	dev->cm_abort = cnic_cm_abort;
	dev->cm_close = cnic_cm_close;
	dev->cm_select_dev = cnic_cm_select_dev;

	cp->ulp_handle[CNIC_ULP_L4] = dev;
	rcu_assign_pointer(cp->ulp_ops[CNIC_ULP_L4], &cm_ulp_ops);
	return 0;

err_out:
	cnic_cm_free_mem(dev);
	return err;
}

static void cnic_cm_cleanup(struct cnic_sock *csk)
{
	clear_bit(SK_F_INUSE, &csk->flags);
#ifndef __VMKLNX__
	if (csk->dst) {
		if (csk->dst->neighbour)
			neigh_release(csk->dst->neighbour);
		dst_release(csk->dst);
		csk->dst = NULL;
	}
#endif
}

static int cnic_cm_shutdown(struct cnic_dev *dev)
{
	struct cnic_local *cp = dev->cnic_priv;
	int i;

	cp->stop_cm(dev);

	unregister_netevent_notifier(&cp->cm_nb);

	tasklet_disable(&cp->cnic_task);

	if (!cp->csk_tbl)
		return 0;

	for (i = 0; i < MAX_CM_SK_TBL_SZ; i++) {
		struct cnic_sock *csk = &cp->csk_tbl[i];

		cnic_cm_cleanup(csk);
	}
	cnic_cm_free_mem(dev);

	return 0;
}

static void cnic_init_context(struct cnic_dev *dev, u32 cid)
{
	u32 cid_addr;
	int i;

	cid_addr = GET_CID_ADDR(cid);

	for (i = 0; i < CTX_SIZE; i += 4)
		cnic_bnx2_ctx_wr(dev, cid_addr, i, 0);
}

static int cnic_setup_5709_context(struct cnic_dev *dev, int valid)
{
	struct cnic_local *cp = dev->cnic_priv;
	int ret = 0, i;
	u32 valid_bit = valid ? BNX2_CTX_HOST_PAGE_TBL_DATA0_VALID : 0;

	if (CHIP_NUM(cp) != CHIP_NUM_5709)
		return 0;

	for (i = 0; i < cp->ctx_blks; i++) {
		int j;
		u32 idx = cp->ctx_arr[i].cid / cp->cids_per_blk;
		u32 val;

		memset(cp->ctx_arr[i].ctx, 0, BCM_PAGE_SIZE);

		CNIC_WR(dev, BNX2_CTX_HOST_PAGE_TBL_DATA0,
			(cp->ctx_arr[i].mapping & 0xffffffff) | valid_bit);
		CNIC_WR(dev, BNX2_CTX_HOST_PAGE_TBL_DATA1,
			(u64) cp->ctx_arr[i].mapping >> 32);
		CNIC_WR(dev, BNX2_CTX_HOST_PAGE_TBL_CTRL, idx |
			BNX2_CTX_HOST_PAGE_TBL_CTRL_WRITE_REQ);
		for (j = 0; j < 10; j++) {

			val = CNIC_RD(dev, BNX2_CTX_HOST_PAGE_TBL_CTRL);
			if (!(val & BNX2_CTX_HOST_PAGE_TBL_CTRL_WRITE_REQ))
				break;
			udelay(5);
		}
		if (val & BNX2_CTX_HOST_PAGE_TBL_CTRL_WRITE_REQ) {
			ret = -EBUSY;
			break;
		}
	}
	return ret;
}

static void cnic_free_irq(struct cnic_dev *dev)
{
	struct cnic_local *cp = dev->cnic_priv;
	struct cnic_eth_dev *ethdev = cp->ethdev;

	if (ethdev->drv_flags & CNIC_DRV_FL_USING_MSIX) {
		cp->disable_int_sync(dev);
		tasklet_disable(&cp->cnic_irq_task);
		free_irq(ethdev->irq_arr[0].vector, dev);
	}
}

static int cnic_init_bnx2_irq(struct cnic_dev *dev)
{
	struct cnic_local *cp = dev->cnic_priv;
	struct cnic_eth_dev *ethdev = cp->ethdev;

	if (ethdev->drv_flags & CNIC_DRV_FL_USING_MSIX) {
		int err, i = 0;
		int sblk_num = cp->status_blk_num;
		u32 base = ((sblk_num - 1) * BNX2_HC_SB_CONFIG_SIZE) +
			   BNX2_HC_SB_CONFIG_1;

		CNIC_WR(dev, base, BNX2_HC_SB_CONFIG_1_ONE_SHOT);

		CNIC_WR(dev, base + BNX2_HC_COMP_PROD_TRIP_OFF, (2 << 16) | 8);
		CNIC_WR(dev, base + BNX2_HC_COM_TICKS_OFF, (64 << 16) | 220);
		CNIC_WR(dev, base + BNX2_HC_CMD_TICKS_OFF, (64 << 16) | 220);

		cp->bnx2_status_blk = cp->status_blk;
		cp->last_status_idx = cp->bnx2_status_blk->status_idx;
		tasklet_init(&cp->cnic_irq_task, &cnic_service_bnx2_msix,
			     (unsigned long) dev);
		err = request_irq(ethdev->irq_arr[0].vector, cnic_irq, 0,
				  "cnic", dev);
		if (err) {
			tasklet_disable(&cp->cnic_irq_task);
			return err;
		}
		while (cp->bnx2_status_blk->status_completion_producer_index &&
		       i < 10) {
			CNIC_WR(dev, BNX2_HC_COALESCE_NOW,
				1 << (11 + sblk_num));
			udelay(10);
			i++;
			barrier();
		}
		if (cp->bnx2_status_blk->status_completion_producer_index) {
			cnic_free_irq(dev);
			goto failed;
		}

	} else {
		struct status_block *sblk = cp->status_blk;
		u32 hc_cmd = CNIC_RD(dev, BNX2_HC_COMMAND);
		int i = 0;

		while (sblk->status_completion_producer_index && i < 10) {
			CNIC_WR(dev, BNX2_HC_COMMAND,
				hc_cmd | BNX2_HC_COMMAND_COAL_NOW_WO_INT);
			udelay(10);
			i++;
			barrier();
		}
		if (sblk->status_completion_producer_index)
			goto failed;

	}
	return 0;

failed:
	printk(KERN_ERR PFX "%s: " "KCQ index not resetting to 0.\n",
	       dev->netdev->name);
	return -EBUSY;
}

static void cnic_enable_bnx2_int(struct cnic_dev *dev)
{
	struct cnic_local *cp = dev->cnic_priv;
	struct cnic_eth_dev *ethdev = cp->ethdev;

	if (!(ethdev->drv_flags & CNIC_DRV_FL_USING_MSIX))
		return;

	CNIC_WR(dev, BNX2_PCICFG_INT_ACK_CMD, BNX2_CNIC_INT_NUM |
		BNX2_PCICFG_INT_ACK_CMD_INDEX_VALID | cp->last_status_idx);
}

static void cnic_disable_bnx2_int_sync(struct cnic_dev *dev)
{
	struct cnic_local *cp = dev->cnic_priv;
	struct cnic_eth_dev *ethdev = cp->ethdev;

	if (!(ethdev->drv_flags & CNIC_DRV_FL_USING_MSIX))
		return;

	CNIC_WR(dev, BNX2_PCICFG_INT_ACK_CMD, BNX2_CNIC_INT_NUM |
		BNX2_PCICFG_INT_ACK_CMD_MASK_INT);
	CNIC_RD(dev, BNX2_PCICFG_INT_ACK_CMD);
	synchronize_irq(ethdev->irq_arr[0].vector);
}

static int cnic_start_bnx2_hw(struct cnic_dev *dev)
{
	struct cnic_local *cp = dev->cnic_priv;
	struct cnic_eth_dev *ethdev = cp->ethdev;
	u32 val;
	int err;

	val = CNIC_RD(dev, BNX2_MQ_CONFIG);
	val &= ~BNX2_MQ_CONFIG_KNL_BYP_BLK_SIZE;
	if (BCM_PAGE_BITS > 12)
		val |= (12 - 8)  << 4;
	else
		val |= (BCM_PAGE_BITS - 8)  << 4;

	CNIC_WR(dev, BNX2_MQ_CONFIG, val);

	CNIC_WR(dev, BNX2_HC_COMP_PROD_TRIP, (2 << 16) | 8);
	CNIC_WR(dev, BNX2_HC_COM_TICKS, (64 << 16) | 220);
	CNIC_WR(dev, BNX2_HC_CMD_TICKS, (64 << 16) | 220);

	err = cnic_setup_5709_context(dev, 1);
	if (err)
		return err;

	cnic_init_context(dev, KWQ_CID);
	cnic_init_context(dev, KCQ_CID);

	cp->kwq_cid_addr = GET_CID_ADDR(KWQ_CID);
	cp->kwq_io_addr = MB_GET_CID_ADDR(KWQ_CID) + L5_KRNLQ_HOST_QIDX;

	cp->max_kwq_idx = MAX_KWQ_IDX;
	cp->kwq_prod_idx = 0;
	cp->kwq_con_idx = 0;
	cp->cnic_local_flags |= CNIC_LCL_FL_KWQ_INIT;
	
	/* Initialize the kernel work queue context. */
	val = KRNLQ_TYPE_TYPE_KRNLQ | KRNLQ_SIZE_TYPE_SIZE |
	      (BCM_PAGE_BITS - 8) | KRNLQ_FLAGS_QE_SELF_SEQ;
	cnic_bnx2_ctx_wr(dev, cp->kwq_cid_addr, L5_KRNLQ_TYPE, val);
	
	val = (BCM_PAGE_SIZE / sizeof(struct kwqe) - 1) << 16;
	cnic_bnx2_ctx_wr(dev, cp->kwq_cid_addr, L5_KRNLQ_QE_SELF_SEQ_MAX, val);
	
	val = ((BCM_PAGE_SIZE / sizeof(struct kwqe)) << 16) | KWQ_PAGE_CNT;
	cnic_bnx2_ctx_wr(dev, cp->kwq_cid_addr, L5_KRNLQ_PGTBL_NPAGES, val);
	
	val = (u32) ((u64) cp->kwq_info.pgtbl_map >> 32);
	cnic_bnx2_ctx_wr(dev, cp->kwq_cid_addr, L5_KRNLQ_PGTBL_HADDR_HI, val);
	
	val = (u32) cp->kwq_info.pgtbl_map;
	cnic_bnx2_ctx_wr(dev, cp->kwq_cid_addr, L5_KRNLQ_PGTBL_HADDR_LO, val);

	cp->kcq_cid_addr = GET_CID_ADDR(KCQ_CID);
	cp->kcq_io_addr = MB_GET_CID_ADDR(KCQ_CID) + L5_KRNLQ_HOST_QIDX;
	
	cp->kcq_prod_idx = 0;
	
	/* Initialize the kernel complete queue context. */
	val = KRNLQ_TYPE_TYPE_KRNLQ | KRNLQ_SIZE_TYPE_SIZE |
	      (BCM_PAGE_BITS - 8) | KRNLQ_FLAGS_QE_SELF_SEQ;
	cnic_bnx2_ctx_wr(dev, cp->kcq_cid_addr, L5_KRNLQ_TYPE, val);
	
	val = (BCM_PAGE_SIZE / sizeof(struct kcqe) - 1) << 16;
	cnic_bnx2_ctx_wr(dev, cp->kcq_cid_addr, L5_KRNLQ_QE_SELF_SEQ_MAX, val);
	
	val = ((BCM_PAGE_SIZE / sizeof(struct kcqe)) << 16) | KCQ_PAGE_CNT;
	cnic_bnx2_ctx_wr(dev, cp->kcq_cid_addr, L5_KRNLQ_PGTBL_NPAGES, val);
	
	val = (u32) ((u64) cp->kcq_info.pgtbl_map >> 32);
	cnic_bnx2_ctx_wr(dev, cp->kcq_cid_addr, L5_KRNLQ_PGTBL_HADDR_HI, val);
	
	val = (u32) cp->kcq_info.pgtbl_map;
	cnic_bnx2_ctx_wr(dev, cp->kcq_cid_addr, L5_KRNLQ_PGTBL_HADDR_LO, val);

	if (ethdev->drv_flags & CNIC_DRV_FL_USING_MSIX) {
		u32 sb_id = cp->status_blk_num;
		u32 sb = KRNLQ_STATUSB_NUM_CNIC_MSIX(sb_id);

		cnic_bnx2_ctx_wr(dev, cp->kwq_cid_addr, L5_KRNLQ_HOST_QIDX, sb);
		cnic_bnx2_ctx_wr(dev, cp->kcq_cid_addr, L5_KRNLQ_HOST_QIDX, sb);
	}

	/* Enable Commnad Scheduler notification when we write to the
	 * host producer index of the kernel contexts. */
	CNIC_WR(dev, BNX2_MQ_KNL_CMD_MASK1, 2);

	/* Enable Command Scheduler notification when we write to either
	 * the Send Queue or Receive Queue producer indexes of the kernel
	 * bypass contexts. */
	CNIC_WR(dev, BNX2_MQ_KNL_BYP_CMD_MASK1, 7);
	CNIC_WR(dev, BNX2_MQ_KNL_BYP_WRITE_MASK1, 7);

	/* Notify COM when the driver post an application buffer. */
	CNIC_WR(dev, BNX2_MQ_KNL_RX_V2P_MASK2, 0x2000);

	/* Set the CP and COM doorbells.  These two processors polls the
	 * doorbell for a non zero value before running.  This must be done
	 * after setting up the kernel queue contexts. */
	cnic_bnx2_reg_wr_ind(dev, BNX2_CP_SCRATCH + 0x20, 1);
	cnic_bnx2_reg_wr_ind(dev, BNX2_COM_SCRATCH + 0x20, 1);

	err = cnic_init_bnx2_irq(dev);
	if (err) {
		printk(KERN_ERR PFX "%s: cnic_init_irq failed\n",
		       dev->netdev->name);
		cnic_bnx2_reg_wr_ind(dev, BNX2_CP_SCRATCH + 0x20, 0);
		cnic_bnx2_reg_wr_ind(dev, BNX2_COM_SCRATCH + 0x20, 0);
		return err;
	}

	return 0;
}

static int cnic_start_hw(struct cnic_dev *dev)
{
	struct cnic_local *cp = dev->cnic_priv;
	struct cnic_eth_dev *ethdev;
	int err;

	if (test_bit(CNIC_F_CNIC_UP, &dev->flags))
		return -EALREADY;

	cp->ethdev = cp->register_cnic(dev->netdev, cp->cnic_ops, dev);
	if (!cp->ethdev) {
		printk(KERN_ERR PFX "%s: register_cnic failed\n",
		       dev->netdev->name);
		return -EBUSY;
	}

	ethdev = cp->ethdev;
	dev->pcidev = ethdev->pdev;
	dev->regview = ethdev->io_base;
	cp->chip_id = ethdev->chip_id;
	pci_dev_get(dev->pcidev);
	cp->func = PCI_FUNC(dev->pcidev->devfn);
	cp->status_blk = ethdev->irq_arr[0].status_blk;
	cp->status_blk_num = ethdev->irq_arr[0].status_blk_num;

	err = cp->alloc_resc(dev);
	if (err) {
		printk(KERN_ERR PFX "%s: allocate resource failure\n",
		       dev->netdev->name);
		goto err1;
	}

	err = cp->start_hw(dev);
	if (err)
		goto err1;

	err = cnic_cm_open(dev);
	if (err)
		goto err1;

	set_bit(CNIC_F_CNIC_UP, &dev->flags);

	cp->enable_int(dev);

	return 0;

err1:
	cp->unregister_cnic(dev->netdev);
	cp->free_resc(dev);
	pci_dev_put(dev->pcidev);
	return err;
}

static void cnic_stop_bnx2_hw(struct cnic_dev *dev)
{
	bnx2_unregister_cnic(dev->netdev);

	cnic_disable_bnx2_int_sync(dev);

	cnic_bnx2_reg_wr_ind(dev, BNX2_CP_SCRATCH + 0x20, 0);
	cnic_bnx2_reg_wr_ind(dev, BNX2_COM_SCRATCH + 0x20, 0);

	cnic_init_context(dev, KWQ_CID);
	cnic_init_context(dev, KCQ_CID);

	cnic_setup_5709_context(dev, 0);
	cnic_free_irq(dev);
	cnic_free_resc(dev);
}

static void cnic_stop_hw(struct cnic_dev *dev)
{
	if (test_bit(CNIC_F_CNIC_UP, &dev->flags)) {
		struct cnic_local *cp = dev->cnic_priv;

		clear_bit(CNIC_F_CNIC_UP, &dev->flags);
		rcu_assign_pointer(cp->ulp_ops[CNIC_ULP_L4], NULL);
#ifndef __VMKLNX__
		synchronize_rcu();
#endif
		cnic_cm_shutdown(dev);
		cp->stop_hw(dev);
		pci_dev_put(dev->pcidev);
	}
}

static struct cnic_dev *alloc_cnic(struct net_device *dev)
{
	struct cnic_dev *cdev;
	struct cnic_local *cp;
	int alloc_size;

	alloc_size = sizeof(struct cnic_dev) + sizeof(struct cnic_local);

	cdev = kmalloc(alloc_size , GFP_KERNEL);
	if (cdev == NULL) {
		printk(KERN_ERR PFX "%s: allocate dev struct failure\n",
		       dev->name);
		return NULL;
	}
	memset(cdev, 0, alloc_size);

	cdev->netdev = dev;
	cdev->cnic_priv = (char *)cdev + sizeof(struct cnic_dev);
	cdev->register_device = cnic_register_device;
	cdev->unregister_device = cnic_unregister_device;
	cp = cdev->cnic_priv;
	cp->dev = cdev;

	spin_lock_init(&cp->cnic_ulp_lock);
	printk(KERN_INFO PFX "Added CNIC device: %s\n", dev->name);

	return cdev;
}

static struct cnic_dev *init_bnx2_cnic(struct net_device *dev)
{
	struct bnx2 *bnx2_bp = netdev_priv(dev);
	struct pci_dev *pdev = bnx2_bp->pdev;
	struct cnic_dev *cdev;
	struct cnic_local *cp;

	if (!pdev)
		return NULL;

	dev_hold(dev);
	pci_dev_get(pdev);
	if (pdev->device == PCI_DEVICE_ID_NX2_5709 ||
	    pdev->device == PCI_DEVICE_ID_NX2_5709S) {
		u8 rev;

		pci_read_config_byte(pdev, PCI_REVISION_ID, &rev);
		if (rev < 0x10) {
			pci_dev_put(pdev);
			dev_put(dev);
			return NULL;
		}
	}
	pci_dev_put(pdev);

	cdev = alloc_cnic(dev);
	if (cdev == NULL) {
		dev_put(dev);
		return NULL;
	}

	set_bit(CNIC_F_BNX2_CLASS, &cdev->flags);
	cdev->submit_kwqes = cnic_submit_bnx2_kwqes;

	cp = cdev->cnic_priv;

	cp->cnic_ops = &cnic_bnx2_ops;
	cp->register_cnic = bnx2_register_cnic;
	cp->unregister_cnic = bnx2_unregister_cnic;
	cp->start_hw = cnic_start_bnx2_hw;
	cp->stop_hw = cnic_stop_bnx2_hw;
	cp->setup_pgtbl = cnic_setup_page_tbl;
	cp->alloc_resc = cnic_alloc_bnx2_resc;
	cp->free_resc = cnic_free_resc;
	cp->start_cm = cnic_cm_init_bnx2_hw;
	cp->stop_cm = cnic_cm_stop_bnx2_hw;
	cp->enable_int = cnic_enable_bnx2_int;
	cp->disable_int_sync = cnic_disable_bnx2_int_sync;
	return cdev;
}

static struct cnic_dev *is_cnic_dev(struct net_device *dev)
{
	struct ethtool_drvinfo drvinfo;
	struct cnic_dev *cdev = NULL;

	if (dev->ethtool_ops && dev->ethtool_ops->get_drvinfo) {
		memset(&drvinfo, 0, sizeof(drvinfo));
		dev->ethtool_ops->get_drvinfo(dev, &drvinfo);
		
		if (!strcmp(drvinfo.driver, "bnx2"))
			cdev = init_bnx2_cnic(dev);
		if (cdev) {
			write_lock(&cnic_dev_lock);
			list_add(&cdev->list, &cnic_dev_list);
			write_unlock(&cnic_dev_lock);
		}
	}
	return cdev;
}

/**
 * IP event handler
 */
#ifndef __VMKLNX__
static int cnic_ip_event(struct notifier_block *this, unsigned long event,
						 void *ptr)
{
	struct in_ifaddr *ifa = (struct in_ifaddr *) ptr;
	struct net_device *netdev = (struct net_device *) ifa->ifa_dev->dev;
	struct cnic_dev *dev;
	int if_type;
	u32 my_dev = 0;

	read_lock(&cnic_dev_lock);
	list_for_each_entry(dev, &cnic_dev_list, list) {
		if (netdev == dev->netdev) {
			my_dev = 1;
			cnic_hold(dev);
			break;
		}
	}
	read_unlock(&cnic_dev_lock);

	if (my_dev) {
		struct cnic_local *cp = dev->cnic_priv;

		rcu_read_lock();
		for (if_type = 0; if_type < MAX_CNIC_ULP_TYPE; if_type++) {
			struct cnic_ulp_ops *ulp_ops;

			ulp_ops = rcu_dereference(cp->ulp_ops[if_type]);
			if (ulp_ops) {
				void *ctx = cp->ulp_handle[if_type];

				ulp_ops->indicate_inetevent(ctx, event);
			}
		}
		rcu_read_unlock();

		cnic_put(dev);
	}

	return NOTIFY_DONE;
}

/**
 * netdev event handler
 */
static int cnic_netdev_event(struct notifier_block *this, unsigned long event,
							 void *ptr)
{
	struct net_device *netdev = ptr;
	struct cnic_dev *dev;
	int if_type;
	u32 my_dev = 0;

	read_lock(&cnic_dev_lock);
	list_for_each_entry(dev, &cnic_dev_list, list) {
		if (netdev == dev->netdev) {
			my_dev = 1;
			cnic_hold(dev);
			break;
		}
	}
	read_unlock(&cnic_dev_lock);

	if (!my_dev && event == NETDEV_REGISTER) {
		/* Check for the hot-plug device */
		dev = is_cnic_dev(netdev);
		if (dev) {
			my_dev = 1;
			cnic_hold(dev);
		}
	}
	if (my_dev) {
		struct cnic_local *cp = dev->cnic_priv;

		if (event == NETDEV_REGISTER)
			cnic_ulp_init(dev);
		else if (event == NETDEV_UNREGISTER)
			cnic_ulp_exit(dev);
		else if (event == NETDEV_UP) {
			mutex_lock(&cnic_lock);
			set_bit(CNIC_F_IF_UP, &dev->flags);
			if (dev->use_count) {
				if (!cnic_start_hw(dev))
					cnic_ulp_start(dev);
			}
			mutex_unlock(&cnic_lock);
		}

		rcu_read_lock();
		for (if_type = 0; if_type < MAX_CNIC_ULP_TYPE; if_type++) {
			struct cnic_ulp_ops *ulp_ops;
			void *ctx;

			ulp_ops = rcu_dereference(cp->ulp_ops[if_type]);
			if (!ulp_ops)
				continue;

			ctx = cp->ulp_handle[if_type];

			ulp_ops->indicate_netevent(ctx, event);
		}
		rcu_read_unlock();

		if (event == NETDEV_GOING_DOWN) {
			mutex_lock(&cnic_lock);
			clear_bit(CNIC_F_IF_UP, &dev->flags);
			set_bit(CNIC_F_IF_GOING_DOWN, &dev->flags);
			cnic_ulp_stop(dev);
			cnic_stop_hw(dev);
			mutex_unlock(&cnic_lock);
		} else if (event == NETDEV_DOWN) {
			mutex_lock(&cnic_lock);
			clear_bit(CNIC_F_IF_GOING_DOWN, &dev->flags);
			mutex_unlock(&cnic_lock);
		} else if (event == NETDEV_UNREGISTER) {
			write_lock(&cnic_dev_lock);
			list_del_init(&dev->list);
			write_unlock(&cnic_dev_lock);

			cnic_put(dev);
			cnic_free_dev(dev);
			goto done;
		}
		cnic_put(dev);
	}
done:
	return NOTIFY_DONE;
}

static struct notifier_block cnic_ip_notifier = {
	cnic_ip_event,
	0
};

static struct notifier_block cnic_netdev_notifier = {
	cnic_netdev_event,
	0
};
#endif

static void cnic_release(void)
{
	struct cnic_dev *dev;

	while (!list_empty(&cnic_dev_list)) {
		dev = list_entry(cnic_dev_list.next, struct cnic_dev, list);
		if (test_bit(CNIC_F_CNIC_UP, &dev->flags))
			cnic_stop_hw(dev);

		list_del_init(&dev->list);
		cnic_free_dev(dev);
	}
}

static int __init cnic_init(void)
{
	int rc = 0;
	struct net_device *dev;

	printk(KERN_INFO "%s", version);

	rtnl_lock();
	/* Find Teton devices */
#if (LINUX_VERSION_CODE >= 0x020618)
	for_each_netdev(&init_net, dev)
#elif (LINUX_VERSION_CODE >= 0x20616)
	for_each_netdev(dev)
#else
	for (dev = dev_base; dev; dev = dev->next)
#endif
		is_cnic_dev(dev);

	rtnl_unlock();

#ifndef __VMKLNX__
	rc = register_inetaddr_notifier(&cnic_ip_notifier);
	if (rc)
		cnic_release();
	rc = register_netdevice_notifier(&cnic_netdev_notifier);
	if (rc) {
		unregister_inetaddr_notifier(&cnic_ip_notifier);
		cnic_release();
	}
#endif
	return rc;
}

static void __exit cnic_exit(void)
{
#ifndef __VMKLNX__
	unregister_inetaddr_notifier(&cnic_ip_notifier);
	unregister_netdevice_notifier(&cnic_netdev_notifier);
#endif
	cnic_release();
	return;
}

module_init(cnic_init);
module_exit(cnic_exit);
