/* 
 * Copyright (c)  2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 * 
 * See LICENSE.bfa for copyright and licensing details.
 */
/** 
 *
 * @brief
 * Contains declarations of OS specific calls required for Brocade FC HBA 
 * HAL implementation. This file has to be provided for each OS instance.
 *
 * Recommendation is to implement all non-performance path functions as 
 * functions.
 *
 * 	TODO: retvals??
 */

/**
 * @file bfa_os.h Linux driver OS specific declarations. 
 */

#ifndef __BFA_OS_H__
#define __BFA_OS_H__

#include <linux/version.h>
#include <linux/pci.h>

#include <linux/dma-mapping.h>

#include <linux/idr.h> // TBD

#include <linux/interrupt.h>
#include <linux/cdev.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <linux/vmalloc.h>

#include <linux/workqueue.h>

#include <scsi/scsi.h>
//#include <scsi/scsi_cmnd.h>
//#include <scsi/scsi_device.h>
#include <scsi/scsi_host.h>

#include <scsi/scsi_tcq.h>
#include <scsi/scsi_transport_fc.h>

#define BFA_ERR			KERN_ERR	
#define BFA_WARNING		KERN_WARNING	
#define BFA_NOTICE		KERN_NOTICE	
#define BFA_INFO		KERN_INFO 
#define BFA_DEBUG		KERN_DEBUG	

#define LOG_BFAD_INIT		0x00000001
#define LOG_FCP_IO		0x00000002

#ifdef DEBUG
#define BFA_LOG_TRACE(bfad, level, mask, fmt, arg...)			\
		BFA_LOG(bfad, level, mask, fmt, ## arg)
#define BFA_DEV_TRACE(bfad, level, fmt, arg...)				\
		BFA_DEV_PRINTF(bfad, level, fmt, ## arg)
#define BFA_TRACE(level, fmt, arg...)					\
		BFA_PRINTF(level, fmt, ## arg)
#else
#define BFA_LOG_TRACE(bfad, level, mask, fmt, arg...)		
#define BFA_DEV_TRACE(bfad, level, fmt, arg...)		
#define BFA_TRACE(level, fmt, arg...)				
#endif

#define BFA_ASSERT(p) do {						\
	if (!(p)) {							\
		printk(KERN_ERR "assert(%s) failed at %s:%d\n",		\
		#p, __FILE__, __LINE__);				\
		BUG();							\
	}								\
} while (0)
			

#define BFA_LOG(bfad, level, mask, fmt, arg...)				\
do { 									\
	if (((mask) & (((bfad_t *)(bfad))->				\
		cfg_data[cfg_log_mask])) || (level[1] <= '3'))		\
		dev_printk(level, &(((bfad_t *)(bfad))			\
			->pcidev->dev), fmt, ##arg);			\
} while (0)

#ifndef BFA_DEV_PRINTF
#define BFA_DEV_PRINTF(bfad, level, fmt, arg...)			\
		dev_printk(level, &(((bfad_t *)(bfad))			\
			->pcidev->dev), fmt, ##arg);			
#endif

#define BFA_PRINTF(level, fmt, arg...)					\
		printk(level fmt, ##arg);				

/*
 * TODO: 
 *	- Need to support selection between tradition interrupt and MSI/MSI-X 
 *	(do we have to support dynamic switching between these modes?)
 * 	- Check silverback APIs
 * 	- How to handle OS multi-processing? Can it be punted to OS Drivers?
 *			Check how silverback handled this?
 *
 * 
 * - TODO - Provide APIs to enumerate children objects.
 * - TODO - Add doxygen tags to embed images
 */

int bfa_os_MWB(void *);

#define bfa_os_mmiowb()		mmiowb()

#define bfa_swap_3b(_x)				\
	((((_x) & 0xff) << 16) |		\
	((_x) & 0x00ff00) |			\
	(((_x) & 0xff0000) >> 16))

#define bfa_swap_8b(_x) 				\
     ((((_x) & 0xff00000000000000ull) >> 56)		\
      | (((_x) & 0x00ff000000000000ull) >> 40)		\
      | (((_x) & 0x0000ff0000000000ull) >> 24)		\
      | (((_x) & 0x000000ff00000000ull) >> 8)		\
      | (((_x) & 0x00000000ff000000ull) << 8)		\
      | (((_x) & 0x0000000000ff0000ull) << 24)		\
      | (((_x) & 0x000000000000ff00ull) << 40)		\
      | (((_x) & 0x00000000000000ffull) << 56))

#define bfa_os_swap32(_x) 			\
	((((_x) & 0xff) << 24) 		|	\
	(((_x) & 0x0000ff00) << 8)	|	\
	(((_x) & 0x00ff0000) >> 8)	|	\
	(((_x) & 0xff000000) >> 24))


#ifndef __BIGENDIAN
#define bfa_os_htons(_x) ((uint16_t)((((_x) & 0xff00) >> 8) | \
				 (((_x) & 0x00ff) << 8)))

#define bfa_os_htonl(_x)	bfa_os_swap32(_x)
#define bfa_os_htonll(_x)	bfa_swap_8b(_x)
#define bfa_os_hton3b(_x)	bfa_swap_3b(_x)		

#define bfa_os_wtole(_x)   (_x)

#else

#define bfa_os_htons(_x)   (_x)
#define bfa_os_htonl(_x)   (_x)
#define bfa_os_hton3b(_x)  (_x)
#define bfa_os_htonll(_x)  (_x)
#define bfa_os_wtole(_x)   bfa_os_swap32(_x)

#endif

#define bfa_os_ntohs(_x)   bfa_os_htons(_x)
#define bfa_os_ntohl(_x)   bfa_os_htonl(_x)
#define bfa_os_ntohll(_x)  bfa_os_htonll(_x)
#define bfa_os_ntoh3b(_x)  bfa_os_hton3b(_x)

#define bfa_os_memset	memset
#define bfa_os_memcpy	memcpy
#define bfa_os_udelay	udelay	
#define bfa_os_vsprintf vsprintf 

typedef void * bfa_os_addr_t;
#define bfa_os_panic()

#define bfa_os_reg_read(_raddr)				\
			bfa_os_wtole(*((volatile uint32_t *)(_raddr)))
#define bfa_os_reg_write(_raddr, _val)		\
			*((volatile uint32_t *)(_raddr)) = bfa_os_wtole((_val))
#define bfa_os_mem_read(_raddr, _off)      \
    		bfa_os_ntohl(((volatile uint32_t *)_raddr)[(_off) >> 2])
#define bfa_os_mem_write(_raddr, _off, _val)   \
            (((volatile uint32_t *)_raddr)[(_off) >> 2]) = bfa_os_htonl(_val)

#define BFA_TRC_TS(_trcm)						\
			({						\
				struct timeval tv;			\
									\
				do_gettimeofday(&tv);			\
				(tv.tv_sec*1000000+tv.tv_usec);		\
			 })

struct bfa_log_mod_s;
void bfa_os_printf(struct bfa_log_mod_s *log_mod, uint32_t msg_id, 
		   const char *fmt, ...);

#endif /* __BFA_OS_H__ */
