/*
 * Brocade Fibre Channel HBA Driver
 * Copyright (c) 2006-2007 Brocade Communications Systems, Inc.
 * All rights reserved.
 */

#include <linux/poll.h>		/* poll_wait() */
#include <linux/spinlock.h>	/* spin_(un)lock_irq**/
#include <asm/uaccess.h>	/* copy_from_user(), copy_to_user() */
#include <linux/fs.h>		/* struct file_operations */
#include <linux/string.h>	/* memset() */

#include "bfad_aen.h"
#include "bfad_ioctl.h"

bfad_t         *bfad_find_bfad_by_inst_no(int inst_no);

static spinlock_t bfad_aen_spinlock = SPIN_LOCK_UNLOCKED;
DECLARE_WAIT_QUEUE_HEAD(bfad_aen_readq);
struct fasync_struct *bfad_aen_asyncq;
int             fasync_enabled;

/**
 * @dg bfad2aen_api Base Linux ==> AEN module interface to support file
 * 			operations
 * @{
 */

/**
 * @brief IOCTL function to copy AEN module data onto user space.
 * @param[in,out] arg   - Ptr to user bfa_ioctl_aen_t where data will be
 * 			copied from/to
 * @retval 0            Success
 * @retval -EIO         Error copying data from user buffer
 * @retval -EINVAL      Invalid instance number passed in user command
 * @retval -ENOMEM      Failed to allocate kernel memory
 * @retval -EFAULT      Error copying data onto user buffer
 */
int32_t
bfad_ioctl_get_aen(unsigned long arg)
{
	bfa_ioctl_aen_t *ioc_aen;
	uint32_t        size;
	unsigned long   rc;
	unsigned long   flags;
	bfad_t         *bfad;
	bfa_ioctl_aen_t iocmd;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg, sizeof(bfa_ioctl_aen_t)))
		return -EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	/*
	 * Allocate memory locally
	 */
	if (iocmd.aen_count > 0)
		size = sizeof(bfa_ioctl_aen_t) +
			((iocmd.aen_count - 1) * sizeof(bfa_aen_entry_t));
	else
		size = sizeof(bfa_ioctl_aen_t);

	ioc_aen = kzalloc(size, GFP_KERNEL);
	if (!ioc_aen)
		return (-ENOMEM);
	/*
	 * Get events from AEN module
	 */
	spin_lock_irqsave(&bfad_aen_spinlock, flags);
	ioc_aen->aen_count =
		bfa_aen_fetch(bfad->aen, ioc_aen->aen_list, iocmd.aen_count);
	spin_unlock_irqrestore(&bfad_aen_spinlock, flags);

	/*
	 * Copy events to user buffer
	 */
	if (ioc_aen->aen_count > 0) {
		ioc_aen->status = BFA_STATUS_OK;
		size = sizeof(bfa_ioctl_aen_t) +
			((ioc_aen->aen_count - 1) * sizeof(bfa_aen_entry_t));
	} else {
		ioc_aen->status = BFA_STATUS_FAILED;
		size = sizeof(bfa_ioctl_aen_t);
	}

	rc = copy_to_user((void __user *)arg, (void *)ioc_aen, size);
	if (rc) {
		/*
		 * Should we log about lost events?
		 * But what about already copied data?
		 */
		kfree(ioc_aen);
		return (-EFAULT);
	}

	kfree(ioc_aen);
	return 0;
}

/**
 * @brief Read function for AEN module file operation through Base Linux driver.
 * @param[in] file  - File ptr
 * @param[in] buf   - Ptr to user space buffer (bfa_aen_entry_t[]) where data
 * 			will be copied to
 * @param[in] count - Amount of data to be read in bytes
 * @param[in] ppos  - Ptr to file position, unused
 * @return Number of bytes copied onto user buffer
 * @retval -EFAULT  Error copying data onto user buffer
 * @retval -EAGAIN  No data available to read, call again later
 */
ssize_t
bfad_aen_read(struct file *file, char __user *buf, size_t buf_size,
	      loff_t *ppos)
{
	bfad_t         *bfad;
	int32_t         entries_read = 0;	/* Total entries read so far */
	int32_t         entries_reqd = 0;	/* How many more entries to be
						 *read */
	bfa_aen_entry_t *aen_entry;
	int32_t         rc = 0;
	int32_t         aen_entry_size = sizeof(bfa_aen_entry_t);
	unsigned long   flags;

	aen_entry = kzalloc(buf_size, GFP_KERNEL);
	if (!aen_entry)
		return (-ENOMEM);

	entries_reqd = (buf_size / aen_entry_size);
#if 0
	DECLARE_WAITQUEUE(wait, current);

	/*
	 * Take care of non blocking read operation
	 */
	add_wait_queue(&bfad_aen_readq, &wait);
	do {
		__set_current_state(TASK_INTERRUPTIBLE);
		/*
		 * TODO for each bfad in the list
		 */
		if (bfa_aen_fetch_count(/* bfad->aen */) > 0) {
			if (file->f_flags & O_NONBLOCK) {
				kfree(aen_entry);
				return (-EAGAIN);
			}
			if (signal_pending(current)) {
				kfree(aen_entry);
				return (-ERESTARTSYS);
			}
			schedule();
		}
	} while (1);
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&bfad_aen_readq, &wait);
#endif

	spin_lock_irqsave(&bfad_aen_spinlock, flags);
	list_for_each_entry(bfad, &bfad_list, list_entry) {
		while (entries_read < entries_reqd) {
			rc = bfa_aen_fetch(bfad->aen, aen_entry + entries_read,
					   entries_reqd);
			if (rc <= 0)
				break;	/* No more events */
			entries_reqd -= rc;
			entries_read += rc;
		}
	}
	spin_unlock_irqrestore(&bfad_aen_spinlock, flags);

	if (!entries_read) {
		kfree(aen_entry);
		return (-EAGAIN);
	}

	rc = copy_to_user((void __user *)(buf), (void *)aen_entry,
			  (entries_read * aen_entry_size));
	kfree(aen_entry);
	if (rc) {
		/*
		 * Should we log about lost events?
		 */
		/*
		 * Could we copy anything at all?
		 */
		if ((entries_read * aen_entry_size) == rc)
			return (-EFAULT);
		return (rc);	/* This may include partial entries!! */
	}

	return (entries_read * aen_entry_size);
}

/**
 * @brief Poll function for AEN module file operation through Base Linux driver.
 * @note  To support poll/select interface for non-blocking read/write file
 * 		operations
 * @param[in] file  - File ptr
 * @param[in] wait  - Ptr to fd for which data availability will be checked
 * @return Mask to inform the current data availability status
 */
uint32_t
bfad_aen_poll(struct file *file, struct poll_table_struct *wait)
{
	bfad_t         *bfad;
	uint32_t        mask = 0;
	unsigned long   flags;

	poll_wait(file, &bfad_aen_readq, wait);

	spin_lock_irqsave(&bfad_aen_spinlock, flags);

	/*
	 * See if there are any events posted
	 */
	list_for_each_entry(bfad, &bfad_list, list_entry) {
		if (bfa_aen_fetch_count(bfad->aen) > 0) {
			mask |= POLLIN | POLLRDNORM;	/* Ok, it's readable */
			break;
		} else {
			/*
			 * Signalled?
			 */
		}
	}

	spin_unlock_irqrestore(&bfad_aen_spinlock, flags);

	return (mask);
}

int
bfad_aen_fasync(int fd, struct file *file, int on)
{
	fasync_enabled = on;
	return fasync_helper(fd, file, on, &bfad_aen_asyncq);
}

/**
 * @}
 */
