/* 
 * Copyright (c)  2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 * 
 * See LICENSE.bfa for copyright and licensing details.
 */

/**
 * @file bfa_attr.c Linux driver configuration interface module.
 */

#include "bfa_os.h"
#include "bfad_drv.h"
#include "bfad_im.h"
#include "bfad_trcmod.h"
#include "bfad_attr.h"

struct Scsi_Host *
bfad_os_starget_to_shost(struct scsi_target *starget)
{
	return dev_to_shost(starget->dev.parent);
}

/*
* TBD: Need to find out the kernel version starting to support
 * FC transport.
 */
/**
 * @brief
 * FC transport template entry, get SCSI host port type.
 */
static void
bfad_im_get_host_port_type(struct Scsi_Host *shost)
{
	bfad_im_port_t *im_port = (bfad_im_port_t *) shost->hostdata[0];
	bfad_t         *bfad = im_port->bfad;
	bfa_pport_attr_t attr;

	bfa_hal_pport_get_attr(&bfad->bfa_hal, &attr);

	switch (attr.port_type) {
	case BFA_PPORT_TYPE_NPORT:
		fc_host_port_type(shost) = FC_PORTTYPE_NPORT;
		break;
	case BFA_PPORT_TYPE_NLPORT:
		fc_host_port_type(shost) = FC_PORTTYPE_NLPORT;
		break;
	case BFA_PPORT_TYPE_P2P:
		fc_host_port_type(shost) = FC_PORTTYPE_PTP;
		break;
	case BFA_PPORT_TYPE_LPORT:
		fc_host_port_type(shost) = FC_PORTTYPE_LPORT;
		break;
	default:
		fc_host_port_type(shost) = FC_PORTTYPE_UNKNOWN;
		break;
	}
}

/**
 * @brief
 * FC transport template entry, get SCSI host port state.
 */
static void
bfad_im_get_host_port_state(struct Scsi_Host *shost)
{
	bfad_im_port_t *im_port = (bfad_im_port_t *) shost->hostdata[0];
	bfad_t         *bfad = im_port->bfad;
	bfa_pport_attr_t attr;

	bfa_hal_pport_get_attr(&bfad->bfa_hal, &attr);

	switch (attr.port_state) {
	case BFA_HAL_PPORT_LINKDOWN:
		fc_host_port_state(shost) = FC_PORTSTATE_LINKDOWN;
		break;
	case BFA_HAL_PPORT_LINKUP:
		fc_host_port_state(shost) = FC_PORTSTATE_ONLINE;
		break;
	case BFA_HAL_PPORT_UNINIT:
	case BFA_HAL_PPORT_ENABLING_QWAIT:
	case BFA_HAL_PPORT_ENABLING:
	case BFA_HAL_PPORT_DISABLING_QWAIT:
	case BFA_HAL_PPORT_DISABLING:
	case BFA_HAL_PPORT_DISABLED:
	case BFA_HAL_PPORT_STOPPED:
	case BFA_HAL_PPORT_IOCDOWN:
	default:
		fc_host_port_state(shost) = FC_PORTSTATE_UNKNOWN;
		break;
	}
}

/**
 * @brief
 * FC transport template entry, get SCSI host active fc4s.
 */
static void
bfad_im_get_host_active_fc4s(struct Scsi_Host *shost)
{
	bfad_im_port_t *im_port = (bfad_im_port_t *) shost->hostdata[0];
	bfad_port_t    *port = im_port->port;

	memset(fc_host_active_fc4s(shost), 0,
	       sizeof(fc_host_active_fc4s(shost)));

	if (port->
	    supported_fc4s & (BFA_PORT_ROLE_FCP_IM | BFA_PORT_ROLE_FCP_TM))
		fc_host_active_fc4s(shost)[2] = 1;

	if (port->supported_fc4s & BFA_PORT_ROLE_FCP_IPFC)
		fc_host_active_fc4s(shost)[3] = 0x20;

	fc_host_active_fc4s(shost)[7] = 1;
}

/**
 * @brief
 * FC transport template entry, get SCSI host link speed.
 */
static void
bfad_im_get_host_speed(struct Scsi_Host *shost)
{
	bfad_im_port_t *im_port = (bfad_im_port_t *) shost->hostdata[0];
	bfad_t         *bfad = im_port->bfad;
	bfa_pport_attr_t attr;

	bfa_hal_pport_get_attr(&bfad->bfa_hal, &attr);
	switch (attr.speed) {
	case BFA_PPORT_SPEED_8GBPS:
		fc_host_speed(shost) = FC_PORTSPEED_8GBIT;
		break;
	case BFA_PPORT_SPEED_4GBPS:
		fc_host_speed(shost) = FC_PORTSPEED_4GBIT;
		break;
	case BFA_PPORT_SPEED_2GBPS:
		fc_host_speed(shost) = FC_PORTSPEED_2GBIT;
		break;
	case BFA_PPORT_SPEED_1GBPS:
		fc_host_speed(shost) = FC_PORTSPEED_1GBIT;
		break;
	default:
		fc_host_speed(shost) = FC_PORTSPEED_UNKNOWN;
		break;
	}
}

/**
 * @brief
 * FC transport template entry, get SCSI host port type.
 */
static void
bfad_im_get_host_fabric_name(struct Scsi_Host *shost)
{
	bfad_im_port_t *im_port = (bfad_im_port_t *) shost->hostdata[0];
	bfad_port_t    *port = im_port->port;
	wwn_t           fabric_nwwn = 0;

	/*
* TBD: Set to fabric nwwn if in switch enviroment. Set to local port
	 * nwwn if private loop or P2P.
	 */
	fabric_nwwn = bfa_fcs_port_get_fabric_name(port->fcs_port);

	fc_host_fabric_name(shost) = fabric_nwwn;

}

/**
 * @brief
 * FC transport template entry, get BFAD statistics.
 */
static struct fc_host_statistics *
bfad_im_get_stats(struct Scsi_Host *shost)
{
	bfad_im_port_t *im_port = (bfad_im_port_t *) shost->hostdata[0];
	bfad_t         *bfad = im_port->bfad;
	struct bfad_hal_comp fcomp;
	struct fc_host_statistics *hstats;
	bfa_status_t    rc;
	unsigned long   flags;

	hstats = &bfad->link_stats;
	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	memset(hstats, 0, sizeof(struct fc_host_statistics));
	rc = bfa_hal_pport_get_stats(&bfad->bfa_hal,
				     (bfa_pport_stats_t *) hstats,
				     bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (rc != BFA_STATUS_OK)
		return NULL;

	wait_for_completion(&fcomp.comp);

	return hstats;
}

/**
 * @brief
 * FC transport template entry, reset BFAD statistics.
 */
static void
bfad_im_reset_stats(struct Scsi_Host *shost)
{
	bfad_im_port_t *im_port = (bfad_im_port_t *) shost->hostdata[0];
	bfad_t         *bfad = im_port->bfad;
	struct bfad_hal_comp fcomp;
	unsigned long   flags;
	bfa_status_t    rc;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	rc = bfa_hal_pport_clear_stats(&bfad->bfa_hal, bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (rc != BFA_STATUS_OK)
		return;

	wait_for_completion(&fcomp.comp);

	return;
}

/**
 * @brief
 * FC transport template entry, get rport loss timeout.
 */
static void
bfad_im_get_rport_loss_tmo(struct fc_rport *rport)
{
	bfad_itnim_data_t *itnim_data = rport->dd_data;
	bfad_itnim_t   *itnim = itnim_data->itnim;
	bfad_t         *bfad = itnim->im->bfad;

	rport->dev_loss_tmo = bfad->cfg_data.rport_del_timeout;
}

/**
 * @brief
 * FC transport template entry, set rport loss timeout.
 */
static void
bfad_im_set_rport_loss_tmo(struct fc_rport *rport, uint32_t timeout)
{
	if (timeout > 0)
		rport->dev_loss_tmo = timeout;
	else {
		bfad_itnim_data_t *itnim_data = rport->dd_data;
		bfad_itnim_t   *itnim = itnim_data->itnim;
		bfad_t         *bfad = itnim->im->bfad;
		rport->dev_loss_tmo = bfad->cfg_data.rport_del_timeout;
	}

}

struct fc_function_template bfad_im_fc_function_template = {

	/*
	 * Target dynamic attributes
	 */
	.get_starget_port_id = bfad_im_get_starget_port_id,
	.show_starget_port_id = 1,
	.get_starget_node_name = bfad_im_get_starget_node_name,
	.show_starget_node_name = 1,
	.get_starget_port_name = bfad_im_get_starget_port_name,
	.show_starget_port_name = 1,

	/*
	 * Host dynamic attribute
	 */
	.get_host_port_id = bfad_im_get_host_port_id,
	.show_host_port_id = 1,

	.issue_fc_host_lip = bfad_im_issue_fc_host_lip,

	/*
	 * TBD: Need to find out the kernel version starting to support
	 * FC transport
	 */
	/*
	 * Host fixed attributes
	 */
	.show_host_node_name = 1,
	.show_host_port_name = 1,
	.show_host_supported_classes = 1,
	.show_host_supported_fc4s = 1,
	.show_host_supported_speeds = 1,
	.show_host_maxframe_size = 1,

	/*
	 * More host dynamic attributes
	 */
	.show_host_port_type = 1,
	.get_host_port_type = bfad_im_get_host_port_type,
	.show_host_port_state = 1,
	.get_host_port_state = bfad_im_get_host_port_state,
	.show_host_active_fc4s = 1,
	.get_host_active_fc4s = bfad_im_get_host_active_fc4s,
	.show_host_speed = 1,
	.get_host_speed = bfad_im_get_host_speed,
	.show_host_fabric_name = 1,
	.get_host_fabric_name = bfad_im_get_host_fabric_name,

	/*
	 * This is host fixed attribute in lk 2.6.16 but changed to dynamic
	 * in lk 2.6.?
	 */
	.show_host_symbolic_name = 1,

	/*
	 * Statistics
	 */
	.get_fc_host_stats = bfad_im_get_stats,
	.reset_fc_host_stats = bfad_im_reset_stats,

	/*
	 * Allocation length for host specific data
	 */
	.dd_fcrport_size = sizeof(bfad_itnim_data_t *),

	/*
	 * Remote port fixed attributes
	 */
	.show_rport_maxframe_size = 1,
	.show_rport_supported_classes = 1,
	.show_rport_dev_loss_tmo = 1,
	.get_rport_dev_loss_tmo = bfad_im_get_rport_loss_tmo,
	.set_rport_dev_loss_tmo = bfad_im_set_rport_loss_tmo,
};

/**
 * @}
 * @dg Scsi_Host_attrs SCSI host attributes
 * @{
 */
static          ssize_t
bfad_im_serial_num_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	bfad_im_port_t *im_port = (bfad_im_port_t *) shost->hostdata[0];
	bfad_t         *bfad = im_port->bfad;
	bfa_ioc_attr_t  ioc_attr;

	memset(&ioc_attr, 0, sizeof(ioc_attr));
	bfa_hal_ioc_get_attr(&bfad->bfa_hal, &ioc_attr);
	return snprintf(buf, PAGE_SIZE, "%s\n",
			ioc_attr.adapter_attr.serial_num);
}

static          ssize_t
bfad_im_model_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	bfad_im_port_t *im_port = (bfad_im_port_t *) shost->hostdata[0];
	bfad_t         *bfad = im_port->bfad;
	bfa_ioc_attr_t  ioc_attr;

	memset(&ioc_attr, 0, sizeof(ioc_attr));
	bfa_hal_ioc_get_attr(&bfad->bfa_hal, &ioc_attr);
	return snprintf(buf, PAGE_SIZE, "%s\n", ioc_attr.adapter_attr.model);
}

static          ssize_t
bfad_im_model_desc_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	bfad_im_port_t *im_port = (bfad_im_port_t *) shost->hostdata[0];
	bfad_t         *bfad = im_port->bfad;
	bfa_ioc_attr_t  ioc_attr;

	memset(&ioc_attr, 0, sizeof(ioc_attr));
	bfa_hal_ioc_get_attr(&bfad->bfa_hal, &ioc_attr);
	return snprintf(buf, PAGE_SIZE, "%s\n",
			ioc_attr.adapter_attr.model_descr);
}

static          ssize_t
bfad_im_node_name_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	bfad_im_port_t *im_port = (bfad_im_port_t *) shost->hostdata[0];
	bfad_port_t    *port = im_port->port;
	uint64_t        nwwn;

	nwwn = bfa_fcs_port_get_nwwn(port->fcs_port);
	return snprintf(buf, PAGE_SIZE, "0x%llx\n", bfa_os_htonll(nwwn));
}

static          ssize_t
bfad_im_symbolic_name_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	bfad_im_port_t *im_port = (bfad_im_port_t *) shost->hostdata[0];
	bfad_t         *bfad = im_port->bfad;
	bfa_ioc_attr_t  ioc_attr;

	memset(&ioc_attr, 0, sizeof(ioc_attr));
	bfa_hal_ioc_get_attr(&bfad->bfa_hal, &ioc_attr);

	return snprintf(buf, PAGE_SIZE, "Brocade %s FV%s DV%s\n",
			ioc_attr.adapter_attr.model,
			ioc_attr.adapter_attr.fw_ver, BFAD_DRIVER_VERSION);
}

static          ssize_t
bfad_im_hw_version_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	bfad_im_port_t *im_port = (bfad_im_port_t *) shost->hostdata[0];
	bfad_t         *bfad = im_port->bfad;
	bfa_ioc_attr_t  ioc_attr;

	memset(&ioc_attr, 0, sizeof(ioc_attr));
	bfa_hal_ioc_get_attr(&bfad->bfa_hal, &ioc_attr);
	return snprintf(buf, PAGE_SIZE, "%s\n", ioc_attr.adapter_attr.hw_ver);
}

static          ssize_t
bfad_im_drv_version_show(struct class_device *cdev, char *buf)
{
	return snprintf(buf, PAGE_SIZE, "%s\n", BFAD_DRIVER_VERSION);
}

static          ssize_t
bfad_im_optionrom_version_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	bfad_im_port_t *im_port = (bfad_im_port_t *) shost->hostdata[0];
	bfad_t         *bfad = im_port->bfad;
	bfa_ioc_attr_t  ioc_attr;

	memset(&ioc_attr, 0, sizeof(ioc_attr));
	bfa_hal_ioc_get_attr(&bfad->bfa_hal, &ioc_attr);
	return snprintf(buf, PAGE_SIZE, "%s\n",
			ioc_attr.adapter_attr.optrom_ver);
}

static          ssize_t
bfad_im_fw_version_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	bfad_im_port_t *im_port = (bfad_im_port_t *) shost->hostdata[0];
	bfad_t         *bfad = im_port->bfad;
	bfa_ioc_attr_t  ioc_attr;

	memset(&ioc_attr, 0, sizeof(ioc_attr));
	bfa_hal_ioc_get_attr(&bfad->bfa_hal, &ioc_attr);
	return snprintf(buf, PAGE_SIZE, "%s\n", ioc_attr.adapter_attr.fw_ver);
}

static          ssize_t
bfad_im_num_of_ports_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	bfad_im_port_t *im_port = (bfad_im_port_t *) shost->hostdata[0];
	bfad_t         *bfad = im_port->bfad;
	bfa_ioc_attr_t  ioc_attr;

	memset(&ioc_attr, 0, sizeof(ioc_attr));
	bfa_hal_ioc_get_attr(&bfad->bfa_hal, &ioc_attr);
	return snprintf(buf, PAGE_SIZE, "%d\n", ioc_attr.adapter_attr.nports);
}

static          ssize_t
bfad_im_drv_name_show(struct class_device *cdev, char *buf)
{
	return snprintf(buf, PAGE_SIZE, "%s\n", BFAD_DRIVER_NAME);
}

static          ssize_t
bfad_im_num_of_discovered_ports_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *shost = class_to_shost(cdev);
	bfad_im_port_t *im_port = (bfad_im_port_t *) shost->hostdata[0];
	bfad_port_t    *port = im_port->port;
	bfad_t         *bfad = im_port->bfad;
	uint32_t        num_ports = 0;
	unsigned long   flags;

	bfad_lock_irqsave(bfad, flags);
	bfa_fcs_port_get_rports(port->fcs_port, NULL, &num_ports);
	bfad_unlock_irqrestore(bfad, flags);

	return snprintf(buf, PAGE_SIZE, "%d\n", num_ports);
}


static          CLASS_DEVICE_ATTR(serial_number, S_IRUGO,
				  bfad_im_serial_num_show, NULL);
static          CLASS_DEVICE_ATTR(model, S_IRUGO, bfad_im_model_show, NULL);
static          CLASS_DEVICE_ATTR(model_description, S_IRUGO,
				  bfad_im_model_desc_show, NULL);
static          CLASS_DEVICE_ATTR(node_name, S_IRUGO, bfad_im_node_name_show,
				  NULL);
static          CLASS_DEVICE_ATTR(symbolic_name, S_IRUGO,
				  bfad_im_symbolic_name_show, NULL);
static          CLASS_DEVICE_ATTR(hardware_version, S_IRUGO,
				  bfad_im_hw_version_show, NULL);
static          CLASS_DEVICE_ATTR(driver_version, S_IRUGO,
				  bfad_im_drv_version_show, NULL);
static          CLASS_DEVICE_ATTR(option_rom_version, S_IRUGO,
				  bfad_im_optionrom_version_show, NULL);
static          CLASS_DEVICE_ATTR(firmware_version, S_IRUGO,
				  bfad_im_fw_version_show, NULL);
static          CLASS_DEVICE_ATTR(number_of_ports, S_IRUGO,
				  bfad_im_num_of_ports_show, NULL);
static          CLASS_DEVICE_ATTR(driver_name, S_IRUGO, bfad_im_drv_name_show,
				  NULL);
static          CLASS_DEVICE_ATTR(number_of_discovered_ports, S_IRUGO,
				  bfad_im_num_of_discovered_ports_show, NULL);


struct class_device_attribute *bfad_im_host_attrs[] = {
	&class_device_attr_serial_number,
	&class_device_attr_model,
	&class_device_attr_model_description,
	&class_device_attr_node_name,
	&class_device_attr_symbolic_name,
	&class_device_attr_hardware_version,
	&class_device_attr_driver_version,
	&class_device_attr_option_rom_version,
	&class_device_attr_firmware_version,
	&class_device_attr_number_of_ports,
	&class_device_attr_driver_name,
	&class_device_attr_number_of_discovered_ports,
	NULL,
};

/**
 * @}
 */
