/* 
 * Copyright (c)  2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 * 
 * See LICENSE.bfa for copyright and licensing details.
 */

/**
 * @brief
 * Contains base driver definitions.
 */

/**
 * @file bfa_drv.h Linux driver data structures.
 */

#ifndef __BFAD_DRV_H__
#define __BFAD_DRV_H__

#include "bfa_os.h"

#include <hal/bfa_hal.h>
#include <hal/bfa_hal_pm.h>
#include <hal/bfa_hal_fcxp.h>
#include <hal/bfa_hal_pport.h>
#include <fcs/bfa_fcs.h>
#include <defs/bfa_defs_pci.h>
#include <defs/bfa_defs_port.h>
#include <defs/bfa_defs_rport.h>
#include <fcs/bfa_fcs_rport.h>
#include <defs/bfa_defs_vport.h>
#include <fcs/bfa_fcs_vport.h>

#include <cs/bfa_plog.h>
#include "aen/bfa_aen.h"
#include <log/bfa_log_linux.h>

#define BFAD_DRIVER_NAME        "bfa"
#ifdef BFA_DRIVER_VERSION
#define BFAD_DRIVER_VERSION    BFA_DRIVER_VERSION
#else
#define BFAD_DRIVER_VERSION    "1.0.0"
#endif

/*
* TBD: PCIE AER is only availabe for the recently release kernels, need to
 * find out the exact kernel version started to support this
 */
#if defined(CONFIG_PCIEPORTBUS) && defined(CONFIG_PCIEAER)
#define BFAD_ENABLE_PCIE_AER(x) pci_enable_pcie_error_reporting(x)
#else
#define BFAD_ENABLE_PCIE_AER(x) {}
#endif

/*
 * Define pm_message_t which came in lk 2.6.11
 * to be backward compatible to older variants of lk 2.6
 */

/*
 * SA_SHIRQ is being pahsed out soon
 */
#define BFAD_IRQ_FLAGS IRQF_SHARED

#ifndef FC_PORTSPEED_8GBIT
#define FC_PORTSPEED_8GBIT 0x10
#endif

/*
 * BFAD flags
 */
#define BFAD_MSIX_ON				0x00000001
#define BFAD_HAL_INIT_DONE			0x00000002
#define BFAD_DRV_INIT_DONE			0x00000004
#define BFAD_CFG_PPORT_DONE			0x00000008
#define BFAD_HAL_START_DONE			0x00000010
#define BFAD_PORT_ONLINE			0x00000020
#define BFAD_RPORT_ONLINE			0x00000040

/*
 * BFAD related definition
 */
#define SCSI_SCAN_DELAY		HZ
#define BFAD_STOP_TIMEOUT	30
#define BFAD_SUSPEND_TIMEOUT	BFAD_STOP_TIMEOUT

/*
 * BFAD configuration parameter default values
 */
#define BFAD_IOC_QUEUE_DEPTH 		2048
#define BFAD_LUN_QUEUE_DEPTH 		32
#define BFAD_DEV_LOSS_TMO		30
#define BFAD_IO_MAX_SGE 		SG_ALL

/*
* TBD: find out the correct kernel version
 */
typedef         irqreturn_t(*bfad_isr_t) (int, void *, struct pt_regs *);

#ifdef CONFIG_PCI_MSI
#define MAX_MSIX_ENTRY 24

struct bfad_msix {
	bfad_isr_t      handler;
	struct msix_entry msix;
};
#endif

typedef enum {
	BFAD_PORT_PHYS_BASE = 0,
	BFAD_PORT_PHYS_VPORT = 1,
	BFAD_PORT_VF_BASE = 2,
	BFAD_PORT_VF_VPORT = 3,
} bfad_port_pvb_type_t;

/*
 * PORT data structure
 */
typedef struct bfad_port_s {
	struct list_head list_entry;
	bfad_t         *bfad;
	bfa_fcs_port_t *fcs_port;
	uint32_t        roles;
	int32_t         flags;
	uint32_t        supported_fc4s;
	uint32_t        pvb_type;
	struct bfad_im_port_s *im_port;	/* IM specific data */
	struct bfad_tm_port_s *tm_port;	/* TM specific data */
	struct bfad_ipfc_port_s *ipfc_port;	/* IPFC specific data */
} bfad_port_t;

/*
 * VPORT data structure
 */
struct bfad_vport_s {
	bfad_port_t     drv_port;
	bfa_fcs_vport_t fcs_vport;
	struct completion *comp_del;
};

/*
 * VF data structure
 */
struct bfad_vf_s {
	bfa_fcs_vf_t    fcs_vf;
	bfad_port_t     base_port;	/* base port for vf */
	bfad_t         *bfad;
};

typedef struct bfad_cfg_param_s {
	uint32_t        rport_del_timeout;
	uint32_t        ioc_queue_depth;
	uint32_t        lun_queue_depth;
	uint32_t        io_max_sge;
	uint32_t        binding_method;
} bfad_cfg_param_t;

/*
 * BFAD (PCI function) data structure
 */
struct bfad_s {
	struct list_head list_entry;
	bfa_hal_t       bfa_hal;
	bfa_fcs_t       bfa_fcs;
	struct pci_dev *pcidev;
	char           *pci_name;
	bfa_hal_pcidev_t hal_pcidev;
	bfa_ioc_pci_attr_t pci_attr;
	unsigned long   pci_bar0_map;
	uint32_t       *pci_bar0_kva;
	struct completion comp;	/* cv for bfad */
	struct completion suspend;	/* cv for bfad suspend */
	struct completion disable_comp;	/* for IOC disable */
	bfa_boolean_t   disable_active;
	bfad_port_t     pport;	/* physical port of the BFAD */
	bfa_hal_meminfo_t meminfo;
	bfa_ioc_cfg_t   ioc_cfg;
	uint32_t        inst_no;	/* BFAD instance number */
	uint32_t        bfad_flags;
	spinlock_t      bfad_lock;
	bfad_cfg_param_t cfg_data;
#ifdef CONFIG_PCI_MSI
	struct bfad_msix msix_tab[MAX_MSIX_ENTRY];
	int             nvec;
#endif

	char            adapter_name[BFA_ADAPTER_SYM_NAME_LEN];
	char            port_name[BFA_ADAPTER_SYM_NAME_LEN];
	struct timer_list hal_tmo;
	unsigned long   hs_start;
	struct bfad_im_s *im;	/* IM specific data */
	struct bfad_tm_s *tm;	/* TM specific data */
	struct bfad_ipfc_s *ipfc;	/* IPFC specific data */
	bfa_log_mod_t   log_data;
	bfa_trc_mod_t  *trcmod;
	bfa_log_mod_t  *logmod;
	bfa_aen_t      *aen;
	bfa_aen_t       aen_buf;
	bfa_plog_t      plog_buf;
	int             ref_count;
	struct fc_host_statistics link_stats;
};

/*
 * RPORT data structure
 */
struct bfad_rport_s {
	bfa_fcs_rport_t fcs_rport;
};

typedef struct bfad_buf_info {
	void           *virt;
	dma_addr_t      phys;
	uint32_t        size;
} bfad_buf_info_t;

typedef struct bfad_fcxp {
	bfad_port_t    *port;
	bfa_hal_rport_t *hal_rport;
	bfa_status_t    req_status;
	uint16_t        tag;
	uint16_t        rsp_len;
	uint16_t        rsp_maxlen;
	uint8_t         use_ireqbuf;
	uint8_t         use_irspbuf;
	uint32_t        num_req_sgles;
	uint32_t        num_rsp_sgles;
	fchs_t          fchs;
	void           *reqbuf_info;
	void           *rspbuf_info;
	bfa_hal_sge_t  *req_sge;
	bfa_hal_sge_t  *rsp_sge;
	fcxp_send_cb_t  send_cbfn;
	void           *send_cbarg;
	void           *bfa_fcxp;
	struct completion comp;	/* cv for completion */
} bfad_fcxp_t;

struct bfad_hal_comp {
	bfa_status_t    status;
	struct completion comp;	/* cv for completion */
};

typedef struct scatterlist sg_t;

/* __kzalloc__{ */
/*
 * For lk does't have kzalloc
 */
/* }__kzalloc__ */

/*
 * BFAD locking macros
 */
#define bfad_lock_irqsave(bfad, flags)				\
	spin_lock_irqsave(&bfad->bfad_lock, flags)
#define bfad_unlock_irqrestore(bfad, flags)			\
	spin_unlock_irqrestore(&bfad->bfad_lock, flags)
#define bfad_lock(bfad) spin_lock(&bfad->bfad_lock)
#define bfad_unlock(bfad) spin_unlock(&bfad->bfad_lock)

/*
* TBD: Need to confirm the kernel version
 */
#define BFAD_WORK_HANDLER(name) void name(void *work)
#define BFAD_INIT_WORK(x, work, func) 			\
	INIT_WORK(&(x)->work, func, &(x)->work)

#define list_remove_head(list, entry, type, member)       	\
do {                                                    	\
	entry = NULL;                                           \
	if (!list_empty(list)) {                                \
		entry = list_entry((list)->next, type, member); 	\
		list_del_init(&entry->member);                  	\
	}								\
} while (0)

#define list_get_first(list, type, member)				\
((list_empty(list)) ? NULL :						\
	list_entry((list)->next, type, member))

bfa_status_t    bfad_vport_create(bfad_t *bfad, uint16_t vf_id,
				  bfa_port_cfg_t *port_cfg);
bfa_status_t    bfad_vf_create(bfad_t *bfad, uint16_t vf_id,
			       bfa_port_cfg_t *port_cfg);
bfa_status_t    bfad_cfg_pport(bfad_t *bfad, bfa_port_role_t role);
bfa_status_t    bfad_fc4_probe(bfad_t *bfad);
bfa_status_t    bfad_drv_init(bfad_t *bfad);
bfad_t         *bfad_find_bfad_by_inst_no(int inst_no);
void            bfad_drv_start(bfad_t *bfad);
void            bfad_uncfg_pport(bfad_t *bfad);
void            bfad_drv_stop(bfad_t *bfad);
void            bfad_remove_intr(bfad_t *bfad);
void            bfad_fc4_probe_undo(bfad_t *bfad);
void            bfad_hal_mem_release(bfad_t *bfad);
void            bfad_hcb_comp(void *arg, bfa_status_t status);

int             bfad_os_ioctl_init(void);
int             bfad_os_ioctl_exit(void);
int             bfad_setup_intr(bfad_t *bfad);
void            bfad_remove_intr(bfad_t *bfad);
int             bfad_os_pci_register_driver(struct pci_driver *drv);
void            bfad_os_pci_unregister_driver(struct pci_driver *drv);
void            bfad_os_device_sysfs_create(bfad_t *);
void            bfad_os_device_sysfs_remove(bfad_t *);
void            bfad_os_pci_set_mwi(struct pci_dev *pdev);
void            bfad_os_idr_init(struct idr *idr);
void            bfad_os_idr_destroy(struct idr *idr);

extern struct idr bfad_im_port_index;
extern struct pci_driver *bfad_pci_driver_p;

extern struct pci_device_id bfad_id_table[];
extern struct semaphore bfad_sem;
extern struct list_head bfad_list;
extern char	*os_name;
extern char	*os_patch;
extern char	*host_name;
extern int	num_rports;
extern int	num_ios;
extern int	num_tms;
extern int	num_fcxps;
extern int	num_ufbufs;
extern int	reqq_size;
extern int	rspq_size;
extern int	num_sgpgs;
extern int      rport_del_timeout;
extern int      msix_disable;
extern int      bfa_ioc_queue_depth;
extern int      bfa_lun_queue_depth;
extern int      bfa_io_max_sge;
extern int      log_level;

extern int      supported_fc4s;

#endif /* __BFAD_DRV_H__ */
