/* 
 * Copyright (c)  2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 * 
 * See LICENSE.bfa for copyright and licensing details.
 */

/**
 * @file bfad_im.c Linux driver IM module.
 */

#include "bfad_drv.h"
#include "bfad_im.h"
#include "bfad_trcmod.h"
#include "bfa_hcb_ioim_macros.h"
#include <linux/spinlock.h>

BFA_TRC_FILE(LDRV, IM);

void bfad_flags_set(bfad_t *bfad, uint32_t flags);

DEFINE_IDR(bfad_im_port_index);
struct scsi_transport_template *bfad_im_scsi_transport_template;

void
bfa_hcb_ioim_done(void *drv, bfad_ioim_t *dio, bfa_ioim_status_t io_status,
		  uint8_t scsi_status, int sns_len, uint8_t *sns_info,
		  int32_t residue)
{
	struct scsi_cmnd *cmnd = (struct scsi_cmnd *)dio;
	bfad_t         *bfad = drv;
	uint8_t         host_status = DID_OK;

	switch (io_status) {
	case BFA_IOIM_STS_OK:
		bfa_trc(bfad, scsi_status);
		cmnd->result = ScsiResult(host_status, scsi_status);
		cmnd->resid  = 0;

		if (sns_len > 0) {
			bfa_trc(bfad, sns_len);
			if (sns_len > SCSI_SENSE_BUFFERSIZE)
				sns_len = SCSI_SENSE_BUFFERSIZE;
			memcpy(cmnd->sense_buffer, sns_info, sns_len);
		}
		if (residue > 0)
			cmnd->resid = residue;
		break;

	case BFA_IOIM_STS_IO_ABORTED:
	case BFA_IOIM_STS_IO_TIMEDOUT:
	default:
		host_status = DID_ERROR;
		cmnd->result = ScsiResult(host_status, 0);
	}

	/*
	 * Unmap DMA, if host is NULL, it means a scsi passthru cmd
	 */
	if (cmnd->device->host != NULL)
		bfad_os_im_dma_unmap(bfad, cmnd);

	cmnd->host_scribble = NULL;
	/*
	 * Call OS IO callbacks (also required for VMware)
	 */
	bfa_trc(bfad, cmnd->result);

	cmnd->scsi_done(cmnd);

	/*
	 * Queue depth adjustment in case of good status completion
	 */
	/*
	 * Queue full handling
	 */
}

void
bfa_hcb_ioim_good_comp(void *drv, bfad_ioim_t *dio)
{
	struct scsi_cmnd *cmnd = (struct scsi_cmnd *)dio;
	bfad_t         *bfad = drv;
	uint8_t         host_status = DID_OK;

	cmnd->result = ScsiResult(host_status, SCSI_STATUS_GOOD);

	/*
	 * Unmap DMA, if host is NULL, it means a scsi passthru cmd
	 */
	if (cmnd->device->host != NULL)
		bfad_os_im_dma_unmap(bfad, cmnd);

	cmnd->host_scribble = NULL;
	/*
	 * Call OS IO callbacks (also required for VMware)
	 */
	bfa_trc_fp(bfad, cmnd->result);
	cmnd->scsi_done(cmnd);
}

void
bfa_hcb_ioim_abort(void *drv, bfad_ioim_t *dio)
{
	struct scsi_cmnd *cmnd = (struct scsi_cmnd *)dio;
	/*
	 * TBD: need bfad_t passed back
	 */
	bfad_t         *bfad = drv;

	cmnd->result = ScsiResult(DID_ERROR, 0);

	/*
	 * Unmap DMA, if host is NULL, it means a scsi passthru cmd
	 */
	if (cmnd->device->host != NULL)
		bfad_os_im_dma_unmap(bfad, cmnd);

	bfa_trc(bfad, cmnd->result);
	cmnd->host_scribble = NULL;
}

void
bfa_hcb_tskim_done(void *bfad, bfad_tskim_t *dtsk,
		   bfa_tskim_status_t tsk_status)
{
	struct scsi_cmnd *cmnd = (struct scsi_cmnd *)dtsk;
	wait_queue_head_t *wq;

	cmnd->SCp.Status |= tsk_status << 1;
	set_bit(IO_DONE_BIT, (unsigned long *)&cmnd->SCp.Status);
	wq = (wait_queue_head_t *) cmnd->SCp.ptr;
	cmnd->SCp.ptr = 0;

	if (wq)
		wake_up(wq);
}

/**
 * @dg Scsi_Host_template SCSI host template
 * @{
 */
/**
 * @brief
 * Scsi_Host template entry, returns BFAD PCI info.
 */
const char     *
bfad_im_info(struct Scsi_Host *shost)
{
	static char     bfa_buf[256];
	bfad_im_port_t *im_port = (bfad_im_port_t *) shost->hostdata[0];
	bfa_ioc_attr_t  ioc_attr;
	bfad_t         *bfad = im_port->bfad;

	memset(&ioc_attr, 0, sizeof(ioc_attr));
	bfa_hal_ioc_get_attr(&bfad->bfa_hal, &ioc_attr);

	bfad_os_dev_init(shost);
	memset(bfa_buf, 0, sizeof(bfa_buf));
	snprintf(bfa_buf, sizeof(bfa_buf),
		 "Brocade FC Adapter, " "model: %s hwpath: %s driver: %s",
		 ioc_attr.adapter_attr.model, bfad->pci_name,
		 BFAD_DRIVER_VERSION);
	return bfa_buf;
}

/**
 * @brief
 * Scsi_Host template entry, aborts the specified SCSI command.
 *
 * Returns: SUCCESS or FAILED.
 */
int
bfad_im_abort_handler(struct scsi_cmnd *cmnd)
{
	struct Scsi_Host *shost = cmnd->device->host;
	bfad_im_port_t *im_port = (bfad_im_port_t *) shost->hostdata[0];
	bfad_t         *bfad = im_port->bfad;
	bfa_hal_ioim_t *hal_io;
	unsigned long   flags;
	uint32_t        timeout;
	int             rc = FAILED;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	hal_io = (bfa_hal_ioim_t *) cmnd->host_scribble;
	if (!hal_io) {
		/*
		 * IO has been completed, retrun success
		 */
		rc = SUCCESS;
		goto out;
	}
	if (hal_io->dio != (bfad_ioim_t *) cmnd) {
		rc = FAILED;
		goto out;
	}

	bfa_trc(bfad, hal_io->iotag);
	bfa_log(bfad->logmod, BFA_LOG_LINUX_SCSI_ABORT,
		im_port->shost->host_no, cmnd, hal_io->iotag);
	bfa_hal_ioim_abort(hal_io);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	/*
	 * Need to wait until the command get aborted
	 */
	timeout = 10;
	while ((bfa_hal_ioim_t *) cmnd->host_scribble == hal_io) {
		bfad_os_spin_os_unlock_irq(shost);
		set_current_state(TASK_UNINTERRUPTIBLE);
		schedule_timeout(timeout);
		bfad_os_spin_os_lock_irq(shost);
		if (timeout < 4 * HZ)
			timeout *= 2;
	}

	/*
	 * Call OS IO callbacks (also required for VMware)
	 */
	cmnd->scsi_done(cmnd);
	bfa_trc(bfad, hal_io->iotag);
	bfa_log(bfad->logmod, BFA_LOG_LINUX_SCSI_ABORT_COMP,
		im_port->shost->host_no, cmnd, hal_io->iotag);
	return SUCCESS;
out:
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	return rc;
}

bfa_status_t
bfad_im_reset_target(bfad_t *bfad, struct scsi_cmnd *cmnd,
		     bfad_itnim_t *itnim, uint32_t target_id,
		     unsigned long *flags)
{
	struct Scsi_Host *shost = cmnd->device->host;
	bfa_hal_tskim_t *tskim;
	bfa_hal_itnim_t *hal_itnim;
	DECLARE_WAIT_QUEUE_HEAD(wq);
	bfa_status_t    rc = BFA_STATUS_OK;
	bfa_tskim_status_t task_status;

	hal_itnim = bfa_fcs_itnim_get_halitn(&itnim->fcs_itnim);
	tskim = bfa_hal_tskim_alloc(&bfad->bfa_hal, (bfad_tskim_t *) cmnd);
	if (!tskim) {
		BFA_DEV_PRINTF(bfad, BFA_ERR,
			       "Target reset," " fail to allocate tskim");
		rc = BFA_STATUS_FAILED;
		goto out;
	}

	/**
	 * Set host_scribble to NULL to avoid aborting a task command if happens.
	 */
	cmnd->host_scribble = NULL;
	cmnd->SCp.ptr = (char *)&wq;
	cmnd->SCp.Status = 0;
	hal_itnim = bfa_fcs_itnim_get_halitn(&itnim->fcs_itnim);
	bfa_hal_tskim_start(tskim, hal_itnim, (lun_t)0,
			    FCP_TM_TARGET_RESET, BFAD_TARGET_RESET_TMO);

	bfad_unlock_irqrestore(bfad, *flags);

	bfad_os_spin_os_unlock_irq(shost);
	wait_event(wq,
		   test_bit(IO_DONE_BIT, (unsigned long *)&cmnd->SCp.Status));
	bfad_os_spin_os_lock_irq(shost);

	bfad_lock_irqsave(bfad, *flags);
	task_status = cmnd->SCp.Status >> 1;
	if (task_status != BFA_TSKIM_STS_OK) {
		BFA_DEV_PRINTF(bfad, BFA_ERR, "LUN failur, status: %d",
			       task_status);
		rc = BFA_STATUS_FAILED;
		goto out;
	}

	/*
	 * TBD: Need a BFA API to verify that NO pending active IO to the
	 * device
	 */
out:
	return rc;
}

/**
 * @brief
 * Scsi_Host template entry, resets a LUN and abort its all commands.
 *
 * Returns: SUCCESS or FAILED.
 *
 */
int
bfad_im_reset_lun_handler(struct scsi_cmnd *cmnd)
{
	struct Scsi_Host *shost = cmnd->device->host;
	bfad_im_port_t *im_port = (bfad_im_port_t *) shost->hostdata[0];
	bfad_itnim_data_t *itnim_data = cmnd->device->hostdata;
	bfad_t         *bfad = im_port->bfad;
	bfa_hal_tskim_t *tskim;
	bfad_itnim_t   *itnim;
	bfa_hal_itnim_t *hal_itnim;
	DECLARE_WAIT_QUEUE_HEAD(wq);
	int             rc = SUCCESS;
	unsigned long   flags;
	bfa_tskim_status_t task_status;

	if (bfad_os_reset_tgt(cmnd)) {
		spin_lock_irqsave(&bfad->bfad_lock, flags);
		/*
		 * execute a target reset if flag not set
		 */
		itnim = itnim_data->itnim;
		if (!itnim) {
			spin_unlock_irqrestore(&bfad->bfad_lock, flags);
			rc = FAILED;
			goto out;
		}
		rc = bfad_im_reset_target(bfad, cmnd, itnim, itnim->scsi_tgt_id,
					  &flags);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
                if (rc == BFA_STATUS_OK)
                        rc = SUCCESS;
                else
                        rc = FAILED;
	} else {
		spin_lock_irqsave(&bfad->bfad_lock, flags);
		itnim = itnim_data->itnim;
		if (!itnim) {
			spin_unlock_irqrestore(&bfad->bfad_lock, flags);
			rc = FAILED;
			goto out;
		}

		tskim = bfa_hal_tskim_alloc(&bfad->bfa_hal,
					    (bfad_tskim_t *) cmnd);
		if (!tskim) {
			BFA_DEV_PRINTF(bfad, BFA_ERR,
				       "LUN reset," " fail to allocate tskim");
			spin_unlock_irqrestore(&bfad->bfad_lock, flags);
			rc = FAILED;
			goto out;
		}

		/**
		 * Set host_scribble to NULL to avoid aborting a task command if happens.
		 */
		cmnd->host_scribble = NULL;
		cmnd->SCp.ptr = (char *)&wq;
		cmnd->SCp.Status = 0;
		hal_itnim = bfa_fcs_itnim_get_halitn(&itnim->fcs_itnim);
		bfa_hal_tskim_start(tskim, hal_itnim,
				    bfad_int_to_lun(cmnd->device->lun),
				    FCP_TM_LUN_RESET, BFAD_LUN_RESET_TMO);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);

		bfad_os_spin_os_unlock_irq(shost);
		wait_event(wq,
			   test_bit(IO_DONE_BIT,
				    (unsigned long *)&cmnd->SCp.Status));
		bfad_os_spin_os_lock_irq(shost);

		task_status = cmnd->SCp.Status >> 1;
		if (task_status != BFA_TSKIM_STS_OK) {
			BFA_DEV_PRINTF(bfad, BFA_ERR, "LUN failure, status: %d",
				       task_status);
			rc = FAILED;
			goto out;
		}

		/*
		 * TBD: Need a BFA API to verify that NO pending active IO to
		 * the device
		 */
	}
out:
	return rc;
}

/**
 * @brief
 * Scsi_Host template entry, resets the bus and abort all commands.
 */
int
bfad_im_reset_bus_handler(struct scsi_cmnd *cmnd)
{
	struct Scsi_Host *shost = cmnd->device->host;
	bfad_im_port_t *im_port = (bfad_im_port_t *) shost->hostdata[0];
	bfad_t         *bfad = im_port->bfad;
	bfad_itnim_t   *itnim;
	unsigned long   flags;
	uint32_t        i, rc, err_cnt = 0;

	bfad_lock_irqsave(bfad, flags);
	for (i = 0; i < MAX_FCP_TARGET; i++) {
		list_for_each_entry(itnim, &im_port->itnim_mapped_list,
				    list_entry) {
			if (itnim->scsi_tgt_id == i) {
				rc = bfad_im_reset_target(bfad, cmnd, itnim, i,
							  &flags);
				if (rc != BFA_STATUS_OK)
					err_cnt++;
				break;
			}
		}
	}
	bfad_unlock_irqrestore(bfad, flags);

	if (err_cnt)
		return FAILED;

	return SUCCESS;
}

/**
 * @brief
 * Scsi_Host template entry slave_destroy.
 */
void
bfad_im_slave_destroy(struct scsi_device *sdev)
{
	sdev->hostdata = NULL;
	return;
}

/**
 * @dg BFA FCS itnim callbacks
 * @{
 */

/**
 * @brief
 * BFA FCS itnim alloc callback, after successful PRLI
 * Context: Interrupt
 */
void
bfa_fcb_itnim_alloc(bfad_t *bfad, bfa_fcs_itnim_t **itnim,
		    bfad_itnim_t **itnim_drv)
{
	*itnim_drv = kzalloc(sizeof(bfad_itnim_t), GFP_ATOMIC);
	if (*itnim_drv == NULL)
		return;

	(*itnim_drv)->im = bfad->im;
	*itnim = &(*itnim_drv)->fcs_itnim;
	(*itnim_drv)->state = ITNIM_STATE_OFFLINE;

	/*
	 * Initiaze the itnim_work
	 */
	bfad_os_itnim_alloc(*itnim_drv);
	bfad_flags_set(bfad, BFAD_RPORT_ONLINE);
}

void
bfad_wwn_to_str(wwn_t wwn, char *str)
{
	sprintf(str, "%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x",
		*((uint8_t *)&wwn),
		*(((uint8_t *)&wwn) + 1),
		*(((uint8_t *)&wwn) + 2),
		*(((uint8_t *)&wwn) + 3),
		*(((uint8_t *)&wwn) + 4),
		*(((uint8_t *)&wwn) + 5),
		*(((uint8_t *)&wwn) + 6),
		*(((uint8_t *)&wwn) + 7));
}

void
bfad_fcid_to_str(uint32_t fcid, char *str)
{
	sprintf(str, "%02x%02x%02x",
		*((uint8_t *)&fcid),
		*(((uint8_t *)&fcid) + 1),
		*(((uint8_t *)&fcid) + 2));
}
		
/**
 * @brief
 * BFA FCS itnim free callback.
 * Context: Interrupt. bfad_lock is held
 */
void
bfa_fcb_itnim_free(bfad_t *bfad, bfad_itnim_t *itnim_drv)
{
	bfad_port_t    *port;
	wwn_t wwpn;
	uint32_t fcid;
	char wwpn_str[32], fcid_str[16];

	itnim_drv->state = ITNIM_STATE_FREE;
	port = bfa_fcs_itnim_get_drvport(&itnim_drv->fcs_itnim);
	itnim_drv->im_port = port->im_port;
	wwpn = bfa_fcs_itnim_get_pwwn(&itnim_drv->fcs_itnim);
	fcid = bfa_fcs_itnim_get_fcid(&itnim_drv->fcs_itnim);
	bfad_wwn_to_str(wwpn, wwpn_str);
	bfad_fcid_to_str(fcid, fcid_str);
	bfa_log(bfad->logmod, BFA_LOG_LINUX_ITNIM_FREE,
		port->im_port->shost->host_no,
		fcid_str, wwpn_str);
	bfad_os_itnim_process(itnim_drv);
}

/**
 * @brief
 * BFA FCS itnim online callback.
 * Context: Interrupt. bfad_lock is held
 */
void
bfa_fcb_itnim_online(bfad_itnim_t *itnim_drv)
{
	bfad_port_t    *port;

	itnim_drv->hal_itnim = bfa_fcs_itnim_get_halitn(&itnim_drv->fcs_itnim);
	itnim_drv->state = ITNIM_STATE_ONLINE;
	port = bfa_fcs_itnim_get_drvport(&itnim_drv->fcs_itnim);
	itnim_drv->im_port = port->im_port;
	bfad_os_itnim_process(itnim_drv);
}

/**
 * @brief
 * BFA FCS itnim offline callback.
 * Context: Interrupt. bfad_lock is held
*/
void
bfa_fcb_itnim_offline(bfad_itnim_t *itnim_drv)
{
	bfad_port_t    *port;

	itnim_drv->state = ITNIM_STATE_OFFLINE;
	port = bfa_fcs_itnim_get_drvport(&itnim_drv->fcs_itnim);
	itnim_drv->im_port = port->im_port;
	bfad_os_itnim_process(itnim_drv);
}

/**
 * @}
 */

/**
 * @brief
 * Allocate a Scsi_Host for a port.
 */
int
bfad_im_scsi_host_alloc(bfad_t *bfad, bfad_im_port_t *im_port)
{
	int             error = 1;

	if (!idr_pre_get(&bfad_im_port_index, GFP_KERNEL)) {
		printk(KERN_WARNING "idr_pre_get failure\n");
		goto out;
	}

	error = bfad_os_idr_get_new(&bfad_im_port_index, &im_port->idr_id);
	if (error) {
		printk(KERN_WARNING "idr_get_new failure\n");
		goto out;
	}

	im_port->shost =
		bfad_os_scsi_host_alloc(&bfad_im_scsi_host_template, im_port,
					bfad);
	if (!im_port->shost) {
		error = 1;
		goto out_free_idr;
	}

	if (bfad_os_scsi_register(im_port->shost, im_port, bfad)) {
		printk(KERN_WARNING "bfad_os_scsi_register failure\n");
		error = 1;
		goto out_unreg;
	}
	if (bfad_os_fc_attach(im_port->shost, &bfad_im_fc_function_template)) {
		printk(KERN_WARNING "bfad_os_fc_attach failure\n");
		error = 1;
		goto out_unreg;
	}

	/*
	 * TBD: How to decide max_id, max_lun, max_cmd_len, can queue?
	 */
	im_port->shost->hostdata[0] = (unsigned long)im_port;
	im_port->shost->unique_id = im_port->idr_id;
	im_port->shost->this_id = -1;
	im_port->shost->max_id = MAX_FCP_TARGET;
	im_port->shost->max_lun = 256;
	im_port->shost->max_cmd_len = 16;
	im_port->shost->can_queue = bfad->cfg_data.ioc_queue_depth;
	bfad_os_transp_templ(im_port->shost, bfad_im_scsi_transport_template);

	bfad_os_select_queue_depths(im_port);
	bfad_os_scsi_set_pci_device(im_port->shost, bfad->pcidev);
	error = bfad_os_scsi_add_host(im_port->shost,
				      (struct device *)&bfad->pcidev->dev);
	if (error) {
		printk(KERN_WARNING "bfad_os_scsi_add_host failure\n");
		goto out_fc_rel;
	}

	/*
	 * setup host fixed attribute if the lk supports
	 */
	bfad_os_fc_host_init(im_port);

	return 0;

out_fc_rel:
	bfad_os_fc_release_transp(im_port->shost);
out_unreg:
	bfad_os_scsi_put_host(im_port->shost);
	bfad_os_scsi_unregister(im_port->shost);

out_free_idr:
	idr_remove(&bfad_im_port_index, im_port->idr_id);
out:
	return error;
}

void
bfad_im_scsi_host_free(bfad_t *bfad, bfad_im_port_t *im_port)
{
	unsigned long flags;

	bfa_trc(bfad, bfad->inst_no);
	bfa_log(bfad->logmod, BFA_LOG_LINUX_SCSI_HOST_FREE,
			im_port->shost->host_no);
	/*
	 * TBD: we may need to defer the processing since some of the functions
	 * are not supposed to be called when lock is held.
	 */

	bfad_os_fc_remove_host(im_port->shost);

	bfad_os_scsi_remove_host(im_port->shost);
	bfad_os_fc_release_transp(im_port->shost);
	bfad_os_scsi_put_host(im_port->shost);
	bfad_os_scsi_unregister(im_port->shost);

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	idr_remove(&bfad_im_port_index, im_port->idr_id);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
}

bfa_status_t
bfad_im_port_new(bfad_t *bfad, bfad_port_t *port)
{
	int             rc = BFA_STATUS_OK;
	bfad_im_port_t *im_port;

	im_port = kzalloc(sizeof(bfad_im_port_t), GFP_ATOMIC);
	if (im_port == NULL) {
		rc = BFA_STATUS_ENOMEM;
		goto ext;
	}
	port->im_port = im_port;
	im_port->port = port;
	im_port->bfad = bfad;

	bfad_os_init_work(im_port);
	INIT_LIST_HEAD(&im_port->itnim_mapped_list);
	INIT_LIST_HEAD(&im_port->binding_list);

ext:
	return rc;
}

void
bfad_im_port_delete(bfad_t *bfad, bfad_port_t *port)
{
	bfad_im_port_t *im_port = port->im_port;

        bfad_os_queue_work(bfad->im->drv_workq,
                                   &im_port->port_delete_work);
}

BFAD_WORK_HANDLER(bfad_im_port_delete_handler)
{	/* work struct is the function arg */
	bfad_im_port_t *im_port =
		container_of(work, bfad_im_port_t, port_delete_work);

	bfad_im_scsi_host_free(im_port->bfad, im_port);
	bfad_im_port_clean(im_port);
	kfree(im_port);
}

void 
bfad_im_port_clean(bfad_im_port_t *im_port)
{
	bfad_fcp_binding_t *bp, *bp_new;
	unsigned long flags;
	bfad_t *bfad =  im_port->bfad;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	list_for_each_entry_safe(bp, bp_new, &im_port->binding_list,
					list_entry) {
		list_del(&bp->list_entry);
		kfree(bp);
	}

	/* the itnim_mapped_list must be empty at this time */
	bfa_assert(list_empty(&im_port->itnim_mapped_list));

	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
}

void
bfad_im_port_online(bfad_t *bfad, bfad_port_t *port)
{
}

void
bfad_im_port_offline(bfad_t *bfad, bfad_port_t *port)
{
}

bfa_status_t
bfad_im_probe(bfad_t *bfad)
{
	bfad_im_t      *im;
	bfa_status_t    rc = BFA_STATUS_OK;

	im = kzalloc(sizeof(bfad_im_t), GFP_KERNEL);
	if (im == NULL) {
		rc = BFA_STATUS_ENOMEM;
		goto ext;
	}

	bfad->im = im;
	im->bfad = bfad;

	if (bfad_os_thread_workq(bfad) != BFA_STATUS_OK) {
		kfree(im);
		rc = BFA_STATUS_FAILED;
	}

ext:
	return rc;
}

void
bfad_im_probe_undo(bfad_t *bfad)
{
	if (bfad->im) {
		bfad_os_destroy_workq(bfad->im);
		kfree(bfad->im);
		bfad->im = NULL;
	}
}
