/*
 * Brocade Fibre Channel HBA Driver
 * Copyright (c) 2006-2008 Brocade Communications Systems, Inc.
 * All rights reserved.
 */

#include <asm/uaccess.h>
#include "bfad_ioctl.h"
#include "bfad_drv.h"
#include "bfad_aen.h"
#include "bfad_im.h"
#include "hal/bfa_hal_ioim.h"
#include "hal/bfa_hal_diag.h"
#include "hal/bfa_hal_debug.h"
#include "fcs/bfa_fcs_fcpim.h"
#include "fcs/bfa_fcs_auth.h"
#include "bfad_trcmod.h"
#include "bfad_sysfs.h"
#include <linux/module.h>

BFA_TRC_FILE(LDRV, IOCTL);

bfad_fcxp_t    *
bfad_fcxp_alloc(bfad_t *bfad)
{
	void           *ptr;

	ptr = kmalloc(sizeof(bfad_fcxp_t), GFP_KERNEL);
	if (ptr)
		memset(ptr, 0, sizeof(bfad_fcxp_t));
	return ptr;
}

void
bfad_fcxp_free(bfad_fcxp_t *fcxp)
{
	kfree(fcxp);
}

struct bfad_scsi_passthru {
	struct completion comp;	/* cv for completion */
	bfa_ioctl_scsi_t iocmd;
};

void
bfad_send_scsi_cb(void *arg)
{
	struct scsi_cmnd *scsi_cmd = arg;
	struct bfad_scsi_passthru *pthru =
		(struct bfad_scsi_passthru *)scsi_cmd->SCp.ptr;

	/*
	 * Wake up the waiting thread
	 */
	complete(&pthru->comp);
}

void
bfad_scsi_sg_data_free(bfad_t *bfad, sg_t *sg_base, uint32_t num_sgles)
{
	int             i;
	sg_t           *sg_info = sg_base;

	for (i = 0; i < num_sgles; sg_info++, i++) {
		if (sg_info->page != NULL) {
			dma_free_coherent(&bfad->pcidev->dev, sg_info->length,
					  sg_info->page, sg_info->dma_address);
		}
	}

	kfree(sg_base);
}

int
bfad_scsi_data_setup(bfad_t *bfad, struct scsi_cmnd *scsi_cmd, char *data_buf,
		     int write, int size)
{
	int             sge_num = 0, cnt, offset = 0;
	sg_t           *sg_info;
	int             rc = 0;

	if (size == 0) {
		bfa_trc(bfad, 0);
		return 0;
	}

	if (size < PAGE_SIZE) {
		scsi_cmd->use_sg = 0;
		scsi_cmd->request_buffer =
#ifdef __VMKERNEL_MODULE__
			dma_alloc_coherent(&bfad->pcidev->dev, size,
					   &(scsi_cmd->request_bufferMA),
					   GFP_KERNEL);
#else

			dma_alloc_coherent(&bfad->pcidev->dev, size,
					   &(scsi_cmd->SCp.dma_handle),
					   GFP_KERNEL);
#endif
		if (!scsi_cmd->request_buffer) {
			bfa_trc(bfad, 0);
			return ENOMEM;
		}
		scsi_cmd->request_bufflen = size;

		if (write) {
			if (copy_from_user
			    ((uint8_t *) scsi_cmd->request_buffer,
			     (void __user *)data_buf, size)) {

#ifdef __VMKERNEL_MODULE__
				dma_free_coherent(&bfad->pcidev->dev, size,
						  scsi_cmd->request_buffer,
						  scsi_cmd->request_bufferMA);
#else
				dma_free_coherent(&bfad->pcidev->dev, size,
						  scsi_cmd->request_buffer,
						  scsi_cmd->SCp.dma_handle);
#endif
				bfa_trc(bfad, 0);
				return EIO;
			}
		}
	} else {
		sge_num = ((size % PAGE_SIZE) ? (size / PAGE_SIZE +
			1) : (size / PAGE_SIZE));
		scsi_cmd->request_buffer =
			kzalloc(sizeof(sg_t[sge_num]), GFP_KERNEL);
		if (!scsi_cmd->request_buffer) {
			bfa_trc(bfad, 0);
			return ENOMEM;
		}

		sg_info = scsi_cmd->request_buffer;

		while (size) {
			cnt = (size > PAGE_SIZE) ? PAGE_SIZE : size;

			sg_info->page =
				dma_alloc_coherent(&bfad->pcidev->dev, cnt,
						   &(sg_info->dma_address),
						   GFP_KERNEL);
			if (!sg_info->page) {
				bfa_trc(bfad, 0);
				rc = ENOMEM;
				goto out_free_mem;
			}

			if (write) {
				if (copy_from_user
				    ((uint8_t *) sg_info->page,
				     (void __user *)(data_buf + offset), cnt)) {
					bfa_trc(bfad, 0);
					rc = EIO;
					goto out_free_mem;
				}
			}

			sg_info->length = cnt;
			size -= cnt;
			offset += cnt;
			sg_info++;
		}
	}

	return 0;

out_free_mem:
	bfad_scsi_sg_data_free(bfad, scsi_cmd->request_buffer, sge_num);
	return rc;
}

int
bfad_scsi_data_copyout(bfad_t *bfad, char *data_buf,
		       struct scsi_cmnd *scsi_cmd, uint32_t size)
{
	int             cnt, offset = 0;
	sg_t           *sg_info;

	if (scsi_cmd->use_sg == 0 && size <= PAGE_SIZE && data_buf != NULL) {
		/*
		 * Copy data to user space
		 */
		if (copy_to_user
		    ((void __user *)(data_buf),
		     (uint8_t *) scsi_cmd->request_buffer, size)) {

			bfa_trc(bfad, 0);
			return EIO;
		}
	} else if (scsi_cmd->use_sg > 0 && size >= PAGE_SIZE
		   && data_buf != NULL) {
		sg_info = scsi_cmd->request_buffer;

		while (size) {
			cnt = (size > PAGE_SIZE) ? PAGE_SIZE : size;

			/*
			 * pci_dma_sync_single_for_device(bfad->pcidev,
			 * sg_info->dma_address, cnt, PCI_DMA_FROMDEVICE);
			 */

			/*
			 * Copy data to user space
			 */
			if (copy_to_user
			    ((void __user *)(data_buf + offset),
			     (uint8_t *) sg_info->page, cnt)) {
				bfa_trc(bfad, 0);
				return EIO;
			}

			sg_info++;
			offset += cnt;
			size -= cnt;
		}
	}

	return 0;
}

unsigned int 
bfad_scsilun_to_int(uint8_t *scsilun)
{
        int i;
        unsigned int lun;

        lun = 0;
        for (i = 0; i < sizeof(lun); i += 2)
                lun = lun | (((scsilun[i] << 8) | scsilun[i + 1]) << (i * 8));
        return lun;
}

void
bfad_scsi_cmd_setup(bfad_t *bfad, struct scsi_cmnd *scsi_cmd,
		    bfa_ioctl_scsi_t *ci, struct bfad_scsi_passthru *pthru)
{
	/*
	 * Setup data direction
	 */
	if (ci->write && ci->iosz > 0)
		scsi_cmd->sc_data_direction = DMA_TO_DEVICE;
	else if (ci->iosz > 0)
		scsi_cmd->sc_data_direction = DMA_FROM_DEVICE;
	else
		scsi_cmd->sc_data_direction = DMA_NONE;

	memcpy(scsi_cmd->cmnd, ci->cdb.scsi_cdb, sizeof(ci->cdb));
	scsi_cmd->cmd_len = sizeof(ci->cdb);
	scsi_cmd->device->tagged_supported = 0;
	scsi_cmd->device->lun = bfad_scsilun_to_int((uint8_t *)&ci->lun);

	scsi_cmd->scsi_done = (void (*)(struct scsi_cmnd *))bfad_send_scsi_cb;
	scsi_cmd->SCp.ptr = (char *)pthru;

}

static int
bfad_ioctl_itnim_get_attr(unsigned long arg)
{
	bfa_ioctl_itnim_attr_t iocmd;
	bfad_t         *bfad;
	bfa_fcs_port_t *fcs_port;
	unsigned long   flags;

	if (copy_from_user((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_itnim_attr_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port =
		bfa_fcs_lookup_port(&bfad->bfa_fcs, iocmd.vf_id, iocmd.lpwwn);
	if (!fcs_port) {
		iocmd.status = BFA_STATUS_UNKNOWN_LWWN;
		bfa_trc(bfad, 0);
	} else {
		iocmd.status =
			bfa_fcs_itnim_attr_get(fcs_port, iocmd.rpwwn,
					       &iocmd.attr);
	}
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (copy_to_user((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_itnim_attr_t))) {
		bfa_trc(bfad, 0);
		return EIO;
	}

	return 0;
}

static int
bfad_ioctl_itnim_get_stats(unsigned long arg)
{
	bfa_ioctl_itnim_stats_t iocmd;
	bfad_t         *bfad;
	bfa_fcs_port_t *fcs_port;
	bfa_fcs_itnim_t *itnim;
	unsigned long   flags;

	if (copy_from_user((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_itnim_stats_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port =
		bfa_fcs_lookup_port(&bfad->bfa_fcs, iocmd.vf_id, iocmd.lpwwn);
	if (!fcs_port) {
		iocmd.status = BFA_STATUS_UNKNOWN_LWWN;
		bfa_trc(bfad, 0);
	} else {
		itnim = bfa_fcs_itnim_lookup(fcs_port, iocmd.rpwwn);
		if (itnim == NULL)
			iocmd.status = BFA_STATUS_UNKNOWN_RWWN;
		else {
			iocmd.status = BFA_STATUS_OK;
			bfa_fcs_itnim_stats_get(fcs_port, iocmd.rpwwn, &iocmd.stats);
			bfa_hal_itnim_get_stats(bfa_fcs_itnim_get_halitn(itnim),
						&iocmd.stats.hal_stats);
		}
	}
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (copy_to_user((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_itnim_stats_t))) {
		bfa_trc(bfad, 0);
		return EIO;
	}

	return 0;
}

static int
bfad_ioctl_itnim_reset_stats(unsigned long arg)
{
	bfa_ioctl_itnim_reset_stats_t iocmd;
	bfad_t                        *bfad;
	bfa_fcs_port_t                *fcs_port;
	bfa_fcs_itnim_t               *itnim;
	unsigned long                 flags;

	if (copy_from_user((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_itnim_reset_stats_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port =
		bfa_fcs_lookup_port(&bfad->bfa_fcs, iocmd.vf_id, iocmd.lpwwn);
	if (!fcs_port) {
		iocmd.status = BFA_STATUS_UNKNOWN_LWWN;
		bfa_trc(bfad, 0);
	} else {
		itnim = bfa_fcs_itnim_lookup(fcs_port, iocmd.rpwwn);
		if (itnim == NULL)
			iocmd.status = BFA_STATUS_UNKNOWN_RWWN;
		else {
			iocmd.status = BFA_STATUS_OK;
			bfa_fcs_itnim_stats_clear(fcs_port, iocmd.rpwwn);
			bfa_hal_itnim_clear_stats(bfa_fcs_itnim_get_halitn(itnim));
		}
	}
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (copy_to_user((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_itnim_reset_stats_t))) {
		bfa_trc(bfad, 0);
		return EIO;
	}

	return 0;
}

static int
bfad_send_scsi_passthru(unsigned long arg)
{
	struct scsi_cmnd *scsi_cmd;
	struct scsi_device *scsi_device;
	struct bfad_scsi_passthru *pthru;
	bfa_ioctl_scsi_t *ci;
	int             rc = 0, total_size;
	uint16_t        sg_cnt = 0;
	unsigned long   flags;
	bfa_hal_ioim_t *hal_io = NULL;
	bfa_fcs_port_t *fcs_port;
	bfa_fcs_itnim_t *fcs_itnim;
	bfa_hal_itnim_t *hal_itnim;
	bfad_t         *bfad;
	uint16_t host_status, scsi_status;

	total_size =
		sizeof(struct scsi_cmnd) + sizeof(struct scsi_device) +
		sizeof(struct bfad_scsi_passthru);
	scsi_cmd = kzalloc(total_size, GFP_KERNEL);
	if (scsi_cmd == NULL)
		return ENOMEM;

	scsi_device = (struct scsi_device *)(scsi_cmd + 1);
	pthru = (struct bfad_scsi_passthru *)(scsi_device + 1);
	ci = &pthru->iocmd;
	scsi_cmd->device = scsi_device;

	if (copy_from_user
	    ((uint8_t *) ci, (void __user *)arg, sizeof(bfa_ioctl_scsi_t))) {
		rc = EIO;
		goto out_free_scsi_cmd;
	}

	bfad = bfad_find_bfad_by_inst_no(ci->bfad_num);
	if (bfad == NULL) {
		rc = EINVAL;
		goto out_free_scsi_cmd;
	}

	bfa_trc(bfad, 0);

	/*
	 * Setup of the data fields in scsi_cmd
	 */
	rc = bfad_scsi_data_setup(bfad, scsi_cmd,
				  (char *)(size_t) ci->iobuf_ptr, ci->write,
				  ci->iosz);
	if (rc != 0) {
		bfa_trc(bfad, 0);
		goto out_free_scsi_cmd;
	}

	/*
	 * Setup Misc scsi_cmd/scsi_device field needed for callback macros
	 */
	bfad_scsi_cmd_setup(bfad, scsi_cmd, ci, pthru);

	if (scsi_cmd->use_sg)
		sg_cnt = scsi_cmd->use_sg;
	else if (scsi_cmd->request_bufflen)
		sg_cnt = 1;

	spin_lock_irqsave(&bfad->bfad_lock, flags);

	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs, ci->vf_id, ci->lpwwn);
	if (!fcs_port) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		ci->status = BFA_STATUS_UNKNOWN_LWWN;
		bfa_trc(bfad, 0);
		goto out_free_scsi_data;
	}

	fcs_itnim = bfa_fcs_itnim_lookup(fcs_port, ci->rpwwn);
	if (!fcs_itnim) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		ci->status = BFA_STATUS_UNKNOWN_RWWN;
		bfa_trc(bfad, 0);
		goto out_free_scsi_data;
	}

	if (fcs_itnim->itnim_drv->state == ITNIM_STATE_OFFLINE) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		ci->status = BFA_STATUS_NO_FCPIM_NEXUS;
		bfa_trc(bfad, 0);
		goto out_free_scsi_data;
	}

	hal_itnim = bfa_fcs_itnim_get_halitn(fcs_itnim);
	hal_io = bfa_hal_ioim_alloc(&bfad->bfa_hal, (bfad_ioim_t *) scsi_cmd,
				    hal_itnim, sg_cnt);
	if (!hal_io) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		ci->status = BFA_STATUS_ENOMEM;
		bfa_trc(bfad, 0);
		goto out_free_scsi_data;
	}

	scsi_cmd->host_scribble = (char *)hal_io;
	init_completion(&pthru->comp);
	bfa_hal_ioim_start(hal_io);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	wait_for_completion(&pthru->comp);

	bfa_trc(bfad, ci->iosz);
	bfa_trc(bfad, ci->write);

	host_status = scsi_cmd->result >> 16;
	scsi_status = scsi_cmd->result & 0xFF;
	if (host_status != DID_OK) {
		ci->status = BFA_STATUS_FAILED;
		bfa_trc(bfad, host_status);
		goto out_free_scsi_data;
	} else if (ci->iosz > 0 && ci->write == 0) {
		rc = bfad_scsi_data_copyout(bfad,
					    (char *)(size_t) ci->
					    iobuf_ptr, scsi_cmd,
					    ci->iosz);
		if (rc != 0) {
			bfa_trc(bfad, 0);
			goto out_free_scsi_data2;
		}
	}

	/* Copy sense data */
	if (ci->snssz > 0 &&
	    copy_to_user((void __user *)(size_t)ci->snsbuf_pt,
		(uint8_t *) scsi_cmd->sense_buffer,
		 (ci->snssz > SCSI_SENSE_BUFFERSIZE ? SCSI_SENSE_BUFFERSIZE :
			ci->snssz))) {
		bfa_trc(bfad, 0);
		rc = EIO;
		goto out_free_scsi_data2; 
	}

	ci->status = BFA_STATUS_OK;
	ci->scsi_status = scsi_status;

out_free_scsi_data:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) ci, sizeof(bfa_ioctl_scsi_t))) {
		bfa_trc(bfad, 0);
		rc = EIO;
	}

out_free_scsi_data2:
	if (scsi_cmd->request_buffer && scsi_cmd->use_sg)
		bfad_scsi_sg_data_free(bfad, scsi_cmd->request_buffer,
				       scsi_cmd->use_sg);
	else if (scsi_cmd->request_buffer)
#ifdef __VMKERNEL_MODULE__
		dma_free_coherent(&bfad->pcidev->dev, scsi_cmd->request_bufflen,
				  scsi_cmd->request_buffer,
				  scsi_cmd->request_bufferMA);
#else
		dma_free_coherent(&bfad->pcidev->dev, scsi_cmd->request_bufflen,
				  scsi_cmd->request_buffer,
				  scsi_cmd->SCp.dma_handle);
#endif
out_free_scsi_cmd:
	kfree(scsi_cmd);

	return rc;
}

void
bfad_send_fcpt_cb(void *bfad, bfa_hal_fcxp_t *fcxp, void *cbarg,
		  bfa_status_t req_status, uint32_t rsp_len, uint32_t resid_len,
		  fchs_t *rsp_fchs)
{
	bfad_fcxp_t    *drv_fcxp = cbarg;

	/*
	 * TBD
	 */
	drv_fcxp->req_status = req_status;
	drv_fcxp->rsp_len = rsp_len;
	bfa_trc((bfad_t *) bfad, req_status);

	/*
	 * bfa_fcxp will be automatically freed by HAL on return from the
	 * callback function
	 */
	drv_fcxp->bfa_fcxp = NULL;

	/*
	 * Wake up the waiting thread
	 */
	complete(&drv_fcxp->comp);
}

void
bfad_fcxp_mem_free(bfad_t *bfad, bfad_buf_info_t *buf_base,
		   uint32_t num_sgles)
{
	int             i;
	bfad_buf_info_t *buf_info = buf_base;

	for (i = 0; i < num_sgles; buf_info++, i++) {
		if (buf_info->virt != NULL) {
			dma_free_coherent(&bfad->pcidev->dev, buf_info->size,
					  buf_info->virt, buf_info->phys);
		}
	}

	kfree(buf_base);
}

bfad_buf_info_t *
bfad_fcxp_mem_alloc(bfad_t *bfad, char *inbuf_ptr, uint32_t size,
		    uint32_t *num_sgles)
{
	int             cnt, offset = 0;
	int             sge_num, total_size;
	bfad_buf_info_t *buf_base, *buf_info;
	bfa_hal_sge_t  *sg_table;

	sge_num = ((size % PAGE_SIZE) ?
			(size / PAGE_SIZE + 1) : (size / PAGE_SIZE));
	total_size = sizeof(bfad_buf_info_t) + sizeof(bfa_hal_sge_t);
	buf_base = kzalloc(total_size * sge_num, GFP_KERNEL);
	sg_table = (bfa_hal_sge_t *) (buf_base + sge_num);

	if (buf_base == NULL)
		return NULL;

	buf_info = buf_base;
	while (size) {
		cnt = (size > PAGE_SIZE) ? PAGE_SIZE : size;
		buf_info->virt =
			dma_alloc_coherent(&bfad->pcidev->dev, cnt,
					   &(buf_info->phys), GFP_KERNEL);
		buf_info->size = cnt;

		if (inbuf_ptr) {
			if (copy_from_user
			    ((uint8_t *) buf_info->virt,
			     (void __user *)(inbuf_ptr + offset), cnt))
				goto out;
		} else {
			memset(buf_info->virt, 0, cnt);
		}

		/*
		 * Setup SG table
		 */
		sg_table->sg_len = cnt;
		sg_table->sg_addr = (void *)(size_t) buf_info->phys;

		offset += cnt;
		size -= cnt;
		buf_info++;
		sg_table++;
	}

	*num_sgles = sge_num;

	return buf_base;

out:
	bfad_fcxp_mem_free(bfad, buf_base, sge_num);
	return NULL;
}

int
bfad_rsp_data_copy(bfad_t *bfad, char *obuf_ptr, bfad_buf_info_t *buf_info,
		   uint32_t size)
{
	int             cnt, offset = 0;

	while (size) {
		cnt = (size > PAGE_SIZE) ? PAGE_SIZE : size;

		if (obuf_ptr) {
			/*
			 * pci_dma_sync_single_for_device(bfad->pcidev,
			 * buf_info->phys, cnt, PCI_DMA_FROMDEVICE);
			 */

			/*
			 * Copy data to user space
			 */
			if (copy_to_user
			    ((void __user *)(obuf_ptr + offset),
			     (uint8_t *) buf_info->virt, cnt))
				return 1;
		}

		buf_info++;
		offset += cnt;
		size -= cnt;
	}

	return 0;
}

uint64_t
bfad_fcxp_get_req_sgaddr_cb(void *bfad_fcxp, int sgeid)
{
	bfad_fcxp_t    *drv_fcxp = bfad_fcxp;
	bfa_hal_sge_t  *sge;
	uint64_t        addr;

	sge = drv_fcxp->req_sge + sgeid;
	addr = (uint64_t) (size_t) sge->sg_addr;

	return (addr);
}

uint32_t
bfad_fcxp_get_req_sglen_cb(void *bfad_fcxp, int sgeid)
{
	bfad_fcxp_t    *drv_fcxp = bfad_fcxp;
	bfa_hal_sge_t  *sge;

	sge = drv_fcxp->req_sge + sgeid;
	return (sge->sg_len);
}

uint64_t
bfad_fcxp_get_rsp_sgaddr_cb(void *bfad_fcxp, int sgeid)
{
	bfad_fcxp_t    *drv_fcxp = bfad_fcxp;
	bfa_hal_sge_t  *sge;
	uint64_t        addr;

	sge = drv_fcxp->rsp_sge + sgeid;
	addr = (uint64_t) (size_t) sge->sg_addr;

	return (addr);
}

uint32_t
bfad_fcxp_get_rsp_sglen_cb(void *bfad_fcxp, int sgeid)
{
	bfad_fcxp_t    *drv_fcxp = bfad_fcxp;
	bfa_hal_sge_t  *sge;

	sge = drv_fcxp->req_sge + sgeid;
	return (sge->sg_len);
}

int
bfad_fcxp_send(bfad_fcxp_t *drv_fcxp, bfa_ioctl_fcpt_t *ci)
{
	bfa_hal_fcxp_t *hal_fcxp;
	bfad_t         *bfad = drv_fcxp->port->bfad;
	bfa_fcxp_get_sgaddr_t req_sgaddr_cb;
	bfa_fcxp_get_sglen_t req_sglen_cb;
	bfa_fcxp_get_sgaddr_t rsp_sgaddr_cb;
	bfa_fcxp_get_sglen_t rsp_sglen_cb;
	int             num_req_sgles, num_rsp_sgles;
	unsigned long   flags;

	if (drv_fcxp->use_ireqbuf && drv_fcxp->use_irspbuf) {
		bfa_trc(bfad, 0);
		num_req_sgles = 0;
		num_rsp_sgles = 0;
		req_sgaddr_cb = NULL;
		req_sglen_cb = NULL;
		rsp_sgaddr_cb = NULL;
		rsp_sglen_cb = NULL;
	} else if (drv_fcxp->use_ireqbuf && !drv_fcxp->use_irspbuf) {
		bfa_trc(bfad, 0);
		num_req_sgles = 0;
		num_rsp_sgles = drv_fcxp->num_rsp_sgles;
		req_sgaddr_cb = NULL;
		req_sglen_cb = NULL;
		rsp_sgaddr_cb = bfad_fcxp_get_rsp_sgaddr_cb;
		rsp_sglen_cb = bfad_fcxp_get_rsp_sglen_cb;
	} else if (!drv_fcxp->use_ireqbuf && drv_fcxp->use_irspbuf) {
		bfa_trc(bfad, 0);
		num_req_sgles = drv_fcxp->num_req_sgles;
		num_rsp_sgles = 0;
		req_sgaddr_cb = bfad_fcxp_get_req_sgaddr_cb;
		req_sglen_cb = bfad_fcxp_get_req_sglen_cb;
		rsp_sgaddr_cb = NULL;
		rsp_sglen_cb = NULL;
	} else {
		bfa_trc(bfad, 0);
		num_req_sgles = drv_fcxp->num_req_sgles;
		num_rsp_sgles = drv_fcxp->num_rsp_sgles;
		req_sgaddr_cb = bfad_fcxp_get_req_sgaddr_cb;
		req_sglen_cb = bfad_fcxp_get_req_sglen_cb;
		rsp_sgaddr_cb = bfad_fcxp_get_rsp_sgaddr_cb;
		rsp_sglen_cb = bfad_fcxp_get_rsp_sglen_cb;
	}

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	hal_fcxp =
		bfa_hal_fcxp_alloc(bfad, &bfad->bfa_hal, num_req_sgles,
				   num_rsp_sgles, req_sgaddr_cb, req_sglen_cb,
				   rsp_sgaddr_cb, rsp_sglen_cb);

	if (!hal_fcxp) {
		bfa_trc(bfad, 0);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		return BFA_STATUS_ENOMEM;
	}

	drv_fcxp->bfa_fcxp = hal_fcxp;
	if (drv_fcxp->use_ireqbuf) {
		drv_fcxp->reqbuf_info = bfa_hal_fcxp_get_reqbuf(hal_fcxp);

		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		if (copy_from_user
		    ((uint8_t *) drv_fcxp->reqbuf_info,
		     (void __user *)(size_t) ci->reqbuf_ptr, ci->reqsz))
			return EIO;
		spin_lock_irqsave(&bfad->bfad_lock, flags);
	}

	if (drv_fcxp->use_irspbuf)
		drv_fcxp->rspbuf_info = bfa_hal_fcxp_get_rspbuf(hal_fcxp);

	bfa_hal_fcxp_send(hal_fcxp, drv_fcxp->hal_rport, ci->vf_id, ci->cts,
			  ci->cos, ci->reqsz, &ci->fchs, bfad_send_fcpt_cb,
			  drv_fcxp, ci->rspsz, ci->tsecs);

	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	return 0;
}

static int
bfad_send_fc_passthru(unsigned long arg)
{
	bfa_ioctl_fcpt_t fc_req, *ci;
	int             rc = 0;
	bfad_fcxp_t    *drv_fcxp;
	bfa_fcs_port_t *fcs_port;
	bfa_fcs_rport_t *fcs_rport;
	bfa_hal_rport_t *hal_rport = NULL;
	bfad_t         *bfad;
	unsigned long   flags;

	ci = &fc_req;

	if (copy_from_user((uint8_t *) ci, (void __user *)arg,
			    sizeof(bfa_ioctl_fcpt_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(ci->bfad_num);
	if (bfad == NULL)
		return EINVAL;

	bfa_trc(bfad, ci->vf_id);
	bfa_trc(bfad, ci->lpwwn);
	bfa_trc(bfad, ci->dpwwn);
	bfa_trc(bfad, ci->tsecs);
	bfa_trc(bfad, ci->cts);
	bfa_trc(bfad, ci->cos);
	bfa_trc(bfad, ci->reqsz);
	bfa_trc(bfad, ci->rspsz);
	bfa_trc(bfad, ci->fchs.d_id);
	bfa_trc(bfad, ci->fchs.s_id);
	bfa_trc(bfad, ci->nport_nexus_needed);

	drv_fcxp = bfad_fcxp_alloc(bfad);
	if (drv_fcxp == NULL) {
		bfa_trc(bfad, 0);
		return ENOMEM;
	}

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs, ci->vf_id, ci->lpwwn);
	if (fcs_port == NULL) {
		ci->status = BFA_STATUS_UNKNOWN_LWWN;
		bfa_trc(bfad, 0);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		goto out_free_fcxp;
	}

	bfa_trc(bfad, fcs_port->pid);

	drv_fcxp->port = bfa_fcs_port_get_drvport(fcs_port);

	/*
	 * TODO - drv_fcxp->port->bfad is null now.
	 */
	if (drv_fcxp->port->bfad == 0) {
		drv_fcxp->port->bfad = bfad;
		bfa_trc(bfad, 0);
	}

	if (ci->nport_nexus_needed) {
		fcs_rport = bfa_fcs_rport_lookup(fcs_port, ci->dpwwn);

		if (fcs_rport == NULL) {
			ci->status = BFA_STATUS_UNKNOWN_RWWN;
			bfa_trc(bfad, 0);
			spin_unlock_irqrestore(&bfad->bfad_lock, flags);
			goto out_free_fcxp;
		}

		hal_rport = bfa_fcs_rport_get_halrport(fcs_rport);
	} else
		drv_fcxp->hal_rport = NULL;

	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (ci->reqsz > BFA_FCXP_MAX_IBUF_SZ) {
		bfa_trc(bfad, 0);
		/*
		 * Allocate DMA memory for req
		 */
		drv_fcxp->reqbuf_info =
			bfad_fcxp_mem_alloc(bfad,
					    (char *)(size_t) ci->reqbuf_ptr,
					    ci->reqsz,
					    &drv_fcxp->num_req_sgles);
		if (!drv_fcxp->reqbuf_info) {
			bfa_trc(bfad, 0);
			rc = ENOMEM;
			goto out_free_fcxp;
		}
		drv_fcxp->req_sge =
			drv_fcxp->reqbuf_info + drv_fcxp->num_req_sgles;
	} else {
		bfa_trc(bfad, ci->reqsz);
		drv_fcxp->use_ireqbuf = 1;
	}

	if (ci->rspsz > BFA_FCXP_MAX_IBUF_SZ) {
		bfa_trc(bfad, 0);
		/*
		 * Allocate DMA memory for rsp
		 */
		drv_fcxp->rspbuf_info =
			bfad_fcxp_mem_alloc(bfad, NULL, ci->rspsz,
					    &drv_fcxp->num_rsp_sgles);
		if (!drv_fcxp->rspbuf_info) {
			bfad_fcxp_mem_free(bfad, drv_fcxp->reqbuf_info,
					   drv_fcxp->num_req_sgles);
			bfa_trc(bfad, 0);
			rc = ENOMEM;
			goto out_free_fcxp;
		}
		drv_fcxp->rsp_sge =
			drv_fcxp->rspbuf_info + drv_fcxp->num_rsp_sgles;
	} else {
		bfa_trc(bfad, ci->rspsz);
		drv_fcxp->use_irspbuf = 1;
	}

	init_completion(&drv_fcxp->comp);
	rc = bfad_fcxp_send(drv_fcxp, ci);
	if (rc == 0) {
		wait_for_completion(&drv_fcxp->comp);
		if (drv_fcxp->req_status == BFA_STATUS_OK) {
			if (drv_fcxp->use_irspbuf) {
				bfa_trc(bfad, 0);
				if (copy_to_user
				    ((void __user *)(size_t) ci->rspbuf_ptr,
				     (uint8_t *) drv_fcxp->rspbuf_info,
				     drv_fcxp->rsp_len)) {
					rc = EIO;
					goto out_mem_free;
				}
			} else {
				rc = bfad_rsp_data_copy(bfad,
							(char *)(size_t) ci->
							rspbuf_ptr,
							drv_fcxp->rspbuf_info,
							ci->rspsz);
				if (rc != 0) {
					rc = EIO;
					goto out_mem_free;
				}
			}
		}
		ci->status = drv_fcxp->req_status;
	}

out_mem_free:
	if (!drv_fcxp->use_ireqbuf)
		bfad_fcxp_mem_free(bfad, drv_fcxp->reqbuf_info,
				   drv_fcxp->num_req_sgles);
	if (!drv_fcxp->use_irspbuf)
		bfad_fcxp_mem_free(bfad, drv_fcxp->rspbuf_info,
				   drv_fcxp->num_rsp_sgles);

out_free_fcxp:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) ci, sizeof(bfa_ioctl_fcpt_t)))
		rc = EIO;

	bfad_fcxp_free(drv_fcxp);
	bfa_trc(bfad, 0);
	return rc;

}

/*
 * IOC
 */
int
bfad_ioctl_ioc_cfg_pport(unsigned long arg)
{
	bfa_ioctl_ioc_cfg_pport_t iocmd;
	bfad_t         *bfad;
	int             rc = 0;

	if (copy_from_user((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_gen_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	iocmd.status = bfad_cfg_pport(bfad, iocmd.role);

	if (copy_to_user((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_ioc_cfg_pport_t))) {
		rc = EIO;
		goto out;
	}

out:
	return rc;
}

int
bfad_ioctl_ioc_enable(unsigned long arg)
{
	bfa_ioctl_gen_t iocmd;
	bfad_t         *bfad;
	int             rc = 0;
	unsigned long   flags;

	if (copy_from_user((uint8_t *) &iocmd,
				(void __user *)arg, sizeof(iocmd)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status = bfa_hal_ioc_enable(&bfad->bfa_hal);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd, sizeof(iocmd))) {
		rc = EIO;
		goto out;
	}

out:
	return rc;
}

int
bfad_ioctl_ioc_disable(unsigned long arg)
{
	bfa_ioctl_gen_t iocmd;
	bfad_t         *bfad;
	int             rc = 0;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg, sizeof(iocmd)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	if (bfad->disable_active) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		return EBUSY;
	}

	bfad->disable_active = BFA_TRUE;
	init_completion(&bfad->disable_comp);
	bfa_hal_ioc_disable(&bfad->bfa_hal);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	wait_for_completion(&bfad->disable_comp);
	bfad->disable_active = BFA_FALSE;

	iocmd.status = BFA_STATUS_OK;
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd, sizeof(iocmd))) {
		rc = EIO;
		goto out;
	}

out:
	return rc;
}

/*
 * TODO - one time init
 */
static void
bfad_set_driver_attr(bfad_t *bfad, bfa_ioc_driver_attr_t *driver_attr)
{
	strcpy(driver_attr->driver, BFAD_DRIVER_NAME);
	strncpy(driver_attr->driver_ver, bfa_version, BFA_VERSION_LEN);
}

int
bfad_ioctl_ioc_get_attr(unsigned long arg)
{
	bfa_ioctl_ioc_attr_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_ioc_attr_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	memset(&iocmd.ioc_attr, 0, sizeof(bfa_ioc_attr_t));
	iocmd.status = BFA_STATUS_OK;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	bfa_hal_ioc_get_attr(&bfad->bfa_hal, &iocmd.ioc_attr);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	bfad_set_driver_attr(bfad, &iocmd.ioc_attr.driver_attr);

	/*
	 * TODO - replace this with version number from the firmware compiled as
	 * part of the driver
	 */
	strcpy(iocmd.ioc_attr.driver_attr.fw_ver,
	       iocmd.ioc_attr.adapter_attr.fw_ver);

	memcpy(&iocmd.ioc_attr.pci_attr, &bfad->pci_attr,
	       sizeof(bfa_ioc_pci_attr_t));

	/*
	 * TODO
	 * Set intr attr
	 */

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_ioc_attr_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_ioc_get_info(unsigned long arg)
{
	int             i;
	bfa_ioctl_ioc_info_t iocmd;
	union attr {
		bfa_pport_attr_t pattr;
		bfa_ioc_attr_t  ioc_attr;
	} attr;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_ioc_info_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	memset(&iocmd, 0, sizeof(iocmd));

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	bfa_hal_pport_get_attr(&bfad->bfa_hal, &attr.pattr);
	iocmd.nwwn = attr.pattr.nwwn;
	iocmd.pwwn = attr.pattr.pwwn;
	memset(&attr.ioc_attr, 0, sizeof(bfa_ioc_attr_t));
	bfa_hal_ioc_get_attr(&bfad->bfa_hal, &attr.ioc_attr);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	strcpy(iocmd.serialnum, attr.ioc_attr.adapter_attr.serial_num);
	strcpy(iocmd.name, bfad->adapter_name);
	strcpy(iocmd.port_name, bfad->port_name);
	strcpy(iocmd.hwpath, bfad->pci_name);
	/*
	 * set adapter hw path
	 */
	strcpy(iocmd.adapter_hwpath, bfad->pci_name);
	i = strlen(iocmd.adapter_hwpath) - 1;
	while (iocmd.adapter_hwpath[i] != '.')
		i--;
	iocmd.adapter_hwpath[i] = '\0';

	iocmd.status = BFA_STATUS_OK;

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_ioc_info_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_ioc_get_stats(unsigned long arg)
{
	bfa_ioctl_ioc_stats_t iocmd;
	bfad_t         *bfad;
	struct bfad_hal_comp fcomp;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_ioc_stats_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);

	if (bfad == NULL)
		return EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status =
		bfa_hal_ioc_get_stats(&bfad->bfa_hal, &iocmd.ioc_stats,
				      bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (iocmd.status != BFA_STATUS_OK) {
		bfa_trc(bfad, iocmd.status);
		goto out;
	}

	wait_for_completion(&fcomp.comp);
	iocmd.status = fcomp.status;

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_ioc_stats_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_ioc_reset_stats(unsigned long arg)
{
	bfa_ioctl_gen_t iocmd;
	bfad_t         *bfad;
	struct bfad_hal_comp fcomp;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg, sizeof(iocmd)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);

	if (bfad == NULL)
		return EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status =
		bfa_hal_ioc_clear_stats(&bfad->bfa_hal, bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (iocmd.status != BFA_STATUS_OK) {
		bfa_trc(bfad, iocmd.status);
		goto out;
	}

	wait_for_completion(&fcomp.comp);
	iocmd.status = fcomp.status;

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd, sizeof(iocmd)))
		return EIO;

	return 0;
}

int
bfad_ioctl_ioc_set_intr(unsigned long arg)
{
	bfa_ioctl_ioc_intr_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_ioc_intr_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);

	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status = bfa_hal_intr_attr_set(&bfad->bfa_hal, &iocmd.attr);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_ioc_intr_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_ioc_set_power(unsigned long arg)
{
	bfa_ioctl_ioc_power_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_ioc_power_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	/*
	 * TODO - ...
	 */
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	bfa_hal_pm_ctl(&bfad->bfa_hal, iocmd.power);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd.status = BFA_STATUS_CMD_NOTSUPP;

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_ioc_power_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_ioc_get_power(unsigned long arg)
{
	bfa_ioctl_ioc_power_t iocmd;
	bfad_t         *bfad;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_ioc_power_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	/*
	 * TODO : API needed to get power level
	 */
	iocmd.status = BFA_STATUS_CMD_NOTSUPP;

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_ioc_power_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_ioc_set_adapter_name(unsigned long arg)
{
	bfa_ioctl_ioc_set_adapter_name_t iocmd;
	bfad_t         *bfad;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_ioc_set_adapter_name_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL || PCI_FUNC(bfad->pcidev->devfn) != 0)
		return EINVAL;

	strcpy(bfad->adapter_name, iocmd.name);
	iocmd.status = BFA_STATUS_OK;

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_ioc_set_adapter_name_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_ioc_set_port_name(unsigned long arg)
{
	/*
	 * TBD: ioctl name need to changed
	 */
	bfa_ioctl_ioc_set_adapter_name_t iocmd;
	bfad_t         *bfad;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_ioc_set_adapter_name_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	strcpy(bfad->port_name, iocmd.name);
	iocmd.status = BFA_STATUS_OK;

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_ioc_set_adapter_name_t)))
		return EIO;

	return 0;
}

/*
 * PORT
 */
int
bfad_ioctl_port_enable(unsigned long arg)
{
	bfa_ioctl_gen_t iocmd;
	bfad_t         *bfad;
	unsigned long flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_gen_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status = bfa_hal_pport_enable(&bfad->bfa_hal);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_gen_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_port_disable(unsigned long arg)
{
	bfa_ioctl_gen_t iocmd;
	bfad_t         *bfad;
	unsigned long flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_gen_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status = bfa_hal_pport_disable(&bfad->bfa_hal);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_gen_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_port_get_attr(unsigned long arg)
{
	bfa_ioctl_port_attr_t iocmd;
	bfa_port_attr_t port_attr;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_port_attr_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	bfa_hal_pport_get_attr(&bfad->bfa_hal, &iocmd.attr);
	bfa_fcs_port_get_attr(&bfad->bfa_fcs.fabric.bport, &port_attr);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (iocmd.attr.topology != BFA_PPORT_TOPOLOGY_NONE)
		iocmd.attr.pid = port_attr.pid;
	else
		iocmd.attr.pid = 0;

	iocmd.attr.port_type = port_attr.port_type;
	iocmd.attr.loopback = port_attr.loopback;
	strncpy(iocmd.attr.port_symname.symname, 
			port_attr.port_cfg.sym_name.symname,
			sizeof(port_attr.port_cfg.sym_name.symname));

	iocmd.status = BFA_STATUS_OK;
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_port_attr_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_port_get_stats(unsigned long arg)
{
	bfa_ioctl_port_stats_t iocmd;
	struct bfad_hal_comp fcomp;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_port_stats_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status =
		bfa_hal_pport_get_stats(&bfad->bfa_hal, &iocmd.stats,
					bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd.status != BFA_STATUS_OK) {
		bfa_trc(bfad, iocmd.status);
		goto out;
	}

	wait_for_completion(&fcomp.comp);
	iocmd.status = fcomp.status;

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_port_stats_t))) {
		bfa_trc(bfad, 0);
		return EIO;
	}

	return 0;
}

int
bfad_ioctl_port_reset_stats(unsigned long arg)
{
	bfa_ioctl_gen_t iocmd;
	bfad_t         *bfad;
	struct bfad_hal_comp fcomp;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_gen_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status =
		bfa_hal_pport_clear_stats(&bfad->bfa_hal, bfad_hcb_comp,
					  &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (iocmd.status != BFA_STATUS_OK) {
		bfa_trc(bfad, iocmd.status);
		goto out;
	}

	wait_for_completion(&fcomp.comp);
	iocmd.status = fcomp.status;

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_gen_t))) {
		bfa_trc(bfad, 0);
		return EIO;
	}

	return 0;
}

int
bfad_ioctl_port_cfg_topology(unsigned long arg)
{
	bfa_ioctl_port_cfg_topo_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_port_cfg_topo_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status = bfa_hal_pport_cfg_topology(&bfad->bfa_hal, iocmd.topo);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_port_cfg_topo_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_port_cfg_speed(unsigned long arg)
{
	bfa_ioctl_port_cfg_speed_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_port_cfg_speed_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status = bfa_hal_pport_cfg_speed(&bfad->bfa_hal, iocmd.speed);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_port_cfg_speed_t))) {
		bfa_trc(bfad, 0);
		return EIO;
	}

	return 0;
}

int
bfad_ioctl_port_cfg_hardalpa(unsigned long arg)
{
	bfa_ioctl_port_alpa_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_port_alpa_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status = bfa_hal_pport_cfg_hardalpa(&bfad->bfa_hal, iocmd.alpa);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_port_alpa_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_port_clr_hardalpa(unsigned long arg)
{
	bfa_ioctl_port_alpa_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_port_alpa_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status = bfa_hal_pport_clr_hardalpa(&bfad->bfa_hal);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_port_alpa_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_port_cfg_maxfrsize(unsigned long arg)
{
	bfa_ioctl_port_cfg_maxfrsize_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_port_cfg_maxfrsize_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status =
		bfa_hal_pport_cfg_maxfrsize(&bfad->bfa_hal, iocmd.maxfrsize);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_port_cfg_maxfrsize_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_port_trunk_enable(unsigned long arg)
{
	bfa_ioctl_port_trunk_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_port_trunk_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status =
		bfa_hal_pport_trunk_enable(&bfad->bfa_hal, iocmd.pp_bmap);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_port_trunk_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_port_trunk_disable(unsigned long arg)
{
	bfa_ioctl_port_trunk_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_port_trunk_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status = bfa_hal_pport_trunk_disable(&bfad->bfa_hal);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_port_trunk_t)))
		return EIO;

	return 0;
}

/*
 * LPORT
 */
int
bfad_ioctl_lport_get_attr(unsigned long arg)
{
	bfa_fcs_port_t *fcs_port;
	bfa_ioctl_lport_attr_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_lport_attr_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs, iocmd.vf_id, iocmd.pwwn);
	if (fcs_port == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd.status = BFA_STATUS_UNKNOWN_LWWN;
		goto out;
	}

	bfa_fcs_port_get_attr(fcs_port, &iocmd.port_attr);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd.status = BFA_STATUS_OK;

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_lport_attr_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_lport_reset_stats(unsigned long arg)
{
	bfa_fcs_port_t *fcs_port;
	bfa_ioctl_lport_reset_stats_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_lport_reset_stats_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs, iocmd.vf_id, iocmd.pwwn);
	if (fcs_port == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd.status = BFA_STATUS_UNKNOWN_LWWN;
		goto out;
	}

	bfa_fcs_port_clear_stats(fcs_port);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd.status = BFA_STATUS_OK;

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_lport_reset_stats_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_lport_get_stats(unsigned long arg)
{
	bfa_fcs_port_t *fcs_port;
	bfa_ioctl_lport_stats_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_lport_stats_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs, iocmd.vf_id, iocmd.pwwn);
	if (fcs_port == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd.status = BFA_STATUS_UNKNOWN_LWWN;
		goto out;
	}

	bfa_fcs_port_get_stats(fcs_port, &iocmd.port_stats);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd.status = BFA_STATUS_OK;

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_lport_stats_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_lport_get_rports(unsigned long arg)
{
	bfa_ioctl_lport_get_rports_t iocmd;
	wwn_t          *rports = NULL;
	bfa_fcs_port_t *fcs_port;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_lport_get_rports_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	if (iocmd.nrports == 0) {
		bfa_trc(bfad, 0);
		return EINVAL;
	}

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs, iocmd.vf_id, iocmd.pwwn);
	if (fcs_port == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		bfa_trc(bfad, 0);
		iocmd.status = BFA_STATUS_UNKNOWN_LWWN;
		goto out;
	}

	rports = kzalloc(sizeof(wwn_t) * iocmd.nrports, GFP_ATOMIC);
	if (rports == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		bfa_trc(bfad, 0);
		iocmd.status = BFA_STATUS_ENOMEM;
		goto out;
	}

	bfa_fcs_port_get_rports(fcs_port, rports, &iocmd.nrports);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (iocmd.nrports
	    && copy_to_user((void __user *)(size_t) iocmd.rbuf_ptr,
			    (uint8_t *) rports,
			    iocmd.nrports * sizeof(wwn_t))) {
		kfree(rports);
		return EIO;
	}

	kfree(rports);
	iocmd.status = BFA_STATUS_OK;

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_lport_get_rports_t)))
		return EIO;

	return 0;
}

/*
 * RPORT
 */
int
bfad_ioctl_rport_get_attr(unsigned long arg)
{
	bfa_ioctl_rport_attr_t iocmd;
	bfa_fcs_port_t *fcs_port;
	bfa_fcs_rport_t *fcs_rport;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_rport_attr_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs, iocmd.vf_id, iocmd.pwwn);
	if (fcs_port == NULL) {
		bfa_trc(bfad, 0);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd.status = BFA_STATUS_UNKNOWN_LWWN;
		goto out;
	}

	fcs_rport = bfa_fcs_rport_lookup(fcs_port, iocmd.rpwwn);
	if (fcs_rport == NULL) {
		bfa_trc(bfad, 0);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd.status = BFA_STATUS_UNKNOWN_RWWN;
		goto out;
	}

	bfa_fcs_rport_get_attr(fcs_rport, &iocmd.attr);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd.status = BFA_STATUS_OK;

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_rport_attr_t))) {
		bfa_trc(bfad, 0);
		return EIO;
	}

	return 0;
}


int
bfad_ioctl_rport_get_addr(unsigned long arg)
{
	bfa_ioctl_rport_scsi_addr_t iocmd;
	bfa_fcs_port_t *fcs_port;
	bfa_fcs_itnim_t *fcs_itnim;
	bfad_itnim_t *drv_itnim;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg, sizeof(iocmd)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs, iocmd.vf_id, iocmd.pwwn);
	if (fcs_port == NULL) {
		bfa_trc(bfad, 0);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd.status = BFA_STATUS_UNKNOWN_LWWN;
		goto out;
	}

	fcs_itnim = bfa_fcs_itnim_lookup(fcs_port, iocmd.rpwwn);
	if (fcs_itnim == NULL) {
		bfa_trc(bfad, 0);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd.status = BFA_STATUS_UNKNOWN_RWWN;
		goto out;
	}

	drv_itnim = fcs_itnim->itnim_drv;

	if (drv_itnim && drv_itnim->im_port) {
#ifdef __VMKERNEL_MODULE__
		int ret;
		char *str = (char *)&drv_itnim->im_port->shost->adapter->name[5];
		ret=get_option(&str, &iocmd.host);
		bfa_trc(bfad, iocmd.host);
		bfa_trc(bfad, ret);
		if (ret == 0) {
			spin_unlock_irqrestore(&bfad->bfad_lock, flags);
			iocmd.status = BFA_STATUS_UNKNOWN_RWWN;
			goto out;
		}
#else
		iocmd.host = drv_itnim->im_port->shost->host_no;
#endif
	} else {
		bfa_trc(bfad, 0);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd.status = BFA_STATUS_UNKNOWN_RWWN;
		goto out;
	}

	iocmd.target = drv_itnim->scsi_tgt_id;
		
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	iocmd.bus = 0;
	iocmd.lun = 0;
	iocmd.status = BFA_STATUS_OK;

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd, sizeof(iocmd)))
		return EIO;

	return 0;
}

int
bfad_ioctl_rport_get_stats(unsigned long arg)
{
	bfa_ioctl_rport_stats_t iocmd;
	bfa_fcs_port_t *fcs_port;
	bfa_fcs_rport_t *fcs_rport;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_rport_stats_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs, iocmd.vf_id, iocmd.pwwn);
	if (fcs_port == NULL) {
		bfa_trc(bfad, 0);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd.status = BFA_STATUS_UNKNOWN_LWWN;
		goto out;
	}

	fcs_rport = bfa_fcs_rport_lookup(fcs_port, iocmd.rpwwn);
	if (fcs_rport == NULL) {
		bfa_trc(bfad, 0);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd.status = BFA_STATUS_UNKNOWN_RWWN;
		goto out;
	}

	bfa_fcs_rport_get_stats(fcs_rport, &iocmd.stats);
	bfa_hal_rport_get_stats(bfa_fcs_rport_get_halrport(fcs_rport),
				&iocmd.stats.hal_stats);

	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd.status = BFA_STATUS_OK;

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_rport_stats_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_rport_clr_stats(unsigned long arg)
{
	bfa_ioctl_rport_reset_stats_t iocmd;
	bfa_fcs_port_t *fcs_port;
	bfa_fcs_rport_t *fcs_rport;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_rport_reset_stats_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs, iocmd.vf_id, iocmd.pwwn);
	if (fcs_port == NULL) {
		bfa_trc(bfad, 0);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd.status = BFA_STATUS_UNKNOWN_LWWN;
		goto out;
	}

	fcs_rport = bfa_fcs_rport_lookup(fcs_port, iocmd.rpwwn);
	if (fcs_rport == NULL) {
		bfa_trc(bfad, 0);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd.status = BFA_STATUS_UNKNOWN_RWWN;
		goto out;
	}

	bfa_fcs_rport_clear_stats(fcs_rport);
	bfa_hal_rport_clear_stats(bfa_fcs_rport_get_halrport(fcs_rport));

	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd.status = BFA_STATUS_OK;

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_rport_reset_stats_t)))
		return EIO;

	return 0;
}

/*
 * VPORT
 */
int
bfad_ioctl_vport_create(unsigned long arg)
{
	bfa_ioctl_vport_create_t iocmd;
	bfad_t         *bfad;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_vport_create_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	bfa_trc(bfad, iocmd.port_cfg.pwwn);
	bfa_trc(bfad, iocmd.port_cfg.nwwn);
	iocmd.status = bfad_vport_create(bfad, iocmd.vf_id, &iocmd.port_cfg);

	if (iocmd.status != BFA_STATUS_OK) {
		BFA_DEV_PRINTF(bfad, BFA_ERR,
			       "Failed to create vport "
			       "wwnn %llx wwpn %llx for vf %d\n",
			       (long long)iocmd.port_cfg.nwwn,
			       (long long)iocmd.port_cfg.pwwn, iocmd.vf_id);
	} else {
		bfad_os_sysfs_lport_add_dir(bfad, iocmd.vf_id, bfad->inst_no,
					    iocmd.port_cfg);
	}

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_vport_create_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_vport_del(unsigned long arg)
{
	bfa_ioctl_vport_t iocmd;
	bfad_vport_t   *bfad_vport;
	bfa_fcs_vport_t *fcs_vport;
	unsigned long   flags;
	bfad_t         *bfad;
	struct completion fcomp;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_vport_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_vport =
		bfa_fcs_vport_lookup(&bfad->bfa_fcs, iocmd.vf_id, iocmd.vpwwn);
	if (fcs_vport == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd.status = BFA_STATUS_UNKNOWN_VWWN;
		goto out;
	}

	bfad_vport = fcs_vport->vport_drv;

	bfad_vport->comp_del = &fcomp;
	init_completion(bfad_vport->comp_del);

	iocmd.status = bfa_fcs_vport_delete(fcs_vport);
	bfa_trc(bfad, iocmd.status);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	wait_for_completion(bfad_vport->comp_del);

	if (iocmd.status != BFA_STATUS_OK) {
		BFA_DEV_PRINTF(bfad, BFA_ERR,
			       "Failed to delete vport "
			       "wwpn %llx for vf %d\n", (long long)iocmd.vpwwn,
			       iocmd.vf_id);
	} else {
		bfad_os_sysfs_lport_delete_dir(bfad, iocmd.vf_id, bfad->inst_no,
					       iocmd.vpwwn);
	}

	bfa_trc(bfad, bfad->inst_no);
	bfad_os_scsi_host_free(bfad, bfad_vport->drv_port.im_port);
	kfree(bfad_vport);

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_vport_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_vport_start(unsigned long arg)
{
	bfa_ioctl_vport_t iocmd;
	bfa_fcs_vport_t *fcs_vport;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_vport_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_vport =
		bfa_fcs_vport_lookup(&bfad->bfa_fcs, iocmd.vf_id, iocmd.vpwwn);
	if (fcs_vport == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd.status = BFA_STATUS_UNKNOWN_VWWN;
		goto out;
	}

	iocmd.status = bfa_fcs_vport_start(fcs_vport);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_vport_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_vport_stop(unsigned long arg)
{
	bfa_ioctl_vport_t iocmd;
	bfa_fcs_vport_t *fcs_vport;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_vport_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_vport =
		bfa_fcs_vport_lookup(&bfad->bfa_fcs, iocmd.vf_id, iocmd.vpwwn);
	if (fcs_vport == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd.status = BFA_STATUS_UNKNOWN_VWWN;
		goto out;
	}

	iocmd.status = bfa_fcs_vport_stop(fcs_vport);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_vport_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_vport_get_attr(unsigned long arg)
{
	bfa_fcs_vport_t *fcs_vport;
	bfa_ioctl_vport_attr_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_vport_attr_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_vport =
		bfa_fcs_vport_lookup(&bfad->bfa_fcs, iocmd.vf_id, iocmd.vpwwn);
	if (fcs_vport == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd.status = BFA_STATUS_UNKNOWN_VWWN;
		goto out;
	}

	bfa_fcs_vport_get_attr(fcs_vport, &iocmd.vport_attr);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd.status = BFA_STATUS_OK;

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_vport_attr_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_vport_get_stats(unsigned long arg)
{
	bfa_fcs_vport_t *fcs_vport;
	bfa_ioctl_vport_stats_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_vport_stats_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_vport =
		bfa_fcs_vport_lookup(&bfad->bfa_fcs, iocmd.vf_id, iocmd.vpwwn);
	if (fcs_vport == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd.status = BFA_STATUS_UNKNOWN_VWWN;
		goto out;
	}

	bfa_fcs_vport_get_stats(fcs_vport, &iocmd.vport_stats);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd.status = BFA_STATUS_OK;

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_vport_stats_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_vport_clr_stats(unsigned long arg)
{
	bfa_fcs_vport_t *fcs_vport;
	bfa_ioctl_vport_reset_stats_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_vport_reset_stats_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_vport =
		bfa_fcs_vport_lookup(&bfad->bfa_fcs, iocmd.vf_id, iocmd.vpwwn);
	if (fcs_vport == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd.status = BFA_STATUS_UNKNOWN_VWWN;
		goto out;
	}

	bfa_fcs_vport_clr_stats(fcs_vport);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd.status = BFA_STATUS_OK;

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_vport_reset_stats_t)))
		return EIO;

	return 0;
}

/*
 * VF
 */
int
bfad_ioctl_vf_mode_enable(unsigned long arg)
{
	bfa_ioctl_vf_t  iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg, sizeof(bfa_ioctl_vf_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	/*
	 * TBD: need to update API to add vf_id
	 */
	/*
	 * iocmd.status = bfa_fcs_vf_mode_enable(&bfad->bfa_fcs, iocmd.vf_id);
	 */
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status = bfa_fcs_vf_mode_enable(&bfad->bfa_fcs, iocmd.vf_id);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd, sizeof(bfa_ioctl_vf_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_vf_mode_disable(unsigned long arg)
{
	bfa_ioctl_vf_t  iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg, sizeof(bfa_ioctl_vf_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status = bfa_fcs_vf_mode_disable(&bfad->bfa_fcs);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd, sizeof(bfa_ioctl_vf_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_vf_add(unsigned long arg)
{
	bfa_ioctl_vf_add_t iocmd;
	bfad_t         *bfad;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_vf_add_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	iocmd.status = bfad_vf_create(bfad, iocmd.vf_id, &iocmd.port_cfg);

	if (iocmd.status != BFA_STATUS_OK) {
		BFA_DEV_PRINTF(bfad, BFA_ERR, "Failed to create vf %d",
			       iocmd.vf_id);
	} else {
		/*
		 * create a sysfs directory entry for the new VF
		 */
		bfad_os_sysfs_vf_add_dir(bfad, bfad->inst_no, iocmd.vf_id);
	}

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_vf_add_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_vf_del(unsigned long arg)
{
	bfa_fcs_vf_t   *fcs_vf;
	bfa_ioctl_vf_t  iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg, sizeof(bfa_ioctl_vf_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_vf = bfa_fcs_vf_lookup(&bfad->bfa_fcs, (uint16_t) iocmd.vf_id);

	if (fcs_vf == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd.status = BFA_STATUS_UNKNOWN_VFID;
		goto out;
	}

	iocmd.status = bfa_fcs_vf_delete(fcs_vf);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (iocmd.status != BFA_STATUS_OK) {
		BFA_DEV_PRINTF(bfad, BFA_ERR, "Failed to delete vf %d",
			       iocmd.vf_id);
	} else {
		/*
		 * delete the sysfs directory entry for the deleted VF
		 */
		bfad_os_sysfs_vf_delete_dir(bfad, iocmd.vf_id);
	}

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd, sizeof(bfa_ioctl_vf_t)))
		return EIO;

	return 0;

}

int
bfad_ioctl_vf_list(unsigned int cmd, unsigned long arg)
{
	bfa_ioctl_vf_list_t iocmd;
	uint16_t       *vf_ids;
	int             nvfs;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_vf_list_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	vf_ids = kzalloc(sizeof(uint16_t) * iocmd.nvfs, GFP_KERNEL);
	if (vf_ids == NULL)
		return ENOMEM;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	if (cmd == IOCMD_VF_LIST)
		bfa_fcs_vf_list(&bfad->bfa_fcs, vf_ids, &nvfs);
	else
		bfa_fcs_vf_list_all(&bfad->bfa_fcs, vf_ids, &nvfs);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	iocmd.nvfs = nvfs;

	if (copy_to_user
	    ((void __user *)(size_t) iocmd.buf_ptr, (uint8_t *) vf_ids,
	     nvfs * sizeof(int))) {
		kfree(vf_ids);
		return EIO;
	}

	kfree(vf_ids);

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_vf_list_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_vf_get_attr(unsigned long arg)
{
	bfa_ioctl_vf_attr_t iocmd;
	bfa_fcs_vf_t   *fcs_vf;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_vf_attr_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	/*
	 * TBD: need to have this BFA API
	 */
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_vf = bfa_fcs_vf_lookup(&bfad->bfa_fcs, iocmd.vf_id);
	if (fcs_vf == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd.status = BFA_STATUS_UNKNOWN_VFID;
		goto out;
	}

	bfa_fcs_vf_get_attr(fcs_vf, &iocmd.attr);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd.status = BFA_STATUS_OK;

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_vf_attr_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_vf_get_stats(unsigned long arg)
{
	bfa_ioctl_vf_stats_t iocmd;
	bfa_fcs_vf_t   *fcs_vf;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_vf_stats_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_vf = bfa_fcs_vf_lookup(&bfad->bfa_fcs, iocmd.vf_id);
	if (fcs_vf == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd.status = BFA_STATUS_UNKNOWN_VFID;
		goto out;
	}

	bfa_fcs_vf_get_stats(fcs_vf, &iocmd.stats);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd.status = BFA_STATUS_OK;

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_vf_stats_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_vf_clr_stats(unsigned long arg)
{
	bfa_ioctl_vf_reset_stats_t iocmd;
	bfa_fcs_vf_t   *fcs_vf;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_vf_reset_stats_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_vf = bfa_fcs_vf_lookup(&bfad->bfa_fcs, iocmd.vf_id);
	if (fcs_vf == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd.status = BFA_STATUS_UNKNOWN_VFID;
		goto out;
	}

	bfa_fcs_vf_clear_stats(fcs_vf);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd.status = BFA_STATUS_OK;

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_vf_reset_stats_t)))
		return EIO;

	return 0;
}

/*
 * FABRIC
 */
int
bfad_ioctl_fabric_get_lports(unsigned long arg)
{
	bfa_ioctl_fabric_get_lports_t iocmd;
	bfa_fcs_vf_t   *fcs_vf;
	wwn_t          *ports;
	uint32_t        nports;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_fabric_get_lports_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	bfa_trc(bfad, iocmd.nports);
	bfa_trc(bfad, iocmd.vf_id);

	if (iocmd.nports == 0) {
		iocmd.status = BFA_STATUS_EINVAL;
		goto out;
	}
	nports = iocmd.nports;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_vf = bfa_fcs_vf_lookup(&bfad->bfa_fcs, iocmd.vf_id);
	if (fcs_vf == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd.status = BFA_STATUS_UNKNOWN_VFID;
		goto out;
	}

	ports = kzalloc(sizeof(wwn_t[iocmd.nports]), GFP_ATOMIC);

	if (ports == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		bfa_trc(bfad, 0);
		return ENOMEM;
	}

	bfa_fcs_vf_get_ports(fcs_vf, ports, &nports);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	iocmd.nports = nports;
	bfa_trc(bfad, iocmd.nports);

	if (copy_to_user
	    ((void __user *)(size_t) iocmd.buf_ptr, (uint8_t *) ports,
	     nports * sizeof(wwn_t))) {
		bfa_trc(bfad, 0);
		kfree(ports);
		return EIO;
	}

	iocmd.status = BFA_STATUS_OK;
	kfree(ports);

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_fabric_get_lports_t)))
		return EIO;

	return 0;
}

/*
 * FLASH
 */
int
bfad_ioctl_flash_get_attr(unsigned long arg)
{
	bfa_ioctl_flash_attr_t iocmd;
	struct bfad_hal_comp fcomp;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_flash_attr_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status =
		bfa_hal_flash_get_attr(&bfad->bfa_hal, &iocmd.attr,
				       bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd.status != BFA_STATUS_OK)
		goto out;

	wait_for_completion(&fcomp.comp);
	iocmd.status = fcomp.status;

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_flash_attr_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_flash_read_part(unsigned long arg)
{
	bfa_ioctl_flash_t iocmd;
	struct bfad_hal_comp fcomp;
	void           *buf;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_flash_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	buf = vmalloc(iocmd.bufsz);
	if (!buf) {
		bfa_trc(bfad, 0x4444);
		return ENOMEM;
	}

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status =
		bfa_hal_flash_read_part(&bfad->bfa_hal, iocmd.type,
					iocmd.instance, buf, iocmd.bufsz, 0,
					bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd.status != BFA_STATUS_OK)
		goto out;

	wait_for_completion(&fcomp.comp);
	iocmd.status = fcomp.status;
	if (iocmd.status != BFA_STATUS_OK)
		goto out;

	if (copy_to_user
	    ((void __user *)(size_t) iocmd.buf_ptr, (uint8_t *) buf,
	     iocmd.bufsz)) {
		vfree(buf);
		return EIO;
	}

out:
	vfree(buf);
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_flash_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_flash_erase_part(unsigned long arg)
{
	bfa_ioctl_flash_t iocmd;
	struct bfad_hal_comp fcomp;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_flash_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status =
		bfa_hal_flash_erase_part(&bfad->bfa_hal, iocmd.type,
					 iocmd.instance, bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd.status != BFA_STATUS_OK)
		goto out;

	wait_for_completion(&fcomp.comp);
	iocmd.status = fcomp.status;

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_flash_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_flash_update_part(unsigned long arg)
{
	bfa_ioctl_flash_t iocmd;
	void           *buffer;
	struct bfad_hal_comp fcomp;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_flash_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	buffer = vmalloc(iocmd.bufsz);
	if (buffer == NULL)
		return ENOMEM;

	if (copy_from_user
	    ((uint8_t *) buffer, (void __user *)(size_t) iocmd.buf_ptr,
	     iocmd.bufsz)) {
		vfree(buffer);
		return EIO;
	}

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status =
		bfa_hal_flash_update_part(&bfad->bfa_hal, iocmd.type,
					  iocmd.instance, buffer, iocmd.bufsz,
					  0, bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd.status != BFA_STATUS_OK)
		goto out;

	wait_for_completion(&fcomp.comp);
	iocmd.status = fcomp.status;

out:
	vfree(buffer);
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_flash_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_flash_mfg_info(unsigned long arg)
{
	bfa_ioctl_flash_mfg_t iocmd;
	struct bfad_hal_comp fcomp;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_flash_mfg_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status =
		bfa_hal_flash_update_part(&bfad->bfa_hal, BFA_FLASH_PART_MFG, 0,
					  &iocmd.mfgblk,
					  sizeof(bfa_mfg_block_t), 0,
					  bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd.status != BFA_STATUS_OK)
		goto out;

	wait_for_completion(&fcomp.comp);
	iocmd.status = fcomp.status;

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_flash_mfg_t)))
		return EIO;

	return 0;
}

/*
 * DIAG
 */
int
bfad_ioctl_diag_regrd(unsigned long arg)
{
	bfa_ioctl_diag_regread_t iocmd;
	uint32_t       *buf;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_diag_regread_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	bfa_trc(bfad, 0x8888);

	buf = kzalloc(iocmd.nwords << 2, GFP_KERNEL);
	if (!buf)
		return ENOMEM;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status =
		bfa_hal_diag_reg_read(&bfad->bfa_hal, iocmd.off, iocmd.nwords,
				      buf);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	bfa_trc(bfad, iocmd.status);
	if (iocmd.status != BFA_STATUS_OK)
		goto out;

	if (copy_to_user
	    ((void __user *)(size_t) iocmd.buf_ptr, (uint8_t *) buf,
	     iocmd.nwords << 2)) {
		kfree(buf);
		return EIO;
	}

out:
	kfree(buf);

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_diag_regread_t)))
		return EIO;

	bfa_trc(bfad, 0x88881);
	return 0;
}

int
bfad_ioctl_diag_regwr(unsigned long arg)
{
	bfa_ioctl_diag_regwrite_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_diag_regwrite_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status =
		bfa_hal_diag_reg_write(&bfad->bfa_hal, iocmd.off, 1,
				       iocmd.word);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_diag_regwrite_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_diag_temp(unsigned long arg)
{
	bfa_ioctl_diag_get_temp_t iocmd;
	struct bfad_hal_comp fcomp;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_diag_get_temp_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status =
		bfa_hal_diag_tsensor_query(&bfad->bfa_hal, &iocmd.temp,
					   bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	bfa_trc(bfad, iocmd.status);
	if (iocmd.status != BFA_STATUS_OK)
		goto out;

	wait_for_completion(&fcomp.comp);
	iocmd.status = fcomp.status;
	bfa_trc(bfad, iocmd.status);

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_diag_get_temp_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_diag_post(unsigned long arg)
{
	return 0;
}

int
bfad_ioctl_diag_memtest(unsigned long arg)
{
	bfa_ioctl_diag_memtest_t iocmd;
	struct bfad_hal_comp fcomp;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_diag_memtest_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_diag_memtest_t)))
		return EIO;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status =
		bfa_hal_diag_memtest(&bfad->bfa_hal, &iocmd.memtest, iocmd.pat,
				     &iocmd.result, bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	bfa_trc(bfad, iocmd.status);
	if (iocmd.status != BFA_STATUS_OK)
		goto out;

	wait_for_completion(&fcomp.comp);
	iocmd.status = fcomp.status;
	bfa_trc(bfad, iocmd.status);

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_diag_memtest_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_diag_loopback(unsigned long arg)
{
	bfa_ioctl_diag_loopback_t iocmd;
	struct bfad_hal_comp fcomp;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_diag_loopback_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status =
		bfa_hal_diag_loopback(&bfad->bfa_hal, iocmd.opmode, iocmd.speed,
				      iocmd.lpcnt, iocmd.pat, &iocmd.result,
				      bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	bfa_trc(bfad, iocmd.status);
	if (iocmd.status != BFA_STATUS_OK)
		goto out;

	wait_for_completion(&fcomp.comp);
	iocmd.status = fcomp.status;
	bfa_trc(bfad, iocmd.status);

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_diag_loopback_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_diag_fwping(unsigned long arg)
{
	bfa_ioctl_diag_fwping_t iocmd;
	struct bfad_hal_comp fcomp;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_diag_fwping_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	bfa_trc(bfad, 0x7777);
	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status =
		bfa_hal_diag_fwping(&bfad->bfa_hal, iocmd.cnt, iocmd.pattern,
				    &iocmd.result, bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	bfa_trc(bfad, iocmd.status);
	if (iocmd.status != BFA_STATUS_OK)
		goto out;

	bfa_trc(bfad, 0x77771);
	wait_for_completion(&fcomp.comp);
	iocmd.status = fcomp.status;
	bfa_trc(bfad, iocmd.status);

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_diag_fwping_t)))
		return EIO;

	bfa_trc(bfad, 0x77772);
	return 0;
}

int
bfad_ioctl_diag_pom(unsigned long arg)
{
	bfa_ioctl_diag_pom_t iocmd;
	bfad_t         *bfad;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_diag_pom_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	/*
	 * TBD: iocmd.status = bfa_hal_diag_pom_query(&bfad->bfa_hal,
	 * &iocmd.pom_attr);
	 */

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_diag_pom_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_diag_sfp(unsigned long arg)
{
	bfa_ioctl_diag_sfp_t iocmd;
	struct bfad_hal_comp fcomp;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_diag_sfp_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status =
		bfa_hal_diag_sfpshow(&bfad->bfa_hal, &iocmd.sfp,
				     iocmd.static_data, bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	bfa_trc(bfad, iocmd.status);
	if (iocmd.status != BFA_STATUS_OK)
		goto out;

	wait_for_completion(&fcomp.comp);
	iocmd.status = fcomp.status;
	bfa_trc(bfad, iocmd.status);

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_diag_sfp_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_diag_led(unsigned long arg)
{
	bfa_ioctl_diag_led_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_diag_led_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status = bfa_hal_diag_ledtest(&bfad->bfa_hal, &iocmd.ledtest);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_diag_led_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_diag_beacon_lport(unsigned long arg)
{
	bfa_ioctl_diag_beacon_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_diag_beacon_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status =
		bfa_hal_diag_beacon_port(&bfad->bfa_hal, 
						iocmd.beacon, 
						iocmd.link_e2e_beacon,
						iocmd.second);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_diag_beacon_t)))
		return EIO;

	return 0;
}

/*
 * DEBUG
 */
int
bfad_ioctl_debug_drv_trace(unsigned long arg)
{
	bfa_ioctl_debug_t iocmd;
	bfad_t         *bfad;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_debug_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	if (iocmd.bufsz < sizeof(bfa_trc_mod_t)) {
		iocmd.status = BFA_STATUS_EINVAL;
		goto out;
	}

	if (copy_to_user
	    ((void __user *)(size_t) iocmd.buf_ptr, (uint8_t *) bfad->trcmod,
	     iocmd.bufsz))
		return EIO;

	iocmd.status = BFA_STATUS_OK;
out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_debug_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_debug_drv_state(unsigned long arg)
{
	bfa_ioctl_debug_t iocmd;
	bfad_t         *bfad;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_debug_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	/*
	 * TBD: we are not support for now
	 */
	iocmd.status = BFA_STATUS_FAILED;
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_debug_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_debug_fw_trace(unsigned long arg)
{
	bfa_ioctl_debug_t iocmd;
	void           *buf;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_debug_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	/*
	 * TBD: Verify the size of the buffer
	 */
	bfa_trc(bfad, iocmd.bufsz);
	if (iocmd.bufsz < sizeof(bfa_trc_mod_t)) {
		bfa_trc(bfad, sizeof(bfa_trc_mod_t));
		iocmd.status = BFA_STATUS_EINVAL;
		goto out;
	}

	buf = vmalloc(iocmd.bufsz);
	if (!buf) {
		bfa_trc(bfad, 0x4444);
		return ENOMEM;
	}

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status = bfa_hal_debug_fwtrc(&bfad->bfa_hal, buf, &iocmd.bufsz);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd.status != BFA_STATUS_OK) {
		bfa_trc(bfad, 0x5555);
		goto out_free_buf;
	}

	bfa_trc(bfad, iocmd.bufsz);
	if (copy_to_user
	    ((void __user *)(size_t) iocmd.buf_ptr, (uint8_t *) buf,
	     iocmd.bufsz)) {
		vfree(buf);
		bfa_trc(bfad, 0x55551);
		return EIO;
	}

out_free_buf:
	vfree(buf);
out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_debug_t)))
		return EIO;

	bfa_trc(bfad, 0x6666);

	return 0;
}

int
bfad_ioctl_debug_fw_state(unsigned long arg)
{
	return 0;
}

int
bfad_ioctl_porglog_get(unsigned long arg)
{
	bfa_ioctl_debug_t iocmd;
	bfad_t         *bfad;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_debug_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	/*
	 * TBD: Verify the size of the buffer
	 */
	bfa_trc(bfad, iocmd.bufsz);

	if (iocmd.bufsz < sizeof(bfa_plog_t)) {
		bfa_trc(bfad, sizeof(bfa_plog_t));
		iocmd.status = BFA_STATUS_EINVAL;
		goto out;
	}

	iocmd.status = BFA_STATUS_OK;

	if (copy_to_user
	    ((void __user *)(size_t) iocmd.buf_ptr,
	     (uint8_t *) &bfad->plog_buf, sizeof(bfa_plog_t))) {
		bfa_trc(bfad, 0);
		return EIO;
	}

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_debug_t))) {
		bfa_trc(bfad, 0);
		return EIO;
	}

	return 0;
}

int
bfad_ioctl_porglog_clear(unsigned long arg)
{
	bfa_ioctl_gen_t iocmd;
	bfad_t         *bfad;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_gen_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	bfa_plog_clear(&bfad->plog_buf);
	iocmd.status = BFA_STATUS_OK;

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_gen_t))) {
		bfa_trc(bfad, 0);
		return EIO;
	}

	return 0;
}

int
bfad_ioctl_porglog_ctl(unsigned long arg)
{
	bfa_ioctl_portlogctl_t iocmd;
	bfad_t         *bfad;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_portlogctl_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	if (iocmd.ctl == BFA_TRUE)
		bfa_plog_enable(&bfad->plog_buf);
	else
		bfa_plog_disable(&bfad->plog_buf);

	iocmd.status = BFA_STATUS_OK;

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_portlogctl_t))) {
		bfa_trc(bfad, 0);
		return EIO;
	}

	return 0;
}

int
bfad_ioctl_log_set_level(unsigned long arg)
{
	bfa_ioctl_log_t iocmd;
	bfad_t         *bfad;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_log_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	bfa_trc(bfad, iocmd.log_level);
	if (iocmd.mod_id == BFA_LOG_MODULE_ID_ALL) {
		iocmd.status = bfa_log_set_level_all(&bfad->log_data, 
						     iocmd.log_level);
	} else if (iocmd.mod_id == BFA_LOG_AEN_ALL) {
		iocmd.status = bfa_log_set_level_aen(&bfad->log_data, 
						     iocmd.log_level);
	} else {
		if (iocmd.mod_id == BFA_LOG_DRV_ALL) 
			iocmd.mod_id = BFA_LOG_LINUX_ID;
		iocmd.status = bfa_log_set_level(&bfad->log_data, 
					iocmd.mod_id, iocmd.log_level);
	}

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_log_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_log_get_level(unsigned long arg)
{
	bfa_ioctl_log_t iocmd;
	bfad_t         *bfad;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_log_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	if ((iocmd.mod_id <= BFA_LOG_UNUSED_ID) || 
	    (iocmd.mod_id > BFA_LOG_MODULE_ID_MAX)) 
		iocmd.mod_id = BFA_LOG_LINUX_ID;

	iocmd.log_level = bfa_log_get_level(&bfad->log_data, iocmd.mod_id);

	iocmd.status = BFA_STATUS_OK;
	bfa_trc(bfad, iocmd.log_level);

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_log_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_qos(unsigned long arg, unsigned int cmd)
{
	bfa_ioctl_gen_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_gen_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	if (cmd == IOCMD_QOS_ENABLE)
		bfa_hal_pport_cfg_qos(&bfad->bfa_hal, BFA_TRUE);
	else if (cmd == IOCMD_QOS_DISABLE)
		bfa_hal_pport_cfg_qos(&bfad->bfa_hal, BFA_FALSE);
	else {
		bfa_trc(bfad, 0);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		return EINVAL;
	}
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	iocmd.status = BFA_STATUS_OK;

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_gen_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_ratelim(unsigned long arg, unsigned int cmd)
{
	bfa_ioctl_gen_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_gen_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	if (cmd == IOCMD_RATELIM_ENABLE)
		bfa_hal_pport_cfg_ratelim(&bfad->bfa_hal, BFA_TRUE);
	else if (cmd == IOCMD_RATELIM_DISABLE)
		bfa_hal_pport_cfg_ratelim(&bfad->bfa_hal, BFA_FALSE);
	else {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		bfa_trc(bfad, 0);
		return EINVAL;
	}
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	iocmd.status = BFA_STATUS_OK;

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_gen_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_fcpim_failover(unsigned long arg)
{
	bfa_ioctl_fcpim_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_fcpim_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	bfa_trc(bfad, iocmd.mpio_mode);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	bfa_hal_fcpim_mpio_mode_set(&bfad->bfa_hal, iocmd.mpio_mode, iocmd.path_tov);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	iocmd.status = BFA_STATUS_OK;

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_fcpim_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_fcpim_qdepth(unsigned long arg)
{
	bfa_ioctl_fcpim_qdepth_t	iocmd;
	bfad_t         			*bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_fcpim_qdepth_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	bfa_trc(bfad, iocmd.q_depth);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	bfa_hal_fcpim_qdepth_set(&bfad->bfa_hal, iocmd.q_depth);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	iocmd.status = BFA_STATUS_OK;

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_fcpim_qdepth_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_boot_cfg(unsigned long arg)
{
	bfa_ioctl_boot_t iocmd;
	bfad_t         *bfad;
	struct bfad_hal_comp fcomp;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_boot_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status =
		bfa_hal_flash_update_part(&bfad->bfa_hal, BFA_FLASH_PART_BOOT,
					  PCI_FUNC(bfad->pcidev->devfn),
					  &iocmd.cfg, sizeof(bfa_boot_cfg_t), 0,
					  bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd.status != BFA_STATUS_OK)
		goto out;

	wait_for_completion(&fcomp.comp);
	iocmd.status = fcomp.status;

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_boot_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_boot_query(unsigned long arg)
{
	bfa_ioctl_boot_t iocmd;
	bfad_t         *bfad;
	struct bfad_hal_comp fcomp;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_boot_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status =
		bfa_hal_flash_read_part(&bfad->bfa_hal, BFA_FLASH_PART_BOOT,
					PCI_FUNC(bfad->pcidev->devfn),
					&iocmd.cfg, sizeof(bfa_boot_cfg_t), 0,
					bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (iocmd.status != BFA_STATUS_OK)
		goto out;

	wait_for_completion(&fcomp.comp);
	iocmd.status = fcomp.status;

out:
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_boot_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_ioc_get_version(unsigned long arg)
{
	bfa_ioctl_ioc_get_version_t iocmd;
	int             rc = 0;

	iocmd.version = BFA_IOCTL_VERSION;
	iocmd.status = BFA_STATUS_OK;

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd, sizeof(iocmd)))
		rc = EIO;

	return rc;
}

/*
 * Assume the instance numbers are in ascending order.
 * Returns the number of ioc instances in iocmd.cnt;
 * iocmd.bm[] has the bits set corresponding to the instances numbers
 */

int
bfad_ioctl_ioc_get_inst(unsigned long arg)
{
	bfad_t         *bfad = NULL;
	bfa_ioctl_ioc_get_inst_t iocmd;
	int idx = 0, bt_idx, rc = 0;
	int bt_sz = sizeof(iocmd.bm[0]) * 8;

	memset(&iocmd, 0, sizeof(iocmd));

	down(&bfad_sem);
	list_for_each_entry(bfad, &bfad_list, list_entry) {
		if (bfad->inst_no >= (idx + 1) * bt_sz) {
			idx++;
			if (idx > sizeof(iocmd.bm)) {
				/*
				 * too many instances
				 */
				rc = EIO;
				goto ext;
			}
		}
		bt_idx = bfad->inst_no - idx * bt_sz;
		if (bt_idx < 0) {
			rc = EIO;	/* inst_no not in ascending order */
			goto ext;
		}
		iocmd.bm[idx] |= (1 << bt_idx);
		iocmd.cnt++;
	}

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd, sizeof(iocmd)))
		rc = EIO;
ext:
	up(&bfad_sem);
	return rc;
}

int
bfad_ioctl_auth_get_attr(unsigned long arg)
{
	bfa_ioctl_auth_attr_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_auth_attr_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status = bfa_fcs_auth_get_attr(&bfad->bfa_fcs, &iocmd.attr);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_auth_attr_t)))
		return EIO;
	return 0;
}

int
bfad_ioctl_auth_get_stats(unsigned long arg)
{
	bfa_ioctl_auth_stats_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_auth_stats_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status = bfa_fcs_auth_get_stats(&bfad->bfa_fcs, &iocmd.stats);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_auth_stats_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_auth_set_policy(unsigned long arg)
{
	bfa_ioctl_auth_policy_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_auth_policy_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status = bfa_fcs_auth_set_policy(&bfad->bfa_fcs, iocmd.mode);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_auth_policy_t)))
		return EIO;

	return 0;

}

int
bfad_ioctl_auth_set_algo(unsigned long arg)
{
	bfa_ioctl_auth_algo_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_auth_algo_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status = bfa_fcs_auth_set_algo(&bfad->bfa_fcs, iocmd.algo);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_auth_algo_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_auth_set_group(unsigned long arg)
{
	bfa_ioctl_auth_group_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_auth_group_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status = bfa_fcs_auth_set_dh_group(&bfad->bfa_fcs, iocmd.group);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_auth_group_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_auth_set_secretstring(unsigned long arg)
{
	bfa_ioctl_auth_secretstring_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_auth_secretstring_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status =
		bfa_fcs_auth_set_secretstring_encrypt(&bfad->bfa_fcs,
		iocmd.encrypt, iocmd.len);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_auth_secretstring_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_auth_set_secretsource(unsigned long arg)
{
	bfa_ioctl_auth_secretsource_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_auth_secretsource_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status =
		bfa_fcs_auth_set_secretsource(&bfad->bfa_fcs, iocmd.source);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_auth_secretsource_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_auth_reset_stats(unsigned long arg)
{
	bfa_ioctl_gen_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_gen_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status = bfa_fcs_auth_reset_stats(&bfad->bfa_fcs);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_gen_t)))
		return EIO;

	return 0;
}

int
bfad_ioctl_auth_reinit(unsigned long arg)
{
	bfa_ioctl_gen_t iocmd;
	bfad_t         *bfad;
	unsigned long   flags;

	if (copy_from_user
	    ((uint8_t *) &iocmd, (void __user *)arg,
	     sizeof(bfa_ioctl_gen_t)))
		return EIO;

	bfad = bfad_find_bfad_by_inst_no(iocmd.bfad_num);
	if (bfad == NULL)
		return EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd.status = bfa_fcs_auth_reinit(&bfad->bfa_fcs);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (copy_to_user
	    ((void __user *)arg, (uint8_t *) &iocmd,
	     sizeof(bfa_ioctl_gen_t)))
		return EIO;

	return 0;

}

int
bfad_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	   unsigned long arg)
{
	int             rc = EINVAL;

	if (!arg)
		goto ext;

	if (cmd == IOCMD_IOC_GET_VERSION) {
		rc = bfad_ioctl_ioc_get_version(arg);
		goto ext;
	}

	if (cmd == IOCMD_IOC_GET_INSTS) {
		rc = bfad_ioctl_ioc_get_inst(arg);
		goto ext;
	}

	switch (cmd) {
	/*
	 * IOC IOCTLs
	 */
	case IOCMD_IOC_ENABLE:
		rc = bfad_ioctl_ioc_enable(arg);
		break;
	case IOCMD_IOC_DISABLE:
		rc = bfad_ioctl_ioc_disable(arg);
		break;
	case IOCMD_IOC_CFG_PPORT:
		rc = bfad_ioctl_ioc_cfg_pport(arg);
		break;
	case IOCMD_IOC_GET_ATTR:
		rc = bfad_ioctl_ioc_get_attr(arg);
		break;
	case IOCMD_IOC_GET_INFO:
		rc = bfad_ioctl_ioc_get_info(arg);
		break;
	case IOCMD_IOC_GET_STATS:
		rc = bfad_ioctl_ioc_get_stats(arg);
		break;
	case IOCMD_IOC_RESET_STATS:
		rc = bfad_ioctl_ioc_reset_stats(arg);
		break;
	case IOCMD_IOC_SET_INTR:
		rc = bfad_ioctl_ioc_set_intr(arg);
		break;
	case IOCMD_IOC_SET_PW:
		rc = bfad_ioctl_ioc_set_power(arg);
		break;
	case IOCMD_IOC_GET_PW:
		rc = bfad_ioctl_ioc_get_power(arg);
		break;
	case IOCMD_IOC_SET_ADAPTER_NAME:
		rc = bfad_ioctl_ioc_set_adapter_name(arg);
		break;
	case IOCMD_IOC_SET_PORT_NAME:
		rc = bfad_ioctl_ioc_set_port_name(arg);
		break;
	/*
	 * PORT IOCTLs
	 */
	case IOCMD_PORT_ENABLE:
		rc = bfad_ioctl_port_enable(arg);
		break;
	case IOCMD_PORT_DISABLE:
		rc = bfad_ioctl_port_disable(arg);
		break;
	case IOCMD_PORT_GET_STATS:
		rc = bfad_ioctl_port_get_stats(arg);
		break;
	case IOCMD_PORT_RESET_STATS:
		rc = bfad_ioctl_port_reset_stats(arg);
		break;
	case IOCMD_PORT_GET_ATTR:
		rc = bfad_ioctl_port_get_attr(arg);
		break;
	case IOCMD_PORT_CFG_TOPO:
		rc = bfad_ioctl_port_cfg_topology(arg);
		break;
	case IOCMD_PORT_CFG_SPEED:
		rc = bfad_ioctl_port_cfg_speed(arg);
		break;
	case IOCMD_PORT_CFG_ALPA:
		rc = bfad_ioctl_port_cfg_hardalpa(arg);
		break;
	case IOCMD_PORT_CLR_ALPA:
		rc = bfad_ioctl_port_clr_hardalpa(arg);
		break;
	case IOCMD_PORT_CFG_MAXFRSZ:
		rc = bfad_ioctl_port_cfg_maxfrsize(arg);
		break;
	case IOCMD_PORT_TRUNK_ENABLE:
		rc = bfad_ioctl_port_trunk_enable(arg);
		break;
	case IOCMD_PORT_TRUNK_DISABLE:
		rc = bfad_ioctl_port_trunk_disable(arg);
		break;
	/*
	 * LPORT IOCTLs
	 */
	case IOCMD_LPORT_GET_ATTR:
		rc = bfad_ioctl_lport_get_attr(arg);
		break;
	case IOCMD_LPORT_GET_STATS:
		rc = bfad_ioctl_lport_get_stats(arg);
		break;
	case IOCMD_LPORT_RESET_STATS:
		rc = bfad_ioctl_lport_reset_stats(arg);
		break;
	case IOCMD_LPORT_GET_RPORTS:
		rc = bfad_ioctl_lport_get_rports(arg);
		break;
	/*
	 * RPORT IOCTLs
	 */
	case IOCMD_RPORT_ADD:
	case IOCMD_RPORT_REMOVE:
		rc = EINVAL;
		break;
	case IOCMD_RPORT_GET_ATTR:
		rc = bfad_ioctl_rport_get_attr(arg);
		break;
	case IOCMD_RPORT_GET_STATS:
		rc = bfad_ioctl_rport_get_stats(arg);
		break;
	case IOCMD_RPORT_RESET_STATS:
		rc = bfad_ioctl_rport_clr_stats(arg);
		break;
	case IOCMD_RPORT_GET_ADDR:
		rc = bfad_ioctl_rport_get_addr(arg);
		break;
	/*
	 * VPORT IOCTLs
	 */
	case IOCMD_VPORT_CREATE:
		rc = bfad_ioctl_vport_create(arg);
		break;
	case IOCMD_VPORT_DEL:
		rc = bfad_ioctl_vport_del(arg);
		break;
	case IOCMD_VPORT_START:
		rc = bfad_ioctl_vport_start(arg);
		break;
	case IOCMD_VPORT_STOP:
		rc = bfad_ioctl_vport_stop(arg);
		break;
	case IOCMD_VPORT_GET_ATTR:
		rc = bfad_ioctl_vport_get_attr(arg);
		break;
	case IOCMD_VPORT_GET_STATS:
		rc = bfad_ioctl_vport_get_stats(arg);
		break;
	case IOCMD_VPORT_RESET_STATS:
		rc = bfad_ioctl_vport_clr_stats(arg);
		break;
	/*
	 * VF IOCTLs
	 */
	case IOCMD_VF_ENABLE:
		rc = bfad_ioctl_vf_mode_enable(arg);
		break;
	case IOCMD_VF_DISABLE:
		rc = bfad_ioctl_vf_mode_disable(arg);
		break;
	case IOCMD_VF_ADD:
		rc = bfad_ioctl_vf_add(arg);
		break;
	case IOCMD_VF_DEL:
		rc = bfad_ioctl_vf_del(arg);
		break;
	case IOCMD_VF_LIST:
	case IOCMD_VF_LIST_ALL:
		rc = bfad_ioctl_vf_list(cmd, arg);
		break;
	case IOCMD_VF_GET_ATTR:
		rc = bfad_ioctl_vf_get_attr(arg);
		break;
	case IOCMD_VF_GET_STATS:
		rc = bfad_ioctl_vf_get_stats(arg);
		break;
	case IOCMD_VF_RESET_STATS:
		rc = bfad_ioctl_vf_clr_stats(arg);
		break;
	/*
	 * FABRIC
	 */
	case IOCMD_FABRIC_GET_LPORTS:
		rc = bfad_ioctl_fabric_get_lports(arg);
		break;
	/*
	 * FLASH IOCTLs
	 */
	case IOCMD_FLASH_GET_ATTR:
		rc = bfad_ioctl_flash_get_attr(arg);
		break;
	case IOCMD_FLASH_ERASE_PART:
		rc = bfad_ioctl_flash_erase_part(arg);
		break;
	case IOCMD_FLASH_UPDATE_PART:
		rc = bfad_ioctl_flash_update_part(arg);
		break;
	case IOCMD_FLASH_MFG_INFO:
		rc = bfad_ioctl_flash_mfg_info(arg);
		break;
	case IOCMD_FLASH_READ_PART:
		rc = bfad_ioctl_flash_read_part(arg);
		break;
	/*
	 * DIAG IOCTLs
	 */
	case IOCMD_DIAG_REGRD:
		rc = bfad_ioctl_diag_regrd(arg);
		break;
	case IOCMD_DIAG_REGWR:
		rc = bfad_ioctl_diag_regwr(arg);
		break;
	case IOCMD_DIAG_TEMP:
		rc = bfad_ioctl_diag_temp(arg);
		break;
	case IOCMD_DIAG_POST:
		rc = bfad_ioctl_diag_post(arg);
		break;
	case IOCMD_DIAG_MEMTEST:
		rc = bfad_ioctl_diag_memtest(arg);
		break;
	case IOCMD_DIAG_LOOPBACK:
		rc = bfad_ioctl_diag_loopback(arg);
		break;
	case IOCMD_DIAG_FWPING:
		rc = bfad_ioctl_diag_fwping(arg);
		break;
	case IOCMD_DIAG_POM:
		rc = bfad_ioctl_diag_pom(arg);
		break;
	case IOCMD_DIAG_SFP:
		rc = bfad_ioctl_diag_sfp(arg);
		break;
	case IOCMD_DIAG_LED:
		rc = bfad_ioctl_diag_led(arg);
		break;
	case IOCMD_DIAG_BEACON_LPORT:
		rc = bfad_ioctl_diag_beacon_lport(arg);
		break;
	/*
	 * DEBUG IOCTLs
	 */
	case IOCMD_DEBUG_DRV_TRACE:
		rc = bfad_ioctl_debug_drv_trace(arg);
		break;
	case IOCMD_DEBUG_DRV_STATE:
		rc = bfad_ioctl_debug_drv_state(arg);
		break;
	case IOCMD_DEBUG_FW_TRACE:
		rc = bfad_ioctl_debug_fw_trace(arg);
		break;
	case IOCMD_DEBUG_FW_STATE:
		rc = bfad_ioctl_debug_fw_state(arg);
		break;

	case IOCMD_DEBUG_PORTLOG:
		rc = bfad_ioctl_porglog_get(arg);
		break;
	case IOCMD_DEBUG_PORTLOG_CLR:
		rc = bfad_ioctl_porglog_clear(arg);
		break;
	case IOCMD_DEBUG_PORTLOG_CTL:
		rc = bfad_ioctl_porglog_ctl(arg);
		break;
	/*
	 * LOG IOCTLs
	 */
	case IOCMD_LOG_SET_LEVEL:
		rc = bfad_ioctl_log_set_level(arg);
		break;
	case IOCMD_LOG_GET_LEVEL:
		rc = bfad_ioctl_log_get_level(arg);
		break;
	case IOCMD_QOS_ENABLE:
	case IOCMD_QOS_DISABLE:
		rc = bfad_ioctl_qos(arg, cmd);
		break;
	/*
	 * RATELIM IOCTLs
	 */
	case IOCMD_RATELIM_ENABLE:
	case IOCMD_RATELIM_DISABLE:
		rc = bfad_ioctl_ratelim(arg, cmd);
		break;
	/*
	 * FCPIM IOCTLs
	 */
	case IOCMD_FCPIM_FAILOVER:
		rc = bfad_ioctl_fcpim_failover(arg);
		break;
	case IOCMD_FCPIM_QDEPTH:
		rc = bfad_ioctl_fcpim_qdepth(arg);
		break;
	/*
	 * BOOT IOCTLs
	 */
	case IOCMD_BOOT_CFG:
		rc = bfad_ioctl_boot_cfg(arg);
		break;
	case IOCMD_BOOT_QUERY:
		rc = bfad_ioctl_boot_query(arg);
		break;
	/*
	 * ITNIM IOCTLs
	 */
	case IOCMD_ITNIM_GET_ATTR:
		rc = bfad_ioctl_itnim_get_attr(arg);
		break;
	case IOCMD_ITNIM_GET_STATS:
		rc = bfad_ioctl_itnim_get_stats(arg);
		break;
	case IOCMD_ITNIM_RESET_STATS:
		rc = bfad_ioctl_itnim_reset_stats(arg);
		break;
	/*
	 * SCSI PASSTHRU
	 */
	case IOCMD_SEND_SCSI:
		rc = bfad_send_scsi_passthru(arg);
		break;
	/*
	 * FC PASSTHRU
	 */
	case IOCMD_SEND_FC:
		rc = bfad_send_fc_passthru(arg);
		break;
	/*
	 * EVENT IOCTLs
	 */
	case IOCMD_GET_AEN:
		rc = bfad_ioctl_get_aen(arg);
		break;
	/*
	 * AUTH IOCTLs
	 */
	case IOCMD_AUTH_GET_ATTR:
		rc = bfad_ioctl_auth_get_attr(arg);
		break;
	case IOCMD_AUTH_GET_STATS:
		rc = bfad_ioctl_auth_get_stats(arg);
		break;
	case IOCMD_AUTH_REINIT:
		rc = bfad_ioctl_auth_reinit(arg);
		break;
	case IOCMD_AUTH_SET_POLICY:
		rc = bfad_ioctl_auth_set_policy(arg);
		break;
	case IOCMD_AUTH_SET_ALGO:
		rc = bfad_ioctl_auth_set_algo(arg);
		break;
	case IOCMD_AUTH_SET_GROUP:
		rc = bfad_ioctl_auth_set_group(arg);
		break;
	case IOCMD_AUTH_SET_SECRETSTR:
		rc = bfad_ioctl_auth_set_secretstring(arg);
		break;
	case IOCMD_AUTH_SET_SECRETSRC:
		rc = bfad_ioctl_auth_set_secretsource(arg);
		break;
	case IOCMD_AUTH_RESET_STATS:
		rc = bfad_ioctl_auth_reset_stats(arg);
		break;
	default:
		rc = EINVAL;
		break;
	}
ext:
	return -rc;
}
