/*
 * Brocade Fibre Channel HBA Driver
 * Copyright (c) 2006-2007 Brocade Communications Systems, Inc.
 * All rights reserved.
 */

#include <linux/module.h>

#include "bfad_ioctl.h"
#include "bfad_drv.h"

#include "bfad_aen.h"

static int      bfad_open(struct inode *inode, struct file *file);
static int      bfad_close(struct inode *inode, struct file *file);
int             bfad_ioctl(struct inode *inode, struct file *file,
			   unsigned int cmd, unsigned long arg);
static ssize_t  bfad_read(struct file *file, char __user *buf, size_t count,
			  loff_t *ppos);
static unsigned int bfad_poll(struct file *file,
			      struct poll_table_struct *wait);
static int      bfad_fasync(int fd, struct file *file, int on);

static struct class *bfad_class;

static int      bfad_major;

bfad_t         *bfad_find_bfad_by_inst_no(int inst_no);

static char    *bfad_dev_name = BFAD_DRIVER_NAME;
static struct file_operations bfad_fops;

int
bfad_os_class_create(struct module *owner, char *name)
{
	bfad_class = class_create(owner, name);
	if (IS_ERR(bfad_class)) {
		printk(KERN_WARNING "Unable to create sysfs class for bfa.\n");
		bfad_class = NULL;
		return 1;
	}

	return 0;
}

void
bfad_os_class_destroy(void)
{
	class_destroy(bfad_class);
	bfad_class = NULL;
}

int
bfad_os_class(void)
{
	return (bfad_class != 0);
}

int
bfad_os_ioctl_init(void)
{
	struct class_device *cd;

	if (bfad_os_class_create(THIS_MODULE, bfad_dev_name)) {
		printk(KERN_WARNING "Unable to create sysfs class for bfa.\n");
		return 1;
	}
	bfad_major = register_chrdev(0, bfad_dev_name, &bfad_fops);
	if (bfad_major < 0) {
		printk(KERN_WARNING "Unable to register chrdev for bfa.\n");
		bfad_os_class_destroy();
		return bfad_major;
	}
	cd = class_device_create(bfad_class, NULL, MKDEV(bfad_major, 0), NULL,
				 "bfa");

	if (IS_ERR(cd))
		return BFA_STATUS_FAILED;

	return BFA_STATUS_OK;
}

int
bfad_os_ioctl_exit(void)
{
	if (!bfad_os_class())
		return 1;

	class_device_destroy(bfad_class, MKDEV(bfad_major, 0));

	unregister_chrdev(bfad_major, bfad_dev_name);
	bfad_os_class_destroy();
	return 0;
}

static long
bfad_compat_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
	return (bfad_ioctl(file->f_dentry->d_inode, file, cmd, arg));
}

static int
bfad_open(struct inode *inode, struct file *file)
{
	bfad_t         *bfad;
	int32_t         bfad_minor;
	unsigned long   flags;

	bfad_minor = MINOR(inode->i_rdev);
	bfad = bfad_find_bfad_by_inst_no(bfad_minor);
	if (bfad == NULL) {
		printk(KERN_WARNING "%s: Invalid minor device %d\n",
			__FUNCTION__, bfad_minor);
		return (-EINVAL);
	}

	file->private_data = (void *)bfad;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	/*
	 * Update the reference count
	 */
	bfad->ref_count++;
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	return 0;
}

static int
bfad_close(struct inode *inode, struct file *file)
{
	bfad_t         *bfad = (bfad_t *) file->private_data;
	int32_t         bfad_minor;
	unsigned long   flags;

	bfad_minor = MINOR(inode->i_rdev);

	if (bfad == NULL) {
		printk(KERN_WARNING "%s: Device %d not open!!\n",
			 __FUNCTION__, bfad_minor);
		return (-EINVAL);
	}

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	/*
	 * Update the reference count
	 */
	bfad->ref_count--;
	if (!bfad->ref_count) {
		file->private_data = 0x0;
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);

		/* 
		 * remove this file from the asynchronously notified file's
		 * bfad_aen_fasync() must not refer to file->private_data
		 * fasync_helper() may sleep, so release the lock
		 */
		bfad_aen_fasync(-1, file, 0);
		return (0);
	}
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	return (0);
}

static          ssize_t
bfad_read(struct file *file, char __user *buf, size_t count, loff_t *ppos)
{
	return bfad_aen_read(file, buf, count, ppos);
}

static unsigned int
bfad_poll(struct file *file, struct poll_table_struct *wait)
{
	return bfad_aen_poll(file, wait);
}

static int
bfad_fasync(int fd, struct file *file, int on)
{
	return bfad_aen_fasync(fd, file, on);
}

/**
 * Define the BFA driver file operations.
 */
static struct file_operations bfad_fops = {
	.owner = THIS_MODULE,
	.ioctl = bfad_ioctl,
	.compat_ioctl = bfad_compat_ioctl,
	.open = bfad_open,
	.release = bfad_close,
	.read = bfad_read,
	.poll = bfad_poll,
	.fasync = bfad_fasync
};
