/* 
 * Copyright (c)  2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 * 
 * See LICENSE.bfa for copyright and licensing details.
 */

/**
 * @file bfad_sysfs.c Linux driver sysfs interface functionality.
 */

/*
 *  SYSFS Directory Structure for Brocade HBAs:
 *
 *  The ROOT_DIR is automatically created by the OS when a HBA is present.
 *  Eg: ROOT_DIR for HBA in slot 6 :
 *      /sys/devices/pci0000:00/0000:00:06.0/0000:10:00.0/    // instance .0
 *      /sys/devices/pci0000:00/0000:00:06.0/0000:10:00.1/    // instance .1
 *
 *  The following sub-directory structure will be created by the driver:
 *
 *  <ROOT_DIR>/bfa/ioc/
 *                 |   /attr
 *                 |   /info
 *                 |   /intr
 *                 |   /power
 *                 |   /stats
 *                 /port/
 *                 |    /alpa
 *                 |    /attr
 *                 |    /enable
 *                 |    /maxfrm
 *                 |    /speed
 *                 |    /stats
 *                 |    /topo
 *                 |    /trunk
 *                 |create_vport
 *                 |delete_vport
 *                 /lport-0/
 *                 |       /attr
 *                 |       /stats
 *                 |       /refresh_rports
 *                 |       /rp-00000000000001/
 *                 |                         /attr
 *                 |                         /stats
 *                 /vp-1122334455667788/                            // VPORT
 *                 |                   /attr
 *                 |                   /stats
 *                 |                   /refresh_rports
 *                 |                   /rp-00000000000001/
 *                 |                                     /attr
 *                 |                                     /stats
 *                 /vf/
 *                   /add
 *                   /delete
 *                   /list
 *                   /listall
 *                   /vf-ACEF/
 *                           /attr
 *                           /stats
 *                           /create_vport
 *                           /delete_vport
 *                           /lport-0/
 *                           |       /attr
 *                           |       /stats
 *                           |       /rp-00000000000001/
 *                           |                         /attr
 *                           |                         /stats
 *                           /vp-01020304050607/
 *                                             /attr
 *                                             /stats
 *                                             /rp-00000000000001/
 *                                                               /attr
 *                                                               /stats
 *
 *
 *  Sub-directories for VPORTs and VFs are created dynamically as and when user
 *  adds them via. sysfs or cli.
 *
 *  As an example, in the following directory tree, VPORT with
 *  PWWN-11:22:33:44:55:66:77 has been added.
 *
 *  Also, a vf with vf id: ACEF has been created with results into a
 *  sub-folder vf-ACEF.
 *
 */

/**
 * common file to be included in OS specific .c files
 */

#include "bfad_drv.h"
#include "bfad_trcmod.h"
#include "bfad_sysfs.h"
#include "fcs/bfa_fcs_fcpim.h"
#include <log/bfa_log_linux.h>
#include <linux/module.h>

#define SYSFS_CONTROL_SUPPORT
#ifdef SYSFS_CONTROL_SUPPORT

#define SYSFS_MAX_RPORTS	(2 * 1024)

/**
 *  forward declarations
 */
static ssize_t   bfad_sysfs_rport_main_show_list(bfad_t *bfad,
						struct kobject *kobj, int vf_id,
						wwn_t pwwn, wwn_t vpwwn,
						char *buf);
char *
bfad_sysfs_private_kobject_get_path(struct kobject *kobj,
					gfp_t gfp_mask);
void            bfad_sysfs_hcb_comp(void *arg, bfa_status_t status);

bfad_t         *bfad_find_bfad_by_inst_no(int inst);

#define MAX_VAL -1
typedef struct {
	int             code;
	char           *str;
} bfa_state_t;

static bfa_state_t port_state_table[] = {
	{0, "Unknown"},
	{BFA_HAL_PPORT_UNINIT, "Uninit"},
	{BFA_HAL_PPORT_ENABLING_QWAIT, "Enabling (qwait)"},
	{BFA_HAL_PPORT_ENABLING, "Enabling"},
	{BFA_HAL_PPORT_LINKDOWN, "Linkdown"},
	{BFA_HAL_PPORT_LINKUP, "Linkup"},
	{BFA_HAL_PPORT_DISABLING_QWAIT, "Disabling (qwait)"},
	{BFA_HAL_PPORT_DISABLING, "Disabling"},
	{BFA_HAL_PPORT_DISABLED, "Disabled"},
	{BFA_HAL_PPORT_STOPPED, "Stopped"},
	{BFA_HAL_PPORT_IOCDOWN, "IOC down"},
	{BFA_HAL_PPORT_MAX_STATE, "Invalid"},
	{MAX_VAL, "--"}
};

static bfa_state_t vport_state_table[] = {
	{BFA_FCS_VPORT_UNINIT, "Uninit"},
	{BFA_FCS_VPORT_CREATED, "Created"},
	{BFA_FCS_VPORT_OFFLINE, "Offline"},
	{BFA_FCS_VPORT_FDISC_SEND, "Fdisc Send"},
	{BFA_FCS_VPORT_FDISC, "Fdisc"},
	{BFA_FCS_VPORT_FDISC_RETRY, "Fdisc Retry"},
	{BFA_FCS_VPORT_ONLINE, "Online"},
	{BFA_FCS_VPORT_DELETING, "Deleting"},
	{BFA_FCS_VPORT_CLEANUP, "Cleanup"},
	{BFA_FCS_VPORT_LOGO_SEND, "Logo Send"},
	{BFA_FCS_VPORT_LOGO, "Logo"},
	{MAX_VAL, "--"},
};

/*
 * rport state
 */
static bfa_state_t rport_state_table[] = {
	{BFA_RPORT_UNINIT, "uninit"},
	{BFA_RPORT_OFFLINE, "offline"},
	{BFA_RPORT_PLOGI, "plogi"},
	{BFA_RPORT_ONLINE, "online"},
	{BFA_RPORT_PLOGI_RETRY, "retry"},
	{BFA_RPORT_NSQUERY, "nsquery"},
	{BFA_RPORT_ADISC, "adisc"},
	{BFA_RPORT_LOGO, "logo"},
	{BFA_RPORT_LOGORCV, "logorecv"},
	{BFA_RPORT_NSDISC, "nsdisc"},
	{MAX_VAL, "--"}
};

/*
 * fcpim state
 */
static bfa_state_t fcpim_state_table[] = {
	{BFA_ITNIM_OFFLINE, "offline"},
	{BFA_ITNIM_PRLI_SEND, "prli send"},
	{BFA_ITNIM_PRLI_SENT, "prli sent"},
	{BFA_ITNIM_PRLI_RETRY, "prli retry"},
	{BFA_ITNIM_HCB_ONLINE, "online callback"},
	{BFA_ITNIM_ONLINE, "online"},
	{BFA_ITNIM_HCB_OFFLINE, "offline callback"},
	{BFA_ITNIM_INITIATIOR, "initiator"},
	{MAX_VAL, "--"}
};

char           *
bfad_sysfs_state_to_string(bfa_rport_state_t state, bfa_state_t *table)
{
	int             j = 0;
	while (table[j].code <= BFA_RPORT_NSDISC) {
		if (table[j].code == state)
			return table[j].str;
		j++;
	}
	return "unknown";
}

/*
 * strcspn() is not found in certain Linux falvors.
 * So, implementing our own.
 */
int
my_strcspn(char *path, const char *sub)
{
	int             i;

	for (i = 0; i <= strlen(path) - strlen(sub); i++) {
		if (strncmp((path + i), sub, strlen(sub)) == 0)
			return i;
	}

	return 0;
}

LIST_HEAD(bfad_kobj_vf_list);	/* !< List maintains the VFs */
LIST_HEAD(bfad_kobj_base_list);	/* !< List maintains the Vports/Rports for base
				 *port */

typedef enum {
	KOBJ_TYPE_NONE = 0,
	KOBJ_TYPE_VF,
	KOBJ_TYPE_VPORT,
	KOBJ_TYPE_RPORT,
	KOBJ_TYPE_LPORT,
	KOBJ_TYPE_ITNIM
} bfad_kobj_type_t;

/**
 * @brief Node which is the directory that has information about the node.
 *  If the directory is a VF node, then it has a list of Vports created
 *  under it.
 */
typedef struct bfad_kobj_s {
	struct list_head kobj_list_entry;
	bfad_kobj_type_t type;	/* !< Type of the node */
	uint8_t         inst;	/* !< Instance = 0 or 1 */
	uint16_t        vf_id;	/* !< VF ID */
	uint64_t        vport_id;	/* !< VPORT ID */
	uint64_t        rport_id;	/* !< RPORT ID */
	uint64_t        pwwn;	/* !< PWWN */
	struct kobject  kobj;	/* !< kobj of the node itself */
	struct list_head kobj_vport_list;	/* !< List of all vports of a
						 *VF */
} bfad_kobj_t;

/**
 * @brief Search function to get a Node that match given values
 * @param[in] path - Path to the IOC
 * @param[in] type - Type of the node
 * @param[in] vf_id - vf_id
 * @param[in] vport_id - vport id
 * @param[in] list - List to be searched
 * @retval kobj if the search is successful
 * @retval NULL if the search fails.
 */
static bfad_kobj_t *
bfad_sysfs_kobj_get(char *ioc_path, bfad_kobj_type_t type, uint16_t vf_id,
		    uint64_t vport_id, struct list_head *list)
{
	bfad_kobj_t    *bfad_kobj = NULL;
	char           *path;

	if (list_empty(list))
		return NULL;

	list_for_each_entry(bfad_kobj, list, kobj_list_entry) {
		path = bfad_sysfs_private_kobject_get_path(&bfad_kobj->kobj,
							   GFP_ATOMIC);
		if (path == NULL)
			continue;
		if (strlen(path) == 0) {
			kfree(path);
			continue;
		}
		if (bfad_kobj->type == type && bfad_kobj->vf_id == vf_id
		    && bfad_kobj->vport_id == vport_id
		    && strstr(path, ioc_path)) {
			kfree(path);
			return bfad_kobj;
		}
		kfree(path);
	}

	return NULL;
}

static bfad_kobj_t *
bfad_sysfs_search_kobj(uint64_t pwwn, bfad_kobj_type_t type, uint16_t vf_id,
		       uint64_t vport_id, wwn_t rport_id,
		       struct list_head *list)
{
	bfad_kobj_t    *bfad_kobj = NULL;

	if (list_empty(list))
		return NULL;

	list_for_each_entry(bfad_kobj, list, kobj_list_entry) {
		if (bfad_kobj->type == type && bfad_kobj->vf_id == vf_id
		    && bfad_kobj->vport_id == vport_id
		    && bfad_kobj->rport_id == rport_id
		    && bfad_kobj->pwwn == pwwn)
			return bfad_kobj;
	}

	return NULL;
}

void
bfad_sysfs_remove_deleted_rports(bfa_fcs_port_t *fcs_port, uint64_t pwwn,
				 uint16_t vf_id, struct list_head *list)
{
	bfa_fcs_rport_t *fcs_rport;
	bfad_kobj_t    *bfad_kobj = NULL;
	struct list_head *pos, *next;

	if (list_empty(list))
		return;

	/*
	 * Delete the ITNIM nodes first
	 */
	list_for_each_safe(pos, next, list) {
		bfad_kobj = list_entry(pos, bfad_kobj_t, kobj_list_entry);
		if (bfad_kobj->type == KOBJ_TYPE_ITNIM
		    && bfad_kobj->vf_id == vf_id && bfad_kobj->pwwn == pwwn) {
			fcs_rport =
				bfa_fcs_rport_lookup(fcs_port,
						     bfad_kobj->rport_id);
			if (fcs_rport == NULL) {
				kobject_unregister(&bfad_kobj->kobj);
				list_del(&bfad_kobj->kobj_list_entry);
				kfree(bfad_kobj);
			}
		}
	}

	/*
	 * Delete the RPORT itself
	 */
	list_for_each_safe(pos, next, list) {
		bfad_kobj = list_entry(pos, bfad_kobj_t, kobj_list_entry);
		if (bfad_kobj->type == KOBJ_TYPE_RPORT
		    && bfad_kobj->vf_id == vf_id && bfad_kobj->pwwn == pwwn) {
			fcs_rport =
				bfa_fcs_rport_lookup(fcs_port,
						     bfad_kobj->rport_id);
			if (fcs_rport == NULL) {
				kobject_unregister(&bfad_kobj->kobj);
				list_del(&bfad_kobj->kobj_list_entry);
				kfree(bfad_kobj);
			}
		}
	}

}

void
bfad_sysfs_delete_all_rports(bfad_t *bfad, bfad_kobj_type_t type,
			     uint16_t vf_id, struct list_head *list)
{
	bfad_kobj_t    *bfad_kobj = NULL;
	bfa_pport_attr_t pattr;
	struct list_head *pos, *next;
	unsigned long	flags;

	if (list_empty(list))
		return;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	bfa_hal_pport_get_attr(&bfad->bfa_hal, &pattr);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	list_for_each_safe(pos, next, list) {
		bfad_kobj = list_entry(pos, bfad_kobj_t, kobj_list_entry);
		if (bfad_kobj->type == type && bfad_kobj->vf_id == vf_id
		    && bfad_kobj->pwwn == pattr.pwwn) {
			kobject_unregister(&bfad_kobj->kobj);
			list_del(&bfad_kobj->kobj_list_entry);
			kfree(bfad_kobj);
		}
	}

}

void
bfad_sysfs_delete_all_rports_of_vport(bfad_t *bfad, bfad_kobj_type_t type,
				      uint16_t vf_id, wwn_t pwwn, wwn_t vpwwn,
				      struct list_head *list)
{
	bfad_kobj_t    *bfad_kobj = NULL;
	struct list_head *pos, *next;

	if (list_empty(list))
		return;

	list_for_each_safe(pos, next, list) {
		bfad_kobj = list_entry(pos, bfad_kobj_t, kobj_list_entry);
		if (bfad_kobj->type == type && bfad_kobj->vf_id == vf_id
		    && bfad_kobj->pwwn == pwwn
		    && bfad_kobj->vport_id == vpwwn) {
			kobject_unregister(&bfad_kobj->kobj);
			list_del(&bfad_kobj->kobj_list_entry);
			kfree(bfad_kobj);
		}
	}

}

/**
 * @brief Function to delete either a VPORT or Lport-0
 * @param[in] path - Path to ioc
 * @param[in] type - Type of the node
 * @param[in] vf_id - vf_id
 * @param[in] vport_id - vport id
 * @param[in] list - List to be searched
 * @retval None
 */
static void
bfad_sysfs_kobj_del_port(char *ioc_path, bfad_kobj_type_t type, uint16_t vf_id,
			 uint64_t vport_id, struct list_head *list)
{
	bfad_kobj_t    *bfad_kobj = NULL;
	char           *path;
	struct list_head *pos, *next;

	list_for_each_safe(pos, next, list) {
		bfad_kobj = list_entry(pos, bfad_kobj_t, kobj_list_entry);
		path = bfad_sysfs_private_kobject_get_path(&bfad_kobj->kobj,
							   GFP_ATOMIC);
		if (bfad_kobj->type == type && bfad_kobj->vf_id == vf_id
		    && bfad_kobj->vport_id == vport_id
		    && strstr(path, ioc_path)) {
			kobject_unregister(&bfad_kobj->kobj);
			list_del(&bfad_kobj->kobj_list_entry);
			kfree(bfad_kobj);
			kfree(path);
			break;
		}
		kfree(path);
	}
}

/**
 * @brief Function to delete all the VPORTs and base port of a VF
 * @param[in] path - IOC path
 * @param[in] type - Type of the node
 * @param[in] vf_id - vf_id
 * @param[in] list - List to be searched
 * @retval None
 */
static void
bfad_sysfs_kobj_del_all_ports(char *ioc_path, bfad_kobj_type_t type,
			      uint16_t vf_id, struct list_head *list)
{
	bfad_kobj_t    *bfad_kobj = NULL;
	char           *path;
	struct list_head *pos, *next;

	list_for_each_safe(pos, next, list) {
		bfad_kobj = list_entry(pos, bfad_kobj_t, kobj_list_entry);
		path = bfad_sysfs_private_kobject_get_path(&bfad_kobj->kobj,
							   GFP_ATOMIC);
		if (bfad_kobj->vf_id == vf_id && bfad_kobj->type == type
		    && strstr(path, ioc_path)) {
			if ((type == KOBJ_TYPE_VPORT
			     && bfad_kobj->vport_id != 0)
			    || type == KOBJ_TYPE_LPORT) {
				kobject_unregister(&bfad_kobj->kobj);
				list_del(&bfad_kobj->kobj_list_entry);
				kfree(bfad_kobj);
				kfree(path);
			}
		}
		kfree(path);
	}

}

static void
bfad_sysfs_kobj_del_vf(char *ioc_path, uint16_t vf_id, struct list_head *list)
{

	if (list_empty(&bfad_kobj_vf_list))
		return;

	bfad_sysfs_kobj_del_all_ports(ioc_path, KOBJ_TYPE_VPORT, vf_id, list);
	bfad_sysfs_kobj_del_port(ioc_path, KOBJ_TYPE_LPORT, vf_id, 0, list);

	return;
}

/**
 * @brief Function to delete all the VFs recursively
 * @param[in] bfad - IOC
 * @retval None
 */
static void
bfad_sysfs_kobj_delete_all_vfs(bfad_t *bfad)
{
	bfad_kobj_t    *bfad_kobj = NULL;
	char           *ioc_path;
	struct list_head *pos, *next;

	if (list_empty(&bfad_kobj_vf_list))
		return;

	ioc_path =
		bfad_sysfs_private_kobject_get_path(&bfad->pcidev->dev.kobj,
						    GFP_ATOMIC);

	/*
	 * delete the vport/lport sub-folders for each of the VF
	 */
	list_for_each_entry(bfad_kobj, &bfad_kobj_vf_list, kobj_list_entry) {
		if (bfad_kobj->type == KOBJ_TYPE_VF && bfad_kobj->vf_id != 0) {
			bfad_sysfs_delete_all_rports(bfad, KOBJ_TYPE_ITNIM,
						     bfad_kobj->vf_id,
						     &bfad_kobj->
						     kobj_vport_list);
			bfad_sysfs_delete_all_rports(bfad, KOBJ_TYPE_RPORT,
						     bfad_kobj->vf_id,
						     &bfad_kobj->
						     kobj_vport_list);
			bfad_sysfs_kobj_del_vf(ioc_path, bfad_kobj->vf_id,
					       &bfad_kobj->kobj_vport_list);
		}
	}

	/*
	 * delete the VFs
	 */
	list_for_each_safe(pos, next, &bfad_kobj_vf_list) {
		bfad_kobj = list_entry(pos, bfad_kobj_t, kobj_list_entry);
		kobject_unregister(&bfad_kobj->kobj);
		list_del(&bfad_kobj->kobj_list_entry);
		kfree(bfad_kobj);
	}

	/*
	 * delete the vports under the base port - under 'bfa'
	 */
	bfad_sysfs_kobj_del_all_ports(ioc_path, KOBJ_TYPE_VPORT, 0,
				      &bfad_kobj_base_list);

	kfree(ioc_path);

}

static void
fill_kobj_path(struct kobject *kobj, char *path, int length)
{
	struct kobject *parent;

	--length;
	for (parent = kobj; parent; parent = parent->parent) {
		int             cur = strlen(kobject_name(parent));
		/*
		 * back up enough to print this name with '/'
		 */
		length -= cur;
		strncpy(path + length, kobject_name(parent), cur);
		*(path + --length) = '/';
	}

	pr_debug("%s: path = '%s'\n", __FUNCTION__, path);
}

static int
get_kobj_path_length(struct kobject *kobj)
{
	int             length = 1;
	struct kobject *parent = kobj;

	/*
	 * walk up the ancestors until we hit the one pointing to the root.
	 * Add 1 to strlen for leading '/' of each level.
	 */
	do {
		if (kobject_name(parent) == NULL)
			return 0;
		length += strlen(kobject_name(parent)) + 1;
		parent = parent->parent;
	} while (parent);
	return length;
}

/**
 * kobject_get_path - generate and return the path associated with a given
 * kobj and kset pair.
 *
 * @kobj:	kobject in question, with which to build the path
 * @gfp_mask:	the allocation type used to allocate the path
 *
 * The result must be freed by the caller with kfree().
 */
char *
bfad_sysfs_private_kobject_get_path(struct kobject *kobj, gfp_t gfp_mask)
{
	char           *path;
	int             len;

	len = get_kobj_path_length(kobj);
	if (len == 0)
		return NULL;
	path = kzalloc(len, gfp_mask);
	if (!path)
		return NULL;
	fill_kobj_path(kobj, path, len);

	return path;
}

char           *
bfad_os_sysfs_get_ioc_path(bfad_t *bfad)
{
	return bfad_sysfs_private_kobject_get_path(&bfad->pcidev->dev.kobj,
						   GFP_ATOMIC);

}

ssize_t
bfad_vf_ctl_show(struct device *dev,
		 struct device_attribute *attr,
		 char *buf)
{
	int             len = 0;

	return len;
}

ssize_t
bfad_vf_ctl_store(struct device *dev,
		  struct device_attribute *attr,
		  const char *buf, size_t count)
{
	int             len = 0;

	return len;
}

ssize_t
bfad_vport_ctl_show(struct device *dev,
		    struct device_attribute *attr,
		    char *buf)
{
	int             len = 0;

	return len;
}

ssize_t
bfad_vport_ctl_store(struct device *dev,
		     struct device_attribute *attr,
		     const char *buf, size_t count)
{
	int             len = 0;

	return len;
}

ssize_t
bfad_fc4types_show(struct device *dev,
		   struct device_attribute *attr,
		   char *buf)
{
	int             len = 0;

	return len;
}

ssize_t
bfad_fc4types_store(struct device *dev,
		    struct device_attribute *attr,
		    const char *buf, size_t count)
{
	int             len = 0;

	return len;
}

ssize_t
bfad_ctl_show(struct device *dev,
	      struct device_attribute *attr,
	      char *buf)
{
	int             len = 0;

	return len;
}

ssize_t
bfad_ctl_store(struct device *dev,
	       struct device_attribute *attr,
	       const char *buf, size_t count)
{
	int             len = 0;

	return len;
}

DEVICE_ATTR(vf_ctl, 0644, bfad_vf_ctl_show, bfad_vf_ctl_store);
DEVICE_ATTR(vport_ctl, 0644, bfad_vport_ctl_show, bfad_vport_ctl_store);
DEVICE_ATTR(fc4type, 0644, bfad_fc4types_show, bfad_fc4types_store);
DEVICE_ATTR(bfad_ctl, 0644, bfad_ctl_show, bfad_ctl_store);

/*
 * IOC Commands
 */
typedef union {
	unsigned char   p[8];
	wwn_t           pn;
} PWWN;

int
bfad_sysfs_snprint_pwwn(wwn_t w, char *buf, int len)
{
	PWWN            p;
	int             i = 0;

	p.pn = w;
	while (i < 8) {
		len += snprintf(buf + len, PAGE_SIZE - len, "%02x", p.p[i]);
		if (i == 7)
			break;
		len += snprintf(buf + len, PAGE_SIZE - len, ":");
		i++;
	}

	return len;
}

int
bfad_sysfs_print_pwwn(wwn_t w, char *buf)
{
	PWWN            p;
	int             i = 0, len = 0;

	p.pn = w;
	while (i < 8) {
		len += snprintf(buf + len, PAGE_SIZE - len, "%02x", p.p[i]);
		if (i == 7)
			break;
		i++;
	}

	return len;
}

bfa_boolean_t
bfad_sysfs_get_port_wwn(char *pid, wwn_t *pn)
{
	PWWN            pp;
	uint8_t         p[8];
	int             i = 0;

	sscanf(pid, "%hhx:%hhx:%hhx:%hhx:%hhx:%hhx:%hhx:%hhx", &p[0], &p[1],
	       &p[2], &p[3], &p[4], &p[5], &p[6], &p[7]);
	while (i < 8) {
		pp.p[i] = p[i];
		i++;
	}
	memcpy(pn, &pp.pn, 8);
	return BFA_TRUE;

}

bfa_boolean_t
bfad_sysfs_get_port_wwn_no_col(char *pid, wwn_t *pn)
{
	PWWN            pp;
	int             i = 0;
	uint8_t		x;
	char            str[100];
	strcpy(str, "");
	while (i < 8) {
		strncpy(str, &pid[i * 2], 2);
		str[2] = '\0';
		sscanf(str, "%hhx", &x);
		pp.p[i] = x;
		i++;
	}
	memcpy(pn, &pp.pn, 8);
	return BFA_TRUE;
}

/**
 * @brief Function to parse the directory path to get ioc instance
 *  VF ID (if applicable) and VPORT ID (if applicable)
 * @param[in]  kobj - kobject of the directory to parse
 * @param[out] ioc_inst - IOC instance
 * @param[out] vf_id - VF ID
 * @param[out] vport_id - PWWN of the VPORT
 * @param[out] rport_id - PWWN of the RPORT
 * @retval bfad_t - if the instance if found and if there is not parsing error
 */
bfad_t         *
bfad_sysfs_get_param(struct kobject *kobj, int *ioc_inst, int *vf_id,
		     uint64_t *vport_id, uint64_t *rport_id)
{
	char           *path, *str, wwn_str[100], tmp_str[100];
	bfad_t         *bfad;
	uint8_t         idx;

	path = bfad_sysfs_private_kobject_get_path(kobj, GFP_ATOMIC);

	/*
	 * extract the instance from the path
	 */
	strcpy(tmp_str, path);
	idx = my_strcspn(tmp_str, "bfa");
	if (idx == 0) {
		kfree(path);
		return NULL;
	}

	tmp_str[idx - 1] = '\0';
	idx = strlen(tmp_str);
	while (tmp_str[idx] != '.')
		idx--;
	sscanf((tmp_str + idx + 1), "%d", ioc_inst);

	bfad = bfad_os_find_bfad_by_path(path);
	if (bfad == NULL) {
		kfree(path);
		return NULL;
	}

	/*
	 * Check if we are at the vf level, if so get the VF ID
	 */
	*vf_id = 0;
	str = strstr(path, "/vf/");
	if (str)
		sscanf(str, "/vf/vf-%X", vf_id);

	*vport_id = 0;
	*rport_id = 0;

	 /* Check if we are at the rport level, if so get the RPORT ID */
	str = strstr(path, "/rp-");
	if (str) {
		sscanf(str, "/rp-%s", wwn_str);
		bfad_sysfs_get_port_wwn_no_col(wwn_str, rport_id);
	} 
	/* Check if we are at the vport level, if so get the VPORT ID */
	str = strstr(path, "/vp-");
	if (str) {
		sscanf(str, "/vp-%s", wwn_str);
		bfad_sysfs_get_port_wwn_no_col(wwn_str, vport_id);
	}

	kfree(path);
	return bfad;
}

#define MAX_IOC_INST	128
struct kobject  ko_bfa_root[MAX_IOC_INST], ko_ioc[MAX_IOC_INST],
	ko_pport[MAX_IOC_INST], ko_lport[MAX_IOC_INST], ko_vf[MAX_IOC_INST],
	ko_vf_lport[MAX_IOC_INST];

#define DEF_ATTR(_obj, _name)		\
static struct attribute _obj = {	\
	.name = _name, .owner = THIS_MODULE, .mode = 0644 };

/*
 * Configured new vport attributes
 */

DEF_ATTR(vport_new_enable, "enable");
DEF_ATTR(vport_new_attr, "attr");
DEF_ATTR(vport_new_stats, "stats");
DEF_ATTR(vport_new_refresh_rports, "refresh_rports");

static struct attribute *bfad_sysfs_vport_new_attrs[] = {
	&vport_new_enable,
	&vport_new_attr,
	&vport_new_stats,
	&vport_new_refresh_rports,
	NULL
};

int
find_kobj_by_path(struct kobject *parent)
{
	char           *path, *ioc_path;
	int             i = -1, found = 0;
	struct kobject *kobj;

	ioc_path = bfad_sysfs_private_kobject_get_path(parent, GFP_ATOMIC);
	if (ioc_path == NULL)
		return i;

	for (i = 0; i < MAX_IOC_INST; i++) {
		kobj = &ko_bfa_root[i];
		if (kobj == NULL)
			continue;
		if (strncmp(kobj->name, "bfa", 3) == 0) {
			path = bfad_sysfs_private_kobject_get_path(kobj,
								   GFP_ATOMIC);
			if (path == NULL)
				continue;
			if (strlen(path) == 0) {
				kfree(path);
				continue;
			}
			if (strstr(path, ioc_path)) {
				found = 1;
				kfree(path);
				break;
			}

			kfree(path);
		}
	}

	kfree(ioc_path);
	if (found)
		return i;
	else
		return -1;
}

/**
 * @brief Function to show VPORT attributes and statistics
 * @param[in] kobj - kobject of the (vport) directory
 * @param[in] attr - Attributes of the directory
 * @param[out] buf - Buffer to be returned with (attribute/statistic) values
 * @retval - size of the 'buf' filled
 */
static          ssize_t
bfad_sysfs_vport_new_show(struct kobject *kobj, struct attribute *attr,
			  char *buf)
{
	int             ioc_inst = 1;
	bfad_t         *bfad;
	int             vf_id = 0, len = 0;
	uint64_t        vport_id, rp;
	bfa_fcs_vport_t *fcs_vport = NULL;
	bfa_pport_attr_t pattr;
	bfa_vport_attr_t vport_attr;
	bfa_vport_stats_t vport_stats;
	unsigned long	flags;

	bfad = bfad_sysfs_get_param(kobj, &ioc_inst, &vf_id, &vport_id, &rp);
	if (bfad == NULL) {
		return snprintf(buf + len, PAGE_SIZE - len,
				"failed to get ioc\n");
	}

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	bfa_hal_pport_get_attr(&bfad->bfa_hal, &pattr);

	fcs_vport = bfa_fcs_vport_lookup(&bfad->bfa_fcs, vf_id, vport_id);
	if (fcs_vport == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		return snprintf(buf + len, PAGE_SIZE - len,
				"failed to locate vport\n");
	}

	if (strcmp(attr->name, "attr") == 0) {

		memset(&vport_attr, 0, sizeof(bfa_vport_attr_t));
		bfa_fcs_vport_get_attr(fcs_vport, &vport_attr);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);

		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %s\n",
				"State",
				bfad_sysfs_state_to_string(vport_attr.
							   vport_state,
							   vport_state_table));

		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %06x\n",
				"Pid", bfa_os_ntoh3b(vport_attr.port_attr.pid));

		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: ", "PWWN");
		len += bfad_sysfs_snprint_pwwn(vport_attr.port_attr.port_cfg.
					       pwwn, buf, len);
		len += snprintf(buf + len, PAGE_SIZE - len, "\n");

		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: ", "NWWN");
		len += bfad_sysfs_snprint_pwwn(vport_attr.port_attr.port_cfg.
					       nwwn, buf, len);
		len += snprintf(buf + len, PAGE_SIZE - len, "\n");

		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %s",
				"Sysname",
				vport_attr.port_attr.port_cfg.sym_name.symname);

		len += snprintf(buf + len, PAGE_SIZE - len, "\n");

		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: ", "Roles");

		if (vport_attr.port_attr.port_cfg.roles == BFA_PORT_ROLE_FCP_IM)
			len += snprintf(buf + len, PAGE_SIZE - len,
					"FCP initiator\n");
		else if (vport_attr.port_attr.port_cfg.roles ==
			 BFA_PORT_ROLE_FCP_TM)
			len += snprintf(buf + len, PAGE_SIZE - len,
					"FCP target\n");
		else if (vport_attr.port_attr.port_cfg.roles ==
			 BFA_PORT_ROLE_FCP_IPFC)
			len += snprintf(buf + len, PAGE_SIZE - len,
					"IP over FC\n");
		else
			len += snprintf(buf + len, PAGE_SIZE - len,
					"invalid !!\n");

		return len;
	} else if (strcmp(attr->name, "stats") == 0) {
		memset(&vport_stats, 0, sizeof(bfa_vport_stats_t));
		bfa_fcs_vport_get_stats(fcs_vport, &vport_stats);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		len += snprintf(buf + len, PAGE_SIZE - len, "Port Stats\n\n");
		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n",
				"NS command retries",
				vport_stats.port_stats.ns_retries);
		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n",
				"NS timeouts",
				vport_stats.port_stats.ns_timeouts);
		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n",
				"NS rejects",
				vport_stats.port_stats.ns_rejects);
		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n\n",
				"No. rescans", vport_stats.port_stats.num_rscn);

		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n",
				"Fdisc retries", vport_stats.fdisc_retries);
		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n",
				"Fdisc timeouts", vport_stats.fdisc_timeouts);
		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n",
				"Fdisc rejects", vport_stats.fdisc_rejects);

		return len;
	} else if (strcmp(attr->name, "refresh_rports") == 0) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		return bfad_sysfs_rport_main_show_list(bfad, kobj, vf_id,
						       pattr.pwwn, vport_id,
						       buf);
	}

	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	return 0;
}

/**
 * @brief Function to start/stop VPORT
 * @param[in] kobj  - kobject of the (vport) directory
 * @param[in] attr  - Attributes of the directory
 * @param[in] buf   - Buffer contains user input
 * @param[in] count - Number of bytes in 'buf'
 * @retval count - Number of bytes of 'buf' used in the function
 */
static          ssize_t
bfad_sysfs_vport_new_store(struct kobject *kobj, struct attribute *attr,
			   const char *buf, size_t count)
{
	int             ioc_inst = 1, value = 0;
	bfad_t         *bfad;
	int             vf_id = 0;
	uint64_t        vport_id, rp;
	bfa_fcs_vport_t *fcs_vport = NULL;
	unsigned long	flags;

	bfad = bfad_sysfs_get_param(kobj, &ioc_inst, &vf_id, &vport_id, &rp);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_vport = bfa_fcs_vport_lookup(&bfad->bfa_fcs, vf_id, vport_id);
	if (fcs_vport == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		return -EINVAL;
	}

	if (strcmp(attr->name, "enable") == 0) {
		sscanf(buf, "%d", &value);
		if (value == 0) {
			/* STOP */
			if (bfa_fcs_vport_stop(fcs_vport) != BFA_STATUS_OK) {
				spin_unlock_irqrestore(&bfad->bfad_lock, flags);
				return -EINVAL;
			}
		} else if (value == 1) {
			/* START */
			if (bfa_fcs_vport_start(fcs_vport) != BFA_STATUS_OK) {
				spin_unlock_irqrestore(&bfad->bfad_lock, flags);
				return -EINVAL;
			}
		} else {
			spin_unlock_irqrestore(&bfad->bfad_lock, flags);
			return -EINVAL;
		}
	}
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	return count;
}

static struct sysfs_ops bfad_sysfs_vport_new_ops = {
	.show = bfad_sysfs_vport_new_show,
	.store = bfad_sysfs_vport_new_store,
};

static struct kobj_type ktype_vport_new = {
	.sysfs_ops = &bfad_sysfs_vport_new_ops,
	.default_attrs = bfad_sysfs_vport_new_attrs,
};

/**
 * @brief Function to display IOC attributes
 * @param[in] bfad  - IOC
 * @param[in] buf   - Buffer to be filled in with attributes
 * @retval - Number of bytes of 'buf' filled in.
 */
ssize_t
bfad_sysfs_attr_show(bfad_t *bfad, char *buf)
{
	int             len = 0;
	bfa_ioc_attr_t  attr;
	bfa_pport_attr_t pattr;
	unsigned long	flags;

	memset(&attr, 0, sizeof(bfa_ioc_attr_t));
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	bfa_hal_ioc_get_attr(&bfad->bfa_hal, &attr);
	bfa_hal_pport_get_attr(&bfad->bfa_hal, &pattr);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	strcpy(attr.driver_attr.fw_ver, attr.adapter_attr.fw_ver);

	memcpy(&attr.pci_attr, &bfad->pci_attr, sizeof(bfa_ioc_pci_attr_t));

	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %s\n",
			"Chip Revision", attr.chip_attr.chip_rev);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %s\n",
			"Manufacturer", attr.adapter_attr.manufacturer);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %s\n", "Serial#",
			attr.adapter_attr.serial_num);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %s\n",
			"Model Description", attr.adapter_attr.model_descr);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: ", "PWWN");
	len += bfad_sysfs_snprint_pwwn(pattr.pwwn, buf, len);
	len += snprintf(buf + len, PAGE_SIZE - len, "\n");
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: ", "NWWN");
	len += bfad_sysfs_snprint_pwwn(pattr.nwwn, buf, len);
	len += snprintf(buf + len, PAGE_SIZE - len, "\n");
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %s\n", "HW Ver",
			attr.adapter_attr.hw_ver);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %s\n", "FW Ver",
			attr.driver_attr.fw_ver);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %s\n", "Optrom Ver",
			attr.driver_attr.ob_ver);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %s\n", "Bios Ver",
			attr.driver_attr.bios_ver);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %s\n", "Ob Ver:",
			attr.driver_attr.ob_ver);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n", "Port Count",
			attr.adapter_attr.nports);

	return len;
}

/**
 * @brief Function to display IOC information
 * @param[in] bfad  - IOC
 * @param[in] buf   - Buffer to be filled in with attributes
 * @retval - Number of bytes of 'buf' filled in.
 */
ssize_t
bfad_sysfs_info_show(bfad_t *bfad, char *buf)
{
	int             len = 0;
	bfa_pport_attr_t pattr;
	bfa_ioc_attr_t  ioc_attr;
	unsigned long	flags;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	bfa_hal_pport_get_attr(&bfad->bfa_hal, &pattr);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: ", "PWWN");
	len += bfad_sysfs_snprint_pwwn(pattr.pwwn, buf, len);
	len += snprintf(buf + len, PAGE_SIZE - len, "\n");
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: ", "NWWN");
	len += bfad_sysfs_snprint_pwwn(pattr.nwwn, buf, len);
	len += snprintf(buf + len, PAGE_SIZE - len, "\n");

	bfa_hal_ioc_get_attr(&bfad->bfa_hal, &ioc_attr);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %s\n", "Name",
			"BFA");
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %s\n", "Serial No",
			ioc_attr.adapter_attr.serial_num);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %s\n", "PCI Name",
			bfad->pci_name);

	return len;
}

/**
 * @brief Function to display IOC statistics
 * @param[in] bfad  - IOC
 * @param[in] buf   - Buffer to be filled in with attributes
 * @retval - Number of bytes of 'buf' filled in.
 */
ssize_t
bfad_sysfs_stats_show(bfad_t *bfad, char *buf)
{
	int             len = 0;
	bfa_status_t    status;
	unsigned long   flags;
	bfa_ioc_stats_t stats;
	struct bfad_hal_comp fcomp;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	status = bfa_hal_ioc_get_stats(&bfad->bfa_hal, &stats,
				       bfad_sysfs_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (status != BFA_STATUS_OK) {
		return snprintf(buf + len, PAGE_SIZE - len,
				"Error getting IOC statistics\n");
	}

	wait_for_completion(&fcomp.comp);

	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n",
			"Heartbeat count", stats.hb_count);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n",
			"Disable Requests", stats.disable_reqs);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n",
			"Enable Requests", stats.enable_reqs);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n",
			"Disable Replies", stats.disable_replies);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n",
			"Enable Replies", stats.enable_replies);

	return len;
}

ssize_t
bfad_sysfs_stats_store(bfad_t *bfad, const char *buf)
{
	int             len = 0;
	bfa_status_t    status;
	unsigned long   flags;
	struct bfad_hal_comp fcomp;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	status = bfa_hal_ioc_clear_stats(&bfad->bfa_hal, bfad_sysfs_hcb_comp,
					 &fcomp);

	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (status != BFA_STATUS_OK)
		return -EINVAL;

	wait_for_completion(&fcomp.comp);

	return len;
}

ssize_t
bfad_sysfs_intr_show(bfad_t *bfad, char *buf)
{
	int             len = 0;

	return len;
}

ssize_t
bfad_sysfs_intr_store(bfad_t *bfad, const char *buf, int latency, int delay)
{
	bfa_ioc_intr_attr_t intr_attr;
	bfa_status_t	    status;
	unsigned long	flags;

	intr_attr.latency = latency;
	intr_attr.delay   = delay;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	status = bfa_hal_intr_attr_set(&bfad->bfa_hal, &intr_attr);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (status != BFA_STATUS_OK)
		return -EIO;

	return 0;
}

ssize_t
bfad_sysfs_power_show(bfad_t *bfad, char *buf)
{
	int             len = 0;

	return len;
}

ssize_t
bfad_sysfs_power_store(bfad_t *bfad, const char *buf)
{
	int             len = 0;

	return len;
}

/**
 * @brief Function to display mpiomode
 * @param[in] bfad  - IOC
 * @param[in] buf   - Buffer contains user input
 * @retval - 0 on SUCCESS
 * @retval - negative value on FAILURE
 */
ssize_t
bfad_sysfs_mpiomode_show(bfad_t *bfad, char *buf)
{
	bfa_boolean_t   mode;

	mode = bfa_hal_fcpim_mpio_mode_get(&bfad->bfa_hal);

	if (mode == BFA_TRUE)
		return snprintf(buf, PAGE_SIZE, "mpiomode is on\n");
	else
		return snprintf(buf, PAGE_SIZE, "mpiomode is off\n");

}

/**
 * @brief Function to change mpiomode
 * @param[in] bfad  - IOC
 * @param[in] buf   - Buffer contains user input
 * @retval - 0 on SUCCESS
 * @retval - negative value on FAILURE
 */
ssize_t
bfad_sysfs_mpiomode_store(bfad_t *bfad, const char *buf)
{
	bfa_boolean_t   mode;
	unsigned long	flags;

	if (strncmp(buf, "on", strlen(buf) - 1) == 0) {
		mode = BFA_TRUE;
	} else if (strncmp(buf, "off", strlen(buf) - 1) == 0) {
		mode = BFA_FALSE;
	} else {
		return -EIO;
	}

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	bfa_hal_fcpim_mpio_mode_set(&bfad->bfa_hal, mode, 0);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	return 0;
}

/**
 * @brief Function to get options based on user input while creating VPORT
 * @param[in] buf  		 - User input
 * @param[out] vport_id  - VPORT ID
 * @param[out] port_cfg  - Contains pwwn(mandatory input),
 *                         nwwn/symbolic name/roles (if provided)
 * @param[out] wwn_str   - wwn as a string
 * @retval - BFA_TRUE if parsing is successful
 * @retval - BFA_FALSE if parsing fails
 */
bfa_boolean_t
bfad_sysfs_parse_vport_options(const char *buf, uint64_t *vport_id,
			       bfa_port_cfg_t *port_cfg, char *wwn_str)
{
	char *option, *value, str[100], *p, *dash = "-", *colon = ":";

	memset(port_cfg, 0, sizeof(bfa_port_cfg_t));

	sscanf(buf, "%s", wwn_str);
	if (strstr(wwn_str, colon) == NULL)
		return BFA_FALSE;

	if (bfad_sysfs_get_port_wwn(wwn_str, &port_cfg->pwwn) == BFA_FALSE)
		return BFA_FALSE;

	memcpy(vport_id, &port_cfg->pwwn, sizeof(uint64_t));

	if (strstr(buf, dash) == NULL)
		return BFA_TRUE;

	strcpy(str, strstr(buf, dash));

	p = str;
	if (!p)
		return BFA_TRUE;

	option = strsep(&p, " ");
	if (option == NULL)
		return BFA_TRUE;

	while ((value = strsep(&p, " "))) {
		if (strcmp(option, "-n") == 0) {
			if (bfad_sysfs_get_port_wwn(wwn_str, &port_cfg->nwwn) ==
			    BFA_FALSE)
				return BFA_FALSE;
		} else if (strcmp(option, "-s") == 0) {
			strcpy(port_cfg->sym_name.symname, value);
		} else if (strcmp(option, "-r") == 0) {
			if (strncmp("fcpim", value, 5) == 0)
				port_cfg->roles = BFA_PORT_ROLE_FCP_IM;
			else if (strncmp("fcprm", value, 5) == 0)
				port_cfg->roles = BFA_PORT_ROLE_FCP_TM;
			else if (strncmp("ipfc", value, 4) == 0)
				port_cfg->roles = BFA_PORT_ROLE_FCP_IPFC;
		} else
			return BFA_TRUE;

		option = strsep(&p, " ");
		if (option == NULL)
			break;
	}

	return BFA_TRUE;
}

static          ssize_t
bfad_sysfs_bfa_show(struct kobject *kobj, struct attribute *attr, char *buf)
{
	return 0;
}

/**
 * @brief Function to add a sysfs directory entry for newly created VPORT
 * (lport). This function will be called when VPORT is created through
 * SYSFS (bfad_sysfs.c) (OR) when user creates VPORT through CLI(bfad_ioctl.c).
 * @param[in] vf_id - VF ID under which the VPORT is created. If the value is
 * 0, then the VPORT is created under the base fabric.
 * @param[in] ioc_inst - IOC instance
 * @param[in] port_cfg - Port configuration
 * @retval - 0 if SUCCESS
 * @retval - negative value in case of FAILURE
 */

int
bfad_os_sysfs_lport_add_dir(bfad_t *bfad, int vf_id, int ioc_inst,
			    bfa_port_cfg_t port_cfg)
{
	bfad_kobj_t    *bfad_kobj, *bfad_vf_kobj = NULL;
	struct kobject *kobj;
	bfa_pport_attr_t pattr;
	struct list_head *kobj_vport_list;
	char           *ioc_path;
	int             i;
	unsigned long	flags;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	bfa_hal_pport_get_attr(&bfad->bfa_hal, &pattr);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	ioc_path =
		bfad_sysfs_private_kobject_get_path(&bfad->pcidev->dev.kobj,
						    GFP_ATOMIC);
	if (vf_id != 0) {
		bfad_vf_kobj =
			bfad_sysfs_kobj_get(ioc_path, KOBJ_TYPE_VF, vf_id, 0,
					&bfad_kobj_vf_list);
		if (bfad_vf_kobj == NULL) {
			kfree(ioc_path);
			return -EINVAL;
		}

		kobj_vport_list = &bfad_vf_kobj->kobj_vport_list;
	} else {
		kobj_vport_list = &bfad_kobj_base_list;
	}

	/*
	 * Check if the sysfs entry if the vport already exists
	 */
	if (bfad_sysfs_kobj_get
	    (ioc_path, KOBJ_TYPE_VPORT, vf_id, port_cfg.pwwn,
	     kobj_vport_list) != NULL) {
		kfree(ioc_path);
		return -EEXIST;
	}
	kfree(ioc_path);

	/*
	 * Create new one
	 */
	bfad_kobj = kzalloc(sizeof(bfad_kobj_t), GFP_KERNEL);
	if (!bfad_kobj)
		return -ENOMEM;

	/*
	 * Create a folder for the new VPORT
	 */
	bfad_kobj->type = KOBJ_TYPE_VPORT;
	bfad_kobj->inst = ioc_inst;
	bfad_kobj->vf_id = vf_id;
	bfad_kobj->pwwn = pattr.pwwn;
	bfad_kobj->vport_id = port_cfg.pwwn;

	kobj = &bfad_kobj->kobj;
	if (vf_id != 0)
		kobj->parent = &bfad_vf_kobj->kobj;
	else {
		i = find_kobj_by_path(&bfad->pcidev->dev.kobj);
		kobj->parent = &ko_bfa_root[i];
	}
	sprintf(kobj->name, "vp-");
	bfad_sysfs_print_pwwn(port_cfg.pwwn, &kobj->name[3]);
	kobj->ktype = &ktype_vport_new;
	if (kobject_register(kobj) == -EINVAL) {
		kfree(bfad_kobj);
		return -EINVAL;
	}

	/*
	 * Add the new vport to the list
	 */
	list_add_tail(&bfad_kobj->kobj_list_entry, kobj_vport_list);

	return 0;
}

/**
 * @brief Function to delete a sysfs directory entry of a VPORT (lport)
 * This function will be called when VPORT is deleted through
 * SYSFS (bfad_sysfs.c) (OR) when user deletes VPORT through CLI(bfad_ioctl.c).
 * @param[in] vf_id - VF ID under which the VPORT is created. If the value is
 * 0, then the VPORT is created under the base fabric.
 * @param[in] ioc_inst - IOC instance
 * @param[in] pwwn  - PWWN of the VPORT to be deleted
 * @retval - 0 if SUCCESS
 * @retval - negative value in case of FAILURE
 */
int
bfad_os_sysfs_lport_delete_dir(bfad_t *bfad, int vf_id, int ioc_inst,
			       wwn_t vpwwn)
{
	bfad_kobj_t    *bfad_vf_kobj;
	struct list_head *kobj_vport_list;
	char           *ioc_path;
	bfa_pport_attr_t pattr;
	unsigned long	flags;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	bfa_hal_pport_get_attr(&bfad->bfa_hal, &pattr);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	ioc_path =
		bfad_sysfs_private_kobject_get_path(&bfad->pcidev->dev.kobj,
						    GFP_ATOMIC);
	if (vf_id != 0) {
		bfad_vf_kobj =
			bfad_sysfs_kobj_get(ioc_path, KOBJ_TYPE_VF, vf_id, 0,
					&bfad_kobj_vf_list);
		if (bfad_vf_kobj == NULL) {
			kfree(ioc_path);
			return -EINVAL;
		}
		kobj_vport_list = &bfad_vf_kobj->kobj_vport_list;
	} else {
		kobj_vport_list = &bfad_kobj_base_list;
	}

	/*
	 * Delete all rports of the vport
	 */
	bfad_sysfs_delete_all_rports_of_vport(bfad, KOBJ_TYPE_ITNIM, vf_id,
					      pattr.pwwn, vpwwn,
					      kobj_vport_list);
	bfad_sysfs_delete_all_rports_of_vport(bfad, KOBJ_TYPE_RPORT, vf_id,
					      pattr.pwwn, vpwwn,
					      kobj_vport_list);

	bfad_sysfs_kobj_del_port(ioc_path, KOBJ_TYPE_VPORT, vf_id, vpwwn,
				 kobj_vport_list);

	kfree(ioc_path);
	return 0;
}

/**
 * @brief Function to create/delete VPORTs under the base fabric
 * @param[in] kobj  - kobject of the base fabric('bfa') directory
 * @param[in] attr  - Attributes of the directory
 * @param[in] buf   - Buffer contains user input
 * @param[in] count - Number of bytes in 'buf'
 * @retval count - Number of bytes of 'buf' used in the function
 */
static          ssize_t
bfad_sysfs_bfa_store(struct kobject *kobj, struct attribute *attr,
		     const char *buf, size_t count)
{
	int             ioc_inst = 1, ret;
	unsigned long   flags;
	uint64_t        vport_id, rp;
	bfad_t         *bfad;
	char            wwn_str[100];
	int             vf_id = 0;
	bfad_vport_t   *bfad_vport;
	bfa_status_t    status;
	bfa_port_cfg_t  port_cfg;
	bfa_fcs_vport_t *fcs_vport = NULL;
	struct completion fcomp;

	bfad = bfad_sysfs_get_param(kobj, &ioc_inst, &vf_id, &vport_id, &rp);
	if (bfad == NULL)
		return -EINVAL;

	if (bfad_sysfs_parse_vport_options(buf, &vport_id, &port_cfg, wwn_str)
	    == BFA_FALSE)
		return -EINVAL;

	if (strcmp(attr->name, "create_vport") == 0) {

		if (port_cfg.roles != BFA_PORT_ROLE_FCP_IM
		    && port_cfg.roles != BFA_PORT_ROLE_FCP_TM)
			return -EINVAL;

		/*
		 * Create the VPORT
		 */
		status = bfad_vport_create(bfad, 0 /* vf_id */, &port_cfg);
		if (status != BFA_STATUS_OK)
			return -EINVAL;

		ret = bfad_os_sysfs_lport_add_dir(bfad, 0, ioc_inst, port_cfg);
		if (ret < 0)
			return ret;

	} else if (strcmp(attr->name, "delete_vport") == 0) {
		spin_lock_irqsave(&bfad->bfad_lock, flags);
		fcs_vport = bfa_fcs_vport_lookup(&bfad->bfa_fcs, 0 /* VF_ID */ ,
						port_cfg.pwwn);
		if (fcs_vport == NULL) {
			spin_unlock_irqrestore(&bfad->bfad_lock, flags);
			return -EINVAL;
		}

		bfad_vport = fcs_vport->vport_drv;

		bfad_vport->comp_del = &fcomp;
		init_completion(bfad_vport->comp_del);

		status = bfa_fcs_vport_delete(fcs_vport);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);

		wait_for_completion(bfad_vport->comp_del);

		if (status == BFA_STATUS_OK)
			bfad_os_sysfs_lport_delete_dir(bfad, 0, ioc_inst,
						       vport_id);
	}

	return count;
}

DEF_ATTR(bfa_lport_create, "create_vport");
DEF_ATTR(bfa_lport_delete, "delete_vport");

static struct attribute *bfad_sysfs_bfa_attrs[] = {
	&bfa_lport_create,
	&bfa_lport_delete,
	NULL,
};

static struct sysfs_ops bfad_sysfs_bfa_ops = {
	.show = bfad_sysfs_bfa_show,
	.store = bfad_sysfs_bfa_store,
};

static struct kobj_type ktype_bfa = {
	.sysfs_ops = &bfad_sysfs_bfa_ops,
	.default_attrs = bfad_sysfs_bfa_attrs,
};

/*
 * Configured lport attributes
 */

DEF_ATTR(lport_attr, "attr");
DEF_ATTR(lport_stats, "stats");
DEF_ATTR(lport_refresh_rports, "refresh_rports");

static struct attribute *bfad_sysfs_lport_attrs[] = {
	&lport_attr,
	&lport_stats,
	&lport_refresh_rports,
	NULL
};

ssize_t
bfad_sysfs_lport_show_attr(bfad_t *bfad, uint16_t vf_id, char *buf,
			   bfa_port_attr_t *attr)
{
	int             len = 0;
	unsigned long	flags;

	/*
	 * attr.port_attr has all the info about the particular physical lport
	 */
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: ", "Port FC Addr");
	if (attr->pid)
		len += snprintf(buf + len, PAGE_SIZE - len, "%06x\n",
				bfa_os_ntoh3b(attr->pid));
	else
		len += snprintf(buf + len, PAGE_SIZE - len, "--\n");

	if (attr->port_type != BFA_PPORT_TYPE_VPORT) {
		bfa_pport_attr_t pp_attr;
		bfa_port_attr_t port_attr;

		spin_lock_irqsave(&bfad->bfad_lock, flags);
		bfa_hal_pport_get_attr(&bfad->bfa_hal, &pp_attr);
		bfa_fcs_port_get_attr(&bfad->bfa_fcs.fabric.bport, &port_attr);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);

		if (pp_attr.topology != BFA_PPORT_TOPOLOGY_NONE)
			pp_attr.pid = port_attr.pid;
		else
			pp_attr.pid = 0;

		pp_attr.port_type = port_attr.port_type;

		len += snprintf(buf + len, PAGE_SIZE - len,
				"%-20s: (State: %s)\n", "Base Port",
				bfad_sysfs_state_to_string(pp_attr.port_state,
							   port_state_table));
	} else {
		bfa_fcs_vport_t *fcs_vport;
		bfa_vport_attr_t vp_attr;

		spin_lock_irqsave(&bfad->bfad_lock, flags);
		fcs_vport =
			bfa_fcs_vport_lookup(&bfad->bfa_fcs, vf_id,
					     attr->port_cfg.pwwn);
		if (fcs_vport == NULL) {
			spin_unlock_irqrestore(&bfad->bfad_lock, flags);
			len += snprintf(buf + len, PAGE_SIZE - len,
					"Could not find fcs_vport\n");
			return len;
		}

		bfa_fcs_vport_get_attr(fcs_vport, &vp_attr);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);

		len += snprintf(buf + len, PAGE_SIZE - len,
				"%-20s: (State: %s)\n", "Virtual Port",
				bfad_sysfs_state_to_string(vp_attr.vport_state,
							   vport_state_table));
	}

	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: ", "Port PWWN");
	len += bfad_sysfs_snprint_pwwn(attr->port_cfg.pwwn, buf, len);
	len += snprintf(buf + len, PAGE_SIZE - len, "\n");
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: ", "Port NWWN");
	len += bfad_sysfs_snprint_pwwn(attr->port_cfg.nwwn, buf, len);
	len += snprintf(buf + len, PAGE_SIZE - len, "\n");
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %s\n",
			"Symbolic name", attr->port_cfg.sym_name.symname);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: ", "Role");
	switch (attr->port_cfg.roles) {
	case BFA_PORT_ROLE_FCP_IM:
		len += snprintf(buf + len, PAGE_SIZE - len, "FCP Initiator\n");
		break;
	case BFA_PORT_ROLE_FCP_TM:
		len += snprintf(buf + len, PAGE_SIZE - len, "FCP Target\n");
		break;
	case BFA_PORT_ROLE_FCP_IPFC:
		len += snprintf(buf + len, PAGE_SIZE - len, "IP over FC\n");
		break;
	default:
		len += snprintf(buf + len, PAGE_SIZE - len, "Unknown: %d\n",
				attr->port_cfg.roles);
		break;
	}

	return len;
}

ssize_t
bfad_sysfs_lport_show_stats(char *buf, bfa_port_stats_t *stats, int len)
{
	len += snprintf(buf + len, PAGE_SIZE - len, "%-25s: %d\n",
			"NS plogi sent", stats->ns_plogi_sent);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-25s: %d\n",
			"NS plogi accepts", stats->ns_plogi_accepts);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-25s: %d\n",
			"NS plogi rsp errors", stats->ns_plogi_rsp_err);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-25s: %d\n",
			"NS plogi accept errors", stats->ns_plogi_acc_err);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-25s: %d\n",
			"NS plogi rejects", stats->ns_rejects);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-25s: %d\n",
			"NS plogi unknown rsp", stats->ns_plogi_unknown_rsp);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-25s: %d\n",
			"NS plogi alloc wait", stats->ns_plogi_alloc_wait);
	len += snprintf(buf + len, PAGE_SIZE - len, "\n");
	len += snprintf(buf + len, PAGE_SIZE - len, "%-25s: %d\n",
			"NS RSPN_ID sent", stats->ns_rspnid_sent);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-25s: %d\n",
			"NS RSPN_ID accepts", stats->ns_rspnid_accepts);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-25s: %d\n",
			"NS RSPN_ID rsp errors", stats->ns_rspnid_rsp_err);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-25s: %d\n",
			"NS RSPN_ID rejects", stats->ns_rspnid_rejects);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-25s: %d\n",
			"NS RSPN_ID alloc wait", stats->ns_rspnid_alloc_wait);
	len += snprintf(buf + len, PAGE_SIZE - len, "\n");
	len += snprintf(buf + len, PAGE_SIZE - len, "%-25s: %d\n",
			"NS RFT_ID sent", stats->ns_rftid_sent);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-25s: %d\n",
			"NS RFT_ID accepts", stats->ns_rftid_accepts);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-25s: %d\n",
			"NS RFT_ID rsp errors", stats->ns_rftid_rsp_err);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-25s: %d\n",
			"NS RFT_ID rejects", stats->ns_rftid_rejects);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-25s: %d\n",
			"NS RFT_ID alloc wait", stats->ns_rftid_alloc_wait);
	len += snprintf(buf + len, PAGE_SIZE - len, "\n");
	len += snprintf(buf + len, PAGE_SIZE - len, "%-25s: %d\n",
			"NS GID_FT sent", stats->ns_gidft_sent);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-25s: %d\n",
			"NS GID_FT accepts", stats->ns_gidft_accepts);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-25s: %d\n",
			"NS GID_FT rsp errors", stats->ns_gidft_rsp_err);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-25s: %d\n",
			"NS GID_FT rejects", stats->ns_gidft_rejects);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-25s: %d\n",
			"NS GID_FT unknown rsp", stats->ns_gidft_unknown_rsp);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-25s: %d\n",
			"NS GID_FT alloc wait", stats->ns_gidft_alloc_wait);
	len += snprintf(buf + len, PAGE_SIZE - len, "\n");
	len += snprintf(buf + len, PAGE_SIZE - len, "%-25s: %d\n",
			"NS timeouts", stats->ns_timeouts);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-25s: %d\n", "NS retries",
			stats->ns_retries);
	len += snprintf(buf + len, PAGE_SIZE - len, "\n");
	len += snprintf(buf + len, PAGE_SIZE - len, "%-25s: %d\n", "Num RSCNs",
			stats->num_rscn);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-25s: %d\n",
			"Num port_id RSCNs", stats->num_portid_rscn);

	return len;
}

/**
 * @brief Function to display LPORT attributes/statistics
 * @param[in] kobj  - kobject of the LPORT directory
 * @param[in] buf   - Buffer to be filled in with attributes/statistics
 * @retval - Number of bytes of 'buf' filled in.
 */
static          ssize_t
bfad_sysfs_lport_show(struct kobject *kobj, struct attribute *attr, char *buf)
{
	int             ioc_inst = 1, len = 0;
	bfad_t         *bfad;
	int             vf_id = 0;
	wwn_t           pwwn, rp;
	bfa_fcs_port_t *fcs_port;
	bfa_port_attr_t pattr;
	bfa_port_stats_t stats;
	unsigned long	flags;

	bfad = bfad_sysfs_get_param(kobj, &ioc_inst, &vf_id, &pwwn, &rp);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs, vf_id, pwwn);
	if (fcs_port == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		return snprintf(buf + len, PAGE_SIZE - len,
				"failed to locate lport\n");
	}

	if (strcmp(attr->name, "attr") == 0) {
		bfa_fcs_port_get_attr(fcs_port, &pattr);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		if (vf_id == 0)
			len += snprintf(buf + len, PAGE_SIZE - len,
					"Base Port Attributes\n");
		else
			len += snprintf(buf + len, PAGE_SIZE - len,
					"Base Port of vf-%X Attributes\n",
					vf_id);
		len += bfad_sysfs_lport_show_attr(bfad, vf_id, buf + len,
						  &pattr);
	} else if (strcmp(attr->name, "stats") == 0) {
		bfa_fcs_port_get_stats(fcs_port, &stats);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		if (vf_id == 0)
			len += snprintf(buf + len, PAGE_SIZE - len,
					"Base Port Statistics\n");
		else
			len += snprintf(buf + len, PAGE_SIZE - len,
					"Base Port of vf-%X Statistics\n",
					vf_id);
		len += bfad_sysfs_lport_show_stats(buf, &stats, len);
	} else if (strcmp(attr->name, "refresh_rports") == 0) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		len += bfad_sysfs_rport_main_show_list(bfad, kobj, vf_id, pwwn,
						       0, buf);
	}
	else {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	}

	return len;
}

static          ssize_t
bfad_sysfs_lport_store(struct kobject *kobj, struct attribute *attr,
		       const char *buf, size_t count)
{
	int             ioc_inst = 1;
	bfad_t         *bfad;
	int             vf_id = 0;
	wwn_t           pwwn, rpwwn;
	bfa_fcs_port_t *fcs_port;
	unsigned long	flags;

	if (strcmp(attr->name, "stats") == 0) {
		bfad = bfad_sysfs_get_param(kobj, &ioc_inst, &vf_id, &pwwn,
			&rpwwn);
		if (bfad == NULL)
			return -EINVAL;

		spin_lock_irqsave(&bfad->bfad_lock, flags);
		fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs, vf_id, pwwn);
		if (fcs_port == NULL) {
			spin_unlock_irqrestore(&bfad->bfad_lock, flags);
			return -EINVAL;
		}

		bfa_fcs_port_clear_stats(fcs_port);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	}

	return count;
}

static struct sysfs_ops bfad_sysfs_lport_ops = {
	.show = bfad_sysfs_lport_show,
	.store = bfad_sysfs_lport_store,
};

static struct kobj_type ktype_lport = {
	.sysfs_ops = &bfad_sysfs_lport_ops,
	.default_attrs = bfad_sysfs_lport_attrs,
};

#ifdef VF_READY
/*
 * Configured New VF ID attributes
 */
DEF_ATTR(vf_new_attr, "attr");
DEF_ATTR(vf_new_stats, "stats");
DEF_ATTR(vf_new_lport_create, "create_vport");
DEF_ATTR(vf_new_lport_delete, "delete_vport");

static struct attribute *bfad_sysfs_vf_new_attrs[] = {
	&vf_new_attr,
	&vf_new_stats,
	&vf_new_lport_create,
	&vf_new_lport_delete,
	NULL
};

/**
 * @brief Function to display attributes/statistics of a VF
 * @param[in] kobj  - kobject of the (vf) directory
 * @param[in] attr  - Attributes of the directory
 * @param[in] buf   - Buffer contains user input
 * @retval - 0 if SUCCESS
 * @retval - negative value in case of FAILURE
 */
static          ssize_t
bfad_sysfs_vf_new_show(struct kobject *kobj, struct attribute *attr, char *buf)
{
	int             ioc_inst = 1;
	bfad_t         *bfad;
	bfa_vf_attr_t   vf_attr;
	bfa_vf_stats_t  stats;
	int             vf_id, len = 0;
	bfa_fcs_vf_t   *fcs_vf;
	uint64_t        dummy_vport_id, dummy_rport_id;
	unsigned long	flags;

	char           *vf_state_str[] = {
		"fabric is not yet initialized",	/* 0 */
		"link is down",	/* 1 */
		"flogi is in progress",	/* 2 */
		"authentication in progress",	/* 3 */
		"fabric is not present",	/* 4 */
		"login to fabric is complete",	/* 5 */
		"EVFP is in progress",	/* 6 */
		"port isolated due to vf_id mismatch",	/* 7 */
	};

	bfad = bfad_sysfs_get_param(kobj, &ioc_inst, &vf_id, &dummy_vport_id,
		&dummy_rport_id);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_vf = bfa_fcs_vf_lookup(&bfad->bfa_fcs, vf_id)
	if (fcs_vf == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		len += snprintf(buf + len, PAGE_SIZE - len,
				"failed to locate vf_id\n");
		return len;
	}

	if (strcmp(attr->name, "attr") == 0) {
		bfa_fcs_vf_get_attr(fcs_vf, &vf_attr);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %s",
				"VF state\n",
				(vf_attr.state <=
				7) ? vf_state_str[vf_attr.state] : "Unknown");
		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: ", "PWWN");
		len += bfad_sysfs_snprint_pwwn(vf_attr.fabric_name, buf, len);
		len += snprintf(buf + len, PAGE_SIZE - len, "\n");

		return len;
	} else if (strcmp(attr->name, "stats") == 0) {
		memset(&stats, 0, sizeof(bfa_vf_stats_t));
		bfa_fcs_vf_get_stats(fcs_vf, &stats);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n",
				"flogi attempts", stats.flogi_attempts);
		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n",
				"flogi timeouts", stats.flogi_timeouts);
		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n",
				"flogi rejects", stats.flogi_rejects);
		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n",
				"auth attempts", stats.auth_attempts);
		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n",
				"auth timeouts", stats.auth_timeouts);
		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n",
				"auth rejects", stats.auth_rejects);
		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n",
				"auth failures", stats.auth_failures);
		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n",
				"evfp attempts", stats.evfp_attempts);
		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n",
				"evfp failures", stats.evfp_failures);
		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n",
				"evfp success", stats.evfp_success);
		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n",
				"evfp isolations", stats.evfp_isolations);
		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n",
				"evfp inbounds", stats.evfp_inbounds);
		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n",
				"fabric offline", stats.fabric_offline);
		len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n",
				"fabric onlines", stats.fabric_onlines);

		return len;
	}
	else {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	}

	return 0;
}

/**
 * @brief Function to create/delete VPORT (lport) under a VF
 * @param[in] kobj  - kobject of the (vf) directory
 * @param[in] attr  - Attributes of the directory
 * @param[in] buf   - Buffer contains user input
 * @param[in] count - Size of buf
 * @retval - 'count' if SUCCESS
 * @retval - negative value in case of FAILURE
 */
static          ssize_t
bfad_sysfs_vf_new_store(struct kobject *kobj, struct attribute *attr,
			const char *buf, size_t count)
{
	int             ioc_inst = 1, ret;
	uint64_t        vport_id, rp;
	bfad_t         *bfad;
	char            wwn_str[100];
	int             vf_id = 0;
	bfa_port_cfg_t  port_cfg;
	bfa_fcs_vport_t *fcs_vport;

	bfad = bfad_sysfs_get_param(kobj, &ioc_inst, &vf_id, &vport_id, &rp);
	if (bfad == NULL)
		return -EINVAL;

	if (bfad_sysfs_parse_vport_options(buf, &vport_id, &port_cfg, wwn_str)
	    == BFA_FALSE)
		return -EINVAL;

	if (strcmp(attr->name, "create_vport") == 0) {

		/*
		 * Create the VPORT
		 */
		if ((bfad_vport_create(bfad, vf_id, &port_cfg))
		    != BFA_STATUS_OK)
			return -EINVAL;

		ret = bfad_os_sysfs_lport_add_dir(bfad, vf_id, ioc_inst,
						  port_cfg);
		if (ret < 0)
			return ret;

	} else if (strcmp(attr->name, "delete_vport") == 0) {
		fcs_vport = bfa_fcs_vport_lookup(&bfad->bfa_fcs, vf_id,
						port_cfg.pwwn);
		if (fcs_vport == NULL)
			return -EINVAL;
		if (bfa_fcs_vport_delete(fcs_vport) != BFA_STATUS_OK)
			bfad_os_sysfs_lport_delete_dir(bfad, vf_id, ioc_inst,
						       vport_id);
	}

	return count;
}

static struct sysfs_ops bfad_sysfs_vf_new_ops = {
	.show = bfad_sysfs_vf_new_show,
	.store = bfad_sysfs_vf_new_store,
};

static struct kobj_type ktype_vf_new = {
	.sysfs_ops = &bfad_sysfs_vf_new_ops,
	.default_attrs = bfad_sysfs_vf_new_attrs,
};

/**
 * @brief Function to display list of VFs
 * @param[in] bfad_fcs_t  - Port' FCS
 * @param[in] nvfs   - Number of VFs to display
 * @param[out] buf   - Buffer to hold VFs to display
 * @retval - Number of bytes of 'buf' filled in.
 */
static          ssize_t
bfad_sysfs_get_vf_list(bfa_fcs_t *bfa_fcs, int nvfs, char *buf)
{
#define MAX_VF_IDS	100
	int             i, len = 0;
	uint16_t       *vf_ids;
	unsigned long	flags;

	if (nvfs == 0 || nvfs > MAX_VF_IDS)
		nvfs = MAX_VF_IDS;

	vf_ids = kzalloc(sizeof(uint16_t) * nvfs, GFP_KERNEL);
	if (vf_ids == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	if (nvfs != MAX_VF_IDS)
		bfa_fcs_vf_list(bfa_fcs, vf_ids, &nvfs);
	else
		bfa_fcs_vf_list_all(bfa_fcs, vf_ids, &nvfs);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	for (i = 0; i < nvfs; i++)
		len += snprintf(buf + len, PAGE_SIZE - len, "%d\n", vf_ids[i]);

	kfree(vf_ids);
	return len;
}

/**
 * @brief Function to display VF attributes
 * @param[in] kobj  - kobject of the (VF) directory
 * @param[in] attr  - Attributes of the directory
 * @param[in] buf   - Buffer contains user input
 * @retval - Number of bytes of 'buf' filled in the function
 */
static          ssize_t
bfad_sysfs_vf_show(struct kobject *kobj, struct attribute *attr, char *buf)
{
	int             ioc_inst = 1, nvfs = 20 /* HACK ALERT */ ;
	bfad_t         *bfad;
	char           *path;

	path = bfad_sysfs_private_kobject_get_path(kobj, GFP_ATOMIC);
	if (strstr(path, ".0/bfa") != NULL)
		ioc_inst = 0;

	kfree(path);

	bfad = bfad_find_bfad_by_inst_no(ioc_inst);
	if (bfad == NULL)
		return -EINVAL;

	if (strcmp(attr->name, "list") == 0) {
		return bfad_sysfs_get_vf_list(&bfad->bfa_fcs, nvfs, buf);
	} else if (strcmp(attr->name, "listall") == 0) {
		return bfad_sysfs_get_vf_list(&bfad->bfa_fcs, 0, buf);
	}

	return 0;
}

/**
 * @brief Function to delete a sysfs directory entry of a VF
 * This function will be called when VF is deleted through
 * SYSFS (bfad_sysfs.c) (OR) when user deletes VF through CLI(bfad_ioctl.c).
 * @param[in] bfad     - IOC
 * @param[in] vf_id    - VF ID of the directory to be deleted
 * @retval - 0
 */
int
bfad_os_sysfs_vf_delete_dir(bfad_t *bfad, uint16_t vf_id)
{
	char           *ioc_path;

	ioc_path =
		bfad_sysfs_private_kobject_get_path(&bfad->pcidev->dev.kobj,
						    GFP_ATOMIC);
	/**search for kobj in the list and unregister it */
	bfad_sysfs_kobj_del_vf(ioc_path, vf_id, &bfad_kobj_vf_list);

	/*
	 * delete the new vf itself
	 */
	bfad_sysfs_kobj_del_port(ioc_path, KOBJ_TYPE_VF, vf_id, 0,
				 &bfad_kobj_vf_list);

	kfree(ioc_path);
	return 0;
}

/**
 * @brief Function to add a sysfs directory entry for newly created VF
 * (lport). This function will be called when VF is created through
 * SYSFS (bfad_sysfs.c) (OR) when user creates VF through CLI(bfad_ioctl.c).
 * @param[in] ioc_inst - IOC instance
 * @param[in] vf_id    - VF ID of the directory to be created
 * @retval - 0 if SUCCESS
 * @retval - negative value in case of FAILURE
 */
int
bfad_os_sysfs_vf_add_dir(bfad_t *bfad, int ioc_inst, uint16_t vf_id)
{
	struct kobject *kobj;
	bfad_kobj_t    *bfad_kobj, *bfad_lport_kobj;
	char           *ioc_path, *path;
	int             i;

	/*
	 * Check if the sysfs entry if the vf already exists
	 */
	ioc_path =
		bfad_sysfs_private_kobject_get_path(&bfad->pcidev->dev.kobj,
						    GFP_ATOMIC);
	list_for_each_entry(bfad_kobj, &bfad_kobj_vf_list, kobj_list_entry) {
		if (bfad_kobj->vf_id == vf_id) {
			path = bfad_sysfs_private_kobject_get_path(&bfad_kobj->
								   kobj,
								   GFP_ATOMIC);
			if (strstr(path, ioc_path)) {
				kfree(ioc_path);
				kfree(path);
				return -EEXIST;
			}
			kfree(path);
		}
	}
	kfree(ioc_path);

	/*
	 * Create sub-directory for the new VF under VF Root
	 */
	bfad_kobj = kzalloc(sizeof(bfad_kobj_t), GFP_KERNEL);
	if (!bfad_kobj)
		return -ENOMEM;

	/*
	 * Create a folder for the new VF
	 */
	bfad_kobj->type = KOBJ_TYPE_VF;
	bfad_kobj->inst = ioc_inst;
	bfad_kobj->vf_id = vf_id;
	bfad_kobj->vport_id = 0;

	i = find_kobj_by_path(&bfad->pcidev->dev.kobj);

	kobj = &bfad_kobj->kobj;
	kobj->parent = &ko_vf[i];
	sprintf(kobj->name, "vf-%X", vf_id);
	kobj->ktype = &ktype_vf_new;
	if (kobject_register(kobj) == -EINVAL) {
		kfree(bfad_kobj);
		return -EINVAL;
	}
	list_add_tail(&bfad_kobj->kobj_list_entry, &bfad_kobj_vf_list);

	INIT_LIST_HEAD(&bfad_kobj->kobj_vport_list);

	/*
	 * Create a folder for lport under the new VF
	 */
	bfad_lport_kobj = kzalloc(sizeof(bfad_kobj_t), GFP_KERNEL);
	if (!bfad_lport_kobj) {
		kobject_unregister(&bfad_kobj->kobj);
		list_del(&bfad_kobj->kobj_list_entry);
		kfree(bfad_kobj);
		return -ENOMEM;
	}

	bfad_lport_kobj->type = KOBJ_TYPE_LPORT;
	bfad_lport_kobj->inst = ioc_inst;
	bfad_lport_kobj->vf_id = vf_id;
	bfad_lport_kobj->vport_id = 0;

	kobj = &bfad_lport_kobj->kobj;
	kobj->parent = &bfad_kobj->kobj;
	sprintf(kobj->name, "lport-0");
	kobj->ktype = &ktype_lport;
	if (kobject_register(kobj) == -EINVAL) {
		kobject_unregister(&bfad_kobj->kobj);
		list_del(&bfad_kobj->kobj_list_entry);
		kfree(bfad_lport_kobj);
		kfree(bfad_kobj);
		return -EINVAL;
	}

	/**add kobjs to the list */
	list_add_tail(&bfad_lport_kobj->kobj_list_entry,
		      &bfad_kobj->kobj_vport_list);

	return 0;

}

/**
 * @brief Function to add/delete, enable/disable VFs
 * @param[in] kobj  - kobject of the 'vf' directory <ROOT/INST/bfa/vf>
 * @param[in] attr  - Attributes of the directory
 * @param[in] buf   - Buffer contains user input
 * @param[in] count - Number of bytes in 'buf'
 * @retval count - Number of bytes of 'buf' used in the function
 */
static          ssize_t
bfad_sysfs_vf_store(struct kobject *kobj, struct attribute *attr,
		    const char *buf, size_t count)
{
	int             ioc_inst = 1, enable;
	bfad_t         *bfad;
	int             vf_id = 0;
	uint64_t        dummy_vport_id, dummy_rport_id;
	bfa_fcs_vf_t   *fcs_vf;
	bfa_status_t    status;
	bfa_port_cfg_t  port_cfg;
	unsigned long	flags;

	bfad = bfad_sysfs_get_param(kobj, &ioc_inst, &vf_id, &dummy_vport_id,
		&dummy_rport_id);
	if (bfad == NULL)
		return -EINVAL;

	if (strcmp(attr->name, "enable") == 0) {
		sscanf(buf, "%d %x", &enable, &vf_id);
		if (enable != 0 || enable != 1)
			return -EINVAL;

		spin_lock_irqsave(&bfad->bfad_lock, flags);
		if (enable == 0) {
			/* DISABLE VF */
			if (bfa_fcs_vf_mode_disable(&bfad->bfa_fcs)
					 != BFA_STATUS_OK) {
				spin_unlock_irqrestore(&bfad->bfad_lock, flags);
				return -EINVAL;
			}
		} else {	/* ENABLE VF */
			if (vf_id == 0)
				vf_id = FC_VF_ID_NULL;

			if (bfa_fcs_vf_mode_enable(&bfad->bfa_fcs, vf_id)
			    != BFA_STATUS_OK) {
				spin_unlock_irqrestore(&bfad->bfad_lock, flags);
				return -EINVAL;
			}
		}
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	} else if (strcmp(attr->name, "add") == 0) {
		sscanf(buf, "%x", &vf_id);
		spin_lock_irqsave(&bfad->bfad_lock, flags);
		if (bfad_vf_create(bfad, f_id, &port_cfg) == BFA_STATUS_OK) {
			if (bfad_os_sysfs_vf_add_dir(bfad, ioc_inst, vf_id)
				 != 0) {
				spin_unlock_irqrestore(&bfad->bfad_lock, flags);
				return -EINVAL;
			}
		}
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	} else if (strcmp(attr->name, "delete") == 0) {
		sscanf(buf, "%x", &vf_id);
		spin_lock_irqsave(&bfad->bfad_lock, flags);
		fcs_vf = bfa_fcs_vf_lookup(&bfad->bfa_fcs, vf_id);
		if (fcs_vf != NULL) {
			status = bfa_fcs_vf_delete(fcs_vf);
			spin_unlock_irqrestore(&bfad->bfad_lock, flags);
			if (status == BFA_STATUS_OK)
				bfad_os_sysfs_vf_delete_dir(bfad, vf_id);
		}
		else {
			spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		}
	}

	return count;
}

DEF_ATTR(kobj_vf_enable, "enable");
DEF_ATTR(kobj_vf_add, "add");
DEF_ATTR(kobj_vf_delete, "delete");
DEF_ATTR(kobj_vf_list, "list");
DEF_ATTR(kobj_vf_listall, "listall");

static struct attribute *bfad_sysfs_vf_attrs[] = {
	&kobj_vf_enable,
	&kobj_vf_add,
	&kobj_vf_delete,
	&kobj_vf_list,
	&kobj_vf_listall,
	NULL,
};

static struct sysfs_ops bfad_sysfs_vf_ops = {
	.show = bfad_sysfs_vf_show,
	.store = bfad_sysfs_vf_store,
};

static struct kobj_type ktype_vf = {
	.sysfs_ops = &bfad_sysfs_vf_ops,
	.default_attrs = bfad_sysfs_vf_attrs,
};
#else
int
bfad_os_sysfs_vf_add_dir(bfad_t *bfad, int inst_no, uint16_t vf_id)
{
	return 0;
}

int
bfad_os_sysfs_vf_delete_dir(bfad_t *bfad, uint16_t vf_id)
{
	return 0;
}
#endif /* VF_READY */

/*
 * IOC Commands
 */

DEF_ATTR(kobj_ioc_attr, "attr");
DEF_ATTR(kobj_ioc_info, "info");
DEF_ATTR(kobj_ioc_stats, "stats");
DEF_ATTR(kobj_ioc_intr, "intr");
DEF_ATTR(kobj_ioc_power, "power");
DEF_ATTR(kobj_ioc_mpiomode, "mpiomode");

static struct attribute *bfad_sysfs_ioc_attrs[] = {
	&kobj_ioc_attr,
	&kobj_ioc_info,
	&kobj_ioc_stats,
	&kobj_ioc_intr,
	&kobj_ioc_power,
	&kobj_ioc_mpiomode,
	NULL,
};

/**
 * @brief Function to show IOC attributes
 * @param[in] kobj  - kobject of the 'ioc' directory <ROOT/INST/bfa/ioc>
 * @param[in] attr  - Attributes of the directory
 * @param[in] buf   - Buffer to be filled in.
 * @retval - Number of bytes of 'buf' filled in.
 */
static          ssize_t
bfad_sysfs_ioc_show(struct kobject *kobj, struct attribute *attr, char *buf)
{
	int             ioc_inst = 1, len = 0;
	bfad_t         *bfad;
	int             dummy_vf_id;
	uint64_t        dummy_vport_id, dummy_rport_id;

	bfad = bfad_sysfs_get_param(kobj, &ioc_inst, &dummy_vf_id,
				&dummy_vport_id, &dummy_rport_id);
	if (bfad == NULL) {
		return len = snprintf(buf + len, PAGE_SIZE - len,
				 "Error finding given IOC(%d)\n", ioc_inst);
	}

	if (strcmp(attr->name, "attr") == 0) {
		return bfad_sysfs_attr_show(bfad, buf);
	} else if (strcmp(attr->name, "info") == 0) {
		return bfad_sysfs_info_show(bfad, buf);
	} else if (strcmp(attr->name, "stats") == 0) {
		return bfad_sysfs_stats_show(bfad, buf);
	} else if (strcmp(attr->name, "intr") == 0) {
		return bfad_sysfs_intr_show(bfad, buf);
	} else if (strcmp(attr->name, "power") == 0) {
		return bfad_sysfs_power_show(bfad, buf);
	} else if (strcmp(attr->name, "mpiomode") == 0) {
		return bfad_sysfs_mpiomode_show(bfad, buf);
	} else
		return 0;
}

/**
 * @brief Function to configure IOC attributes e.g. power/stats/interrupt
 * @param[in] kobj  - kobject of the 'ioc' directory <ROOT/INST/bfa/ioc>
 * @param[in] attr  - Attributes of the directory
 * @param[in] buf   - Buffer contains user input
 * @param[in] count - Number of bytes in 'buf'
 * @retval count - Number of bytes of 'buf' used in the function
 */
static          ssize_t
bfad_sysfs_ioc_store(struct kobject *kobj, struct attribute *attr,
		     const char *buf, size_t count)
{
	int             ioc_inst = 1;
	bfad_t         *bfad;
	int             dummy_vf_id;
	uint64_t        dummy_vport_id, dummy_rport_id;

	bfad = bfad_sysfs_get_param(kobj, &ioc_inst, &dummy_vf_id,
				&dummy_vport_id, &dummy_rport_id);
	if (bfad == NULL)
		return -EINVAL;

	if (strcmp(attr->name, "stats") == 0) {
		bfad_sysfs_stats_store(bfad, buf);
	} else if (strcmp(attr->name, "intr") == 0) {
		int             delay = 100, latency = 100;	/* HACK Alert
								 *!! */
		bfad_sysfs_intr_store(bfad, buf, delay, latency);
	} else if (strcmp(attr->name, "power") == 0) {
		bfad_sysfs_power_store(bfad, buf);
	} else if (strcmp(attr->name, "mpiomode") == 0) {
		bfad_sysfs_mpiomode_store(bfad, buf);
	}

	return count;
}

static struct sysfs_ops bfad_sysfs_ioc_ops = {
	.show = bfad_sysfs_ioc_show,
	.store = bfad_sysfs_ioc_store,
};

static struct kobj_type ktype_ioc = {
	.sysfs_ops = &bfad_sysfs_ioc_ops,
	.default_attrs = bfad_sysfs_ioc_attrs,
};

/*
 * ITNIM Commands
 */

ssize_t
bfad_sysfs_itnim_show_attr(bfad_t *bfad, int vf_id, wwn_t pwwn, wwn_t rpwwn,
			   char *buf)
{
	int             len = 0;
	bfa_itnim_attr_t attr;
	bfa_fcs_port_t *fcs_port;
	unsigned long	flags;
	bfa_status_t	status;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs, vf_id, pwwn);
	if (fcs_port == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		return len += snprintf(buf + len, PAGE_SIZE - len,
				 "can't find FCS port\n");
	}

	status = bfa_fcs_itnim_attr_get(fcs_port, rpwwn, &attr); 
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (status == BFA_STATUS_OK) {
		len += snprintf(buf + len, PAGE_SIZE - len,
				"FCP IM state: %s\n",
				bfad_sysfs_state_to_string(attr.state,
							   fcpim_state_table));
	} else {
		len += snprintf(buf + len, PAGE_SIZE - len,
				"error getting FCP IM state\n");
	}

	return len;

}

ssize_t
bfad_sysfs_itnim_show_stats(bfad_t *bfad, int vf_id, wwn_t pwwn, wwn_t rpwwn,
			    char *buf)
{
	int             len = 0;
	bfa_itnim_stats_t stats;
	bfa_fcs_port_t *fcs_port;
	unsigned long	flags;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs, vf_id, pwwn);
	if (fcs_port == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		return len +=
			snprintf(buf + len, PAGE_SIZE - len,
				 "can't find FCS port\n");
	}

	if (bfa_fcs_itnim_stats_get(fcs_port, rpwwn, &stats) != BFA_STATUS_OK) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		len += snprintf(buf + len, PAGE_SIZE - len,
				"error getting itnim stats");
		return len;
	}

	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	len += snprintf(buf + len, PAGE_SIZE - len, "%-40s: %-10d\n",
			"Num rport onlines", stats.onlines);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-40s: %-10d\n",
			"Num rport offlines", stats.offlines);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-40s: %-10d\n",
			"Num PRLI sent", stats.prli_sent);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-40s: %-10d\n",
			"Num PRLI Accepts", stats.prli_rsp_acc);

	len += snprintf(buf + len, PAGE_SIZE - len, "%-40s: %-10d\n",
			"Num PRLI error responses", stats.prli_rsp_err);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-40s: %-10d\n",
			"Num PRLI responses parse err",
			stats.prli_rsp_parse_err);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-40s: %-10d\n",
			"Num fcxp alloc waits", stats.fcxp_alloc_wait);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-40s: %-10d\n",
			"Num Second level err recovery", stats.sler);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-40s: %-10d\n",
			"Num times in initiator mode", stats.initiator);

	return len;

}

DEF_ATTR(kobj_itnim_attr, "attr");
DEF_ATTR(kobj_itnim_stats, "stats");

static struct attribute *bfad_sysfs_itnim_attrs[] = {
	&kobj_itnim_attr,
	&kobj_itnim_stats,
	NULL,
};

/**
 * @brief Function to show RPORT-ITNIM attributes
 * @param[in] kobj  - kobject of the 'ioc' directory <ROOT/INST/bfa/ioc>
 * @param[in] attr  - Attributes of the directory
 * @param[in] buf   - Buffer to be filled in.
 * @retval - Number of bytes of 'buf' filled in.
 */
static          ssize_t
bfad_sysfs_itnim_show(struct kobject *kobj, struct attribute *attr, char *buf)
{
	int             ioc_inst = 1, len = 0;
	bfad_t         *bfad;
	int             vf_id;
	uint64_t        rpwwn, vpwwn;
	bfa_pport_attr_t pattr;
	unsigned long	flags;

	bfad = bfad_sysfs_get_param(kobj, &ioc_inst, &vf_id, &vpwwn, &rpwwn);
	if (bfad == NULL) {
		return len = snprintf(buf + len, PAGE_SIZE - len,
				 "Error finding given IOC(%d)\n", ioc_inst);
	}

	if (vpwwn == 0) {
		spin_lock_irqsave(&bfad->bfad_lock, flags);
		bfa_hal_pport_get_attr(&bfad->bfa_hal, &pattr);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		vpwwn = pattr.pwwn;
	}

	if (strcmp(attr->name, "attr") == 0) {
		return bfad_sysfs_itnim_show_attr(bfad, vf_id, vpwwn,
						  rpwwn, buf);
	} else if (strcmp(attr->name, "stats") == 0) {
		return bfad_sysfs_itnim_show_stats(bfad, vf_id, vpwwn,
						   rpwwn, buf);
	}

	return 0;

}

/**
 * @brief Function to configure RPORT-ITNIM attributes e.g. statsclear
 * @param[in] kobj  - kobject of the 'ioc' directory <ROOT/INST/bfa/ioc>
 * @param[in] attr  - Attributes of the directory
 * @param[in] buf   - Buffer contains user input
 * @param[in] count - Number of bytes in 'buf'
 * @retval count - Number of bytes of 'buf' used in the function
 */
ssize_t
bfad_sysfs_itnim_store(struct kobject *kobj, struct attribute *attr,
		       const char *buf, size_t count)
{
	int             ioc_inst = 1;
	bfad_t         *bfad;
	int             vf_id;
	uint64_t        rpwwn, vpwwn;
	bfa_fcs_port_t *fcs_port;
	bfa_fcs_itnim_t *itnim;
	bfa_pport_attr_t pattr;
	unsigned long	flags;

	bfad = bfad_sysfs_get_param(kobj, &ioc_inst, &vf_id, &vpwwn, &rpwwn);
	if (bfad == NULL)
		return count;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	if (vpwwn == 0) {
		bfa_hal_pport_get_attr(&bfad->bfa_hal, &pattr);
		vpwwn = pattr.pwwn;
	}

	if (strcmp(attr->name, "stats") == 0) {
		fcs_port =
			bfa_fcs_lookup_port(&bfad->bfa_fcs, vf_id, vpwwn);
		if (fcs_port == NULL) {
			spin_unlock_irqrestore(&bfad->bfad_lock, flags);
			return count;
		}

		itnim = bfa_fcs_itnim_lookup(fcs_port, rpwwn);
		if (itnim == NULL) {
			spin_unlock_irqrestore(&bfad->bfad_lock, flags);
			return count;
		}

		bfa_fcs_itnim_stats_clear(fcs_port, rpwwn);
		bfa_hal_itnim_clear_stats(bfa_fcs_itnim_get_halitn(itnim));
	}

	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	return count;
}

static struct sysfs_ops bfad_sysfs_itnim_ops = {
	.show = bfad_sysfs_itnim_show,
	.store = bfad_sysfs_itnim_store,
};

static struct kobj_type ktype_itnim = {
	.sysfs_ops = &bfad_sysfs_itnim_ops,
	.default_attrs = bfad_sysfs_itnim_attrs,
};

/*
 * RPORT Commands
 */

ssize_t
bfad_sysfs_rport_show_attr(bfad_t *bfad, int vf_id, wwn_t pwwn, wwn_t rpwwn,
			   char *buf)
{
	int             len = 0;
	bfa_rport_attr_t attr;
	bfa_fcs_port_t *fcs_port;
	bfa_fcs_rport_t *fcs_rport;
	unsigned long	flags;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs, vf_id, pwwn);
	if (fcs_port == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		return len += snprintf(buf + len, PAGE_SIZE - len,
				 "can't find FCS port\n");
	}

	fcs_rport = bfa_fcs_rport_lookup(fcs_port, rpwwn);
	if (fcs_rport == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		return len += snprintf(buf + len, PAGE_SIZE - len,
				 "can't find FCS port\n");
	}

	bfa_fcs_rport_get_attr(fcs_rport, &attr);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %06x\n",
			"FC Address", bfa_os_ntoh3b(attr.pid));
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: ", "PWWN");
	len += bfad_sysfs_snprint_pwwn(attr.pwwn, buf, len);
	len += snprintf(buf + len, PAGE_SIZE - len, "\n");
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: ", "NWWN");
	len += bfad_sysfs_snprint_pwwn(attr.nwwn, buf, len);
	len += snprintf(buf + len, PAGE_SIZE - len, "\n");
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %s\n", "Status",
		bfad_sysfs_state_to_string(attr.state, rport_state_table));
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: ",
			"Class of Service");

	if (attr.fc_cos == FC_CLASS_2) {
		len += snprintf(buf + len, PAGE_SIZE - len, "Class-2\n");
	} else if (attr.fc_cos == FC_CLASS_3) {
		len += snprintf(buf + len, PAGE_SIZE - len, "Class-3\n");
	} else if (attr.fc_cos == FC_CLASS_2_3) {
		len += snprintf(buf + len, PAGE_SIZE - len,
				"Class-2, Class-3\n");
	}

	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n",
			"Data Field Size", attr.df_sz);

	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: ",
			"Cnt.Incr.Sec_Cnt");

	if (attr.cisc)
		len += snprintf(buf + len, PAGE_SIZE - len, "%s\n",
				"Supported\n");
	else
		len += snprintf(buf + len, PAGE_SIZE - len, "%s\n",
				"Not Supported\n");

	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %s\n", "Speed",
			"TBD");
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %s\n",
			"QOS Priority", "TBD");
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %s\n",
			"Symbolic Name", attr.symname.symname);

	return len;

}

ssize_t
bfad_sysfs_rport_show_stats(bfad_t *bfad, int vf_id, wwn_t pwwn, wwn_t rpwwn,
			    char *buf)
{
	int             len = 0;
	bfa_rport_stats_t stats;
	bfa_fcs_port_t *fcs_port;
	bfa_fcs_rport_t *fcs_rport;
	unsigned long	flags;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs, vf_id, pwwn);
	if (fcs_port == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		return len += snprintf(buf + len, PAGE_SIZE - len,
				 "can't find FCS port\n");
	}

	fcs_rport = bfa_fcs_rport_lookup(fcs_port, rpwwn);
	if (fcs_rport == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		return len += snprintf(buf + len, PAGE_SIZE - len,
				 "can't find FCS port for\n");
	}

	bfa_fcs_rport_get_stats(fcs_rport, &stats);
	bfa_hal_rport_get_stats(bfa_fcs_rport_get_halrport(fcs_rport),
				&stats.hal_stats);

	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %-10d\n",
			"Offlines", stats.offlines);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %-10d\n", "Onlines",
			stats.onlines);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %-10d\n", "Rscns",
			stats.rscns);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %-10d\n",
			"plogi sent", stats.plogis);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %-10d\n",
			"plogi accepts", stats.plogi_accs);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %-10d\n",
			"plogi timeouts", stats.plogi_timeouts);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %-10d\n",
			"plogi failed", stats.plogi_failed);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %-10d\n",
			"plogi rejects", stats.plogi_rejects);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %-10d\n",
			"plogi rcvd", stats.plogi_rcvd);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %-10d\n",
			"prlis received", stats.prli_rcvd);

	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %-10d\n",
			"logo sent", stats.logos);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %-10d\n",
			"logo accepts", stats.logo_accs);

	return len;

}

DEF_ATTR(kobj_rport_attr, "attr");
DEF_ATTR(kobj_rport_stats, "stats");

static struct attribute *bfad_sysfs_rport_attrs[] = {
	&kobj_rport_attr,
	&kobj_rport_stats,
	NULL,
};

/**
 * @brief Function to show RPORT attributes
 * @param[in] kobj  - kobject of the 'ioc' directory <ROOT/INST/bfa/ioc>
 * @param[in] attr  - Attributes of the directory
 * @param[in] buf   - Buffer to be filled in.
 * @retval - Number of bytes of 'buf' filled in.
 */
static          ssize_t
bfad_sysfs_rport_show(struct kobject *kobj, struct attribute *attr, char *buf)
{
	int             ioc_inst = 1, len = 0;
	bfad_t         *bfad;
	int             vf_id;
	bfa_pport_attr_t pattr;
	uint64_t        rpwwn, vpwwn;
	unsigned long	flags;

	bfad = bfad_sysfs_get_param(kobj, &ioc_inst, &vf_id, &vpwwn, &rpwwn);
	if (bfad == NULL) {
		return len = snprintf(buf + len, PAGE_SIZE - len,
				 "Error finding given IOC(%d)\n", ioc_inst);
	}

	if (vpwwn == 0) {
		spin_lock_irqsave(&bfad->bfad_lock, flags);
		bfa_hal_pport_get_attr(&bfad->bfa_hal, &pattr);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		vpwwn = pattr.pwwn;
	}

	if (strcmp(attr->name, "attr") == 0) {
		return bfad_sysfs_rport_show_attr(bfad, vf_id, vpwwn,
						  rpwwn, buf);
	} else if (strcmp(attr->name, "stats") == 0) {
		return bfad_sysfs_rport_show_stats(bfad, vf_id, vpwwn,
						   rpwwn, buf);
	}

	return 0;

}

/**
 * @brief Function to configure RPORT attributes e.g. statsclear
 * @param[in] kobj  - kobject of the 'ioc' directory <ROOT/INST/bfa/ioc>
 * @param[in] attr  - Attributes of the directory
 * @param[in] buf   - Buffer contains user input
 * @param[in] count - Number of bytes in 'buf'
 * @retval count - Number of bytes of 'buf' used in the function
 */
ssize_t
bfad_sysfs_rport_store(struct kobject *kobj, struct attribute *attr,
		       const char *buf, size_t count)
{
	int             ioc_inst = 1;
	bfad_t         *bfad;
	int             vf_id;
	uint64_t        rpwwn, vpwwn;
	bfa_fcs_port_t *fcs_port;
	bfa_pport_attr_t pattr;
	bfa_fcs_rport_t *fcs_rport;
	unsigned long	flags;

	bfad = bfad_sysfs_get_param(kobj, &ioc_inst, &vf_id, &vpwwn, &rpwwn);
	if (bfad == NULL)
		return count;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	
	if (vpwwn == 0) {
		bfa_hal_pport_get_attr(&bfad->bfa_hal, &pattr);
		vpwwn = pattr.pwwn;
	}

	if (strcmp(attr->name, "stats") == 0) {
		fcs_port =
			bfa_fcs_lookup_port(&bfad->bfa_fcs, vf_id, vpwwn);
		if (fcs_port == NULL) {
			spin_unlock_irqrestore(&bfad->bfad_lock, flags);
			return count;
		}

		fcs_rport = bfa_fcs_rport_lookup(fcs_port, rpwwn);
		if (fcs_rport == NULL) {
			spin_unlock_irqrestore(&bfad->bfad_lock, flags);
			return count;
		}

		bfa_fcs_rport_clear_stats(fcs_rport);
		bfa_hal_rport_clear_stats(bfa_fcs_rport_get_halrport(fcs_rport));
	}

	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	return count;
}

static struct sysfs_ops bfad_sysfs_rport_ops = {
	.show = bfad_sysfs_rport_show,
	.store = bfad_sysfs_rport_store,
};

static struct kobj_type ktype_rport = {
	.sysfs_ops = &bfad_sysfs_rport_ops,
	.default_attrs = bfad_sysfs_rport_attrs,
};

ssize_t
bfad_sysfs_rport_main_show_list(bfad_t *bfad, struct kobject *kobj, int vf_id,
				wwn_t pwwn, wwn_t vpwwn, char *buf)
{
	int             len = 0, i, index;
	wwn_t          *rports = NULL;
	char           *ioc_path, *port_path;
	uint32_t        nrports = SYSFS_MAX_RPORTS;
	bfad_kobj_t    *bfad_kobj, *bfad_vf_kobj, *bfad_lport_kobj;
	unsigned long   flags;
	bfa_fcs_port_t *fcs_port;
	bfa_fcs_rport_t *fcs_rport;
	struct kobject *parent_kobj, *parent_itnim;
	struct list_head *kobj_vport_list;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs, vf_id, vpwwn?vpwwn:pwwn);
	if (fcs_port == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		return len += snprintf(buf + len, PAGE_SIZE - len,
				 "can't find FCS port\n");
	}

	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	rports = kzalloc(sizeof(wwn_t) * nrports, GFP_KERNEL);
	if (rports == NULL) {
		return len +=
			snprintf(buf + len, PAGE_SIZE - len, "memory sqeeze\n");
	}

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	bfa_fcs_port_get_rports(fcs_port, rports, &nrports);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	ioc_path =
		bfad_sysfs_private_kobject_get_path(&bfad->pcidev->dev.kobj,
						    GFP_ATOMIC);
	if (vf_id != 0) {
		bfad_vf_kobj = bfad_sysfs_kobj_get(ioc_path, KOBJ_TYPE_VF,
				vf_id, 0, &bfad_kobj_vf_list);
		if (bfad_vf_kobj == NULL) {
			kfree(ioc_path);
			return len += snprintf(buf + len, PAGE_SIZE - len,
				"can't locate kobj vf_id: %x\n", vf_id);
		}

		kobj_vport_list = &bfad_vf_kobj->kobj_vport_list;
	} else {
		kobj_vport_list = &bfad_kobj_base_list;
	}

	port_path = bfad_sysfs_private_kobject_get_path(kobj, GFP_ATOMIC);
	if (strstr(port_path, "lport") && vf_id == 0) {
		index = find_kobj_by_path(&bfad->pcidev->dev.kobj);
		parent_kobj = &ko_lport[index];
	} else {
		/*
		 * find the parent for the lport
		 */
		bfad_lport_kobj = bfad_sysfs_kobj_get(ioc_path,
			 KOBJ_TYPE_VPORT, vf_id, vpwwn, kobj_vport_list);
		if (bfad_lport_kobj == NULL) {
			kfree(ioc_path);
			kfree(port_path);
			return len += snprintf(buf + len, PAGE_SIZE - len,
			 "Parent can not be located to create rport\n");
		}
		parent_kobj = &bfad_lport_kobj->kobj;
	}
	kfree(port_path);
	kfree(ioc_path);

	len += snprintf(buf + len, PAGE_SIZE - len, "No. of rports: %d\n",
			nrports);

	/*
	 * Create a directory entry for each rport
	 */
	for (i = 0; i < nrports; i++) {
		fcs_rport = bfa_fcs_rport_lookup(fcs_port, rports[i]);
		if (fcs_rport == NULL) {
			kfree(rports);
			return len += snprintf(buf + len, PAGE_SIZE - len,
					 "can't find FCS port \n");
		}

		if (bfad_sysfs_search_kobj(pwwn, KOBJ_TYPE_RPORT, vf_id, vpwwn,
				 rports[i], kobj_vport_list))
			continue;

		bfad_kobj = kzalloc(sizeof(bfad_kobj_t), GFP_KERNEL);
		if (!bfad_kobj) {
			return len += snprintf(buf + len, PAGE_SIZE - len,
					 "memory sqeeze\n");
		}

		sprintf(bfad_kobj->kobj.name, "rp-");
		bfad_sysfs_print_pwwn(rports[i], &bfad_kobj->kobj.name[3]);

		bfad_kobj->type = KOBJ_TYPE_RPORT;
		bfad_kobj->inst = bfad->inst_no;
		bfad_kobj->vf_id = vf_id;
		bfad_kobj->rport_id = rports[i];
		bfad_kobj->vport_id = vpwwn;
		bfad_kobj->pwwn = pwwn;
		bfad_kobj->kobj.ktype = &ktype_rport;
		bfad_kobj->kobj.parent = parent_kobj;
		if (kobject_register(&bfad_kobj->kobj) == -EINVAL) {
			return len += snprintf(buf + len, PAGE_SIZE - len,
					 "failed to register kobj\n");
			return 0;
		}

		list_add_tail(&bfad_kobj->kobj_list_entry, kobj_vport_list);

		parent_itnim = &bfad_kobj->kobj;

		/*
		 * Create itinm node here
		 */
		bfad_kobj = kzalloc(sizeof(bfad_kobj_t), GFP_KERNEL);
		if (!bfad_kobj) {
			return len += snprintf(buf + len, PAGE_SIZE - len,
					 "memory sqeeze\n");
		}

		sprintf(bfad_kobj->kobj.name, "itnim");

		bfad_kobj->type = KOBJ_TYPE_ITNIM;
		bfad_kobj->inst = bfad->inst_no;
		bfad_kobj->vf_id = vf_id;
		bfad_kobj->rport_id = rports[i];
		bfad_kobj->vport_id = vpwwn;
		bfad_kobj->pwwn = pwwn;
		bfad_kobj->kobj.ktype = &ktype_itnim;
		bfad_kobj->kobj.parent = parent_itnim;
		if (kobject_register(&bfad_kobj->kobj) == -EINVAL) {
			return len += snprintf(buf + len, PAGE_SIZE - len,
					 "failed to register kobj\n");
			return 0;
		}

		list_add_tail(&bfad_kobj->kobj_list_entry, kobj_vport_list);
	}

	/*
	 * Remove deleted rports
	 */
	bfad_sysfs_remove_deleted_rports(fcs_port, pwwn, vf_id,
					 kobj_vport_list);

	kfree(rports);
	return len;

}

/**
 * @brief Function to display physical port status
 * @param[in] bfad  - IOC
 * @param[in] buf   - Buffer to be filled in with attributes
 * @retval - Number of bytes of 'buf' filled in.
 */
ssize_t
bfad_sysfs_pport_enable_show(bfad_t *bfad, char *buf)
{
	int             len = 0;

	return len;
}

/**
 * @brief Function to enable/disable physical port
 * @param[in] bfad  - IOC
 * @param[in] buf   - Buffer contains user input
 * @retval - 0 on SUCCESS
 * @retval - negative value on FAILURE
 */
ssize_t
bfad_sysfs_pport_enable_store(bfad_t *bfad, const char *buf)
{
	unsigned long	flags;
	if (strncmp("enable", buf, strlen(buf) - 1) == 0) {
		spin_lock_irqsave(&bfad->bfad_lock, flags);
		bfa_hal_pport_enable(&bfad->bfa_hal);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	} else if (strncmp("disable", buf, strlen(buf) - 1) == 0) {
		spin_lock_irqsave(&bfad->bfad_lock, flags);
		bfa_hal_pport_disable(&bfad->bfa_hal);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	} else
		return -EIO;

	return 0;
}

/**
 * @brief Function to display physical port attribute values
 * @param[in] bfad  - IOC
 * @param[in] buf   - Buffer to be filled in with attributes
 * @retval - Number of bytes of 'buf' filled in.
 */
ssize_t
bfad_sysfs_pport_attr_show(bfad_t *bfad, char *buf)
{
	int             len = 0;
	bfa_pport_attr_t pattr;
	bfa_port_attr_t attr;
	unsigned long	flags;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	bfa_hal_pport_get_attr(&bfad->bfa_hal, &pattr);
	bfa_fcs_port_get_attr(&bfad->bfa_fcs.fabric.bport, &attr);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	pattr.port_type = attr.port_type;

	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: ", "PWWN");
	len += bfad_sysfs_snprint_pwwn(pattr.pwwn, buf, len);
	len += snprintf(buf + len, PAGE_SIZE - len, "\n");
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: ", "NWWN");
	len += bfad_sysfs_snprint_pwwn(pattr.nwwn, buf, len);
	len += snprintf(buf + len, PAGE_SIZE - len, "\n");
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %06x\n", "FC Addr",
			bfa_os_ntoh3b(attr.pid));
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %s\n", "Media",
			"sw");
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: ", "State");
	if (pattr.port_state >= BFA_HAL_PPORT_UNINIT
	    && pattr.port_state <= BFA_HAL_PPORT_MAX_STATE)
		len += snprintf(buf + len, PAGE_SIZE - len, "%s",
				bfad_sysfs_state_to_string(pattr.port_state,
							   port_state_table));
	else
		len += snprintf(buf + len, PAGE_SIZE - len, "%s", "invalid");
	len += snprintf(buf + len, PAGE_SIZE - len, "\n");

	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: ", "Port Type");
	switch (pattr.port_type) {
	case BFA_PPORT_TYPE_UNKNOWN:
		len += snprintf(buf + len, PAGE_SIZE - len, "%s", "unknown");
		break;
	case BFA_PPORT_TYPE_TRUNKED:
		len += snprintf(buf + len, PAGE_SIZE - len, "%s", "trunked");
		break;
	case BFA_PPORT_TYPE_NPORT:
		len += snprintf(buf + len, PAGE_SIZE - len, "%s", "nport");
		break;
	case BFA_PPORT_TYPE_NLPORT:
		len += snprintf(buf + len, PAGE_SIZE - len, "%s", "nlport");
	case BFA_PPORT_TYPE_LPORT:
		len += snprintf(buf + len, PAGE_SIZE - len, "%s", "lport");
		break;
	case BFA_PPORT_TYPE_P2P:
		len += snprintf(buf + len, PAGE_SIZE - len, "%s", "p2p");
		break;
	case BFA_PPORT_TYPE_VPORT:
		len += snprintf(buf + len, PAGE_SIZE - len, "%s", "vport");
		break;
	default:
		len += snprintf(buf + len, PAGE_SIZE - len, "%s(%d)", "invalid",
				pattr.port_type);
		break;
	}
	len += snprintf(buf + len, PAGE_SIZE - len, "\n");

	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: ",
			"Supported Classes");
	switch (pattr.port_type) {
	case FC_CLASS_2:
		len += snprintf(buf + len, PAGE_SIZE - len, "%s", "unknown");
		break;
	case FC_CLASS_3:
		len += snprintf(buf + len, PAGE_SIZE - len, "%s", "trunked");
		break;
	default:
		len += snprintf(buf + len, PAGE_SIZE - len, "%s", "--");
		break;
	}
	len += snprintf(buf + len, PAGE_SIZE - len, "\n");

	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %s\n",
			"Symbolic Name", pattr.port_symname.symname);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %d\n",
			"Recv. bb credits", pattr.pport_cfg.rx_bbcredit);

	return len;
}

/**
 * @brief Function to change physical port attribute values
 * @param[in] bfad  - IOC
 * @param[in] buf   - Buffer contains user input
 * @retval - 0 on SUCCESS
 * @retval - negative value on FAILURE
 */
ssize_t
bfad_sysfs_pport_attr_store(bfad_t *bfad, const char *buf)
{
	int             len = 0;

	return len;
}

void
bfad_sysfs_hcb_comp(void *arg, bfa_status_t status)
{
	struct bfad_hal_comp *fcomp = (struct bfad_hal_comp *)arg;

	fcomp->status = status;
	complete(&fcomp->comp);
}

/**
 * @brief Function to display physical port statistics
 * @param[in] bfad  - IOC
 * @param[in] buf   - Buffer to be filled in with attributes
 * @retval - Number of bytes of 'buf' filled in.
 */
ssize_t
bfad_sysfs_pport_stats_show(bfad_t *bfad, char *buf)
{
	int             len = 0;
	bfa_pport_stats_t stats;
	bfa_status_t    status;
	struct bfad_hal_comp fcomp;
	unsigned long   flags;

	memset(&stats, 0, sizeof(bfa_pport_stats_t));

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	status = bfa_hal_pport_get_stats(&bfad->bfa_hal, &stats,
					 bfad_sysfs_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (status != BFA_STATUS_OK) {
		return snprintf(buf + len, PAGE_SIZE - len,
				"failed to get stats\n");
	}

	wait_for_completion(&fcomp.comp);

	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %lld\n",
			"Secs Since Reset", (long long)stats.secs_since_reset);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %lld\n",
			"Tx. Frames", (long long)stats.tx_frames);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %lld\n",
			"Tx. Words", (long long)stats.tx_words);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %lld\n",
			"Rx. Frames", (long long)stats.rx_frames);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %lld\n",
			"Rx. Words", (long long)stats.rx_words);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %lld\n",
			"LIPs seen", (long long)stats.lip_count);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %lld\n",
			"NOS count", (long long)stats.nos_count);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %lld\n",
			"Errored Frames", (long long)stats.error_frames);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %lld\n",
			"Dropped Frames", (long long)stats.dropped_frames);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %lld\n",
			"No Link Failures", (long long)stats.link_failures);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %lld\n",
			"No Loss of Sync", (long long)stats.loss_of_syncs);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %lld\n",
			"No Loss of Sig", (long long)stats.loss_of_signals);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %lld\n",
			"Primitive Seq.Err", (long long)stats.primseq_errs);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %lld\n",
			"Invalid ordered set", (long long)stats.bad_os_count);
	len += snprintf(buf + len, PAGE_SIZE - len, "%-20s: %lld\n",
			"Invalid CRCs", (long long)stats.invalid_crcs);
	return len;
}

/**
 * @brief Function to clear physical port statistics
 * @param[in] bfad  - IOC
 * @param[in] buf   - Buffer contains user input
 * @retval - 0 on SUCCESS
 */
ssize_t
bfad_sysfs_pport_stats_store(bfad_t *bfad, const char *buf)
{
	struct bfad_hal_comp fcomp;
	unsigned long   flags;
	bfa_status_t    status;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	status = bfa_hal_pport_clear_stats(&bfad->bfa_hal, bfad_sysfs_hcb_comp,
					   &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (status != BFA_STATUS_OK)
		return -EIO;

	wait_for_completion(&fcomp.comp);

	return 0;
}

/**
 * @brief Function to display physical port topology
 * @param[in] bfad  - IOC
 * @param[in] buf   - Buffer to be filled in with attributes
 * @retval - Number of bytes of 'buf' filled in.
 */
ssize_t
bfad_sysfs_pport_topo_show(bfad_t *bfad, char *buf)
{
	int             len = 0;
	bfa_pport_topology_t topology;
	bfa_pport_attr_t attr;
	unsigned long	flags;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	bfa_hal_pport_get_attr(&bfad->bfa_hal, &attr);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	topology = attr.topology;
	if (topology == BFA_PPORT_TOPOLOGY_P2P) {
		len += snprintf(buf + len, PAGE_SIZE - len,
			"Current Topology: %s\n", "p2p");
	} else if (topology == BFA_PPORT_TOPOLOGY_LOOP) {
		len += snprintf(buf + len, PAGE_SIZE - len,
			"Current Topology: %s\n", "loop");
	} else if (topology == BFA_PPORT_TOPOLOGY_AUTO) {
		len += snprintf(buf + len, PAGE_SIZE - len,
			"Current Topology: %s\n", "auto");
	} else {
		len += snprintf(buf + len, PAGE_SIZE - len, "%s (%d)\n",
			"Current Topology:invalid !!", topology);
	}

	topology = attr.pport_cfg.topology;
	if (topology == BFA_PPORT_TOPOLOGY_P2P) {
		len += snprintf(buf + len, PAGE_SIZE - len,
			"Configured Topology: %s\n", "p2p");
	} else if (topology == BFA_PPORT_TOPOLOGY_LOOP) {
		len += snprintf(buf + len, PAGE_SIZE - len,
			"Configured Topology: %s\n", "loop");
	} else if (topology == BFA_PPORT_TOPOLOGY_AUTO) {
		len += snprintf(buf + len, PAGE_SIZE - len,
			"Configured Topology: %s\n", "auto");
	} else {
		len += snprintf(buf + len, PAGE_SIZE - len, "%s (%d)\n",
			"Configured Topology:invalid !!", topology);
	}

	return len;
}

/**
 * @brief Function to change physical port topology
 * @param[in] bfad  - IOC
 * @param[in] buf   - Buffer contains user input
 * @retval - 0 on SUCCESS
 * @retval - negative value on FAILURE
 */
ssize_t
bfad_sysfs_pport_topo_store(bfad_t *bfad, const char *buf)
{
	bfa_status_t    status;
	bfa_pport_topology_t topo;
	unsigned long	flags;

	if (strncmp("p2p", buf, strlen(buf) - 1) == 0)
		topo = BFA_PPORT_TOPOLOGY_P2P;
	else if (strncmp("loop", buf, strlen(buf) - 1) == 0)
		topo = BFA_PPORT_TOPOLOGY_LOOP;
	else if (strncmp("auto", buf, strlen(buf) - 1) == 0)
		topo = BFA_PPORT_TOPOLOGY_AUTO;
	else
		topo = BFA_PPORT_TOPOLOGY_NONE;

	if (topo == BFA_PPORT_TOPOLOGY_NONE)
		return -EIO;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	status = bfa_hal_pport_cfg_topology(&bfad->bfa_hal, topo);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (status != BFA_STATUS_OK)
		return -EIO;

	return 0;
}

char           *
bfad_sysfs_get_speed_str(int speed)
{
	static bfa_state_t port_spd[] = {
		{BFA_PPORT_SPEED_UNKNOWN, "---"},
		{BFA_PPORT_SPEED_1GBPS, "1G "},
		{BFA_PPORT_SPEED_2GBPS, "2G "},
		{BFA_PPORT_SPEED_4GBPS, "4G "},
		{BFA_PPORT_SPEED_8GBPS, "8G "},
		{BFA_PPORT_SPEED_AUTO, "Auto"},
	};
	int             j = 0;

	while (port_spd[j].code <= BFA_PPORT_SPEED_AUTO) {
		if (port_spd[j].code == speed)
			return port_spd[j].str;
		j++;
	}

	return "Invalid";
}

/**
 * @brief Function to display physical port speed
 * @param[in] bfad  - IOC
 * @param[in] buf   - Buffer to be filled in with attributes
 * @retval - Number of bytes of 'buf' filled in.
 */
ssize_t
bfad_sysfs_pport_speed_show(bfad_t *bfad, char *buf)
{
	int             len = 0;
	bfa_pport_attr_t attr;
	unsigned long	flags;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	bfa_hal_pport_get_attr(&bfad->bfa_hal, &attr);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	len += snprintf(buf + len, PAGE_SIZE - len, "Current Speed   : %s\n",
			bfad_sysfs_get_speed_str(attr.speed));

	len += snprintf(buf + len, PAGE_SIZE - len, "Configured Speed: %s\n",
			bfad_sysfs_get_speed_str(attr.pport_cfg.speed));

	return len;
}

/**
 * @brief Function to change physical port speed
 * @param[in] bfad  - IOC
 * @param[in] buf   - Buffer contains user input
 * @retval - 0 on SUCCESS
 * @retval - negative value on FAILURE
 */
ssize_t
bfad_sysfs_pport_speed_store(bfad_t *bfad, const char *buf)
{
	bfa_status_t    status;
	bfa_pport_speed_t speed;
	unsigned long	flags;

	if (strncmp("1gbps", buf, strlen(buf) - 1) == 0) {
		speed = BFA_PPORT_SPEED_1GBPS;
	} else if (strncmp("2gbps", buf, strlen(buf) - 1) == 0) {
		speed = BFA_PPORT_SPEED_2GBPS;
	} else if (strncmp("4gbps", buf, strlen(buf) - 1) == 0) {
		speed = BFA_PPORT_SPEED_4GBPS;
	} else if (strncmp("8gbps", buf, strlen(buf) - 1) == 0) {
		speed = BFA_PPORT_SPEED_8GBPS;
	} else if (strncmp("auto", buf, strlen(buf) - 1) == 0) {
		speed = BFA_PPORT_SPEED_AUTO;
	} else
		return -EIO;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	status = bfa_hal_pport_cfg_speed(&bfad->bfa_hal, speed);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (status != BFA_STATUS_OK)
		return -EIO;

	return 0;
}

/**
 * @brief Function to display physical port hard alpa value
 * @param[in] bfad  - IOC
 * @param[in] buf   - Buffer to be filled in with attributes
 * @retval - Number of bytes of 'buf' filled in.
 */
ssize_t
bfad_sysfs_pport_alpa_show(bfad_t *bfad, char *buf)
{
	int             len = 0;
	uint8_t         hardalpa;

	if (bfa_hal_pport_get_hardalpa(&bfad->bfa_hal, &hardalpa) == BFA_TRUE) {
		len += snprintf(buf + len, PAGE_SIZE - len, "Hardalpa: %02x\n",
				hardalpa);
	} else {
		len += snprintf(buf + len, PAGE_SIZE - len,
				"Hardalpa not configured\n");
	}

	return len;
}

/**
 * @brief Function to change physical port hard alpa value
 * @param[in] bfad  - IOC
 * @param[in] buf   - Buffer contains user input
 * @retval - 0 on SUCCESS
 * @retval - negative value on FAILURE
 */
ssize_t
bfad_sysfs_pport_alpa_store(bfad_t *bfad, const char *buf)
{
	char            str[100];
	uint8_t         alpa;
	unsigned long	flags;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	if (strncmp("hardalpa", buf, strlen(buf) - 1) == 0) {
		sscanf(buf, "%s %c", str, &alpa);
		if (bfa_hal_pport_cfg_hardalpa(&bfad->bfa_hal, alpa)
		    != BFA_STATUS_OK) {
			spin_unlock_irqrestore(&bfad->bfad_lock, flags);
			return -EIO;
		}
	} else if (strncmp("clearalpa", buf, strlen(buf) - 1) == 0) {
		if (bfa_hal_pport_clr_hardalpa(&bfad->bfa_hal)
		    != BFA_STATUS_OK) {
			spin_unlock_irqrestore(&bfad->bfad_lock, flags);
			return -EIO;
		}
	}
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	return 0;
}

/**
 * @brief Function to show physical port maxfrm
 * @param[in] bfad  - IOC
 * @param[in] buf   - Buffer to be filled in the function
 * @retval - number of bytes of 'buf' filled
 */
ssize_t
bfad_sysfs_pport_maxfrm_show(bfad_t *bfad, char *buf)
{
	int             len = 0;

	return snprintf(buf + len, PAGE_SIZE - len, "%d\n",
			bfa_hal_pport_get_maxfrsize(&bfad->bfa_hal));
}

/**
 * @brief Function to change physical port maxfrm value
 * @param[in] bfad  - IOC
 * @param[in] buf   - Buffer contains user input
 * @retval - 0 on SUCCESS
 * @retval - negative value on FAILURE
 */
ssize_t
bfad_sysfs_pport_maxfrm_store(bfad_t *bfad, const char *buf)
{
	int             maxfrsize;
	unsigned long	flags;

	sscanf(buf, "%d", &maxfrsize);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	if (bfa_hal_pport_cfg_maxfrsize(&bfad->bfa_hal, (uint16_t) maxfrsize)
	    != BFA_STATUS_OK) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		return -EIO;
	}

	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	return 0;
}

/**
 * @brief Function to display physical port trunking
 * @param[in] bfad  - IOC
 * @param[in] buf   - Buffer to be filled in with attributes
 * @retval - Number of bytes of 'buf' filled in.
 */
ssize_t
bfad_sysfs_pport_trunk_show(bfad_t *bfad, char *buf)
{
	int             len = 0;
	uint32_t        bitmap;

	if (bfa_hal_pport_trunk_query(&bfad->bfa_hal, &bitmap) == BFA_TRUE) {
		len += snprintf(buf + len, PAGE_SIZE - len, "%02x\n", bitmap);
	} else {
		len += snprintf(buf + len, PAGE_SIZE - len, "Not trunked\n");
	}

	return len;
}

/**
 * @brief Function to change physical port trunking
 * @param[in] bfad  - IOC
 * @param[in] buf   - Buffer contains user input
 * @retval - 0 on SUCCESS
 * @retval - negative value on FAILURE
 */
ssize_t
bfad_sysfs_pport_trunk_store(bfad_t *bfad, const char *buf)
{
	char            str[100];
	uint32_t        pp_bmap;
	unsigned long	flags;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	if (strncmp("on", buf, strlen(buf) - 1) == 0) {
		sscanf(buf, "%s %x", str, &pp_bmap);
		if (bfa_hal_pport_trunk_enable(&bfad->bfa_hal,
			(uint8_t) pp_bmap) != BFA_STATUS_OK) {
			spin_unlock_irqrestore(&bfad->bfad_lock, flags);
			return -EIO;
		}
	} else if (strncmp("off", buf, strlen(buf) - 1) == 0) {
		if (bfa_hal_pport_trunk_disable(&bfad->bfa_hal)
		    != BFA_STATUS_OK) {
			spin_unlock_irqrestore(&bfad->bfad_lock, flags);
			return -EIO;
		}
	}
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	return 0;
}

DEF_ATTR(kobj_pport_enable, "enable");
DEF_ATTR(kobj_pport_attr, "attr");
DEF_ATTR(kobj_pport_stats, "stats");
DEF_ATTR(kobj_pport_topo, "topo");
DEF_ATTR(kobj_pport_speed, "speed");
DEF_ATTR(kobj_pport_alpa, "alpa");
DEF_ATTR(kobj_pport_maxfrm, "maxfrm");
DEF_ATTR(kobj_pport_trunk, "trunk");

static struct attribute *bfad_sysfs_pport_attrs[] = {
	&kobj_pport_enable,
	&kobj_pport_attr,
	&kobj_pport_stats,
	&kobj_pport_topo,
	&kobj_pport_speed,
	&kobj_pport_alpa,
	&kobj_pport_maxfrm,
	&kobj_pport_trunk,
	NULL
};

/**
 * @brief Function to show physical ports parameters
 * @param[in] kobj - kobject of the (vport) directory
 * @param[in] attr - Attributes of the directory
 * @param[out] buf - Buffer to be returned with (attribute/statistic) values
 * @retval - size of the 'buf' filled
 */
static          ssize_t
bfad_sysfs_pport_show(struct kobject *kobj, struct attribute *attr, char *buf)
{
	int             ioc_inst = 1;
	bfad_t         *bfad;
	int             dummy_vf_id;
	uint64_t        dummy_vport_id, dummy_rport_id;

	bfad = bfad_sysfs_get_param(kobj, &ioc_inst, &dummy_vf_id,
					&dummy_vport_id, &dummy_rport_id);
	if (bfad == NULL)
		return -EINVAL;

	if (strcmp(attr->name, "enable") == 0) {
		return bfad_sysfs_pport_enable_show(bfad, buf);
	} else if (strcmp(attr->name, "attr") == 0) {
		return bfad_sysfs_pport_attr_show(bfad, buf);
	} else if (strcmp(attr->name, "stats") == 0) {
		return bfad_sysfs_pport_stats_show(bfad, buf);
	} else if (strcmp(attr->name, "topo") == 0) {
		return bfad_sysfs_pport_topo_show(bfad, buf);
	} else if (strcmp(attr->name, "speed") == 0) {
		return bfad_sysfs_pport_speed_show(bfad, buf);
	} else if (strcmp(attr->name, "alpa") == 0) {
		return bfad_sysfs_pport_alpa_show(bfad, buf);
	} else if (strcmp(attr->name, "maxfrm") == 0) {
		return bfad_sysfs_pport_maxfrm_show(bfad, buf);
	} else if (strcmp(attr->name, "trunk") == 0) {
		return bfad_sysfs_pport_trunk_show(bfad, buf);
	}

	return 0;
}

/**
 * @brief Function to change physical port parameters e.g. speed/topology etc.
 * @param[in] kobj  - kobject of the (port) directory <ROOT/INST/bfa/port>
 * @param[in] attr  - Attributes of the directory
 * @param[in] buf   - Buffer contains user input
 * @param[in] count - Number of bytes in 'buf'
 * @retval count - Number of bytes of 'buf' used in the function
 */
static          ssize_t
bfad_sysfs_pport_store(struct kobject *kobj, struct attribute *attr,
		       const char *buf, size_t count)
{
	int             ioc_inst = 1;
	bfad_t         *bfad;
	int             dummy_vf_id;
	uint64_t        dummy_vport_id, dummy_rport_id;

	bfad = bfad_sysfs_get_param(kobj, &ioc_inst, &dummy_vf_id,
				&dummy_vport_id, &dummy_rport_id);
	if (bfad == NULL)
		return -EINVAL;

	if (strcmp(attr->name, "enable") == 0) {
		bfad_sysfs_pport_enable_store(bfad, buf);
	} else if (strcmp(attr->name, "stats") == 0) {
		bfad_sysfs_pport_stats_store(bfad, buf);
	} else if (strcmp(attr->name, "topo") == 0) {
		bfad_sysfs_pport_topo_store(bfad, buf);
	} else if (strcmp(attr->name, "speed") == 0) {
		bfad_sysfs_pport_speed_store(bfad, buf);
	} else if (strcmp(attr->name, "alpa") == 0) {
		bfad_sysfs_pport_alpa_store(bfad, buf);
	} else if (strcmp(attr->name, "maxfrm") == 0) {
		bfad_sysfs_pport_maxfrm_store(bfad, buf);
	} else if (strcmp(attr->name, "trunk") == 0) {
		bfad_sysfs_pport_trunk_store(bfad, buf);
	}

	return count;
}

static struct sysfs_ops bfad_sysfs_pport_ops = {
	.show = bfad_sysfs_pport_show,
	.store = bfad_sysfs_pport_store,
};

static struct kobj_type ktype_pport = {
	.sysfs_ops = &bfad_sysfs_pport_ops,
	.default_attrs = bfad_sysfs_pport_attrs,
};

/**
 * @brief
 * Create the device sysfs attribute for PCI function level control .
 */
int             bfad_sysfs_first_time;
void
bfad_device_sysfs_create(bfad_t *bfad)
{
	struct kobject *kobj;
	int             i;

	if (bfad_sysfs_first_time) {
		bfad_sysfs_first_time = 0;
		INIT_LIST_HEAD(&bfad_kobj_vf_list);
		INIT_LIST_HEAD(&bfad_kobj_base_list);
	}

	/*
	 * create other configurable device parameters
	 */
	if (device_create_file(&bfad->pcidev->dev, &dev_attr_vf_ctl))
		return;
	if (device_create_file(&bfad->pcidev->dev, &dev_attr_vport_ctl))
		return;
	if (device_create_file(&bfad->pcidev->dev, &dev_attr_fc4type))
		return;
	if (device_create_file(&bfad->pcidev->dev, &dev_attr_bfad_ctl))
		return;

	if (bfad->inst_no >= MAX_IOC_INST)
		return;

	/*
	 * Look for next ko_bfa_root to use
	 */
	for (i = 0; i < MAX_IOC_INST; i++) {
		kobj = &ko_bfa_root[i];
		if (strncmp(kobj->name, "bfa", 3) != 0)
			break;
	}

	/*
	 * Create the BFA sub-directory under the IOC and register it
	 */
	kobj = &ko_bfa_root[i];
	memset(kobj, 0, sizeof(struct kobject));
	kobj->parent = &bfad->pcidev->dev.kobj;
	strcpy(kobj->name, "bfa");
	kobj->ktype = &ktype_bfa;
	if (kobject_register(kobj) == -EINVAL)
		return;

	/*
	 * Create sub-directory for ioc port
	 */
	kobj = &ko_ioc[i];
	memset(kobj, 0, sizeof(struct kobject));
	kobj->parent = &ko_bfa_root[i];
	strcpy(kobj->name, "ioc");
	kobj->ktype = &ktype_ioc;
	if (kobject_register(kobj) == -EINVAL)
		return;

	/*
	 * Create sub-directory for physical port
	 */
	kobj = &ko_pport[i];
	memset(kobj, 0, sizeof(struct kobject));
	kobj->parent = &ko_bfa_root[i];
	strcpy(kobj->name, "port");
	kobj->ktype = &ktype_pport;
	if (kobject_register(kobj) == -EINVAL)
		return;

	/*
	 * Create sub-directory for lport
	 */
	kobj = &ko_lport[i];
	memset(kobj, 0, sizeof(struct kobject));
	kobj->parent = &ko_bfa_root[i];
	strcpy(kobj->name, "lport-0");
	kobj->ktype = &ktype_lport;
	if (kobject_register(kobj) == -EINVAL)
		return;

#ifdef VF_READY
	/*
	 * Create sub-directory for VF
	 */
	kobj = &ko_vf[i];
	memset(kobj, 0, sizeof(struct kobject));
	kobj->parent = &ko_bfa_root[i];
	strcpy(kobj->name, "vf");
	kobj->ktype = &ktype_vf;
	if (kobject_register(kobj) == -EINVAL)
		return;
#endif

}

void
bfad_device_sysfs_remove(bfad_t *bfad)
{
	int             i;

	device_remove_file(&bfad->pcidev->dev, &dev_attr_vf_ctl);
	device_remove_file(&bfad->pcidev->dev, &dev_attr_vport_ctl);
	device_remove_file(&bfad->pcidev->dev, &dev_attr_fc4type);
	device_remove_file(&bfad->pcidev->dev, &dev_attr_bfad_ctl);

	if (bfad->inst_no >= MAX_IOC_INST) {
		printk(KERN_WARNING "ERROR(sysfs_remove): bfad->inst_no: %d\n",
		       bfad->inst_no);
		return;
	}

	/*
	 * Delete kobjs for all the VFs created
	 */
	bfad_sysfs_kobj_delete_all_vfs(bfad);

	/*
	 * Delete Rports of the base port
	 */
	bfad_sysfs_delete_all_rports(bfad, KOBJ_TYPE_ITNIM, 0,
				     &bfad_kobj_base_list);
	bfad_sysfs_delete_all_rports(bfad, KOBJ_TYPE_RPORT, 0,
				     &bfad_kobj_base_list);

	/*
	 * remove other configurable device parameters
	 */
	i = find_kobj_by_path(&bfad->pcidev->dev.kobj);
	if (i == -1)
		return;

#ifdef VF_READY
	kobject_unregister(&ko_vf[i]);
#endif
	kobject_unregister(&ko_lport[i]);
	kobject_unregister(&ko_pport[i]);
	kobject_unregister(&ko_ioc[i]);
	kobject_unregister(&ko_bfa_root[i]);
	strcpy(ko_bfa_root[i].name, "");

}

void
bfad_os_device_sysfs_create(bfad_t *bfad)
{
	bfad_device_sysfs_create(bfad);
}

void
bfad_os_device_sysfs_remove(bfad_t *bfad)
{
	bfad_device_sysfs_remove(bfad);
}

#endif /* SYSFS_CONTROL_SUPPORT */
