/* 
 * Copyright (c)  2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 * 
 * See LICENSE.bfa for copyright and licensing details.
 */
#include "fcs_fabric.h"
#include "fcs_port.h"
#include "fcs_vport.h"
#include "fcs_trcmod.h"
#include "fcs_fcxp.h"
#include "fcbuild.h"
#include "bfa_os_inc.h"
#include <protocol/md5.h>
#include <protocol/sha1.h>
#include <log/bfa_log_fcs.h>
#include <hal/bfa_hal_wwn.h>
#include <hal/bfa_hal_pport.h>
#include <protocol/fc_sp.h>
#include <fcs/bfa_fcs_auth.h>
#include "bfa_os.h"
#include "fcs_auth.h"
#include <aen/bfal_aen_audit.h>
#include <aen/bfal_aen_port.h>

BFA_TRC_FILE(FCS, AUTH);

#define BFA_FCS_AUTH_RETRY_DELAY		(2000)	/* milliseconds */

#define TEST_PASSWORD "swnewsecret" /* HACK Alert !! HACK for testing */

/**
 * FC-SP Authentication State machine implementation
 */
/**
 * Authentication state machine events
 */
typedef enum {
	BFA_FCS_AUTH_SM_START 	 		= 1,	/*!< start - send negotiate		*/
	BFA_FCS_AUTH_SM_RETRY_OP		= 2,	/*!< retry if timed out			*/
	BFA_FCS_AUTH_SM_RECV_CHALLENGE	= 3,	/*!< received challenge			*/
	BFA_FCS_AUTH_SM_RECV_REJECT		= 4,	/*!< received reject			*/
	BFA_FCS_AUTH_SM_RECV_SUCCESS	= 5,	/*!< received auth success		*/
	BFA_FCS_AUTH_SM_FCXP_CONTINUE	= 6,	/*!< fcxp available				*/
	BFA_FCS_AUTH_SM_FCXP_SENT	 	= 7,	/*!< fcxp sent completion		*/
	BFA_FCS_AUTH_SM_DELAYED	 		= 8,	/*!< timeout delay event		*/
	BFA_FCS_AUTH_SM_STOP	 		= 9,	/*!< stop event					*/
	BFA_FCS_AUTH_SM_RECV_NEG		= 10,	/*!< received negotiate 		*/
} bfa_fcs_auth_event_t;

static uint32_t global_auth_txn_id = FC_START_TXN_ID;

/*
 * forward declarations
 */
extern void bfa_fcs_auth_finished(bfa_fcs_fabric_t *fabric, auth_status_t status);
extern void fc_els_req_build(fchs_t *fchs, uint32_t d_id, uint32_t s_id, uint16_t ox_id);
static int 	bfa_fcs_auth_fill_reply(uint8_t *in_buf, uint8_t *out_buf, bfa_fcs_auth_t *auth);
static int 	bfa_fcs_auth_fill_negotiate(uint8_t *buf, bfa_fcs_auth_t *auth);
static void bfa_fcs_auth_acc_send(void *auth_cbarg, bfa_hal_fcxp_t *fcxp);
static void bfa_fcs_auth_send(void *auth_cbarg, bfa_hal_fcxp_t  *fcxp);
static void bfa_fcs_auth_delay(void *cbarg);
static void bfa_fcs_auth_restart(struct bfa_fcs_s *port);

static void bfa_fcs_auth_sm_uninit(bfa_fcs_auth_t *auth, bfa_fcs_auth_event_t event);
static void bfa_fcs_auth_sm_negotiate_send(bfa_fcs_auth_t *auth, bfa_fcs_auth_event_t event);
static void bfa_fcs_auth_sm_challenge_wait(bfa_fcs_auth_t *auth, bfa_fcs_auth_event_t event);
static void bfa_fcs_auth_sm_negotiate_retry(bfa_fcs_auth_t *auth, bfa_fcs_auth_event_t event);
static void bfa_fcs_auth_sm_reply_send(bfa_fcs_auth_t *auth, bfa_fcs_auth_event_t event);
static void bfa_fcs_auth_sm_status_wait(bfa_fcs_auth_t *auth, bfa_fcs_auth_event_t event);
static void bfa_fcs_auth_sm_success(bfa_fcs_auth_t *auth, bfa_fcs_auth_event_t event);
static void bfa_fcs_auth_sm_failed(bfa_fcs_auth_t *auth, bfa_fcs_auth_event_t event);

void bfa_fcs_auth_start(bfa_fcs_fabric_t *fabric);

static bfa_sm_table_t auth_sm_table[] = {
    { BFA_SM(bfa_fcs_auth_sm_uninit), BFA_AUTH_UNINIT },
    { BFA_SM(bfa_fcs_auth_sm_negotiate_send), BFA_AUTH_NEG_SEND },
    { BFA_SM(bfa_fcs_auth_sm_challenge_wait), BFA_AUTH_CHAL_WAIT },
    { BFA_SM(bfa_fcs_auth_sm_negotiate_retry), BFA_AUTH_NEG_RETRY },
    { BFA_SM(bfa_fcs_auth_sm_reply_send), BFA_AUTH_REPLY_SEND },
    { BFA_SM(bfa_fcs_auth_sm_status_wait), BFA_AUTH_STATUS_WAIT },
    { BFA_SM(bfa_fcs_auth_sm_success), BFA_AUTH_SUCCESS },
    { BFA_SM(bfa_fcs_auth_sm_failed), BFA_AUTH_FAILED },
};

/*
 *****************************************************************************
 * Internal functions
 *****************************************************************************
 */

/**
 * @brief exponentiation by squaring algorithm
 * @param [in] x   - value to be encrypted/decrypted
 * @param [in] n   - pubic[encryption]/private[decryption] key
 * @param [in] len - key length
 * @retval output
 */
static uint32_t bfad_auth_power (uint32_t x, uint32_t n, uint32_t len)
{
        long result = 1;
        while ( n ) {
                if ( n & 1 ) {
                        result = (result * x) % len;
                        n = n-1;
                }
                x = (x*x) % len;
                n = n/2;
        }
        return result;
}

/**
 * @brief function to decrept given an array of encrypted values
 * @param [out] str - decrypted string
 * @param [in]  in  - arry of values encrypted
 * @param [in]  len - length of the array in
 * @retval none
 */
void
bfad_auth_decrypt_string (char *str, uint32_t in[], uint32_t len)
{
        uint32_t result;
        size_t   i=0;

        while ( i < len ) {
                result = bfad_auth_power (in[i], PRIVATE_KEY, KEY_LEN);
                str[i++] = result & 0xff;
        }

}


/*
 * @brief function to post events depending on authentication enabled/disabled
 * @param[in] auth  - auth
 * @param[in] event - BFA_AUDIT_AEN_AUTH_DISABLE/BFA_AUDIT_AEN_AUTH_ENABLE
 * @return none
 */
static void
bfa_fcs_auth_aen_audit_post(bfa_fcs_auth_t *auth, bfa_audit_aen_event_t event)
{
	bfa_aen_data_t  aen_data;
	bfa_fcs_t	*fcs = auth->fabric->fcs;

	bfa_log_mod_t	*logmod = fcs->logm;
	wwn_t		pwwn = auth->fabric->bport.port_cfg.pwwn;
	char		pwwn_buf[BFA_STRING_32];
	char		*pwwn_ptr;

	pwwn_ptr = wwn2str(pwwn_buf, sizeof(pwwn_buf), pwwn);

	switch (event) {
	case BFA_AUDIT_AEN_AUTH_ENABLE:
		bfa_log(logmod, BFAL_AEN_AUDIT_AUTH_ENABLE, pwwn_ptr);
		break;
	case BFA_AUDIT_AEN_AUTH_DISABLE:
		bfa_log(logmod, BFAL_AEN_AUDIT_AUTH_DISABLE, pwwn_ptr);
		break;
	default: 
		break;
	}

	aen_data.audit.pwwn = pwwn;
	bfa_aen_post(fcs->aen, BFA_AEN_CAT_AUDIT, event, &aen_data);
}

/*
 * @brief function to post events depending on authentication success/failure
 * @param[in] auth  - auth
 * @param[in] event - BFA_PORT_AEN_AUTH_OFF/BFA_PORT_AEN_AUTH_ON
 * @return none
 */
static void
bfa_fcs_auth_aen_post(bfa_fcs_auth_t *auth, bfa_port_aen_event_t event)
{
	bfa_aen_data_t  aen_data;
	bfa_fcs_t	*fcs = auth->fabric->fcs;

	bfa_log_mod_t	*logmod = fcs->logm;
	wwn_t		pwwn = auth->fabric->bport.port_cfg.pwwn;
	char		pwwn_buf[BFA_STRING_32];
	char		*pwwn_ptr;

	pwwn_ptr = wwn2str(pwwn_buf, sizeof(pwwn_buf), pwwn);

	switch (event) {
	case BFA_PORT_AEN_AUTH_ON:
		bfa_log(logmod, BFAL_AEN_PORT_AUTH_ON, pwwn_ptr);
		break;
	case BFA_PORT_AEN_AUTH_OFF:
		bfa_log(logmod, BFAL_AEN_PORT_AUTH_OFF, pwwn_ptr);
		break;
	default: 
		break;
	}

	aen_data.port.pwwn = pwwn;
	bfa_aen_post(fcs->aen, BFA_AEN_CAT_PORT, event, &aen_data);
}

/*
 * @brief Authentication packets send completion callback
 * @param[in] fcsarg
 * @param[in] fcxp 	 	- fcxp
 * @param[in] cbarg  	- callback argument which is fcs
 * @param[in] status 	- status of the send
 * @param[in] rsp_len 	- response payload length
 * @param[in] resid_len
 * @param[in] rspfchs 	- response fchs
 *
 * @return - none
 */
static void
bfa_fcs_auth_send_comp( void *fcsarg, bfa_hal_fcxp_t *fcxp, void *cbarg,
						  bfa_status_t status, uint32_t rsp_len,
						  uint32_t resid_len, fchs_t *rspfchs)
{
	bfa_fcs_auth_t*		auth = (bfa_fcs_auth_t *)cbarg;
	auth_els_msg_t*		els_cmd;
	fc_ls_rjt_t* 		ls_rjt;

	bfa_trc(auth->fabric->fcs, status);

	if (status != BFA_STATUS_OK) {
		bfa_trc(auth->fabric->fcs, status);
		bfa_sm_send_event(auth, BFA_FCS_AUTH_SM_RETRY_OP);
		return;
	}

	els_cmd = (auth_els_msg_t *)(BFA_FCXP_RSP_PLD(fcxp));

	switch (els_cmd->auth_els_code) {

    case FC_ELS_ACC :
		bfa_trc(auth->fabric->fcs, els_cmd->auth_els_code);
        break;

   	case FC_ELS_LS_RJT :
		ls_rjt = (fc_ls_rjt_t *)BFA_FCXP_RSP_PLD(fcxp);

		bfa_trc(auth->fabric->fcs, ls_rjt->reason_code);
		bfa_trc(auth->fabric->fcs, ls_rjt->reason_code_expl);
		bfa_sm_send_event(auth, BFA_FCS_AUTH_SM_RETRY_OP);
        break;

    default :
        bfa_assert(0);
		bfa_sm_send_event(auth, BFA_FCS_AUTH_SM_RETRY_OP);
        break;
	}

}

void
bfa_fcs_auth_uf_recv(bfa_fcs_fabric_t *fabric, int len)
{
	int					state;
	bfa_fcs_auth_t*		auth = &fabric->auth;
	auth_els_msg_t*		els_cmd;

	bfa_trc(auth->fabric->fcs, len);

	els_cmd = (auth_els_msg_t *)auth->response;

	bfa_trc(auth->fabric->fcs, els_cmd->auth_els_code);
	bfa_trc(auth->fabric->fcs, els_cmd->auth_msg_code);

	bfa_fcs_auth_acc_send(auth, NULL);

	/* If authentication is disabled, don't process packets */
	state = bfa_sm_to_state(auth_sm_table, fabric->auth.sm);
	if ((fabric->auth.policy != BFA_TRUE) ||
        (state < BFA_AUTH_UNINIT || state > BFA_AUTH_FAILED)) {
		return;
	}

	switch (els_cmd->auth_msg_code) {

	case FC_AUTH_MC_AUTH_NEG:
		/*
		 * FC-SP Specification. Section 5.10
		 * The switch may restart authentication by sending negotiate message
		 */
		bfa_sm_send_event(auth, BFA_FCS_AUTH_SM_RECV_NEG);
		break;

	case FC_AUTH_MC_DHCHAP_CHAL:
		bfa_sm_send_event(auth, BFA_FCS_AUTH_SM_RECV_CHALLENGE);
		break;	

	case FC_AUTH_MC_AUTH_RJT:
		bfa_sm_send_event(auth, BFA_FCS_AUTH_SM_RECV_REJECT);
		break;

	case FC_AUTH_MC_DHCHAP_SUCC:
		/* switch is sending multiple SUCCESS messages some times */
		if (auth->status != FC_AUTH_STATE_SUCCESS) {
		    bfa_sm_send_event(auth, BFA_FCS_AUTH_SM_RECV_SUCCESS);
		}
		break;

	default:
		bfa_trc(auth->fabric->fcs, els_cmd->auth_msg_code);
		bfa_assert(0);
		break;

	}

}

void
compute_md5_rsa(uint8_t *key, uint8_t len, unsigned char *digest)
{
  	MD5_CTX context;

  	MD5Init (&context);
  	MD5Update (&context, key, len);
  	MD5Final (&context);

	memcpy(digest, context.digest, FC_MD5_DIGEST_LEN);
}
void

compute_sha1_digest(uint8_t *key, uint8_t len, unsigned char *digest)
{
	SHA1Context sha;


  	SHA1Reset (&sha);
        SHA1Input(&sha, (const unsigned char *) key, len);
	SHA1Result(&sha, digest);

}

/*
 * @brief fill authentication reply packet
 *
 * @param[in] in_buf 	- challenge packet received from the switch
 * @param[out] out_buf 	- reply packet
 * @param[in] auth 		- auth
 *
 * @return - lenght of the replay packet
 */
int
bfa_fcs_auth_fill_reply(uint8_t *in_buf, uint8_t *out_buf, bfa_fcs_auth_t *auth)
{
	uint32_t*	value;		/* value of the parameter */
	int		size = 0, key_length = 0, digest_len; 
	uint8_t		key[FC_MAX_CHAP_KEY_LEN], secret[FC_MAX_AUTH_SECRET_LEN],
			digest[FC_MAX_CHAP_KEY_LEN], *ptr;

	auth_dhchap_chal_msg_t*		chal = (auth_dhchap_chal_msg_t *)in_buf;
	auth_dhchap_reply_msg_t*	msg  = (auth_dhchap_reply_msg_t *)out_buf;

	bfa_assert(msg);

	bfa_trc(auth->fabric->fcs, bfa_os_htonl(chal->hash_id));

	bfa_assert ( (bfa_os_htonl(chal->hash_id) == FC_AUTH_HASH_FUNC_MD5) ||
		     (bfa_os_htonl(chal->hash_id) == FC_AUTH_HASH_FUNC_SHA_1) );

	/* get the MD5 digest */
	/* HACK Alert !! HACK for testing */
	if (strlen(auth->secret) == 0) {
		strncpy(secret, TEST_PASSWORD, sizeof (TEST_PASSWORD)); 
		secret[sizeof (TEST_PASSWORD)] = '\0';
	}
	else {
		strncpy(secret, auth->secret, strlen(auth->secret));
		secret[strlen(auth->secret)] = '\0';
	}

	key[0] = (unsigned char )bfa_os_htonl(chal->hdr.trans_id);
	key_length = 1;

	memcpy(&key[key_length], secret, strlen(secret));
	key_length += strlen (secret);

	memcpy(&key[key_length], &chal->chal_val[0], bfa_os_htonl(chal->chal_val_len));
	key_length += bfa_os_htonl(chal->chal_val_len);

	bfa_trc(auth->fabric->fcs, 101010);
	bfa_trc(auth->fabric->fcs, bfa_os_htonl(chal->hash_id));

	if (chal->hash_id == bfa_os_htonl(FC_AUTH_HASH_FUNC_MD5)) {
	    compute_md5_rsa(key, (uint8_t)key_length, digest);
	    digest_len = FC_MD5_DIGEST_LEN;
	}
	else {
	    compute_sha1_digest(key, (uint8_t)key_length, digest);
	    digest_len = FC_SHA1_DIGEST_LEN;
	}

	/* fill in the packet */	
	msg->hdr.auth_els_code = FC_AUTH_ELS_COMMAND_CODE;
	msg->hdr.auth_els_flag = 0;
	msg->hdr.auth_msg_code = FC_AUTH_MC_DHCHAP_REPLY;
	msg->hdr.proto_version = FC_AUTH_PROTO_VER_1;
	msg->hdr.trans_id = chal->hdr.trans_id;

	msg->hdr.msg_len  = 
			bfa_os_htonl(
			  FC_AUTH_RESP_VALUE_LEN_SZ
			+ digest_len
			+ FC_AUTH_PROTO_PARAM_LEN_SZ
			+ FC_AUTH_PROTO_PARAM_LEN_SZ);

	value = (uint32_t *)((uint8_t *)msg + sizeof (auth_els_msg_t));
	*value = bfa_os_htonl(digest_len);

	ptr = (uint8_t *)((uint8_t *)value + sizeof (uint32_t));
	memcpy(ptr, digest, digest_len);

	/* DHgIDList value is 0 for DH NULL option */
	value = (uint32_t *)(ptr + digest_len);
	*value = 0; /* @todo, should read from the config */

	/* Challenge value length, shall be 0 since we are not doing bi-directional */
	value = (uint32_t *)((uint8_t *)value + sizeof (uint32_t));
	*value = 0; /* @todo, should read from the config */

	size = 	sizeof (auth_dhchap_reply_msg_t)
			+ FC_AUTH_RESP_VALUE_LEN_SZ 
			+ digest_len
			+ FC_AUTH_PROTO_PARAM_VAL_SZ
			+ FC_AUTH_PROTO_PARAM_VAL_SZ;
			
	return size;
}

/*
 * @brief fill authentication negotiate packet
 * @param[out] buf 	- negotiate packet
 * @return - length of the negotiate packet
 */
int
bfa_fcs_auth_fill_negotiate(uint8_t *buf, bfa_fcs_auth_t *auth)
{
	int 		acount, /* hash algorithm count */
			dcount = FC_MAX_DHG_SUPPORTED, nproto = FC_MAX_PROTO_SUPPORTED;
	uint32_t*	value;	/* value of the parameter */
	int		size = 0;

	auth_dhchap_neg_msg_t*	msg = (auth_dhchap_neg_msg_t *)buf;
	dhchap_param_format_t*  param;

	bfa_assert(msg);


	if (auth->algo == BFA_AUTH_ALGO_MD5 || auth->algo == BFA_AUTH_ALGO_SHA1)
	    acount = 1;
	else
	    acount = 2;

	bfa_trc(auth->fabric->fcs, acount);

        size = (sizeof(auth_dhchap_neg_msg_t) +
                sizeof(dhchap_param_format_t) +
                sizeof(dhchap_param_format_t) +
                (acount + dcount) * (FC_AUTH_PROTO_PARAM_VAL_SZ));

	bfa_trc(auth->fabric->fcs, 4545);
	bfa_trc(auth->fabric->fcs, size);

	msg->hdr.auth_els_code = FC_AUTH_ELS_COMMAND_CODE;
	msg->hdr.auth_els_flag = 0;
	msg->hdr.auth_msg_code = FC_AUTH_MC_AUTH_NEG;
	msg->hdr.proto_version = FC_AUTH_PROTO_VER_1;
   	msg->hdr.msg_len  =
			bfa_os_htonl(		
			sizeof (auth_name_t)
                        + sizeof (dhchap_param_format_t)
			+ acount * FC_AUTH_PROTO_PARAM_VAL_SZ
                        + sizeof (dhchap_param_format_t)
			+ dcount * FC_AUTH_PROTO_PARAM_VAL_SZ
                        + FC_AUTH_PROTO_PARAM_LEN_SZ
                        + nproto * FC_AUTH_PROTO_PARAM_VAL_SZ
                        + FC_AUTH_NUM_USABLE_PROTO_LEN_SZ);

	bfa_trc(auth->fabric->fcs, bfa_os_htonl(msg->hdr.msg_len));

	global_auth_txn_id++;
	msg->hdr.trans_id = bfa_os_htonl(global_auth_txn_id);
	bfa_trc(auth->fabric->fcs, global_auth_txn_id);

	memcpy(&msg->nego.auth_ini_name.name, &auth->fabric->bport.port_cfg.pwwn, sizeof (wwn_t));
	msg->nego.auth_ini_name.name_tag = bfa_os_htons(1);
	msg->nego.auth_ini_name.name_len = bfa_os_htons(sizeof (wwn_t));
	msg->nego.usable_auth_protos = bfa_os_htonl(nproto);

	msg->nego.proto_params[0].proto_param_len = 
				bfa_os_htonl(
				sizeof (dhchap_param_format_t) +
				acount * FC_AUTH_PROTO_PARAM_VAL_SZ + 
				sizeof (dhchap_param_format_t) +
				dcount * FC_AUTH_PROTO_PARAM_VAL_SZ +
				FC_AUTH_PROTO_PARAM_LEN_SZ);

	msg->nego.proto_params[0].proto_id = bfa_os_htonl(FC_AUTH_PROTO_DHCHAP);

	bfa_trc(auth->fabric->fcs, bfa_os_htonl (msg->nego.proto_params[0].proto_param_len)); 

	/* fill in the algo information */
	param = (dhchap_param_format_t *) ((uint8_t *)msg + sizeof(auth_dhchap_neg_msg_t));
	param->tag = bfa_os_htons(FC_AUTH_NEG_DHCHAP_HASHLIST);
	value = (uint32_t *)((uint8_t *)param + sizeof (dhchap_param_format_t));

	if (auth->algo == BFA_AUTH_ALGO_MD5) {
	    param->word_cnt = bfa_os_htons(1);
	    *value = bfa_os_htonl(FC_AUTH_HASH_FUNC_MD5);
	}
	else if (auth->algo == BFA_AUTH_ALGO_SHA1) {
	    param->word_cnt = bfa_os_htons(1);
	    *value = bfa_os_htonl(FC_AUTH_HASH_FUNC_SHA_1);
	}
	else if (auth->algo == BFA_AUTH_ALGO_MS) {
	    param->word_cnt = bfa_os_htons(2);
	    *value = bfa_os_htonl(FC_AUTH_HASH_FUNC_MD5);
	    value += 4;
	    *value = bfa_os_htonl(FC_AUTH_HASH_FUNC_SHA_1);
	}
	else if (auth->algo == BFA_AUTH_ALGO_SM) {
	    param->word_cnt = bfa_os_htons(2);
	    *value = bfa_os_htonl(FC_AUTH_HASH_FUNC_SHA_1);
	    value += 4;
	    *value = bfa_os_htonl(FC_AUTH_HASH_FUNC_MD5);
	}

	/* fill in the DG information */
	param = (dhchap_param_format_t *)((uint8_t *)param + sizeof
			(dhchap_param_format_t) + 
			acount * FC_AUTH_PROTO_PARAM_VAL_SZ);
	param->tag = bfa_os_htons(FC_AUTH_NEG_DHCHAP_DHG_ID_LIST);
	param->word_cnt = bfa_os_htons(1);
	value = (uint32_t *)((uint8_t *)param + sizeof (dhchap_param_format_t));
	*value = bfa_os_htonl(FC_AUTH_DH_GID_0_DHG_NULL); /* @todo, should read from the config */

	return size;
}

/**
 * @brief allocate and send authentication packets
 *
 * @param[in] auth 	- auth
 * @param[in] fcxp_alloced - fcxp
 *
 * @return - BFA_TRUE if successful otherwise BFA_FALSE
 */
static void
bfa_fcs_auth_send(void *auth_cbarg, bfa_hal_fcxp_t *fcxp_alloced)
{
	bfa_fcs_auth_t*	auth = (bfa_fcs_auth_t *)auth_cbarg;
    fchs_t*         in_fchs = (fchs_t *)(auth->response - sizeof (fchs_t));
	bfa_hal_t*		hal = auth->fabric->fcs->hal;
	bfa_hal_fcxp_t*		fcxp;
	uint16_t		reqlen = 0;
	fchs_t			fchs;
	uint32_t 		d_id = bfa_os_hton3b(FC_FABRIC_PORT), s_id = 0, ox_id = 0;
	uint8_t*		ptr;

	s_id = auth->fabric->bport.pid;

	fcxp = fcxp_alloced ? fcxp_alloced : bfa_fcs_fcxp_alloc(auth->fabric->fcs);
	if (!fcxp) {
		bfa_hal_fcxp_alloc_wait(hal, &auth->fabric->fcxp_wqe,
								bfa_fcs_auth_send, auth);
		return;
	}

	bfa_assert(fcxp);

	auth->fcxp = fcxp;

	/**
	 * @todo Get parameters from HAL.
	 */
	fc_els_req_build(&fchs, d_id, s_id, (uint16_t)ox_id);

	if (auth->sentcode == FC_AUTH_MC_AUTH_NEG) {
		auth->stats.auth_tx_stats.auth_negs++;
		reqlen = (uint16_t)bfa_fcs_auth_fill_negotiate(bfa_hal_fcxp_get_reqbuf(fcxp), auth);
	}
	else if (auth->sentcode == FC_AUTH_MC_DHCHAP_REPLY) {
		auth->stats.auth_tx_stats.dhchap_replies++;
		reqlen = (uint16_t)bfa_fcs_auth_fill_reply(auth->response, 
							 bfa_hal_fcxp_get_reqbuf(fcxp), auth);
		memcpy(&fchs, in_fchs, sizeof (fchs_t));
		fchs.d_id = d_id;
		fchs.s_id = s_id;
		fchs.ox_id = in_fchs->ox_id;
		ptr = (uint8_t *)bfa_hal_fcxp_get_reqbuf(fcxp);
		bfa_trc(auth->fabric->fcs, reqlen);
	}
	else
		bfa_assert(0);

	bfa_hal_fcxp_send(fcxp,
					  NULL,
					  auth->fabric->vf_id,
					  BFA_FALSE, FC_CLASS_3,
					  reqlen,
					  &fchs,
					  bfa_fcs_auth_send_comp,
					  auth,
					  FC_MAX_PDUSZ,
					  FC_RA_TOV);

	bfa_sm_send_event(auth, BFA_FCS_AUTH_SM_FCXP_SENT);

}

static void
bfa_fcs_acc_send_comp( void *fcsarg, bfa_hal_fcxp_t *fcxp, void *cbarg,
						  bfa_status_t status, uint32_t rsp_len,
						  uint32_t resid_len, fchs_t *rspfchs)
{
	bfa_fcs_auth_t*	auth = (bfa_fcs_auth_t *)cbarg;

	bfa_trc(auth->fabric->fcs, 8888);
}

static void
bfa_fcs_auth_acc_send(void *auth_cbarg, bfa_hal_fcxp_t*	fcxp_alloced)
{
	bfa_fcs_auth_t*		auth = (bfa_fcs_auth_t *)auth_cbarg;
	fchs_t*			in_fchs = (fchs_t *)(auth->response - sizeof (fchs_t));
	bfa_hal_t*		hal = auth->fabric->fcs->hal;
	bfa_hal_fcxp_t*		fcxp;
	uint16_t		len = 0;
	fchs_t			fchs;

	bfa_trc(auth->fabric->fcs, 7777);

	fcxp = fcxp_alloced ? fcxp_alloced : bfa_fcs_fcxp_alloc(auth->fabric->fcs);
	if (!fcxp) {
		bfa_hal_fcxp_alloc_wait(hal, &auth->fabric->fcxp_wqe,
								bfa_fcs_auth_acc_send, auth);
		return;
	}

	bfa_assert(fcxp);

	auth->fcxp = fcxp;

	len = fc_ls_acc_build(&fchs, bfa_hal_fcxp_get_reqbuf(fcxp),
						  in_fchs->s_id, in_fchs->d_id, in_fchs->ox_id);

	bfa_hal_fcxp_send(fcxp,
					  NULL,
					  auth->fabric->vf_id,
					  BFA_FALSE, FC_CLASS_3,
					  len,
					  &fchs,
					  bfa_fcs_acc_send_comp,
					  auth,
					  FC_MAX_PDUSZ,
					  0);

}

static void
bfa_fcs_auth_delay(void *cbarg)
{
	bfa_fcs_auth_t	*auth = cbarg;
	bfa_trc(auth->fabric->fcs, 9999);
	bfa_trc(auth->fabric->fcs, auth->sentcode);
	bfa_trc(auth->fabric->fcs, bfa_sm_to_state(auth_sm_table, auth->sm));

	bfa_sm_send_event(auth, BFA_FCS_AUTH_SM_DELAYED);
}

/*
 *****************************************************************************
 * State Machine functions
 *****************************************************************************
 */

/*
 * @brief authentication start - send negotiate packet
 * @param[in] auth 	- auth
 * @param[in] event - event
 *
 * @return - none
 */
void
bfa_fcs_auth_sm_uninit(bfa_fcs_auth_t *auth, bfa_fcs_auth_event_t event)
{
	auth_rjt_msg_t*     rej_msg;

	bfa_trc(auth->fabric->fcs, event);

	auth->status = FC_AUTH_STATE_INPROGRESS;
	auth->fcxp = NULL;
	auth->rjt_code = 0;
	auth->rjt_code_exp = 0;

	switch (event) {

	case BFA_FCS_AUTH_SM_START:
		auth->nretries = FC_MAX_AUTH_RETRIES;
		bfa_sm_set_state(auth, bfa_fcs_auth_sm_negotiate_send);
		auth->sentcode = FC_AUTH_MC_AUTH_NEG;
		auth->response = NULL;
		bfa_fcs_auth_send(auth, NULL);
		break;

	case BFA_FCS_AUTH_SM_STOP:
		if (auth->fcxp != NULL)
			bfa_hal_fcxp_discard(auth->fcxp);

		break;

	case BFA_FCS_AUTH_SM_RECV_NEG:
		bfa_fcs_auth_restart(auth->fabric->fcs);
		break;

	case BFA_FCS_AUTH_SM_RECV_REJECT:
		rej_msg = (auth_rjt_msg_t *)auth->response;

		bfa_trc(auth->fabric->fcs, rej_msg->reason_code);
		bfa_trc(auth->fabric->fcs, rej_msg->reason_code_exp);

		if ((rej_msg->reason_code == FC_AUTH_RJT_CODE_LOGICAL_ERR) &&
		    (rej_msg->reason_code_exp == FC_AUTH_CEXP_RESTART_AUTH_PROTO)) {
			bfa_fcs_auth_restart(auth->fabric->fcs);
		}

		break;

	default:
		bfa_assert(0);
	}
}

/*
 * @brief authentication negotiate is being sent
 * @param[in] auth 	- auth
 * @param[in] event - event
 *
 * @return - none
 */
void
bfa_fcs_auth_sm_negotiate_send(bfa_fcs_auth_t *auth, bfa_fcs_auth_event_t event)
{
	auth_rjt_msg_t*     rej_msg;

	bfa_trc(auth->fabric->fcs, event);

	switch (event) {

    case BFA_FCS_AUTH_SM_FCXP_SENT:
		/* start a timer to handle switch not sending either challenge/status */
		bfa_hal_timer_start(auth->fabric->fcs->hal, &auth->delay_timer,
							bfa_fcs_auth_delay, auth,
							BFA_FCS_AUTH_RETRY_DELAY);

		bfa_sm_set_state(auth, bfa_fcs_auth_sm_challenge_wait);
		break;

	case BFA_FCS_AUTH_SM_STOP:
		bfa_sm_set_state(auth, bfa_fcs_auth_sm_uninit);
		if (auth->fcxp != NULL)
			bfa_hal_fcxp_discard(auth->fcxp);

		break;

	case BFA_FCS_AUTH_SM_RECV_NEG:
		bfa_fcs_auth_restart(auth->fabric->fcs);
		break;

	case BFA_FCS_AUTH_SM_RECV_REJECT:
		rej_msg = (auth_rjt_msg_t *)auth->response;

		bfa_trc(auth->fabric->fcs, rej_msg->reason_code);
		bfa_trc(auth->fabric->fcs, rej_msg->reason_code_exp);

		if ((rej_msg->reason_code == FC_AUTH_RJT_CODE_LOGICAL_ERR) &&
		    (rej_msg->reason_code_exp == FC_AUTH_CEXP_RESTART_AUTH_PROTO)) {
			bfa_fcs_auth_restart(auth->fabric->fcs);
		}

		break;

	default:
		bfa_assert(0);
	
	}

}

/*
 * @brief authentication negotiate sent - waiting for challenge message
 * @param[in] auth 	- auth
 * @param[in] event - event
 *
 * @return - none
 */
void
bfa_fcs_auth_sm_challenge_wait(bfa_fcs_auth_t *auth, bfa_fcs_auth_event_t event)
{
	auth_dhchap_chal_msg_t*    chal;
	auth_rjt_msg_t*            rej_msg;

	bfa_trc(auth->fabric->fcs, event);

	switch (event) {
	case BFA_FCS_AUTH_SM_RECV_NEG:
		bfa_fcs_auth_restart(auth->fabric->fcs);
		break;

	case BFA_FCS_AUTH_SM_DELAYED:
		bfa_sm_set_state(auth, bfa_fcs_auth_sm_negotiate_retry);
		bfa_sm_send_event(auth, BFA_FCS_AUTH_SM_DELAYED);
		break;

    	case BFA_FCS_AUTH_SM_RETRY_OP:
		bfa_hal_timer_stop(&auth->delay_timer);
		if (auth->nretries) {
			bfa_trc(auth->fabric->fcs, auth->nretries);
		    	auth->nretries--;
			bfa_sm_set_state(auth, bfa_fcs_auth_sm_negotiate_retry);
			bfa_hal_timer_start(auth->fabric->fcs->hal, &auth->delay_timer,
								bfa_fcs_auth_delay, auth,
								BFA_FCS_AUTH_RETRY_DELAY);
		}
		else {
			bfa_trc(auth->fabric->fcs, auth->nretries);
			auth->stats.auth_failures++;

			bfa_fcs_auth_aen_post(auth, BFA_PORT_AEN_AUTH_OFF);

			auth->status = FC_AUTH_STATE_FAILED;
			auth->rjt_code = BFA_AUTH_FAIL_TIMEOUT;
			auth->rjt_code_exp = BFA_AUTH_FAIL_TIMEOUT;
			bfa_sm_set_state(auth, bfa_fcs_auth_sm_failed);
			bfa_fcs_auth_finished(auth->fabric, FC_AUTH_STATE_FAILED);
		}
		break;

	case BFA_FCS_AUTH_SM_RECV_CHALLENGE:
		bfa_hal_timer_stop(&auth->delay_timer);
		chal = (auth_dhchap_chal_msg_t *)auth->response;
		bfa_trc(auth->fabric->fcs, bfa_os_htonl(chal->hdr.trans_id));

		auth->stats.auth_rx_stats.dhchap_challenges++;
		auth->sentcode = FC_AUTH_MC_DHCHAP_REPLY;
		bfa_sm_set_state(auth, bfa_fcs_auth_sm_reply_send);
		bfa_fcs_auth_send(auth, NULL);
		break;

	case BFA_FCS_AUTH_SM_RECV_REJECT:
		bfa_hal_timer_stop(&auth->delay_timer);
		rej_msg = (auth_rjt_msg_t *)auth->response;

		bfa_trc(auth->fabric->fcs, rej_msg->reason_code);
		bfa_trc(auth->fabric->fcs, rej_msg->reason_code_exp);

		if ((rej_msg->reason_code == FC_AUTH_RJT_CODE_LOGICAL_ERR) &&
		    (rej_msg->reason_code_exp == FC_AUTH_CEXP_RESTART_AUTH_PROTO)) {
			bfa_fcs_auth_restart(auth->fabric->fcs);
			break;
		}

		auth->stats.auth_rx_stats.auth_rjts++;

		bfa_fcs_auth_aen_post(auth, BFA_PORT_AEN_AUTH_OFF);

		auth->status = FC_AUTH_STATE_FAILED;
		auth->rjt_code = rej_msg->reason_code;
		auth->rjt_code_exp = rej_msg->reason_code_exp;
		bfa_sm_set_state(auth, bfa_fcs_auth_sm_failed);
		bfa_fcs_auth_finished(auth->fabric, FC_AUTH_STATE_FAILED);
		break;

	case BFA_FCS_AUTH_SM_STOP:
		bfa_hal_timer_stop(&auth->delay_timer);
		bfa_sm_set_state(auth, bfa_fcs_auth_sm_uninit);
		if (auth->fcxp != NULL)
			bfa_hal_fcxp_discard(auth->fcxp);

		break;

	default:
		bfa_assert(0);
	
	}

}

/*
 * @brief authentication retry function, handles timeouts
 * @param[in] auth 	- auth
 * @param[in] event - event
 *
 * @return - none
 */
void
bfa_fcs_auth_sm_negotiate_retry(bfa_fcs_auth_t *auth, bfa_fcs_auth_event_t event)
{
	auth_rjt_msg_t*     rej_msg;

	bfa_trc(auth->fabric->fcs, event);

	switch (event) {

	case BFA_FCS_AUTH_SM_DELAYED:
		if (auth->nretries) {
			bfa_trc(auth->fabric->fcs, auth->nretries);
		    	auth->nretries--;
			auth->sentcode = FC_AUTH_MC_AUTH_NEG;
			auth->response = NULL;
			bfa_sm_set_state(auth, bfa_fcs_auth_sm_negotiate_send);
			bfa_fcs_auth_send(auth, NULL);
		}
		else {
			bfa_trc(auth->fabric->fcs, event);

			bfa_fcs_auth_aen_post(auth, BFA_PORT_AEN_AUTH_OFF);

			auth->status = FC_AUTH_STATE_FAILED;
			auth->rjt_code = BFA_AUTH_FAIL_TIMEOUT;
			auth->rjt_code_exp = BFA_AUTH_FAIL_TIMEOUT;
			auth->stats.auth_failures++;
			bfa_sm_set_state(auth, bfa_fcs_auth_sm_failed);
			bfa_fcs_auth_finished(auth->fabric, FC_AUTH_STATE_FAILED);
		}
		
		break;

	case BFA_FCS_AUTH_SM_STOP:
		bfa_sm_set_state(auth, bfa_fcs_auth_sm_uninit);
		if (auth->fcxp != NULL)
			bfa_hal_fcxp_discard(auth->fcxp);

		break;

	case BFA_FCS_AUTH_SM_RECV_NEG:
		bfa_fcs_auth_restart(auth->fabric->fcs);
		break;

	case BFA_FCS_AUTH_SM_RECV_REJECT:
		rej_msg = (auth_rjt_msg_t *)auth->response;

		bfa_trc(auth->fabric->fcs, rej_msg->reason_code);
		bfa_trc(auth->fabric->fcs, rej_msg->reason_code_exp);

		if ((rej_msg->reason_code == FC_AUTH_RJT_CODE_LOGICAL_ERR) &&
		    (rej_msg->reason_code_exp == FC_AUTH_CEXP_RESTART_AUTH_PROTO)) {
			bfa_fcs_auth_restart(auth->fabric->fcs);
		}

		break;

	default:
		bfa_assert(0);
	}
}

/*
 * @brief authentication reply is being sent
 * @param[in] auth 	- auth
 * @param[in] event - event
 *
 * @return - none
 */
void
bfa_fcs_auth_sm_reply_send(bfa_fcs_auth_t *auth, bfa_fcs_auth_event_t event)
{
	auth_rjt_msg_t*     rej_msg;

	bfa_trc(auth->fabric->fcs, event);

	switch (event) {

    case BFA_FCS_AUTH_SM_FCXP_SENT:
		/* start a timer to handle switch not sending either challenge/status */
		bfa_hal_timer_start(auth->fabric->fcs->hal, &auth->delay_timer,
							bfa_fcs_auth_delay, auth,
							BFA_FCS_AUTH_RETRY_DELAY);
		bfa_sm_set_state(auth, bfa_fcs_auth_sm_status_wait);
		break;

	case BFA_FCS_AUTH_SM_STOP:
		bfa_sm_set_state(auth, bfa_fcs_auth_sm_uninit);
		if (auth->fcxp != NULL)
			bfa_hal_fcxp_discard(auth->fcxp);

		break;

	case BFA_FCS_AUTH_SM_RECV_NEG:
		bfa_fcs_auth_restart(auth->fabric->fcs);
		break;

	case BFA_FCS_AUTH_SM_RECV_REJECT:
		rej_msg = (auth_rjt_msg_t *)auth->response;

		bfa_trc(auth->fabric->fcs, rej_msg->reason_code);
		bfa_trc(auth->fabric->fcs, rej_msg->reason_code_exp);

		if ((rej_msg->reason_code == FC_AUTH_RJT_CODE_LOGICAL_ERR) &&
		    (rej_msg->reason_code_exp == FC_AUTH_CEXP_RESTART_AUTH_PROTO)) {
			bfa_fcs_auth_restart(auth->fabric->fcs);
		}

		break;

	default:
		bfa_assert(0);
	
	}

}

/*
 * @brief authentication negotiate sent state event handler
 * @param[in] auth 	- auth
 * @param[in] event - event
 *
 * @return - none
 */
void
bfa_fcs_auth_sm_status_wait(bfa_fcs_auth_t *auth, bfa_fcs_auth_event_t event)
{
	auth_rjt_msg_t*	rej_msg;

	bfa_trc(auth->fabric->fcs, event);


	switch (event) {

	case BFA_FCS_AUTH_SM_DELAYED:
		bfa_sm_set_state(auth, bfa_fcs_auth_sm_negotiate_retry);
		bfa_sm_send_event(auth, BFA_FCS_AUTH_SM_DELAYED);
		break;


    	case BFA_FCS_AUTH_SM_RETRY_OP:
		bfa_hal_timer_stop(&auth->delay_timer);
		if (auth->nretries) {
			bfa_trc(auth->fabric->fcs, auth->nretries);
		    auth->nretries--;
			bfa_sm_set_state(auth, bfa_fcs_auth_sm_negotiate_retry);
			bfa_hal_timer_start(auth->fabric->fcs->hal, &auth->delay_timer,
								bfa_fcs_auth_delay, auth,
								BFA_FCS_AUTH_RETRY_DELAY);
		}
		else {
			bfa_trc(auth->fabric->fcs, auth->nretries);
			auth->stats.auth_failures++;

			bfa_fcs_auth_aen_post(auth, BFA_PORT_AEN_AUTH_OFF);

			auth->status = FC_AUTH_STATE_FAILED;
			auth->rjt_code = BFA_AUTH_FAIL_TIMEOUT;
			auth->rjt_code_exp = BFA_AUTH_FAIL_TIMEOUT;
			bfa_sm_set_state(auth, bfa_fcs_auth_sm_failed);
			bfa_fcs_auth_finished(auth->fabric, FC_AUTH_STATE_FAILED);
		}
		break;

	case BFA_FCS_AUTH_SM_RECV_REJECT:
		bfa_hal_timer_stop(&auth->delay_timer);
		rej_msg = (auth_rjt_msg_t *)auth->response;

		bfa_trc(auth->fabric->fcs, rej_msg->reason_code);
		bfa_trc(auth->fabric->fcs, rej_msg->reason_code_exp);

		if ((rej_msg->reason_code == FC_AUTH_RJT_CODE_LOGICAL_ERR) &&
		    (rej_msg->reason_code_exp == FC_AUTH_CEXP_RESTART_AUTH_PROTO)) {
			bfa_fcs_auth_restart(auth->fabric->fcs);
			break;
		}

		auth->stats.auth_rx_stats.auth_rjts++;

		bfa_fcs_auth_aen_post(auth, BFA_PORT_AEN_AUTH_OFF);

		auth->status = FC_AUTH_STATE_FAILED;
		auth->rjt_code = rej_msg->reason_code;
		auth->rjt_code_exp = rej_msg->reason_code_exp;
		bfa_sm_set_state(auth, bfa_fcs_auth_sm_failed);
		bfa_fcs_auth_finished(auth->fabric, FC_AUTH_STATE_FAILED);
		break;

	case BFA_FCS_AUTH_SM_RECV_SUCCESS:
		bfa_hal_timer_stop(&auth->delay_timer);
		auth->stats.auth_successes++;
		auth->stats.auth_rx_stats.dhchap_successes++;

		bfa_fcs_auth_aen_post(auth, BFA_PORT_AEN_AUTH_ON);

		auth->status = FC_AUTH_STATE_SUCCESS;
		bfa_sm_set_state(auth, bfa_fcs_auth_sm_success);
		bfa_fcs_auth_finished(auth->fabric, FC_AUTH_STATE_SUCCESS);
		break;

	case BFA_FCS_AUTH_SM_STOP:
		bfa_hal_timer_stop(&auth->delay_timer);
		bfa_sm_set_state(auth, bfa_fcs_auth_sm_uninit);
		if (auth->fcxp != NULL)
			bfa_hal_fcxp_discard(auth->fcxp);

		break;

	default:
		bfa_assert(0);
	
	}

}

/*
 * @brief authentication successfully completed
 * @param[in] auth  - auth
 * @param[in] event - event
 *
 * @return - none
 */
void
bfa_fcs_auth_sm_success(bfa_fcs_auth_t *auth, bfa_fcs_auth_event_t event)
{
	auth_rjt_msg_t*     rej_msg;

	bfa_trc(auth->fabric->fcs, event);
	bfa_trc(auth->fabric->fcs, 9999);

	switch (event) {
	case BFA_FCS_AUTH_SM_STOP:
		bfa_sm_set_state(auth, bfa_fcs_auth_sm_uninit);
		if (auth->fcxp != NULL)
			bfa_hal_fcxp_discard(auth->fcxp);
		break;

	case BFA_FCS_AUTH_SM_RECV_NEG:
		bfa_fcs_auth_restart(auth->fabric->fcs);
		break;

	case BFA_FCS_AUTH_SM_RECV_REJECT:
		rej_msg = (auth_rjt_msg_t *)auth->response;

		bfa_trc(auth->fabric->fcs, rej_msg->reason_code);
		bfa_trc(auth->fabric->fcs, rej_msg->reason_code_exp);

		if ((rej_msg->reason_code == FC_AUTH_RJT_CODE_LOGICAL_ERR) &&
		    (rej_msg->reason_code_exp == FC_AUTH_CEXP_RESTART_AUTH_PROTO)) {
			bfa_fcs_auth_restart(auth->fabric->fcs);
		}

		break;

	default:
		bfa_assert(0);
		break;
	}
}

/*
 * @brief authentication failed
 * @param[in] auth  - auth
 * @param[in] event - event
 *
 * @return - none
 */
void
bfa_fcs_auth_sm_failed(bfa_fcs_auth_t *auth, bfa_fcs_auth_event_t event)
{
	auth_rjt_msg_t*     rej_msg;

	bfa_trc(auth->fabric->fcs, event);
	bfa_trc(auth->fabric->fcs, 6666);

	switch (event) {
	case BFA_FCS_AUTH_SM_STOP:
		bfa_sm_set_state(auth, bfa_fcs_auth_sm_uninit);
		if (auth->fcxp != NULL)
			bfa_hal_fcxp_discard(auth->fcxp);
		break;

	case BFA_FCS_AUTH_SM_RECV_NEG:
		bfa_fcs_auth_restart(auth->fabric->fcs);
		break;

	case BFA_FCS_AUTH_SM_RECV_REJECT:
		rej_msg = (auth_rjt_msg_t *)auth->response;

		bfa_trc(auth->fabric->fcs, rej_msg->reason_code);
		bfa_trc(auth->fabric->fcs, rej_msg->reason_code_exp);

		if ((rej_msg->reason_code == FC_AUTH_RJT_CODE_LOGICAL_ERR) &&
		    (rej_msg->reason_code_exp == FC_AUTH_CEXP_RESTART_AUTH_PROTO)) {
			bfa_fcs_auth_restart(auth->fabric->fcs);
		}

		break;

	default:
		bfa_assert(0);
		break;
	}
}

/*
 *****************************************************************************
 * Externalized BFA FCS functions
 *****************************************************************************
 */
/**
 * @brief get attributes
 * @param[in] port   - port
 * @param[in] policy - enable/disable
 *
 * @return : BFA_STATUS_OK if successful, else BFA_STATUS_FAILED for failure
 */
bfa_status_t
bfa_fcs_auth_get_attr(bfa_fcs_t *port, bfa_auth_attr_t *attr)
{
	bfa_os_memset(attr, 0, sizeof(bfa_auth_attr_t));

	if (port->fabric.auth.algo >= BFA_AUTH_ALGO_MD5 &&
	    port->fabric.auth.algo <= BFA_AUTH_ALGO_SM)
		attr->algo = port->fabric.auth.algo;
	else
		attr->algo = port->fabric.auth.algo = BFA_AUTH_ALGO_MD5;

	if (port->fabric.auth.group == FC_AUTH_DH_GID_0_DHG_NULL) {
		attr->dh_grp = BFA_AUTH_GROUP_DHNULL;
	}
	else if (port->fabric.auth.group == FC_AUTH_DH_GID_1_DHG_1024) {
		attr->dh_grp = BFA_AUTH_GROUP_DH1024;
	}
	else if (port->fabric.auth.group == FC_AUTH_DH_GID_2_DHG_1280) {
		attr->dh_grp = BFA_AUTH_GROUP_DH1280;
	}
	else if (port->fabric.auth.group == FC_AUTH_DH_GID_3_DHG_1536) {
		attr->dh_grp = BFA_AUTH_GROUP_DH1536;
	}
	else { /* Use the default value as we support only the DHNULL group */
		attr->dh_grp = BFA_AUTH_GROUP_DHNULL;
	}

	if (port->fabric.auth.policy == BFA_FALSE)
		attr->status = BFA_AUTH_STATUS_NONE;
	else
		attr->status = bfa_sm_to_state(auth_sm_table, port->fabric.auth.sm);
	if (port->fabric.auth.secret_len > 0)
		attr->secret_set = BFA_TRUE;
	else 
		attr->secret_set = BFA_FALSE;
	attr->rjt_code = port->fabric.auth.rjt_code;
	attr->rjt_code_exp = port->fabric.auth.rjt_code_exp;

	return BFA_STATUS_OK;
}

/**
 * @brief enable/disable authentication
 * @param[in] port 	 - port
 * @param[in] policy - enable/disable
 *
 * @return : BFA_STATUS_OK
 */
bfa_status_t
bfa_fcs_auth_set_policy(bfa_fcs_t *port, bfa_boolean_t policy)
{
	bfa_fcs_auth_t*		auth = &port->fabric.auth;
	bfa_pport_attr_t 	attr;
	int					state;

	auth->fabric = &port->fabric;

	if (auth->algo < BFA_AUTH_ALGO_MD5 || auth->algo > BFA_AUTH_ALGO_SM)
		auth->algo = BFA_AUTH_ALGO_MD5; 

	/* Initialize the state machine, if not already initialized */
	state = bfa_sm_to_state(auth_sm_table, auth->sm);
	if (state < BFA_AUTH_UNINIT || state > BFA_AUTH_FAILED) {
		bfa_sm_set_state(auth, bfa_fcs_auth_sm_uninit);
	}

	/* 
	 * If the link is already up and authentication is enabled now,
	 * and authentication is not already successful,
	 * then restart authentication process
	 */
	if (auth->policy == BFA_FALSE && policy == BFA_TRUE &&
		(auth->status != FC_AUTH_STATE_SUCCESS)) {
		bfa_hal_pport_get_attr(port->hal, &attr);
		if ( attr.port_state == BFA_HAL_PPORT_LINKUP)
			bfa_fcs_auth_restart(port);
	}

	auth->policy = policy;
	port->fabric.auth_reqd = policy;
	bfa_fcs_auth_aen_audit_post(auth, (policy == BFA_TRUE) ? BFA_AUDIT_AEN_AUTH_ENABLE : BFA_AUDIT_AEN_AUTH_DISABLE);

	return BFA_STATUS_OK;
}

/**
 * @brief get the current status of authentication
 * @param[in] port 	 - port
 * @param[in] policy - enable/disable
 *
 * @return : bfa_auth_status_t
 */
bfa_auth_status_t
bfa_fcs_auth_get_status(bfa_fcs_t *port)
{
	if (port->fabric.auth.policy == BFA_FALSE)
		return BFA_AUTH_STATUS_NONE;
	else
		return bfa_sm_to_state(auth_sm_table, port->fabric.auth.sm);

}

/**
 * @brief set the authentication algorithm
 * @param[in] port - port
 * @param[in] algo - authentication algorithm
 *
 * @return : BFA_STATUS_OK if successful, else BFA_STATUS_FAILED
 */
bfa_status_t
bfa_fcs_auth_set_algo(bfa_fcs_t *port, bfa_auth_algo_t algo)
{
	bfa_fcs_auth_t	*auth = &port->fabric.auth;

	auth->algo = algo;

	return BFA_STATUS_OK;
	
}

/**
 * @brief get statistics
 * @param[in] 	port  - port
 * @param[out] 	stats - pointer to the authentication statistics of the port
 *
 * @return : BFA_STATUS_OK
 */
bfa_status_t
bfa_fcs_auth_get_stats(bfa_fcs_t *port, bfa_auth_stats_t *stats)
{
	bfa_fcs_auth_t	*auth = &port->fabric.auth;

	memcpy(stats, &auth->stats, sizeof (bfa_auth_stats_t));

	return BFA_STATUS_OK;
}

/**
 * @brief Get DH Group
 * @param[in] port  - port
 * @param[in] group - DH Group
 *
 * @return : BFA_TRUE if successful else BFA_FALSE
 */
bfa_status_t
bfa_fcs_auth_set_dh_group(bfa_fcs_t *port, int group)
{
	if (group == FC_AUTH_DH_GID_0_DHG_NULL) {
		port->fabric.auth.group = group;
		return BFA_STATUS_OK;
	}

	return BFA_STATUS_FAILED;
}

/**
 * @brief Set secret string
 * @param[in] port  - port
 * @param[in] secret- secret string
 *
* @return : BFA_TRUE
 */
bfa_status_t
bfa_fcs_auth_set_secretstring(bfa_fcs_t *port, char *secret)
{
	strncpy(port->fabric.auth.secret, secret, strlen(secret));
	port->fabric.auth.secret[strlen(secret)] = '\0';

	return BFA_STATUS_OK;
}

/**
 * @brief Set encrypted secret string
 * @param[in] port  - port
 * @param[in] secret- secret string
 *
* @return : BFA_TRUE
 */
bfa_status_t
bfa_fcs_auth_set_secretstring_encrypt(bfa_fcs_t *port, uint32_t secret[],
	uint32_t len)
{
	bfad_auth_decrypt_string (port->fabric.auth.secret, secret, len);
	port->fabric.auth.secret[len] = '\0';
	port->fabric.auth.secret_len = len & 0xff;
	return BFA_STATUS_OK;
}

/**
 * @brief Set secret source
 * @param[in] port  - port
 * @param[in] src   - secret source
 *
 * @return : BFA_STATUS_OK if successful, otherwise BFA_STATUS_FAILED
 */
bfa_status_t
bfa_fcs_auth_set_secretsource(bfa_fcs_t *port, bfa_auth_secretsource_t src)
{
    bfa_fcs_auth_t  *auth = &port->fabric.auth;

	if (src == BFA_AUTH_SECSRC_LOCAL) {
		auth->source = src;
		return BFA_STATUS_OK;
	}

	return BFA_STATUS_FAILED;
}

/**
 * @brief reset statistics
 * @param[in] port  - port
 *
 * @return : BFA_STATUS_OK
 */
bfa_status_t
bfa_fcs_auth_reset_stats(bfa_fcs_t *port)
{
	bfa_fcs_auth_t	*auth = &port->fabric.auth;

	memset(&auth->stats, 0x00, sizeof (bfa_auth_stats_t));

	return BFA_STATUS_OK;
}

/**
 * @brief re-init authentication
 * @param[in] port  - port
 *
 * @return : BFA_STATUS_FAILED
 */
bfa_status_t
bfa_fcs_auth_reinit(bfa_fcs_t *port)
{
	return BFA_STATUS_FAILED;
}

/**
 * @brief restart authentication
 * @param[in] port  - port
 *
 * @return : BFA_STATUS_OK
 */
void
bfa_fcs_auth_restart(bfa_fcs_t *port)
{
	bfa_fcs_auth_start(&port->fabric);
}

/*
 *****************************************************************************
 * Externalized BFA private functions
 *****************************************************************************
 */

/*
 * @brief function called to fire up authentication
 * @param[in] fabric - fabric
 * @return - none
 */
void
bfa_fcs_auth_start(bfa_fcs_fabric_t *fabric)
{
	bfa_fcs_auth_t*		auth = &fabric->auth;

	bfa_trc(fabric->fcs, BFA_FCS_AUTH_SM_START);

	auth->fabric = fabric;
	bfa_fcs_auth_sm_uninit(auth, BFA_FCS_AUTH_SM_START);
}

void
bfa_fcs_auth_stop(bfa_fcs_fabric_t *fabric)
{
	bfa_fcs_auth_t*		auth = &fabric->auth;
	int					state;

	bfa_trc(fabric->fcs, BFA_FCS_AUTH_SM_STOP);

	state = bfa_sm_to_state(auth_sm_table, fabric->auth.sm);
    	if (state >= BFA_AUTH_UNINIT && state <= BFA_AUTH_FAILED) {
	    bfa_sm_send_event(auth, BFA_FCS_AUTH_SM_STOP);
	}
}

/* vim:set ts=4 sw=4 noet ff=unix ft=c tw=78 ai: */

/*
 * emacs.Local Variables:
 * emacs.mode:C
 * emacs.c-basic-offset:4
 * emacs.indent-tabs-mode:t
 * emacs.tab-width:4
 * emacs.End:
 */
