/*
 * Copyright (c)  2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

/**
 * @file bfa_fcs.c BFA FCS main
 */

#include <fcs/bfa_fcs.h>
#include "fcs_pport.h"
#include "fcs_uf.h"
#include "fcs_vport.h"
#include "fcs_rport.h"
#include "fcs_fabric.h"
#include "fcs_fcpim.h"
#include "fcs_fcptm.h"
#include "fcbuild.h"
#include <fcb/bfa_fcb.h>

/**
 * FCS sub-modules
 */
typedef struct {
	void            (*modinit) (bfa_fcs_t *fcs);
	void            (*modexit) (bfa_fcs_t *fcs);
} bfa_fcs_mod_t;

#define BFA_FCS_MODULE(_mod)	{ _mod ## _modinit, _mod ## _modexit }

static bfa_fcs_mod_t fcs_modules[] = {
	BFA_FCS_MODULE(bfa_fcs_pport),
	BFA_FCS_MODULE(bfa_fcs_uf),
	BFA_FCS_MODULE(bfa_fcs_fabric),
	BFA_FCS_MODULE(bfa_fcs_vport),
	BFA_FCS_MODULE(bfa_fcs_rport),
	BFA_FCS_MODULE(bfa_fcs_fcpim),
};

/**
 * @dg fcs_api BFA FCS API
 * @{
 */

static void
bfa_fcs_exit_comp(void *fcs_cbarg)
{
	bfa_fcs_t      *fcs = fcs_cbarg;

	bfa_fcb_exit(fcs->bfad);
}

/**
 * @}
 */

/**
 * @dg fcs_api BFA FCS API
 * @{
 */

/**
 * 	@brief
 * 		FCS instance initialization.
 *
 * 	@param[in]		fcs		FCS instance
 * 	@param[in]		hal		HAL instance
 * 	@param[in]		bfad		BFA driver instance
 *
 * 	@return None
 */
void
bfa_fcs_init(bfa_fcs_t *fcs, bfa_hal_t *hal, bfad_t *bfad)
{
	int             i;
	bfa_fcs_mod_t  *mod;

	fcs->hal = hal;
	fcs->bfad = bfad;

	bfa_hal_fcs_attach(hal);
	fcbuild_init();

	for (i = 0; i < sizeof(fcs_modules) / sizeof(fcs_modules[0]); i++) {
		mod = &fcs_modules[i];
		mod->modinit(fcs);
	}
}

/**
 * 	@brief
 * 		FCS driver details initialization.
 *
 * 	@param[in]		fcs		FCS instance
 * 	@param[in]		driver_info	Driver Details
 *
 * 	@return None
 */
void
bfa_fcs_driver_info_init(bfa_fcs_t *fcs, bfa_fcs_driver_info_t *driver_info)
{

	fcs->driver_info = *driver_info;

	bfa_fcs_fabric_psymb_init(&fcs->fabric);
}

/**
 * 	@brief
 * 		FCS instance cleanup and exit.
 *
 * 	@param[in]		fcs			FCS instance
 * 	@return None
 */
void
bfa_fcs_exit(bfa_fcs_t *fcs)
{
	bfa_fcs_mod_t  *mod;
	int             nmods, i;

	bfa_wc_init(&fcs->wc, bfa_fcs_exit_comp, fcs);

	nmods = sizeof(fcs_modules) / sizeof(fcs_modules[0]);

	for (i = 0; i < nmods; i++) {
		bfa_wc_up(&fcs->wc);

		mod = &fcs_modules[i];
		mod->modexit(fcs);
	}

	bfa_wc_wait(&fcs->wc);
}

/**
 * @brief Initialize FCS tracing.
 */
void
bfa_fcs_trc_init(bfa_fcs_t *fcs, bfa_trc_mod_t *trcmod)
{
	fcs->trcmod = trcmod;
}

/**
 * @brief Initialize FCS logging.
 */
void
bfa_fcs_log_init(bfa_fcs_t *fcs, bfa_log_mod_t *logmod)
{
	fcs->logm = logmod;
}

/**
 * @brief Initialize AEN in BFA HAL module.
 */
void
bfa_fcs_aen_init(bfa_fcs_t *fcs, bfa_aen_t *aen)
{
	fcs->aen = aen;
}

void
bfa_fcs_modexit_comp(bfa_fcs_t *fcs)
{
	bfa_wc_down(&fcs->wc);
}

/**
 * @}
 */
