/*
 * Copyright (c)  2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

/**
 * @file bfa_fcs_port.c BFA FCS port
 */

#include <fcs/bfa_fcs.h>
#include <fcs/bfa_fcs_port.h>
#include <fcb/bfa_fcb_port.h>
#include <fcs/bfa_fcs_rport.h>
#include <hal/bfa_hal_pport.h>
#include <log/bfa_log_fcs.h>
#include <cs/bfa_stats.h>
#include "fcs.h"
#include "fcs_port.h"
#include "fcs_vport.h"
#include "fcs_rport.h"
#include "fcs_fcxp.h"
#include "fcs_trcmod.h"
#include "port_priv.h"
#include <aen/bfal_aen_lport.h>

BFA_TRC_FILE(FCS, PORT);

/**
 * Forward declarations
 */

static void     bfa_fcs_port_aen_post(bfa_fcs_port_t *port,
				      bfa_lport_aen_event_t event);
static void     bfa_fcs_port_send_ls_rjt(bfa_fcs_port_t *port,
					 fchs_t *rx_fchs, uint8_t reason_code,
					 uint8_t reason_code_expl);
static void     bfa_fcs_port_plogi(bfa_fcs_port_t *port, fchs_t *rx_fchs,
				   fc_plogi_t *plogi);
static void     bfa_fcs_port_online_actions(bfa_fcs_port_t *port);
static void     bfa_fcs_port_offline_actions(bfa_fcs_port_t *port);
static void     bfa_fcs_port_unknown_init(bfa_fcs_port_t *port);
static void     bfa_fcs_port_unknown_online(bfa_fcs_port_t *port);
static void     bfa_fcs_port_unknown_offline(bfa_fcs_port_t *port);
static void     bfa_fcs_port_deleted(bfa_fcs_port_t *port);
static void     bfa_fcs_port_echo(bfa_fcs_port_t *port, fchs_t *rx_fchs,
				  fc_echo_t *echo, uint16_t len);
static void     bfa_fcs_port_rnid(bfa_fcs_port_t *port, fchs_t *rx_fchs,
				  fc_rnid_cmd_t *rnid, uint16_t len);
static void     bfa_fs_port_get_gen_topo_data(bfa_fcs_port_t *port,
					      fc_rnid_general_topology_data_t *
					      gen_topo_data);

struct {
	void            (*init) (bfa_fcs_port_t *port);
	void            (*online) (bfa_fcs_port_t *port);
	void            (*offline) (bfa_fcs_port_t *port);
} __port_action[] = {
	{
	bfa_fcs_port_unknown_init, bfa_fcs_port_unknown_online,
			bfa_fcs_port_unknown_offline}, {
	bfa_fcs_port_fab_init, bfa_fcs_port_fab_online,
			bfa_fcs_port_fab_offline}, {
	bfa_fcs_port_loop_init, bfa_fcs_port_loop_online,
			bfa_fcs_port_loop_offline}, {
	bfa_fcs_port_n2n_init, bfa_fcs_port_n2n_online,
			bfa_fcs_port_n2n_offline},};

/**
 * @dg fcs_port_sm FCS logical port state machine
 * @{
 */

typedef enum {
	BFA_FCS_PORT_SM_CREATE = 1,
	BFA_FCS_PORT_SM_ONLINE = 2,
	BFA_FCS_PORT_SM_OFFLINE = 3,
	BFA_FCS_PORT_SM_DELETE = 4,
	BFA_FCS_PORT_SM_DELRPORT = 5,
} bfa_fcs_port_event_t;

static void     bfa_fcs_port_sm_uninit(bfa_fcs_port_t *port,
				       bfa_fcs_port_event_t event);
static void     bfa_fcs_port_sm_init(bfa_fcs_port_t *port,
				     bfa_fcs_port_event_t event);
static void     bfa_fcs_port_sm_online(bfa_fcs_port_t *port,
				       bfa_fcs_port_event_t event);
static void     bfa_fcs_port_sm_offline(bfa_fcs_port_t *port,
					bfa_fcs_port_event_t event);
static void     bfa_fcs_port_sm_deleting(bfa_fcs_port_t *port,
					 bfa_fcs_port_event_t event);

static void
bfa_fcs_port_sm_uninit(bfa_fcs_port_t *port, bfa_fcs_port_event_t event)
{
	bfa_trc(port->fcs, port->port_cfg.pwwn);
	bfa_trc(port->fcs, event);

	switch (event) {
	case BFA_FCS_PORT_SM_CREATE:
		bfa_sm_set_state(port, bfa_fcs_port_sm_init);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_fcs_port_sm_init(bfa_fcs_port_t *port, bfa_fcs_port_event_t event)
{
	bfa_trc(port->fcs, port->port_cfg.pwwn);
	bfa_trc(port->fcs, event);

	switch (event) {
	case BFA_FCS_PORT_SM_ONLINE:
		bfa_sm_set_state(port, bfa_fcs_port_sm_online);
		bfa_fcs_port_online_actions(port);
		break;

	case BFA_FCS_PORT_SM_DELETE:
		bfa_sm_set_state(port, bfa_fcs_port_sm_uninit);
		bfa_fcs_port_deleted(port);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_fcs_port_sm_online(bfa_fcs_port_t *port, bfa_fcs_port_event_t event)
{
	bfa_fcs_rport_t *rport;
	bfa_q_t        *qe, *qen;

	bfa_trc(port->fcs, port->port_cfg.pwwn);
	bfa_trc(port->fcs, event);

	switch (event) {
	case BFA_FCS_PORT_SM_OFFLINE:
		bfa_sm_set_state(port, bfa_fcs_port_sm_offline);
		bfa_fcs_port_offline_actions(port);
		break;

	case BFA_FCS_PORT_SM_DELETE:

		__port_action[port->fabric->fab_type].offline(port);

		if (port->num_rports == 0) {
			bfa_sm_set_state(port, bfa_fcs_port_sm_uninit);
			bfa_fcs_port_deleted(port);
		} else {
			bfa_sm_set_state(port, bfa_fcs_port_sm_deleting);
			bfa_q_iter_safe(&port->rport_q, qe, qen) {
				rport = (bfa_fcs_rport_t *) qe;
				bfa_fcs_rport_delete(rport);
			}
		}
		break;

	case BFA_FCS_PORT_SM_DELRPORT:
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_fcs_port_sm_offline(bfa_fcs_port_t *port, bfa_fcs_port_event_t event)
{
	bfa_fcs_rport_t *rport;
	bfa_q_t        *qe, *qen;

	bfa_trc(port->fcs, port->port_cfg.pwwn);
	bfa_trc(port->fcs, event);

	switch (event) {
	case BFA_FCS_PORT_SM_ONLINE:
		bfa_sm_set_state(port, bfa_fcs_port_sm_online);
		bfa_fcs_port_online_actions(port);
		break;

	case BFA_FCS_PORT_SM_DELETE:
		if (port->num_rports == 0) {
			bfa_sm_set_state(port, bfa_fcs_port_sm_uninit);
			bfa_fcs_port_deleted(port);
		} else {
			bfa_sm_set_state(port, bfa_fcs_port_sm_deleting);
			bfa_q_iter_safe(&port->rport_q, qe, qen) {
				rport = (bfa_fcs_rport_t *) qe;
				bfa_fcs_rport_delete(rport);
			}
		}
		break;

	case BFA_FCS_PORT_SM_DELRPORT:
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_fcs_port_sm_deleting(bfa_fcs_port_t *port, bfa_fcs_port_event_t event)
{
	bfa_trc(port->fcs, port->port_cfg.pwwn);
	bfa_trc(port->fcs, event);

	switch (event) {
	case BFA_FCS_PORT_SM_DELRPORT:
		if (port->num_rports == 0) {
			bfa_sm_set_state(port, bfa_fcs_port_sm_uninit);
			bfa_fcs_port_deleted(port);
		}
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @}
 */

/**
 * @dg fcs_port_pvt
 * @{
 */

/**
 * Send AEN notification
 */
static void
bfa_fcs_port_aen_post(bfa_fcs_port_t *port, bfa_lport_aen_event_t event)
{
	bfa_aen_data_t  aen_data;
	bfa_log_mod_t	*logmod = port->fcs->logm;
	bfa_port_role_t	role = port->port_cfg.roles;
	wwn_t		lpwwn = bfa_fcs_port_get_pwwn(port);
	char		lpwwn_buf[BFA_STRING_32];
	char		*lpwwn_ptr;
	char		*role_str[BFA_PORT_ROLE_FCP_MAX/2 + 1] = 
				{"Initiator", "Target", "IPFC"};

        lpwwn_ptr = wwn2str(lpwwn_buf, sizeof(lpwwn_buf), lpwwn);

	bfa_assert(role <= BFA_PORT_ROLE_FCP_MAX);

	switch (event) {
	case BFA_LPORT_AEN_ONLINE:
		bfa_log(logmod, BFAL_AEN_LPORT_ONLINE, lpwwn_ptr, 
				role_str[role/2]);
		break;
	case BFA_LPORT_AEN_OFFLINE:
		bfa_log(logmod, BFAL_AEN_LPORT_OFFLINE, lpwwn_ptr, 
				role_str[role/2]);
		break;
	case BFA_LPORT_AEN_NEW:
		bfa_log(logmod, BFAL_AEN_LPORT_NEW, lpwwn_ptr, 
				role_str[role/2]);
		break;
	case BFA_LPORT_AEN_DELETE:
		bfa_log(logmod, BFAL_AEN_LPORT_DELETE, lpwwn_ptr, 
				role_str[role/2]);
		break;
	case BFA_LPORT_AEN_DISCONNECT:
		bfa_log(logmod, BFAL_AEN_LPORT_DISCONNECT, lpwwn_ptr, 
				role_str[role/2]);
		break;
	default: 
		break;
	}

	aen_data.lport.vf_id = port->fabric->vf_id;
	aen_data.lport.roles = role;
	aen_data.lport.lpwwn = lpwwn;
	bfa_aen_post(port->fcs->aen, BFA_AEN_CAT_LPORT, event, &aen_data);
}

/*
 * Send a LS reject
 */
static void
bfa_fcs_port_send_ls_rjt(bfa_fcs_port_t *port, fchs_t *rx_fchs,
			 uint8_t reason_code, uint8_t reason_code_expl)
{
	fchs_t          fchs;
	bfa_hal_fcxp_t *fcxp;
	bfa_hal_rport_t *hal_rport = NULL;
	int             len;

	bfa_trc(port->fcs, rx_fchs->s_id);

	fcxp = bfa_fcs_fcxp_alloc(port->fcs);
	if (!fcxp)
		return;

	len = fc_ls_rjt_build(&fchs, bfa_hal_fcxp_get_reqbuf(fcxp),
			      rx_fchs->s_id, bfa_fcs_port_get_fcid(port),
			      rx_fchs->ox_id, reason_code, reason_code_expl);

	bfa_hal_fcxp_send(fcxp, hal_rport, port->fabric->vf_id, BFA_FALSE,
			  FC_CLASS_3, len, &fchs, NULL, NULL, FC_MAX_PDUSZ, 0);
}

/**
 * Process incoming plogi from a remote port.
 */
static void
bfa_fcs_port_plogi(bfa_fcs_port_t *port, fchs_t *rx_fchs, fc_plogi_t *plogi)
{
	bfa_fcs_rport_t *rport;

	bfa_trc(port->fcs, rx_fchs->d_id);
	bfa_trc(port->fcs, rx_fchs->s_id);

	if (fc_plogi_parse(rx_fchs) != FC_PARSE_OK) {
		bfa_trc(port->fcs, rx_fchs->s_id);
		/*
		 * send a LS reject
		 */
		bfa_fcs_port_send_ls_rjt(port, rx_fchs,
					 FC_LS_RJT_RSN_PROTOCOL_ERROR,
					 FC_LS_RJT_EXP_SPARMS_ERR_OPTIONS);
		return;
	}

	/**
	 * @todo Non-switched P2P mode to handle local port address assignment.
	 */

	/**
	 * First, check if we know the device by pwwn.
	 */
	rport = bfa_fcs_port_get_rport_by_pwwn(port, plogi->port_name);
	if (rport) {
		bfa_fcs_rport_plogi(rport, rx_fchs, plogi);
		return;
	}

	/**
	 * Next, lookup rport by PID.
	 */
	rport = bfa_fcs_port_get_rport_by_pid(port, rx_fchs->s_id);
	if (!rport) {
		/**
		 * Inbound PLOGI from a new device.
		 */
		bfa_fcs_rport_plogi_create(port, rx_fchs, plogi);
		return;
	}

	/**
	 * Rport is known only by PID.
	 */
	if (rport->pwwn) {
		/**
		 * This is a different device with the same pid. Old device
		 * disappeared. Send implicit LOGO to old device.
		 */
		bfa_assert(rport->pwwn != plogi->port_name);
		bfa_fcs_rport_logo_imp(rport);

		/**
		 * Inbound PLOGI from a new device (with old PID).
		 */
		bfa_fcs_rport_plogi_create(port, rx_fchs, plogi);
		return;
	}

	/**
	 * PLOGI crossing each other.
	 */
	bfa_assert(rport->pwwn == WWN_NULL);
	bfa_fcs_rport_plogi(rport, rx_fchs, plogi);
}

/*
 * Process incoming ECHO.
 * Since it does not require a login, it is processed here.
 */
static void
bfa_fcs_port_echo(bfa_fcs_port_t *port, fchs_t *rx_fchs, fc_echo_t *echo,
		  uint16_t rx_len)
{
	fchs_t          fchs;
	bfa_hal_fcxp_t *fcxp;
	bfa_hal_rport_t *hal_rport = NULL;
	int             len, pyld_len;

	bfa_trc(port->fcs, rx_fchs->s_id);
	bfa_trc(port->fcs, rx_fchs->d_id);
	bfa_trc(port->fcs, rx_len);

	fcxp = bfa_fcs_fcxp_alloc(port->fcs);
	if (!fcxp)
		return;

	len = fc_ls_acc_build(&fchs, bfa_hal_fcxp_get_reqbuf(fcxp),
			      rx_fchs->s_id, bfa_fcs_port_get_fcid(port),
			      rx_fchs->ox_id);

	/*
	 * Copy the payload (if any) from the echo frame
	 */
	pyld_len = rx_len - sizeof(fchs_t);
	bfa_trc(port->fcs, pyld_len);

	if (pyld_len > len)
		memcpy(((uint8_t *) bfa_hal_fcxp_get_reqbuf(fcxp)) +
		       sizeof(fc_echo_t), (echo + 1),
		       (pyld_len - sizeof(fc_echo_t)));

	bfa_hal_fcxp_send(fcxp, hal_rport, port->fabric->vf_id, BFA_FALSE,
			  FC_CLASS_3, pyld_len, &fchs, NULL, NULL, FC_MAX_PDUSZ,
			  0);
}

/*
 * Process incoming RNID.
 * Since it does not require a login, it is processed here.
 */
static void
bfa_fcs_port_rnid(bfa_fcs_port_t *port, fchs_t *rx_fchs, fc_rnid_cmd_t *rnid,
		  uint16_t rx_len)
{
	fc_rnid_common_id_data_t common_id_data;
	fc_rnid_general_topology_data_t gen_topo_data;
	fchs_t          fchs;
	bfa_hal_fcxp_t *fcxp;
	bfa_hal_rport_t *hal_rport = NULL;
	uint16_t        len;
	uint32_t        data_format;

	bfa_trc(port->fcs, rx_fchs->s_id);
	bfa_trc(port->fcs, rx_fchs->d_id);
	bfa_trc(port->fcs, rx_len);

	fcxp = bfa_fcs_fcxp_alloc(port->fcs);
	if (!fcxp)
		return;

	/*
	 * Check Node Indentification Data Format
	 * We only support General Topology Discovery Format.
	 * For any other requested Data Formats, we return Common Node Id Data
	 * only, as per FC-LS.
	 */
	bfa_trc(port->fcs, rnid->node_id_data_format);
	if (rnid->node_id_data_format == RNID_NODEID_DATA_FORMAT_DISCOVERY) {
		data_format = RNID_NODEID_DATA_FORMAT_DISCOVERY;
		/*
		 * Get General topology data for this port
		 */
		bfa_fs_port_get_gen_topo_data(port, &gen_topo_data);
	} else {
		data_format = RNID_NODEID_DATA_FORMAT_COMMON;
	}

	/*
	 * Copy the Node Id Info
	 */
	common_id_data.port_name = bfa_fcs_port_get_pwwn(port);
	common_id_data.node_name = bfa_fcs_port_get_nwwn(port);

	len = fc_rnid_acc_build(&fchs, bfa_hal_fcxp_get_reqbuf(fcxp),
				rx_fchs->s_id, bfa_fcs_port_get_fcid(port),
				rx_fchs->ox_id, data_format, &common_id_data,
				&gen_topo_data);

	bfa_hal_fcxp_send(fcxp, hal_rport, port->fabric->vf_id, BFA_FALSE,
			  FC_CLASS_3, len, &fchs, NULL, NULL, FC_MAX_PDUSZ, 0);

	return;
}

/*
 *  Fill out General Topolpgy Discovery Data for RNID ELS.
 */
static void
bfa_fs_port_get_gen_topo_data(bfa_fcs_port_t *port,
			      fc_rnid_general_topology_data_t *gen_topo_data)
{

	bfa_os_memset(gen_topo_data, 0,
		      sizeof(fc_rnid_general_topology_data_t));

	gen_topo_data->asso_type = bfa_os_htonl(RNID_ASSOCIATED_TYPE_HOST);
	gen_topo_data->phy_port_num = 0;	/* @todo */
	gen_topo_data->num_attached_nodes = bfa_os_htonl(1);
}

static void
bfa_fcs_port_online_actions(bfa_fcs_port_t *port)
{
	bfa_trc(port->fcs, port->fabric->oper_type);

	__port_action[port->fabric->fab_type].init(port);
	__port_action[port->fabric->fab_type].online(port);

	bfa_fcs_port_aen_post(port, BFA_LPORT_AEN_ONLINE);
	bfa_fcb_port_online(port->fcs->bfad, port->port_cfg.roles,
			    port->fabric->vf_drv,
			    (port->vport == NULL) ?
				 NULL : port->vport->vport_drv);
}

static void
bfa_fcs_port_offline_actions(bfa_fcs_port_t *port)
{
	bfa_q_t        *qe, *qen;
	bfa_fcs_rport_t *rport;

	bfa_trc(port->fcs, port->fabric->oper_type);

	__port_action[port->fabric->fab_type].offline(port);

	if (bfa_fcs_port_is_online(bfa_fcs_get_base_port(port->fcs)) 
			== BFA_TRUE) {
		bfa_fcs_port_aen_post(port, BFA_LPORT_AEN_DISCONNECT);
	} else {
		bfa_fcs_port_aen_post(port, BFA_LPORT_AEN_OFFLINE);
	}
	bfa_fcb_port_offline(port->fcs->bfad, port->port_cfg.roles,
			     port->fabric->vf_drv,
			     (port->vport == NULL) ?
				 NULL : port->vport->vport_drv);

	bfa_q_iter_safe(&port->rport_q, qe, qen) {
		rport = (bfa_fcs_rport_t *) qe;
		bfa_fcs_rport_offline(rport);
	}
}

static void
bfa_fcs_port_unknown_init(bfa_fcs_port_t *port)
{
	bfa_assert(0);
}

static void
bfa_fcs_port_unknown_online(bfa_fcs_port_t *port)
{
	bfa_assert(0);
}

static void
bfa_fcs_port_unknown_offline(bfa_fcs_port_t *port)
{
	bfa_assert(0);
}

static void
bfa_fcs_port_deleted(bfa_fcs_port_t *port)
{
	bfa_fcs_port_aen_post(port, BFA_LPORT_AEN_DELETE);

	/* Base port will be deleted by the OS driver */
	if (port->vport) {
		bfa_fcb_port_delete(port->fcs->bfad, port->port_cfg.roles,
				port->fabric->vf_drv,
				port->vport ? port->vport->vport_drv : NULL);
		bfa_fcs_vport_delete_comp(port->vport);
	}
	else {
		 bfa_fcs_fabric_port_delete_comp(port->fabric);
	}
}

/**
 * @}
 */

/**
 * @dg fcs_lport_api BFA FCS port API
 * @{
 */
/**
 * @brief
 *   Module initialization
 */
void
bfa_fcs_port_modinit(bfa_fcs_t *fcs)
{

}

/**
 * @brief
 *   Module cleanup
 */
void
bfa_fcs_port_modexit(bfa_fcs_t *fcs)
{
	bfa_fcs_modexit_comp(fcs);
}

/**
 * @brief
 * 		Unsolicited frame receive handling.
 */
void
bfa_fcs_port_uf_recv(bfa_fcs_port_t *lport, fchs_t *fchs, uint16_t len)
{
	uint32_t        pid = fchs->s_id;
	bfa_fcs_rport_t *rport = NULL;
	fc_els_cmd_t   *els_cmd = (fc_els_cmd_t *) (fchs + 1);

	bfa_stats(lport, uf_recvs);

	if (!bfa_fcs_port_is_online(lport)) {
		bfa_stats(lport, uf_recv_drops);
		return;
	}

	/**
	 * First, handle ELSs that donot require a login.
	 */
	/*
	 * Handle PLOGI first
	 */
	if ((fchs->type == FC_TYPE_ELS) &&
		(els_cmd->els_code == FC_ELS_PLOGI)) {
		bfa_fcs_port_plogi(lport, fchs, (fc_plogi_t *) els_cmd);
		return;
	}

	/*
	 * Handle ECHO separately.
	 */
	if ((fchs->type == FC_TYPE_ELS) && (els_cmd->els_code == FC_ELS_ECHO)) {
		bfa_fcs_port_echo(lport, fchs, (fc_echo_t *) els_cmd, len);
		return;
	}

	/*
	 * Handle RNID separately.
	 */
	if ((fchs->type == FC_TYPE_ELS) && (els_cmd->els_code == FC_ELS_RNID)) {
		bfa_fcs_port_rnid(lport, fchs, (fc_rnid_cmd_t *) els_cmd, len);
		return;
	}

	/**
	 * look for a matching remote port ID
	 */
	rport = bfa_fcs_port_get_rport_by_pid(lport, pid);
	if (rport) {
		bfa_trc(rport->fcs, fchs->s_id);
		bfa_trc(rport->fcs, fchs->d_id);
		bfa_trc(rport->fcs, fchs->type);

		bfa_fcs_rport_uf_recv(rport, fchs, len);
		return;
	}

	/**
	 * Only handles ELS frames for now.
	 */
	if (fchs->type != FC_TYPE_ELS) {
		bfa_trc(lport->fcs, fchs->type);
		bfa_assert(0);
		return;
	}

	bfa_trc(lport->fcs, els_cmd->els_code);
	if (els_cmd->els_code == FC_ELS_RSCN) {
		bfa_fcs_port_scn_process_rscn(lport, fchs, len);
		return;
	}

	if (els_cmd->els_code == FC_ELS_LOGO) {
		/**
		 * @todo Handle LOGO frames received.
		 */
		bfa_trc(lport->fcs, els_cmd->els_code);
		return;
	}

	if (els_cmd->els_code == FC_ELS_PRLI) {
		/**
		 * @todo Handle PRLI frames received.
		 */
		bfa_trc(lport->fcs, els_cmd->els_code);
		return;
	}

	/**
	 * Unhandled ELS frames. Send a LS_RJT.
	 */
	bfa_fcs_port_send_ls_rjt(lport, fchs, FC_LS_RJT_RSN_CMD_NOT_SUPP,
				 FC_LS_RJT_EXP_NO_ADDL_INFO);

}

/**
 * @brief
 *   PID based Lookup for a R-Port in the Port R-Port Queue
 */
bfa_fcs_rport_t *
bfa_fcs_port_get_rport_by_pid(bfa_fcs_port_t *port, uint32_t pid)
{
	bfa_fcs_rport_t *rport;
	bfa_q_t        *qe;

	bfa_q_iter(&port->rport_q, qe) {
		rport = (bfa_fcs_rport_t *) qe;
		if (rport->pid == pid)
			return rport;
	}

	bfa_trc(port->fcs, pid);
	return NULL;
}

/**
 * @brief
 *   PWWN based Lookup for a R-Port in the Port R-Port Queue
 */
bfa_fcs_rport_t *
bfa_fcs_port_get_rport_by_pwwn(bfa_fcs_port_t *port, wwn_t pwwn)
{
	bfa_fcs_rport_t *rport;
	bfa_q_t        *qe;

	bfa_q_iter(&port->rport_q, qe) {
		rport = (bfa_fcs_rport_t *) qe;
		if (wwn_is_equal(rport->pwwn, pwwn))
			return rport;
	}

	bfa_trc(port->fcs, pwwn);
	return (NULL);
}

/**
 * @brief
 *   NWWN based Lookup for a R-Port in the Port R-Port Queue
 */
bfa_fcs_rport_t *
bfa_fcs_port_get_rport_by_nwwn(bfa_fcs_port_t *port, wwn_t nwwn)
{
	bfa_fcs_rport_t *rport;
	bfa_q_t        *qe;

	bfa_q_iter(&port->rport_q, qe) {
		rport = (bfa_fcs_rport_t *) qe;
		if (wwn_is_equal(rport->nwwn, nwwn))
			return rport;
	}

	bfa_trc(port->fcs, nwwn);
	return (NULL);
}

/**
 * Called by rport module when new rports are discovered.
 */
void
bfa_fcs_port_add_rport(bfa_fcs_port_t *port, bfa_fcs_rport_t *rport)
{
	bfa_q_enq(&port->rport_q, rport);
	port->num_rports++;
}

/**
 * Called by rport module to when rports are deleted.
 */
void
bfa_fcs_port_del_rport(bfa_fcs_port_t *port, bfa_fcs_rport_t *rport)
{
	bfa_assert(bfa_q_is_on_q(&port->rport_q, rport));
	bfa_q_qe_deq(rport);
	port->num_rports--;

	bfa_sm_send_event(port, BFA_FCS_PORT_SM_DELRPORT);
}

/**
 * Called by fabric for base port when fabric login is complete.
 * Called by vport for virtual ports when FDISC is complete.
 */
void
bfa_fcs_port_online(bfa_fcs_port_t *port)
{
	bfa_sm_send_event(port, BFA_FCS_PORT_SM_ONLINE);
}

/**
 * Called by fabric for base port when fabric goes offline.
 * Called by vport for virtual ports when virtual port becomes offline.
 */
void
bfa_fcs_port_offline(bfa_fcs_port_t *port)
{
	bfa_sm_send_event(port, BFA_FCS_PORT_SM_OFFLINE);
}

/**
 * Called by fabric to delete base lport and associated resources.
 *
 * Called by vport to delete lport and associated resources. Should call
 * bfa_fcs_vport_delete_comp() for vports on completion.
 */
void
bfa_fcs_port_delete(bfa_fcs_port_t *port)
{
	bfa_sm_send_event(port, BFA_FCS_PORT_SM_DELETE);
}

/**
 * Called by fabric in private loop topology to process LIP event.
 */
void
bfa_fcs_port_lip(bfa_fcs_port_t *port)
{
}

/**
 * Return TRUE if port is online, else return FALSE
 */
bfa_boolean_t
bfa_fcs_port_is_online(bfa_fcs_port_t *port)
{
	return (bfa_sm_cmp_state(port, bfa_fcs_port_sm_online));
}

/**
 * Logical port initialization of base or virtual port.
 * Called by fabric for base port or by vport for virtual ports.
 */
void
bfa_fcs_lport_init(bfa_fcs_port_t *lport, bfa_fcs_t *fcs, uint16_t vf_id,
		   bfa_port_cfg_t *port_cfg, bfa_fcs_vport_t *vport)
{
	lport->fcs = fcs;
	lport->fabric = bfa_fcs_vf_lookup(fcs, vf_id);
	lport->port_cfg = *port_cfg;
	lport->vport = vport;

	bfa_q_init(&lport->rport_q);
	lport->num_rports = 0;

	lport->bfad_port =
		bfa_fcb_port_new(fcs->bfad, lport, lport->port_cfg.roles,
				lport->fabric->vf_drv,
				vport ? vport->vport_drv : NULL);
	bfa_fcs_port_aen_post(lport, BFA_LPORT_AEN_NEW);

	bfa_sm_set_state(lport, bfa_fcs_port_sm_uninit);
	bfa_sm_send_event(lport, BFA_FCS_PORT_SM_CREATE);
}

/**
 * @}
 */

/**
 * @dg fcs_lport_api
 * @{
 */

void
bfa_fcs_port_get_attr(bfa_fcs_port_t *port, bfa_port_attr_t *port_attr)
{
	if (bfa_sm_cmp_state(port, bfa_fcs_port_sm_online))
		port_attr->pid = port->pid;
	else
		port_attr->pid = 0;

	port_attr->port_cfg = port->port_cfg;

	if (port->fabric) {
		port_attr->port_type = bfa_fcs_fabric_port_type(port->fabric);
		port_attr->loopback = bfa_fcs_fabric_is_loopback(port->fabric);

		if (port->vport != NULL)
			port_attr->port_type = BFA_PPORT_TYPE_VPORT;

	} else {
		port_attr->port_type = BFA_PPORT_TYPE_UNKNOWN;
		port_attr->state = BFA_PORT_UNINIT;
	}

}

/**
 * @}
 */
