/*
 * Copyright (c)  2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

/**
 * @file fabric.c Fabric module implementation.
 */

#include "fcs_fabric.h"
#include "fcs.h"
#include "fcs_port.h"
#include "fcs_vport.h"
#include "fcs_trcmod.h"
#include "fcs_fcxp.h"
#include "fcbuild.h"
#include "protocol/fc_sp.h"
#include "fcs/bfa_fcs_auth.h"
#include "fcs_auth.h"
#include <log/bfa_log_fcs.h>
#include <hal/bfa_hal_wwn.h>
#include <hal/bfa_hal_pport.h>

BFA_TRC_FILE(FCS, FABRIC);

#define BFA_FCS_FABRIC_RETRY_DELAY		(2000)	/* milliseconds */
#define BFA_FCS_FABRIC_CLEANUP_DELAY	(10000)	/* milliseconds */

#define bfa_fcs_fabric_set_opertype(__fabric) do {          \
    if (bfa_hal_pport_get_topology((__fabric)->fcs->hal)    \
				== BFA_PPORT_TOPOLOGY_P2P)   \
	    (__fabric)->oper_type = BFA_PPORT_TYPE_NPORT;       \
    else                                                    \
	    (__fabric)->oper_type = BFA_PPORT_TYPE_NLPORT;      \
} while (0)

/*
 * forward declarations
 */
static void     bfa_fcs_fabric_init(bfa_fcs_fabric_t *fabric);
static void     bfa_fcs_fabric_send_flogi(void *fabric_cbarg,
					  bfa_hal_fcxp_t *fcxp_alloced);
static void     bfa_fcs_fabric_notify_online(bfa_fcs_fabric_t *fabric);
static void     bfa_fcs_fabric_notify_offline(bfa_fcs_fabric_t *fabric);
static void     bfa_fcs_fabric_delay(void *cbarg);
static void     bfa_fcs_fabric_delete(bfa_fcs_fabric_t *fabric);
static void     bfa_fcs_fabric_delete_comp(void *cbarg);
static void     bfa_fcs_fabric_process_uf(bfa_fcs_fabric_t *fabric,
					  fchs_t *fchs, uint16_t len);
static void     bfa_fcs_fabric_process_flogi(bfa_fcs_fabric_t *fabric,
					     fchs_t *fchs, uint16_t len);
static void     bfa_fcs_fabric_send_flogi_acc(void *fabric_cbarg,
					      bfa_hal_fcxp_t *fcxp_alloced);
static void     bfa_fcs_fabric_flogiacc_comp(void *fcsarg,
					     bfa_hal_fcxp_t *fcxp, void *cbarg,
					     bfa_status_t status,
					     uint32_t rsp_len,
					     uint32_t resid_len,
					     fchs_t *rspfchs);
/**
 * @dg fcs_fabric_sm fabric state machine functions
 * @{
 * @img fabric_sm.jpg
 */

/**
 * Fabric state machine events
 */
typedef enum {
	BFA_FCS_FABRIC_SM_CREATE = 1,	/* !< fabric create from driver */
	BFA_FCS_FABRIC_SM_DELETE = 2,	/* !< fabric delete from driver */
	BFA_FCS_FABRIC_SM_LINK_DOWN = 3, /* !< link down from port */
	BFA_FCS_FABRIC_SM_LINK_UP = 4,	/* !< link up from port */
	BFA_FCS_FABRIC_SM_CONT_OP = 5,	/* !< continue op from flogi/auth */
	BFA_FCS_FABRIC_SM_RETRY_OP = 6,	/* !< continue op from flogi/auth */
	BFA_FCS_FABRIC_SM_NO_FABRIC = 7, /* !< no fabric from flogi/auth */
	BFA_FCS_FABRIC_SM_PERF_EVFP = 8, /* !< perform EVFP from flogi/auth */
	BFA_FCS_FABRIC_SM_ISOLATE = 9,   /* !< isolate from EVFP processing */
	BFA_FCS_FABRIC_SM_NO_TAGGING = 10, /* !< no VFT tagging from EVFP */
	BFA_FCS_FABRIC_SM_FCXP_SENT = 11, /* !< fcxp sent completion */
	BFA_FCS_FABRIC_SM_DELAYED = 12,	/* !< timeout delay event */
	BFA_FCS_FABRIC_SM_AUTH_FAILED = 13,	/* !< authentication failed */
	BFA_FCS_FABRIC_SM_AUTH_SUCCESS = 14, /* !< authentication successful */
	BFA_FCS_FABRIC_SM_DELCOMP = 15,	/* !< all vports deleted event */
	BFA_FCS_FABRIC_SM_LOOPBACK = 16,	/* !< Received our own FLOGI */
} bfa_fcs_fabric_event_t;

static void     bfa_fcs_fabric_sm_uninit(bfa_fcs_fabric_t *fabric,
					 bfa_fcs_fabric_event_t event);
static void     bfa_fcs_fabric_sm_linkdown(bfa_fcs_fabric_t *fabric,
					   bfa_fcs_fabric_event_t event);
static void     bfa_fcs_fabric_sm_flogi_send(bfa_fcs_fabric_t *fabric,
					     bfa_fcs_fabric_event_t event);

static void     bfa_fcs_fabric_sm_flogi(bfa_fcs_fabric_t *fabric,
					bfa_fcs_fabric_event_t event);
static void     bfa_fcs_fabric_sm_flogi_retry(bfa_fcs_fabric_t *fabric,
					      bfa_fcs_fabric_event_t event);
static void     bfa_fcs_fabric_sm_auth(bfa_fcs_fabric_t *fabric,
				       bfa_fcs_fabric_event_t event);
static void     bfa_fcs_fabric_sm_auth_failed(bfa_fcs_fabric_t *fabric,
					      bfa_fcs_fabric_event_t event);
static void     bfa_fcs_fabric_sm_loopback(bfa_fcs_fabric_t *fabric,
					   bfa_fcs_fabric_event_t event);
static void     bfa_fcs_fabric_sm_nofabric(bfa_fcs_fabric_t *fabric,
					   bfa_fcs_fabric_event_t event);
static void     bfa_fcs_fabric_sm_online(bfa_fcs_fabric_t *fabric,
					 bfa_fcs_fabric_event_t event);
static void     bfa_fcs_fabric_sm_evfp(bfa_fcs_fabric_t *fabric,
				       bfa_fcs_fabric_event_t event);
static void     bfa_fcs_fabric_sm_evfp_done(bfa_fcs_fabric_t *fabric,
					    bfa_fcs_fabric_event_t event);
static void     bfa_fcs_fabric_sm_isolated(bfa_fcs_fabric_t *fabric,
					   bfa_fcs_fabric_event_t event);
static void     bfa_fcs_fabric_sm_deleting(bfa_fcs_fabric_t *fabric,
					   bfa_fcs_fabric_event_t event);
/**
 * @brief
 *   Beginning state before fabric creation.
 */
static void
bfa_fcs_fabric_sm_uninit(bfa_fcs_fabric_t *fabric,
			 bfa_fcs_fabric_event_t event)
{
	bfa_trc(fabric->fcs, fabric->bport.port_cfg.pwwn);
	bfa_trc(fabric->fcs, event);

	switch (event) {
	case BFA_FCS_FABRIC_SM_CREATE:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_linkdown);
		bfa_fcs_fabric_init(fabric);
		bfa_fcs_lport_init(&fabric->bport, fabric->fcs, FC_VF_ID_NULL,
				   &fabric->bport.port_cfg, NULL);
		break;

	default:
		bfa_assert(event != BFA_FCS_FABRIC_SM_CREATE);
	}
}

/**
 * @brief
 *   Link is down, awaiting LINK UP event from port. This is also the
 *   first state at fabric creation.
 */
static void
bfa_fcs_fabric_sm_linkdown(bfa_fcs_fabric_t *fabric,
			   bfa_fcs_fabric_event_t event)
{
	bfa_trc(fabric->fcs, fabric->bport.port_cfg.pwwn);
	bfa_trc(fabric->fcs, event);

	switch (event) {
	case BFA_FCS_FABRIC_SM_LINK_UP:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_flogi_send);
		bfa_fcs_fabric_send_flogi(fabric, NULL);
		break;

	case BFA_FCS_FABRIC_SM_RETRY_OP:
		break;

	case BFA_FCS_FABRIC_SM_DELETE:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_deleting);
		bfa_fcs_fabric_delete(fabric);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 *   Flogi is being sent.
 */
static void
bfa_fcs_fabric_sm_flogi_send(bfa_fcs_fabric_t *fabric,
			     bfa_fcs_fabric_event_t event)
{
	bfa_trc(fabric->fcs, fabric->bport.port_cfg.pwwn);
	bfa_trc(fabric->fcs, event);

	switch (event) {
	case BFA_FCS_FABRIC_SM_FCXP_SENT:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_flogi);
		break;

	case BFA_FCS_FABRIC_SM_LINK_DOWN:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_linkdown);
		bfa_hal_fcxp_walloc_cancel(fabric->fcs->hal, &fabric->fcxp_wqe);
		break;

	case BFA_FCS_FABRIC_SM_DELETE:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_deleting);
		bfa_hal_fcxp_walloc_cancel(fabric->fcs->hal, &fabric->fcxp_wqe);
		bfa_fcs_fabric_delete(fabric);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 *   FLOGI is in progress, awaiting FLOGI reply.
 */
static void
bfa_fcs_fabric_sm_flogi(bfa_fcs_fabric_t *fabric, bfa_fcs_fabric_event_t event)
{
	bfa_trc(fabric->fcs, fabric->bport.port_cfg.pwwn);
	bfa_trc(fabric->fcs, event);

	switch (event) {
	case BFA_FCS_FABRIC_SM_CONT_OP:

		bfa_hal_pport_set_tx_bbcredit(fabric->fcs->hal,
					      fabric->bb_credit);
		fabric->fab_type = BFA_FCS_FABRIC_SWITCHED;

		if (fabric->auth_reqd) {
			bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_auth);
			bfa_trc(fabric->fcs, event);
			bfa_fcs_auth_start(fabric);
		} else {
			bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_online);
			bfa_fcs_fabric_notify_online(fabric);
		}
		break;

	case BFA_FCS_FABRIC_SM_RETRY_OP:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_flogi_retry);
		bfa_hal_timer_start(fabric->fcs->hal, &fabric->delay_timer,
				    bfa_fcs_fabric_delay, fabric,
				    BFA_FCS_FABRIC_RETRY_DELAY);
		break;

	case BFA_FCS_FABRIC_SM_LOOPBACK:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_loopback);
		bfa_hal_fcxp_discard(fabric->fcxp);
		bfa_fcs_fabric_set_opertype(fabric);
		break;

	case BFA_FCS_FABRIC_SM_NO_FABRIC:
		fabric->fab_type = BFA_FCS_FABRIC_N2N;
		bfa_hal_pport_set_tx_bbcredit(fabric->fcs->hal,
					      fabric->bb_credit);
		bfa_fcs_fabric_notify_online(fabric);
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_nofabric);
		break;

	case BFA_FCS_FABRIC_SM_LINK_DOWN:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_linkdown);
		bfa_hal_fcxp_discard(fabric->fcxp);
		break;

	case BFA_FCS_FABRIC_SM_DELETE:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_deleting);
		bfa_hal_fcxp_discard(fabric->fcxp);
		bfa_fcs_fabric_delete(fabric);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 */
static void
bfa_fcs_fabric_sm_flogi_retry(bfa_fcs_fabric_t *fabric,
			      bfa_fcs_fabric_event_t event)
{
	bfa_trc(fabric->fcs, fabric->bport.port_cfg.pwwn);
	bfa_trc(fabric->fcs, event);

	switch (event) {
	case BFA_FCS_FABRIC_SM_DELAYED:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_flogi_send);
		bfa_fcs_fabric_send_flogi(fabric, NULL);
		break;

	case BFA_FCS_FABRIC_SM_LINK_DOWN:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_linkdown);
		bfa_hal_timer_stop(&fabric->delay_timer);
		break;

	case BFA_FCS_FABRIC_SM_DELETE:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_deleting);
		bfa_hal_timer_stop(&fabric->delay_timer);
		bfa_fcs_fabric_delete(fabric);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 *   Authentication is in progress, awaiting authentication results.
 */
static void
bfa_fcs_fabric_sm_auth(bfa_fcs_fabric_t *fabric, bfa_fcs_fabric_event_t event)
{
	bfa_trc(fabric->fcs, fabric->bport.port_cfg.pwwn);
	bfa_trc(fabric->fcs, event);

	switch (event) {
	case BFA_FCS_FABRIC_SM_AUTH_FAILED:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_auth_failed);
		break;

	case BFA_FCS_FABRIC_SM_AUTH_SUCCESS:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_online);
		bfa_fcs_fabric_notify_online(fabric);
		break;

	case BFA_FCS_FABRIC_SM_PERF_EVFP:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_evfp);
		break;

	case BFA_FCS_FABRIC_SM_LINK_DOWN:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_linkdown);
		bfa_fcs_auth_stop(fabric);
		break;

	case BFA_FCS_FABRIC_SM_DELETE:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_deleting);
		bfa_fcs_auth_stop(fabric);
		bfa_fcs_fabric_delete(fabric);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 *   Authentication failed
 */
static void
bfa_fcs_fabric_sm_auth_failed(bfa_fcs_fabric_t *fabric,
			      bfa_fcs_fabric_event_t event)
{
	bfa_trc(fabric->fcs, fabric->bport.port_cfg.pwwn);
	bfa_trc(fabric->fcs, event);

	switch (event) {
	case BFA_FCS_FABRIC_SM_LINK_DOWN:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_linkdown);
		bfa_fcs_auth_stop(fabric);
		break;

	case BFA_FCS_FABRIC_SM_DELETE:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_deleting);
		bfa_fcs_auth_stop(fabric);
		bfa_fcs_fabric_delete(fabric);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 *   Port is in loopback mode.
 */
static void
bfa_fcs_fabric_sm_loopback(bfa_fcs_fabric_t *fabric,
			   bfa_fcs_fabric_event_t event)
{
	bfa_trc(fabric->fcs, fabric->bport.port_cfg.pwwn);
	bfa_trc(fabric->fcs, event);

	switch (event) {
	case BFA_FCS_FABRIC_SM_LINK_DOWN:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_linkdown);
		bfa_fcs_fabric_notify_offline(fabric);
		break;

	case BFA_FCS_FABRIC_SM_DELETE:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_deleting);
		bfa_fcs_fabric_delete(fabric);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 *   There is no attached fabric - private loop or NPort-to-NPort topology.
 */
static void
bfa_fcs_fabric_sm_nofabric(bfa_fcs_fabric_t *fabric,
			   bfa_fcs_fabric_event_t event)
{
	bfa_trc(fabric->fcs, fabric->bport.port_cfg.pwwn);
	bfa_trc(fabric->fcs, event);

	switch (event) {
	case BFA_FCS_FABRIC_SM_LINK_DOWN:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_linkdown);
		bfa_fcs_fabric_notify_offline(fabric);
		break;

	case BFA_FCS_FABRIC_SM_DELETE:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_deleting);
		bfa_fcs_fabric_delete(fabric);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 *   Fabric is online - normal operating state.
 */
static void
bfa_fcs_fabric_sm_online(bfa_fcs_fabric_t *fabric,
			 bfa_fcs_fabric_event_t event)
{
	bfa_trc(fabric->fcs, fabric->bport.port_cfg.pwwn);
	bfa_trc(fabric->fcs, event);

	switch (event) {
	case BFA_FCS_FABRIC_SM_LINK_DOWN:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_linkdown);
		if (fabric->auth_reqd)
			bfa_fcs_auth_stop(fabric);
		bfa_fcs_fabric_notify_offline(fabric);
		break;

	case BFA_FCS_FABRIC_SM_DELETE:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_deleting);
		if (fabric->auth_reqd)
			bfa_fcs_auth_stop(fabric);
		bfa_fcs_fabric_delete(fabric);
		break;

	case BFA_FCS_FABRIC_SM_AUTH_FAILED:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_auth_failed);
		break;

	case BFA_FCS_FABRIC_SM_AUTH_SUCCESS:
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 *   Exchanging virtual fabric parameters.
 */
static void
bfa_fcs_fabric_sm_evfp(bfa_fcs_fabric_t *fabric, bfa_fcs_fabric_event_t event)
{
	bfa_trc(fabric->fcs, fabric->bport.port_cfg.pwwn);
	bfa_trc(fabric->fcs, event);

	switch (event) {
	case BFA_FCS_FABRIC_SM_CONT_OP:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_evfp_done);
		break;

	case BFA_FCS_FABRIC_SM_ISOLATE:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_isolated);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 *   EVFP exchange complete and VFT tagging is enabled.
 */
static void
bfa_fcs_fabric_sm_evfp_done(bfa_fcs_fabric_t *fabric,
			    bfa_fcs_fabric_event_t event)
{
	bfa_trc(fabric->fcs, fabric->bport.port_cfg.pwwn);
	bfa_trc(fabric->fcs, event);
}

/**
 * @brief
 *   Port is isolated after EVFP exchange due to VF_ID mismatch (N and F).
 */
static void
bfa_fcs_fabric_sm_isolated(bfa_fcs_fabric_t *fabric,
			   bfa_fcs_fabric_event_t event)
{
	bfa_trc(fabric->fcs, fabric->bport.port_cfg.pwwn);
	bfa_trc(fabric->fcs, event);

	bfa_log(fabric->fcs->logm, BFA_LOG_FCS_FABRIC_ISOLATED,
		fabric->bport.port_cfg.pwwn, fabric->fcs->port_vfid,
		fabric->event_arg.swp_vfid);
}

/**
 * @brief
 *   Fabric is being deleted, awaiting vport delete completions.
 */
static void
bfa_fcs_fabric_sm_deleting(bfa_fcs_fabric_t *fabric,
			   bfa_fcs_fabric_event_t event)
{
	bfa_trc(fabric->fcs, fabric->bport.port_cfg.pwwn);
	bfa_trc(fabric->fcs, event);

	switch (event) {
	case BFA_FCS_FABRIC_SM_DELCOMP:
		bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_uninit);
		bfa_fcs_modexit_comp(fabric->fcs);
		break;

	case BFA_FCS_FABRIC_SM_LINK_UP:
		break;

	case BFA_FCS_FABRIC_SM_LINK_DOWN:
		bfa_fcs_fabric_notify_offline(fabric);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @}
 */

/**
 * @dg fcs_fabric_private fabric private functions
 * @{
 */

static void
bfa_fcs_fabric_init(bfa_fcs_fabric_t *fabric)
{
	bfa_port_cfg_t *port_cfg = &fabric->bport.port_cfg;

	port_cfg->roles = BFA_PORT_ROLE_FCP_IM;
	port_cfg->nwwn = bfa_hal_wwn_get_naa2(fabric->fcs->hal);
	port_cfg->pwwn = bfa_hal_wwn_get_naa1(fabric->fcs->hal);
}

/**
 * @brief
 * Port Symbolic Name Creation for base port.
 */
void
bfa_fcs_fabric_psymb_init(bfa_fcs_fabric_t *fabric)
{
	bfa_port_cfg_t *port_cfg = &fabric->bport.port_cfg;
	bfa_ioc_attr_t  ioc_attr;
	bfa_fcs_driver_info_t  *driver_info = &fabric->fcs->driver_info;

	bfa_hal_ioc_get_attr(fabric->fcs->hal, &ioc_attr);

	sprintf((char *)&port_cfg->sym_name, "%s%s%s%s%s%s%s%s%s", 
			ioc_attr.adapter_attr.model,
			BFA_FCS_PORT_SYMBNAME_SEPARATOR,
			(char *)driver_info->version,
			BFA_FCS_PORT_SYMBNAME_SEPARATOR,
			(char *)driver_info->host_machine_name,
			(char *)BFA_FCS_PORT_SYMBNAME_SEPARATOR,
			(char *)driver_info->host_os_name,
			(char *)BFA_FCS_PORT_SYMBNAME_SEPARATOR,
			(char *)driver_info->host_os_patch);
}

/**
 * @brief
 *   Flogi completion callback.
 */
static void
bfa_fcs_fabric_flogi_comp(void *fcsarg, bfa_hal_fcxp_t *fcxp, void *cbarg,
			  bfa_status_t status, uint32_t rsp_len,
			  uint32_t resid_len, fchs_t *rspfchs)
{
	bfa_fcs_fabric_t *fabric = cbarg;
	fc_flogi_t     *flogi_resp;
	fc_els_cmd_t   *els_cmd;
	fc_ls_rjt_t    *ls_rjt;

	bfa_trc(fabric->fcs, fabric->bport.port_cfg.pwwn);

	if (status != BFA_STATUS_OK) {
		fabric->stats.flogi_rsp_err++;
		bfa_trc(fabric->fcs, status);
		bfa_sm_send_event(fabric, BFA_FCS_FABRIC_SM_RETRY_OP);
		return;
	}

	els_cmd = (fc_els_cmd_t *) BFA_FCXP_RSP_PLD(fcxp);

	switch (els_cmd->els_code) {
	case FC_ELS_ACC:
		if (rsp_len < sizeof(fc_flogi_t)) {
			bfa_trc(fabric->fcs, rsp_len);
			fabric->stats.flogi_acc_err++;
			bfa_sm_send_event(fabric, BFA_FCS_FABRIC_SM_RETRY_OP);
			return;
		}
		break;

	case FC_ELS_LS_RJT:
		ls_rjt = (fc_ls_rjt_t *) BFA_FCXP_RSP_PLD(fcxp);

		fabric->stats.flogi_rejects++;
		bfa_trc(fabric->fcs, ls_rjt->reason_code);
		bfa_trc(fabric->fcs, ls_rjt->reason_code_expl);
		bfa_sm_send_event(fabric, BFA_FCS_FABRIC_SM_RETRY_OP);
		return;

	default:
		bfa_trc(fabric->fcs, els_cmd->els_code);
		fabric->stats.flogi_unknown_rsp++;
		bfa_assert(0);
		bfa_sm_send_event(fabric, BFA_FCS_FABRIC_SM_RETRY_OP);
		return;
	}

	flogi_resp = (fc_flogi_t *) els_cmd;

	bfa_trc(fabric->fcs, fabric->bport.pid);
	bfa_trc(fabric->fcs, fabric->bb_credit);

	/*
	 * If attached to a F-port, Check if the PID is assigned
	 */
	if ((flogi_resp->csp.port_type) && (rspfchs->d_id == 0)) {
		bfa_trc(fabric->fcs, rspfchs->d_id);
		fabric->stats.flogi_acc_err++;
		bfa_sm_send_event(fabric, BFA_FCS_FABRIC_SM_RETRY_OP);
		return;
	}

	fabric->bb_credit = bfa_os_ntohs(flogi_resp->csp.bbcred);

	fabric->stats.flogi_accepts++;
	/*
	 * Check port type. It should be 1 = F-port.
	 */
	if (flogi_resp->csp.port_type) {
		fabric->bport.pid = rspfchs->d_id;
		fabric->is_npiv = flogi_resp->csp.npiv_supp;
		fabric->is_auth = flogi_resp->csp.security;
		bfa_sm_send_event(fabric, BFA_FCS_FABRIC_SM_CONT_OP);
	} else {
		/*
		 * Nport-2-Nport direct attached
		 */
		fabric->bport.port_topo.pn2n.rem_port_wwn =
			flogi_resp->port_name;
		bfa_sm_send_event(fabric, BFA_FCS_FABRIC_SM_NO_FABRIC);
	}
}

/**
 * @brief
 * 		Allocate and send FLOGI.
 */
static void
bfa_fcs_fabric_send_flogi(void *fabric_cbarg, bfa_hal_fcxp_t *fcxp_alloced)
{
	bfa_fcs_fabric_t *fabric = fabric_cbarg;
	bfa_hal_t      *hal = fabric->fcs->hal;
	bfa_hal_fcxp_t *fcxp;
	uint16_t        reqlen;
	fchs_t          fchs;
	bfa_port_cfg_t *pcfg = &fabric->bport.port_cfg;
	uint8_t         alpa = 0;

	fcxp = fcxp_alloced ? fcxp_alloced : bfa_fcs_fcxp_alloc(fabric->fcs);
	if (!fcxp) {
		fabric->stats.flogi_alloc_wait++;
		bfa_hal_fcxp_alloc_wait(hal, &fabric->fcxp_wqe,
					bfa_fcs_fabric_send_flogi, fabric);
		return;
	}
	fabric->fcxp = fcxp;

	if (bfa_hal_pport_get_topology(hal) == BFA_PPORT_TOPOLOGY_LOOP)
		alpa = bfa_hal_pport_get_myalpa(hal);

	reqlen = fc_flogi_build(&fchs, bfa_hal_fcxp_get_reqbuf(fcxp),
				bfa_os_hton3b(alpa), 0, pcfg->pwwn, pcfg->nwwn,
				bfa_hal_pport_get_maxfrsize(hal), BFA_TRUE, fabric->auth_reqd,
				bfa_hal_pport_get_rx_bbcredit(hal));

	bfa_hal_fcxp_send(fcxp, NULL, fabric->vf_id, BFA_FALSE, FC_CLASS_3,
			  reqlen, &fchs, bfa_fcs_fabric_flogi_comp, fabric,
			  FC_MAX_PDUSZ, 2 * FC_ED_TOV);

	fabric->stats.flogi_sent++;
	bfa_sm_send_event(fabric, BFA_FCS_FABRIC_SM_FCXP_SENT);
}

static void
bfa_fcs_fabric_notify_online(bfa_fcs_fabric_t *fabric)
{
	bfa_fcs_vport_t *vport;
	bfa_q_t        *qe, *qen;

	bfa_trc(fabric->fcs, fabric->fabric_name);

	bfa_fcs_fabric_set_opertype(fabric);
	fabric->stats.fabric_onlines++;

	/**
	 * notify online event to base and then virtual ports
	 */
	bfa_fcs_port_online(&fabric->bport);

	bfa_q_iter_safe(&fabric->vport_q, qe, qen) {
		vport = (bfa_fcs_vport_t *) qe;
		bfa_fcs_vport_online(vport);
	}
}

static void
bfa_fcs_fabric_notify_offline(bfa_fcs_fabric_t *fabric)
{
	bfa_fcs_vport_t *vport;
	bfa_q_t        *qe, *qen;

	bfa_trc(fabric->fcs, fabric->fabric_name);
	fabric->stats.fabric_offlines++;

	/**
	 * notify offline event first to vports and then base port.
	 */
	bfa_q_iter_safe(&fabric->vport_q, qe, qen) {
		vport = (bfa_fcs_vport_t *) qe;
		bfa_fcs_vport_offline(vport);
	}

	bfa_fcs_port_offline(&fabric->bport);

	/**
	 * @todo
	 * 	 if (!mpiomode)
	 * 		start timer;
	 */
}

static void
bfa_fcs_fabric_delay(void *cbarg)
{
	bfa_fcs_fabric_t *fabric = cbarg;

	bfa_sm_send_event(fabric, BFA_FCS_FABRIC_SM_DELAYED);
}

/**
 * Delete all vports and wait for vport delete completions.
 */
static void
bfa_fcs_fabric_delete(bfa_fcs_fabric_t *fabric)
{
	bfa_fcs_vport_t *vport;
	bfa_q_t        *qe, *qen;

	bfa_q_iter_safe(&fabric->vport_q, qe, qen) {
		vport = (bfa_fcs_vport_t *) qe;
		bfa_fcs_vport_delete(vport);
	}

	bfa_fcs_port_delete(&fabric->bport);
	bfa_wc_wait(&fabric->wc);
}

static void
bfa_fcs_fabric_delete_comp(void *cbarg)
{
	bfa_fcs_fabric_t *fabric = cbarg;

	bfa_sm_send_event(fabric, BFA_FCS_FABRIC_SM_DELCOMP);
}

/**
 * @}
 */

/**
 * @dg fcs_fabric_public fabric public functions
 * @img fabric_pub.jpg
 * @{
 */

/**
 * @brief
 *   Module initialization
 */
void
bfa_fcs_fabric_modinit(bfa_fcs_t *fcs)
{
	bfa_fcs_fabric_t *fabric;

	fabric = &fcs->fabric;
	bfa_os_memset(fabric, 0, sizeof(bfa_fcs_fabric_t));

	/**
	 * Initialize base fabric.
	 */
	fabric->fcs = fcs;
	bfa_q_init(&fabric->vport_q);
	bfa_q_init(&fabric->vf_q);

	/**
	 * Initialize fabric delete completion handler. Fabric deletion is complete
	 * when the last vport delete is complete.
	 */
	bfa_wc_init(&fabric->wc, bfa_fcs_fabric_delete_comp, fabric);
	bfa_wc_up(&fabric->wc); /* For the base port */

	bfa_sm_set_state(fabric, bfa_fcs_fabric_sm_uninit);
	bfa_sm_send_event(fabric, BFA_FCS_FABRIC_SM_CREATE);
	bfa_trc(fcs, 0);
}

/**
 * @brief
 *   Module cleanup
 */
void
bfa_fcs_fabric_modexit(bfa_fcs_t *fcs)
{
	bfa_fcs_fabric_t *fabric;

	bfa_trc(fcs, 0);

	/**
	 * Cleanup base fabric.
	 */
	fabric = &fcs->fabric;
	bfa_sm_send_event(fabric, BFA_FCS_FABRIC_SM_DELETE);
}

/**
 * @brief
 *   Suspend fabric activity as part of driver suspend.
 */
void
bfa_fcs_fabric_modsusp(bfa_fcs_t *fcs)
{
}

bfa_boolean_t
bfa_fcs_fabric_is_loopback(bfa_fcs_fabric_t *fabric)
{
	return (bfa_sm_cmp_state(fabric, bfa_fcs_fabric_sm_loopback));
}

bfa_pport_type_t
bfa_fcs_fabric_port_type(bfa_fcs_fabric_t *fabric)
{
	return fabric->oper_type;
}

/**
 * @brief
 *   Link up notification from HAL physical port module.
 */
void
bfa_fcs_fabric_link_up(bfa_fcs_fabric_t *fabric)
{
	bfa_trc(fabric->fcs, fabric->bport.port_cfg.pwwn);
	bfa_sm_send_event(fabric, BFA_FCS_FABRIC_SM_LINK_UP);
}

/**
 * @brief
 *   Link down notification from HAL physical port module.
 */
void
bfa_fcs_fabric_link_down(bfa_fcs_fabric_t *fabric)
{
	bfa_trc(fabric->fcs, fabric->bport.port_cfg.pwwn);
	bfa_sm_send_event(fabric, BFA_FCS_FABRIC_SM_LINK_DOWN);
}

/**
 * @brief
 *   A child vport is being created in the fabric.
 *
 *   Call from vport module at vport creation. A list of base port and vports
 *   belonging to a fabric is maintained to propagate link events.
 *
 *   param[in] fabric - Fabric instance. This can be a base fabric or vf.
 *   param[in] vport  - Vport being created.
 *
 *   @return None (always succeeds)
 */
void
bfa_fcs_fabric_addvport(bfa_fcs_fabric_t *fabric, bfa_fcs_vport_t *vport)
{
	/**
	 * - add vport to fabric's vport_q
	 */
	bfa_trc(fabric->fcs, fabric->vf_id);

	bfa_q_enq(&fabric->vport_q, vport);
	fabric->num_vports++;
	bfa_wc_up(&fabric->wc);
}

/**
 * @brief
 *   A child vport is being deleted from fabric.
 *
 *   Vport is being deleted.
 */
void
bfa_fcs_fabric_delvport(bfa_fcs_fabric_t *fabric, bfa_fcs_vport_t *vport)
{
	bfa_q_qe_deq(vport);
	fabric->num_vports--;
	bfa_wc_down(&fabric->wc);
}

/**
 * @brief
 *   Base port is deleted.
 */
void
bfa_fcs_fabric_port_delete_comp(bfa_fcs_fabric_t *fabric)
{
    bfa_wc_down(&fabric->wc);
}


/**
 * @brief
 *    Check if fabric is online.
 *
 *   param[in] fabric - Fabric instance. This can be a base fabric or vf.
 *
 *   @return  TRUE/FALSE
 */
int
bfa_fcs_fabric_is_online(bfa_fcs_fabric_t *fabric)
{
	return (bfa_sm_cmp_state(fabric, bfa_fcs_fabric_sm_online));
}

/**
 * 	@brief
 *
 */
bfa_status_t
bfa_fcs_fabric_addvf(bfa_fcs_fabric_t *vf, bfa_fcs_t *fcs,
		     bfa_port_cfg_t *port_cfg, bfad_vf_t *vf_drv)
{
	bfa_sm_set_state(vf, bfa_fcs_fabric_sm_uninit);
	return BFA_STATUS_OK;
}

/**
 * Lookup for a vport withing a fabric given its pwwn
 */
bfa_fcs_vport_t *
bfa_fcs_fabric_vport_lookup(bfa_fcs_fabric_t *fabric, wwn_t pwwn)
{
	bfa_fcs_vport_t *vport;
	bfa_q_t        *qe;

	bfa_q_iter(&fabric->vport_q, qe) {
		vport = (bfa_fcs_vport_t *) qe;
		if (bfa_fcs_port_get_pwwn(&vport->lport) == pwwn)
			return vport;
	}

	return NULL;
}

/**
 * @brief
 *    In a given fabric, return the number of lports.
 *
 *   param[in] fabric - Fabric instance. This can be a base fabric or vf.
 *
*   @return : 1 or more.
 */
uint16_t
bfa_fcs_fabric_vport_count(bfa_fcs_fabric_t *fabric)
{
	return (fabric->num_vports);
}

/**
 * @brief
 * 		Unsolicited frame receive handling.
 */
void
bfa_fcs_fabric_uf_recv(bfa_fcs_fabric_t *fabric, fchs_t *fchs, uint16_t len)
{
	uint32_t        pid = fchs->d_id;
	bfa_fcs_vport_t *vport;
	bfa_q_t        *qe;
	fc_els_cmd_t   *els_cmd = (fc_els_cmd_t *) (fchs + 1);
	fc_plogi_t     *flogi = (fc_plogi_t *) els_cmd;

	bfa_trc(fabric->fcs, len);
	bfa_trc(fabric->fcs, pid);

	/**
	 * Look for our own FLOGI frames being looped back. This means an
	 * external loopback cable is in place. Our own FLOGI frames are
	 * sometimes looped back when switch port gets temporarily bypassed.
	 */
	if ((pid == bfa_os_ntoh3b(FC_FABRIC_PORT)) 
	    && (els_cmd->els_code == FC_ELS_FLOGI)
	    && (flogi->port_name == bfa_fcs_port_get_pwwn(&fabric->bport))) {
		bfa_sm_send_event(fabric, BFA_FCS_FABRIC_SM_LOOPBACK);
		return;
	}

	/**
	 * FLOGI/EVFP exchanges should be consumed by base fabric.
	 */
	if (fchs->d_id == bfa_os_hton3b(FC_FABRIC_PORT)) {
		bfa_trc(fabric->fcs, pid);
		bfa_fcs_fabric_process_uf(fabric, fchs, len);
		return;
	}

	if (fabric->bport.pid == pid) {
		/**
		 * All authentication frames should be routed to auth
		 */
		bfa_trc(fabric->fcs, els_cmd->els_code);
		if (els_cmd->els_code == FC_ELS_AUTH) {
			bfa_trc(fabric->fcs, els_cmd->els_code);
			fabric->auth.response = (uint8_t *) els_cmd;
			bfa_fcs_auth_uf_recv(fabric, len);
			return;
		}

		bfa_trc(fabric->fcs, *(uint8_t *) ((uint8_t *) fchs));
		bfa_fcs_port_uf_recv(&fabric->bport, fchs, len);
		return;
	}

	/**
	 * look for a matching local port ID
	 */
	bfa_q_iter(&fabric->vport_q, qe) {
		vport = (bfa_fcs_vport_t *) qe;
		if (vport->lport.pid == pid) {
			bfa_fcs_port_uf_recv(&vport->lport, fchs, len);
			return;
		}
	}
	if (fabric->fab_type == BFA_FCS_FABRIC_N2N) {
		/*
		 * if this is a Plogi Frame, pass it to port
		 */
		fc_els_cmd_t   *els_cmd = (fc_els_cmd_t *) (fchs + 1);
		bfa_trc(fabric->fcs, els_cmd->els_code);
		if (els_cmd->els_code == FC_ELS_PLOGI) {
			fabric->bport.pid = fchs->d_id;
			bfa_fcs_port_uf_recv(&fabric->bport, fchs, len);
		}
	}
}

/**
 * @brief
 * 		Unsolicited frames to be processed by fabric.
 */
static void
bfa_fcs_fabric_process_uf(bfa_fcs_fabric_t *fabric, fchs_t *fchs,
			  uint16_t len)
{
	fc_els_cmd_t   *els_cmd = (fc_els_cmd_t *) (fchs + 1);

	bfa_trc(fabric->fcs, els_cmd->els_code);

	switch (els_cmd->els_code) {
	case FC_ELS_FLOGI:
		bfa_fcs_fabric_process_flogi(fabric, fchs, len);
		break;

	default:
		/*
		 * need to generate a LS_RJT
		 */
		break;
	}
}

/**
 * @brief
 * 	Process	incoming FLOGI
 */
static void
bfa_fcs_fabric_process_flogi(bfa_fcs_fabric_t *fabric, fchs_t *fchs,
			     uint16_t len)
{
	fc_flogi_t     *flogi = (fc_flogi_t *) (fchs + 1);
	bfa_fcs_port_t *bport = &fabric->bport;

	bfa_trc(fabric->fcs, fchs->s_id);

	fabric->stats.flogi_rcvd++;
	/*
	 * Check port type. It should be 0 = n-port.
	 */
	if (flogi->csp.port_type) {
		/*
		 * @todo: may need to send a LS_RJT
		 */
		bfa_trc(fabric->fcs, flogi->port_name);
		fabric->stats.flogi_rejected++;
		return;
	}

	fabric->bb_credit = bfa_os_ntohs(flogi->csp.bbcred);
	bport->port_topo.pn2n.rem_port_wwn = flogi->port_name;
	bport->port_topo.pn2n.reply_oxid = fchs->ox_id;

	/*
	 * Send a Flogi Acc
	 */
	bfa_fcs_fabric_send_flogi_acc(fabric, NULL);
	bfa_sm_send_event(fabric, BFA_FCS_FABRIC_SM_NO_FABRIC);
}

static void
bfa_fcs_fabric_send_flogi_acc(void *fabric_cbarg, bfa_hal_fcxp_t *fcxp_alloced)
{
	bfa_fcs_fabric_t *fabric = fabric_cbarg;
	bfa_port_cfg_t *pcfg = &fabric->bport.port_cfg;
	bfa_fcs_port_n2n_t *n2n_port = &fabric->bport.port_topo.pn2n;
	bfa_hal_t      *hal = fabric->fcs->hal;
	bfa_hal_fcxp_t *fcxp;
	uint16_t        reqlen;
	fchs_t          fchs;

	fcxp = fcxp_alloced ? fcxp_alloced : bfa_fcs_fcxp_alloc(fabric->fcs);
	if (!fcxp) {
		bfa_hal_fcxp_alloc_wait(hal, &fabric->fcxp_wqe,
					bfa_fcs_fabric_send_flogi_acc, fabric);
		return;
	}
	fabric->fcxp = fcxp;

	reqlen = fc_flogi_acc_build(&fchs, bfa_hal_fcxp_get_reqbuf(fcxp),
				bfa_os_hton3b(FC_FABRIC_PORT),
				n2n_port->reply_oxid, pcfg->pwwn,
				pcfg->nwwn,
				bfa_hal_pport_get_maxfrsize(hal),
				bfa_hal_pport_get_rx_bbcredit(hal));

	bfa_hal_fcxp_send(fcxp, NULL, fabric->vf_id, BFA_FALSE, FC_CLASS_3,
			 reqlen, &fchs, bfa_fcs_fabric_flogiacc_comp, fabric,
			 FC_MAX_PDUSZ, 0);	/* Timeout 0 indicates no
						 * response expected
						 */
}

/**
 * @brief
 *   Flogi Acc completion callback.
 */
static void
bfa_fcs_fabric_flogiacc_comp(void *fcsarg, bfa_hal_fcxp_t *fcxp, void *cbarg,
			     bfa_status_t status, uint32_t rsp_len,
			     uint32_t resid_len, fchs_t *rspfchs)
{
	bfa_fcs_fabric_t *fabric = cbarg;

	bfa_trc(fabric->fcs, status);
}

/*
 * @brief function called after the completion of the authentication
 *
 * @param[in] fabric - fabric
 * @param[in] result - 1
 *
 * @return - none
 */
void
bfa_fcs_auth_finished(bfa_fcs_fabric_t *fabric, auth_status_t status)
{
	bfa_trc(fabric->fcs, status);

	if (status == FC_AUTH_STATE_SUCCESS)
		bfa_sm_send_event(fabric, BFA_FCS_FABRIC_SM_AUTH_SUCCESS);
	else
		bfa_sm_send_event(fabric, BFA_FCS_FABRIC_SM_AUTH_FAILED);
}

/**
 * @}
 */
