/*
 * Copyright (c)  2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

/**
 * @file fcpim.c - FCP initiator mode i-t nexus state machine
 */

#include <hal/bfa_hal.h>
#include <hal/bfa_hal_timer.h>
#include <hal/bfa_hal_fcxp.h>
#include <hal/bfa_hal_rport.h>
#include <hcb/bfa_hcb_fcxp.h>
#include <fcs/bfa_fcs_port.h>
#include <fcs/bfa_fcs_vport.h>
#include <fcs/bfa_fcs_fcpim.h>
#include <hcb/bfa_hcb_itnim.h>
#include <fcb/bfa_fcb_fcpim.h>
#include "fcs_fcpim.h"
#include "fcs_rport.h"
#include "fcs_port.h"
#include "fcs_trcmod.h"
#include "fcs_fcxp.h"
#include "fcs.h"
#include <aen/bfal_aen_itnim.h>

BFA_TRC_FILE(FCS, FCPIM);

/*
 * forward declarations
 */
static void     bfa_fcs_itnim_timeout(void *arg);
static void     bfa_fcs_itnim_free(bfa_fcs_itnim_t *itnim);
static void     bfa_fcs_itnim_send_prli(void *itnim_cbarg,
					bfa_hal_fcxp_t *fcxp_alloced);
static void     bfa_fcs_itnim_prli_response(void *fcsarg, bfa_hal_fcxp_t *fcxp,
					    void *cbarg,
					    bfa_status_t req_status,
					    uint32_t rsp_len,
					    uint32_t resid_len,
					    fchs_t *rsp_fchs);
static void     bfa_fcs_itnim_aen_post(bfa_fcs_itnim_t *itnim,
				       bfa_itnim_aen_event_t event);

/**
 * @dg fcs_itnim_sm FCS itnim state machine events
 * @{
 */

typedef enum {
	BFA_FCS_ITNIM_SM_ONLINE = 1,	/* !< rport online event */
	BFA_FCS_ITNIM_SM_OFFLINE = 2,	/* !< rport offline */
	BFA_FCS_ITNIM_SM_FRMSENT = 3,	/* !< prli frame is sent */
	BFA_FCS_ITNIM_SM_RSP_OK = 4,	/* !< good response */
	BFA_FCS_ITNIM_SM_RSP_ERROR = 5,	/* !< error response */
	BFA_FCS_ITNIM_SM_TIMEOUT = 6,	/* !< delay timeout */
	BFA_FCS_ITNIM_SM_HCB_OFFLINE = 7, /* !< HAL online callback */
	BFA_FCS_ITNIM_SM_HCB_ONLINE = 8, /* !< HAL pffline callback */
	BFA_FCS_ITNIM_SM_INITIATOR = 9,	/* !< rport is initiator */
	BFA_FCS_ITNIM_SM_DELETE = 10,	/* !< delete event from rport */
	BFA_FCS_ITNIM_SM_PRLO = 11,	/* !< delete event from rport */
} bfa_fcs_itnim_event_t;

static void     bfa_fcs_itnim_sm_offline(bfa_fcs_itnim_t *itnim,
					 bfa_fcs_itnim_event_t event);
static void     bfa_fcs_itnim_sm_prli_send(bfa_fcs_itnim_t *itnim,
					   bfa_fcs_itnim_event_t event);
static void     bfa_fcs_itnim_sm_prli(bfa_fcs_itnim_t *itnim,
				      bfa_fcs_itnim_event_t event);
static void     bfa_fcs_itnim_sm_prli_retry(bfa_fcs_itnim_t *itnim,
					    bfa_fcs_itnim_event_t event);
static void     bfa_fcs_itnim_sm_hcb_online(bfa_fcs_itnim_t *itnim,
					    bfa_fcs_itnim_event_t event);
static void     bfa_fcs_itnim_sm_online(bfa_fcs_itnim_t *itnim,
					bfa_fcs_itnim_event_t event);
static void     bfa_fcs_itnim_sm_hcb_offline(bfa_fcs_itnim_t *itnim,
					     bfa_fcs_itnim_event_t event);
static void     bfa_fcs_itnim_sm_initiator(bfa_fcs_itnim_t *itnim,
					   bfa_fcs_itnim_event_t event);

static bfa_sm_table_t itnim_sm_table[] = {
	{BFA_SM(bfa_fcs_itnim_sm_offline), BFA_ITNIM_OFFLINE},
	{BFA_SM(bfa_fcs_itnim_sm_prli_send), BFA_ITNIM_PRLI_SEND},
	{BFA_SM(bfa_fcs_itnim_sm_prli), BFA_ITNIM_PRLI_SENT},
	{BFA_SM(bfa_fcs_itnim_sm_prli_retry), BFA_ITNIM_PRLI_RETRY},
	{BFA_SM(bfa_fcs_itnim_sm_hcb_online), BFA_ITNIM_HCB_ONLINE},
	{BFA_SM(bfa_fcs_itnim_sm_online), BFA_ITNIM_ONLINE},
	{BFA_SM(bfa_fcs_itnim_sm_hcb_offline), BFA_ITNIM_HCB_OFFLINE},
	{BFA_SM(bfa_fcs_itnim_sm_initiator), BFA_ITNIM_INITIATIOR},
};

/**
 * @dg fcs_itnim_sm FCS itnim state machine
 * @{
 */

static void
bfa_fcs_itnim_sm_offline(bfa_fcs_itnim_t *itnim, bfa_fcs_itnim_event_t event)
{
	bfa_trc(itnim->fcs, itnim->rport->pwwn);
	bfa_trc(itnim->fcs, event);

	switch (event) {
	case BFA_FCS_ITNIM_SM_ONLINE:
		bfa_sm_set_state(itnim, bfa_fcs_itnim_sm_prli_send);
		bfa_fcs_itnim_send_prli(itnim, NULL);
		break;

	case BFA_FCS_ITNIM_SM_OFFLINE:
		bfa_fcs_rport_itnim_ack(itnim->rport);
		break;

	case BFA_FCS_ITNIM_SM_INITIATOR:
		bfa_sm_set_state(itnim, bfa_fcs_itnim_sm_initiator);
		break;

	case BFA_FCS_ITNIM_SM_DELETE:
		bfa_fcs_itnim_free(itnim);
		break;

	default:
		bfa_assert(0);
	}

}

static void
bfa_fcs_itnim_sm_prli_send(bfa_fcs_itnim_t *itnim, bfa_fcs_itnim_event_t event)
{
	bfa_trc(itnim->fcs, itnim->rport->pwwn);
	bfa_trc(itnim->fcs, event);

	switch (event) {
	case BFA_FCS_ITNIM_SM_FRMSENT:
		bfa_sm_set_state(itnim, bfa_fcs_itnim_sm_prli);
		break;

	case BFA_FCS_ITNIM_SM_INITIATOR:
		bfa_sm_set_state(itnim, bfa_fcs_itnim_sm_initiator);
		bfa_hal_fcxp_walloc_cancel(itnim->fcs->hal, &itnim->fcxp_wqe);
		break;

	case BFA_FCS_ITNIM_SM_OFFLINE:
		bfa_sm_set_state(itnim, bfa_fcs_itnim_sm_offline);
		bfa_hal_fcxp_walloc_cancel(itnim->fcs->hal, &itnim->fcxp_wqe);
		bfa_fcs_rport_itnim_ack(itnim->rport);
		break;

	case BFA_FCS_ITNIM_SM_DELETE:
		bfa_sm_set_state(itnim, bfa_fcs_itnim_sm_offline);
		bfa_hal_fcxp_walloc_cancel(itnim->fcs->hal, &itnim->fcxp_wqe);
		bfa_fcs_itnim_free(itnim);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_fcs_itnim_sm_prli(bfa_fcs_itnim_t *itnim, bfa_fcs_itnim_event_t event)
{
	bfa_trc(itnim->fcs, itnim->rport->pwwn);
	bfa_trc(itnim->fcs, event);

	switch (event) {
	case BFA_FCS_ITNIM_SM_RSP_OK:
		bfa_sm_set_state(itnim, bfa_fcs_itnim_sm_hcb_online);
		bfa_hal_itnim_online(itnim->hal_itnim, itnim->seq_rec);
		break;

	case BFA_FCS_ITNIM_SM_RSP_ERROR:
		bfa_sm_set_state(itnim, bfa_fcs_itnim_sm_prli_retry);
		bfa_hal_timer_start(itnim->fcs->hal, &itnim->timer,
				    bfa_fcs_itnim_timeout, itnim,
				    BFA_FCS_RETRY_TIMEOUT);
		break;

	case BFA_FCS_ITNIM_SM_OFFLINE:
		bfa_sm_set_state(itnim, bfa_fcs_itnim_sm_offline);
		bfa_hal_fcxp_discard(itnim->fcxp);
		bfa_fcs_rport_itnim_ack(itnim->rport);
		break;

	case BFA_FCS_ITNIM_SM_INITIATOR:
		bfa_sm_set_state(itnim, bfa_fcs_itnim_sm_initiator);
		bfa_hal_fcxp_discard(itnim->fcxp);
		break;

	case BFA_FCS_ITNIM_SM_DELETE:
		bfa_sm_set_state(itnim, bfa_fcs_itnim_sm_offline);
		bfa_hal_fcxp_discard(itnim->fcxp);
		bfa_fcs_itnim_free(itnim);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_fcs_itnim_sm_prli_retry(bfa_fcs_itnim_t *itnim,
			    bfa_fcs_itnim_event_t event)
{
	bfa_trc(itnim->fcs, itnim->rport->pwwn);
	bfa_trc(itnim->fcs, event);

	switch (event) {
	case BFA_FCS_ITNIM_SM_TIMEOUT:
		bfa_sm_set_state(itnim, bfa_fcs_itnim_sm_prli_send);
		bfa_fcs_itnim_send_prli(itnim, NULL);
		break;

	case BFA_FCS_ITNIM_SM_OFFLINE:
		bfa_sm_set_state(itnim, bfa_fcs_itnim_sm_offline);
		bfa_hal_timer_stop(&itnim->timer);
		bfa_fcs_rport_itnim_ack(itnim->rport);
		break;

	case BFA_FCS_ITNIM_SM_INITIATOR:
		bfa_sm_set_state(itnim, bfa_fcs_itnim_sm_initiator);
		bfa_hal_timer_stop(&itnim->timer);
		break;

	case BFA_FCS_ITNIM_SM_DELETE:
		bfa_sm_set_state(itnim, bfa_fcs_itnim_sm_offline);
		bfa_hal_timer_stop(&itnim->timer);
		bfa_fcs_itnim_free(itnim);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_fcs_itnim_sm_hcb_online(bfa_fcs_itnim_t *itnim,
			    bfa_fcs_itnim_event_t event)
{
	bfa_trc(itnim->fcs, itnim->rport->pwwn);
	bfa_trc(itnim->fcs, event);

	switch (event) {
	case BFA_FCS_ITNIM_SM_HCB_ONLINE:
		bfa_sm_set_state(itnim, bfa_fcs_itnim_sm_online);
		bfa_fcb_itnim_online(itnim->itnim_drv);
		bfa_fcs_itnim_aen_post(itnim, BFA_ITNIM_AEN_ONLINE);
		break;

	case BFA_FCS_ITNIM_SM_OFFLINE:
		bfa_sm_set_state(itnim, bfa_fcs_itnim_sm_offline);
		bfa_hal_itnim_offline(itnim->hal_itnim);
		bfa_fcs_rport_itnim_ack(itnim->rport);
		break;

	case BFA_FCS_ITNIM_SM_DELETE:
		bfa_sm_set_state(itnim, bfa_fcs_itnim_sm_offline);
		bfa_fcs_itnim_free(itnim);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_fcs_itnim_sm_online(bfa_fcs_itnim_t *itnim, bfa_fcs_itnim_event_t event)
{
	bfa_trc(itnim->fcs, itnim->rport->pwwn);
	bfa_trc(itnim->fcs, event);

	switch (event) {
	case BFA_FCS_ITNIM_SM_OFFLINE:
		bfa_sm_set_state(itnim, bfa_fcs_itnim_sm_hcb_offline);
		bfa_fcb_itnim_offline(itnim->itnim_drv);
		bfa_hal_itnim_offline(itnim->hal_itnim);
		if (bfa_fcs_port_is_online(itnim->rport->port) == BFA_TRUE) {
			bfa_fcs_itnim_aen_post(itnim, BFA_ITNIM_AEN_DISCONNECT);
		} else {
			bfa_fcs_itnim_aen_post(itnim, BFA_ITNIM_AEN_OFFLINE);
		}
		break;

	case BFA_FCS_ITNIM_SM_DELETE:
		bfa_sm_set_state(itnim, bfa_fcs_itnim_sm_offline);
		bfa_fcs_itnim_free(itnim);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_fcs_itnim_sm_hcb_offline(bfa_fcs_itnim_t *itnim,
			     bfa_fcs_itnim_event_t event)
{
	bfa_trc(itnim->fcs, itnim->rport->pwwn);
	bfa_trc(itnim->fcs, event);

	switch (event) {
	case BFA_FCS_ITNIM_SM_HCB_OFFLINE:
		bfa_sm_set_state(itnim, bfa_fcs_itnim_sm_offline);
		bfa_fcs_rport_itnim_ack(itnim->rport);
		break;

	case BFA_FCS_ITNIM_SM_DELETE:
		bfa_sm_set_state(itnim, bfa_fcs_itnim_sm_offline);
		bfa_fcs_itnim_free(itnim);
		break;

	default:
		bfa_assert(0);
	}
}

/*
 * This state is set when a discovered rport is also in intiator mode.
 * This ITN is marked as no_op and is not active and will not be truned into
 * online state.
 */
static void
bfa_fcs_itnim_sm_initiator(bfa_fcs_itnim_t *itnim, bfa_fcs_itnim_event_t event)
{
	bfa_trc(itnim->fcs, itnim->rport->pwwn);
	bfa_trc(itnim->fcs, event);

	switch (event) {
	case BFA_FCS_ITNIM_SM_OFFLINE:
		bfa_sm_set_state(itnim, bfa_fcs_itnim_sm_offline);
		bfa_fcs_rport_itnim_ack(itnim->rport);
		break;

	case BFA_FCS_ITNIM_SM_ONLINE:
	case BFA_FCS_ITNIM_SM_INITIATOR:
		break;

	case BFA_FCS_ITNIM_SM_DELETE:
		bfa_sm_set_state(itnim, bfa_fcs_itnim_sm_offline);
		bfa_fcs_itnim_free(itnim);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @}
 */

/**
 * @dg itnim_private FCS ITNIM private interfaces
 * @{
 */

static void
bfa_fcs_itnim_aen_post(bfa_fcs_itnim_t *itnim, bfa_itnim_aen_event_t event)
{
	bfa_fcs_rport_t *rport = itnim->rport;
	bfa_aen_data_t  aen_data;
	bfa_log_mod_t	*logmod = rport->fcs->logm;
	wwn_t		lpwwn = bfa_fcs_port_get_pwwn(rport->port);
	wwn_t		rpwwn = rport->pwwn;
	char		lpwwn_buf[BFA_STRING_32];
	char		rpwwn_buf[BFA_STRING_32];
	char		*lpwwn_ptr;
	char		*rpwwn_ptr;

	/* Don't post events for well known addresses */
	if (bfa_os_ntoh3b(rport->pid) > 0xFFF000) {
		return;
	}

    lpwwn_ptr = wwn2str(lpwwn_buf, sizeof(lpwwn_buf), lpwwn);
    rpwwn_ptr = wwn2str(rpwwn_buf, sizeof(rpwwn_buf), rpwwn);	

	switch (event) {
	case BFA_ITNIM_AEN_NEW:
		bfa_log(logmod, BFAL_AEN_ITNIM_NEW, lpwwn_ptr);
		break;
	case BFA_ITNIM_AEN_DELETE:
		bfa_log(logmod, BFAL_AEN_ITNIM_DELETE, rpwwn_ptr, lpwwn_ptr);
		break;
	case BFA_ITNIM_AEN_ONLINE:
		bfa_log(logmod, BFAL_AEN_ITNIM_ONLINE, rpwwn_ptr, lpwwn_ptr);
		break;
	case BFA_ITNIM_AEN_OFFLINE:
		bfa_log(logmod, BFAL_AEN_ITNIM_OFFLINE, rpwwn_ptr, lpwwn_ptr);
		break;
	case BFA_ITNIM_AEN_DISCONNECT:
		bfa_log(logmod, BFAL_AEN_ITNIM_DISCONNECT, rpwwn_ptr, lpwwn_ptr);
		break;
	default: 
		break;
	}

	aen_data.itnim.vf_id = rport->port->fabric->vf_id;
	aen_data.itnim.lpwwn = lpwwn;
	aen_data.itnim.rpwwn = rpwwn;
	
	bfa_aen_post(rport->fcs->aen, BFA_AEN_CAT_ITNIM, event, &aen_data);
}

static void
bfa_fcs_itnim_send_prli(void *itnim_cbarg, bfa_hal_fcxp_t *fcxp_alloced)
{
	bfa_fcs_itnim_t *itnim = itnim_cbarg;
	bfa_fcs_rport_t *rport = itnim->rport;
	bfa_fcs_port_t *port = rport->port;
	fchs_t          fchs;
	bfa_hal_fcxp_t *fcxp;
	int             len;

	bfa_trc(itnim->fcs, itnim->rport->pwwn);

	fcxp = fcxp_alloced ? fcxp_alloced : bfa_fcs_fcxp_alloc(port->fcs);
	if (!fcxp) {
		itnim->stats.fcxp_alloc_wait++;
		bfa_hal_fcxp_alloc_wait(port->fcs->hal, &itnim->fcxp_wqe,
					bfa_fcs_itnim_send_prli, itnim);
		return;
	}
	itnim->fcxp = fcxp;

	len = fc_prli_build(&fchs, bfa_hal_fcxp_get_reqbuf(fcxp),
			    itnim->rport->pid, bfa_fcs_port_get_fcid(port), 0);

	bfa_hal_fcxp_send(fcxp, rport->hal_rport, port->fabric->vf_id,
			  BFA_FALSE, FC_CLASS_3, len, &fchs,
			  bfa_fcs_itnim_prli_response, (void *)itnim,
			  FC_MAX_PDUSZ, FC_RA_TOV);

	itnim->stats.prli_sent++;
	bfa_sm_send_event(itnim, BFA_FCS_ITNIM_SM_FRMSENT);
}

static void
bfa_fcs_itnim_prli_response(void *fcsarg, bfa_hal_fcxp_t *fcxp, void *cbarg,
			    bfa_status_t req_status, uint32_t rsp_len,
			    uint32_t resid_len, fchs_t *rsp_fchs)
{
	bfa_fcs_itnim_t *itnim = (bfa_fcs_itnim_t *) cbarg;
	fc_els_cmd_t   *els_cmd;
	fc_prli_t      *prli_resp;
	fc_ls_rjt_t    *ls_rjt;
	fc_prli_params_t *sparams;

	bfa_trc(itnim->fcs, req_status);

	/*
	 * Sanity Checks
	 */
	if (req_status != BFA_STATUS_OK) {
		itnim->stats.prli_rsp_err++;
		bfa_sm_send_event(itnim, BFA_FCS_ITNIM_SM_RSP_ERROR);
		return;
	}

	els_cmd = (fc_els_cmd_t *) BFA_FCXP_RSP_PLD(fcxp);

	if (els_cmd->els_code == FC_ELS_ACC) {
		prli_resp = (fc_prli_t *) els_cmd;

		if (fc_prli_rsp_parse(prli_resp, rsp_len) != FC_PARSE_OK) {
			bfa_trc(itnim->fcs, rsp_len);
			/*
			 * Check if this  r-port is also in Initiator mode.
			 * If so, we need to set this ITN as a no-op.
			 */
			if (prli_resp->parampage.servparams.initiator) {
				bfa_trc(itnim->fcs, prli_resp->parampage.type);
				itnim->stats.initiator++;
				itnim->rport->scsi_function = BFA_RPORT_INITIATOR;
				bfa_sm_send_event(itnim,
						  BFA_FCS_ITNIM_SM_INITIATOR);
			}

			itnim->stats.prli_rsp_parse_err++;
			return;
		}
		itnim->rport->scsi_function = BFA_RPORT_TARGET;

		sparams = &prli_resp->parampage.servparams;
		itnim->seq_rec       = sparams->retry;
		itnim->rec_support   = sparams->rec_support;
		itnim->task_retry_id = sparams->task_retry_id;
		itnim->conf_comp     = sparams->confirm;

		itnim->stats.prli_rsp_acc++;
		bfa_sm_send_event(itnim, BFA_FCS_ITNIM_SM_RSP_OK);
	} else {
		ls_rjt = (fc_ls_rjt_t *) BFA_FCXP_RSP_PLD(fcxp);

		bfa_trc(itnim->fcs, ls_rjt->reason_code);
		bfa_trc(itnim->fcs, ls_rjt->reason_code_expl);

		itnim->stats.prli_rsp_rjt++;
		bfa_sm_send_event(itnim, BFA_FCS_ITNIM_SM_RSP_ERROR);
	}
}

static void
bfa_fcs_itnim_timeout(void *arg)
{
	bfa_fcs_itnim_t *itnim = (bfa_fcs_itnim_t *) arg;

	itnim->stats.timeout++;
	bfa_sm_send_event(itnim, BFA_FCS_ITNIM_SM_TIMEOUT);
}

static void
bfa_fcs_itnim_free(bfa_fcs_itnim_t *itnim)
{
	bfa_hal_itnim_delete(itnim->hal_itnim);
	bfa_fcs_itnim_aen_post(itnim, BFA_ITNIM_AEN_DELETE);
	bfa_fcb_itnim_free(itnim->fcs->bfad, itnim->itnim_drv);
}

/**
 * @}
 */

/**
 * @dg itnim_public FCS ITNIM public interfaces
 * @{
 */

/**
 * @brief
 * 	Called by rport when a new rport is created.
 *
 * @param[in] rport	-  remote port.
 */
bfa_fcs_itnim_t *
bfa_fcs_itnim_create(bfa_fcs_rport_t *rport)
{
	bfa_fcs_port_t *port = rport->port;
	bfa_fcs_itnim_t *itnim;
	bfad_itnim_t   *itnim_drv;
	bfa_hal_itnim_t *hal_itnim;

	/*
	 * call bfad to allocate the itnim
	 */
	bfa_fcb_itnim_alloc(port->fcs->bfad, &itnim, &itnim_drv);
	if (itnim == NULL) {
		bfa_trc(port->fcs, rport->pwwn);
		return NULL;
	}

	/*
	 * Initialize itnim
	 */
	itnim->rport = rport;
	itnim->fcs = rport->fcs;
	itnim->itnim_drv = itnim_drv;

	/*
	 * call HAL to create the itnim
	 */
	hal_itnim =
		bfa_hal_itnim_create(port->fcs->hal, rport->hal_rport, itnim);

	if (hal_itnim == NULL) {
		bfa_trc(port->fcs, rport->pwwn);
		bfa_fcb_itnim_free(port->fcs->bfad, itnim_drv);
		bfa_assert(0);
		return NULL;
	}

	itnim->hal_itnim     = hal_itnim;
	itnim->seq_rec       = BFA_FALSE;
	itnim->rec_support   = BFA_FALSE;
	itnim->conf_comp     = BFA_FALSE;
	itnim->task_retry_id = BFA_FALSE;

	/*
	 * Set State machine
	 */
	bfa_sm_set_state(itnim, bfa_fcs_itnim_sm_offline);
	bfa_fcs_itnim_aen_post(itnim, BFA_ITNIM_AEN_NEW);

	return itnim;
}

/**
 * @brief
 * 	Called by rport to delete  the instance of FCPIM.
 *
 * @param[in] rport	-  remote port.
 */
void
bfa_fcs_itnim_delete(bfa_fcs_itnim_t *itnim)
{
	bfa_trc(itnim->fcs, itnim->rport->pid);
	bfa_sm_send_event(itnim, BFA_FCS_ITNIM_SM_DELETE);
}

/**
 * Notification from rport that PLOGI is complete to initiate FC-4 session.
 */
void
bfa_fcs_itnim_rport_online(bfa_fcs_itnim_t *itnim)
{
	itnim->stats.onlines++;
	bfa_sm_send_event(itnim, BFA_FCS_ITNIM_SM_ONLINE);
}

/**
 * Called by rport to handle a remote device offline.
 */
void
bfa_fcs_itnim_rport_offline(bfa_fcs_itnim_t *itnim)
{
	itnim->stats.offlines++;
	bfa_sm_send_event(itnim, BFA_FCS_ITNIM_SM_OFFLINE);
}

/**
 * Called by rport when remote port is known to be an initiator from
 * PRLI received.
 */
void
bfa_fcs_itnim_is_initiator(bfa_fcs_itnim_t *itnim)
{
	bfa_trc(itnim->fcs, itnim->rport->pid);
	itnim->stats.initiator++;
	bfa_sm_send_event(itnim, BFA_FCS_ITNIM_SM_INITIATOR);
}

/**
 * Called by rport to check if the itnim is online.
 */
bfa_status_t
bfa_fcs_itnim_get_online_state(bfa_fcs_itnim_t *itnim)
{
	bfa_trc(itnim->fcs, itnim->rport->pid);
	switch (bfa_sm_to_state(itnim_sm_table, itnim->sm)) {
		case  BFA_ITNIM_ONLINE :
		case  BFA_ITNIM_INITIATIOR :
			return BFA_STATUS_OK;

		default :
			return BFA_STATUS_NO_FCPIM_NEXUS;
			
	}
}

/**
 * HAL completion callback for bfa_hal_itnim_online().
 */
void
bfa_hcb_itnim_online(void *cbarg)
{
	bfa_fcs_itnim_t *itnim = (bfa_fcs_itnim_t *) cbarg;

	bfa_trc(itnim->fcs, itnim->rport->pwwn);
	bfa_sm_send_event(itnim, BFA_FCS_ITNIM_SM_HCB_ONLINE);
}

/**
 * HAL completion callback for bfa_hal_itnim_offline().
 */
void
bfa_hcb_itnim_offline(void *cb_arg)
{
	bfa_fcs_itnim_t *itnim = (bfa_fcs_itnim_t *) cb_arg;

	bfa_trc(itnim->fcs, itnim->rport->pwwn);
	bfa_sm_send_event(itnim, BFA_FCS_ITNIM_SM_HCB_OFFLINE);
}

/**
 * @brief
 * 		HAL notification to FCS/driver for second level error recovery.
 *
 * Atleast one I/O request has timedout and target is unresponsive to
 * repeated abort requests. Second level error recovery should be initiated
 * by starting implicit logout and recovery procedures.
 */
void
bfa_hcb_itnim_sler(void *cb_arg)
{
	bfa_fcs_itnim_t *itnim = (bfa_fcs_itnim_t *) cb_arg;

	itnim->stats.sler++;
	bfa_trc(itnim->fcs, itnim->rport->pwwn);
	bfa_fcs_rport_logo_imp(itnim->rport);
}

bfa_fcs_itnim_t *
bfa_fcs_itnim_lookup(bfa_fcs_port_t *port, wwn_t rpwwn)
{
	bfa_fcs_rport_t *rport;
	rport = bfa_fcs_rport_lookup(port, rpwwn);

	if (!rport)
		return NULL;

	bfa_assert(rport->itnim != NULL);
	return (rport->itnim);
}

bfa_status_t
bfa_fcs_itnim_attr_get(bfa_fcs_port_t *port, wwn_t rpwwn,
		       bfa_itnim_attr_t *attr)
{
	bfa_fcs_itnim_t *itnim = NULL;

	itnim = bfa_fcs_itnim_lookup(port, rpwwn);

	if (itnim == NULL)
		return BFA_STATUS_NO_FCPIM_NEXUS;

	attr->state         = bfa_sm_to_state(itnim_sm_table, itnim->sm);
	attr->retry         = itnim->seq_rec;
	attr->rec_support   = itnim->rec_support;
	attr->conf_comp     = itnim->conf_comp;
	attr->task_retry_id = itnim->task_retry_id;

	return BFA_STATUS_OK;
}

bfa_status_t
bfa_fcs_itnim_stats_get(bfa_fcs_port_t *port, wwn_t rpwwn,
			bfa_itnim_stats_t *stats)
{
	bfa_fcs_itnim_t *itnim = NULL;

	bfa_assert(port != NULL);

	itnim = bfa_fcs_itnim_lookup(port, rpwwn);

	if (itnim == NULL)
		return BFA_STATUS_NO_FCPIM_NEXUS;

	bfa_os_memcpy(stats, &itnim->stats, sizeof(bfa_itnim_stats_t));

	return BFA_STATUS_OK;
}

bfa_status_t
bfa_fcs_itnim_stats_clear(bfa_fcs_port_t *port, wwn_t rpwwn)
{
	bfa_fcs_itnim_t *itnim = NULL;

	bfa_assert(port != NULL);

	itnim = bfa_fcs_itnim_lookup(port, rpwwn);

	if (itnim == NULL)
		return BFA_STATUS_NO_FCPIM_NEXUS;

	bfa_os_memset(&itnim->stats, 0, sizeof(bfa_itnim_stats_t));
	return BFA_STATUS_OK;
}

void
bfa_fcs_fcpim_uf_recv(bfa_fcs_itnim_t *itnim, fchs_t *fchs, uint16_t len)
{
	fc_els_cmd_t   *els_cmd;

	bfa_trc(itnim->fcs, fchs->type);

	if (fchs->type != FC_TYPE_ELS)
		return;

	els_cmd = (fc_els_cmd_t *) (fchs + 1);

	bfa_trc(itnim->fcs, els_cmd->els_code);

	switch (els_cmd->els_code) {
	case FC_ELS_PRLO:
		// bfa_sm_send_event(itnim, BFA_FCS_ITNIM_SM_PRLO);
		break;

	default:
		bfa_assert(0);
	}
}

void
bfa_fcs_itnim_pause(bfa_fcs_itnim_t *itnim)
{
}

void
bfa_fcs_itnim_resume(bfa_fcs_itnim_t *itnim)
{
}

/**
 * @brief
 *   Module initialization
 */
void
bfa_fcs_fcpim_modinit(bfa_fcs_t *fcs)
{
}

/**
 * @brief
 *   Module cleanup
 */
void
bfa_fcs_fcpim_modexit(bfa_fcs_t *fcs)
{
	bfa_fcs_modexit_comp(fcs);
}

/**
 * @}
 */
