/*
 * Copyright (c)  2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

/**
 * @file port_api.c BFA FCS port
 */

/**
 * @brief fdmi_sm_info Port FDMI  State Machine
 *
 */
#include <hal/bfa_hal.h>
#include <hal/bfa_hal_timer.h>
#include <hal/bfa_hal_fcxp.h>
#include <hal/bfa_hal_pport.h>
#include <hcb/bfa_hcb_fcxp.h>
#include <fcs/bfa_fcs_port.h>
#include <fcs/bfa_fcs_fdmi.h>
#include <protocol/fdmi.h>
#include <protocol/fc.h>
#include "fcs_port.h"
#include "fcs_rport.h"
#include "port_priv.h"
#include "fcs_trcmod.h"
#include "fcs_fcxp.h"

BFA_TRC_FILE(FCS, FDMI);

#define BFA_FCS_FDMI_CMD_MAX_RETRIES 2

/*
 * forward declarations
 */
static void     bfa_fcs_port_fdmi_send_rhba(void *fdmi_cbarg,
					    bfa_hal_fcxp_t *fcxp_alloced);
static void     bfa_fcs_port_fdmi_send_rprt(void *fdmi_cbarg,
					    bfa_hal_fcxp_t *fcxp_alloced);
static void     bfa_fcs_port_fdmi_send_rpa(void *fdmi_cbarg,
					   bfa_hal_fcxp_t *fcxp_alloced);
static void     bfa_fcs_port_fdmi_rhba_response(void *fcsarg,
						bfa_hal_fcxp_t *fcxp,
						void *cbarg,
						bfa_status_t req_status,
						uint32_t rsp_len,
						uint32_t resid_len,
						fchs_t *rsp_fchs);
static void     bfa_fcs_port_fdmi_rprt_response(void *fcsarg,
						bfa_hal_fcxp_t *fcxp,
						void *cbarg,
						bfa_status_t req_status,
						uint32_t rsp_len,
						uint32_t resid_len,
						fchs_t *rsp_fchs);
static void     bfa_fcs_port_fdmi_rpa_response(void *fcsarg,
					       bfa_hal_fcxp_t *fcxp,
					       void *cbarg,
					       bfa_status_t req_status,
					       uint32_t rsp_len,
					       uint32_t resid_len,
					       fchs_t *rsp_fchs);
static void     bfa_fcs_port_fdmi_timeout(void *arg);
static uint16_t bfa_fcs_port_fdmi_build_rhba_pyld(bfa_fcs_port_fdmi_t *fdmi,
						  uint8_t *pyld);
static uint16_t bfa_fcs_port_fdmi_build_rprt_pyld(bfa_fcs_port_fdmi_t *fdmi,
						  uint8_t *pyld);
static uint16_t bfa_fcs_port_fdmi_build_rpa_pyld(bfa_fcs_port_fdmi_t *fdmi,
						 uint8_t *pyld);
static uint16_t bfa_fcs_port_fdmi_build_portattr_block(bfa_fcs_port_fdmi_t *
						       fdmi, uint8_t *pyld);
void            bfa_fcs_fdmi_get_hbaattr(bfa_fcs_port_fdmi_t *fdmi,
					 bfa_fcs_fdmi_hba_attr_t *hba_attr);
void            bfa_fcs_fdmi_get_portattr(bfa_fcs_port_fdmi_t *fdmi,
					  bfa_fcs_fdmi_port_attr_t *port_attr);
/**
 * @dg fcs_fdmi_sm FCS FDMI state machine
 * @{
 */

/**
 *  FDMI State Machine events
 */
typedef enum {
	FDMISM_EVENT_PORT_ONLINE = 1,
	FDMISM_EVENT_PORT_OFFLINE = 2,
	FDMISM_EVENT_RSP_OK = 4,
	FDMISM_EVENT_RSP_ERROR = 5,
	FDMISM_EVENT_TIMEOUT = 6,
	FDMISM_EVENT_RHBA_SENT = 7,
	FDMISM_EVENT_RPRT_SENT = 8,
	FDMISM_EVENT_RPA_SENT = 9,
} port_fdmi_event_t;

static void     bfa_fcs_port_fdmi_sm_offline(bfa_fcs_port_fdmi_t *fdmi,
					     port_fdmi_event_t event);
static void     bfa_fcs_port_fdmi_sm_sending_rhba(bfa_fcs_port_fdmi_t *fdmi,
						  port_fdmi_event_t event);
static void     bfa_fcs_port_fdmi_sm_rhba(bfa_fcs_port_fdmi_t *fdmi,
					  port_fdmi_event_t event);
static void     bfa_fcs_port_fdmi_sm_rhba_retry(bfa_fcs_port_fdmi_t *fdmi,
						port_fdmi_event_t event);

static void     bfa_fcs_port_fdmi_sm_sending_rprt(bfa_fcs_port_fdmi_t *fdmi,
						  port_fdmi_event_t event);
static void     bfa_fcs_port_fdmi_sm_rprt(bfa_fcs_port_fdmi_t *fdmi,
					  port_fdmi_event_t event);
static void     bfa_fcs_port_fdmi_sm_rprt_retry(bfa_fcs_port_fdmi_t *fdmi,
						port_fdmi_event_t event);

static void     bfa_fcs_port_fdmi_sm_sending_rpa(bfa_fcs_port_fdmi_t *fdmi,
						 port_fdmi_event_t event);
static void     bfa_fcs_port_fdmi_sm_rpa(bfa_fcs_port_fdmi_t *fdmi,
					 port_fdmi_event_t event);
static void     bfa_fcs_port_fdmi_sm_rpa_retry(bfa_fcs_port_fdmi_t *fdmi,
					       port_fdmi_event_t event);

static void     bfa_fcs_port_fdmi_sm_online(bfa_fcs_port_fdmi_t *fdmi,
					    port_fdmi_event_t event);
/**
 * @brief
 * 		Start in offline state - awaiting MS to send start.
 */
static void
bfa_fcs_port_fdmi_sm_offline(bfa_fcs_port_fdmi_t *fdmi,
			     port_fdmi_event_t event)
{
	bfa_fcs_port_t *port = fdmi->ms->port;

	bfa_trc(port->fcs, port->port_cfg.pwwn);
	bfa_trc(port->fcs, event);

	fdmi->retry_cnt = 0;

	switch (event) {
	case FDMISM_EVENT_PORT_ONLINE:
		if (port->vport) {
			/*
			 * For Vports, register a new port.
			 */
			bfa_sm_set_state(fdmi,
					 bfa_fcs_port_fdmi_sm_sending_rprt);
			bfa_fcs_port_fdmi_send_rprt(fdmi, NULL);
		} else {
			/*
			 * For a base port, we should first register the HBA
			 * atribute. The HBA attribute also contains the base
			 *  port registration.
			 */
			bfa_sm_set_state(fdmi,
					 bfa_fcs_port_fdmi_sm_sending_rhba);
			bfa_fcs_port_fdmi_send_rhba(fdmi, NULL);
		}
		break;

	case FDMISM_EVENT_PORT_OFFLINE:
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_fcs_port_fdmi_sm_sending_rhba(bfa_fcs_port_fdmi_t *fdmi,
				  port_fdmi_event_t event)
{
	bfa_fcs_port_t *port = fdmi->ms->port;

	bfa_trc(port->fcs, port->port_cfg.pwwn);
	bfa_trc(port->fcs, event);

	switch (event) {
	case FDMISM_EVENT_RHBA_SENT:
		bfa_sm_set_state(fdmi, bfa_fcs_port_fdmi_sm_rhba);
		break;

	case FDMISM_EVENT_PORT_OFFLINE:
		bfa_sm_set_state(fdmi, bfa_fcs_port_fdmi_sm_offline);
		bfa_hal_fcxp_walloc_cancel(BFA_FCS_GET_HAL_FROM_PORT(port),
					   &fdmi->fcxp_wqe);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_fcs_port_fdmi_sm_rhba(bfa_fcs_port_fdmi_t *fdmi, port_fdmi_event_t event)
{
	bfa_fcs_port_t *port = fdmi->ms->port;

	bfa_trc(port->fcs, port->port_cfg.pwwn);
	bfa_trc(port->fcs, event);

	switch (event) {
	case FDMISM_EVENT_RSP_ERROR:
		/*
		 * if max retries have not been reached, start timer for a
		 * delayed retry
		 */
		if (fdmi->retry_cnt++ < BFA_FCS_FDMI_CMD_MAX_RETRIES) {
			bfa_sm_set_state(fdmi, bfa_fcs_port_fdmi_sm_rhba_retry);
			bfa_hal_timer_start(BFA_FCS_GET_HAL_FROM_PORT(port),
					    &fdmi->timer,
					    bfa_fcs_port_fdmi_timeout, fdmi,
					    BFA_FCS_RETRY_TIMEOUT);
		} else {
			/*
			 * set state to offline
			 */
			bfa_sm_set_state(fdmi, bfa_fcs_port_fdmi_sm_offline);
		}
		break;

	case FDMISM_EVENT_RSP_OK:
		/*
		 * Initiate Register Port Attributes
		 */
		bfa_sm_set_state(fdmi, bfa_fcs_port_fdmi_sm_sending_rpa);
		fdmi->retry_cnt = 0;
		bfa_fcs_port_fdmi_send_rpa(fdmi, NULL);
		break;

	case FDMISM_EVENT_PORT_OFFLINE:
		bfa_hal_fcxp_discard(fdmi->fcxp);
		bfa_sm_set_state(fdmi, bfa_fcs_port_fdmi_sm_offline);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_fcs_port_fdmi_sm_rhba_retry(bfa_fcs_port_fdmi_t *fdmi,
				port_fdmi_event_t event)
{
	bfa_fcs_port_t *port = fdmi->ms->port;

	bfa_trc(port->fcs, port->port_cfg.pwwn);
	bfa_trc(port->fcs, event);

	switch (event) {
	case FDMISM_EVENT_TIMEOUT:
		/*
		 * Retry Timer Expired. Re-send
		 */
		bfa_sm_set_state(fdmi, bfa_fcs_port_fdmi_sm_sending_rhba);
		bfa_fcs_port_fdmi_send_rhba(fdmi, NULL);
		break;

	case FDMISM_EVENT_PORT_OFFLINE:
		bfa_sm_set_state(fdmi, bfa_fcs_port_fdmi_sm_offline);
		bfa_hal_timer_stop(&fdmi->timer);
		break;

	default:
		bfa_assert(0);
	}
}

/*
* RPRT : Register Port
 */
static void
bfa_fcs_port_fdmi_sm_sending_rprt(bfa_fcs_port_fdmi_t *fdmi,
				  port_fdmi_event_t event)
{
	bfa_fcs_port_t *port = fdmi->ms->port;

	bfa_trc(port->fcs, port->port_cfg.pwwn);
	bfa_trc(port->fcs, event);

	switch (event) {
	case FDMISM_EVENT_RPRT_SENT:
		bfa_sm_set_state(fdmi, bfa_fcs_port_fdmi_sm_rprt);
		break;

	case FDMISM_EVENT_PORT_OFFLINE:
		bfa_sm_set_state(fdmi, bfa_fcs_port_fdmi_sm_offline);
		bfa_hal_fcxp_walloc_cancel(BFA_FCS_GET_HAL_FROM_PORT(port),
					   &fdmi->fcxp_wqe);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_fcs_port_fdmi_sm_rprt(bfa_fcs_port_fdmi_t *fdmi, port_fdmi_event_t event)
{
	bfa_fcs_port_t *port = fdmi->ms->port;

	bfa_trc(port->fcs, port->port_cfg.pwwn);
	bfa_trc(port->fcs, event);

	switch (event) {
	case FDMISM_EVENT_RSP_ERROR:
		/*
		 * if max retries have not been reached, start timer for a
		 * delayed retry
		 */
		if (fdmi->retry_cnt++ < BFA_FCS_FDMI_CMD_MAX_RETRIES) {
			bfa_sm_set_state(fdmi, bfa_fcs_port_fdmi_sm_rprt_retry);
			bfa_hal_timer_start(BFA_FCS_GET_HAL_FROM_PORT(port),
					    &fdmi->timer,
					    bfa_fcs_port_fdmi_timeout, fdmi,
					    BFA_FCS_RETRY_TIMEOUT);

		} else {
			/*
			 * set state to offline
			 */
			bfa_sm_set_state(fdmi, bfa_fcs_port_fdmi_sm_offline);
			fdmi->retry_cnt = 0;
		}
		break;

	case FDMISM_EVENT_RSP_OK:
		fdmi->retry_cnt = 0;
		bfa_sm_set_state(fdmi, bfa_fcs_port_fdmi_sm_online);
		break;

	case FDMISM_EVENT_PORT_OFFLINE:
		bfa_hal_fcxp_discard(fdmi->fcxp);
		bfa_sm_set_state(fdmi, bfa_fcs_port_fdmi_sm_offline);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_fcs_port_fdmi_sm_rprt_retry(bfa_fcs_port_fdmi_t *fdmi,
				port_fdmi_event_t event)
{
	bfa_fcs_port_t *port = fdmi->ms->port;

	bfa_trc(port->fcs, port->port_cfg.pwwn);
	bfa_trc(port->fcs, event);

	switch (event) {
	case FDMISM_EVENT_TIMEOUT:
		/*
		 * Retry Timer Expired. Re-send
		 */
		bfa_sm_set_state(fdmi, bfa_fcs_port_fdmi_sm_sending_rprt);
		bfa_fcs_port_fdmi_send_rprt(fdmi, NULL);
		break;

	case FDMISM_EVENT_PORT_OFFLINE:
		bfa_sm_set_state(fdmi, bfa_fcs_port_fdmi_sm_offline);
		bfa_hal_timer_stop(&fdmi->timer);
		break;

	default:
		bfa_assert(0);
	}
}

/*
 * Register Port Attributes
 */
static void
bfa_fcs_port_fdmi_sm_sending_rpa(bfa_fcs_port_fdmi_t *fdmi,
				 port_fdmi_event_t event)
{
	bfa_fcs_port_t *port = fdmi->ms->port;

	bfa_trc(port->fcs, port->port_cfg.pwwn);
	bfa_trc(port->fcs, event);

	switch (event) {
	case FDMISM_EVENT_RPA_SENT:
		bfa_sm_set_state(fdmi, bfa_fcs_port_fdmi_sm_rpa);
		break;

	case FDMISM_EVENT_PORT_OFFLINE:
		bfa_sm_set_state(fdmi, bfa_fcs_port_fdmi_sm_offline);
		bfa_hal_fcxp_walloc_cancel(BFA_FCS_GET_HAL_FROM_PORT(port),
					   &fdmi->fcxp_wqe);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_fcs_port_fdmi_sm_rpa(bfa_fcs_port_fdmi_t *fdmi, port_fdmi_event_t event)
{
	bfa_fcs_port_t *port = fdmi->ms->port;

	bfa_trc(port->fcs, port->port_cfg.pwwn);
	bfa_trc(port->fcs, event);

	switch (event) {
	case FDMISM_EVENT_RSP_ERROR:
		/*
		 * if max retries have not been reached, start timer for a
		 * delayed retry
		 */
		if (fdmi->retry_cnt++ < BFA_FCS_FDMI_CMD_MAX_RETRIES) {
			bfa_sm_set_state(fdmi, bfa_fcs_port_fdmi_sm_rpa_retry);
			bfa_hal_timer_start(BFA_FCS_GET_HAL_FROM_PORT(port),
					    &fdmi->timer,
					    bfa_fcs_port_fdmi_timeout, fdmi,
					    BFA_FCS_RETRY_TIMEOUT);
		} else {
			/*
			 * set state to offline
			 */
			bfa_sm_set_state(fdmi, bfa_fcs_port_fdmi_sm_offline);
			fdmi->retry_cnt = 0;
		}
		break;

	case FDMISM_EVENT_RSP_OK:
		bfa_sm_set_state(fdmi, bfa_fcs_port_fdmi_sm_online);
		fdmi->retry_cnt = 0;
		break;

	case FDMISM_EVENT_PORT_OFFLINE:
		bfa_hal_fcxp_discard(fdmi->fcxp);
		bfa_sm_set_state(fdmi, bfa_fcs_port_fdmi_sm_offline);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_fcs_port_fdmi_sm_rpa_retry(bfa_fcs_port_fdmi_t *fdmi,
			       port_fdmi_event_t event)
{
	bfa_fcs_port_t *port = fdmi->ms->port;

	bfa_trc(port->fcs, port->port_cfg.pwwn);
	bfa_trc(port->fcs, event);

	switch (event) {
	case FDMISM_EVENT_TIMEOUT:
		/*
		 * Retry Timer Expired. Re-send
		 */
		bfa_sm_set_state(fdmi, bfa_fcs_port_fdmi_sm_sending_rpa);
		bfa_fcs_port_fdmi_send_rpa(fdmi, NULL);
		break;

	case FDMISM_EVENT_PORT_OFFLINE:
		bfa_sm_set_state(fdmi, bfa_fcs_port_fdmi_sm_offline);
		bfa_hal_timer_stop(&fdmi->timer);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_fcs_port_fdmi_sm_online(bfa_fcs_port_fdmi_t *fdmi, port_fdmi_event_t event)
{
	bfa_fcs_port_t *port = fdmi->ms->port;

	bfa_trc(port->fcs, port->port_cfg.pwwn);
	bfa_trc(port->fcs, event);

	switch (event) {
	case FDMISM_EVENT_PORT_OFFLINE:
		bfa_sm_set_state(fdmi, bfa_fcs_port_fdmi_sm_offline);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @}
 */
/**
*  RHBA : Register HBA Attributes.
 */
static void
bfa_fcs_port_fdmi_send_rhba(void *fdmi_cbarg, bfa_hal_fcxp_t *fcxp_alloced)
{
	bfa_fcs_port_fdmi_t *fdmi = fdmi_cbarg;
	bfa_fcs_port_t *port = fdmi->ms->port;
	fchs_t          fchs;
	int             len, attr_len;
	bfa_hal_fcxp_t *fcxp;
	uint8_t        *pyld;

	bfa_trc(port->fcs, port->port_cfg.pwwn);

fcxp = fcxp_alloced ? fcxp_alloced : bfa_fcs_fcxp_alloc(port->fcs);
	if (!fcxp) {
		bfa_hal_fcxp_alloc_wait(port->fcs->hal, &fdmi->fcxp_wqe,
					bfa_fcs_port_fdmi_send_rhba, fdmi);
		return;
	}
	fdmi->fcxp = fcxp;

	pyld = bfa_hal_fcxp_get_reqbuf(fcxp);
	bfa_os_memset(pyld, 0, FC_MAX_PDUSZ);

	len = fc_fdmi_reqhdr_build(&fchs, pyld, bfa_fcs_port_get_fcid(port),
				   FDMI_RHBA);

	attr_len =
		bfa_fcs_port_fdmi_build_rhba_pyld(fdmi,
						  (uint8_t *) ((ct_hdr_t *) pyld
							       + 1));

	bfa_hal_fcxp_send(fcxp, NULL, port->fabric->vf_id, BFA_FALSE,
			  FC_CLASS_3, (len + attr_len), &fchs,
			  bfa_fcs_port_fdmi_rhba_response, (void *)fdmi,
			  FC_MAX_PDUSZ, FC_RA_TOV);

	bfa_sm_send_event(fdmi, FDMISM_EVENT_RHBA_SENT);
}

static          uint16_t
bfa_fcs_port_fdmi_build_rhba_pyld(bfa_fcs_port_fdmi_t *fdmi, uint8_t *pyld)
{
	bfa_fcs_port_t *port = fdmi->ms->port;
	bfa_fcs_fdmi_hba_attr_t hba_attr;	/* @todo */
	bfa_fcs_fdmi_hba_attr_t *fcs_hba_attr = &hba_attr;	/* @todo */
	fdmi_rhba_t    *rhba = (fdmi_rhba_t *) pyld;
	fdmi_attr_t    *attr;
	uint8_t        *curr_ptr;
	uint16_t        len;

	/*
	 * get hba attributes
	 */
bfa_fcs_fdmi_get_hbaattr(fdmi, fcs_hba_attr);	/* @todo : need to fix
							 *this */

	rhba->hba_id = bfa_fcs_port_get_pwwn(port);
	rhba->port_list.num_ports = bfa_os_htonl(1);
	rhba->port_list.port_entry = bfa_fcs_port_get_pwwn(port);

	len = sizeof(rhba->hba_id) + sizeof(rhba->port_list);

	rhba->hba_attr_blk.attr_count =
		bfa_os_htonl(FDMI_HBA_ATTRIB_MAX_TYPE - 1);
	len += sizeof(rhba->hba_attr_blk.attr_count);

	/*
	 * fill out the invididual entries of the HBA attrib Block
	 */
	curr_ptr = (uint8_t *) &rhba->hba_attr_blk.hba_attr;

	/*
	 * Node Name
	 */
	attr = (fdmi_attr_t *) curr_ptr;
	attr->type = bfa_os_htons(FDMI_HBA_ATTRIB_NODENAME);
	attr->len = sizeof(wwn_t);
	memcpy(attr->value, &bfa_fcs_port_get_nwwn(port), attr->len);
	curr_ptr += sizeof(attr->type) + sizeof(attr->len) + attr->len;
	len += attr->len;
	attr->len =
		bfa_os_htons(attr->len + sizeof(attr->type) +
			     sizeof(attr->len));

	/*
	 * Manufacturer
	 */
	attr = (fdmi_attr_t *) curr_ptr;
	attr->type = bfa_os_htons(FDMI_HBA_ATTRIB_MANUFACTURER);
	attr->len = (uint16_t) strlen(fcs_hba_attr->manufacturer);
	memcpy(attr->value, fcs_hba_attr->manufacturer, attr->len);
	attr->len = fc_roundup(attr->len, sizeof(uint32_t));	/* variable
								 *fields need
								 *to be 4 byte
								 *aligned */
	curr_ptr += sizeof(attr->type) + sizeof(attr->len) + attr->len;
	len += attr->len;
	attr->len =
		bfa_os_htons(attr->len + sizeof(attr->type) +
			     sizeof(attr->len));

	/*
	 * Serial Number
	 */
	attr = (fdmi_attr_t *) curr_ptr;
	attr->type = bfa_os_htons(FDMI_HBA_ATTRIB_SERIALNUM);
	attr->len = (uint16_t) strlen(fcs_hba_attr->serial_num);
	memcpy(attr->value, fcs_hba_attr->serial_num, attr->len);
	attr->len = fc_roundup(attr->len, sizeof(uint32_t));	/* variable
								 *fields need
								 *to be 4 byte
								 *aligned */
	curr_ptr += sizeof(attr->type) + sizeof(attr->len) + attr->len;
	len += attr->len;
	attr->len =
		bfa_os_htons(attr->len + sizeof(attr->type) +
			     sizeof(attr->len));

	/*
	 * Model
	 */
	attr = (fdmi_attr_t *) curr_ptr;
	attr->type = bfa_os_htons(FDMI_HBA_ATTRIB_MODEL);
	attr->len = (uint16_t) strlen(fcs_hba_attr->model);
	memcpy(attr->value, fcs_hba_attr->model, attr->len);
	attr->len = fc_roundup(attr->len, sizeof(uint32_t));	/* variable
								 *fields need
								 *to be 4 byte
								 *aligned */
	curr_ptr += sizeof(attr->type) + sizeof(attr->len) + attr->len;
	len += attr->len;
	attr->len =
		bfa_os_htons(attr->len + sizeof(attr->type) +
			     sizeof(attr->len));

	/*
	 * Model Desc
	 */
	attr = (fdmi_attr_t *) curr_ptr;
	attr->type = bfa_os_htons(FDMI_HBA_ATTRIB_MODEL_DESC);
	attr->len = (uint16_t) strlen(fcs_hba_attr->model_desc);
	memcpy(attr->value, fcs_hba_attr->model_desc, attr->len);
	attr->len = fc_roundup(attr->len, sizeof(uint32_t));	/* variable
								 *fields need
								 *to be 4 byte
								 *aligned */
	curr_ptr += sizeof(attr->type) + sizeof(attr->len) + attr->len;
	len += attr->len;
	attr->len =
		bfa_os_htons(attr->len + sizeof(attr->type) +
			     sizeof(attr->len));

	/*
	 * H/W Version
	 */
	attr = (fdmi_attr_t *) curr_ptr;
	attr->type = bfa_os_htons(FDMI_HBA_ATTRIB_HW_VERSION);
	attr->len = (uint16_t) strlen(fcs_hba_attr->hw_version);
	memcpy(attr->value, fcs_hba_attr->hw_version, attr->len);
	attr->len = fc_roundup(attr->len, sizeof(uint32_t));	/* variable
								 *fields need
								 *to be 4 byte
								 *aligned */
	curr_ptr += sizeof(attr->type) + sizeof(attr->len) + attr->len;
	len += attr->len;
	attr->len =
		bfa_os_htons(attr->len + sizeof(attr->type) +
			     sizeof(attr->len));

	/*
	 * Driver Version
	 */
	attr = (fdmi_attr_t *) curr_ptr;
	attr->type = bfa_os_htons(FDMI_HBA_ATTRIB_DRIVER_VERSION);
	attr->len = (uint16_t) strlen(fcs_hba_attr->driver_version);
	memcpy(attr->value, fcs_hba_attr->driver_version, attr->len);
	attr->len = fc_roundup(attr->len, sizeof(uint32_t));	/* variable
								 *fields need
								 *to be 4 byte
								 *aligned */
	curr_ptr += sizeof(attr->type) + sizeof(attr->len) + attr->len;
	len += attr->len;;
	attr->len =
		bfa_os_htons(attr->len + sizeof(attr->type) +
			     sizeof(attr->len));

	/*
	 * Option Rom Version
	 */
	attr = (fdmi_attr_t *) curr_ptr;
	attr->type = bfa_os_htons(FDMI_HBA_ATTRIB_ROM_VERSION);
	attr->len = (uint16_t) strlen(fcs_hba_attr->option_rom_ver);
	memcpy(attr->value, fcs_hba_attr->option_rom_ver, attr->len);
	attr->len = fc_roundup(attr->len, sizeof(uint32_t));	/* variable
								 *fields need
								 *to be 4 byte
								 *aligned */
	curr_ptr += sizeof(attr->type) + sizeof(attr->len) + attr->len;
	len += attr->len;
	attr->len =
		bfa_os_htons(attr->len + sizeof(attr->type) +
			     sizeof(attr->len));

	/*
	 * f/w Version
	 */
	attr = (fdmi_attr_t *) curr_ptr;
	attr->type = bfa_os_htons(FDMI_HBA_ATTRIB_FW_VERSION);
	attr->len = (uint16_t) strlen(fcs_hba_attr->fw_version);
	memcpy(attr->value, fcs_hba_attr->fw_version, attr->len);
	attr->len = fc_roundup(attr->len, sizeof(uint32_t));	/* variable
								 *fields need
								 *to be 4 byte
								 *aligned */
	curr_ptr += sizeof(attr->type) + sizeof(attr->len) + attr->len;
	len += attr->len;
	attr->len =
		bfa_os_htons(attr->len + sizeof(attr->type) +
			     sizeof(attr->len));

	/*
	 * OS Name
	 */
	attr = (fdmi_attr_t *) curr_ptr;
	attr->type = bfa_os_htons(FDMI_HBA_ATTRIB_OS_NAME);
	attr->len = (uint16_t) strlen(fcs_hba_attr->os_name);
	memcpy(attr->value, fcs_hba_attr->os_name, attr->len);
	attr->len = fc_roundup(attr->len, sizeof(uint32_t));	/* variable
								 *fields need
								 *to be 4 byte
								 *aligned */
	curr_ptr += sizeof(attr->type) + sizeof(attr->len) + attr->len;
	len += attr->len;
	attr->len =
		bfa_os_htons(attr->len + sizeof(attr->type) +
			     sizeof(attr->len));

	/*
	 * MAX_CT_PAYLOAD
	 */
	attr = (fdmi_attr_t *) curr_ptr;
	attr->type = bfa_os_htons(FDMI_HBA_ATTRIB_MAX_CT);
	attr->len = sizeof(fcs_hba_attr->max_ct_pyld);
	memcpy(attr->value, &fcs_hba_attr->max_ct_pyld, attr->len);
	len += attr->len;
	attr->len =
		bfa_os_htons(attr->len + sizeof(attr->type) +
			     sizeof(attr->len));

	/*
	 * Update size of payload
	 */
	len += ((sizeof(attr->type) +
		 sizeof(attr->len)) * (FDMI_HBA_ATTRIB_MAX_TYPE - 1));
	return len;
}

static void
bfa_fcs_port_fdmi_rhba_response(void *fcsarg, bfa_hal_fcxp_t *fcxp,
				void *cbarg, bfa_status_t req_status,
				uint32_t rsp_len, uint32_t resid_len,
				fchs_t *rsp_fchs)
{
	bfa_fcs_port_fdmi_t *fdmi = (bfa_fcs_port_fdmi_t *) cbarg;
	bfa_fcs_port_t *port = fdmi->ms->port;
	ct_hdr_t       *cthdr = NULL;

	bfa_trc(port->fcs, port->port_cfg.pwwn);

	/*
	 * Sanity Checks
	 */
	if (req_status != BFA_STATUS_OK) {
		bfa_trc(port->fcs, req_status);
		bfa_sm_send_event(fdmi, FDMISM_EVENT_RSP_ERROR);
		return;
	}

	cthdr = (ct_hdr_t *) BFA_FCXP_RSP_PLD(fcxp);
	cthdr->cmd_rsp_code = bfa_os_ntohs(cthdr->cmd_rsp_code);

	if (cthdr->cmd_rsp_code == CT_RSP_ACCEPT) {
		bfa_sm_send_event(fdmi, FDMISM_EVENT_RSP_OK);
		return;
	}

	bfa_trc(port->fcs, cthdr->reason_code);
	bfa_trc(port->fcs, cthdr->exp_code);
	bfa_sm_send_event(fdmi, FDMISM_EVENT_RSP_ERROR);
}

/**
*  RPRT : Register Port
 */
static void
bfa_fcs_port_fdmi_send_rprt(void *fdmi_cbarg, bfa_hal_fcxp_t *fcxp_alloced)
{
	bfa_fcs_port_fdmi_t *fdmi = fdmi_cbarg;
	bfa_fcs_port_t *port = fdmi->ms->port;
	fchs_t          fchs;
	uint16_t        len, attr_len;
	bfa_hal_fcxp_t *fcxp;
	uint8_t        *pyld;

	bfa_trc(port->fcs, port->port_cfg.pwwn);

fcxp = fcxp_alloced ? fcxp_alloced : bfa_fcs_fcxp_alloc(port->fcs);
	if (!fcxp) {
		bfa_hal_fcxp_alloc_wait(port->fcs->hal, &fdmi->fcxp_wqe,
					bfa_fcs_port_fdmi_send_rprt, fdmi);
		return;
	}
	fdmi->fcxp = fcxp;

	pyld = bfa_hal_fcxp_get_reqbuf(fcxp);
	bfa_os_memset(pyld, 0, FC_MAX_PDUSZ);

	len = fc_fdmi_reqhdr_build(&fchs, pyld, bfa_fcs_port_get_fcid(port),
				   FDMI_RPRT);

	attr_len =
		bfa_fcs_port_fdmi_build_rprt_pyld(fdmi,
						  (uint8_t *) ((ct_hdr_t *) pyld
							       + 1));

	bfa_hal_fcxp_send(fcxp, NULL, port->fabric->vf_id, BFA_FALSE,
			  FC_CLASS_3, len + attr_len, &fchs,
			  bfa_fcs_port_fdmi_rprt_response, (void *)fdmi,
			  FC_MAX_PDUSZ, FC_RA_TOV);

	bfa_sm_send_event(fdmi, FDMISM_EVENT_RPRT_SENT);
}

/**
 * This routine builds Port Attribute Block that used in RPA, RPRT commands.
 */
static          uint16_t
bfa_fcs_port_fdmi_build_portattr_block(bfa_fcs_port_fdmi_t *fdmi,
				       uint8_t *pyld)
{
	bfa_fcs_fdmi_port_attr_t fcs_port_attr;
	fdmi_port_attr_t *port_attrib = (fdmi_port_attr_t *) pyld;
	fdmi_attr_t    *attr;
	uint8_t        *curr_ptr;
	uint16_t        len;

	/*
	 * get port attributes
	 */
	bfa_fcs_fdmi_get_portattr(fdmi, &fcs_port_attr);

	port_attrib->attr_count = bfa_os_htonl(FDMI_PORT_ATTR_MAX_TYPE - 1);
	len = sizeof(port_attrib->attr_count);

	/*
	 * fill out the invididual entries
	 */
	curr_ptr = (uint8_t *) &port_attrib->port_attr;

	/*
	 * FC4 Types
	 */
	attr = (fdmi_attr_t *) curr_ptr;
	attr->type = bfa_os_htons(FDMI_PORT_ATTRIB_FC4_TYPES);
	attr->len = sizeof(fcs_port_attr.supp_fc4_types);
	memcpy(attr->value, fcs_port_attr.supp_fc4_types, attr->len);
	curr_ptr += sizeof(attr->type) + sizeof(attr->len) + attr->len;
	len += attr->len;
	attr->len =
		bfa_os_htons(attr->len + sizeof(attr->type) +
			     sizeof(attr->len));

	/*
	 * Supported Speed
	 */
	attr = (fdmi_attr_t *) curr_ptr;
	attr->type = bfa_os_htons(FDMI_PORT_ATTRIB_SUPP_SPEED);
	attr->len = sizeof(fcs_port_attr.supp_speed);
	memcpy(attr->value, &fcs_port_attr.supp_speed, attr->len);
	curr_ptr += sizeof(attr->type) + sizeof(attr->len) + attr->len;
	len += attr->len;
	attr->len =
		bfa_os_htons(attr->len + sizeof(attr->type) +
			     sizeof(attr->len));

	/*
	 * current Port Speed
	 */
	attr = (fdmi_attr_t *) curr_ptr;
	attr->type = bfa_os_htons(FDMI_PORT_ATTRIB_PORT_SPEED);
	attr->len = sizeof(fcs_port_attr.curr_speed);
	memcpy(attr->value, &fcs_port_attr.curr_speed, attr->len);
	curr_ptr += sizeof(attr->type) + sizeof(attr->len) + attr->len;
	len += attr->len;
	attr->len =
		bfa_os_htons(attr->len + sizeof(attr->type) +
			     sizeof(attr->len));

	/*
	 * max frame size
	 */
	attr = (fdmi_attr_t *) curr_ptr;
	attr->type = bfa_os_htons(FDMI_PORT_ATTRIB_FRAME_SIZE);
	attr->len = sizeof(fcs_port_attr.max_frm_size);
	memcpy(attr->value, &fcs_port_attr.max_frm_size, attr->len);
	curr_ptr += sizeof(attr->type) + sizeof(attr->len) + attr->len;
	len += attr->len;
	attr->len =
		bfa_os_htons(attr->len + sizeof(attr->type) +
			     sizeof(attr->len));

	/*
	 * OS Device Name
	 */
	attr = (fdmi_attr_t *) curr_ptr;
	attr->type = bfa_os_htons(FDMI_PORT_ATTRIB_DEV_NAME);
	attr->len = (uint16_t) strlen(fcs_port_attr.os_device_name);
	memcpy(attr->value, fcs_port_attr.os_device_name, attr->len);
	attr->len = fc_roundup(attr->len, sizeof(uint32_t));	/* variable
								 *fields need
								 *to be 4 byte
								 *aligned */
	curr_ptr += sizeof(attr->type) + sizeof(attr->len) + attr->len;
	len += attr->len;
	attr->len =
		bfa_os_htons(attr->len + sizeof(attr->type) +
			     sizeof(attr->len));

	/*
	 * Host Name
	 */
	attr = (fdmi_attr_t *) curr_ptr;
	attr->type = bfa_os_htons(FDMI_PORT_ATTRIB_HOST_NAME);
	attr->len = (uint16_t) strlen(fcs_port_attr.host_name);
	memcpy(attr->value, fcs_port_attr.host_name, attr->len);
	attr->len = fc_roundup(attr->len, sizeof(uint32_t));	/* variable
								 *fields need
								 *to be 4 byte
								 *aligned */
	curr_ptr += sizeof(attr->type) + sizeof(attr->len) + attr->len;
	len += attr->len;
	attr->len =
		bfa_os_htons(attr->len + sizeof(attr->type) +
			     sizeof(attr->len));

	/*
	 * Update size of payload
	 */
	len += ((sizeof(attr->type) +
		 sizeof(attr->len)) * (FDMI_PORT_ATTR_MAX_TYPE - 1));
	return len;
}

static          uint16_t
bfa_fcs_port_fdmi_build_rprt_pyld(bfa_fcs_port_fdmi_t *fdmi, uint8_t *pyld)
{
	bfa_fcs_port_t *port = fdmi->ms->port;
	fdmi_rprt_t    *rprt = (fdmi_rprt_t *) pyld;
	uint16_t        len;

	rprt->hba_id = bfa_fcs_port_get_pwwn(bfa_fcs_get_base_port(port->fcs));
	rprt->port_name = bfa_fcs_port_get_pwwn(port);

	len = bfa_fcs_port_fdmi_build_portattr_block(fdmi,
				(uint8_t *) &rprt->port_attr_blk);

	len += sizeof(rprt->hba_id) + sizeof(rprt->port_name);

	return len;
}

static void
bfa_fcs_port_fdmi_rprt_response(void *fcsarg, bfa_hal_fcxp_t *fcxp,
				void *cbarg, bfa_status_t req_status,
				uint32_t rsp_len, uint32_t resid_len,
				fchs_t *rsp_fchs)
{
	bfa_fcs_port_fdmi_t *fdmi = (bfa_fcs_port_fdmi_t *) cbarg;
	bfa_fcs_port_t *port = fdmi->ms->port;
	ct_hdr_t       *cthdr = NULL;

	bfa_trc(port->fcs, port->port_cfg.pwwn);

	/*
	 * Sanity Checks
	 */
	if (req_status != BFA_STATUS_OK) {
		bfa_trc(port->fcs, req_status);
		bfa_sm_send_event(fdmi, FDMISM_EVENT_RSP_ERROR);
		return;
	}

	cthdr = (ct_hdr_t *) BFA_FCXP_RSP_PLD(fcxp);
	cthdr->cmd_rsp_code = bfa_os_ntohs(cthdr->cmd_rsp_code);

	if (cthdr->cmd_rsp_code == CT_RSP_ACCEPT) {
		bfa_sm_send_event(fdmi, FDMISM_EVENT_RSP_OK);
		return;
	}

	bfa_trc(port->fcs, cthdr->reason_code);
	bfa_trc(port->fcs, cthdr->exp_code);
	bfa_sm_send_event(fdmi, FDMISM_EVENT_RSP_ERROR);
}

/**
*  RPA : Register Port Attributes.
 */
static void
bfa_fcs_port_fdmi_send_rpa(void *fdmi_cbarg, bfa_hal_fcxp_t *fcxp_alloced)
{
	bfa_fcs_port_fdmi_t *fdmi = fdmi_cbarg;
	bfa_fcs_port_t *port = fdmi->ms->port;
	fchs_t          fchs;
	uint16_t        len, attr_len;
	bfa_hal_fcxp_t *fcxp;
	uint8_t        *pyld;

	bfa_trc(port->fcs, port->port_cfg.pwwn);

fcxp = fcxp_alloced ? fcxp_alloced : bfa_fcs_fcxp_alloc(port->fcs);
	if (!fcxp) {
		bfa_hal_fcxp_alloc_wait(port->fcs->hal, &fdmi->fcxp_wqe,
					bfa_fcs_port_fdmi_send_rpa, fdmi);
		return;
	}
	fdmi->fcxp = fcxp;

	pyld = bfa_hal_fcxp_get_reqbuf(fcxp);
	bfa_os_memset(pyld, 0, FC_MAX_PDUSZ);

	len = fc_fdmi_reqhdr_build(&fchs, pyld, bfa_fcs_port_get_fcid(port),
				   FDMI_RPA);

	attr_len =
		bfa_fcs_port_fdmi_build_rpa_pyld(fdmi,
						 (uint8_t *) ((ct_hdr_t *) pyld
							      + 1));

	bfa_hal_fcxp_send(fcxp, NULL, port->fabric->vf_id, BFA_FALSE,
			  FC_CLASS_3, len + attr_len, &fchs,
			  bfa_fcs_port_fdmi_rpa_response, (void *)fdmi,
			  FC_MAX_PDUSZ, FC_RA_TOV);

	bfa_sm_send_event(fdmi, FDMISM_EVENT_RPA_SENT);
}

static          uint16_t
bfa_fcs_port_fdmi_build_rpa_pyld(bfa_fcs_port_fdmi_t *fdmi, uint8_t *pyld)
{
	bfa_fcs_port_t *port = fdmi->ms->port;
	fdmi_rpa_t     *rpa = (fdmi_rpa_t *) pyld;
	uint16_t        len;

	rpa->port_name = bfa_fcs_port_get_pwwn(port);

	len = bfa_fcs_port_fdmi_build_portattr_block(fdmi,
				(uint8_t *) &rpa->port_attr_blk);

	len += sizeof(rpa->port_name);

	return len;
}

static void
bfa_fcs_port_fdmi_rpa_response(void *fcsarg, bfa_hal_fcxp_t *fcxp, void *cbarg,
			       bfa_status_t req_status, uint32_t rsp_len,
			       uint32_t resid_len, fchs_t *rsp_fchs)
{
	bfa_fcs_port_fdmi_t *fdmi = (bfa_fcs_port_fdmi_t *) cbarg;
	bfa_fcs_port_t *port = fdmi->ms->port;
	ct_hdr_t       *cthdr = NULL;

	bfa_trc(port->fcs, port->port_cfg.pwwn);

	/*
	 * Sanity Checks
	 */
	if (req_status != BFA_STATUS_OK) {
		bfa_trc(port->fcs, req_status);
		bfa_sm_send_event(fdmi, FDMISM_EVENT_RSP_ERROR);
		return;
	}

	cthdr = (ct_hdr_t *) BFA_FCXP_RSP_PLD(fcxp);
	cthdr->cmd_rsp_code = bfa_os_ntohs(cthdr->cmd_rsp_code);

	if (cthdr->cmd_rsp_code == CT_RSP_ACCEPT) {
		bfa_sm_send_event(fdmi, FDMISM_EVENT_RSP_OK);
		return;
	}

	bfa_trc(port->fcs, cthdr->reason_code);
	bfa_trc(port->fcs, cthdr->exp_code);
	bfa_sm_send_event(fdmi, FDMISM_EVENT_RSP_ERROR);
}

static void
bfa_fcs_port_fdmi_timeout(void *arg)
{
	bfa_fcs_port_fdmi_t *fdmi = (bfa_fcs_port_fdmi_t *) arg;

	bfa_sm_send_event(fdmi, FDMISM_EVENT_TIMEOUT);
}

/*
 * @todo
 *  These Values will be eventually got from Driver or mgmt interface.
 */

#define BFA_FCS_FDMI_MODEL 	  	   "BFA-8xxx"
#define BFA_FCS_FDMI_MODELDESC 	   "Brocade Fibre Channel 8 Gig Adapter"
#define BFA_FCS_FDMI_HW_VER 	   "v-1.0.0"
#define BFA_FCS_FDMI_DRIVER_VER	   "v-1.0.0"
#define BFA_FCS_FDMI_ROM_VER 	   "v-1.0.0"
#define BFA_FCS_FDMI_FW_VER 	   "v-1.0.0"
#define BFA_FCS_FDMI_OS_NAME 		"TBD : Host Operating System Name"

void
bfa_fcs_fdmi_get_hbaattr(bfa_fcs_port_fdmi_t *fdmi,
			 bfa_fcs_fdmi_hba_attr_t *hba_attr)
{
	bfa_fcs_port_t *port = fdmi->ms->port;
	bfa_ioc_attr_t  ioc_attr;

	bfa_os_memset(hba_attr, 0, sizeof(bfa_fcs_fdmi_hba_attr_t));

	bfa_hal_ioc_get_attr(port->fcs->hal, &ioc_attr);

	strncpy(hba_attr->manufacturer, ioc_attr.adapter_attr.manufacturer,
		sizeof(ioc_attr.adapter_attr.manufacturer));

	strncpy(hba_attr->serial_num, ioc_attr.adapter_attr.serial_num,
		sizeof(ioc_attr.adapter_attr.serial_num));

	strncpy(hba_attr->model, BFA_FCS_FDMI_MODEL,
		sizeof(BFA_FCS_FDMI_MODEL));

	strncpy(hba_attr->model_desc, BFA_FCS_FDMI_MODELDESC,
		sizeof(BFA_FCS_FDMI_MODELDESC));

	strncpy(hba_attr->hw_version, BFA_FCS_FDMI_HW_VER,
		sizeof(BFA_FCS_FDMI_HW_VER));

	strncpy(hba_attr->driver_version, BFA_FCS_FDMI_DRIVER_VER,
		sizeof(BFA_FCS_FDMI_DRIVER_VER));

	strncpy(hba_attr->option_rom_ver, BFA_FCS_FDMI_ROM_VER,
		sizeof(BFA_FCS_FDMI_ROM_VER));

	strncpy(hba_attr->fw_version, BFA_FCS_FDMI_FW_VER,
		sizeof(BFA_FCS_FDMI_FW_VER));

	strncpy(hba_attr->os_name, BFA_FCS_FDMI_OS_NAME,
		sizeof(BFA_FCS_FDMI_OS_NAME));

	hba_attr->max_ct_pyld = bfa_os_htonl(FC_MAX_PDUSZ);
}

#define BFA_FCS_FDMI_HOST_NAME 			"TBD : Host Machine name"
#define BFA_FCS_FDMI_OS_DEVICE_NAME 	"TBD : OS device Name"

/*
 * @todo
*   Host Name, OS device Name :
 *       These Values will be eventually got from Driver or mgmt interface.
 */
void
bfa_fcs_fdmi_get_portattr(bfa_fcs_port_fdmi_t *fdmi,
			  bfa_fcs_fdmi_port_attr_t *port_attr)
{
	bfa_fcs_port_t *port = fdmi->ms->port;
	bfa_pport_attr_t pport_attr;

	bfa_os_memset(port_attr, 0, sizeof(bfa_fcs_fdmi_port_attr_t));

	/*
	 * get pport attributes from hal
	 */
	bfa_hal_pport_get_attr(port->fcs->hal, &pport_attr);

	/*
	 * get FC4 type Bitmask
	 */
	fc_get_fc4type_bitmask(FC_TYPE_FCP, port_attr->supp_fc4_types);

	/*
	 * Supported Speeds
	 */
	port_attr->supp_speed = bfa_os_htonl(BFA_FCS_FDMI_SUPORTED_SPEEDS);

	/*
	 * Current Speed
	 */
	port_attr->curr_speed = bfa_os_htonl(pport_attr.speed);

	/*
	 * Max PDU Size.
	 */
	port_attr->max_frm_size = bfa_os_htonl(FC_MAX_PDUSZ);

	/*
	 * OS device Name
	 */
	strncpy(port_attr->os_device_name, BFA_FCS_FDMI_OS_DEVICE_NAME,
		sizeof(BFA_FCS_FDMI_OS_DEVICE_NAME));

	/*
	 * Host name
	 */
	strncpy(port_attr->host_name, BFA_FCS_FDMI_HOST_NAME,
		sizeof(BFA_FCS_FDMI_HOST_NAME));

}

/**
* @brief : Public Functions
 */
void
bfa_fcs_port_fdmi_init(bfa_fcs_port_ms_t *ms)
{
	bfa_fcs_port_fdmi_t *fdmi = &ms->fdmi;

	fdmi->ms = ms;
	bfa_sm_set_state(fdmi, bfa_fcs_port_fdmi_sm_offline);
}

void
bfa_fcs_port_fdmi_offline(bfa_fcs_port_ms_t *ms)
{
	bfa_fcs_port_fdmi_t *fdmi = &ms->fdmi;

	fdmi->ms = ms;
	bfa_sm_send_event(fdmi, FDMISM_EVENT_PORT_OFFLINE);
}

void
bfa_fcs_port_fdmi_online(bfa_fcs_port_ms_t *ms)
{
	bfa_fcs_port_fdmi_t *fdmi = &ms->fdmi;

	fdmi->ms = ms;
	bfa_sm_send_event(fdmi, FDMISM_EVENT_PORT_ONLINE);
}
