/*
 * Copyright (c)  2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

/**
 * @file port_loop.c vport private loop implementation.
 */
#include <hal/bfa_hal.h>
#include <hal/bfa_hal_timer.h>
#include <hal/bfa_hal_fcxp.h>
#include <hcb/bfa_hcb_fcxp.h>
#include <fcs/bfa_fcs_port.h>
#include <fcs/bfa_fcs_vport.h>
#include "fcs_port.h"
#include "fcs_rport.h"
#include "port_priv.h"
#include "fcs_trcmod.h"

BFA_TRC_FILE(FCS, LOOP);

/**
 * @brief
 *   ALPA to LIXA bitmap mapping
 *
 *   ALPA 0x00 (Word 0, Bit 30) is invalid for N_Ports. Also Word 0 Bit 31
 * is for L_bit (login required) and is filled as ALPA 0x00 here.
 */
const uint8_t   port_loop_alpa_map[] = {
	0xEF, 0xE8, 0xE4, 0xE2, 0xE1, 0xE0, 0xDC, 0xDA,	/* Word 3 Bits 0..7 */
	0xD9, 0xD6, 0xD5, 0xD4, 0xD3, 0xD2, 0xD1, 0xCE,	/* Word 3 Bits 8..15 */
	0xCD, 0xCC, 0xCB, 0xCA, 0xC9, 0xC7, 0xC6, 0xC5,	/* Word 3 Bits 16..23 */
	0xC3, 0xBC, 0xBA, 0xB9, 0xB6, 0xB5, 0xB4, 0xB3,	/* Word 3 Bits 24..31 */

	0xB2, 0xB1, 0xAE, 0xAD, 0xAC, 0xAB, 0xAA, 0xA9,	/* Word 2 Bits 0..7 */
	0xA7, 0xA6, 0xA5, 0xA3, 0x9F, 0x9E, 0x9D, 0x9B,	/* Word 2 Bits 8..15 */
	0x98, 0x97, 0x90, 0x8F, 0x88, 0x84, 0x82, 0x81,	/* Word 2 Bits 16..23 */
	0x80, 0x7C, 0x7A, 0x79, 0x76, 0x75, 0x74, 0x73,	/* Word 2 Bits 24..31 */

	0x72, 0x71, 0x6E, 0x6D, 0x6C, 0x6B, 0x6A, 0x69,	/* Word 1 Bits 0..7 */
	0x67, 0x66, 0x65, 0x63, 0x5C, 0x5A, 0x59, 0x56,	/* Word 1 Bits 8..15 */
	0x55, 0x54, 0x53, 0x52, 0x51, 0x4E, 0x4D, 0x4C,	/* Word 1 Bits 16..23 */
	0x4B, 0x4A, 0x49, 0x47, 0x46, 0x45, 0x43, 0x3C,	/* Word 1 Bits 24..31 */

	0x3A, 0x39, 0x36, 0x35, 0x34, 0x33, 0x32, 0x31,	/* Word 0 Bits 0..7 */
	0x2E, 0x2D, 0x2C, 0x2B, 0x2A, 0x29, 0x27, 0x26,	/* Word 0 Bits 8..15 */
	0x25, 0x23, 0x1F, 0x1E, 0x1D, 0x1B, 0x18, 0x17,	/* Word 0 Bits 16..23 */
	0x10, 0x0F, 0x08, 0x04, 0x02, 0x01, 0x00, 0x00,	/* Word 0 Bits 24..31 */
};

/*
 * Local Functions
 */
bfa_status_t    bfa_fcs_port_loop_send_plogi(bfa_fcs_port_t *port,
					     uint8_t alpa);

void            bfa_fcs_port_loop_plogi_response(void *fcsarg,
						 bfa_hal_fcxp_t *fcxp,
						 void *cbarg,
						 bfa_status_t req_status,
						 uint32_t rsp_len,
						 uint32_t resid_len,
						 fchs_t *rsp_fchs);

bfa_status_t    bfa_fcs_port_loop_send_adisc(bfa_fcs_port_t *port,
					     uint8_t alpa);

void            bfa_fcs_port_loop_adisc_response(void *fcsarg,
						 bfa_hal_fcxp_t *fcxp,
						 void *cbarg,
						 bfa_status_t req_status,
						 uint32_t rsp_len,
						 uint32_t resid_len,
						 fchs_t *rsp_fchs);

bfa_status_t    bfa_fcs_port_loop_send_plogi_acc(bfa_fcs_port_t *port,
						 uint8_t alpa);

void            bfa_fcs_port_loop_plogi_acc_response(void *fcsarg,
						     bfa_hal_fcxp_t *fcxp,
						     void *cbarg,
						     bfa_status_t req_status,
						     uint32_t rsp_len,
						     uint32_t resid_len,
						     fchs_t *rsp_fchs);

bfa_status_t    bfa_fcs_port_loop_send_adisc_acc(bfa_fcs_port_t *port,
						 uint8_t alpa);

void            bfa_fcs_port_loop_adisc_acc_response(void *fcsarg,
						     bfa_hal_fcxp_t *fcxp,
						     void *cbarg,
						     bfa_status_t req_status,
						     uint32_t rsp_len,
						     uint32_t resid_len,
						     fchs_t *rsp_fchs);
/**
 * @brief
 *   Called by port to initializar in provate LOOP topology.
 */
void
bfa_fcs_port_loop_init(bfa_fcs_port_t *port)
{
}

/**
 * @brief
 *   Called by port to notify transition to online state.
 */
void
bfa_fcs_port_loop_online(bfa_fcs_port_t *port)
{

	uint8_t         num_alpa = port->port_topo.ploop.num_alpa;
	uint8_t        *alpa_pos_map = port->port_topo.ploop.alpa_pos_map;
	bfa_fcs_rport_t *r_port;
	int             ii = 0;

	/*
	 * If the port role is Initiator Mode, create Rports.
	 */
	if (port->port_cfg.roles == BFA_PORT_ROLE_FCP_IM) {
		/*
		 * Check if the ALPA positional bitmap is available.
		 * if not, we send PLOGI to all possible ALPAs.
		 */
		if (num_alpa > 0) {
			for (ii = 0; ii < num_alpa; ii++) {
				/*
				 * ignore ALPA of bfa port
				 */
				if (alpa_pos_map[ii] != port->pid) {
					r_port = bfa_fcs_rport_create(port,
						alpa_pos_map[ii]);
				}
			}
		} else {
			for (ii = 0; ii < MAX_ALPA_COUNT; ii++) {
				/*
				 * ignore ALPA of bfa port
				 */
				if ((port_loop_alpa_map[ii] > 0)
				    && (port_loop_alpa_map[ii] != port->pid))
					bfa_fcs_port_loop_send_plogi(port,
						port_loop_alpa_map[ii]);
				/**TBD */
			}
		}
	} else {
		/*
		 * TBD Target Mode ??
		 */
	}

}

/**
 * @brief
 *   Called by port to notify transition to offline state.
 */
void
bfa_fcs_port_loop_offline(bfa_fcs_port_t *port)
{

}

/**
 * @brief
 *   Called by port to notify a LIP on the loop.
 */
void
bfa_fcs_port_loop_lip(bfa_fcs_port_t *port)
{
}

/**
 * @}
 * Local Functions.
 */
bfa_status_t
bfa_fcs_port_loop_send_plogi(bfa_fcs_port_t *port, uint8_t alpa)
{
	fchs_t          fchs;
	bfa_hal_fcxp_t *fcxp = NULL;
	int             len;

	bfa_trc(port->fcs, alpa);

	fcxp = bfa_hal_fcxp_alloc(NULL, port->fcs->hal, 0, 0, NULL, NULL, NULL,
				  NULL);
	bfa_assert(fcxp);

	len = fc_plogi_build(&fchs, bfa_hal_fcxp_get_reqbuf(fcxp), alpa,
			     bfa_fcs_port_get_fcid(port), 0,
			     port->port_cfg.pwwn, port->port_cfg.nwwn,
			     BFA_FCS_PORT_MAX_DATA_SIZE);

	bfa_hal_fcxp_send(fcxp, NULL, port->fabric->vf_id, BFA_FALSE,
			  FC_CLASS_3, len, &fchs,
			  bfa_fcs_port_loop_plogi_response, (void *)port,
			  FC_MAX_PDUSZ, FC_RA_TOV);

	return BFA_STATUS_OK;
}

/**
 * @brief
 *   Called by fcxp to notify the Plogi response
 */
void
bfa_fcs_port_loop_plogi_response(void *fcsarg, bfa_hal_fcxp_t *fcxp,
				 void *cbarg, bfa_status_t req_status,
				 uint32_t rsp_len, uint32_t resid_len,
				 fchs_t *rsp_fchs)
{
	bfa_fcs_port_t *port = (bfa_fcs_port_t *) cbarg;
	fc_plogi_t     *plogi_resp;
	fc_els_cmd_t   *els_cmd;

	bfa_trc(port->fcs, req_status);

	/*
	 * Sanity Checks
	 */
	if (req_status != BFA_STATUS_OK) {
		bfa_trc(port->fcs, req_status);
		/*
		 * @todo
		 * This could mean that the device with this APLA does not
		 * exist on the loop.
		 */

		return;
	}

	els_cmd = (fc_els_cmd_t *) BFA_FCXP_RSP_PLD(fcxp);
	plogi_resp = (fc_plogi_t *) els_cmd;

	if (els_cmd->els_code == FC_ELS_ACC) {
		bfa_fcs_rport_start(port, rsp_fchs, plogi_resp);
	} else {
		bfa_trc(port->fcs, plogi_resp->els_cmd.els_code);
		bfa_assert(0);
	}
}

bfa_status_t
bfa_fcs_port_loop_send_plogi_acc(bfa_fcs_port_t *port, uint8_t alpa)
{
	fchs_t          fchs;
	bfa_hal_fcxp_t *fcxp;
	int             len;

	bfa_trc(port->fcs, alpa);

	fcxp = bfa_hal_fcxp_alloc(NULL, port->fcs->hal, 0, 0, NULL, NULL, NULL,
				  NULL);
	bfa_assert(fcxp);

	len = fc_plogi_acc_build(&fchs, bfa_hal_fcxp_get_reqbuf(fcxp), alpa,
				 bfa_fcs_port_get_fcid(port), 0,
				 port->port_cfg.pwwn, port->port_cfg.nwwn,
				 BFA_FCS_PORT_MAX_DATA_SIZE);

	bfa_hal_fcxp_send(fcxp, NULL, port->fabric->vf_id, BFA_FALSE,
				 FC_CLASS_3, len, &fchs,
				 bfa_fcs_port_loop_plogi_acc_response,
				 (void *)port, FC_MAX_PDUSZ, 0); /* No response
								  * expected
								  */

	return BFA_STATUS_OK;
}

/*
 *  Plogi Acc Response
 * We donot do any processing here.
 */
void
bfa_fcs_port_loop_plogi_acc_response(void *fcsarg, bfa_hal_fcxp_t *fcxp,
				     void *cbarg, bfa_status_t req_status,
				     uint32_t rsp_len, uint32_t resid_len,
				     fchs_t *rsp_fchs)
{

	bfa_fcs_port_t *port = (bfa_fcs_port_t *) cbarg;

	bfa_trc(port->fcs, port->pid);

	/*
	 * Sanity Checks
	 */
	if (req_status != BFA_STATUS_OK) {
		bfa_trc(port->fcs, req_status);
		return;
	}
}

bfa_status_t
bfa_fcs_port_loop_send_adisc(bfa_fcs_port_t *port, uint8_t alpa)
{
	fchs_t          fchs;
	bfa_hal_fcxp_t *fcxp;
	int             len;

	bfa_trc(port->fcs, alpa);

	fcxp = bfa_hal_fcxp_alloc(NULL, port->fcs->hal, 0, 0, NULL, NULL, NULL,
				  NULL);
	bfa_assert(fcxp);

	len = fc_adisc_build(&fchs, bfa_hal_fcxp_get_reqbuf(fcxp), alpa,
			     bfa_fcs_port_get_fcid(port), 0,
			     port->port_cfg.pwwn, port->port_cfg.nwwn);

	bfa_hal_fcxp_send(fcxp, NULL, port->fabric->vf_id, BFA_FALSE,
			  FC_CLASS_3, len, &fchs,
			  bfa_fcs_port_loop_adisc_response, (void *)port,
			  FC_MAX_PDUSZ, FC_RA_TOV);

	return BFA_STATUS_OK;
}

/**
 * @brief
 *   Called by fcxp to notify the ADISC response
 */
void
bfa_fcs_port_loop_adisc_response(void *fcsarg, bfa_hal_fcxp_t *fcxp,
				 void *cbarg, bfa_status_t req_status,
				 uint32_t rsp_len, uint32_t resid_len,
				 fchs_t *rsp_fchs)
{
	bfa_fcs_port_t *port = (bfa_fcs_port_t *) cbarg;
	bfa_fcs_rport_t *rport;
	fc_adisc_t     *adisc_resp;
	fc_els_cmd_t   *els_cmd;
	uint32_t        pid = rsp_fchs->s_id;

	bfa_trc(port->fcs, req_status);

	/*
	 * Sanity Checks
	 */
	if (req_status != BFA_STATUS_OK) {
		/*
		 * TBD : we may need to retry certain requests
		 */
		bfa_hal_fcxp_free(fcxp);
		return;
	}

	els_cmd = (fc_els_cmd_t *) BFA_FCXP_RSP_PLD(fcxp);
	adisc_resp = (fc_adisc_t *) els_cmd;

	if (els_cmd->els_code == FC_ELS_ACC) {
	} else {
		bfa_trc(port->fcs, adisc_resp->els_cmd.els_code);

		/*
		 * TBD: we may need to check for reject codes and retry
		 */
		rport = bfa_fcs_port_get_rport_by_pid(port, pid);
		if (rport) {
			bfa_q_qe_deq(rport);
			bfa_fcs_rport_delete(rport);
		}

	}
	return;
}

bfa_status_t
bfa_fcs_port_loop_send_adisc_acc(bfa_fcs_port_t *port, uint8_t alpa)
{
	fchs_t          fchs;
	bfa_hal_fcxp_t *fcxp;
	int             len;

	bfa_trc(port->fcs, alpa);

	fcxp = bfa_hal_fcxp_alloc(NULL, port->fcs->hal, 0, 0, NULL, NULL, NULL,
				  NULL);
	bfa_assert(fcxp);

	len = fc_adisc_acc_build(&fchs, bfa_hal_fcxp_get_reqbuf(fcxp), alpa,
				 bfa_fcs_port_get_fcid(port), 0,
				 port->port_cfg.pwwn, port->port_cfg.nwwn);

	bfa_hal_fcxp_send(fcxp, NULL, port->fabric->vf_id, BFA_FALSE,
				FC_CLASS_3, len, &fchs,
				bfa_fcs_port_loop_adisc_acc_response,
				(void *)port, FC_MAX_PDUSZ, 0); /* no reponse
								 * expected
								 */

	return BFA_STATUS_OK;
}

/*
 *  Adisc Acc Response
 * We donot do any processing here.
 */
void
bfa_fcs_port_loop_adisc_acc_response(void *fcsarg, bfa_hal_fcxp_t *fcxp,
				     void *cbarg, bfa_status_t req_status,
				     uint32_t rsp_len, uint32_t resid_len,
				     fchs_t *rsp_fchs)
{

	bfa_fcs_port_t *port = (bfa_fcs_port_t *) cbarg;

	bfa_trc(port->fcs, port->pid);

	/*
	 * Sanity Checks
	 */
	if (req_status != BFA_STATUS_OK) {
		bfa_trc(port->fcs, req_status);
		return;
	}
}
