/*
 * Copyright (c)  2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

/**
 * @file rport.c Remote port implementation.
 */

#include <hal/bfa_hal.h>
#include <hal/bfa_hal_timer.h>
#include <hal/bfa_hal_fcxp.h>
#include <hal/bfa_hal_rport.h>
#include <hal/bfa_hal_pport.h>
#include <hcb/bfa_hcb_fcxp.h>
#include <fcs/bfa_fcs_port.h>
#include <fcs/bfa_fcs_vport.h>
#include <fcs/bfa_fcs_rport.h>
#include <hcb/bfa_hcb_rport.h>
#include <fcb/bfa_fcb_rport.h>
#include "fcs.h"
#include "fcbuild.h"
#include "fcs_vport.h"
#include "fcs_port.h"
#include "fcs_rport.h"
#include "fcs_fcpim.h"
#include "fcs_fcptm.h"
#include "port_priv.h"
#include "fcs_trcmod.h"
#include "fcs_fcxp.h"
#include <aen/bfal_aen_rport.h>

BFA_TRC_FILE(FCS, RPORT);

#define BFA_FCS_RPORT_MAX_RETRIES		(5)

static uint32_t 
bfa_fcs_rport_del_timeout = BFA_FCS_RPORT_DEF_DEL_TIMEOUT * 1000; /* In millisecs */
/*
 * forward declarations
 */
static bfa_fcs_rport_t *bfa_fcs_rport_alloc(bfa_fcs_port_t *port, wwn_t pwwn,
					    uint32_t rpid);
static void     bfa_fcs_rport_free(bfa_fcs_rport_t *rport);
static void     bfa_fcs_rport_hal_online(bfa_fcs_rport_t *rport);
static void     bfa_fcs_rport_online_action(bfa_fcs_rport_t *rport);
static void     bfa_fcs_rport_offline_action(bfa_fcs_rport_t *rport);
static void     bfa_fcs_rport_update(bfa_fcs_rport_t *rport,
				     fc_plogi_t *plogi);
static void     bfa_fcs_rport_fc4_pause(bfa_fcs_rport_t *rport);
static void     bfa_fcs_rport_fc4_resume(bfa_fcs_rport_t *rport);
static void     bfa_fcs_rport_timeout(void *arg);
static void     bfa_fcs_rport_send_plogi(void *rport_cbarg,
					 bfa_hal_fcxp_t *fcxp_alloced);
static void     bfa_fcs_rport_send_plogiacc(void *rport_cbarg,
					    bfa_hal_fcxp_t *fcxp_alloced);
static void     bfa_fcs_rport_plogi_response(void *fcsarg,
					     bfa_hal_fcxp_t *fcxp, void *cbarg,
					     bfa_status_t req_status,
					     uint32_t rsp_len,
					     uint32_t resid_len,
					     fchs_t *rsp_fchs);
static void     bfa_fcs_rport_send_adisc(void *rport_cbarg,
					 bfa_hal_fcxp_t *fcxp_alloced);
static void     bfa_fcs_rport_adisc_response(void *fcsarg,
					     bfa_hal_fcxp_t *fcxp, void *cbarg,
					     bfa_status_t req_status,
					     uint32_t rsp_len,
					     uint32_t resid_len,
					     fchs_t *rsp_fchs);
static void     bfa_fcs_rport_send_gpnid(void *rport_cbarg,
					 bfa_hal_fcxp_t *fcxp_alloced);
static void     bfa_fcs_rport_gpnid_response(void *fcsarg,
					     bfa_hal_fcxp_t *fcxp, void *cbarg,
					     bfa_status_t req_status,
					     uint32_t rsp_len,
					     uint32_t resid_len,
					     fchs_t *rsp_fchs);
static void     bfa_fcs_rport_send_logo(void *rport_cbarg,
					bfa_hal_fcxp_t *fcxp_alloced);
static void     bfa_fcs_rport_send_logo_acc(void *rport_cbarg);
static void     bfa_fcs_rport_process_prli(bfa_fcs_rport_t *rport,
					   fchs_t *rx_fchs, uint16_t len);
static void     bfa_fcs_rport_send_ls_rjt(bfa_fcs_rport_t *rport,
					  fchs_t *rx_fchs, uint8_t reason_code,
					  uint8_t reason_code_expl);
static void 	bfa_fcs_rport_process_adisc(bfa_fcs_rport_t *rport, 
					  fchs_t *rx_fchs, uint16_t len);
#ifdef RPSC_TEST
static void     bfa_fcs_rport_send_rpsc(void *rport_cbarg,
					bfa_hal_fcxp_t *fcxp_alloced);
static void     bfa_fcs_rport_rpsc_response(void *fcsarg, bfa_hal_fcxp_t *fcxp,
					    void *cbarg,
					    bfa_status_t req_status,
					    uint32_t rsp_len,
					    uint32_t resid_len,
					    fchs_t *rsp_fchs);
#endif /* RPSC_TEST */

/**
 * @dg fcs_rport_sm FCS rport state machine events
 * @{
 */

typedef enum {
	RPSM_EVENT_PLOGI_SEND  = 1,      /*!< new rport; start with PLOGI       */
	RPSM_EVENT_PLOGI_RCVD  = 2,      /*!< Inbound PLOGI from remote port    */
	RPSM_EVENT_PLOGI_COMP  = 3,      /*!< PLOGI completed to rport          */
	RPSM_EVENT_LOGO_RCVD   = 4,      /*!< LOGO from remote device           */
	RPSM_EVENT_LOGO_IMP    = 5,      /*!< implicit logo for SLER            */
	RPSM_EVENT_FCXP_SENT   = 6,      /*!< Frame from has been sent          */
	RPSM_EVENT_DELETE      = 7,     /*!< RPORT delete request              */
	RPSM_EVENT_SCN         = 8,      /*!< state change notification         */
	RPSM_EVENT_ACCEPTED    = 9,      /*!< Good response from remote device  */
	RPSM_EVENT_FAILED      = 10,     /*!< Request to rport failed.          */
	RPSM_EVENT_TIMEOUT     = 11,     /*!< Rport SM timeout event            */
	RPSM_EVENT_HCB_ONLINE  = 12,     /*!< HAL rport online callback         */
	RPSM_EVENT_HCB_OFFLINE = 13,     /*!< HAL rport offline callback        */
	RPSM_EVENT_FC4_OFFLINE = 14,     /*!< FC-4 offline complete             */
} rport_event_t;

static void     bfa_fcs_rport_sm_uninit(bfa_fcs_rport_t *rport,
					rport_event_t event);
static void     bfa_fcs_rport_sm_plogi_sending(bfa_fcs_rport_t *rport,
					       rport_event_t event);
static void     bfa_fcs_rport_sm_plogiacc_sending(bfa_fcs_rport_t *rport,
						  rport_event_t event);
static void     bfa_fcs_rport_sm_plogi_retry(bfa_fcs_rport_t *rport,
					     rport_event_t event);
static void     bfa_fcs_rport_sm_plogi(bfa_fcs_rport_t *rport,
				       rport_event_t event);
static void     bfa_fcs_rport_sm_hal_online(bfa_fcs_rport_t *rport,
					    rport_event_t event);
static void     bfa_fcs_rport_sm_online(bfa_fcs_rport_t *rport,
					rport_event_t event);
static void     bfa_fcs_rport_sm_nsquery_sending(bfa_fcs_rport_t *rport,
						 rport_event_t event);
static void     bfa_fcs_rport_sm_nsquery(bfa_fcs_rport_t *rport,
					 rport_event_t event);
static void     bfa_fcs_rport_sm_adisc_sending(bfa_fcs_rport_t *rport,
					       rport_event_t event);
static void     bfa_fcs_rport_sm_adisc(bfa_fcs_rport_t *rport,
				       rport_event_t event);
static void     bfa_fcs_rport_sm_fc4_logorcv(bfa_fcs_rport_t *rport,
					     rport_event_t event);
static void     bfa_fcs_rport_sm_fc4_logosend(bfa_fcs_rport_t *rport,
					      rport_event_t event);
static void     bfa_fcs_rport_sm_fc4_offline(bfa_fcs_rport_t *rport,
					     rport_event_t event);
static void     bfa_fcs_rport_sm_hcb_offline(bfa_fcs_rport_t *rport,
					     rport_event_t event);
static void     bfa_fcs_rport_sm_hcb_logorcv(bfa_fcs_rport_t *rport,
					     rport_event_t event);
static void     bfa_fcs_rport_sm_hcb_logosend(bfa_fcs_rport_t *rport,
					      rport_event_t event);
static void     bfa_fcs_rport_sm_logo_sending(bfa_fcs_rport_t *rport,
					      rport_event_t event);
static void     bfa_fcs_rport_sm_offline(bfa_fcs_rport_t *rport,
					 rport_event_t event);
static void     bfa_fcs_rport_sm_nsdisc_sending(bfa_fcs_rport_t *rport,
						rport_event_t event);
static void     bfa_fcs_rport_sm_nsdisc_retry(bfa_fcs_rport_t *rport,
					      rport_event_t event);
static void     bfa_fcs_rport_sm_nsdisc_sent(bfa_fcs_rport_t *rport,
					     rport_event_t event);
static void     bfa_fcs_rport_sm_nsdisc_sent(bfa_fcs_rport_t *rport,
					     rport_event_t event);

static bfa_sm_table_t rport_sm_table[] = {
	{BFA_SM(bfa_fcs_rport_sm_uninit), BFA_RPORT_UNINIT},
	{BFA_SM(bfa_fcs_rport_sm_plogi_sending), BFA_RPORT_PLOGI},
	{BFA_SM(bfa_fcs_rport_sm_plogiacc_sending), BFA_RPORT_ONLINE},
	{BFA_SM(bfa_fcs_rport_sm_plogi_retry), BFA_RPORT_PLOGI_RETRY},
	{BFA_SM(bfa_fcs_rport_sm_plogi), BFA_RPORT_PLOGI},
	{BFA_SM(bfa_fcs_rport_sm_hal_online), BFA_RPORT_ONLINE},
	{BFA_SM(bfa_fcs_rport_sm_online), BFA_RPORT_ONLINE},
	{BFA_SM(bfa_fcs_rport_sm_nsquery_sending), BFA_RPORT_NSQUERY},
	{BFA_SM(bfa_fcs_rport_sm_nsquery), BFA_RPORT_NSQUERY},
	{BFA_SM(bfa_fcs_rport_sm_adisc_sending), BFA_RPORT_ADISC},
	{BFA_SM(bfa_fcs_rport_sm_adisc), BFA_RPORT_ADISC},
	{BFA_SM(bfa_fcs_rport_sm_fc4_logorcv), BFA_RPORT_LOGORCV},
	{BFA_SM(bfa_fcs_rport_sm_fc4_logosend), BFA_RPORT_LOGO},
	{BFA_SM(bfa_fcs_rport_sm_fc4_offline), BFA_RPORT_OFFLINE},
	{BFA_SM(bfa_fcs_rport_sm_hcb_offline), BFA_RPORT_OFFLINE},
	{BFA_SM(bfa_fcs_rport_sm_hcb_logorcv), BFA_RPORT_LOGORCV},
	{BFA_SM(bfa_fcs_rport_sm_hcb_logosend), BFA_RPORT_LOGO},
	{BFA_SM(bfa_fcs_rport_sm_logo_sending), BFA_RPORT_LOGO},
	{BFA_SM(bfa_fcs_rport_sm_offline), BFA_RPORT_OFFLINE},
	{BFA_SM(bfa_fcs_rport_sm_nsdisc_sending), BFA_RPORT_NSDISC},
	{BFA_SM(bfa_fcs_rport_sm_nsdisc_retry), BFA_RPORT_NSDISC},
	{BFA_SM(bfa_fcs_rport_sm_nsdisc_sent), BFA_RPORT_NSDISC},
};

/**
 * @brief
 * 		Beginning state.
 */
static void
bfa_fcs_rport_sm_uninit(bfa_fcs_rport_t *rport, rport_event_t event)
{
	bfa_trc(rport->fcs, rport->pwwn);
	bfa_trc(rport->fcs, rport->pid);
	bfa_trc(rport->fcs, event);

	switch (event) {
	case RPSM_EVENT_PLOGI_SEND:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_plogi_sending);
		rport->plogi_retries = 0;
		bfa_fcs_rport_send_plogi(rport, NULL);
		break;

	case RPSM_EVENT_PLOGI_RCVD:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_plogiacc_sending);
		bfa_fcs_rport_send_plogiacc(rport, NULL);
		break;

	case RPSM_EVENT_PLOGI_COMP:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_hal_online);
		bfa_fcs_rport_hal_online(rport);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * 		PLOGI is being sent.
 */
static void
bfa_fcs_rport_sm_plogi_sending(bfa_fcs_rport_t *rport, rport_event_t event)
{
	bfa_trc(rport->fcs, rport->pwwn);
	bfa_trc(rport->fcs, rport->pid);
	bfa_trc(rport->fcs, event);

	switch (event) {
	case RPSM_EVENT_FCXP_SENT:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_plogi);
		break;

	case RPSM_EVENT_DELETE:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_uninit);
		bfa_hal_fcxp_walloc_cancel(rport->fcs->hal, &rport->fcxp_wqe);
		bfa_fcs_rport_free(rport);
		break;

	case RPSM_EVENT_PLOGI_RCVD:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_plogiacc_sending);
		bfa_hal_fcxp_walloc_cancel(rport->fcs->hal, &rport->fcxp_wqe);
		bfa_fcs_rport_send_plogiacc(rport, NULL);
		break;

	case RPSM_EVENT_LOGO_IMP:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_offline);
		bfa_hal_fcxp_walloc_cancel(rport->fcs->hal, &rport->fcxp_wqe);
		bfa_hal_timer_start(rport->fcs->hal, &rport->timer,
				    bfa_fcs_rport_timeout, rport,
				    bfa_fcs_rport_del_timeout);
		break;

	case RPSM_EVENT_SCN:
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * 		PLOGI is being sent.
 */
static void
bfa_fcs_rport_sm_plogiacc_sending(bfa_fcs_rport_t *rport, rport_event_t event)
{
	bfa_trc(rport->fcs, rport->pwwn);
	bfa_trc(rport->fcs, rport->pid);
	bfa_trc(rport->fcs, event);

	switch (event) {
	case RPSM_EVENT_FCXP_SENT:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_hal_online);
		bfa_fcs_rport_hal_online(rport);
		break;

	case RPSM_EVENT_DELETE:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_uninit);
		bfa_hal_fcxp_walloc_cancel(rport->fcs->hal, &rport->fcxp_wqe);
		bfa_fcs_rport_free(rport);
		break;

	case RPSM_EVENT_SCN:
		/**
		 * Ignore, SCN is possibly online notification.
		 */
		break;

	case RPSM_EVENT_LOGO_IMP:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_offline);
		bfa_hal_fcxp_walloc_cancel(rport->fcs->hal, &rport->fcxp_wqe);
		bfa_hal_timer_start(rport->fcs->hal, &rport->timer,
				    bfa_fcs_rport_timeout, rport,
				    bfa_fcs_rport_del_timeout);
		break;

	case RPSM_EVENT_HCB_OFFLINE:
		/**
		 * Ignore HAL callback, on a PLOGI receive we call hal offline.
		 */
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * 		PLOGI is sent.
 */
static void
bfa_fcs_rport_sm_plogi_retry(bfa_fcs_rport_t *rport, rport_event_t event)
{
	bfa_trc(rport->fcs, rport->pwwn);
	bfa_trc(rport->fcs, rport->pid);
	bfa_trc(rport->fcs, event);

	switch (event) {
	case RPSM_EVENT_SCN:
		bfa_hal_timer_stop(&rport->timer);
		/*
		 * !! fall through !!
		 */

	case RPSM_EVENT_TIMEOUT:
		rport->plogi_retries++;
		if (rport->plogi_retries < BFA_FCS_RPORT_MAX_RETRIES) {
			bfa_sm_set_state(rport, bfa_fcs_rport_sm_plogi_sending);
			bfa_fcs_rport_send_plogi(rport, NULL);
		} else {
			bfa_sm_set_state(rport, bfa_fcs_rport_sm_offline);
			bfa_hal_timer_start(rport->fcs->hal, &rport->timer,
					    bfa_fcs_rport_timeout, rport,
					    bfa_fcs_rport_del_timeout);
		}
		break;

	case RPSM_EVENT_DELETE:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_uninit);
		bfa_hal_timer_stop(&rport->timer);
		bfa_fcs_rport_free(rport);
		break;

	case RPSM_EVENT_LOGO_RCVD:
		break;

	case RPSM_EVENT_PLOGI_RCVD:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_plogiacc_sending);
		bfa_hal_timer_stop(&rport->timer);
		bfa_fcs_rport_send_plogiacc(rport, NULL);
		break;

	case RPSM_EVENT_LOGO_IMP:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_offline);
		bfa_hal_timer_stop(&rport->timer);
		bfa_hal_timer_start(rport->fcs->hal, &rport->timer,
				    bfa_fcs_rport_timeout, rport,
				    bfa_fcs_rport_del_timeout);
		break;

	case RPSM_EVENT_PLOGI_COMP:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_hal_online);
		bfa_hal_timer_stop(&rport->timer);
		bfa_fcs_rport_hal_online(rport);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * 		PLOGI is sent.
 */
static void
bfa_fcs_rport_sm_plogi(bfa_fcs_rport_t *rport, rport_event_t event)
{
	bfa_trc(rport->fcs, rport->pwwn);
	bfa_trc(rport->fcs, rport->pid);
	bfa_trc(rport->fcs, event);

	switch (event) {
	case RPSM_EVENT_ACCEPTED:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_hal_online);
		rport->plogi_retries = 0;
		bfa_fcs_rport_hal_online(rport);
		break;

	case RPSM_EVENT_FAILED:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_plogi_retry);
		bfa_hal_timer_start(rport->fcs->hal, &rport->timer,
				    bfa_fcs_rport_timeout, rport,
				    BFA_FCS_RETRY_TIMEOUT);
		break;

	case RPSM_EVENT_LOGO_RCVD:
	case RPSM_EVENT_LOGO_IMP:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_offline);
		bfa_hal_fcxp_discard(rport->fcxp);
		bfa_hal_timer_start(rport->fcs->hal, &rport->timer,
				    bfa_fcs_rport_timeout, rport,
				    bfa_fcs_rport_del_timeout);
		break;

	case RPSM_EVENT_PLOGI_RCVD:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_plogiacc_sending);
		bfa_hal_fcxp_discard(rport->fcxp);
		bfa_fcs_rport_send_plogiacc(rport, NULL);
		break;

	case RPSM_EVENT_SCN:
		/**
		 * Ignore SCN - wait for PLOGI response.
		 */
		break;

	case RPSM_EVENT_DELETE:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_uninit);
		bfa_hal_fcxp_discard(rport->fcxp);
		bfa_fcs_rport_free(rport);
		break;

	case RPSM_EVENT_PLOGI_COMP:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_hal_online);
		bfa_hal_fcxp_discard(rport->fcxp);
		bfa_fcs_rport_hal_online(rport);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * 		PLOGI is complete. Awaiting HAL rport online callback. FC-4s
 * 		are offline.
 */
static void
bfa_fcs_rport_sm_hal_online(bfa_fcs_rport_t *rport, rport_event_t event)
{
	bfa_trc(rport->fcs, rport->pwwn);
	bfa_trc(rport->fcs, rport->pid);
	bfa_trc(rport->fcs, event);

	switch (event) {
	case RPSM_EVENT_HCB_ONLINE:

#ifdef RPSC_TEST
		bfa_fcs_rport_send_rpsc(rport, NULL);
#endif
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_online);
		bfa_fcs_rport_online_action(rport);
		break;

	case RPSM_EVENT_LOGO_RCVD:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_hcb_logorcv);
		bfa_hal_rport_offline(rport->hal_rport);
		break;

	case RPSM_EVENT_LOGO_IMP:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_hcb_offline);
		bfa_hal_rport_offline(rport->hal_rport);
		break;

	case RPSM_EVENT_PLOGI_RCVD:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_plogiacc_sending);
		bfa_hal_rport_offline(rport->hal_rport);
		bfa_fcs_rport_send_plogiacc(rport, NULL);
		break;

	case RPSM_EVENT_DELETE:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_hcb_logosend);
		bfa_hal_rport_offline(rport->hal_rport);
		break;

	case RPSM_EVENT_SCN:
		/**
		 * @todo
		 * Ignore SCN - PLOGI just completed, FC-4 login should detect
		 * device failures.
		 */
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * 		Rport is ONLINE. FC-4s active.
 */
static void
bfa_fcs_rport_sm_online(bfa_fcs_rport_t *rport, rport_event_t event)
{
	bfa_trc(rport->fcs, rport->pwwn);
	bfa_trc(rport->fcs, rport->pid);
	bfa_trc(rport->fcs, event);

	switch (event) {
	case RPSM_EVENT_SCN:
		/**
		 * Pause FC-4 activity till rport is authenticated.
		 * In switched fabrics, check presence of device in nameserver
		 * first.
		 */
		bfa_fcs_rport_fc4_pause(rport);

		if (bfa_fcs_fabric_is_switched(rport->port->fabric)) {
			bfa_sm_set_state(rport,
					 bfa_fcs_rport_sm_nsquery_sending);
			rport->ns_retries = 0;
			bfa_fcs_rport_send_gpnid(rport, NULL);
		} else {
			bfa_sm_set_state(rport, bfa_fcs_rport_sm_adisc_sending);
			bfa_fcs_rport_send_adisc(rport, NULL);
		}
		break;

	case RPSM_EVENT_PLOGI_RCVD:
	case RPSM_EVENT_LOGO_IMP:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_fc4_offline);
		bfa_fcs_rport_offline_action(rport);
		break;

	case RPSM_EVENT_DELETE:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_fc4_logosend);
		bfa_fcs_rport_offline_action(rport);
		break;

	case RPSM_EVENT_LOGO_RCVD:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_fc4_logorcv);
		bfa_fcs_rport_offline_action(rport);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * 		An SCN event is received in ONLINE state. NS query is being sent
 * 		prior to ADISC authentication with rport. FC-4s are paused.
 */
static void
bfa_fcs_rport_sm_nsquery_sending(bfa_fcs_rport_t *rport, rport_event_t event)
{
	bfa_trc(rport->fcs, rport->pwwn);
	bfa_trc(rport->fcs, rport->pid);
	bfa_trc(rport->fcs, event);

	switch (event) {
	case RPSM_EVENT_FCXP_SENT:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_nsquery);
		break;

	case RPSM_EVENT_DELETE:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_fc4_logosend);
		bfa_hal_fcxp_walloc_cancel(rport->fcs->hal, &rport->fcxp_wqe);
		bfa_fcs_rport_offline_action(rport);
		break;

	case RPSM_EVENT_SCN:
		/**
		 * ignore SCN, wait for response to query itself
		 */
		break;

	case RPSM_EVENT_LOGO_RCVD:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_fc4_logorcv);
		bfa_hal_fcxp_walloc_cancel(rport->fcs->hal, &rport->fcxp_wqe);
		bfa_fcs_rport_offline_action(rport);
		break;

	case RPSM_EVENT_LOGO_IMP:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_offline);
		bfa_hal_fcxp_walloc_cancel(rport->fcs->hal, &rport->fcxp_wqe);
		bfa_hal_timer_start(rport->fcs->hal, &rport->timer,
				    bfa_fcs_rport_timeout, rport,
				    bfa_fcs_rport_del_timeout);
		break;

	case RPSM_EVENT_PLOGI_RCVD:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_fc4_offline);
		bfa_hal_fcxp_walloc_cancel(rport->fcs->hal, &rport->fcxp_wqe);
		bfa_fcs_rport_offline_action(rport);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * 	An SCN event is received in ONLINE state. NS query is sent to rport.
 * 	FC-4s are paused.
 */
static void
bfa_fcs_rport_sm_nsquery(bfa_fcs_rport_t *rport, rport_event_t event)
{
	bfa_trc(rport->fcs, rport->pwwn);
	bfa_trc(rport->fcs, rport->pid);
	bfa_trc(rport->fcs, event);

	switch (event) {
	case RPSM_EVENT_ACCEPTED:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_adisc_sending);
		bfa_fcs_rport_send_adisc(rport, NULL);
		break;

	case RPSM_EVENT_FAILED:
		rport->ns_retries++;
		if (rport->ns_retries < BFA_FCS_RPORT_MAX_RETRIES) {
			bfa_sm_set_state(rport,
					 bfa_fcs_rport_sm_nsquery_sending);
			bfa_fcs_rport_send_gpnid(rport, NULL);
		} else {
			bfa_sm_set_state(rport, bfa_fcs_rport_sm_fc4_offline);
			bfa_fcs_rport_offline_action(rport);
		}
		break;

	case RPSM_EVENT_DELETE:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_fc4_logosend);
		bfa_hal_fcxp_discard(rport->fcxp);
		bfa_fcs_rport_offline_action(rport);
		break;

	case RPSM_EVENT_SCN:
		break;

	case RPSM_EVENT_LOGO_RCVD:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_fc4_logorcv);
		bfa_hal_fcxp_discard(rport->fcxp);
		bfa_fcs_rport_offline_action(rport);
		break;

	case RPSM_EVENT_PLOGI_RCVD:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_fc4_offline);
		bfa_hal_fcxp_discard(rport->fcxp);
		bfa_fcs_rport_offline_action(rport);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * 	An SCN event is received in ONLINE state. ADISC is being sent for
 * 	authenticating with rport. FC-4s are paused.
 */
static void
bfa_fcs_rport_sm_adisc_sending(bfa_fcs_rport_t *rport, rport_event_t event)
{
	bfa_trc(rport->fcs, rport->pwwn);
	bfa_trc(rport->fcs, rport->pid);
	bfa_trc(rport->fcs, event);

	switch (event) {
	case RPSM_EVENT_FCXP_SENT:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_adisc);
		break;

	case RPSM_EVENT_DELETE:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_fc4_logosend);
		bfa_hal_fcxp_walloc_cancel(rport->fcs->hal, &rport->fcxp_wqe);
		bfa_fcs_rport_offline_action(rport);
		break;

	case RPSM_EVENT_LOGO_IMP:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_fc4_offline);
		bfa_hal_fcxp_walloc_cancel(rport->fcs->hal, &rport->fcxp_wqe);
		bfa_fcs_rport_offline_action(rport);
		break;

	case RPSM_EVENT_LOGO_RCVD:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_fc4_logorcv);
		bfa_hal_fcxp_walloc_cancel(rport->fcs->hal, &rport->fcxp_wqe);
		bfa_fcs_rport_offline_action(rport);
		break;

	case RPSM_EVENT_SCN:
		break;

	case RPSM_EVENT_PLOGI_RCVD:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_fc4_offline);
		bfa_hal_fcxp_walloc_cancel(rport->fcs->hal, &rport->fcxp_wqe);
		bfa_fcs_rport_offline_action(rport);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * 		An SCN event is received in ONLINE state. ADISC is to rport.
 * 		FC-4s are paused.
 */
static void
bfa_fcs_rport_sm_adisc(bfa_fcs_rport_t *rport, rport_event_t event)
{
	bfa_trc(rport->fcs, rport->pwwn);
	bfa_trc(rport->fcs, rport->pid);
	bfa_trc(rport->fcs, event);

	switch (event) {
	case RPSM_EVENT_ACCEPTED:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_online);
		bfa_fcs_rport_fc4_resume(rport);
		break;

	case RPSM_EVENT_PLOGI_RCVD:
		/**
		 * Too complex to cleanup FC-4 & rport and then acc to PLOGI.
		 * At least go offline when a PLOGI is received.
		 */
		bfa_hal_fcxp_discard(rport->fcxp);
		/*
		 * !!! fall through !!!
		 */

	case RPSM_EVENT_FAILED:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_fc4_offline);
		bfa_fcs_rport_offline_action(rport);
		break;

	case RPSM_EVENT_DELETE:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_fc4_logosend);
		bfa_hal_fcxp_discard(rport->fcxp);
		bfa_fcs_rport_offline_action(rport);
		break;

	case RPSM_EVENT_SCN:
		/**
		 * already processing RSCN
		 */
		break;

	case RPSM_EVENT_LOGO_IMP:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_fc4_offline);
		bfa_hal_fcxp_discard(rport->fcxp);
		bfa_fcs_rport_offline_action(rport);
		break;

	case RPSM_EVENT_LOGO_RCVD:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_fc4_logorcv);
		bfa_hal_fcxp_discard(rport->fcxp);
		bfa_fcs_rport_offline_action(rport);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * 		Rport has sent LOGO. Awaiting FC-4 offline completion callback.
 */
static void
bfa_fcs_rport_sm_fc4_logorcv(bfa_fcs_rport_t *rport, rport_event_t event)
{
	bfa_trc(rport->fcs, rport->pwwn);
	bfa_trc(rport->fcs, rport->pid);
	bfa_trc(rport->fcs, event);

	switch (event) {
	case RPSM_EVENT_FC4_OFFLINE:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_hcb_logorcv);
		bfa_hal_rport_offline(rport->hal_rport);
		break;

	case RPSM_EVENT_DELETE:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_fc4_logosend);
		break;

	case RPSM_EVENT_LOGO_RCVD:
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * 		LOGO needs to be sent to rport. Awaiting FC-4 offline completion
 * 		callback.
 */
static void
bfa_fcs_rport_sm_fc4_logosend(bfa_fcs_rport_t *rport, rport_event_t event)
{
	bfa_trc(rport->fcs, rport->pwwn);
	bfa_trc(rport->fcs, rport->pid);
	bfa_trc(rport->fcs, event);

	switch (event) {
	case RPSM_EVENT_FC4_OFFLINE:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_hcb_logosend);
		bfa_hal_rport_offline(rport->hal_rport);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * 	Rport is going offline. Awaiting FC-4 offline completion callback.
 */
static void
bfa_fcs_rport_sm_fc4_offline(bfa_fcs_rport_t *rport, rport_event_t event)
{
	bfa_trc(rport->fcs, rport->pwwn);
	bfa_trc(rport->fcs, rport->pid);
	bfa_trc(rport->fcs, event);

	switch (event) {
	case RPSM_EVENT_FC4_OFFLINE:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_hcb_offline);
		bfa_hal_rport_offline(rport->hal_rport);
		break;

	case RPSM_EVENT_SCN:
	case RPSM_EVENT_LOGO_IMP:
	case RPSM_EVENT_LOGO_RCVD:
		/**
		 * rport is already going offline.
		 * SCN - ignore and wait till transitioning to offline state
		 */
		break;

	case RPSM_EVENT_DELETE:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_fc4_logosend);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * 		Rport is offline. FC-4s are offline. Awaiting HAL rport offline
 * 		callback.
 */
static void
bfa_fcs_rport_sm_hcb_offline(bfa_fcs_rport_t *rport, rport_event_t event)
{
	bfa_trc(rport->fcs, rport->pwwn);
	bfa_trc(rport->fcs, rport->pid);
	bfa_trc(rport->fcs, event);

	switch (event) {
	case RPSM_EVENT_HCB_OFFLINE:
		if (bfa_fcs_port_is_online(rport->port)) {
			bfa_sm_set_state(rport, bfa_fcs_rport_sm_nsdisc_sending);
			rport->ns_retries = 0;
			bfa_fcs_rport_send_gpnid(rport, NULL);
		} else {
			bfa_sm_set_state(rport, bfa_fcs_rport_sm_offline);
			bfa_hal_timer_start(rport->fcs->hal, &rport->timer,
				    bfa_fcs_rport_timeout, rport,
				    bfa_fcs_rport_del_timeout);
		}
		break;

	case RPSM_EVENT_DELETE:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_uninit);
		bfa_fcs_rport_free(rport);
		break;

	case RPSM_EVENT_SCN:
	case RPSM_EVENT_LOGO_RCVD:
		/**
		 * Ignore, already offline.
		 */
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * 		Rport is offline. FC-4s are offline. Awaiting HAL rport offline
 * 		callback to send LOGO accept.
 */
static void
bfa_fcs_rport_sm_hcb_logorcv(bfa_fcs_rport_t *rport, rport_event_t event)
{
	bfa_trc(rport->fcs, rport->pwwn);
	bfa_trc(rport->fcs, rport->pid);
	bfa_trc(rport->fcs, event);

	switch (event) {
	case RPSM_EVENT_HCB_OFFLINE:
		bfa_fcs_rport_send_logo_acc(rport);
		if (bfa_fcs_port_is_online(rport->port)) {
			bfa_sm_set_state(rport, bfa_fcs_rport_sm_nsdisc_sending);
			rport->ns_retries = 0;
			bfa_fcs_rport_send_gpnid(rport, NULL);
		} else {
			bfa_sm_set_state(rport, bfa_fcs_rport_sm_offline);
			bfa_hal_timer_start(rport->fcs->hal, &rport->timer,
				    bfa_fcs_rport_timeout, rport,
				    bfa_fcs_rport_del_timeout);
		}
		break;

	case RPSM_EVENT_DELETE:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_hcb_logosend);
		break;

	case RPSM_EVENT_LOGO_IMP:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_hcb_offline);
		break;

	case RPSM_EVENT_LOGO_RCVD:
		/**
		 * Ignore - already processing a LOGO.
		 */
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * 		Rport is being deleted. FC-4s are offline. Awaiting HAL rport offline
 * 		callback to send LOGO.
 */
static void
bfa_fcs_rport_sm_hcb_logosend(bfa_fcs_rport_t *rport, rport_event_t event)
{
	bfa_trc(rport->fcs, rport->pwwn);
	bfa_trc(rport->fcs, rport->pid);
	bfa_trc(rport->fcs, event);

	switch (event) {
	case RPSM_EVENT_HCB_OFFLINE:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_logo_sending);
		bfa_fcs_rport_send_logo(rport, NULL);
		break;

	case RPSM_EVENT_LOGO_RCVD:
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * 		Rport is being deleted. FC-4s are offline. LOGO is being sent.
 */
static void
bfa_fcs_rport_sm_logo_sending(bfa_fcs_rport_t *rport, rport_event_t event)
{
	bfa_trc(rport->fcs, rport->pwwn);
	bfa_trc(rport->fcs, rport->pid);
	bfa_trc(rport->fcs, event);

	switch (event) {
	case RPSM_EVENT_FCXP_SENT:
		/* Once LOGO is sent, we donot wait for the response */
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_uninit);
		bfa_fcs_rport_free(rport);
		break;

	case RPSM_EVENT_SCN:
		break;

	case RPSM_EVENT_LOGO_RCVD:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_uninit);
		bfa_hal_fcxp_walloc_cancel(rport->fcs->hal, &rport->fcxp_wqe);
		bfa_fcs_rport_free(rport);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * 		Rport is offline. FC-4s are offline. HAL rport is offline.
 * 		Timer active to delete stale rport.
 */
static void
bfa_fcs_rport_sm_offline(bfa_fcs_rport_t *rport, rport_event_t event)
{
	bfa_trc(rport->fcs, rport->pwwn);
	bfa_trc(rport->fcs, rport->pid);
	bfa_trc(rport->fcs, event);

	switch (event) {
	case RPSM_EVENT_TIMEOUT:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_uninit);
		bfa_fcs_rport_free(rport);
		break;

	case RPSM_EVENT_SCN:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_nsdisc_sending);
		bfa_hal_timer_stop(&rport->timer);
		rport->ns_retries = 0;
		bfa_fcs_rport_send_gpnid(rport, NULL);
		break;

	case RPSM_EVENT_DELETE:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_uninit);
		bfa_hal_timer_stop(&rport->timer);
		bfa_fcs_rport_free(rport);
		break;

	case RPSM_EVENT_PLOGI_RCVD:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_plogiacc_sending);
		bfa_hal_timer_stop(&rport->timer);
		bfa_fcs_rport_send_plogiacc(rport, NULL);
		break;

	case RPSM_EVENT_LOGO_RCVD:
	case RPSM_EVENT_LOGO_IMP:
		break;

	case RPSM_EVENT_PLOGI_COMP:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_hal_online);
		bfa_hal_timer_stop(&rport->timer);
		bfa_fcs_rport_hal_online(rport);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * 	Rport address has changed. Nameserver discovery request is being sent.
 */
static void
bfa_fcs_rport_sm_nsdisc_sending(bfa_fcs_rport_t *rport, rport_event_t event)
{
	bfa_trc(rport->fcs, rport->pwwn);
	bfa_trc(rport->fcs, rport->pid);
	bfa_trc(rport->fcs, event);

	switch (event) {
	case RPSM_EVENT_FCXP_SENT:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_nsdisc_sent);
		break;

	case RPSM_EVENT_DELETE:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_uninit);
		bfa_hal_fcxp_walloc_cancel(rport->fcs->hal, &rport->fcxp_wqe);
		bfa_fcs_rport_free(rport);
		break;

	case RPSM_EVENT_PLOGI_RCVD:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_plogiacc_sending);
		bfa_hal_fcxp_walloc_cancel(rport->fcs->hal, &rport->fcxp_wqe);
		bfa_fcs_rport_send_plogiacc(rport, NULL);
		break;

	case RPSM_EVENT_SCN:
	case RPSM_EVENT_LOGO_RCVD:
		break;

	case RPSM_EVENT_PLOGI_COMP:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_hal_online);
		bfa_hal_fcxp_walloc_cancel(rport->fcs->hal, &rport->fcxp_wqe);
		bfa_fcs_rport_hal_online(rport);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * 		Nameserver discovery failed. Waiting for timeout to retry.
 */
static void
bfa_fcs_rport_sm_nsdisc_retry(bfa_fcs_rport_t *rport, rport_event_t event)
{
	bfa_trc(rport->fcs, rport->pwwn);
	bfa_trc(rport->fcs, rport->pid);
	bfa_trc(rport->fcs, event);

	switch (event) {
	case RPSM_EVENT_TIMEOUT:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_nsdisc_sending);
		bfa_fcs_rport_send_gpnid(rport, NULL);
		break;

	case RPSM_EVENT_SCN:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_nsdisc_sending);
		bfa_hal_timer_stop(&rport->timer);
		rport->ns_retries = 0;
		bfa_fcs_rport_send_gpnid(rport, NULL);
		break;

	case RPSM_EVENT_DELETE:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_uninit);
		bfa_hal_timer_stop(&rport->timer);
		bfa_fcs_rport_free(rport);
		break;

	case RPSM_EVENT_PLOGI_RCVD:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_plogiacc_sending);
		bfa_hal_timer_stop(&rport->timer);
		bfa_fcs_rport_send_plogiacc(rport, NULL);
		break;

	case RPSM_EVENT_LOGO_IMP:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_offline);
		bfa_hal_timer_stop(&rport->timer);
		bfa_hal_timer_start(rport->fcs->hal, &rport->timer,
				    bfa_fcs_rport_timeout, rport,
				    bfa_fcs_rport_del_timeout);
		break;

	case RPSM_EVENT_LOGO_RCVD:
		bfa_fcs_rport_send_logo_acc(rport);
		break;

	case RPSM_EVENT_PLOGI_COMP:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_hal_online);
		bfa_hal_timer_stop(&rport->timer);
		bfa_fcs_rport_hal_online(rport);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * 		Rport address has changed. Nameserver discovery request is sent.
 */
static void
bfa_fcs_rport_sm_nsdisc_sent(bfa_fcs_rport_t *rport, rport_event_t event)
{
	bfa_trc(rport->fcs, rport->pwwn);
	bfa_trc(rport->fcs, rport->pid);
	bfa_trc(rport->fcs, event);

	switch (event) {
	case RPSM_EVENT_ACCEPTED:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_plogi_sending);
		bfa_fcs_rport_send_plogi(rport, NULL);
		break;

	case RPSM_EVENT_FAILED:
		rport->ns_retries++;
		if (rport->ns_retries < BFA_FCS_RPORT_MAX_RETRIES) {
			bfa_sm_set_state(rport, bfa_fcs_rport_sm_nsdisc_sending);
			bfa_fcs_rport_send_gpnid(rport, NULL);
		} else {
			bfa_sm_set_state(rport, bfa_fcs_rport_sm_offline);
			bfa_hal_timer_start(rport->fcs->hal, &rport->timer,
				    bfa_fcs_rport_timeout, rport,
				    bfa_fcs_rport_del_timeout);
		};
		break;

	case RPSM_EVENT_DELETE:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_uninit);
		bfa_hal_fcxp_discard(rport->fcxp);
		bfa_fcs_rport_free(rport);
		break;

	case RPSM_EVENT_PLOGI_RCVD:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_plogiacc_sending);
		bfa_hal_fcxp_discard(rport->fcxp);
		bfa_fcs_rport_send_plogiacc(rport, NULL);
		break;

	case RPSM_EVENT_LOGO_IMP:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_offline);
		bfa_hal_fcxp_discard(rport->fcxp);
		bfa_hal_timer_start(rport->fcs->hal, &rport->timer,
				    bfa_fcs_rport_timeout, rport,
				    bfa_fcs_rport_del_timeout);
		break;

	case RPSM_EVENT_SCN:
		/**
		 * ignore, wait for NS query response
		 */
		break;

	case RPSM_EVENT_LOGO_RCVD:
		/**
		 * Not logged-in yet. Accept LOGO.
		 */
		bfa_fcs_rport_send_logo_acc(rport);
		break;

	case RPSM_EVENT_PLOGI_COMP:
		bfa_sm_set_state(rport, bfa_fcs_rport_sm_hal_online);
		bfa_hal_fcxp_discard(rport->fcxp);
		bfa_fcs_rport_hal_online(rport);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @}
 */

/**
 * @dg fcs_rport_private FCS RPORT provate functions
 * @{
 */

static void
bfa_fcs_rport_send_plogi(void *rport_cbarg, bfa_hal_fcxp_t *fcxp_alloced)
{
	bfa_fcs_rport_t *rport = rport_cbarg;
	bfa_fcs_port_t *port = rport->port;
	fchs_t          fchs;
	int             len;
	bfa_hal_fcxp_t *fcxp;

	bfa_trc(rport->fcs, rport->pwwn);

	fcxp = fcxp_alloced ? fcxp_alloced : bfa_fcs_fcxp_alloc(port->fcs);
	if (!fcxp) {
		bfa_hal_fcxp_alloc_wait(port->fcs->hal, &rport->fcxp_wqe,
					bfa_fcs_rport_send_plogi, rport);
		return;
	}
	rport->fcxp = fcxp;

	len = fc_plogi_build(&fchs, bfa_hal_fcxp_get_reqbuf(fcxp), rport->pid,
			     bfa_fcs_port_get_fcid(port), 0,
			     port->port_cfg.pwwn, port->port_cfg.nwwn,
			     BFA_FCS_PORT_MAX_DATA_SIZE);

	bfa_hal_fcxp_send(fcxp, NULL, port->fabric->vf_id, BFA_FALSE,
			  FC_CLASS_3, len, &fchs, bfa_fcs_rport_plogi_response,
			  (void *)rport, FC_MAX_PDUSZ, FC_RA_TOV);

	rport->stats.plogis++;
	bfa_sm_send_event(rport, RPSM_EVENT_FCXP_SENT);
}

static void
bfa_fcs_rport_plogi_response(void *fcsarg, bfa_hal_fcxp_t *fcxp, void *cbarg,
			     bfa_status_t req_status, uint32_t rsp_len,
			     uint32_t resid_len, fchs_t *rsp_fchs)
{
	bfa_fcs_rport_t *rport = (bfa_fcs_rport_t *) cbarg;
	fc_plogi_t	*plogi_rsp;
	fc_ls_rjt_t	*ls_rjt;
	bfa_fcs_rport_t	*twin;
	bfa_q_t		*qe;

	bfa_trc(rport->fcs, rport->pwwn);

	/*
	 * Sanity Checks
	 */
	if (req_status != BFA_STATUS_OK) {
		bfa_trc(rport->fcs, req_status);
		rport->stats.plogi_failed++;
		bfa_sm_send_event(rport, RPSM_EVENT_FAILED);
		return;
	}

	plogi_rsp = (fc_plogi_t *) BFA_FCXP_RSP_PLD(fcxp);

	/**
	 * Check for failure first.
	 */
	if (plogi_rsp->els_cmd.els_code != FC_ELS_ACC) {
		ls_rjt = (fc_ls_rjt_t *) BFA_FCXP_RSP_PLD(fcxp);

		bfa_trc(rport->fcs, ls_rjt->reason_code);
		bfa_trc(rport->fcs, ls_rjt->reason_code_expl);

		rport->stats.plogi_rejects++;
		bfa_sm_send_event(rport, RPSM_EVENT_FAILED);
		return;
	}

	/**
	 * PLOGI is complete. Make sure this device is not one of the known
	 * device with a new FC port address.
	 */
	bfa_q_iter(&rport->port->rport_q, qe) {
		twin = (bfa_fcs_rport_t *) qe;
		if (twin == rport)
			continue;
		if (!rport->pwwn && (plogi_rsp->port_name == twin->pwwn)) {
			bfa_trc(rport->fcs, twin->pid);
			bfa_trc(rport->fcs, rport->pid);

			bfa_fcs_rport_delete(rport);

			bfa_fcs_rport_update(twin, plogi_rsp);
			twin->pid = rsp_fchs->s_id;
			bfa_sm_send_event(twin, RPSM_EVENT_PLOGI_COMP);
			return;
		}
	}

	/**
	 * Normal login path -- no evil twins.
	 */
	rport->stats.plogi_accs++;
	bfa_fcs_rport_update(rport, plogi_rsp);
	bfa_sm_send_event(rport, RPSM_EVENT_ACCEPTED);
}

static void
bfa_fcs_rport_send_plogiacc(void *rport_cbarg, bfa_hal_fcxp_t *fcxp_alloced)
{
	bfa_fcs_rport_t *rport = rport_cbarg;
	bfa_fcs_port_t *port = rport->port;
	fchs_t          fchs;
	int             len;
	bfa_hal_fcxp_t *fcxp;

	bfa_trc(rport->fcs, rport->pwwn);
	bfa_trc(rport->fcs, rport->reply_oxid);

	fcxp = fcxp_alloced ? fcxp_alloced : bfa_fcs_fcxp_alloc(port->fcs);
	if (!fcxp) {
		bfa_hal_fcxp_alloc_wait(port->fcs->hal, &rport->fcxp_wqe,
					bfa_fcs_rport_send_plogiacc, rport);
		return;
	}
	rport->fcxp = fcxp;

	len = fc_plogi_acc_build(&fchs, bfa_hal_fcxp_get_reqbuf(fcxp),
				 rport->pid, bfa_fcs_port_get_fcid(port),
				 rport->reply_oxid, port->port_cfg.pwwn,
				 port->port_cfg.nwwn,
				 BFA_FCS_PORT_MAX_DATA_SIZE);

	bfa_hal_fcxp_send(fcxp, NULL, port->fabric->vf_id, BFA_FALSE,
			  FC_CLASS_3, len, &fchs, NULL, NULL, FC_MAX_PDUSZ, 0);

	bfa_sm_send_event(rport, RPSM_EVENT_FCXP_SENT);
}

static void
bfa_fcs_rport_send_adisc(void *rport_cbarg, bfa_hal_fcxp_t *fcxp_alloced)
{
	bfa_fcs_rport_t *rport = rport_cbarg;
	bfa_fcs_port_t *port = rport->port;
	fchs_t          fchs;
	int             len;
	bfa_hal_fcxp_t *fcxp;

	bfa_trc(rport->fcs, rport->pwwn);

	fcxp = fcxp_alloced ? fcxp_alloced : bfa_fcs_fcxp_alloc(port->fcs);
	if (!fcxp) {
		bfa_hal_fcxp_alloc_wait(port->fcs->hal, &rport->fcxp_wqe,
					bfa_fcs_rport_send_adisc, rport);
		return;
	}
	rport->fcxp = fcxp;

	len = fc_adisc_build(&fchs, bfa_hal_fcxp_get_reqbuf(fcxp), rport->pid,
			     bfa_fcs_port_get_fcid(port), 0,
			     port->port_cfg.pwwn, port->port_cfg.nwwn);

	bfa_hal_fcxp_send(fcxp, NULL, port->fabric->vf_id, BFA_FALSE,
			  FC_CLASS_3, len, &fchs, bfa_fcs_rport_adisc_response,
			  rport, FC_MAX_PDUSZ, FC_RA_TOV);

	rport->stats.adisc_sent++;
	bfa_sm_send_event(rport, RPSM_EVENT_FCXP_SENT);
}

static void
bfa_fcs_rport_adisc_response(void *fcsarg, bfa_hal_fcxp_t *fcxp, void *cbarg,
			     bfa_status_t req_status, uint32_t rsp_len,
			     uint32_t resid_len, fchs_t *rsp_fchs)
{
	bfa_fcs_rport_t *rport = (bfa_fcs_rport_t *) cbarg;
	void           *pld = bfa_hal_fcxp_get_rspbuf(fcxp);
	fc_ls_rjt_t    *ls_rjt;

	if (req_status != BFA_STATUS_OK) {
		bfa_trc(rport->fcs, req_status);
		rport->stats.adisc_failed++;
		bfa_sm_send_event(rport, RPSM_EVENT_FAILED);
		return;
	}

	if (fc_adisc_rsp_parse((fc_adisc_t *)pld, rsp_len, rport->pwwn, 
						rport->nwwn)  == FC_PARSE_OK) {
		rport->stats.adisc_accs++;
		bfa_sm_send_event(rport, RPSM_EVENT_ACCEPTED);
		return;
	}

	rport->stats.adisc_rejects++;
	ls_rjt = pld;
	bfa_trc(rport->fcs, ls_rjt->els_cmd.els_code);
	bfa_trc(rport->fcs, ls_rjt->reason_code);
	bfa_trc(rport->fcs, ls_rjt->reason_code_expl);
	bfa_sm_send_event(rport, RPSM_EVENT_FAILED);
}

#ifdef RPSC_TEST

static void
bfa_fcs_rport_send_rpsc(void *rport_cbarg, bfa_hal_fcxp_t *fcxp_alloced)
{
	bfa_fcs_rport_t *rport = rport_cbarg;
	bfa_fcs_port_t *port = rport->port;
	fchs_t          fchs;
	int             len;
	bfa_hal_fcxp_t *fcxp;

	bfa_trc(rport->fcs, rport->pwwn);

	fcxp = fcxp_alloced ? fcxp_alloced : bfa_fcs_fcxp_alloc(port->fcs);
	if (!fcxp) {
		bfa_hal_fcxp_alloc_wait(port->fcs->hal, &rport->fcxp_wqe,
					bfa_fcs_rport_send_rpsc, rport);
		return;
	}
	rport->fcxp = fcxp;

	len = fc_rpsc_build(&fchs, bfa_hal_fcxp_get_reqbuf(fcxp), rport->pid,
			    bfa_fcs_port_get_fcid(port), 0);

	bfa_hal_fcxp_send(fcxp, NULL, port->fabric->vf_id, BFA_FALSE,
			  FC_CLASS_3, len, &fchs, bfa_fcs_rport_rpsc_response,
			  rport, FC_MAX_PDUSZ, FC_RA_TOV);

}

static void
bfa_fcs_rport_rpsc_response(void *fcsarg, bfa_hal_fcxp_t *fcxp, void *cbarg,
			    bfa_status_t req_status, uint32_t rsp_len,
			    uint32_t resid_len, fchs_t *rsp_fchs)
{
	bfa_fcs_rport_t *rport = (bfa_fcs_rport_t *) cbarg;
	fc_ls_rjt_t    *ls_rjt;
	fc_rpsc_acc_t  *rpsc_acc;
	uint16_t        num_ents;
	fc_rpsc_speed_info_t *speed_info;

	bfa_trc(rport->fcs, req_status);

	if (req_status != BFA_STATUS_OK) {
		bfa_trc(rport->fcs, req_status);
		return;
	}

	rpsc_acc = (fc_rpsc_acc_t *) BFA_FCXP_RSP_PLD(fcxp);
	if (rpsc_acc->command == FC_ELS_ACC) {
		num_ents = bfa_os_ntohs(rpsc_acc->num_entries);
		bfa_trc(rport->fcs, num_ents);
		if (num_ents > 0) {
			speed_info = rpsc_acc->speed_info;
			bfa_trc(rport->fcs,
				bfa_os_ntohs(speed_info->port_speed_cap));
			bfa_trc(rport->fcs,
				bfa_os_ntohs(speed_info->port_op_speed));
		}
	} else {
		ls_rjt = (fc_ls_rjt_t *) BFA_FCXP_RSP_PLD(fcxp);
		bfa_trc(rport->fcs, ls_rjt->reason_code);
		bfa_trc(rport->fcs, ls_rjt->reason_code_expl);

	}
}
#endif /* RPSC_TEST */

static void
bfa_fcs_rport_send_gpnid(void *rport_cbarg, bfa_hal_fcxp_t *fcxp_alloced)
{
	bfa_fcs_rport_t *rport = rport_cbarg;
	bfa_fcs_port_t *port = rport->port;
	fchs_t          fchs;
	bfa_hal_fcxp_t *fcxp;
	int             len;

	bfa_trc(rport->fcs, rport->pid);

	fcxp = fcxp_alloced ? fcxp_alloced : bfa_fcs_fcxp_alloc(port->fcs);
	if (!fcxp) {
		bfa_hal_fcxp_alloc_wait(port->fcs->hal, &rport->fcxp_wqe,
					bfa_fcs_rport_send_gpnid, rport);
		return;
	}
	rport->fcxp = fcxp;

	len = fc_gpnid_build(&fchs, bfa_hal_fcxp_get_reqbuf(fcxp),
			     bfa_fcs_port_get_fcid(port), 0, rport->pid);

	bfa_hal_fcxp_send(fcxp, NULL, port->fabric->vf_id, BFA_FALSE,
			  FC_CLASS_3, len, &fchs, bfa_fcs_rport_gpnid_response,
			  (void *)rport, FC_MAX_PDUSZ, FC_RA_TOV);

	bfa_sm_send_event(rport, RPSM_EVENT_FCXP_SENT);
}

static void
bfa_fcs_rport_gpnid_response(void *fcsarg, bfa_hal_fcxp_t *fcxp, void *cbarg,
			     bfa_status_t req_status, uint32_t rsp_len,
			     uint32_t resid_len, fchs_t *rsp_fchs)
{
	bfa_fcs_rport_t *rport = (bfa_fcs_rport_t *) cbarg;
	ct_hdr_t       *cthdr;
	wwn_t          *gpnid_rsp;

	bfa_trc(rport->fcs, rport->pwwn);

	cthdr = (ct_hdr_t *) BFA_FCXP_RSP_PLD(fcxp);
	cthdr->cmd_rsp_code = bfa_os_ntohs(cthdr->cmd_rsp_code);

	if (cthdr->cmd_rsp_code == CT_RSP_ACCEPT) {
		/*
		 * Check if the pwwn is the same as in rport
		 */
		gpnid_rsp = (wwn_t *) (cthdr + 1);

		if (*gpnid_rsp == rport->pwwn) {
			/*
			 * device is online
			 */
			bfa_sm_send_event(rport, RPSM_EVENT_ACCEPTED);
		} else {
			/*
			 * It is a different device.
			 * TBD : WIll need special handling for this case ??
			 */
			bfa_sm_send_event(rport, RPSM_EVENT_FAILED);
		}
		return;
	}

	/*
	 * Reject Response
	 */
	switch (cthdr->reason_code) {
	case CT_RSN_LOGICAL_BUSY:
		/*
		 * Need to retry
		 */
		bfa_sm_send_event(rport, RPSM_EVENT_TIMEOUT);
		break;

	case CT_RSN_UNABLE_TO_PERF:
		/*
		 * device doesn't exist : Start timer to cleanup this later.
		 */
		bfa_sm_send_event(rport, RPSM_EVENT_FAILED);
		break;

	default:
		bfa_sm_send_event(rport, RPSM_EVENT_FAILED);
		break;
	}
}

/**
 * @brief
 *    Called to send a logout to the rport.
 */
static void
bfa_fcs_rport_send_logo(void *rport_cbarg, bfa_hal_fcxp_t *fcxp_alloced)
{
	bfa_fcs_rport_t *rport = rport_cbarg;
	bfa_fcs_port_t *port;
	fchs_t          fchs;
	bfa_hal_fcxp_t *fcxp;
	uint16_t        len;

	bfa_trc(rport->fcs, rport->pid);

	port = rport->port;

	fcxp = fcxp_alloced ? fcxp_alloced : bfa_fcs_fcxp_alloc(port->fcs);
	if (!fcxp) {
		bfa_hal_fcxp_alloc_wait(port->fcs->hal, &rport->fcxp_wqe,
					bfa_fcs_rport_send_logo, rport);
		return;
	}
	rport->fcxp = fcxp;

	len = fc_logo_build(&fchs, bfa_hal_fcxp_get_reqbuf(fcxp), rport->pid,
			    bfa_fcs_port_get_fcid(port), 0,
			    bfa_fcs_port_get_pwwn(port));

	bfa_hal_fcxp_send(fcxp, NULL, port->fabric->vf_id, BFA_FALSE,
			  FC_CLASS_3, len, &fchs, NULL,
			  rport, FC_MAX_PDUSZ, FC_ED_TOV);

	rport->stats.logos++;
	bfa_hal_fcxp_discard(rport->fcxp);
	bfa_sm_send_event(rport, RPSM_EVENT_FCXP_SENT);
}

/**
 * @brief
 *    Send ACC for a LOGO received.
 */
static void
bfa_fcs_rport_send_logo_acc(void *rport_cbarg)
{
	bfa_fcs_rport_t *rport = rport_cbarg;
	bfa_fcs_port_t *port;
	fchs_t          fchs;
	bfa_hal_fcxp_t *fcxp;
	uint16_t        len;

	bfa_trc(rport->fcs, rport->pid);

	port = rport->port;

	fcxp = bfa_fcs_fcxp_alloc(port->fcs);
	if (!fcxp)
		return;

	rport->stats.logo_rcvd++;
	len = fc_logo_acc_build(&fchs, bfa_hal_fcxp_get_reqbuf(fcxp),
				rport->pid, bfa_fcs_port_get_fcid(port),
				rport->reply_oxid);

	bfa_hal_fcxp_send(fcxp, NULL, port->fabric->vf_id, BFA_FALSE,
			  FC_CLASS_3, len, &fchs, NULL, NULL, FC_MAX_PDUSZ, 0);
}

/**
 * 	@brief
 *     This routine will be called by bfa_timer on timer timeouts.
 *
 * 	@param[in] 	rport 			- pointer to bfa_fcs_port_ns_t.
 * 	@param[out]	rport_status 	- pointer to return vport status in
 *
 * 	@return
 * 		void
 *
* 	Special Considerations:
 *
 * 	@note
 */
static void
bfa_fcs_rport_timeout(void *arg)
{
	bfa_fcs_rport_t *rport = (bfa_fcs_rport_t *) arg;

	rport->stats.plogi_timeouts++;
	bfa_sm_send_event(rport, RPSM_EVENT_TIMEOUT);
}

static void
bfa_fcs_rport_process_prli(bfa_fcs_rport_t *rport, fchs_t *rx_fchs,
			   uint16_t len)
{
	bfa_hal_fcxp_t *fcxp;
	fchs_t          fchs;
	bfa_fcs_port_t *port = rport->port;
	fc_prli_t      *prli;

	bfa_trc(port->fcs, rx_fchs->s_id);
	bfa_trc(port->fcs, rx_fchs->d_id);

	rport->stats.prli_rcvd++;

	if (BFA_FCS_VPORT_IS_TARGET_MODE(port)) {
		/*
		 * Target Mode : Let the fcptm handle it
		 */
		bfa_fcs_itntm_rx_prli(rport->itntm, rx_fchs, len);
		return;
	}

	/*
	 * We are either in Initiator or ipfc Mode
	 */
	prli = (fc_prli_t *) (rx_fchs + 1);

	if (prli->parampage.servparams.initiator) {
		bfa_trc(rport->fcs, prli->parampage.type);
		rport->scsi_function = BFA_RPORT_INITIATOR;
		bfa_fcs_itnim_is_initiator(rport->itnim);
	} else {
		/*
		 * @todo: PRLI from a target ?
		 */
		bfa_trc(port->fcs, rx_fchs->s_id);
		rport->scsi_function = BFA_RPORT_TARGET;
	}

	fcxp = bfa_fcs_fcxp_alloc(port->fcs);
	if (!fcxp)
		return;

	len = fc_prli_acc_build(&fchs, bfa_hal_fcxp_get_reqbuf(fcxp),
				rx_fchs->s_id, bfa_fcs_port_get_fcid(port),
				rx_fchs->ox_id, port->port_cfg.roles);

	bfa_hal_fcxp_send(fcxp, NULL, port->fabric->vf_id, BFA_FALSE,
			  FC_CLASS_3, len, &fchs, NULL, NULL, FC_MAX_PDUSZ, 0);
}

static void
bfa_fcs_rport_process_rpsc(bfa_fcs_rport_t *rport, fchs_t *rx_fchs,
			   uint16_t len)
{
	bfa_hal_fcxp_t *fcxp;
	fchs_t          fchs;
	bfa_fcs_port_t *port = rport->port;
	fc_rpsc_speed_info_t speeds;
	bfa_pport_attr_t pport_attr;

	bfa_trc(port->fcs, rx_fchs->s_id);
	bfa_trc(port->fcs, rx_fchs->d_id);

	speeds.port_speed_cap =
		RPSC_SPEED_CAP_1G | RPSC_SPEED_CAP_2G | RPSC_SPEED_CAP_4G |
		RPSC_SPEED_CAP_8G;

	/*
	 * get curent speed from pport attributes from hal
	 */
	bfa_hal_pport_get_attr(port->fcs->hal, &pport_attr);

	speeds.port_op_speed = fc_bfa_speed_to_rpsc_operspeed(pport_attr.speed);

	fcxp = bfa_fcs_fcxp_alloc(port->fcs);
	if (!fcxp)
		return;

	len = fc_rpsc_acc_build(&fchs, bfa_hal_fcxp_get_reqbuf(fcxp),
				rx_fchs->s_id, bfa_fcs_port_get_fcid(port),
				rx_fchs->ox_id, &speeds);

	bfa_hal_fcxp_send(fcxp, NULL, port->fabric->vf_id, BFA_FALSE,
			  FC_CLASS_3, len, &fchs, NULL, NULL, FC_MAX_PDUSZ, 0);
}

static void
bfa_fcs_rport_process_adisc(bfa_fcs_rport_t *rport, fchs_t *rx_fchs,
			   uint16_t len)
{
	bfa_hal_fcxp_t *fcxp;
	fchs_t          fchs;
	bfa_fcs_port_t *port = rport->port;
	fc_adisc_t      *adisc;

	bfa_trc(port->fcs, rx_fchs->s_id);
	bfa_trc(port->fcs, rx_fchs->d_id);

	rport->stats.adisc_rcvd++;

	if (BFA_FCS_VPORT_IS_TARGET_MODE(port)) {
		/* @todo :  Target Mode  handling */
		bfa_trc(port->fcs, rx_fchs->d_id);
		bfa_assert(0);
		return;
	}

	adisc = (fc_adisc_t *) (rx_fchs + 1);

	/* Accept if the itnim for this rport is online. Else reject the ADISC */
	if (bfa_fcs_itnim_get_online_state(rport->itnim) == BFA_STATUS_OK ) {

		fcxp = bfa_fcs_fcxp_alloc(port->fcs);
		if (!fcxp)
			return;

		len = fc_adisc_acc_build(&fchs, bfa_hal_fcxp_get_reqbuf(fcxp),
					rx_fchs->s_id, bfa_fcs_port_get_fcid(port),
					rx_fchs->ox_id, port->port_cfg.pwwn, port->port_cfg.nwwn);

		bfa_hal_fcxp_send(fcxp, NULL, port->fabric->vf_id, BFA_FALSE,
				  FC_CLASS_3, len, &fchs, NULL, NULL, FC_MAX_PDUSZ, 0);
	} else {
		rport->stats.adisc_rejected++;
		bfa_fcs_rport_send_ls_rjt(rport, rx_fchs,
					  FC_LS_RJT_RSN_UNABLE_TO_PERF_CMD,
					  FC_LS_RJT_EXP_LOGIN_REQUIRED);
	}
	
}

static void
bfa_fcs_rport_hal_online(bfa_fcs_rport_t *rport)
{
	bfa_fcs_port_t *port = rport->port;
	bfa_hal_rport_info_t rport_info;

	rport_info.pid = rport->pid;
	rport_info.local_pid = port->pid;
	rport_info.vf_id = port->fabric->vf_id;
	rport_info.vf_en = port->fabric->is_vf;
	rport_info.fc_class = rport->fc_cos;
	rport_info.cisc = rport->cisc;
	rport_info.max_frmsz = rport->maxfrsize;
	bfa_hal_rport_online(rport->hal_rport, &rport_info);
}

static void
bfa_fcs_rport_fc4_pause(bfa_fcs_rport_t *rport)
{
	if (bfa_fcs_port_is_initiator(rport->port))
		bfa_fcs_itnim_pause(rport->itnim);

	if (bfa_fcs_port_is_target(rport->port))
		bfa_fcs_itntm_pause(rport->itntm);
}

static void
bfa_fcs_rport_fc4_resume(bfa_fcs_rport_t *rport)
{
	if (bfa_fcs_port_is_initiator(rport->port))
		bfa_fcs_itnim_resume(rport->itnim);

	if (bfa_fcs_port_is_target(rport->port))
		bfa_fcs_itntm_resume(rport->itntm);
}

static bfa_fcs_rport_t *
bfa_fcs_rport_alloc(bfa_fcs_port_t *port, wwn_t pwwn, uint32_t rpid)
{
	bfa_fcs_t      *fcs = port->fcs;
	bfa_fcs_rport_t *rport;
	bfad_rport_t   *rport_drv;

	/**
	 * allocate rport
	 */
	if (bfa_fcb_rport_alloc(fcs->bfad, &rport, &rport_drv)
			!= BFA_STATUS_OK) {
		bfa_trc(fcs, rpid);
		return NULL;
	}

	/*
	 * Initialize r-port
	 */
	rport->port = port;
	rport->fcs = fcs;
	rport->rp_drv = rport_drv;
	rport->pid = rpid;
	rport->pwwn = pwwn;

	/**
	 * allocate HAL rport
	 */
	rport->hal_rport = bfa_hal_rport_create(port->fcs->hal, rport);
	if (!rport->hal_rport) {
		bfa_trc(fcs, rpid);
		bfa_fcb_rport_free(fcs->bfad, &rport_drv);
		return NULL;
	}

	/**
	 * allocate FC-4s
	 */
	bfa_assert(bfa_fcs_port_is_initiator(port) ^
		   bfa_fcs_port_is_target(port));

	if (bfa_fcs_port_is_initiator(port)) {
		rport->itnim = bfa_fcs_itnim_create(rport);
		if (!rport->itnim) {
			bfa_trc(fcs, rpid);
			bfa_hal_rport_delete(rport->hal_rport);
			bfa_fcb_rport_free(fcs->bfad, &rport_drv);
			return NULL;
		}
	}

	if (bfa_fcs_port_is_target(port)) {
		rport->itntm = bfa_fcs_itntm_create(rport);
		if (!rport->itntm) {
			bfa_trc(fcs, rpid);
			bfa_hal_rport_delete(rport->hal_rport);
			bfa_fcb_rport_free(fcs->bfad, &rport_drv);
			return NULL;
		}
	}

	bfa_fcs_port_add_rport(port, rport);

	bfa_sm_set_state(rport, bfa_fcs_rport_sm_uninit);

	return rport;
}

/**
 * @brief Delete an rport.
 */
static void
bfa_fcs_rport_free(bfa_fcs_rport_t *rport)
{
	bfa_fcs_port_t *port = rport->port;

	/**
	 * - delete FC-4s
	 * - delete HAL rport
	 * - remove from queue of rports
	 */
	if (bfa_fcs_port_is_initiator(port))
		bfa_fcs_itnim_delete(rport->itnim);

	if (bfa_fcs_port_is_target(port))
		bfa_fcs_itntm_delete(rport->itntm);

	bfa_hal_rport_delete(rport->hal_rport);
	bfa_fcs_port_del_rport(port, rport);
	bfa_fcb_rport_free(rport->fcs->bfad, &rport->rp_drv);
}

static void
bfa_fcs_rport_aen_post(bfa_fcs_rport_t *rport, bfa_rport_aen_event_t event)
{
	bfa_aen_data_t  aen_data;
	bfa_log_mod_t	*logmod = rport->fcs->logm;
	wwn_t		lpwwn = bfa_fcs_port_get_pwwn(rport->port);
	wwn_t		rpwwn = rport->pwwn;
	char		lpwwn_buf[BFA_STRING_32];
	char		rpwwn_buf[BFA_STRING_32];
	char		*lpwwn_ptr;
	char		*rpwwn_ptr;

    lpwwn_ptr = wwn2str(lpwwn_buf, sizeof(lpwwn_buf), lpwwn);
    rpwwn_ptr = wwn2str(rpwwn_buf, sizeof(rpwwn_buf), rpwwn);

	switch (event) {
	case BFA_RPORT_AEN_ONLINE:
		bfa_log(logmod, BFAL_AEN_RPORT_ONLINE, rpwwn_ptr, lpwwn_ptr);
		break;
	case BFA_RPORT_AEN_OFFLINE:
		bfa_log(logmod, BFAL_AEN_RPORT_OFFLINE, rpwwn_ptr, lpwwn_ptr);
		break;
	case BFA_RPORT_AEN_DISCONNECT:
		bfa_log(logmod, BFAL_AEN_RPORT_DISCONNECT, rpwwn_ptr, lpwwn_ptr);
		break;
	default: 
		break;
	}

	aen_data.rport.vf_id = rport->port->fabric->vf_id;
	aen_data.rport.lpwwn = lpwwn;
	aen_data.rport.rpwwn = rpwwn;
	bfa_aen_post(rport->fcs->aen, BFA_AEN_CAT_RPORT, event, &aen_data);
}

static void
bfa_fcs_rport_online_action(bfa_fcs_rport_t *rport)
{
	bfa_fcs_port_t *port = rport->port;

	rport->stats.onlines++;

	if (bfa_fcs_port_is_initiator(port))
		bfa_fcs_itnim_rport_online(rport->itnim);

	if (bfa_fcs_port_is_target(port))
		bfa_fcs_itntm_rport_online(rport->itntm);

	/* Don't post events for well known addresses */
	if (bfa_os_ntoh3b(rport->pid) < 0xFFF000)  
		bfa_fcs_rport_aen_post(rport, BFA_RPORT_AEN_ONLINE);
}

static void
bfa_fcs_rport_offline_action(bfa_fcs_rport_t *rport)
{
	bfa_fcs_port_t *port = rport->port;

	rport->stats.offlines++;

	/* Don't post events for well known addresses */
	if (bfa_os_ntoh3b(rport->pid) < 0xFFF000) {
		if (bfa_fcs_port_is_online(rport->port) == BFA_TRUE) {
			bfa_fcs_rport_aen_post(rport, BFA_RPORT_AEN_DISCONNECT);
		} else {
			bfa_fcs_rport_aen_post(rport, BFA_RPORT_AEN_OFFLINE);
		}
	}

	if (bfa_fcs_port_is_initiator(port))
		bfa_fcs_itnim_rport_offline(rport->itnim);

	if (bfa_fcs_port_is_target(port))
		bfa_fcs_itntm_rport_offline(rport->itntm);
}

/**
 * Update rport parameters from PLOGI or PLOGI accept.
 */
static void
bfa_fcs_rport_update(bfa_fcs_rport_t *rport, fc_plogi_t *plogi)
{
	/**
	 * - port name
	 * - node name
	 */
	rport->pwwn = plogi->port_name;
	rport->nwwn = plogi->node_name;

	/**
	 * - class of service
	 */
	rport->fc_cos = 0;
	if (plogi->class3.class_valid)
		rport->fc_cos = FC_CLASS_3;

	if (plogi->class2.class_valid)
		rport->fc_cos |= FC_CLASS_2;

	/**
	 * - CISC
	 * - MAX receive frame size
	 */
	rport->cisc = plogi->csp.cisc;
	rport->maxfrsize = bfa_os_ntohs(plogi->class3.rxsz);
}

/**
 * @brief
 *   Called to handle LOGO received from an existing remote port.
 */
static void
bfa_fcs_rport_process_logo(bfa_fcs_rport_t *rport, fchs_t *fchs)
{
	rport->reply_oxid = fchs->ox_id;
	bfa_trc(rport->fcs, rport->reply_oxid);

	rport->stats.logo_rcvd++;
	bfa_sm_send_event(rport, RPSM_EVENT_LOGO_RCVD);
}

/**
 * @}
 */

/**
 * @dg fcs_rport_public FCS rport public interfaces
 * @{
 */

/**
 * @brief
 * 	Called by bport/vport to create a remote port instance for a discovered
 * 	remote device.
 *
 * @param[in] port	- base port or vport
 * @param[in] rpid	- remote port ID
 *
 * @return None
 */
bfa_fcs_rport_t *
bfa_fcs_rport_create(bfa_fcs_port_t *port, uint32_t rpid)
{
	bfa_fcs_rport_t *rport;

	rport = bfa_fcs_rport_alloc(port, WWN_NULL, rpid);
	if (!rport)
		return NULL;

	bfa_sm_send_event(rport, RPSM_EVENT_PLOGI_SEND);
	return rport;
}

/**
 * @brief
 * Called by bport in private loop topology to indicate that a
 * rport has been discovered and plogi has been completed.
 *
 * @param[in] port	- base port or vport
 * @param[in] rpid	- remote port ID
 */
void
bfa_fcs_rport_start(bfa_fcs_port_t *port, fchs_t *fchs, fc_plogi_t *plogi)
{
	bfa_fcs_rport_t *rport;

	rport = bfa_fcs_rport_alloc(port, WWN_NULL, fchs->s_id);
	if (!rport)
		return;

	bfa_fcs_rport_update(rport, plogi);

	bfa_sm_send_event(rport, RPSM_EVENT_PLOGI_COMP);
}

/**
 * @brief
 *   Called by bport/vport to handle PLOGI received from a new remote port.
 *   If an existing rport does a plogi, it will be handled separately.
 */
void
bfa_fcs_rport_plogi_create(bfa_fcs_port_t *port, fchs_t *fchs,
			   fc_plogi_t *plogi)
{
	bfa_fcs_rport_t *rport;

	rport = bfa_fcs_rport_alloc(port, plogi->port_name, fchs->s_id);
	if (!rport)
		return;

	bfa_fcs_rport_update(rport, plogi);

	rport->reply_oxid = fchs->ox_id;
	bfa_trc(rport->fcs, rport->reply_oxid);

	rport->stats.plogi_rcvd++;
	bfa_sm_send_event(rport, RPSM_EVENT_PLOGI_RCVD);
}

static int
wwn_compare(wwn_t wwn1, wwn_t wwn2)
{
	uint8_t        *b1 = (uint8_t *) &wwn1;
	uint8_t        *b2 = (uint8_t *) &wwn2;
	int             i;

	for (i = 0; i < sizeof(wwn_t); i++) {
		if (b1[i] < b2[i])
			return -1;
		if (b1[i] > b2[i])
			return 1;
	}
	return 0;
}

/**
 * @brief
 *   Called by bport/vport to handle PLOGI received from an existing
 * 	 remote port.
 */
void
bfa_fcs_rport_plogi(bfa_fcs_rport_t *rport, fchs_t *rx_fchs,
		    fc_plogi_t *plogi)
{
	/**
	 * @todo Handle P2P and initiator-initiator.
	 */

	bfa_fcs_rport_update(rport, plogi);

	rport->reply_oxid = rx_fchs->ox_id;
	bfa_trc(rport->fcs, rport->reply_oxid);

	/**
	 * In Switched fabric topology,
	 * PLOGI to each other. If our pwwn is smaller, ignore.
	 * If the link topology is N2N,
	 * this Plogi should be accepted.
	 */
	if ((wwn_compare(rport->port->port_cfg.pwwn, rport->pwwn) == -1)
	    && (bfa_fcs_fabric_is_switched(rport->port->fabric)))
		return;

	rport->stats.plogi_rcvd++;
	bfa_sm_send_event(rport, RPSM_EVENT_PLOGI_RCVD);
}

/**
 * @brief
 * Called by bport/vport to delete a remote port instance.
 *
 * Rport delete is called under the following conditions:
 * 		- vport is deleted
 * 		- vf is deleted
 * 		- explicit request from OS to delete rport (vmware)
 */
void
bfa_fcs_rport_delete(bfa_fcs_rport_t *rport)
{
	bfa_sm_send_event(rport, RPSM_EVENT_DELETE);
}

/**
 * @brief
 * Called by bport/vport to  when a target goes offline.
 *
 */
void
bfa_fcs_rport_offline(bfa_fcs_rport_t *rport)
{
	bfa_sm_send_event(rport, RPSM_EVENT_LOGO_IMP);
}

/**
 * @brief
 *   Called by bport/vport to notify SCN for the remote port
 */
void
bfa_fcs_rport_scn(bfa_fcs_rport_t *rport)
{
	bfa_sm_send_event(rport, RPSM_EVENT_SCN);
}

/**
 * @brief
 *   Called by  fcpim to notify that the ITN cleanup is done.
 */
void
bfa_fcs_rport_itnim_ack(bfa_fcs_rport_t *rport)
{
	bfa_sm_send_event(rport, RPSM_EVENT_FC4_OFFLINE);
}

/**
 * @brief
 *   Called by fcptm to notify that the ITN cleanup is done.
 */
void
bfa_fcs_rport_itntm_ack(bfa_fcs_rport_t *rport)
{
	bfa_sm_send_event(rport, RPSM_EVENT_FC4_OFFLINE);
}

/**
 * 	@brief
 *     This routine HAL callback for bfa_hal_rport_online() call.
 *
 * 	@param[in] 	cb_arg	-  rport struct.
 *
 * 	@return
 * 		void
 *
 * 	Special Considerations:
 *
 * 	@note
 */
void
bfa_hcb_rport_online(void *cbarg)
{

	bfa_fcs_rport_t *rport = (bfa_fcs_rport_t *) cbarg;

	bfa_trc(rport->fcs, rport->pwwn);
	bfa_sm_send_event(rport, RPSM_EVENT_HCB_ONLINE);
}

/**
 * 	@brief
 *     This routine HAL callback for bfa_hal_rport_offline() call.
 *
 * 	@param[in] 	rport 	-
 *
 * 	@return
 * 		void
 *
 * 	Special Considerations:
 *
 * 	@note
 */
void
bfa_hcb_rport_offline(void *cbarg)
{
	bfa_fcs_rport_t *rport = (bfa_fcs_rport_t *) cbarg;

	bfa_trc(rport->fcs, rport->pwwn);
	bfa_sm_send_event(rport, RPSM_EVENT_HCB_OFFLINE);
}

/**
 * @brief
 * 		Called to process any unsolicted frames from this remote port
 */
void
bfa_fcs_rport_logo_imp(bfa_fcs_rport_t *rport)
{
	bfa_sm_send_event(rport, RPSM_EVENT_LOGO_IMP);
}

/**
 * @brief
 * 		Called to process any unsolicted frames from this remote port
 */
void
bfa_fcs_rport_uf_recv(bfa_fcs_rport_t *rport, fchs_t *fchs, uint16_t len)
{
	bfa_fcs_port_t *port = rport->port;
	fc_els_cmd_t   *els_cmd;

	bfa_trc(rport->fcs, fchs->s_id);
	bfa_trc(rport->fcs, fchs->d_id);
	bfa_trc(rport->fcs, fchs->type);

	if (fchs->type != FC_TYPE_ELS)
		return;

	els_cmd = (fc_els_cmd_t *) (fchs + 1);

	bfa_trc(rport->fcs, els_cmd->els_code);

	switch (els_cmd->els_code) {
	case FC_ELS_LOGO:
		bfa_fcs_rport_process_logo(rport, fchs);
		break;

	case FC_ELS_ADISC:
		bfa_fcs_rport_process_adisc(rport, fchs, len);
		break;

	case FC_ELS_PRLO:
		if (bfa_fcs_port_is_initiator(port))
			bfa_fcs_fcpim_uf_recv(rport->itnim, fchs, len);

		if (bfa_fcs_port_is_target(port))
			bfa_fcs_fcptm_uf_recv(rport->itntm, fchs, len);
		break;

	case FC_ELS_PRLI:
		bfa_fcs_rport_process_prli(rport, fchs, len);
		break;

	case FC_ELS_RPSC:
		bfa_fcs_rport_process_rpsc(rport, fchs, len);
		break;

	case FC_ELS_SCR:
		bfa_fcs_rport_send_ls_rjt(rport, fchs,
					  FC_LS_RJT_RSN_CMD_NOT_SUPP,
					  FC_LS_RJT_EXP_NO_ADDL_INFO);
		break;

	default:
		break;
	}
}

/*
 * @brief
 * Send a LS reject
 */
static void
bfa_fcs_rport_send_ls_rjt(bfa_fcs_rport_t *rport, fchs_t *rx_fchs,
			  uint8_t reason_code, uint8_t reason_code_expl)
{
	bfa_fcs_port_t *port = rport->port;
	fchs_t          fchs;
	bfa_hal_fcxp_t *fcxp;
	int             len;

	bfa_trc(rport->fcs, rx_fchs->s_id);

	fcxp = bfa_fcs_fcxp_alloc(rport->fcs);
	if (!fcxp)
		return;

	len = fc_ls_rjt_build(&fchs, bfa_hal_fcxp_get_reqbuf(fcxp),
			      rx_fchs->s_id, bfa_fcs_port_get_fcid(port),
			      rx_fchs->ox_id, reason_code, reason_code_expl);

	bfa_hal_fcxp_send(fcxp, NULL, port->fabric->vf_id,
			  BFA_FALSE, FC_CLASS_3, len, &fchs, NULL, NULL,
			  FC_MAX_PDUSZ, 0);
}

/**
 * @brief
 *   Module initialization
 */
void
bfa_fcs_rport_modinit(bfa_fcs_t *fcs)
{
}

/**
 * @brief
 *   Module cleanup
 */
void
bfa_fcs_rport_modexit(bfa_fcs_t *fcs)
{
	bfa_fcs_modexit_comp(fcs);
}

/**
 * Return state of rport.
 */
int
bfa_fcs_rport_get_state(bfa_fcs_rport_t *rport)
{
	return bfa_sm_to_state(rport_sm_table, rport->sm);
}

/**
 * 	@brief
 * 		 Called by the Driver to set rport delete/ageout timeout
 *
 * 	@param[in]		rport timeout value in seconds.
 *
 * 	@return None
 */
void
bfa_fcs_rport_set_del_timeout(uint8_t rport_tmo)
{
	/* convert to Millisecs */
	if (rport_tmo > 0)  {
		bfa_fcs_rport_del_timeout = rport_tmo * 1000;
	}
}
