/*
 * Copyright (c)  2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

#include <hal/bfa_hal.h>
#include <hal/bfa_hal_timer.h>
#include <hal/bfa_hal_fcxp.h>
#include <hcb/bfa_hcb_fcxp.h>
#include <fcs/bfa_fcs_port.h>
#include <fcs/bfa_fcs_vport.h>
#include "fcs_port.h"
#include "fcs_rport.h"
#include "port_priv.h"
#include "fcs_trcmod.h"
#include "fcs_fcxp.h"

BFA_TRC_FILE(FCS, SCN);

/*
 * forward declarations
 */
static void     bfa_fcs_port_scn_send_scr(void *scn_cbarg,
					  bfa_hal_fcxp_t *fcxp_alloced);
static void     bfa_fcs_port_scn_scr_response(void *fcsarg,
					      bfa_hal_fcxp_t *fcxp,
					      void *cbarg,
					      bfa_status_t req_status,
					      uint32_t rsp_len,
					      uint32_t resid_len,
					      fchs_t *rsp_fchs);
static void     bfa_fcs_port_scn_send_ls_acc(bfa_fcs_port_t *port,
					     fchs_t *rx_fchs);
static void     bfa_fcs_port_scn_timeout(void *arg);

/**
 * @dg fcs_scm_sm FCS SCN state machine
 * @{
 */

/**
 * VPort SCN State Machine events
 */
typedef enum {
	SCNSM_EVENT_PORT_ONLINE = 1,
	SCNSM_EVENT_PORT_OFFLINE = 2,
	SCNSM_EVENT_RSP_OK = 3,
	SCNSM_EVENT_RSP_ERROR = 4,
	SCNSM_EVENT_TIMEOUT = 5,
	SCNSM_EVENT_SCR_SENT = 6,
} port_scn_event_t;

static void     bfa_fcs_port_scn_sm_offline(bfa_fcs_port_scn_t *scn,
					    port_scn_event_t event);
static void     bfa_fcs_port_scn_sm_sending_scr(bfa_fcs_port_scn_t *scn,
						port_scn_event_t event);
static void     bfa_fcs_port_scn_sm_scr(bfa_fcs_port_scn_t *scn,
					port_scn_event_t event);
static void     bfa_fcs_port_scn_sm_scr_retry(bfa_fcs_port_scn_t *scn,
					      port_scn_event_t event);
static void     bfa_fcs_port_scn_sm_online(bfa_fcs_port_scn_t *scn,
					   port_scn_event_t event);

/**
 * @brief
 * 		Starting state - awaiting link up.
 */
static void
bfa_fcs_port_scn_sm_offline(bfa_fcs_port_scn_t *scn, port_scn_event_t event)
{
	switch (event) {
	case SCNSM_EVENT_PORT_ONLINE:
		bfa_sm_set_state(scn, bfa_fcs_port_scn_sm_sending_scr);
		bfa_fcs_port_scn_send_scr(scn, NULL);
		break;

	case SCNSM_EVENT_PORT_OFFLINE:
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_fcs_port_scn_sm_sending_scr(bfa_fcs_port_scn_t *scn,
				port_scn_event_t event)
{
	switch (event) {
	case SCNSM_EVENT_SCR_SENT:
		bfa_sm_set_state(scn, bfa_fcs_port_scn_sm_scr);
		break;

	case SCNSM_EVENT_PORT_OFFLINE:
		bfa_sm_set_state(scn, bfa_fcs_port_scn_sm_offline);
		bfa_hal_fcxp_walloc_cancel(scn->port->fcs->hal, &scn->fcxp_wqe);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_fcs_port_scn_sm_scr(bfa_fcs_port_scn_t *scn, port_scn_event_t event)
{
	bfa_fcs_port_t *port = scn->port;

	switch (event) {
	case SCNSM_EVENT_RSP_OK:
		bfa_sm_set_state(scn, bfa_fcs_port_scn_sm_online);
		break;

	case SCNSM_EVENT_RSP_ERROR:
		bfa_sm_set_state(scn, bfa_fcs_port_scn_sm_scr_retry);
		bfa_hal_timer_start(port->fcs->hal, &scn->timer,
				    bfa_fcs_port_scn_timeout, scn,
				    BFA_FCS_RETRY_TIMEOUT);
		break;

	case SCNSM_EVENT_PORT_OFFLINE:
		bfa_sm_set_state(scn, bfa_fcs_port_scn_sm_offline);
		bfa_hal_fcxp_discard(scn->fcxp);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_fcs_port_scn_sm_scr_retry(bfa_fcs_port_scn_t *scn, port_scn_event_t event)
{
	switch (event) {
	case SCNSM_EVENT_TIMEOUT:
		bfa_sm_set_state(scn, bfa_fcs_port_scn_sm_sending_scr);
		bfa_fcs_port_scn_send_scr(scn, NULL);
		break;

	case SCNSM_EVENT_PORT_OFFLINE:
		bfa_sm_set_state(scn, bfa_fcs_port_scn_sm_offline);
		bfa_hal_timer_stop(&scn->timer);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_fcs_port_scn_sm_online(bfa_fcs_port_scn_t *scn, port_scn_event_t event)
{
	switch (event) {
	case SCNSM_EVENT_PORT_OFFLINE:
		bfa_sm_set_state(scn, bfa_fcs_port_scn_sm_offline);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @}
 */

/**
 * @dg fcs_scn_private FCS SCN private functions
 * @{
 */

/**
 * This routine will be called to send a SCR command.
 */
static void
bfa_fcs_port_scn_send_scr(void *scn_cbarg, bfa_hal_fcxp_t *fcxp_alloced)
{
	bfa_fcs_port_scn_t *scn = scn_cbarg;
	bfa_fcs_port_t *port = scn->port;
	fchs_t          fchs;
	int             len;
	bfa_hal_fcxp_t *fcxp;

	bfa_trc(port->fcs, port->pid);
	bfa_trc(port->fcs, port->port_cfg.pwwn);

fcxp = fcxp_alloced ? fcxp_alloced : bfa_fcs_fcxp_alloc(port->fcs);
	if (!fcxp) {
		bfa_hal_fcxp_alloc_wait(port->fcs->hal, &scn->fcxp_wqe,
					bfa_fcs_port_scn_send_scr, scn);
		return;
	}
	scn->fcxp = fcxp;

	len = fc_scr_build(&fchs, bfa_hal_fcxp_get_reqbuf(fcxp), port->pid, 0);

	bfa_hal_fcxp_send(fcxp, NULL, port->fabric->vf_id, BFA_FALSE,
			  FC_CLASS_3, len, &fchs, bfa_fcs_port_scn_scr_response,
			  (void *)scn, FC_MAX_PDUSZ, FC_RA_TOV);

	bfa_sm_send_event(scn, SCNSM_EVENT_SCR_SENT);
}

static void
bfa_fcs_port_scn_scr_response(void *fcsarg, bfa_hal_fcxp_t *fcxp, void *cbarg,
			      bfa_status_t req_status, uint32_t rsp_len,
			      uint32_t resid_len, fchs_t *rsp_fchs)
{
	bfa_fcs_port_scn_t *scn = (bfa_fcs_port_scn_t *) cbarg;
	bfa_fcs_port_t *port = scn->port;
	fc_els_cmd_t   *els_cmd;
	fc_ls_rjt_t    *ls_rjt;

	bfa_trc(port->fcs, port->port_cfg.pwwn);

	/*
	 * Sanity Checks
	 */
	if (req_status != BFA_STATUS_OK) {
		bfa_trc(port->fcs, req_status);
		bfa_sm_send_event(scn, SCNSM_EVENT_RSP_ERROR);
		return;
	}

	els_cmd = (fc_els_cmd_t *) BFA_FCXP_RSP_PLD(fcxp);

	switch (els_cmd->els_code) {

	case FC_ELS_ACC:
		bfa_sm_send_event(scn, SCNSM_EVENT_RSP_OK);
		break;

	case FC_ELS_LS_RJT:

		ls_rjt = (fc_ls_rjt_t *) BFA_FCXP_RSP_PLD(fcxp);

		bfa_trc(port->fcs, ls_rjt->reason_code);
		bfa_trc(port->fcs, ls_rjt->reason_code_expl);

		bfa_sm_send_event(scn, SCNSM_EVENT_RSP_ERROR);
		break;

	default:
		bfa_sm_send_event(scn, SCNSM_EVENT_RSP_ERROR);
	}
}

/*
 * Send a LS Accept
 */
static void
bfa_fcs_port_scn_send_ls_acc(bfa_fcs_port_t *port, fchs_t *rx_fchs)
{
	fchs_t          fchs;
	bfa_hal_fcxp_t *fcxp;
	bfa_hal_rport_t *hal_rport = NULL;
	int             len;

	bfa_trc(port->fcs, rx_fchs->s_id);

	fcxp = bfa_fcs_fcxp_alloc(port->fcs);
	if (!fcxp)
		return;

	len = fc_ls_acc_build(&fchs, bfa_hal_fcxp_get_reqbuf(fcxp),
			      rx_fchs->s_id, bfa_fcs_port_get_fcid(port),
			      rx_fchs->ox_id);

	bfa_hal_fcxp_send(fcxp, hal_rport, port->fabric->vf_id, BFA_FALSE,
			  FC_CLASS_3, len, &fchs, NULL, NULL, FC_MAX_PDUSZ, 0);
}

/**
 * 	@brief
 *     This routine will be called by bfa_timer on timer timeouts.
 *
 * 	@param[in] 	vport 			- pointer to bfa_fcs_port_t.
 * 	@param[out]	vport_status 	- pointer to return vport status in
 *
 * 	@return
 * 		void
 *
* 	Special Considerations:
 *
 * 	@note
 */
static void
bfa_fcs_port_scn_timeout(void *arg)
{
	bfa_fcs_port_scn_t *scn = (bfa_fcs_port_scn_t *) arg;

	bfa_sm_send_event(scn, SCNSM_EVENT_TIMEOUT);
}

/**
 * @}
 */

/**
 * @dg fcs_scn_public FCS state change notification public interfaces
 * @{
 */

/*
 * Functions called by port/fab
 */
void
bfa_fcs_port_scn_init(bfa_fcs_port_t *port)
{
	bfa_fcs_port_scn_t *scn = BFA_FCS_GET_SCN_FROM_PORT(port);

	scn->port = port;
	bfa_sm_set_state(scn, bfa_fcs_port_scn_sm_offline);
}

void
bfa_fcs_port_scn_offline(bfa_fcs_port_t *port)
{
	bfa_fcs_port_scn_t *scn = BFA_FCS_GET_SCN_FROM_PORT(port);

	scn->port = port;
	bfa_sm_send_event(scn, SCNSM_EVENT_PORT_OFFLINE);
}

void
bfa_fcs_port_scn_online(bfa_fcs_port_t *port)
{
	bfa_fcs_port_scn_t *scn = BFA_FCS_GET_SCN_FROM_PORT(port);

	scn->port = port;
	bfa_sm_send_event(scn, SCNSM_EVENT_PORT_ONLINE);
}

static void
bfa_fcs_port_scn_portid_rscn(bfa_fcs_port_t *port, uint32_t rpid)
{
	bfa_fcs_rport_t *rport;

	bfa_trc(port->fcs, rpid);

	/**
	 * If this is an unknown device, then it just came online.
	 * Otherwise let rport handle the RSCN event.
	 */
	rport = bfa_fcs_port_get_rport_by_pid(port, rpid);
	if (rport == NULL)
		rport = bfa_fcs_rport_create(port, rpid);
	else
		bfa_fcs_rport_scn(rport);
}

/**
 * rscn format based PID comparison
 */
#define __fc_pid_match(__c0, __c1, __fmt)		\
	(((__fmt) == FC_RSCN_FORMAT_FABRIC) ||		\
	 (((__fmt) == FC_RSCN_FORMAT_DOMAIN) &&		\
	  ((__c0)[0] == (__c1)[0])) ||				\
	 (((__fmt) == FC_RSCN_FORMAT_AREA) &&		\
	  ((__c0)[0] == (__c1)[0]) &&				\
	  ((__c0)[1] == (__c1)[1])))

static void
bfa_fcs_port_scn_multiport_rscn(bfa_fcs_port_t *port, fc_rscn_format_t format,
				uint32_t rscn_pid)
{
	bfa_fcs_rport_t *rport;
	bfa_q_t        *qe, *qe_next;
	uint8_t        *c0, *c1;

	bfa_trc(port->fcs, format);
	bfa_trc(port->fcs, rscn_pid);

	c0 = (uint8_t *) &rscn_pid;

	bfa_q_iter_safe(&port->rport_q, qe, qe_next) {
		rport = (bfa_fcs_rport_t *) qe;
		c1 = (uint8_t *) &rport->pid;
		if (__fc_pid_match(c0, c1, format))
			bfa_fcs_rport_scn(rport);
	}
}

void
bfa_fcs_port_scn_process_rscn(bfa_fcs_port_t *port, fchs_t *fchs,
			      uint32_t len)
{
	fc_rscn_pl_t   *rscn = (fc_rscn_pl_t *) (fchs + 1);
	int             num_entries;
	uint32_t        rscn_pid;
	bfa_boolean_t   nsquery = BFA_FALSE;
	int             i = 0;

	num_entries =
		(bfa_os_ntohs(rscn->payldlen) -
		 sizeof(uint32_t)) / sizeof(rscn->event[0]);

	bfa_trc(port->fcs, num_entries);

	port->stats.num_rscn++;

	bfa_fcs_port_scn_send_ls_acc(port, fchs);

	for (i = 0; i < num_entries; i++) {
		rscn_pid = rscn->event[i].portid;

		bfa_trc(port->fcs, rscn->event[i].format);
		bfa_trc(port->fcs, rscn_pid);

		switch (rscn->event[i].format) {
		case FC_RSCN_FORMAT_PORTID:
			port->stats.num_portid_rscn++;
			bfa_fcs_port_scn_portid_rscn(port, rscn_pid);
			break;

		case FC_RSCN_FORMAT_AREA:
		case FC_RSCN_FORMAT_DOMAIN:
		case FC_RSCN_FORMAT_FABRIC:
			nsquery = BFA_TRUE;
			bfa_fcs_port_scn_multiport_rscn(port,
							rscn->event[i].format,
							rscn_pid);
			break;

		default:
			bfa_assert(0);
			nsquery = BFA_TRUE;
		}
	}

	/**
	 * If any of area, domain or fabric RSCN is received, do a fresh discovery
	 * to find new devices.
	 */
	if (nsquery)
		bfa_fcs_port_ns_query(port);
}

/**
 * @}
 */
