/*
 * Copyright (c)  2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

/**
 * @file bfa_fcs_vport.c FCS virtual port state machine
 */

#include <hal/bfa_hal.h>
#include <hal/bfa_hal_timer.h>
#include <hal/bfa_hal_fcxp.h>
#include <hcb/bfa_hcb_fcxp.h>
#include <fcs/bfa_fcs_port.h>
#include <fcs/bfa_fcs_vport.h>
#include <fcb/bfa_fcb_port.h>
#include <hcb/bfa_hcb_fcxp.h>
#include <fcbuild.h>
#include "fcs.h"
#include "fcs_fabric.h"
#include "fcs_port.h"
#include "fcs_vport.h"
#include "fcs_trcmod.h"
#include "fcs_fcxp.h"
#include "port_priv.h"

BFA_TRC_FILE(FCS, VPORT);

#define __vport_fcs(__vp)       (__vp)->lport.fcs
#define __vport_pwwn(__vp)      (__vp)->lport.port_cfg.pwwn
#define __vport_nwwn(__vp)      (__vp)->lport.port_cfg.nwwn
#define __vport_hal(__vp)       (__vp)->lport.fcs->hal
#define __vport_fcid(__vp)      (__vp)->lport.pid
#define __vport_fabric(__vp)    (__vp)->lport.fabric
#define __vport_vfid(__vp)      (__vp)->lport.fabric->vf_id

#define BFA_FCS_VPORT_MAX_RETRIES  5
/*
 * Forward declarations
 */
static void     bfa_fcs_vport_fdisc_send(void *vport_cbarg,
					 bfa_hal_fcxp_t *fcxp_alloced);
static void     bfa_fcs_vport_timeout(void *vport_arg);
static void     bfa_fcs_vport_fdisc_response(void *fcsarg,
					     bfa_hal_fcxp_t *fcxp, void *cbarg,
					     bfa_status_t req_status,
					     uint32_t rsp_len,
					     uint32_t resid_len,
					     fchs_t *rsp_fchs);
static void     bfa_fcs_vport_logo_send(void *vport_arg,
					bfa_hal_fcxp_t *fcxp_alloced);
static void     bfa_fcs_vport_logo_response(void *fcsarg, bfa_hal_fcxp_t *fcxp,
					    void *vport_cbarg,
					    bfa_status_t req_status,
					    uint32_t rsp_len,
					    uint32_t resid_len,
					    fchs_t *rsp_fchs);
static void     bfa_fcs_vport_free(bfa_fcs_vport_t *vport);

/**
 * @dg fcs_vport_sm FCS virtual port state machine
 * @{
 */

/**
 * VPort State Machine events
 */
typedef enum {
	BFA_FCS_VPORT_SM_CREATE = 1,	/* !< vport create event */
	BFA_FCS_VPORT_SM_DELETE = 2,	/* !< vport delete event */
	BFA_FCS_VPORT_SM_START = 3,	/* !< vport start request */
	BFA_FCS_VPORT_SM_STOP = 4,	/* !< stop: unsupported */
	BFA_FCS_VPORT_SM_ONLINE = 5,	/* !< fabric online */
	BFA_FCS_VPORT_SM_OFFLINE = 6,	/* !< fabric offline event */
	BFA_FCS_VPORT_SM_FRMSENT = 7,	/* !< fdisc/logo sent events */
	BFA_FCS_VPORT_SM_RSP_OK = 8,	/* !< good response */
	BFA_FCS_VPORT_SM_RSP_ERROR = 9,	/* !< failed/bad response */
	BFA_FCS_VPORT_SM_TIMEOUT = 10,	/* !< delay timer event */
	BFA_FCS_VPORT_SM_DELCOMP = 11,	/* !< lport delete completion */
} bfa_fcs_vport_event_t;

static void     bfa_fcs_vport_sm_uninit(bfa_fcs_vport_t *vport,
					bfa_fcs_vport_event_t event);
static void     bfa_fcs_vport_sm_created(bfa_fcs_vport_t *vport,
					 bfa_fcs_vport_event_t event);
static void     bfa_fcs_vport_sm_offline(bfa_fcs_vport_t *vport,
					 bfa_fcs_vport_event_t event);
static void     bfa_fcs_vport_sm_fdisc_send(bfa_fcs_vport_t *vport,
					    bfa_fcs_vport_event_t event);
static void     bfa_fcs_vport_sm_fdisc(bfa_fcs_vport_t *vport,
				       bfa_fcs_vport_event_t event);
static void     bfa_fcs_vport_sm_fdisc_retry(bfa_fcs_vport_t *vport,
					     bfa_fcs_vport_event_t event);
static void     bfa_fcs_vport_sm_online(bfa_fcs_vport_t *vport,
					bfa_fcs_vport_event_t event);
static void     bfa_fcs_vport_sm_deleting(bfa_fcs_vport_t *vport,
					  bfa_fcs_vport_event_t event);
static void     bfa_fcs_vport_sm_cleanup(bfa_fcs_vport_t *vport,
					 bfa_fcs_vport_event_t event);
static void     bfa_fcs_vport_sm_logo_send(bfa_fcs_vport_t *vport,
					   bfa_fcs_vport_event_t event);
static void     bfa_fcs_vport_sm_logo(bfa_fcs_vport_t *vport,
				      bfa_fcs_vport_event_t event);
static void     bfa_fcs_vport_sm_error(bfa_fcs_vport_t *vport,
				      bfa_fcs_vport_event_t event);

bfa_sm_table_t  vport_sm_table[] = {
	{BFA_SM(bfa_fcs_vport_sm_uninit), BFA_FCS_VPORT_UNINIT},
	{BFA_SM(bfa_fcs_vport_sm_created), BFA_FCS_VPORT_CREATED},
	{BFA_SM(bfa_fcs_vport_sm_offline), BFA_FCS_VPORT_OFFLINE},
	{BFA_SM(bfa_fcs_vport_sm_fdisc_send), BFA_FCS_VPORT_FDISC_SEND},
	{BFA_SM(bfa_fcs_vport_sm_fdisc), BFA_FCS_VPORT_FDISC},
	{BFA_SM(bfa_fcs_vport_sm_fdisc_retry), BFA_FCS_VPORT_FDISC_RETRY},
	{BFA_SM(bfa_fcs_vport_sm_online), BFA_FCS_VPORT_ONLINE},
	{BFA_SM(bfa_fcs_vport_sm_deleting), BFA_FCS_VPORT_DELETING},
	{BFA_SM(bfa_fcs_vport_sm_cleanup), BFA_FCS_VPORT_CLEANUP},
	{BFA_SM(bfa_fcs_vport_sm_logo_send), BFA_FCS_VPORT_LOGO_SEND},
	{BFA_SM(bfa_fcs_vport_sm_logo), BFA_FCS_VPORT_LOGO},
	{BFA_SM(bfa_fcs_vport_sm_error), BFA_FCS_VPORT_ERROR}
};

/**
 * Beginning state.
 */
static void
bfa_fcs_vport_sm_uninit(bfa_fcs_vport_t *vport, bfa_fcs_vport_event_t event)
{
	bfa_trc(__vport_fcs(vport), __vport_pwwn(vport));
	bfa_trc(__vport_fcs(vport), event);

	switch (event) {
	case BFA_FCS_VPORT_SM_CREATE:
		bfa_sm_set_state(vport, bfa_fcs_vport_sm_created);
		bfa_fcs_fabric_addvport(__vport_fabric(vport), vport);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * Created state - a start event is required to start up the state machine.
 */
static void
bfa_fcs_vport_sm_created(bfa_fcs_vport_t *vport, bfa_fcs_vport_event_t event)
{
	bfa_trc(__vport_fcs(vport), __vport_pwwn(vport));
	bfa_trc(__vport_fcs(vport), event);

	switch (event) {
	case BFA_FCS_VPORT_SM_START:
		if (bfa_fcs_fabric_is_online(__vport_fabric(vport))
		    && bfa_fcs_fabric_npiv_capable(__vport_fabric(vport))) {
			bfa_sm_set_state(vport, bfa_fcs_vport_sm_fdisc_send);
			bfa_fcs_vport_fdisc_send(vport, NULL);
		} else {
			/**
			 * Fabric is offline or not NPIV capable, stay in
			 * offline state.
			 */
			vport->vport_stats.fab_no_npiv++;
			bfa_sm_set_state(vport, bfa_fcs_vport_sm_offline);
		}
		break;

	case BFA_FCS_VPORT_SM_DELETE:
		bfa_sm_set_state(vport, bfa_fcs_vport_sm_cleanup);
		bfa_fcs_port_delete(&vport->lport);
		break;

	case BFA_FCS_VPORT_SM_ONLINE:
	case BFA_FCS_VPORT_SM_OFFLINE:
		/**
		 * Ignore ONLINE/OFFLINE events from fabric till vport is started.
		 */
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * Offline state - awaiting ONLINE event from fabric SM.
 */
static void
bfa_fcs_vport_sm_offline(bfa_fcs_vport_t *vport, bfa_fcs_vport_event_t event)
{
	bfa_trc(__vport_fcs(vport), __vport_pwwn(vport));
	bfa_trc(__vport_fcs(vport), event);

	switch (event) {
	case BFA_FCS_VPORT_SM_DELETE:
		bfa_sm_set_state(vport, bfa_fcs_vport_sm_cleanup);
		bfa_fcs_port_delete(&vport->lport);
		break;

	case BFA_FCS_VPORT_SM_ONLINE:
		bfa_sm_set_state(vport, bfa_fcs_vport_sm_fdisc_send);
		vport->fdisc_retries = 0;
		bfa_fcs_vport_fdisc_send(vport, NULL);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * Fabric Online and FDSIC is being sent.
 */
static void
bfa_fcs_vport_sm_fdisc_send(bfa_fcs_vport_t *vport,
			    bfa_fcs_vport_event_t event)
{
	bfa_trc(__vport_fcs(vport), __vport_pwwn(vport));
	bfa_trc(__vport_fcs(vport), event);

	switch (event) {
	case BFA_FCS_VPORT_SM_DELETE:
		bfa_sm_set_state(vport, bfa_fcs_vport_sm_cleanup);
		bfa_hal_fcxp_walloc_cancel(__vport_hal(vport),
					   &vport->fcxp_wqe);
		bfa_fcs_port_delete(&vport->lport);
		break;

	case BFA_FCS_VPORT_SM_FRMSENT:
		bfa_sm_set_state(vport, bfa_fcs_vport_sm_fdisc);
		break;

	case BFA_FCS_VPORT_SM_OFFLINE:
		bfa_sm_set_state(vport, bfa_fcs_vport_sm_offline);
		bfa_hal_fcxp_walloc_cancel(__vport_hal(vport),
					   &vport->fcxp_wqe);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * FDISC is sent and awaiting reply from fabric.
 */
static void
bfa_fcs_vport_sm_fdisc(bfa_fcs_vport_t *vport, bfa_fcs_vport_event_t event)
{
	bfa_trc(__vport_fcs(vport), __vport_pwwn(vport));
	bfa_trc(__vport_fcs(vport), event);

	switch (event) {
	case BFA_FCS_VPORT_SM_DELETE:
		bfa_sm_set_state(vport, bfa_fcs_vport_sm_logo_send);
		bfa_hal_fcxp_discard(vport->fcxp);
		bfa_fcs_vport_logo_send(vport, NULL);
		break;

	case BFA_FCS_VPORT_SM_OFFLINE:
		bfa_sm_set_state(vport, bfa_fcs_vport_sm_offline);
		bfa_hal_fcxp_discard(vport->fcxp);
		break;

	case BFA_FCS_VPORT_SM_RSP_OK:
		bfa_sm_set_state(vport, bfa_fcs_vport_sm_online);
		bfa_fcs_port_online(&vport->lport);
		break;

	case BFA_FCS_VPORT_SM_RSP_ERROR:
		bfa_sm_set_state(vport, bfa_fcs_vport_sm_fdisc_retry);
		bfa_hal_timer_start(__vport_hal(vport), &vport->timer,
				    bfa_fcs_vport_timeout, vport,
				    BFA_FCS_RETRY_TIMEOUT);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * FDISC attempt failed - a timer is active to retry FDISC.
 */
static void
bfa_fcs_vport_sm_fdisc_retry(bfa_fcs_vport_t *vport,
			     bfa_fcs_vport_event_t event)
{
	bfa_trc(__vport_fcs(vport), __vport_pwwn(vport));
	bfa_trc(__vport_fcs(vport), event);

	switch (event) {
	case BFA_FCS_VPORT_SM_DELETE:
		bfa_sm_set_state(vport, bfa_fcs_vport_sm_cleanup);
		bfa_hal_timer_stop(&vport->timer);
		bfa_fcs_port_delete(&vport->lport);
		break;

	case BFA_FCS_VPORT_SM_OFFLINE:
		bfa_sm_set_state(vport, bfa_fcs_vport_sm_offline);
		bfa_hal_timer_stop(&vport->timer);
		break;

	case BFA_FCS_VPORT_SM_TIMEOUT:
		bfa_sm_set_state(vport, bfa_fcs_vport_sm_fdisc_send);
		vport->vport_stats.fdisc_retries++;
		vport->fdisc_retries++;
		bfa_fcs_vport_fdisc_send(vport, NULL);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * Vport is online (FDISC is complete).
 */
static void
bfa_fcs_vport_sm_online(bfa_fcs_vport_t *vport, bfa_fcs_vport_event_t event)
{
	bfa_trc(__vport_fcs(vport), __vport_pwwn(vport));
	bfa_trc(__vport_fcs(vport), event);

	switch (event) {
	case BFA_FCS_VPORT_SM_DELETE:
		bfa_sm_set_state(vport, bfa_fcs_vport_sm_deleting);
		bfa_fcs_port_delete(&vport->lport);
		break;

	case BFA_FCS_VPORT_SM_OFFLINE:
		bfa_sm_set_state(vport, bfa_fcs_vport_sm_offline);
		bfa_fcs_port_offline(&vport->lport);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * Vport is being deleted - awaiting lport delete completion to send
 * LOGO to fabric.
 */
static void
bfa_fcs_vport_sm_deleting(bfa_fcs_vport_t *vport, bfa_fcs_vport_event_t event)
{
	bfa_trc(__vport_fcs(vport), __vport_pwwn(vport));
	bfa_trc(__vport_fcs(vport), event);

	switch (event) {
	case BFA_FCS_VPORT_SM_DELETE:
		break;

	case BFA_FCS_VPORT_SM_DELCOMP:
		bfa_sm_set_state(vport, bfa_fcs_vport_sm_logo_send);
		bfa_fcs_vport_logo_send(vport, NULL);
		break;

	case BFA_FCS_VPORT_SM_OFFLINE:
		bfa_sm_set_state(vport, bfa_fcs_vport_sm_cleanup);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 *  Error State. 
 *   This state will be set when the Vport Creation fails due to errors like Dup WWN.
 *   In this state only operation allowed is a Vport Delete.
 */
static void
bfa_fcs_vport_sm_error(bfa_fcs_vport_t *vport, bfa_fcs_vport_event_t event)
{
	bfa_trc(__vport_fcs(vport), __vport_pwwn(vport));
	bfa_trc(__vport_fcs(vport), event);

	switch (event) {
	case BFA_FCS_VPORT_SM_DELETE:
		bfa_sm_set_state(vport, bfa_fcs_vport_sm_uninit);
		bfa_fcs_vport_free(vport);
		break;

	default:
		bfa_trc(__vport_fcs(vport), event);
	}
}

/**
 * Lport cleanup is in progress since vport is being deleted. Fabric is
 * offline, so no LOGO is needed to complete vport deletion.
 */
static void
bfa_fcs_vport_sm_cleanup(bfa_fcs_vport_t *vport, bfa_fcs_vport_event_t event)
{
	bfa_trc(__vport_fcs(vport), __vport_pwwn(vport));
	bfa_trc(__vport_fcs(vport), event);

	switch (event) {
	case BFA_FCS_VPORT_SM_DELCOMP:
		bfa_sm_set_state(vport, bfa_fcs_vport_sm_uninit);
		bfa_fcs_vport_free(vport);
		break;

	case BFA_FCS_VPORT_SM_DELETE:
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * LOGO is being sent to fabric. Vport delete is in progress. Lport delete
 * cleanup is done.
 */
static void
bfa_fcs_vport_sm_logo_send(bfa_fcs_vport_t *vport, bfa_fcs_vport_event_t event)
{
	bfa_trc(__vport_fcs(vport), __vport_pwwn(vport));
	bfa_trc(__vport_fcs(vport), event);

	switch (event) {
	case BFA_FCS_VPORT_SM_FRMSENT:
		bfa_sm_set_state(vport, bfa_fcs_vport_sm_logo);
		break;

	case BFA_FCS_VPORT_SM_OFFLINE:
		bfa_sm_set_state(vport, bfa_fcs_vport_sm_deleting);
		bfa_hal_fcxp_walloc_cancel(__vport_hal(vport),
					   &vport->fcxp_wqe);
		bfa_fcs_port_delete(&vport->lport);
		break;

	case BFA_FCS_VPORT_SM_DELETE:
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * LOGO is sent to fabric. Vport delete is in progress. Lport delete cleanup
 * is done.
 */
static void
bfa_fcs_vport_sm_logo(bfa_fcs_vport_t *vport, bfa_fcs_vport_event_t event)
{
	bfa_trc(__vport_fcs(vport), __vport_pwwn(vport));
	bfa_trc(__vport_fcs(vport), event);

	switch (event) {
	case BFA_FCS_VPORT_SM_OFFLINE:
		bfa_hal_fcxp_discard(vport->fcxp);
		/*
		 * !!! fall through !!!
		 */

	case BFA_FCS_VPORT_SM_RSP_OK:
	case BFA_FCS_VPORT_SM_RSP_ERROR:
		bfa_sm_set_state(vport, bfa_fcs_vport_sm_uninit);
		bfa_fcs_vport_free(vport);
		break;

	case BFA_FCS_VPORT_SM_DELETE:
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @}
 */

/**
 * @dg fcs_vport_private FCS virtual port private functions
 * @{
 */

/**
 * This routine will be called to send a FDISC command.
 */
static void
bfa_fcs_vport_fdisc_send(void *vport_arg, bfa_hal_fcxp_t *fcxp_alloced)
{
	bfa_fcs_vport_t *vport = vport_arg;
	uint32_t        len;
	fchs_t          fchs;
	bfa_hal_fcxp_t *fcxp;
	bfa_hal_rport_t *hal_rport = NULL;

	bfa_trc(__vport_fcs(vport), __vport_pwwn(vport));

	fcxp = fcxp_alloced ? fcxp_alloced :
		bfa_fcs_fcxp_alloc(__vport_fcs(vport));
	if (!fcxp) {
		vport->vport_stats.fdisc_alloc_wait++;
		bfa_hal_fcxp_alloc_wait(__vport_hal(vport), &vport->fcxp_wqe,
					bfa_fcs_vport_fdisc_send, vport);
		return;
	}
	vport->fcxp = fcxp;

	len = fc_fdisc_build(&fchs, bfa_hal_fcxp_get_reqbuf(fcxp), 0, 0,
			     __vport_pwwn(vport), __vport_nwwn(vport),
			     BFA_FCS_PORT_MAX_DATA_SIZE);

	bfa_hal_fcxp_send(fcxp, hal_rport, __vport_vfid(vport), BFA_FALSE,
			  FC_CLASS_3, len, &fchs, bfa_fcs_vport_fdisc_response,
			  vport, FC_MAX_PDUSZ, FC_RA_TOV);

	vport->vport_stats.fdisc_sent++;
	bfa_sm_send_event(vport, BFA_FCS_VPORT_SM_FRMSENT);
}

/**
 * FDISC Response
 */
static void
bfa_fcs_vport_fdisc_response(void *fcsarg, bfa_hal_fcxp_t *fcxp,
			     void *vport_arg, bfa_status_t req_status,
			     uint32_t rsp_len, uint32_t resid_len,
			     fchs_t *rsp_fchs)
{
	bfa_fcs_vport_t *vport = (bfa_fcs_vport_t *) vport_arg;
	fc_els_cmd_t   *els_cmd;
	fc_ls_rjt_t    *ls_rjt;

	bfa_trc(__vport_fcs(vport), __vport_pwwn(vport));

	if (req_status != BFA_STATUS_OK) {
		bfa_trc(__vport_fcs(vport), req_status);
		vport->vport_stats.fdisc_rsp_err++;
		bfa_sm_send_event(vport, BFA_FCS_VPORT_SM_RSP_ERROR);
		return;
	}

	els_cmd = (fc_els_cmd_t *) BFA_FCXP_RSP_PLD(fcxp);
	bfa_trc(__vport_fcs(vport), els_cmd->els_code);

	switch (els_cmd->els_code) {
	case FC_ELS_ACC:
		if (rsp_len < sizeof(fc_flogi_t)) {
			bfa_trc(__vport_fcs(vport), rsp_len);
			vport->vport_stats.fdisc_acc_bad++;
			bfa_sm_send_event(vport, BFA_FCS_VPORT_SM_RSP_ERROR);
			break;
		}

		/*
		 * Initialiaze the V-Port fields
		 */
		__vport_fcid(vport) = rsp_fchs->d_id;
		vport->vport_stats.fdisc_accepts++;
		bfa_sm_send_event(vport, BFA_FCS_VPORT_SM_RSP_OK);
		break;

	case FC_ELS_LS_RJT:
		ls_rjt = (fc_ls_rjt_t *) BFA_FCXP_RSP_PLD(fcxp);
		bfa_trc(__vport_fcs(vport), ls_rjt->reason_code);
		bfa_trc(__vport_fcs(vport), ls_rjt->reason_code_expl);
		vport->vport_stats.fdisc_rejects++;

		/* For certain reason codes, we don't want to retry. */
		switch (ls_rjt->reason_code_expl) {
		case  FC_LS_RJT_EXP_INV_PORT_NAME  : /* by brocade */
		case  FC_LS_RJT_EXP_INVALID_NPORT_ID : /* by Cisco */
			if (vport->fdisc_retries < BFA_FCS_VPORT_MAX_RETRIES) 
				bfa_sm_send_event(vport, BFA_FCS_VPORT_SM_RSP_ERROR);
			else 
				bfa_sm_set_state(vport, bfa_fcs_vport_sm_error);
					
			break;

		default :
			bfa_sm_send_event(vport, BFA_FCS_VPORT_SM_RSP_ERROR);
			break;
		}
		break;

	default:
		bfa_assert(0);
		vport->vport_stats.fdisc_unknown_rsp++;
		bfa_sm_send_event(vport, BFA_FCS_VPORT_SM_RSP_ERROR);
	}
}

/**
 * @brief
 * 	Called to send a logout to the fabric. Used when a V-Port is
 * 	deleted/stopped.
 */
static void
bfa_fcs_vport_logo_send(void *vport_arg, bfa_hal_fcxp_t *fcxp_alloced)
{
	bfa_fcs_vport_t *vport = vport_arg;
	fchs_t          fchs;
	uint16_t        len;
	bfa_hal_fcxp_t *fcxp;

	bfa_trc(__vport_fcs(vport), __vport_pwwn(vport));

	fcxp = fcxp_alloced ? fcxp_alloced :
		bfa_fcs_fcxp_alloc(__vport_fcs(vport));
	if (!fcxp) {
		vport->vport_stats.logo_alloc_wait++;
		bfa_hal_fcxp_alloc_wait(__vport_hal(vport), &vport->fcxp_wqe,
					bfa_fcs_vport_logo_send, vport);
		return;
	}
	vport->fcxp = fcxp;

	len = fc_logo_build(&fchs, bfa_hal_fcxp_get_reqbuf(fcxp),
			    bfa_os_hton3b(FC_FABRIC_PORT), __vport_fcid(vport),
			    0, __vport_pwwn(vport));

	bfa_hal_fcxp_send(fcxp, NULL, __vport_vfid(vport), BFA_FALSE,
			  FC_CLASS_3, len, &fchs, bfa_fcs_vport_logo_response,
			  vport, FC_MAX_PDUSZ, FC_RA_TOV);

	vport->vport_stats.logo_sent++;

	bfa_sm_send_event(vport, BFA_FCS_VPORT_SM_FRMSENT);
}

/*
 * LOGO Response
 */
static void
bfa_fcs_vport_logo_response(void *fcsarg, bfa_hal_fcxp_t *fcxp,
			    void *vport_arg, bfa_status_t req_status,
			    uint32_t rsp_len, uint32_t resid_len,
			    fchs_t *rsp_fchs)
{
	bfa_fcs_vport_t *vport = (bfa_fcs_vport_t *) vport_arg;
	fc_els_cmd_t   *els_cmd;
	fc_ls_rjt_t    *ls_rjt;

	bfa_trc(__vport_fcs(vport), __vport_pwwn(vport));

	/*
	 * Sanity Checks
	 */
	if (req_status != BFA_STATUS_OK) {
		bfa_trc(__vport_fcs(vport), req_status);
		vport->vport_stats.logo_rsp_err++;
		bfa_sm_send_event(vport, BFA_FCS_VPORT_SM_RSP_ERROR);
		return;
	}

	els_cmd = (fc_els_cmd_t *) BFA_FCXP_RSP_PLD(fcxp);

	switch (els_cmd->els_code) {

	case FC_ELS_ACC:
		vport->vport_stats.logo_accepts++;
		bfa_sm_send_event(vport, BFA_FCS_VPORT_SM_RSP_OK);
		break;

	case FC_ELS_LS_RJT:
		ls_rjt = (fc_ls_rjt_t *) BFA_FCXP_RSP_PLD(fcxp);
		bfa_trc(__vport_fcs(vport), ls_rjt->reason_code);
		bfa_trc(__vport_fcs(vport), ls_rjt->reason_code_expl);
		vport->vport_stats.logo_rejects++;
		bfa_sm_send_event(vport, BFA_FCS_VPORT_SM_RSP_ERROR);
		break;

	default:
		bfa_assert(0);
		vport->vport_stats.logo_unknown_rsp++;
		bfa_sm_send_event(vport, BFA_FCS_VPORT_SM_RSP_ERROR);
	}
}

/**
 * 	@brief
 *     This routine will be called by bfa_timer on timer timeouts.
 *
 * 	@param[in] 	vport 		- pointer to bfa_fcs_vport_t.
 * 	@param[out]	vport_status 	- pointer to return vport status in
 *
 * 	@return
 * 		void
 *
 * 	Special Considerations:
 *
 * 	@note
 */
static void
bfa_fcs_vport_timeout(void *vport_arg)
{
	bfa_fcs_vport_t *vport = (bfa_fcs_vport_t *) vport_arg;

	vport->vport_stats.fdisc_timeouts++;
	bfa_sm_send_event(vport, BFA_FCS_VPORT_SM_TIMEOUT);
}

static void
bfa_fcs_vport_free(bfa_fcs_vport_t *vport)
{
	bfa_fcs_fabric_delvport(__vport_fabric(vport), vport);
	bfa_fcb_vport_delete(vport->vport_drv);
}

/**
 * @}
 */

/**
 * @dg fcs_vport_public FCS virtual port public interfaces
 * @{
 */

/**
 * Online notification from fabric SM.
 */
void
bfa_fcs_vport_online(bfa_fcs_vport_t *vport)
{
	vport->vport_stats.fab_online++;
	bfa_sm_send_event(vport, BFA_FCS_VPORT_SM_ONLINE);
}

/**
 * Offline notification from fabric SM.
 */
void
bfa_fcs_vport_offline(bfa_fcs_vport_t *vport)
{
	vport->vport_stats.fab_offline++;
	bfa_sm_send_event(vport, BFA_FCS_VPORT_SM_OFFLINE);
}

/**
 * Cleanup notification from fabric SM on link timer expiry.
 */
void
bfa_fcs_vport_cleanup(bfa_fcs_vport_t *vport)
{
	vport->vport_stats.fab_cleanup++;
}

/**
 * Delete completion callback from associated lport
 */
void
bfa_fcs_vport_delete_comp(bfa_fcs_vport_t *vport)
{
	bfa_sm_send_event(vport, BFA_FCS_VPORT_SM_DELCOMP);
}

/**
 * @brief
 *   Module initialization
 */
void
bfa_fcs_vport_modinit(bfa_fcs_t *fcs)
{
}

/**
 * @brief
 *   Module cleanup
 */
void
bfa_fcs_vport_modexit(bfa_fcs_t *fcs)
{
	bfa_fcs_modexit_comp(fcs);
}

/**
 * @}
 */

/**
 * @dg fcs_vport_api Virtual port API
 * @{
 */

/**
 * 	@brief
 *  	Use this function to instantiate a new FCS vport object. This
 * 	function will not trigger any HW initialization process (which will be
 * 	done in vport_start() call)
 *
 * 	@param[in] vport	- 	pointer to bfa_fcs_vport_t. This space
 *					needs to be allocated by the driver.
 * 	@param[in] fcs 		- 	FCS instance
 * 	@param[in] vport_cfg	- 	vport configuration
 * 	@param[in] vf_id    	- 	VF_ID if vport is created within a VF.
 *                          		FC_VF_ID_NULL to specify base fabric.
 * 	@param[in] vport_drv 	- 	Opaque handle back to the driver's vport
 * 					structure
 *
 * 	@retval BFA_STATUS_OK - on success.
 * 	@retval BFA_STATUS_FAILED - on failure.
 */
bfa_status_t
bfa_fcs_vport_create(bfa_fcs_vport_t *vport, bfa_fcs_t *fcs, uint16_t vf_id,
		     bfa_port_cfg_t *vport_cfg, bfad_vport_t *vport_drv)
{
	if (vport_cfg->pwwn == 0)
		return (BFA_STATUS_INVALID_WWN);

	if (bfa_fcs_port_get_pwwn(&fcs->fabric.bport) == vport_cfg->pwwn)
		return BFA_STATUS_VPORT_WWN_BP;

	if (bfa_fcs_vport_lookup(fcs, vf_id, vport_cfg->pwwn) != NULL)
		return BFA_STATUS_VPORT_EXISTS;

	if (bfa_fcs_fabric_vport_count(&fcs->fabric) == BFA_FCS_MAX_VPORTS_SUPP)
		return BFA_STATUS_VPORT_MAX;

	vport->vport_drv = vport_drv;
	bfa_sm_set_state(vport, bfa_fcs_vport_sm_uninit);

	bfa_fcs_lport_init(&vport->lport, fcs, vf_id, vport_cfg, vport);

	bfa_sm_send_event(vport, BFA_FCS_VPORT_SM_CREATE);

	return BFA_STATUS_OK;
}

/**
 *  @brief
 *  	Use this function initialize the vport.
 *
 *  @param[in] vport - pointer to bfa_fcs_vport_t.
 *
 *  @returns None
 */
bfa_status_t
bfa_fcs_vport_start(bfa_fcs_vport_t *vport)
{
	bfa_sm_send_event(vport, BFA_FCS_VPORT_SM_START);

	return BFA_STATUS_OK;
}

/**
 * 	@brief
 *  	Use this function quiese the vport object. This function will return
 * 	immediately, when the vport is actually stopped, the
 * 	bfa_drv_vport_stop_cb() will be called.
 *
 * 	@param[in] vport - pointer to bfa_fcs_vport_t.
 *
 * 	@return None
 */
bfa_status_t
bfa_fcs_vport_stop(bfa_fcs_vport_t *vport)
{
	bfa_sm_send_event(vport, BFA_FCS_VPORT_SM_STOP);

	return BFA_STATUS_OK;
}

/**
 * 	@brief
 *  	Use this function to delete a vport object. Fabric object should
 * 		be stopped before this function call.
 *
 * 	@param[in] vport - pointer to bfa_fcs_vport_t.
 *
 * 	@return     None
 */
bfa_status_t
bfa_fcs_vport_delete(bfa_fcs_vport_t *vport)
{
	bfa_sm_send_event(vport, BFA_FCS_VPORT_SM_DELETE);

	return BFA_STATUS_OK;
}

/**
 * 	@brief
 *  	Use this function to get vport's current status info.
 *
 * 	@param[in] 	vport 		pointer to bfa_fcs_vport_t.
 * 	@param[out]	attr 		pointer to return vport attributes
 *
 * 	@return None
 */
void
bfa_fcs_vport_get_attr(bfa_fcs_vport_t *vport, bfa_vport_attr_t *attr)
{
	if (vport == NULL || attr == NULL)
		return;

	bfa_os_memset(attr, 0, sizeof(bfa_vport_attr_t));

	bfa_fcs_port_get_attr(&vport->lport, &attr->port_attr);
	attr->vport_state = bfa_sm_to_state(vport_sm_table, vport->sm);
}

/**
 * 	@brief
 *  	Use this function to get vport's statistics.
 *
 * 	@param[in] 	vport 		pointer to bfa_fcs_vport_t.
 * 	@param[out]	stats		pointer to return vport statistics in
 *
 * 	@return None
 */
void
bfa_fcs_vport_get_stats(bfa_fcs_vport_t *vport, bfa_vport_stats_t *stats)
{
	*stats = vport->vport_stats;
}

/**
 * 	@brief
 *  	Use this function to clear vport's statistics.
 *
 * 	@param[in] 	vport 		pointer to bfa_fcs_vport_t.
 *
 * 	@return None
 */
void
bfa_fcs_vport_clr_stats(bfa_fcs_vport_t *vport)
{
	bfa_os_memset(&vport->vport_stats, 0, sizeof(bfa_vport_stats_t));
}

/**
 * @brief
 *      Lookup a virtual port. Excludes base port from lookup.
 */
bfa_fcs_vport_t *
bfa_fcs_vport_lookup(bfa_fcs_t *fcs, uint16_t vf_id, wwn_t vpwwn)
{
	bfa_fcs_vport_t *vport;
	bfa_fcs_fabric_t *fabric;

	bfa_trc(fcs, vf_id);
	bfa_trc(fcs, vpwwn);

	fabric = bfa_fcs_vf_lookup(fcs, vf_id);
	if (!fabric) {
		bfa_trc(fcs, vf_id);
		return NULL;
	}

	vport = bfa_fcs_fabric_vport_lookup(fabric, vpwwn);
	return vport;
}

/**
 * @}
 */
