/*
 * Copyright (c) 2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

#include <hal/bfa_hal.h>
#include <log/bfa_log_hal.h>

BFA_TRC_FILE(HAL, FCPIM);
BFA_HAL_MODULE(fcpim);

/**
 * @dg hal_fcpim_mod HAL FCP Initiator Mode module
 * @{
 */

/**
 * @brief
 * 		Compute and return memory needed by FCP(im) module.
 */
static void
bfa_hal_fcpim_meminfo(bfa_ioc_cfg_t *cfg, uint32_t *km_len, uint32_t *dm_len)
{
	bfa_hal_itnim_meminfo(cfg, km_len, dm_len);

	/**
	 * IO memory
	 */
	if (cfg->fwcfg.num_ioim_reqs < BFA_HAL_IOIM_MIN)
		cfg->fwcfg.num_ioim_reqs = BFA_HAL_IOIM_MIN;
	*km_len += cfg->fwcfg.num_ioim_reqs *
	  (sizeof(bfa_hal_ioim_t) + sizeof(bfa_hal_ioim_sp_t));

	*dm_len += cfg->fwcfg.num_ioim_reqs * BFA_FWI_IOIM_SNSLEN;
					       

	/**
	 * task management command memory
	 */
	if (cfg->fwcfg.num_tskim_reqs < BFA_HAL_TSKIM_MIN)
		cfg->fwcfg.num_tskim_reqs = BFA_HAL_TSKIM_MIN;
	*km_len += cfg->fwcfg.num_tskim_reqs * sizeof(bfa_hal_tskim_t);
}

/**
 * @brief
 *
 */
static void
bfa_hal_fcpim_attach(bfa_hal_t *hal, void *bfad, bfa_ioc_cfg_t *cfg,
		     bfa_hal_meminfo_t *meminfo, bfa_hal_pcidev_t *pcidev)
{
	bfa_hal_fcpim_mod_t *fcpim = BFA_HAL_FCPIM_MOD(hal);

	bfa_trc(hal, cfg->mpio_mode);
	bfa_trc(hal, cfg->path_tov);
	bfa_trc(hal, cfg->fwcfg.num_itnims);
	bfa_trc(hal, cfg->fwcfg.num_ioim_reqs);
	bfa_trc(hal, cfg->fwcfg.num_tskim_reqs);

	fcpim->hal            = hal;
	fcpim->num_itnims     = cfg->fwcfg.num_itnims;
	fcpim->num_ioim_reqs  = cfg->fwcfg.num_ioim_reqs;
	fcpim->num_tskim_reqs = cfg->fwcfg.num_tskim_reqs;
	fcpim->mpio_mode      = cfg->mpio_mode;
	fcpim->path_tov       = cfg->path_tov;

	bfa_hal_itnim_attach(fcpim, meminfo);
	bfa_hal_tskim_attach(fcpim, meminfo);
	bfa_hal_ioim_attach(fcpim, meminfo);
}

static void
bfa_hal_fcpim_initdone(bfa_hal_t *hal)
{
}

static void
bfa_hal_fcpim_detach(bfa_hal_t *hal)
{
	bfa_hal_fcpim_mod_t *fcpim = BFA_HAL_FCPIM_MOD(hal);

	bfa_hal_ioim_detach(fcpim);
	bfa_hal_tskim_detach(fcpim);
}

static void
bfa_hal_fcpim_start(bfa_hal_t *hal)
{
}

static void
bfa_hal_fcpim_stop(bfa_hal_t *hal)
{
}

static void
bfa_hal_fcpim_iocdisable(bfa_hal_t *hal)
{
	bfa_hal_fcpim_mod_t *fcpim = BFA_HAL_FCPIM_MOD(hal);
	bfa_hal_itnim_t *itnim;
	bfa_q_t        *qe, *qen;

	bfa_q_iter_safe(&fcpim->itnim_q, qe, qen) {
		itnim = (bfa_hal_itnim_t *) qe;
		bfa_hal_itnim_iocdisable(itnim);
	}
}

void
bfa_hal_fcpim_mpio_mode_set(bfa_hal_t *hal, bfa_boolean_t mpio_mode,
			    uint16_t path_tov)
{
	bfa_hal_fcpim_mod_t *fcpim = BFA_HAL_FCPIM_MOD(hal);

	fcpim->mpio_mode = mpio_mode;
	if (!mpio_mode) {
		if (path_tov)
			fcpim->path_tov = path_tov * 1000;
		else
			fcpim->path_tov = BFA_HAL_FCPIM_PATH_TOV_DEF;
	}
}

bfa_boolean_t
bfa_hal_fcpim_mpio_mode_get(bfa_hal_t *hal)
{
	bfa_hal_fcpim_mod_t *fcpim = BFA_HAL_FCPIM_MOD(hal);

	return (fcpim->mpio_mode);
}

uint16_t
bfa_hal_fcpim_path_tov_get(bfa_hal_t *hal)
{
	bfa_hal_fcpim_mod_t *fcpim = BFA_HAL_FCPIM_MOD(hal);

	return (fcpim->path_tov / 1000);
}

void
bfa_hal_fcpim_qdepth_set(bfa_hal_t *hal, uint16_t q_depth)
{
	bfa_hal_fcpim_mod_t *fcpim = BFA_HAL_FCPIM_MOD(hal);

	bfa_assert(q_depth <= BFA_IOC_MAX_Q_DEPTH);

	fcpim->q_depth = q_depth;
}

uint16_t
bfa_hal_fcpim_qdepth_get(bfa_hal_t *hal)
{
	bfa_hal_fcpim_mod_t *fcpim = BFA_HAL_FCPIM_MOD(hal);

	return (fcpim->q_depth);
}

/**
 * @}
 */
