/*
 * Copyright (c) 2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

#include <hal/bfa_hal.h>
#include <hcb/bfa_hcb_fcxp.h>
#include <fwi/bfa_fwi_uf.h>
#include <cs/bfa_debug.h>

BFA_TRC_FILE(HAL, FCXP);
BFA_HAL_MODULE(fcxp);

/**
 * forward declarations
 */
static void     __bfa_hal_fcxp_send_cbfn(void *cbarg, bfa_boolean_t complete);
static void     hal_fcxp_rx_plog(bfa_hal_t *hal, bfa_hal_fcxp_t *fcxp,
				 bfa_fwi_fcxp_send_rsp_t *fcxp_rsp);
static void     hal_fcxp_tx_plog(bfa_hal_t *hal, uint32_t reqlen,
				 bfa_hal_fcxp_t *fcxp, fchs_t *fchs);

/**
 * @dg fcxp_pvt HAL FCXP private functions
 * @{
 */

static void
claim_fcxp_req_rsp_mem(bfa_hal_fcxp_mod_t *mod, bfa_hal_meminfo_t *mi)
{
	uint8_t        *dm_kva = NULL;
	uint64_t        dm_pa;
	uint32_t        buf_pool_sz = 0;

	dm_kva = bfa_hal_meminfo_dma_virt(mi);
	dm_pa = bfa_hal_meminfo_dma_phys(mi);

	buf_pool_sz += sizeof(bfa_hal_fcxp_buf_t) * mod->num_fcxps;

	/*
	 * Initialize the fcxp req payload list
	 */
	mod->req_pld_list_kva = (bfa_hal_fcxp_buf_t *) dm_kva;
	mod->req_pld_list_pa = dm_pa;
	dm_kva += buf_pool_sz;
	dm_pa += buf_pool_sz;
	bfa_os_memset(mod->req_pld_list_kva, 0, buf_pool_sz);

	/*
	 * Initialize the fcxp rsp payload list
	 */
	mod->rsp_pld_list_kva = (bfa_hal_fcxp_buf_t *) dm_kva;
	mod->rsp_pld_list_pa = dm_pa;
	dm_kva += buf_pool_sz;
	dm_pa += buf_pool_sz;
	bfa_os_memset(mod->rsp_pld_list_kva, 0, buf_pool_sz);

	bfa_hal_meminfo_dma_virt(mi) = dm_kva;
	bfa_hal_meminfo_dma_phys(mi) = dm_pa;
}

static void
claim_fcxps_mem(bfa_hal_fcxp_mod_t *mod, bfa_hal_meminfo_t *mi)
{
	uint16_t        i;
	bfa_hal_fcxp_t *fcxp;

	fcxp = (bfa_hal_fcxp_t *) bfa_hal_meminfo_kva(mi);
	bfa_os_memset(fcxp, 0, sizeof(bfa_hal_fcxp_t) * mod->num_fcxps);

	bfa_q_init(&mod->fcxp_free_q);
	bfa_q_init(&mod->fcxp_active_q);

	mod->fcxp_list = fcxp;

	for (i = 0; i < mod->num_fcxps; i++) {
		fcxp->fcxp_mod = mod;
		fcxp->fcxp_tag = i;

		bfa_q_enq(&mod->fcxp_free_q, fcxp);
		fcxp = fcxp + 1;
	}

	bfa_hal_meminfo_kva(mi) = (void *)fcxp;
}

static void
bfa_hal_fcxp_meminfo(bfa_ioc_cfg_t *cfg, uint32_t *ndm_len, uint32_t *dm_len)
{
	uint16_t        num_fcxp_reqs = cfg->fwcfg.num_fcxp_reqs;

	if (num_fcxp_reqs == 0)
		return;

	/*
	 * Account for req/rsp payload
	 */
	*dm_len += sizeof(bfa_hal_fcxp_buf_t) * num_fcxp_reqs * 2;

	/*
	 * Account for fcxp structs
	 */
	*ndm_len += sizeof(bfa_hal_fcxp_t) * num_fcxp_reqs;
}

static void
bfa_hal_fcxp_attach(bfa_hal_t *hal, void *bfad, bfa_ioc_cfg_t *cfg,
		    bfa_hal_meminfo_t *meminfo, bfa_hal_pcidev_t *pcidev)
{
	bfa_hal_fcxp_mod_t *mod = BFA_HAL_FCXP_MOD(hal);

	bfa_os_memset(mod, 0, sizeof(bfa_hal_fcxp_mod_t));
	mod->hal = hal;
	mod->num_fcxps = cfg->fwcfg.num_fcxp_reqs;
	bfa_q_init(&mod->wait_q);

	claim_fcxp_req_rsp_mem(mod, meminfo);
	claim_fcxps_mem(mod, meminfo);
}

static void
bfa_hal_fcxp_initdone(bfa_hal_t *hal)
{
}

static void
bfa_hal_fcxp_detach(bfa_hal_t *hal)
{
}

static void
bfa_hal_fcxp_start(bfa_hal_t *hal)
{
}

static void
bfa_hal_fcxp_stop(bfa_hal_t *hal)
{
}

static void
bfa_hal_fcxp_iocdisable(bfa_hal_t *hal)
{
	bfa_hal_fcxp_mod_t *mod = BFA_HAL_FCXP_MOD(hal);
	bfa_hal_fcxp_t *fcxp;
	bfa_q_t        *qe, *qen;

	bfa_q_iter_safe(&mod->fcxp_active_q, qe, qen) {
		fcxp = (bfa_hal_fcxp_t *) qe;
		if (fcxp->caller == NULL)
			fcxp->send_cbfn(fcxp->caller, fcxp, fcxp->send_cbarg,
					BFA_STATUS_IOC_FAILURE, 0, 0, NULL);
		else {
			fcxp->rsp_status = BFA_STATUS_IOC_FAILURE;
			bfa_hcb_queue(hal, &fcxp->hcb_qe,
				      __bfa_hal_fcxp_send_cbfn, fcxp);
		}
	}
}

static bfa_hal_fcxp_t *
bfa_hal_fcxp_get(bfa_hal_fcxp_mod_t *fm)
{
	bfa_hal_fcxp_t *fcxp;

	bfa_q_deq(&fm->fcxp_free_q, &fcxp);

	if (fcxp)
		bfa_q_enq(&fm->fcxp_active_q, fcxp);

	return (fcxp);
}

static void
bfa_hal_fcxp_put(bfa_hal_fcxp_t *fcxp)
{
	bfa_hal_fcxp_mod_t *mod = fcxp->fcxp_mod;
	bfa_hal_fcxp_wqe_t *wqe;

	bfa_q_deq(&mod->wait_q, &wqe);
	if (wqe) {
		bfa_trc(mod->hal, fcxp->fcxp_tag);
		wqe->alloc_cbfn(wqe->alloc_cbarg, fcxp);
		return;
	}

	bfa_assert(bfa_q_is_on_q(&mod->fcxp_active_q, fcxp));
	bfa_q_qe_deq(fcxp);
	bfa_q_enq(&mod->fcxp_free_q, fcxp);
}

static void
bfa_hal_fcxp_null_comp(void *bfad_fcxp, bfa_hal_fcxp_t *fcxp, void *cbarg,
		       bfa_status_t req_status, uint32_t rsp_len,
		       uint32_t resid_len, fchs_t *rsp_fchs)
{
	/**discarded fcxp completion */
}

static void
__bfa_hal_fcxp_send_cbfn(void *cbarg, bfa_boolean_t complete)
{
	bfa_hal_fcxp_t *fcxp = cbarg;

	if (complete) {
		fcxp->send_cbfn(fcxp->caller, fcxp, fcxp->send_cbarg,
				fcxp->rsp_status, fcxp->rsp_len,
				fcxp->residue_len, &fcxp->rsp_fchs);
	} else {
		bfa_hal_fcxp_free(fcxp);
	}
}

static void
hal_fcxp_send_comp(bfa_hal_t *hal, bfa_fwi_fcxp_send_rsp_t *fcxp_rsp)
{
	bfa_hal_fcxp_mod_t	*mod = BFA_HAL_FCXP_MOD(hal);
	bfa_hal_fcxp_t		*fcxp;
	uint16_t		fcxp_tag = bfa_os_ntohs(fcxp_rsp->fcxp_tag);

	bfa_trc(hal, fcxp_tag);

	fcxp_rsp->rsp_len = bfa_os_ntohl(fcxp_rsp->rsp_len);

	/**
	 * @todo f/w should not set residue to non-0 when everything
	 *       is received.
	 */
	if (fcxp_rsp->req_status == BFA_STATUS_OK)
		fcxp_rsp->residue_len = 0;
	else
		fcxp_rsp->residue_len = bfa_os_ntohl(fcxp_rsp->residue_len);

	fcxp = BFA_HAL_FCXP_FROM_TAG(mod, fcxp_tag);

	bfa_assert(fcxp->send_cbfn != NULL);

	hal_fcxp_rx_plog(mod->hal, fcxp, fcxp_rsp);

	if (fcxp->send_cbfn != NULL) {
		if (fcxp->caller == NULL) {
			bfa_trc(mod->hal, fcxp->fcxp_tag);

			fcxp->send_cbfn(fcxp->caller, fcxp, fcxp->send_cbarg,
					fcxp_rsp->req_status, fcxp_rsp->rsp_len,
					fcxp_rsp->residue_len, &fcxp_rsp->fchs);
			/*
			 * fcxp automatically freed on return from the callback
			 */
			bfa_hal_fcxp_free(fcxp);
		} else {
			bfa_trc(mod->hal, fcxp->fcxp_tag);
			fcxp->rsp_status = fcxp_rsp->req_status;
			fcxp->rsp_len = fcxp_rsp->rsp_len;
			fcxp->residue_len = fcxp_rsp->residue_len;
			fcxp->rsp_fchs = fcxp_rsp->fchs;

			bfa_hcb_queue(hal, &fcxp->hcb_qe,
				      __bfa_hal_fcxp_send_cbfn, fcxp);
		}
	} else {
		bfa_trc(hal, fcxp_tag);
	}
}

static void
hal_fcxp_set_local_sges(bfa_sge_t *sge, uint32_t reqlen, uint64_t req_pa)
{
	bfa_addr_t      sga_zero = { {0} };

	sge->sg_len = reqlen;
	sge->flags = BFA_SGE_DATA_LAST;
	bfa_hal_dma_addr_set(sge[0].sga, req_pa);
	bfa_sge_to_be(sge);
	sge++;

	sge->sga = sga_zero;
	sge->sg_len = reqlen;
	sge->flags = BFA_SGE_PGDLEN;
	bfa_sge_to_be(sge);
}

static void
hal_fcxp_tx_plog(bfa_hal_t *hal, uint32_t reqlen, bfa_hal_fcxp_t *fcxp,
		 fchs_t *fchs)
{
	/*
* TODO: TX ox_id
	 */
	if (reqlen > 0) {
		if (fcxp->use_ireqbuf) {
			uint32_t        pld_w0 =
				*((uint32_t *) BFA_FCXP_REQ_PLD(fcxp));

			bfa_plog_fchdr_and_pl(hal->plog, BFA_PL_MID_HAL_FCXP,
					      BFA_PL_EID_TX,
					      reqlen + sizeof(fchs_t), fchs,
					      pld_w0);
		} else {
			bfa_plog_fchdr(hal->plog, BFA_PL_MID_HAL_FCXP,
				       BFA_PL_EID_TX, reqlen + sizeof(fchs_t),
				       fchs);
		}
	} else {
		bfa_plog_fchdr(hal->plog, BFA_PL_MID_HAL_FCXP, BFA_PL_EID_TX,
			       reqlen + sizeof(fchs_t), fchs);
	}
}

static void
hal_fcxp_rx_plog(bfa_hal_t *hal, bfa_hal_fcxp_t *fcxp,
		 bfa_fwi_fcxp_send_rsp_t *fcxp_rsp)
{
	if (fcxp_rsp->rsp_len > 0) {
		if (fcxp->use_irspbuf) {
			uint32_t        pld_w0 =
				*((uint32_t *) BFA_FCXP_RSP_PLD(fcxp));

			bfa_plog_fchdr_and_pl(hal->plog, BFA_PL_MID_HAL_FCXP,
					      BFA_PL_EID_RX,
					      (uint16_t) fcxp_rsp->rsp_len,
					      &fcxp_rsp->fchs, pld_w0);
		} else {
			bfa_plog_fchdr(hal->plog, BFA_PL_MID_HAL_FCXP,
				       BFA_PL_EID_RX,
				       (uint16_t) fcxp_rsp->rsp_len,
				       &fcxp_rsp->fchs);
		}
	} else {
		bfa_plog_fchdr(hal->plog, BFA_PL_MID_HAL_FCXP, BFA_PL_EID_RX,
			       (uint16_t) fcxp_rsp->rsp_len, &fcxp_rsp->fchs);
	}
}

/**
 * @dg hal_fcxp_api HAL FCXP API
 * @{
 */

/**
 * @brief
 * Allocate an FCXP instance to send a response or to send a request
 * that has a response. Request/response buffers are allocated by caller.
 *
 * @param[in]	hal		BFA hal instance
 * @param[in]	nreq_sgles	Number of SG elements required for request
 * 				buffer. 0, if fcxp internal buffers are	used.
 * 				Use bfa_hal_fcxp_get_reqbuf() to get the
 * 				internal req buffer.
 * @param[in]	req_sgles	SG elements describing request buffer. Will be
 * 				copied in by HAL and hence can be freed on
 * 				return from this function.
 * @param[in]	get_req_sga	function ptr to be called to get a request SG
 * 				Address (given the sge index).
 * @param[in]	get_req_sglen	function ptr to be called to get a request SG
 * 				len (given the sge index).
 * @param[in]	get_rsp_sga	function ptr to be called to get a response SG
 * 				Address (given the sge index).
 * @param[in]	get_rsp_sglen	function ptr to be called to get a response SG
 * 				len (given the sge index).
 *
 * @return FCXP instance. NULL on failure.
 */
bfa_hal_fcxp_t *
bfa_hal_fcxp_alloc(void *caller, bfa_hal_t *hal, int nreq_sgles,
		   int nrsp_sgles, bfa_fcxp_get_sgaddr_t req_sga_cbfn,
		   bfa_fcxp_get_sglen_t req_sglen_cbfn,
		   bfa_fcxp_get_sgaddr_t rsp_sga_cbfn,
		   bfa_fcxp_get_sglen_t rsp_sglen_cbfn)
{
	bfa_hal_fcxp_t *fcxp = NULL;
	uint32_t        nreq_sgpg, nrsp_sgpg;

	bfa_assert(hal != NULL);

	fcxp = bfa_hal_fcxp_get(BFA_HAL_FCXP_MOD(hal));
	if (fcxp == NULL)
		return (NULL);

	bfa_trc(hal, fcxp->fcxp_tag);

	fcxp->caller = caller;

	if (nreq_sgles == 0) {
		fcxp->use_ireqbuf = 1;
	} else {
		bfa_assert(req_sga_cbfn != NULL);
		bfa_assert(req_sglen_cbfn != NULL);

		fcxp->use_ireqbuf = 0;
		fcxp->req_sga_cbfn = req_sga_cbfn;
		fcxp->req_sglen_cbfn = req_sglen_cbfn;

		fcxp->nreq_sgles = nreq_sgles;

		/*
		 * alloc required sgpgs
		 */
		if (nreq_sgles > BFA_SGE_DATA_MAX_INLINE) {
			nreq_sgpg = BFA_SGPG_NPAGE(nreq_sgles);

			if (bfa_hal_sgpg_malloc
			    (hal, &fcxp->req_sgpg_q, nreq_sgpg)
			    != BFA_STATUS_OK) {
				/* bfa_hal_sgpg_wait(hal, &fcxp->req_sgpg_wqe,
				nreq_sgpg); */
				/*
				 * TODO
				 */
			}
		}
	}

	if (nrsp_sgles == 0) {
		fcxp->use_irspbuf = 1;
	} else {
		bfa_assert(rsp_sga_cbfn != NULL);
		bfa_assert(rsp_sglen_cbfn != NULL);

		fcxp->use_irspbuf = 0;
		fcxp->rsp_sga_cbfn = rsp_sga_cbfn;
		fcxp->rsp_sglen_cbfn = rsp_sglen_cbfn;

		fcxp->nrsp_sgles = nrsp_sgles;
		/*
		 * alloc required sgpgs
		 */
		if (nrsp_sgles > BFA_SGE_DATA_MAX_INLINE) {
			nrsp_sgpg = BFA_SGPG_NPAGE(nreq_sgles);

			if (bfa_hal_sgpg_malloc
			    (hal, &fcxp->rsp_sgpg_q, nrsp_sgpg)
			    != BFA_STATUS_OK) {
				/* bfa_hal_sgpg_wait(hal, &fcxp->rsp_sgpg_wqe,
				nrsp_sgpg); */
				/*
				 * TODO
				 */
			}
		}
	}

	return (fcxp);
}

/**
 * @brief
 * Get the internal request buffer pointer
 *
 * @param[in]	fcxp	BFA HAL fcxp pointer
 *
 * @return 		pointer to the internal request buffer
 */
void *
bfa_hal_fcxp_get_reqbuf(bfa_hal_fcxp_t *fcxp)
{
	bfa_hal_fcxp_mod_t *mod = fcxp->fcxp_mod;

	bfa_assert(fcxp->use_ireqbuf == 1);
	return (&mod->req_pld_list_kva[fcxp->fcxp_tag]);
}

/**
 * @brief
 * Get the internal response buffer pointer
 *
 * @param[in]	fcxp	BFA HAL fcxp pointer
 *
 * @return		pointer to the internal request buffer
 */
void *
bfa_hal_fcxp_get_rspbuf(bfa_hal_fcxp_t *fcxp)
{
	bfa_hal_fcxp_mod_t *mod = fcxp->fcxp_mod;

	bfa_assert(fcxp->use_irspbuf == 1);
	return (&mod->rsp_pld_list_kva[fcxp->fcxp_tag]);
}

/**
 * @brief
 * 		Free the BFA FCXP
 *
 * @param[in]	fcxp			BFA HAL fcxp pointer
 *
 * @return 		void
 */
void
bfa_hal_fcxp_free(bfa_hal_fcxp_t *fcxp)
{
	bfa_hal_fcxp_mod_t *mod = fcxp->fcxp_mod;

	bfa_assert(fcxp != NULL);
	bfa_trc(mod->hal, fcxp->fcxp_tag);
	bfa_hal_fcxp_put(fcxp);
}

/**
 * @brief
 * Send a FCXP request
 *
 * @param[in]	fcxp	BFA HAL fcxp pointer
 * @param[in]	rport	BFA HAL rport pointer. Could be left NULL for WKA rports
 * @param[in]	vf_id	virtual Fabric ID
 * @param[in]	cts	use Continous sequence
 * @param[in]	cos	fc Class of Service
 * @param[in]	reqlen	request length, does not include FCHS length
 * @param[in]	fchs	fc Header Pointer. The header content will be copied
 *			in by HAL.
 *
 * @param[in]	cbfn	call back function to be called on receiving
 * 								the response
 * @param[in]	cbarg	arg for cbfn
 * @param[in]	rsp_timeout
 *			response timeout
 *
 * @return		bfa_status_t
 */
void
bfa_hal_fcxp_send(bfa_hal_fcxp_t *fcxp, bfa_hal_rport_t *rport,
		  uint16_t vf_id, bfa_boolean_t cts, fc_cos_t cos,
		  uint32_t reqlen, fchs_t *fchs, bfa_hcb_fcxp_send_t cbfn,
		  void *cbarg, uint32_t rsp_maxlen, uint8_t rsp_timeout)
{
	bfa_fwi_fcxp_send_req_t *send_req;
	bfa_hal_t      *hal = fcxp->fcxp_mod->hal;

	send_req = bfa_hal_reqq_next(hal, BFA_HAL_REQQ_FCXP);

	bfa_fwi_h2i_set(send_req->mh, BFA_MC_FCXP, BFA_FWI_FCXP_H2I_SEND_REQ,
			bfa_hal_ioc_lpuid(hal));

	bfa_trc(hal, fcxp->fcxp_tag);

	send_req->fcxp_tag = bfa_os_htons(fcxp->fcxp_tag);

	if (rport != NULL) {
		if (rport->rport_info.max_frmsz == 0) {
			bfa_assert(0);
			send_req->max_frmsz = bfa_os_htons(FC_MAX_PDUSZ);
		} else
			send_req->max_frmsz =
				bfa_os_htons(rport->rport_info.max_frmsz);
		send_req->rport_fw_hndl = rport->fw_handle;
	} else {
		send_req->max_frmsz = bfa_os_htons(FC_MAX_PDUSZ);
		send_req->rport_fw_hndl = 0;
	}

	send_req->vf_id = bfa_os_htons(vf_id);
	send_req->req_type = 0;
	send_req->class = cos;
	send_req->rsp_timeout = rsp_timeout;
	send_req->cts = cts;

	bfa_os_memcpy((void *)&send_req->fchs, (void *)fchs, sizeof(fchs_t));

	send_req->req_len = bfa_os_htonl(reqlen);
	send_req->rsp_maxlen = bfa_os_htonl(rsp_maxlen);

	/*
	 * setup req sgles
	 */
	if (fcxp->use_ireqbuf == 1) {
		hal_fcxp_set_local_sges(send_req->req_sge, reqlen,
					BFA_FCXP_REQ_PLD_PA(fcxp));
	} else {
		if (fcxp->nreq_sgles > 0) {
			bfa_assert(fcxp->nreq_sgles == 1);
			hal_fcxp_set_local_sges(send_req->req_sge, reqlen,
						fcxp->req_sga_cbfn(fcxp->caller,
								   0));
		} else {
			bfa_assert(reqlen == 0);
			hal_fcxp_set_local_sges(send_req->rsp_sge, 0, 0);
		}
	}

	/*
	 * setup rsp sgles
	 */
	if (fcxp->use_irspbuf == 1) {
		bfa_assert(rsp_maxlen <= BFA_FCXP_MAX_IBUF_SZ);

		hal_fcxp_set_local_sges(send_req->rsp_sge, rsp_maxlen,
					BFA_FCXP_RSP_PLD_PA(fcxp));

	} else {
		if (fcxp->nrsp_sgles > 0) {
			bfa_assert(fcxp->nrsp_sgles == 1);
			hal_fcxp_set_local_sges(send_req->rsp_sge, rsp_maxlen,
						fcxp->rsp_sga_cbfn(fcxp->caller,
								   0));
		} else {
			bfa_assert(rsp_maxlen == 0);
			hal_fcxp_set_local_sges(send_req->rsp_sge, 0, 0);
		}
	}

	hal_fcxp_tx_plog(hal, reqlen, fcxp, fchs);

fcxp->send_cbfn = cbfn ? cbfn : bfa_hal_fcxp_null_comp;
	fcxp->send_cbarg = cbarg;

	bfa_hal_reqq_produce(hal, BFA_HAL_REQQ_FCXP);

	bfa_trc(hal, bfa_hal_reqq_pi(hal, BFA_HAL_REQQ_FCXP));
	bfa_trc(hal, bfa_hal_reqq_ci(hal, BFA_HAL_REQQ_FCXP));
}

/**
 * @brief
 * Abort a BFA FCXP
 *
 * @param[in]	fcxp	BFA HAL fcxp pointer
 *
 * @return 		void
 */
bfa_status_t
bfa_hal_fcxp_abort(bfa_hal_fcxp_t *fcxp)
{
	bfa_assert(0);
	return (BFA_STATUS_OK);
}

void
bfa_hal_fcxp_alloc_wait(bfa_hal_t *hal, bfa_hal_fcxp_wqe_t *wqe,
			bfa_fcxp_alloc_cbfn_t alloc_cbfn, void *alloc_cbarg)
{
	bfa_hal_fcxp_mod_t *mod = BFA_HAL_FCXP_MOD(hal);

	bfa_assert(bfa_q_is_empty(&mod->fcxp_free_q));

	wqe->alloc_cbfn = alloc_cbfn;
	wqe->alloc_cbarg = alloc_cbarg;
	bfa_q_enq(&mod->wait_q, wqe);
}

void
bfa_hal_fcxp_walloc_cancel(bfa_hal_t *hal, bfa_hal_fcxp_wqe_t *wqe)
{
	bfa_hal_fcxp_mod_t *mod = BFA_HAL_FCXP_MOD(hal);

	bfa_assert(bfa_q_is_on_q(&mod->wait_q, wqe));
	bfa_q_qe_deq(wqe);
}

void
bfa_hal_fcxp_discard(bfa_hal_fcxp_t *fcxp)
{
	fcxp->send_cbfn = bfa_hal_fcxp_null_comp;
}

/**
 * @}
 */

/**
 * @dg hal_fcxp_public HAL FCXP public functions
 * @{
 */

void
bfa_hal_fcxp_isr(bfa_hal_t *hal, bfa_msg_t *msg)
{
	switch (msg->mhdr.msg_id) {
	case BFA_FWI_FCXP_I2H_SEND_RSP:
		hal_fcxp_send_comp(hal, (bfa_fwi_fcxp_send_rsp_t *) msg);
		break;

	default:
		bfa_trc(hal, msg->mhdr.msg_id);
		bfa_assert(0);
	}
}

/**
 * @}
 */
