/*
 * Copyright (c) 2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

#include <hal/bfa_hal.h>
#include <hal/bfa_hal_flash.h>
#include <hcb/bfa_hcb.h>
#include <cs/bfa_debug.h>

BFA_TRC_FILE(HAL, FLASH);
BFA_HAL_MODULE(flash);

#define BFA_HAL_FLASH_DMA_BUF_SZ  	0x010000	/*! 64k dma buffer */

/*
 *****************************************************************************
 *
 * Internal functions
 *
 *****************************************************************************
 */

static void
__bfa_hcb_flash(void *cbarg, bfa_boolean_t complete)
{
	bfa_hal_flash_mod_t *mod = cbarg;

	if (complete)
		mod->cbfn(mod->cbarg, mod->status);
	else
		mod->op_busy = 0;
}

/**
 * @brief Set flash request scatter gather element.
 *
 * @param[in] mod - hal flash module
 * @param[in] sge - scatter gather element
 * @param[in] len - scatter gather len
 */
static void
bfa_hal_flash_sge(bfa_hal_flash_mod_t *mod, bfa_sge_t *sge, uint32_t len)
{
	bfa_addr_t      sga_zero = { {0} };

	sge->sg_len = len;
	sge->flags = BFA_SGE_DATA_LAST;
	bfa_hal_dma_addr_set(sge->sga, mod->dbuf_pa);
	bfa_sge_to_be(sge);

	sge++;
	sge->sg_len = len;
	sge->flags = BFA_SGE_PGDLEN;
	sge->sga = sga_zero;
	bfa_sge_to_be(sge);
}

/**
 * @brief Send flash attribute query request.
 *
 * @param[in] cbarg - callback argument
 */
static void
bfa_hal_flash_query_send(void *cbarg)
{
	bfa_hal_flash_mod_t *mod = cbarg;
	bfa_fwi_flash_query_req_t *msg;

	msg = bfa_hal_reqq_next(mod->hal, BFA_HAL_REQQ_FLASH);

	if (!msg) {
		bfa_hal_reqq_wait(mod->hal, BFA_HAL_REQQ_FLASH,
				  &mod->reqq_wait);
		return;
	}

	bfa_fwi_h2i_set(msg->mh, BFA_MC_FLASH, BFA_FLASH_H2I_QUERY_REQ,
			mod->hal->ioc.pci_info.pci_func);
	bfa_hal_flash_sge(mod, msg->req_sge, sizeof(bfa_flash_attr_t));
	bfa_hal_reqq_produce(mod->hal, BFA_HAL_REQQ_FLASH);

	return;
}

/**
 * @brief Send flash write request.
 *
 * @param[in] cbarg - callback argument
 */
static void
bfa_hal_flash_write_send(void *cbarg)
{
	bfa_hal_flash_mod_t *mod = cbarg;
	bfa_fwi_flash_write_req_t *msg;
	uint32_t        len;

	msg = bfa_hal_reqq_next(mod->hal, BFA_HAL_REQQ_FLASH);

	if (!msg) {
		bfa_hal_reqq_wait(mod->hal, BFA_HAL_REQQ_FLASH,
							&mod->reqq_wait);
		return;
	}

	msg->type = bfa_os_ntohl(mod->type);
	msg->instance = bfa_os_ntohl(mod->instance);
	msg->offset = bfa_os_ntohl(mod->addr_off + mod->offset);
	len = (mod->residue < BFA_HAL_FLASH_DMA_BUF_SZ) ?
				mod->residue : BFA_HAL_FLASH_DMA_BUF_SZ;
	msg->length = bfa_os_ntohl(len);

	/*
	 * indicate if it's the last msg of the whole write operation
	 */
msg->last = (len == mod->residue) ? bfa_os_ntohl(1) : 0;

	bfa_fwi_h2i_set(msg->mh, BFA_MC_FLASH, BFA_FLASH_H2I_WRITE_REQ,
					mod->hal->ioc.pci_info.pci_func);
	bfa_hal_flash_sge(mod, msg->req_sge, len);

	memcpy(mod->dbuf_kva, mod->ubuf + mod->offset, len);
	bfa_hal_reqq_produce(mod->hal, BFA_HAL_REQQ_FLASH);

	mod->residue -= len;
	mod->offset += len;

	return;
}

/**
 * @brief Send flash read request.
 *
 * @param[in] cbarg - callback argument
 */
static void
bfa_hal_flash_read_send(void *cbarg)
{
	bfa_hal_flash_mod_t *mod = cbarg;
	bfa_fwi_flash_read_req_t *msg;
	uint32_t        len;

	msg = bfa_hal_reqq_next(mod->hal, BFA_HAL_REQQ_FLASH);

	if (!msg) {
		bfa_hal_reqq_wait(mod->hal, BFA_HAL_REQQ_FLASH,
				  &mod->reqq_wait);
		return;
	}

	msg->type = bfa_os_ntohl(mod->type);
	msg->instance = bfa_os_ntohl(mod->instance);
	msg->offset = bfa_os_ntohl(mod->addr_off + mod->offset);
	len = (mod->residue < BFA_HAL_FLASH_DMA_BUF_SZ) ?
				mod->residue : BFA_HAL_FLASH_DMA_BUF_SZ;
	msg->length = bfa_os_ntohl(len);
	bfa_fwi_h2i_set(msg->mh, BFA_MC_FLASH, BFA_FLASH_H2I_READ_REQ,
			mod->hal->ioc.pci_info.pci_func);
	bfa_hal_flash_sge(mod, msg->req_sge, len);
	bfa_hal_reqq_produce(mod->hal, BFA_HAL_REQQ_FLASH);

	return;
}

/**
 * @brief Send flash erase request.
 *
 * @param[in] cbarg - callback argument
 */
static void
bfa_hal_flash_erase_send(void *cbarg)
{
	bfa_hal_flash_mod_t *mod = cbarg;
	bfa_fwi_flash_erase_req_t *msg;

	msg = bfa_hal_reqq_next(mod->hal, BFA_HAL_REQQ_FLASH);

	if (!msg) {
		bfa_hal_reqq_wait(mod->hal, BFA_HAL_REQQ_FLASH,
							&mod->reqq_wait);
		return;
	}

	msg->type = bfa_os_ntohl(mod->type);
	msg->instance = bfa_os_ntohl(mod->instance);
	bfa_fwi_h2i_set(msg->mh, BFA_MC_FLASH, BFA_FLASH_H2I_ERASE_REQ,
					mod->hal->ioc.pci_info.pci_func);
	bfa_hal_reqq_produce(mod->hal, BFA_HAL_REQQ_FLASH);

	return;
}

/**
 * @brief Update virtual and physical memory for flash.
 *
 * @param[in] cfg - ioc config
 * @param[in] ndm_len - non-dma memory length
 * @param[in] dm_len - dma memory length
 */
static void
bfa_hal_flash_meminfo(bfa_ioc_cfg_t *cfg, uint32_t *ndm_len,
		      uint32_t *dm_len)
{
	*dm_len += BFA_HAL_FLASH_DMA_BUF_SZ;
}

/**
 * @brief Attach virtual and physical memory for flash.
 *
 * @param[in] hal - hal structure
 * @param[in] bfad - driver structure
 * @param[in] cfg - ioc config
 * @param[in] meminfo - hal memory info
 * @param[in] pcidev - pci device
 */
static void
bfa_hal_flash_attach(bfa_hal_t *hal, void *bfad, bfa_ioc_cfg_t *cfg,
		     bfa_hal_meminfo_t *meminfo, bfa_hal_pcidev_t *pcidev)
{
	uint8_t        *dm_kva = NULL;
	uint64_t        dm_pa;
	bfa_hal_flash_mod_t *flash_mod = BFA_HAL_FLASH_MOD(hal);

	flash_mod->hal = hal;

	dm_kva = bfa_hal_meminfo_dma_virt(meminfo);
	dm_pa = bfa_hal_meminfo_dma_phys(meminfo);

	flash_mod->dbuf_kva = dm_kva;
	flash_mod->dbuf_pa = dm_pa;
	dm_kva += BFA_HAL_FLASH_DMA_BUF_SZ;
	dm_pa += BFA_HAL_FLASH_DMA_BUF_SZ;
	bfa_os_memset(flash_mod->dbuf_kva, 0, BFA_HAL_FLASH_DMA_BUF_SZ);

	bfa_hal_meminfo_dma_virt(meminfo) = dm_kva;
	bfa_hal_meminfo_dma_phys(meminfo) = dm_pa;
}

static void
bfa_hal_flash_initdone(bfa_hal_t *hal)
{
}

static void
bfa_hal_flash_detach(bfa_hal_t *hal)
{
}

static void
bfa_hal_flash_start(bfa_hal_t *hal)
{
}

static void
bfa_hal_flash_stop(bfa_hal_t *hal)
{
}

static void
bfa_hal_flash_iocdisable(bfa_hal_t *hal)
{
	bfa_hal_flash_mod_t *mod = &(hal->modules.flash_mod); 

	bfa_trc(hal, mod->op_busy);
	if (!mod->op_busy)
		return;

	mod->status = BFA_STATUS_IOC_FAILURE;
	bfa_hcb_queue(hal, &mod->hcb_qe, __bfa_hcb_flash, mod);
}

/*
 *****************************************************************************
 *
 * Externalized functions
 *
 *****************************************************************************
 */

/**
 * @ig hal_flash_api
 * @{
 */

/**
 * @brief Get flash attribute.
 *
 * @param[in] hal - hal structure
 * @param[in] attr - flash attribute structure
 * @param[in] cbfn - callback function
 * @param[in] cbarg - callback argument
 *
 * Return status.
 */
bfa_status_t
bfa_hal_flash_get_attr(bfa_hal_t *hal, bfa_flash_attr_t *attr,
		       bfa_hcb_flash_t cbfn, void *cbarg)
{
	bfa_hal_flash_mod_t *mod = BFA_HAL_FLASH_MOD(hal);

	bfa_trc(hal, BFA_FLASH_H2I_QUERY_REQ);

	if (!bfa_hal_ioc_is_operational(hal))
		return BFA_STATUS_IOC_NON_OP;

	if (mod->op_busy) {
		bfa_trc(hal, mod->op_busy);
		return (BFA_STATUS_DEVBUSY);
	}

	mod->op_busy = 1;

	mod->cbfn = cbfn;
	mod->cbarg = cbarg;
	mod->ubuf = (uint8_t *) attr;
	bfa_hal_reqq_winit(&mod->reqq_wait, bfa_hal_flash_query_send, mod);
	bfa_hal_flash_query_send(mod);

	return (BFA_STATUS_OK);
}

/**
 * @brief Erase flash partition.
 *
 * @param[in] hal - hal structure
 * @param[in] type - flash partition type
 * @param[in] instance - flash partition instance
 * @param[in] cbfn - callback function
 * @param[in] cbarg - callback argument
 *
 * Return status.
 */
bfa_status_t
bfa_hal_flash_erase_part(bfa_hal_t *hal, bfa_flash_part_type_t type,
			 uint32_t instance, bfa_hcb_flash_t cbfn, void *cbarg)
{
	bfa_hal_flash_mod_t *mod = BFA_HAL_FLASH_MOD(hal);

	bfa_trc(hal, BFA_FLASH_H2I_ERASE_REQ);
	bfa_trc(hal, type);
	bfa_trc(hal, instance);

	if (!bfa_hal_ioc_is_operational(hal))
		return BFA_STATUS_IOC_NON_OP;

	if (mod->op_busy) {
		bfa_trc(hal, mod->op_busy);
		return (BFA_STATUS_DEVBUSY);
	}

	mod->op_busy = 1;

	mod->cbfn = cbfn;
	mod->cbarg = cbarg;
	mod->type = type;
	mod->instance = instance;
	bfa_hal_reqq_winit(&mod->reqq_wait, bfa_hal_flash_erase_send, mod);

	bfa_hal_flash_erase_send(mod);

	return (BFA_STATUS_OK);
}

/**
 * @brief Update flash partition.
 *
 * @param[in] hal - hal structure
 * @param[in] type - flash partition type
 * @param[in] instance - flash partition instance
 * @param[in] buf - update data buffer
 * @param[in] len - data buffer length
 * @param[in] offset - offset relative to the partition starting address
 * @param[in] cbfn - callback function
 * @param[in] cbarg - callback argument
 *
 * Return status.
 */
bfa_status_t
bfa_hal_flash_update_part(bfa_hal_t *hal, bfa_flash_part_type_t type,
			  uint32_t instance, void *buf, uint32_t len,
			  uint32_t offset, bfa_hcb_flash_t cbfn, void *cbarg)
{
	bfa_hal_flash_mod_t *mod = BFA_HAL_FLASH_MOD(hal);

	bfa_trc(hal, BFA_FLASH_H2I_WRITE_REQ);
	bfa_trc(hal, type);
	bfa_trc(hal, instance);
	bfa_trc(hal, len);
	bfa_trc(hal, offset);

	if (!bfa_hal_ioc_is_operational(hal))
		return BFA_STATUS_IOC_NON_OP;

	/**
	 * 'len' must be in word (4-byte) boundary
	 * 'offset' must be in sector (16kb) boundary
	 */
	if ((len & 0x03) || (offset & 0x00003FFF))
		return (BFA_STATUS_FLASH_BAD_LEN);

	if (mod->op_busy) {
		bfa_trc(hal, mod->op_busy);
		return (BFA_STATUS_DEVBUSY);
	}

	mod->op_busy = 1;

	mod->cbfn = cbfn;
	mod->cbarg = cbarg;
	mod->type = type;
	mod->instance = instance;
	mod->residue = len;
	mod->offset = 0;
	mod->addr_off = offset;
	mod->ubuf = buf;
	bfa_hal_reqq_winit(&mod->reqq_wait, bfa_hal_flash_write_send, mod);

	bfa_hal_flash_write_send(mod);

	return (BFA_STATUS_OK);
}

/**
 * @brief Read flash partition.
 *
 * @param[in] hal - hal structure
 * @param[in] type - flash partition type
 * @param[in] instance - flash partition instance
 * @param[in] buf - read data buffer
 * @param[in] len - data buffer length
 * @param[in] offset - offset relative to the partition starting address
 * @param[in] cbfn - callback function
 * @param[in] cbarg - callback argument
 *
 * Return status.
 */
bfa_status_t
bfa_hal_flash_read_part(bfa_hal_t *hal, bfa_flash_part_type_t type,
			uint32_t instance, void *buf, uint32_t len,
			uint32_t offset, bfa_hcb_flash_t cbfn, void *cbarg)
{
	bfa_hal_flash_mod_t *mod = BFA_HAL_FLASH_MOD(hal);

	bfa_trc(hal, BFA_FLASH_H2I_READ_REQ);
	bfa_trc(hal, type);
	bfa_trc(hal, instance);
	bfa_trc(hal, len);
	bfa_trc(hal, offset);

	if (!bfa_hal_ioc_is_operational(hal))
		return BFA_STATUS_IOC_NON_OP;

	/**
	 * 'len' must be in word (4-byte) boundary
	 * 'offset' must be in sector (16kb) boundary
	 */
	if ((len & 0x03) || (offset & 0x00003FFF))
		return (BFA_STATUS_FLASH_BAD_LEN);

	if (mod->op_busy) {
		bfa_trc(hal, mod->op_busy);
		return (BFA_STATUS_DEVBUSY);
	}

	mod->op_busy = 1;

	mod->cbfn = cbfn;
	mod->cbarg = cbarg;
	mod->type = type;
	mod->instance = instance;
	mod->residue = len;
	mod->offset = 0;
	mod->addr_off = offset;
	mod->ubuf = buf;
	bfa_hal_reqq_winit(&mod->reqq_wait, bfa_hal_flash_read_send, mod);
	bfa_hal_flash_read_send(mod);

	return (BFA_STATUS_OK);
}

/**
 * @brief Process flash response messages upon receiving interrupts.
 *
 * @param[in] hal - hal structure
 * @param[in] msg - message structure
 */
void
bfa_hal_flash_intr(bfa_hal_t *hal, bfa_msg_t *msg)
{
	uint32_t        status;
	bfa_hal_flash_mod_t *mod = BFA_HAL_FLASH_MOD(hal);

	union {
		bfa_fwi_flash_query_rsp_t *query;
		bfa_fwi_flash_erase_rsp_t *erase;
		bfa_fwi_flash_write_rsp_t *write;
		bfa_fwi_flash_read_rsp_t *read;
		bfa_msg_t      *msg;
	} m;

	m.msg = msg;
	bfa_trc(hal, msg->mhdr.msg_id);

	if (!mod->op_busy) {
		/*
		 * receiving response after ioc failure 
		 */
		bfa_trc(hal, 0x9999);
		return;
	}

	switch (msg->mhdr.msg_id) {
	case BFA_FLASH_I2H_QUERY_RSP:
		status = bfa_os_ntohl(m.query->status);
		bfa_trc(hal, status);

		if (status == BFA_STATUS_OK) {
			uint32_t        i;
			bfa_flash_attr_t *attr, *f;

			attr = (bfa_flash_attr_t *) mod->ubuf;
			f = (bfa_flash_attr_t *) mod->dbuf_kva;
			attr->status = bfa_os_ntohl(f->status);
			attr->npart = bfa_os_ntohl(f->npart);
			bfa_trc(hal, attr->status);
			bfa_trc(hal, attr->npart);
			for (i = 0; i < attr->npart; i++) {
				attr->part[i].part_type =
					bfa_os_ntohl(f->part[i].part_type);
				attr->part[i].part_instance =
					bfa_os_ntohl(f->part[i].part_instance);
				attr->part[i].part_off =
					bfa_os_ntohl(f->part[i].part_off);
				attr->part[i].part_size =
					bfa_os_ntohl(f->part[i].part_size);
				attr->part[i].part_len =
					bfa_os_ntohl(f->part[i].part_len);
				attr->part[i].part_status =
					bfa_os_ntohl(f->part[i].part_status);
			}

		}

		mod->status = status;
		bfa_hcb_queue(mod->hal, &mod->hcb_qe, __bfa_hcb_flash, mod);
		break;
	case BFA_FLASH_I2H_ERASE_RSP:
		status = bfa_os_ntohl(m.erase->status);
		bfa_trc(hal, status);

		mod->status = status;
		bfa_hcb_queue(mod->hal, &mod->hcb_qe, __bfa_hcb_flash, mod);
		break;
	case BFA_FLASH_I2H_WRITE_RSP:
		status = bfa_os_ntohl(m.write->status);
		bfa_trc(hal, status);

		if (status != BFA_STATUS_OK || mod->residue == 0) {
			mod->status = status;
			bfa_hcb_queue(mod->hal, &mod->hcb_qe, __bfa_hcb_flash,
				      mod);
		} else {
			bfa_trc(hal, mod->offset);
			bfa_hal_flash_write_send(mod);
		}
		break;
	case BFA_FLASH_I2H_READ_RSP:
		status = bfa_os_ntohl(m.read->status);
		bfa_trc(hal, status);

		if (status != BFA_STATUS_OK) {
			mod->status = status;
			bfa_hcb_queue(mod->hal, &mod->hcb_qe, __bfa_hcb_flash,
				      mod);
		} else {
			uint32_t        len = bfa_os_ntohl(m.read->length);
			bfa_trc(hal, mod->offset);
			bfa_trc(hal, len);

			memcpy(mod->ubuf + mod->offset, mod->dbuf_kva, len);
			mod->residue -= len;
			mod->offset += len;

			if (mod->residue == 0) {
				mod->status = status;
				bfa_hcb_queue(mod->hal, &mod->hcb_qe,
					      __bfa_hcb_flash, mod);
			} else {
				bfa_hal_flash_read_send(mod);
			}
		}
		break;
	default:
		bfa_assert(0);
	}
}

/**
 * @}
 */
