/*
 * Copyright (c) 2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

/**
 * @file bfa_hal_flash_raw.c flash raw operations
 *
 */

#include <bfa_os.h>
#include <fwi/bfa_fwi.h>
#include <hal_reg.h>
#include <defs/bfa_defs_status.h>
#include <hal/bfa_hal_flash_raw.h>

/*
 * register definitions
 */
#define FLI_CMD_REG                     0x0001d000
#define FLI_WRDATA_REG                  0x0001d00c
#define FLI_RDDATA_REG                  0x0001d010
#define FLI_ADDR_REG                    0x0001d004
#define FLI_DEV_STATUS_REG              0x0001d014

#define BFA_HAL_FLASH_FIFO_SIZE         128	/* fifo size */
#define BFA_HAL_FLASH_CHECK_MAX		10000	/* max # of status check */
#define BFA_HAL_FLASH_BLOCKING_OP_MAX	1000000	/* max # of blocking op check */
#define BFA_HAL_FLASH_WIP_MASK		0x01	/* write in progress bit mask */

typedef enum {
	BFA_HAL_FLASH_FAST_READ		= 0x0b,	/* fast read */
	BFA_HAL_FLASH_WRITE_ENABLE	= 0x06,	/* write enable */
	BFA_HAL_FLASH_SECTOR_ERASE	= 0xd8,	/* sector erase */
	BFA_HAL_FLASH_WRITE		= 0x02,	/* write */
	BFA_HAL_FLASH_READ_STATUS	= 0x05,	/* read status */
} bfa_hal_flash_cmd_t;

/**
 * @brief hardware error definition
 */
typedef enum {
	BFA_HAL_FLASH_NOT_PRESENT	= -1,	/*!< flash not present */
	BFA_HAL_FLASH_UNINIT		= -2,	/*!< flash not initialized */
	BFA_HAL_FLASH_BAD		= -3,	/*!< flash bad */
	BFA_HAL_FLASH_BUSY		= -4,	/*!< flash busy */
	BFA_HAL_FLASH_ERR_CMD_ACT	= -5,	/*!< command active never
						 * cleared
						 */
	BFA_HAL_FLASH_ERR_FIFO_CNT	= -6, 	/*!< fifo count never cleared */
	BFA_HAL_FLASH_ERR_WIP		= -7,	/*!< write-in-progress never
						 * cleared
						 */
	BFA_HAL_FLASH_ERR_TIMEOUT	= -8,	/*!< fli timeout */
	BFA_HAL_FLASH_ERR_LEN		= -9,	/*!< invalid length */
} bfa_hal_flash_err_t;

/**
 * @brief flash command register data structure
 */
typedef union {
	struct {
#ifdef __BIGENDIAN
		uint32_t        act:1;
		uint32_t        rsv:1;
		uint32_t        write_cnt:9;
		uint32_t        read_cnt:9;
		uint32_t        addr_cnt:4;
		uint32_t        cmd:8;
#else
		uint32_t        cmd:8;
		uint32_t        addr_cnt:4;
		uint32_t        read_cnt:9;
		uint32_t        write_cnt:9;
		uint32_t        rsv:1;
		uint32_t        act:1;
#endif
	} r;
	uint32_t        i;
} bfa_hal_flash_cmd_reg_t;

/**
 * @brief flash device status register data structure
 */
typedef union {
	struct {
#ifdef __BIGENDIAN
		uint32_t        rsv:21;
		uint32_t        fifo_cnt:6;
		uint32_t        busy:1;
		uint32_t        init_status:1;
		uint32_t        present:1;
		uint32_t        bad:1;
		uint32_t        good:1;
#else
		uint32_t        good:1;
		uint32_t        bad:1;
		uint32_t        present:1;
		uint32_t        init_status:1;
		uint32_t        busy:1;
		uint32_t        fifo_cnt:6;
		uint32_t        rsv:21;
#endif
	} r;
	uint32_t        i;
} bfa_hal_flash_dev_status_reg_t;

/**
 * @brief flash address register data structure
 */
typedef union {
	struct {
#ifdef __BIGENDIAN
		uint32_t        addr:24;
		uint32_t        dummy:8;
#else
		uint32_t        dummy:8;
		uint32_t        addr:24;
#endif
	} r;
	uint32_t        i;
} bfa_hal_flash_addr_reg_t;

/**
 * dg flash_raw_private Flash raw private functions
 * @{
 */

/**
 * @brief
 * Flush FLI data fifo.
 *
 * @param[in] pci_bar - pci bar address
 * @param[in] dev_status - device status
 *
 * Return 0 on success, negative error number on error.
 */
static          int32_t
bfa_hal_flash_fifo_flush(bfa_os_addr_t pci_bar,
			 bfa_hal_flash_dev_status_reg_t dev_status)
{
	int32_t         i;
	uint32_t        t;

	for (i = 0; i < dev_status.r.fifo_cnt; i += 4)
		t = bfa_hal_reg_read(pci_bar + FLI_RDDATA_REG);

	/*
	 * Check the device status. It may take some time.
	 */
	for (i = 0; i < BFA_HAL_FLASH_CHECK_MAX; i++) {
		dev_status.i = bfa_hal_reg_read(pci_bar + FLI_DEV_STATUS_REG);
		if (!dev_status.r.fifo_cnt)
			break;
	}

	if (dev_status.r.fifo_cnt)
		return (BFA_HAL_FLASH_ERR_FIFO_CNT);

	return 0;
}

/**
 * @brief
 * Enable flash write.
 *
 * @param[in] pci_bar - pci bar address
 *
 * Return 0 on success, negative error number on error.
 */
static          int32_t
bfa_hal_flash_write_enable(bfa_os_addr_t pci_bar)
{
	int32_t         i;
	bfa_hal_flash_cmd_reg_t cmd;

	cmd.r.act = 1;
	cmd.r.write_cnt = 0;
	cmd.r.read_cnt = 0;
	cmd.r.addr_cnt = 0;
	cmd.r.cmd = BFA_HAL_FLASH_WRITE_ENABLE;
	bfa_hal_reg_write((pci_bar + FLI_CMD_REG), cmd.i);

	for (i = 0; i < BFA_HAL_FLASH_CHECK_MAX; i++) {
		cmd.i = bfa_hal_reg_read(pci_bar + FLI_CMD_REG);
		if (!cmd.r.act)
			break;
	}

	/*
	 * return error if active bit is never cleared
	 */
	if (cmd.r.act)
		return (BFA_HAL_FLASH_ERR_CMD_ACT);

	return 0;
}

/**
 * @brief
 * Read flash status.
 *
 * @param[in] pci_bar - pci bar address
 *
 * Return 0 on success, negative error number on error.
*/
static          int32_t
bfa_hal_flash_status_read(bfa_os_addr_t pci_bar)
{
	bfa_hal_flash_cmd_reg_t cmd;
	bfa_hal_flash_dev_status_reg_t dev_status;
	int32_t         i, status;

	dev_status.i = bfa_hal_reg_read(pci_bar + FLI_DEV_STATUS_REG);

	if (dev_status.r.init_status)
		return (BFA_HAL_FLASH_UNINIT);

	if (!dev_status.r.present)
		return (BFA_HAL_FLASH_NOT_PRESENT);

	if (dev_status.r.busy)
		return (BFA_HAL_FLASH_BUSY);

	if (dev_status.r.fifo_cnt) {
		status = bfa_hal_flash_fifo_flush(pci_bar, dev_status);
		if (status < 0)
			return (status);
	}

	cmd.r.act = 1;
	cmd.r.write_cnt = 0;
	cmd.r.read_cnt = 4;
	cmd.r.addr_cnt = 0;
	cmd.r.cmd = BFA_HAL_FLASH_READ_STATUS;
	bfa_hal_reg_write((pci_bar + FLI_CMD_REG), cmd.i);

	for (i = 0; i < BFA_HAL_FLASH_CHECK_MAX; i++) {
		cmd.i = bfa_hal_reg_read(pci_bar + FLI_CMD_REG);
		if (!cmd.r.act)
			break;
	}

	/*
	 * return error if active bit is never cleared
	 */
	if (cmd.r.act)
		return (BFA_HAL_FLASH_ERR_CMD_ACT);

	status = bfa_hal_reg_read(pci_bar + FLI_RDDATA_REG) >> 24;

	/*
	 * Make sure data fifo is completely flashed out.
	 * It may take time.
	 */
	for (i = 0; i < BFA_HAL_FLASH_CHECK_MAX; i++) {
		dev_status.i = bfa_hal_reg_read(pci_bar + FLI_DEV_STATUS_REG);
		if (!dev_status.r.fifo_cnt)
			break;
	}

	/*
	 * flush fifo if fifo count is never 0
	 */
	if (dev_status.r.fifo_cnt)
		bfa_hal_flash_fifo_flush(pci_bar, dev_status);

	return (status);
}

/**
 * @brief
 * Start flash read operation.
 *
 * @param[in] pci_bar - pci bar address
 * @param[in] offset - flash address offset
 * @param[in] len - read data length
 * @param[in] buf - read data buffer
 *
 * Return 0 on success, negative error number on error.
 */
static          int32_t
bfa_hal_flash_read_start(bfa_os_addr_t pci_bar, uint32_t offset, uint32_t len,
			 char *buf)
{
	int32_t         status;
	bfa_hal_flash_cmd_reg_t cmd;
	bfa_hal_flash_addr_reg_t addr;

	/*
	 * len must be mutiple of 4 and not exceeding fifo size
	 */
	if (len == 0 || len > BFA_HAL_FLASH_FIFO_SIZE || (len & 0x03) != 0)
		return (BFA_HAL_FLASH_ERR_LEN);

	/*
	 * check status
	 */
	status = bfa_hal_flash_status_read(pci_bar);
	if (status < 0)
		return (status);

	/*
	 * check if write-in-progress bit is cleared
	 */
	if (status & BFA_HAL_FLASH_WIP_MASK)
		return (BFA_HAL_FLASH_ERR_WIP);

	/*
	 * set up address
	 */
	addr.r.addr = offset & 0x00ffffff;
	addr.r.dummy = 0;
	bfa_hal_reg_write((pci_bar + FLI_ADDR_REG), addr.i);

	/*
	 * set up cmd
	 */
	cmd.r.act = 1;
	cmd.r.write_cnt = 0;
	cmd.r.read_cnt = len;
	cmd.r.addr_cnt = 4;
	cmd.r.cmd = BFA_HAL_FLASH_FAST_READ;
	bfa_hal_reg_write((pci_bar + FLI_CMD_REG), cmd.i);

	return 0;
}

/**
 * @brief
 * Check flash read operation.

 * @param[in] pci_bar - pci bar address
 *
 * Return flash device status, 1 if busy, 0 if not.
 */
static          int32_t
bfa_hal_flash_read_check(bfa_os_addr_t pci_bar)
{
	bfa_hal_flash_dev_status_reg_t dev_status;

	dev_status.i = bfa_hal_reg_read(pci_bar + FLI_DEV_STATUS_REG);
	return (dev_status.r.busy);
}

/**
 * @brief
 * End flash read operation.
 *
 * @param[in] pci_bar - pci bar address
 * @param[in] len - read data length
 * @param[in] buf - read data buffer
 *
 */
static void
bfa_hal_flash_read_end(bfa_os_addr_t pci_bar, uint32_t len, char *buf)
{

	int32_t         i;
	bfa_hal_flash_dev_status_reg_t dev_status;

	/*
	 * read data fifo up to 32 words
	 */
	for (i = 0; i < len; i += 4) {
		uint32_t        w = bfa_hal_reg_read(pci_bar + FLI_RDDATA_REG);
		*((uint32_t *) (buf + i)) = bfa_os_ntohl(w);
	}

	/*
	 * Make sure fifo count is 0. It may take time.
	 */
	for (i = 0; i < BFA_HAL_FLASH_CHECK_MAX; i++) {
		dev_status.i = bfa_hal_reg_read(pci_bar + FLI_DEV_STATUS_REG);
		if (!dev_status.r.fifo_cnt)
			break;
	}

	if (dev_status.r.fifo_cnt)
		bfa_hal_flash_fifo_flush(pci_bar, dev_status);
}

/**
 * @brief
 * Start flash write operation.
 *
 * @param[in] pci_bar - pci bar address
 * @param[in] offset - flash address offset
 * @param[in] len - write data length
 * @param[in] buf - write data buffer
 *
 * Return 0 on success, negative error number on error.
 */
static          int32_t
bfa_hal_flash_write_start(bfa_os_addr_t pci_bar, uint32_t offset, uint32_t len,
			  char *buf)
{
	int32_t         i, status;
	bfa_hal_flash_cmd_reg_t cmd;
	bfa_hal_flash_addr_reg_t addr;

	/*
	 * len must be mutiple of 4 and not exceeding fifo size
	 */
	if (len == 0 || len > BFA_HAL_FLASH_FIFO_SIZE || (len & 0x03) != 0)
		return (BFA_HAL_FLASH_ERR_LEN);

	/*
	 * check status
	 */
	status = bfa_hal_flash_status_read(pci_bar);
	if (status < 0)
		return (status);

	/*
	 * check if write-in-progress bit is cleared
	 */
	if (status & BFA_HAL_FLASH_WIP_MASK)
		return (BFA_HAL_FLASH_ERR_WIP);

	/*
	 * write enable
	 */
	status = bfa_hal_flash_write_enable(pci_bar);
	if (status < 0)
		return (status);

	/*
	 * set up address
	 */
	addr.r.addr = offset & 0x00ffffff;
	addr.r.dummy = 0;
	bfa_hal_reg_write((pci_bar + FLI_ADDR_REG), addr.i);

	/*
	 * set up data fifo up to 32 words
	 */
	for (i = 0; i < len; i += 4) {
		uint32_t        data = *((uint32_t *) (buf + i));
		data = bfa_os_ntohl(data);
		bfa_hal_reg_write((pci_bar + FLI_WRDATA_REG), data);
	}

	/*
	 * set up cmd
	 */
	cmd.r.act = 1;
	cmd.r.write_cnt = len;
	cmd.r.read_cnt = 0;
	cmd.r.addr_cnt = 3;
	cmd.r.cmd = BFA_HAL_FLASH_WRITE;
	bfa_hal_reg_write((pci_bar + FLI_CMD_REG), cmd.i);

	return 0;
}

/**
 * @brief
 * Check flash write operation.
 *
 * @param[in] pci_bar - pci bar address
 *
 * Return flash device status, 1 if busy, 0 if not.
 */
static          int32_t
bfa_hal_flash_write_check(bfa_os_addr_t pci_bar)
{
	int32_t         status;
	bfa_hal_flash_dev_status_reg_t dev_status;

	dev_status.i = bfa_hal_reg_read(pci_bar + FLI_DEV_STATUS_REG);
	if (dev_status.r.busy)
		return 1;

	status = bfa_hal_flash_status_read(pci_bar);
	if (status < 0 || (status & BFA_HAL_FLASH_WIP_MASK))
		return 1;
	else
		return 0;
}

/**
 * @brief
 * Start flash sector erase operation.
 *
 * @param[in] pci_bar - pci bar address
 * @param[in] offset - flash address offset
 *
 * Return 0 on success, negative error number on error.
 */
static          int32_t
bfa_hal_flash_sector_erase_start(bfa_os_addr_t pci_bar, uint32_t offset)
{
	int32_t         status;
	bfa_hal_flash_cmd_reg_t cmd;
	bfa_hal_flash_addr_reg_t addr;

	/*
	 * check status
	 */
	status = bfa_hal_flash_status_read(pci_bar);
	if (status < 0)
		return (status);

	/*
	 * check if write-in-progress bit is cleared
	 */
	if (status & BFA_HAL_FLASH_WIP_MASK)
		return (BFA_HAL_FLASH_ERR_WIP);

	/*
	 * write enable
	 */
	status = bfa_hal_flash_write_enable(pci_bar);
	if (status < 0)
		return (status);

	/*
	 * ASIC bug workaround on sector erase setup
	 */
	addr.r.addr = 0;
	addr.r.dummy = 0;
	bfa_hal_reg_write((pci_bar + FLI_ADDR_REG), addr.i);

	bfa_hal_reg_write((pci_bar + FLI_WRDATA_REG), offset >> 16);

	cmd.r.act = 1;
	cmd.r.write_cnt = 3;
	cmd.r.read_cnt = 0;
	cmd.r.addr_cnt = 0;
	cmd.r.cmd = BFA_HAL_FLASH_SECTOR_ERASE;
	bfa_hal_reg_write((pci_bar + FLI_CMD_REG), cmd.i);

	return 0;
}

/**
 * @brief
 * Check flash erase operation.
 *
 * @param[in] pci_bar - pci bar address
 *
 * Return flash erase status, 1 if busy, 0 if not.
 */
static          int32_t
bfa_hal_flash_erase_check(bfa_os_addr_t pci_bar)
{
	int32_t         status;

	status = bfa_hal_flash_status_read(pci_bar);
	if (status < 0 || (status & BFA_HAL_FLASH_WIP_MASK))
		return 1;
	else
		return 0;
}

/**
 * @}
 *
 * dg flash_raw_public Flash raw operation public functions
 * @{
 */

/**
 * @brief
 * Perform flash raw read.
 *
 * @param[in] pci_bar - pci bar address
 * @param[in] offset - flash partition address offset
 * @param[in] buf - read data buffer
 * @param[in] len - read data length
 *
 * Return status.
 */
bfa_status_t
bfa_hal_flash_raw_read(bfa_os_addr_t pci_bar, uint32_t offset, char *buf,
		       uint32_t len)
{
	int32_t         n, status;
	uint32_t        off, l, residue;

	residue = len;
	off = 0;

	while (residue) {
		l = (residue > BFA_HAL_FLASH_FIFO_SIZE) ?
					BFA_HAL_FLASH_FIFO_SIZE : residue;

		status = bfa_hal_flash_read_start(pci_bar, offset + off, l,
								&buf[off]);
		if (status < 0)
			return (BFA_STATUS_FAILED);

		n = BFA_HAL_FLASH_BLOCKING_OP_MAX;
		while (bfa_hal_flash_read_check(pci_bar)) {
			if (--n <= 0)
				return (BFA_STATUS_FAILED);
		}

		bfa_hal_flash_read_end(pci_bar, l, &buf[off]);

		residue -= l;
		off += l;
	}

	return (BFA_STATUS_OK);
}

/**
 * @brief
 * Perform flash raw write.
 *
 * @param[in] pci_bar - pci bar address
 * @param[in] offset - flash partition address offset
 * @param[in] buf - write data buffer
 * @param[in] len - write data length
 *
 * Return status.
 */
bfa_status_t
bfa_hal_flash_raw_write(bfa_os_addr_t pci_bar, uint32_t offset, char *buf,
			uint32_t len)
{
	int32_t         n, status;
	uint32_t        off, l, residue;

	status = bfa_hal_flash_sector_erase_start(pci_bar, offset);
	if (status < 0)
		return (BFA_STATUS_FAILED);

	n = BFA_HAL_FLASH_BLOCKING_OP_MAX;
	while (bfa_hal_flash_erase_check(pci_bar)) {
		if (--n <= 0)
			return (BFA_STATUS_FAILED);

		bfa_os_udelay(10000);
	}

	residue = len;
	off = 0;
	while (residue) {
		l = (residue > BFA_HAL_FLASH_FIFO_SIZE) ?
					BFA_HAL_FLASH_FIFO_SIZE : residue;

		status = bfa_hal_flash_write_start(pci_bar, offset + off, l,
								&buf[off]);
		if (status < 0)
			return (BFA_STATUS_FAILED);

		n = BFA_HAL_FLASH_BLOCKING_OP_MAX;
		while (bfa_hal_flash_write_check(pci_bar)) {
			if (--n <= 0)
				return (BFA_STATUS_FAILED);
		}

		residue -= l;
		off += l;
	}

	return (BFA_STATUS_OK);
}

/**
 * @}
 */
