/*
 * Copyright (c) 2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */
#include "bfa_hal_ioc_priv.h"
#include "hal_callback.h"
#include <cs/bfa_debug.h>
#include <hal_priv.h>
#include <hal/bfa_hal_wwn.h>
#include <hal/bfa_hal_pm.h>
#include <log/bfa_log_hal.h>
#include <fwi/bfa_fwi_boot.h>
#include <cs/bfa_plog.h>
#include <aen/bfal_aen_ioc.h>

BFA_TRC_FILE(HAL, IOC);

/**
 * IOC local definitions
 */
#define BFA_HAL_IOC_TOV		2000	/* msecs */
#define BFA_HAL_IOC_HB_TOV      2000	/* msecs */
#define BFA_HAL_IOC_HB_FAIL_MAX 10
#define BFA_HAL_IOC_HWINIT_MAX	2
#define BFA_HAL_FWIMG_MINSZ     (16 * 1024)

#define bfa_hal_ioc_timer_start(__ioc)					\
	bfa_hal_timer_start((__ioc)->hal, &(__ioc)->ioc_timer,          \
			bfa_hal_ioc_timeout, (__ioc), BFA_HAL_IOC_TOV)
#define bfa_hal_ioc_timer_stop(__ioc)   bfa_hal_timer_stop(&(__ioc)->ioc_timer)

/*
 * forward declarations
 */
static void     bfa_hal_hw_sem_get(bfa_hal_ioc_t *ioc);
static void     bfa_hal_hw_sem_release(bfa_hal_ioc_t *ioc);
static void     bfa_hal_hw_sem_get_cancel(bfa_hal_ioc_t *ioc);
static void     bfa_hal_ioc_hwinit(bfa_hal_ioc_t *ioc, bfa_boolean_t force);
static void     bfa_hal_ioc_timeout(void *ioc);
static void     bfa_hal_ioc_send_cfg(bfa_hal_ioc_t *ioc);
static void     bfa_hal_ioc_send_enable(bfa_hal_ioc_t *ioc);
static void     bfa_hal_ioc_send_disable(bfa_hal_ioc_t *ioc);
static void     bfa_hal_ioc_hb_monitor(bfa_hal_ioc_t *ioc);
static void     bfa_hal_ioc_hb_stop(bfa_hal_ioc_t *ioc);
static void     bfa_hal_ioc_initdone_submod(bfa_hal_t *hal);
static void     bfa_hal_ioc_start_submod(bfa_hal_t *hal);
static void     bfa_hal_ioc_iocdisable_submod(bfa_hal_t *hal);
static void     __bfa_hcb_init(void *cbarg, bfa_boolean_t complete);
static void     __bfa_hcb_stop(void *cbarg, bfa_boolean_t complete);
static void     __bfa_hcb_ioc_disable(void *cbarg, bfa_boolean_t complete);
static void     bfa_hal_ioc_callback(bfa_hal_ioc_t *ioc);

static void     bfa_hal_ioc_stats_clear(void *cbarg);
static void     bfa_hal_ioc_stats_swap(bfa_ioc_stats_t *d,
				       bfa_ioc_stats_t *s);
static void     __bfa_hcb_ioc_stats_clr(void *cbarg, bfa_boolean_t complete);
static void     bfa_hal_ioc_stats_clr_timeout(void *cbarg);
static void     __bfa_hcb_ioc_stats(void *cbarg, bfa_boolean_t complete);
static void     bfa_hal_ioc_stats_timeout(void *cbarg);
static void     bfa_hal_ioc_reset_queues(bfa_hal_ioc_t *ioc);
static void     bfa_hal_ioc_reset(bfa_hal_ioc_t *ioc, bfa_boolean_t force);

/**
 * @ig hal_ioc_sm
 * @{
 * @img ioc_sm.jpg
 */

/**
 * IOC state machine events
 */
typedef enum {
	IOC_E_INIT	= 1,
	IOC_E_ENABLE	= 2,
	IOC_E_DISABLE	= 3,
	IOC_E_TIMEOUT	= 4,
	IOC_E_FWREADY	= 5,
	IOC_E_FWRSP_CFG	= 6,
	IOC_E_FWRSP_ENABLE = 7,
	IOC_E_FWRSP_DISABLE = 8,
	IOC_E_HBFAIL	= 9,	/* !< heartbeat failure */
	IOC_E_SEMLOCKED	= 10,	/* !< h/w semaphore is locked */
} ioc_event_t;

static void     bfa_hal_ioc_sm_reset(bfa_hal_ioc_t *ioc, ioc_event_t event);
static void     bfa_hal_ioc_sm_semwait(bfa_hal_ioc_t *ioc, ioc_event_t event);
static void     bfa_hal_ioc_sm_hwinit(bfa_hal_ioc_t *ioc, ioc_event_t event);
static void     bfa_hal_ioc_sm_cfginit(bfa_hal_ioc_t *ioc, ioc_event_t event);
static void     bfa_hal_ioc_sm_cfgdone(bfa_hal_ioc_t *ioc, ioc_event_t event);
static void     bfa_hal_ioc_sm_op(bfa_hal_ioc_t *ioc, ioc_event_t event);
static void     bfa_hal_ioc_sm_hwfail(bfa_hal_ioc_t *ioc, ioc_event_t event);
static void     bfa_hal_ioc_sm_hbfail(bfa_hal_ioc_t *ioc, ioc_event_t event);
static void     bfa_hal_ioc_sm_disabling(bfa_hal_ioc_t *ioc,
					 ioc_event_t event);
static void     bfa_hal_ioc_sm_disabled(bfa_hal_ioc_t *ioc, ioc_event_t event);
static void     bfa_hal_ioc_sm_recover(bfa_hal_ioc_t *ioc, ioc_event_t event);

static bfa_sm_table_t ioc_sm_table[] = {
	{BFA_SM(bfa_hal_ioc_sm_reset), BFA_IOC_RESET},
	{BFA_SM(bfa_hal_ioc_sm_semwait), BFA_IOC_SEMWAIT},
	{BFA_SM(bfa_hal_ioc_sm_hwinit), BFA_IOC_HWINIT},
	{BFA_SM(bfa_hal_ioc_sm_cfginit), BFA_IOC_CFGINIT},
	{BFA_SM(bfa_hal_ioc_sm_cfgdone), BFA_IOC_CFGDONE},
	{BFA_SM(bfa_hal_ioc_sm_op), BFA_IOC_OPERATIONAL},
	{BFA_SM(bfa_hal_ioc_sm_hwfail), BFA_IOC_HWFAIL},
	{BFA_SM(bfa_hal_ioc_sm_hbfail), BFA_IOC_HBFAIL},
	{BFA_SM(bfa_hal_ioc_sm_disabling), BFA_IOC_DISABLING},
	{BFA_SM(bfa_hal_ioc_sm_disabled), BFA_IOC_DISABLED},
	{BFA_SM(bfa_hal_ioc_sm_recover), BFA_IOC_RECOVER},
};

/**
 * @brief
 * Beginning state. IOC is in reset state.
 */
static void
bfa_hal_ioc_sm_reset(bfa_hal_ioc_t *ioc, ioc_event_t event)
{
	bfa_trc(ioc->hal, event);

	switch (event) {
	case IOC_E_INIT:
		bfa_sm_set_state(ioc, bfa_hal_ioc_sm_semwait);
		bfa_hal_hw_sem_get(ioc);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * Awaiting release of h/w semaphore.
 */
static void
bfa_hal_ioc_sm_semwait(bfa_hal_ioc_t *ioc, ioc_event_t event)
{
	bfa_trc(ioc->hal, event);

	switch (event) {
	case IOC_E_SEMLOCKED:
		bfa_sm_set_state(ioc, bfa_hal_ioc_sm_hwinit);
		bfa_hal_ioc_timer_start(ioc);
		ioc->hwinit_count = 0;
		bfa_hal_ioc_reset(ioc, BFA_FALSE);
		break;

	case IOC_E_DISABLE:
		bfa_sm_set_state(ioc, bfa_hal_ioc_sm_disabled);
		bfa_hal_intr_disable(ioc->hal);
		bfa_hal_hw_sem_get_cancel(ioc);
		bfa_hal_ioc_callback(ioc);
		break;

	case IOC_E_FWREADY:
		bfa_sm_set_state(ioc, bfa_hal_ioc_sm_cfginit);
		bfa_hal_hw_sem_get_cancel(ioc);
		bfa_hal_ioc_timer_start(ioc);
		bfa_hal_ioc_send_cfg(ioc);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * Hardware is being initialized. Interrupts are enabled.
 * Holding hardware semaphore lock.
 */
static void
bfa_hal_ioc_sm_hwinit(bfa_hal_ioc_t *ioc, ioc_event_t event)
{
	bfa_trc(ioc->hal, event);

	switch (event) {
	case IOC_E_FWREADY:
	case IOC_E_FWRSP_ENABLE:
		bfa_sm_set_state(ioc, bfa_hal_ioc_sm_cfginit);
		bfa_hal_ioc_timer_stop(ioc);
		bfa_hal_hw_sem_release(ioc);

		bfa_hal_ioc_timer_start(ioc);
		bfa_hal_ioc_send_cfg(ioc);
		break;

	case IOC_E_TIMEOUT:
		ioc->hwinit_count++;
		if (ioc->hwinit_count < BFA_HAL_IOC_HWINIT_MAX) {
			bfa_hal_ioc_timer_start(ioc);
			bfa_hal_ioc_reset(ioc, BFA_TRUE);
			break;
		}

		bfa_hal_hw_sem_release(ioc);
		bfa_sm_set_state(ioc, bfa_hal_ioc_sm_hwfail);
		bfa_hal_ioc_callback(ioc);
		break;

	case IOC_E_DISABLE:
		bfa_sm_set_state(ioc, bfa_hal_ioc_sm_disabled);
		bfa_hal_hw_sem_release(ioc);
		bfa_hal_ioc_timer_stop(ioc);
		bfa_hal_ioc_callback(ioc);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * IOC configuration in progress. Timer is active.
 */
static void
bfa_hal_ioc_sm_cfginit(bfa_hal_ioc_t *ioc, ioc_event_t event)
{
	bfa_trc(ioc->hal, event);

	switch (event) {
	case IOC_E_FWRSP_CFG:
		bfa_hal_ioc_timer_stop(ioc);
		if (ioc->ioc_started) {
			bfa_sm_set_state(ioc, bfa_hal_ioc_sm_op);
			bfa_hal_ioc_initdone_submod(ioc->hal);
			bfa_hal_ioc_start_submod(ioc->hal);
			bfa_hal_ioc_hb_monitor(ioc);
		} else {
			bfa_sm_set_state(ioc, bfa_hal_ioc_sm_cfgdone);
			bfa_hal_ioc_initdone_submod(ioc->hal);
			bfa_hal_ioc_callback(ioc);
		}
		break;

	case IOC_E_TIMEOUT:
		bfa_sm_set_state(ioc, bfa_hal_ioc_sm_hwfail);
		bfa_hal_ioc_callback(ioc);
		break;

	case IOC_E_DISABLE:
		bfa_sm_set_state(ioc, bfa_hal_ioc_sm_disabled);
		bfa_hal_ioc_timer_stop(ioc);
		bfa_hal_ioc_callback(ioc);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * Send AEN notification
 */
static void
bfa_hal_ioc_aen_post(bfa_hal_t *hal, bfa_ioc_aen_event_t event)
{
	bfa_aen_data_t  aen_data;
	bfa_log_mod_t	*logmod = hal->logm;
	int32_t		inst_num = bfa_aen_get_inst(hal->aen);

	switch (event) {
	case BFA_IOC_AEN_HBGOOD:
		bfa_log(logmod, BFAL_AEN_IOC_HBGOOD, inst_num);
		break;
	case BFA_IOC_AEN_HBFAIL:
		bfa_log(logmod, BFAL_AEN_IOC_HBFAIL, inst_num);
		break;
	case BFA_IOC_AEN_ENABLE:
		bfa_log(logmod, BFAL_AEN_IOC_ENABLE, inst_num);
		break;
	case BFA_IOC_AEN_DISABLE:
		bfa_log(logmod, BFAL_AEN_IOC_DISABLE, inst_num);
		break;
	default: 
		break;
	}

	aen_data.ioc.pwwn = bfa_hal_ioc_get_pwwn(hal);
	bfa_aen_post(hal->aen, BFA_AEN_CAT_IOC, event, &aen_data);
}

static void
bfa_hal_ioc_sm_cfgdone(bfa_hal_ioc_t *ioc, ioc_event_t event)
{
	bfa_trc(ioc->hal, event);

	switch (event) {
	case IOC_E_ENABLE:
		bfa_sm_set_state(ioc, bfa_hal_ioc_sm_op);
		bfa_hal_ioc_aen_post(ioc->hal, BFA_IOC_AEN_ENABLE);
		bfa_hal_ioc_start_submod(ioc->hal);
		bfa_hal_ioc_hb_monitor(ioc);
		break;

	case IOC_E_DISABLE:
		bfa_sm_set_state(ioc, bfa_hal_ioc_sm_disabling);
		bfa_hal_ioc_timer_start(ioc);
		bfa_hal_ioc_send_disable(ioc);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_hal_ioc_sm_op(bfa_hal_ioc_t *ioc, ioc_event_t event)
{
	bfa_trc(ioc->hal, event);

	switch (event) {
	case IOC_E_DISABLE:
		bfa_sm_set_state(ioc, bfa_hal_ioc_sm_disabling);
		bfa_hal_ioc_aen_post(ioc->hal, BFA_IOC_AEN_DISABLE);
		bfa_hal_ioc_hb_stop(ioc);
		bfa_hal_ioc_timer_start(ioc);
		bfa_hal_ioc_send_disable(ioc);
		break;

	case IOC_E_HBFAIL:
		if (ioc->auto_recover) {
			bfa_sm_set_state(ioc, bfa_hal_ioc_sm_recover);
			bfa_hal_ioc_timer_start(ioc);
		} else {
			bfa_sm_set_state(ioc, bfa_hal_ioc_sm_hbfail);
			bfa_hal_reg_write(ioc->bfa_regs.ioc_fwstate,
					  BFA_FWI_IOC_HBFAIL);
		}
		bfa_hal_ioc_iocdisable_submod(ioc->hal);
		bfa_hal_ioc_aen_post(ioc->hal, BFA_IOC_AEN_HBFAIL);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_hal_ioc_sm_disabling(bfa_hal_ioc_t *ioc, ioc_event_t event)
{
	bfa_trc(ioc->hal, event);

	switch (event) {
	case IOC_E_FWRSP_DISABLE:
		bfa_hal_ioc_timer_stop(ioc);
		/*
		 * !!! fall through !!!
		 */

	case IOC_E_TIMEOUT:
		bfa_sm_set_state(ioc, bfa_hal_ioc_sm_disabled);
		bfa_hal_intr_disable(ioc->hal);
		bfa_hal_ioc_iocdisable_submod(ioc->hal);
		bfa_hal_ioc_callback(ioc);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_hal_ioc_sm_disabled(bfa_hal_ioc_t *ioc, ioc_event_t event)
{
	bfa_trc(ioc->hal, event);

	switch (event) {
	case IOC_E_ENABLE:
		bfa_sm_set_state(ioc, bfa_hal_ioc_sm_semwait);
		bfa_hal_ioc_aen_post(ioc->hal, BFA_IOC_AEN_ENABLE);
		bfa_hal_hw_sem_get(ioc);
		break;

	case IOC_E_DISABLE:
		bfa_hal_ioc_callback(ioc);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * Hardware initialization failed.
 */
static void
bfa_hal_ioc_sm_hwfail(bfa_hal_ioc_t *ioc, ioc_event_t event)
{
	bfa_trc(ioc->hal, event);

	switch (event) {
	case IOC_E_DISABLE:
		bfa_sm_set_state(ioc, bfa_hal_ioc_sm_disabled);
		bfa_hal_ioc_callback(ioc);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * IOC heartbeat failure.
 */
static void
bfa_hal_ioc_sm_hbfail(bfa_hal_ioc_t *ioc, ioc_event_t event)
{
	bfa_trc(ioc->hal, event);

	switch (event) {
	case IOC_E_DISABLE:
		bfa_sm_set_state(ioc, bfa_hal_ioc_sm_disabled);
		bfa_hal_ioc_callback(ioc);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * IOC heartbeat failure.
 */
static void
bfa_hal_ioc_sm_recover(bfa_hal_ioc_t *ioc, ioc_event_t event)
{
	bfa_trc(ioc->hal, event);

	switch (event) {
	case IOC_E_DISABLE:
		bfa_sm_set_state(ioc, bfa_hal_ioc_sm_disabled);
		bfa_hal_ioc_timer_stop(ioc);
		bfa_hal_ioc_callback(ioc);
		break;

	case IOC_E_TIMEOUT:
		bfa_sm_set_state(ioc, bfa_hal_ioc_sm_semwait);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @}
 */

/**
 * @dg hal_ioc_pvt HAL IOC private functions
 * @{
 */

static void
bfa_hal_cqs_sz(bfa_ioc_cfg_t *cfg, uint32_t *dm_len)
{
	int             i, per_reqq_sz, per_rspq_sz;

	per_reqq_sz = BFA_ROUNDUP((cfg->num_reqq_elems * BFA_LMSG_SZ),
							BFA_DMA_ALIGN_SZ);
	per_rspq_sz = BFA_ROUNDUP((cfg->num_rspq_elems * BFA_LMSG_SZ),
							BFA_DMA_ALIGN_SZ);

	/*
	 * Calculate CQ size
	 */
	for (i = 0; i < cfg->fwcfg.num_cqs; i++) {
		*dm_len = *dm_len + per_reqq_sz;
		*dm_len = *dm_len + per_rspq_sz;
	}

	/*
	 * Calculate Shadow CI/PI size
	 */
	for (i = 0; i < cfg->fwcfg.num_cqs; i++)
		*dm_len += (2 * BFA_HAL_CACHELINE_SZ);
}

static void
bfa_hal_fw_cfg_sz(bfa_ioc_cfg_t *cfg, uint32_t *dm_len)
{
	*dm_len +=
		BFA_ROUNDUP(sizeof(bfa_ioc_cfg_info_t), BFA_HAL_CACHELINE_SZ);
	*dm_len +=
		BFA_ROUNDUP(sizeof(bfa_ioc_cfgrsp_info_t),
			    BFA_HAL_CACHELINE_SZ);
	*dm_len += BFA_ROUNDUP(sizeof(bfa_ioc_stats_t), BFA_HAL_CACHELINE_SZ);
}

static void
bfa_hal_ioc_sem_timeout(void *ioc_arg)
{
	bfa_hal_ioc_t  *ioc = (bfa_hal_ioc_t *) ioc_arg;

	bfa_hal_hw_sem_get(ioc);
}

static void
bfa_hal_hw_sem_get(bfa_hal_ioc_t *ioc)
{
	uint32_t        r32;

    /**
     * First read to the semaphore register will return 0, subsequent reads
     * will return 1. Semaphore is released by writing 0 to the register
     */
	r32 = bfa_hal_reg_read(ioc->bfa_regs.host_sem0_reg);
	if (r32 == 0) {
		bfa_sm_send_event(ioc, IOC_E_SEMLOCKED);
		return;
	}

	bfa_hal_timer_start(ioc->hal, &ioc->sem_timer, bfa_hal_ioc_sem_timeout,
			    ioc, BFA_HAL_IOC_TOV);
}

static void
bfa_hal_hw_sem_release(bfa_hal_ioc_t *ioc)
{
	bfa_hal_reg_write(ioc->bfa_regs.host_sem0_reg, 1);
}

static void
bfa_hal_hw_sem_get_cancel(bfa_hal_ioc_t *ioc)
{
	bfa_hal_timer_stop(&ioc->sem_timer);
}

/**
 * @brief
 * Initialize LPU local memory (aka secondary memory / SRAM)
 */
static void
bfa_hal_ioc_lmem_init(bfa_hal_ioc_t *ioc)
{
	uint32_t        pss_ctl;
	int             i;
#define PSS_LMEM_INIT_TIME  10000

	pss_ctl = bfa_hal_reg_read(ioc->bfa_regs.pss_ctl_reg);
	pss_ctl &= ~__PSS_LMEM_RESET;
	pss_ctl |= __PSS_LMEM_INIT_EN;
	pss_ctl |= __PSS_I2C_CLK_DIV(3UL);	/* i2c workaround 12.5khz clock
						 */
	bfa_hal_reg_write(ioc->bfa_regs.pss_ctl_reg, pss_ctl);

    /**
     * wait for memory initialization to be complete
     */
	i = 0;
	do {
		pss_ctl = bfa_hal_reg_read(ioc->bfa_regs.pss_ctl_reg);
		i++;
	} while (!(pss_ctl & __PSS_LMEM_INIT_DONE) && (i < PSS_LMEM_INIT_TIME));

    /**
     * If memory initialization is not successful, IOC timeout will catch
	 * such failures.
     */
	bfa_assert(pss_ctl & __PSS_LMEM_INIT_DONE);
	bfa_trc(ioc->hal, pss_ctl);

	pss_ctl &= ~(__PSS_LMEM_INIT_DONE | __PSS_LMEM_INIT_EN);
	bfa_hal_reg_write(ioc->bfa_regs.pss_ctl_reg, pss_ctl);
}

static void
bfa_hal_lpu_start(bfa_hal_ioc_t *ioc)
{
	uint32_t        pss_ctl;

    /**
     * Take processor out of reset.
     */
	pss_ctl = bfa_hal_reg_read(ioc->bfa_regs.pss_ctl_reg);
	pss_ctl &= ~__PSS_LPU0_RESET;

	bfa_hal_reg_write(ioc->bfa_regs.pss_ctl_reg, pss_ctl);
}

/**
 * @brief
 * Return true if current running version is same as driver version.
 */
static bfa_boolean_t
bfa_hal_ioc_fwver_check(bfa_hal_ioc_t *ioc)
{
	bfa_hal_t	*hal = ioc->hal;
	uint32_t        pgnum, pgoff;
	uint32_t        loff = 0;
	int             i;
	bfa_fwi_ioc_image_hdr_t fwhdr, *drv_fwhdr;
	uint32_t	*fwsig = (uint32_t *) &fwhdr;

	pgnum = PSS_SMEM_PGNUM(loff);
	pgoff = PSS_SMEM_PGOFF(loff);
	bfa_hal_reg_write(ioc->bfa_regs.host_page_num_fn, pgnum);

	for (i = 0; i < (sizeof(bfa_fwi_ioc_image_hdr_t) / sizeof(uint32_t));
	     i++) {
		fwsig[i] =
			bfa_hal_mem_read(ioc->bfa_regs.smem_page_start, loff);
		loff += sizeof(uint32_t);
	}

	bfa_trc(hal, fwhdr.signature);
	bfa_trc(hal, fwhdr.cksum);
	bfa_trc(hal, fwhdr.exec);
	bfa_trc(hal, fwhdr.param);

	drv_fwhdr = (bfa_fwi_ioc_image_hdr_t *) bfa_hal_fwimg;

	bfa_trc(hal, drv_fwhdr->signature);
	bfa_trc(hal, drv_fwhdr->cksum);
	bfa_trc(hal, drv_fwhdr->exec);
	bfa_trc(hal, drv_fwhdr->param);

	return (((fwhdr.signature == drv_fwhdr->signature))
		&& (fwhdr.cksum == drv_fwhdr->cksum)
		&& (fwhdr.exec == drv_fwhdr->exec)
		);
}

/**
 * @img ioc_init_logic.jpg
 */
static void
bfa_hal_ioc_hwinit(bfa_hal_ioc_t *ioc, bfa_boolean_t force)
{
	bfa_fwi_ioc_state_t ioc_fwstate;

	ioc_fwstate = bfa_hal_reg_read(ioc->bfa_regs.ioc_fwstate);
	if (force)
		ioc_fwstate = BFA_FWI_IOC_UNINIT;

	bfa_trc(ioc->hal, ioc_fwstate);

	/**
	 * If hardware initialization is in progress (initialized by other IOC),
	 * just wait for an initialization completion interrupt.
	 */
	if ((ioc_fwstate == BFA_FWI_IOC_HWINIT)
	    || (ioc_fwstate == BFA_FWI_IOC_INITING)) {
		bfa_trc(ioc->hal, ioc_fwstate);
		return;
	}

	/**
	 * If IOC function is disabled and firmware version is same,
	 * just re-enable IOC.
	 */
	if ((ioc_fwstate == BFA_FWI_IOC_DISABLED)
	    && bfa_hal_ioc_fwver_check(ioc)) {
		bfa_trc(ioc->hal, ioc_fwstate);
		bfa_hal_ioc_send_enable(ioc);
		return;
	}

	/**
	 * Initialize the h/w for any other states.
	 */
	bfa_hal_ioc_boot(ioc->hal, BFA_FWI_BOOT_TYPE_NORMAL,
			 BFA_FWI_BOOT_PARAM_VAL_DEFAULT);
}

static void
bfa_hal_ioc_timeout(void *ioc_arg)
{
	bfa_hal_ioc_t  *ioc = (bfa_hal_ioc_t *) ioc_arg;

	bfa_trc(ioc->hal, 0);
	bfa_sm_send_event(ioc, IOC_E_TIMEOUT);
}

static void
bfa_hal_ioc_mbox_send(bfa_hal_ioc_t *ioc, void *ioc_msg, int len)
{
	uint32_t       *msgp = (uint32_t *) ioc_msg;
	uint32_t        i;

	bfa_trc(ioc->hal, len);

	bfa_assert(len <= BFA_MBOX_REG_LEN);

	/*
	 * first write msg to mailbox registers
	 */
	for (i = 0; i < len / sizeof(uint32_t); i++)
		bfa_hal_reg_write(ioc->bfa_regs.hfn_mbox + i * sizeof(uint32_t),
							bfa_os_wtole(msgp[i]));

	/*
	 * write 1 to mailbox CMD to trigger LPU event
	 */
	bfa_hal_reg_write(ioc->bfa_regs.hfn_mbox_cmd, 1);
}

/**
 * Use the Mailbox interface to send BFA_IOC_H2I_CFG_REQ
 */
static void
bfa_hal_ioc_send_cfg(bfa_hal_ioc_t *ioc)
{
	int             i;
	bfa_ioc_cfg_req_t cfg_req;
	bfa_ioc_cfg_info_t *cfg_info = ioc->cfginfo;
	bfa_ioc_cfg_t  *hal_cfg = &ioc->cfg;

	bfa_assert(hal_cfg->fwcfg.num_cqs <= BFA_IOC_MAX_CQS);
	bfa_trc(ioc->hal, hal_cfg->fwcfg.num_cqs);

	/**
	* initialize IOC configuration info
	*/
	cfg_info->endian_sig = BFA_FWI_IOC_ENDIAN_SIG;
	cfg_info->num_cqs = hal_cfg->fwcfg.num_cqs;

	bfa_hal_dmabe_addr_set(cfg_info->cfgrsp_addr, ioc->cfgrsp_dma.pa);
	bfa_hal_dmabe_addr_set(cfg_info->stats_addr, ioc->stats_pa);

	/**
	* dma map REQ and RSP circular queues and shadow pointers
	*/
	for (i = 0; i < hal_cfg->fwcfg.num_cqs; i++) {
		bfa_hal_dmabe_addr_set(cfg_info->req_cq_ba[i],
				       ioc->req_cq_ba[i].pa);
		bfa_hal_dmabe_addr_set(cfg_info->req_shadow_ci[i],
				       ioc->req_cq_shadow_ci[i].pa);
		cfg_info->req_cq_elems[i] =
			bfa_os_htons(hal_cfg->num_reqq_elems);

		bfa_hal_dmabe_addr_set(cfg_info->rsp_cq_ba[i],
				       ioc->rsp_cq_ba[i].pa);
		bfa_hal_dmabe_addr_set(cfg_info->rsp_shadow_pi[i],
				       ioc->rsp_cq_shadow_pi[i].pa);
		cfg_info->rsp_cq_elems[i] =
			bfa_os_htons(hal_cfg->num_rspq_elems);
	}

	/**
	* dma map IOC configuration itself
	*/
	bfa_fwi_h2i_set(cfg_req.mh, BFA_MC_IOC, BFA_IOC_H2I_CFG_REQ,
			bfa_hal_ioc_lpuid(ioc->hal));
	bfa_hal_dmabe_addr_set(cfg_req.ioc_cfg_dma_addr, ioc->cfg_info.pa);

	bfa_hal_ioc_mbox_send(ioc, &cfg_req, sizeof(bfa_ioc_cfg_req_t));
}

static void
bfa_hal_ioc_send_enable(bfa_hal_ioc_t *ioc)
{
	bfa_ioc_enable_req_t enable_req;

	bfa_fwi_h2i_set(enable_req.mh, BFA_MC_IOC, BFA_IOC_H2I_ENABLE_REQ,
			bfa_hal_ioc_lpuid(ioc->hal));
	bfa_hal_ioc_mbox_send(ioc, &enable_req, sizeof(bfa_ioc_enable_req_t));
}

static void
bfa_hal_ioc_send_disable(bfa_hal_ioc_t *ioc)
{
	bfa_ioc_enable_req_t disable_req;

	bfa_fwi_h2i_set(disable_req.mh, BFA_MC_IOC, BFA_IOC_H2I_DISABLE_REQ,
			bfa_hal_ioc_lpuid(ioc->hal));
	bfa_hal_ioc_mbox_send(ioc, &disable_req, sizeof(bfa_ioc_disable_req_t));
}

/**
 * Firmware failure detected. Start recovery actions.
 */
static void
bfa_hal_ioc_recover(bfa_hal_ioc_t *ioc)
{
	bfa_sm_send_event(ioc, IOC_E_HBFAIL);
}

static void
bfa_hal_ioc_hb_check(void *cbarg)
{
	bfa_hal_ioc_t  *ioc = cbarg;
	uint32_t        hb_count;

	hb_count = bfa_hal_reg_read(ioc->bfa_regs.heartbeat);
	if (ioc->hb_count == hb_count) {
		ioc->hb_fail++;
	} else {
		ioc->hb_count = hb_count;
		ioc->hb_fail = 0;
	}

	if (ioc->hb_fail >= BFA_HAL_IOC_HB_FAIL_MAX) {
		bfa_log(ioc->hal->logm, BFA_LOG_HAL_HEARTBEAT_FAILURE,
			hb_count);
		ioc->hb_fail = 0;
		bfa_hal_ioc_recover(ioc);
		return;
	}

	bfa_hal_timer_start(ioc->hal, &ioc->ioc_timer, bfa_hal_ioc_hb_check,
			    ioc, BFA_HAL_IOC_HB_TOV);
}

static void
bfa_hal_ioc_hb_monitor(bfa_hal_ioc_t *ioc)
{
	ioc->hb_fail = 0;
	ioc->hb_count = bfa_hal_reg_read(ioc->bfa_regs.heartbeat);
	bfa_hal_timer_start(ioc->hal, &ioc->ioc_timer, bfa_hal_ioc_hb_check,
			    ioc, BFA_HAL_IOC_HB_TOV);
}

static void
bfa_hal_ioc_hb_stop(bfa_hal_ioc_t *ioc)
{
	bfa_hal_timer_stop(&ioc->ioc_timer);
}

static void
bfa_hal_reg_init(bfa_hal_ioc_t *ioc)
{
	bfa_os_addr_t   rb;
	int             i;
	int             fn = ioc->pci_info.pci_func;

	rb = ioc->pci_info.pci_bar_kva;

	if (fn == 0) {
		ioc->bfa_regs.intr_status = (rb + HOSTFN0_INT_STATUS);
		ioc->bfa_regs.intr_mask = (rb + HOSTFN0_INT_MSK);
		ioc->bfa_regs.hfn_mbox_cmd = (rb + HOSTFN0_LPU0_CMD_STAT);
		ioc->bfa_regs.hfn_mbox = (rb + HOSTFN0_LPU0_MBOX_0);
		ioc->bfa_regs.lpu_mbox_cmd = (rb + LPU0_HOSTFN0_CMD_STAT);
		ioc->bfa_regs.lpu_mbox = (rb + LPU0_HOSTFN0_MBOX_0);
		ioc->bfa_regs.lpu_ioc_status = (rb + LPU0_HOSTFN0_CMD_STAT);
		ioc->bfa_regs.host_page_num_fn = (rb + HOST_PAGE_NUM_FN0);
		ioc->bfa_regs.heartbeat = (rb + BFA_IOC0_HBEAT_REG);
		ioc->bfa_regs.ioc_fwstate = (rb + BFA_IOC0_STATE_REG);
	} else {
		ioc->bfa_regs.intr_status = (rb + HOSTFN1_INT_STATUS);
		ioc->bfa_regs.intr_mask = (rb + HOSTFN1_INT_MSK);
		ioc->bfa_regs.hfn_mbox_cmd = (rb + HOSTFN1_LPU1_CMD_STAT);
		ioc->bfa_regs.hfn_mbox = (rb + HOSTFN1_LPU1_MBOX_0);
		ioc->bfa_regs.lpu_mbox_cmd = (rb + LPU1_HOSTFN1_CMD_STAT);
		ioc->bfa_regs.lpu_mbox = (rb + LPU1_HOSTFN1_MBOX_0);
		ioc->bfa_regs.lpu_ioc_status = (rb + LPU1_HOSTFN1_CMD_STAT);
		ioc->bfa_regs.host_page_num_fn = (rb + HOST_PAGE_NUM_FN1);
		ioc->bfa_regs.heartbeat = (rb + BFA_IOC1_HBEAT_REG);
		ioc->bfa_regs.ioc_fwstate = (rb + BFA_IOC1_STATE_REG);
	}

	for (i = 0; i < BFA_IOC_MAX_CQS; i++) {
		int             q;

		/*
		 * CPE registers
		 */
		q = CPE_Q_NUM(fn, i);
		ioc->bfa_regs.cpe_q_pi[i] = (rb + CPE_Q_PI(q));
		ioc->bfa_regs.cpe_q_ci[i] = (rb + CPE_Q_CI(q));
		ioc->bfa_regs.cpe_q_depth[i] = (rb + CPE_Q_DEPTH(q));

		/*
		 * RME registers
		 */
		q = CPE_Q_NUM(fn, i);
		ioc->bfa_regs.rme_q_pi[i] = (rb + RME_Q_PI(q));
		ioc->bfa_regs.rme_q_ci[i] = (rb + RME_Q_CI(q));
		ioc->bfa_regs.rme_q_depth[i] = (rb + RME_Q_DEPTH(q));
	}

	/*
	 * PSS control registers
	 */
	ioc->bfa_regs.pss_ctl_reg = (rb + PSS_CTL_REG);
	ioc->bfa_regs.app_pll_400_ctl_reg = (rb + APP_PLL_400_CTL_REG);
	ioc->bfa_regs.app_pll_212_ctl_reg = (rb + APP_PLL_212_CTL_REG);

	/*
	 * host semaphore registers
	 */
	ioc->bfa_regs.host_sem0_reg = (rb + HOST_SEM0_REG);
	ioc->bfa_regs.host_sem1_reg = (rb + HOST_SEM1_REG);
	ioc->bfa_regs.host_sem2_reg = (rb + HOST_SEM2_REG);
	ioc->bfa_regs.host_sem3_reg = (rb + HOST_SEM3_REG);

	/**
	* sram memory access
	*/
	ioc->bfa_regs.smem_page_start = (rb + PSS_SMEM_PAGE_START);
}

static void
bfa_hal_init_struct(bfa_hal_t *hal, void *bfad, bfa_ioc_cfg_t *cfg,
		    bfa_hal_pcidev_t *pcidev)
{
	bfa_hal_ioc_t  *ioc = &hal->ioc;

	/*
	 * Init ioc structure
	 */
	bfa_os_memset((void *)ioc, 0, sizeof(bfa_hal_ioc_t));

	hal->bfad = bfad;
	ioc->hal = hal;

	ioc->cfg = *cfg;
	ioc->pci_info = *pcidev;

	bfa_hal_reg_init(ioc);
}

static void
bfa_hal_ioc_mem_claim(bfa_hal_t *hal, bfa_ioc_cfg_t *cfg,
		      bfa_hal_meminfo_t *meminfo)
{
	uint8_t        *dm_kva;
	uint64_t        dm_pa;
	int             i, per_reqq_sz, per_rspq_sz;
	bfa_hal_ioc_t  *ioc = &hal->ioc;

	dm_kva = bfa_hal_meminfo_dma_virt(meminfo);
	dm_pa = bfa_hal_meminfo_dma_phys(meminfo);

	/*
	 * Claim DMA-able memory for the request/response queues and for shadow
	 * ci/pi registers
	 */
	per_reqq_sz = BFA_ROUNDUP((cfg->num_reqq_elems * BFA_LMSG_SZ),
							BFA_DMA_ALIGN_SZ);
	per_rspq_sz = BFA_ROUNDUP((cfg->num_rspq_elems * BFA_LMSG_SZ),
							BFA_DMA_ALIGN_SZ);

	for (i = 0; i < cfg->fwcfg.num_cqs; i++) {
		ioc->req_cq_ba[i].kva = dm_kva;
		ioc->req_cq_ba[i].pa = dm_pa;
		dm_kva += per_reqq_sz;
		dm_pa += per_reqq_sz;

		ioc->rsp_cq_ba[i].kva = dm_kva;
		ioc->rsp_cq_ba[i].pa = dm_pa;
		dm_kva += per_rspq_sz;
		dm_pa += per_rspq_sz;
	}

	for (i = 0; i < cfg->fwcfg.num_cqs; i++) {
		ioc->req_cq_shadow_ci[i].kva = dm_kva;
		ioc->req_cq_shadow_ci[i].pa = dm_pa;
		dm_kva += BFA_HAL_CACHELINE_SZ;
		dm_pa += BFA_HAL_CACHELINE_SZ;

		ioc->rsp_cq_shadow_pi[i].kva = dm_kva;
		ioc->rsp_cq_shadow_pi[i].pa = dm_pa;
		dm_kva += BFA_HAL_CACHELINE_SZ;
		dm_pa += BFA_HAL_CACHELINE_SZ;
	}

	/*
	 * Claim DMA-able memory for the config info page
	 */
	hal->ioc.cfg_info.kva = dm_kva;
	hal->ioc.cfg_info.pa = dm_pa;
	hal->ioc.cfginfo = (bfa_ioc_cfg_info_t *) dm_kva;
	dm_kva += BFA_ROUNDUP(sizeof(bfa_ioc_cfg_info_t), BFA_HAL_CACHELINE_SZ);
	dm_pa += BFA_ROUNDUP(sizeof(bfa_ioc_cfg_info_t), BFA_HAL_CACHELINE_SZ);

	/*
	 * Claim DMA-able memory for the config response
	 */
	hal->ioc.cfgrsp_dma.kva = dm_kva;
	hal->ioc.cfgrsp_dma.pa = dm_pa;
	hal->ioc.cfgrsp = (bfa_ioc_cfgrsp_info_t *) dm_kva;

	dm_kva +=
		BFA_ROUNDUP(sizeof(bfa_ioc_cfgrsp_info_t),
			    BFA_HAL_CACHELINE_SZ);
	dm_pa += BFA_ROUNDUP(sizeof(bfa_ioc_cfgrsp_info_t),
			     BFA_HAL_CACHELINE_SZ);

	/*
	 * Claim DMA-able memory for ioc stats
	 */
	hal->ioc.stats_kva = dm_kva;
	hal->ioc.stats_pa = dm_pa;
	hal->ioc.stats = (bfa_ioc_stats_t *) dm_kva;
	dm_kva += BFA_ROUNDUP(sizeof(bfa_ioc_stats_t), BFA_HAL_CACHELINE_SZ);
	dm_pa += BFA_ROUNDUP(sizeof(bfa_ioc_stats_t), BFA_HAL_CACHELINE_SZ);

	bfa_hal_meminfo_dma_virt(meminfo) = dm_kva;
	bfa_hal_meminfo_dma_phys(meminfo) = dm_pa;
}

/**
 * @brief
 * HAL submodules initialization completion notification.
 */
static void
bfa_hal_ioc_initdone_submod(bfa_hal_t *hal)
{
	int             i;

	for (i = 0; hal_mods[i]; i++)
		hal_mods[i]->initdone(hal);
}

/**
 * @brief
 * Start HAL submodules.
 */
static void
bfa_hal_ioc_start_submod(bfa_hal_t *hal)
{
	int             i;

	for (i = 0; hal_mods[i]; i++)
		hal_mods[i]->start(hal);
}

/**
 * @brief
 * Notify sub-modules of hardware failure.
 */
static void
bfa_hal_ioc_iocdisable_submod(bfa_hal_t *hal)
{
	int             i;

	for (i = 0; hal_mods[i]; i++)
		hal_mods[i]->iocdisable(hal);
}

static void
__bfa_hcb_init(void *cbarg, bfa_boolean_t complete)
{
	bfa_hal_ioc_t  *ioc = cbarg;

	if (complete) {
		if (bfa_sm_cmp_state(ioc, bfa_hal_ioc_sm_cfgdone))
			bfa_hcb_init(ioc->hal->bfad, BFA_STATUS_OK);
		else
			bfa_hcb_init(ioc->hal->bfad, BFA_STATUS_FAILED);
	}
}

static void
__bfa_hcb_stop(void *cbarg, bfa_boolean_t complete)
{
	bfa_hal_ioc_t  *ioc = cbarg;

	if (complete)
		bfa_hcb_stop(ioc->hal->bfad, BFA_STATUS_OK);
}

static void
__bfa_hcb_ioc_disable(void *cbarg, bfa_boolean_t complete)
{
	bfa_hal_ioc_t  *ioc = cbarg;

	if (complete)
		bfa_hcb_ioc_disable(ioc->hal->bfad);
}

static void
bfa_hal_ioc_callback(bfa_hal_ioc_t *ioc)
{
	if (ioc->init_cb) {
		ioc->init_cb = BFA_FALSE;
		bfa_hcb_queue(ioc->hal, &ioc->init_hcb_qe, __bfa_hcb_init, ioc);
	}

	if (ioc->disable_cb) {
		ioc->disable_cb = BFA_FALSE;
		bfa_hcb_queue(ioc->hal, &ioc->dis_hcb_qe, __bfa_hcb_ioc_disable,
			      ioc);
	}

	if (ioc->stop_cb) {
		ioc->stop_cb = BFA_FALSE;
		bfa_hcb_queue(ioc->hal, &ioc->stop_hcb_qe, __bfa_hcb_stop, ioc);
	}
}

#define BFA_MFG_NAME "Brocade"
static void
bfa_hal_ioc_get_adapter_attr(bfa_hal_t *hal, bfa_adapter_attr_t *ad_attr)
{
	bfa_ioc_cfgrsp_info_t *cfginfo;
	char            model[BFA_ADAPTER_MODEL_NAME_LEN];

	cfginfo = hal->ioc.cfgrsp;
	bfa_os_memcpy((void *)&ad_attr->serial_num,
		      (void *)cfginfo->brcd_serialnum,
		      BFA_ADAPTER_SERIAL_NUM_LEN);

	bfa_os_memcpy(&ad_attr->fw_ver, cfginfo->fw_version, BFA_VERSION_LEN);
	bfa_os_memcpy(&ad_attr->manufacturer, BFA_MFG_NAME,
		      BFA_ADAPTER_MFG_NAME_LEN);
	ad_attr->nports = BFA_FWI_ADAPTER_GETP(NPORTS, cfginfo->adapter_prop);
	ad_attr->max_speed = BFA_FWI_ADAPTER_GETP(SPEED, cfginfo->adapter_prop);

	/**
	 * model name
	 */
	strcpy(model, "Brocade-??5");
	model[8] = '0' + BFA_FWI_ADAPTER_GETP(SPEED, cfginfo->adapter_prop);
	model[9] = '0' + ad_attr->nports;

	if (BFA_FWI_ADAPTER_IS_SPECIAL(cfginfo->adapter_prop))
		ad_attr->prototype = 1;
	else
		ad_attr->prototype = 0;

	bfa_os_memcpy(&ad_attr->model, model, BFA_ADAPTER_MODEL_NAME_LEN);
	bfa_os_memcpy(&ad_attr->model_descr, &ad_attr->model,
		      BFA_ADAPTER_MODEL_NAME_LEN);

	ad_attr->pwwn = bfa_hal_ioc_get_pwwn(hal);

	ad_attr->pcie_gen = cfginfo->pcie_gen;
	ad_attr->pcie_lanes = cfginfo->pcie_lanes;
	ad_attr->pcie_lanes_orig = cfginfo->pcie_lanes_orig;
	ad_attr->asic_rev = cfginfo->asic_rev;

	/**
	* @todo
	 * ad_attr->node_symname
	 * ad_attr->hw_ver
	 * ad_attr->optrom_ver
	 */
}

/**
 * @brief
 * Update HAL configuration from firmware configuration.
 */
static void
bfa_hal_ioc_cfg_update(bfa_hal_ioc_t *ioc)
{
	bfa_ioc_cfgrsp_info_t *cfgrsp  = ioc->cfgrsp;
	bfa_ioc_fwcfg_t       *fwcfg   = &cfgrsp->fwcfg;
	bfa_ioc_cfg_info_t    *cfginfo = ioc->cfginfo;

	cfgrsp->adapter_prop  = bfa_os_ntohl(cfgrsp->adapter_prop);
	cfgrsp->maxfrsize     = bfa_os_ntohs(cfgrsp->maxfrsize);

	fwcfg->num_cqs        = fwcfg->num_cqs;
	fwcfg->num_itnims     = bfa_os_ntohs(fwcfg->num_itnims);
	fwcfg->num_ioim_reqs  = bfa_os_ntohs(fwcfg->num_ioim_reqs);
	fwcfg->num_tskim_reqs = bfa_os_ntohs(fwcfg->num_tskim_reqs);
	fwcfg->num_fcxp_reqs  = bfa_os_ntohs(fwcfg->num_fcxp_reqs);
	fwcfg->num_uf_bufs    = bfa_os_ntohs(fwcfg->num_uf_bufs);
	fwcfg->num_rports     = bfa_os_ntohs(fwcfg->num_rports);
	fwcfg->num_itntms     = bfa_os_ntohs(fwcfg->num_itntms);

	cfginfo->intr_attr.coalesce = cfgrsp->intr_attr.coalesce;
	cfginfo->intr_attr.delay    = bfa_os_ntohs(cfgrsp->intr_attr.delay);
	cfginfo->intr_attr.latency  = bfa_os_ntohs(cfgrsp->intr_attr.latency);
}

static void
bfa_hal_ioc_stats_clear(void *cbarg)
{
	bfa_hal_ioc_t  *ioc = (bfa_hal_ioc_t *) cbarg;
	bfa_ioc_stats_clr_req_t stats_req;

	bfa_hal_timer_start(ioc->hal, &ioc->stats_timer,
			    bfa_hal_ioc_stats_clr_timeout, ioc,
			    BFA_HAL_IOC_TOV);

	bfa_fwi_h2i_set(stats_req.mh, BFA_MC_IOC, BFA_IOC_H2I_CLEAR_STATS_REQ,
			bfa_hal_ioc_lpuid(ioc->hal));
	bfa_hal_ioc_mbox_send(ioc, &stats_req, sizeof(bfa_ioc_stats_clr_req_t));

	return;
}

static void
bfa_hal_ioc_stats_swap(bfa_ioc_stats_t *d, bfa_ioc_stats_t *s)
{
	uint32_t       *dip = (uint32_t *) d;
	uint32_t       *sip = (uint32_t *) s;
	int             i;

	for (i = 0; i < (sizeof(bfa_ioc_stats_t) / sizeof(uint32_t)); i++)
		dip[i] = bfa_os_ntohl(sip[i]);
}

static void
__bfa_hcb_ioc_stats_clr(void *cbarg, bfa_boolean_t complete)
{
	bfa_hal_ioc_t  *ioc = cbarg;

	if (complete) {
		ioc->stats_cbfn(ioc->stats_cbarg, ioc->stats_status);
	} else {
		ioc->stats_busy = BFA_FALSE;
		ioc->stats_status = BFA_STATUS_OK;
	}
}

static void
bfa_hal_ioc_stats_clr_timeout(void *cbarg)
{
	bfa_hal_ioc_t  *ioc = (bfa_hal_ioc_t *) cbarg;

	bfa_trc(ioc->hal, 0);

	ioc->stats_status = BFA_STATUS_ETIMER;
	bfa_hcb_queue(ioc->hal, &ioc->stats_hcb_qe, __bfa_hcb_ioc_stats_clr,
		      ioc);
}

static void
__bfa_hcb_ioc_stats(void *cbarg, bfa_boolean_t complete)
{
	bfa_hal_ioc_t  *ioc = cbarg;

	if (complete) {
		if (ioc->stats_status == BFA_STATUS_OK)
			bfa_hal_ioc_stats_swap(ioc->stats_ret, ioc->stats);
		ioc->stats_cbfn(ioc->stats_cbarg, ioc->stats_status);
	} else {
		ioc->stats_busy = BFA_FALSE;
		ioc->stats_status = BFA_STATUS_OK;
	}
}

static void
bfa_hal_ioc_stats_timeout(void *cbarg)
{
	bfa_hal_ioc_t  *ioc = (bfa_hal_ioc_t *) cbarg;

	bfa_trc(ioc->hal, 0);

	ioc->stats_status = BFA_STATUS_ETIMER;
	bfa_hcb_queue(ioc->hal, &ioc->stats_hcb_qe, __bfa_hcb_ioc_stats, ioc);
}

/**
 * @brief
 *      Initiate a full firmware download.
 */
static void
bfa_hal_ioc_download_fw(bfa_hal_t *hal, uint32_t boot_type,
			uint32_t boot_param)
{
	bfa_hal_ioc_t  *ioc = &hal->ioc;
	uint32_t       *fwimg = bfa_hal_fwimg;
	uint32_t        pgnum, pgoff;
	uint32_t        loff = 0;
	uint32_t        i;

	/**
	 * Initialize LMEM first before code download
	 */
	bfa_hal_ioc_lmem_init(ioc);

	/**
	 * Flash based firmware boot
	 */
	bfa_trc(hal, bfa_hal_fwimg_size);
	if (bfa_hal_fwimg_size < BFA_HAL_FWIMG_MINSZ)
		boot_type = BFA_FWI_BOOT_TYPE_FLASH;
	fwimg[BFA_FWI_BOOT_TYPE_OFF / sizeof(uint32_t)] =
		bfa_os_swap32(boot_type);
	fwimg[BFA_FWI_BOOT_PARAM_OFF / sizeof(uint32_t)] =
		bfa_os_swap32(boot_param);

	pgnum = PSS_SMEM_PGNUM(loff);
	pgoff = PSS_SMEM_PGOFF(loff);

	bfa_hal_reg_write(ioc->bfa_regs.host_page_num_fn, pgnum);

	for (i = 0; i < bfa_hal_fwimg_size; i++) {
	/**
	  * write smem
	  */
		bfa_hal_mem_write(ioc->bfa_regs.smem_page_start, loff,
				  fwimg[i]);
		loff += sizeof(uint32_t);

	/**
	  * handle page offset wrap around
	  */
		loff = PSS_SMEM_PGOFF(loff);
		if (loff == 0) {
			pgnum++;
			bfa_hal_reg_write(ioc->bfa_regs.host_page_num_fn,
					  pgnum);
		}
	}

	bfa_hal_reg_write(ioc->bfa_regs.host_page_num_fn, PSS_SMEM_PGNUM(0));
}

static void
bfa_hal_ioc_stats_query(bfa_hal_ioc_t *ioc)
{
	bfa_ioc_stats_get_req_t stats_req;

	bfa_hal_timer_start(ioc->hal, &ioc->stats_timer,
			    bfa_hal_ioc_stats_timeout, ioc, BFA_HAL_IOC_TOV);

	bfa_fwi_h2i_set(stats_req.mh, BFA_MC_IOC, BFA_IOC_H2I_GET_STATS_REQ,
			bfa_hal_ioc_lpuid(ioc->hal));
	bfa_hal_ioc_mbox_send(ioc, &stats_req, sizeof(bfa_ioc_stats_get_req_t));
}

static void
bfa_hal_ioc_reset_queues(bfa_hal_ioc_t *ioc)
{
	int             q;

	for (q = 0; q < BFA_IOC_MAX_CQS; q++) {
		bfa_hal_rspq_ci(ioc->hal, q) = 0;
		bfa_hal_reqq_pi(ioc->hal, q) = 0;
	}
}

static void
bfa_hal_ioc_reset(bfa_hal_ioc_t *ioc, bfa_boolean_t force)
{
	bfa_hal_ioc_reset_queues(ioc);
	bfa_hal_ioc_hwinit(ioc, force);
	if (ioc->ioc_started)
		bfa_hal_intr_enable(ioc->hal);
}

/**
 * @}
 */

/**
 * @ig hal_ioc_public
 * @{
 */

void
bfa_hal_ioc_set_snsbase(bfa_hal_t *hal, uint64_t snsbase_pa)
{
	bfa_hal_ioc_t      *ioc = &hal->ioc;

	ioc->cfginfo->sense_buf_len = (BFA_FWI_IOIM_SNSLEN - 1);
	bfa_hal_dmabe_addr_set(ioc->cfginfo->ioim_snsbase, snsbase_pa);
}

wwn_t
bfa_hal_ioc_get_pwwn(bfa_hal_t *hal)
{
	return bfa_hal_wwn_get_naa1(hal);
}

bfa_status_t
bfa_hal_ioc_pll_init(bfa_hal_t *hal)
{
	bfa_hal_ioc_t  *ioc = &hal->ioc;
	uint32_t        r32;
	bfa_os_addr_t   rb;

	rb = ioc->pci_info.pci_bar_kva;
	bfa_hal_reg_write((rb + BFA_IOC0_STATE_REG), BFA_FWI_IOC_UNINIT);
	bfa_hal_reg_write((rb + BFA_IOC1_STATE_REG), BFA_FWI_IOC_UNINIT);

	bfa_hal_reg_write((rb + HOSTFN0_INT_MSK), 0xffffffffU);
	bfa_hal_reg_write((rb + HOSTFN1_INT_MSK), 0xffffffffU);

	bfa_hal_reg_write(ioc->bfa_regs.app_pll_212_ctl_reg,
			  __APP_PLL_212_LOGIC_SOFT_RESET);
	bfa_hal_reg_write(ioc->bfa_regs.app_pll_212_ctl_reg,
			  __APP_PLL_212_BYPASS |
			  __APP_PLL_212_LOGIC_SOFT_RESET);
	bfa_hal_reg_write(ioc->bfa_regs.app_pll_400_ctl_reg,
			  __APP_PLL_400_LOGIC_SOFT_RESET);
	bfa_hal_reg_write(ioc->bfa_regs.app_pll_400_ctl_reg,
			  __APP_PLL_400_BYPASS |
			  __APP_PLL_400_LOGIC_SOFT_RESET);
	bfa_os_udelay(2);
	bfa_hal_reg_write(ioc->bfa_regs.app_pll_212_ctl_reg,
			  __APP_PLL_212_LOGIC_SOFT_RESET);
	bfa_hal_reg_write(ioc->bfa_regs.app_pll_400_ctl_reg,
			  __APP_PLL_400_LOGIC_SOFT_RESET);

	bfa_hal_reg_write(ioc->bfa_regs.app_pll_212_ctl_reg,
			  __APP_PLL_212_ENABLE | __APP_PLL_212_LRESETN |
			  __APP_PLL_212_P0_1(3U) | __APP_PLL_212_JITLMT0_1(3U) |
			  __APP_PLL_212_CNTLMT0_1(3U) |
			  __APP_PLL_212_LOGIC_SOFT_RESET);
	bfa_hal_reg_write(ioc->bfa_regs.app_pll_400_ctl_reg,
			  __APP_PLL_400_ENABLE | __APP_PLL_400_LRESETN |
			  __APP_PLL_400_RSEL200500 | __APP_PLL_400_P0_1(3U) |
			  __APP_PLL_400_JITLMT0_1(3U) |
			  __APP_PLL_400_CNTLMT0_1(3U) |
			  __APP_PLL_400_LOGIC_SOFT_RESET);

	/**
	 * Wait for PLLs to lock.
	 */
	bfa_os_udelay(2000);
	bfa_hal_reg_write((rb + HOSTFN0_INT_STATUS), 0xffffffffU);
	bfa_hal_reg_write((rb + HOSTFN1_INT_STATUS), 0xffffffffU);

	bfa_hal_reg_write(ioc->bfa_regs.app_pll_212_ctl_reg,
			  __APP_PLL_212_ENABLE | __APP_PLL_212_LRESETN |
			  __APP_PLL_212_P0_1(3U) | __APP_PLL_212_JITLMT0_1(3U) |
			  __APP_PLL_212_CNTLMT0_1(3U));
	bfa_hal_reg_write(ioc->bfa_regs.app_pll_400_ctl_reg,
			  __APP_PLL_400_ENABLE | __APP_PLL_400_LRESETN |
			  __APP_PLL_400_RSEL200500 | __APP_PLL_400_P0_1(3U) |
			  __APP_PLL_400_JITLMT0_1(3U) |
			  __APP_PLL_400_CNTLMT0_1(3U));

	r32 = bfa_hal_reg_read(ioc->bfa_regs.app_pll_212_ctl_reg);
	if (!(r32 & __P_212_PLL_LOCK)) {
		bfa_trc(ioc->hal, r32);
		return BFA_STATUS_FAILED;
	}

	r32 = bfa_hal_reg_read(ioc->bfa_regs.app_pll_400_ctl_reg);
	if (!(r32 & __P_400_PLL_LOCK)) {
		bfa_trc(ioc->hal, r32);
		return BFA_STATUS_FAILED;
	}

	return BFA_STATUS_OK;
}

void
bfa_hal_ioc_boot(bfa_hal_t *hal, uint32_t boot_type, uint32_t boot_param)
{
	bfa_hal_ioc_t  *ioc = &hal->ioc;
	bfa_os_addr_t   rb;

	if (bfa_hal_ioc_pll_init(ioc->hal) != BFA_STATUS_OK)
		return;

	/**
	 * Initialize IOC state of all functions on a chip reset.
	 */
	rb = ioc->pci_info.pci_bar_kva;
	bfa_hal_reg_write((rb + BFA_IOC0_STATE_REG), BFA_FWI_IOC_INITING);
	bfa_hal_reg_write((rb + BFA_IOC1_STATE_REG), BFA_FWI_IOC_INITING);

	bfa_hal_ioc_download_fw(ioc->hal, boot_type, boot_param);

	/**
	 * Enable interrupts just before starting LPU
	 */
	bfa_hal_intr_enable(ioc->hal);
	bfa_hal_lpu_start(ioc);
}

bfa_status_t
bfa_hal_ioc_get_stats(bfa_hal_t *hal, bfa_ioc_stats_t *stats,
		      bfa_hcb_ioc_t cbfn, void *cbarg)
{
	bfa_hal_ioc_t  *ioc = &hal->ioc;

	if (ioc->stats_busy) {
		bfa_trc(hal, ioc->stats_busy);
		return (BFA_STATUS_DEVBUSY);
	}

	ioc->stats_busy = BFA_TRUE;
	ioc->stats_ret = stats;
	ioc->stats_cbfn = cbfn;
	ioc->stats_cbarg = cbarg;

	bfa_hal_ioc_stats_query(ioc);

	return (BFA_STATUS_OK);
}

bfa_status_t
bfa_hal_ioc_clear_stats(bfa_hal_t *hal, bfa_hcb_ioc_t cbfn, void *cbarg)
{
	bfa_hal_ioc_t  *ioc = &hal->ioc;

	if (ioc->stats_busy) {
		bfa_trc(hal, ioc->stats_busy);
		return (BFA_STATUS_DEVBUSY);
	}

	ioc->stats_busy = BFA_TRUE;
	ioc->stats_cbfn = cbfn;
	ioc->stats_cbarg = cbarg;

	bfa_hal_ioc_stats_clear(ioc);
	return (BFA_STATUS_OK);
}

void
bfa_hal_ioc_get_attr(bfa_hal_t *hal, bfa_ioc_attr_t *ioc_attr)
{
	bfa_hal_ioc_t      *ioc = &hal->ioc;

	bfa_os_memset((void *)ioc_attr, 0, sizeof(bfa_ioc_attr_t));

	ioc_attr->state = bfa_sm_to_state(ioc_sm_table, hal->ioc.sm);
	ioc_attr->fwcfg = hal->ioc.cfg;
	ioc_attr->intr_attr = ioc->cfginfo->intr_attr;
	bfa_hal_ioc_get_adapter_attr(hal, &ioc_attr->adapter_attr);

	ioc_attr->chip_attr.chip_rev[0] = 'R';
	ioc_attr->chip_attr.chip_rev[1] = 'e';
	ioc_attr->chip_attr.chip_rev[2] = 'v';
	ioc_attr->chip_attr.chip_rev[3] = '-';
	ioc_attr->chip_attr.chip_rev[4] = ioc_attr->adapter_attr.asic_rev;
	ioc_attr->chip_attr.chip_rev[5] = '\0';
}

void
bfa_hal_pm_ctl(bfa_hal_t *hal, bfa_hal_pm_cmd_t cmd)
{
}

void
bfa_hal_debug(bfa_hal_t *hal)
{
	bfa_hal_ioc_t  *ioc = &hal->ioc;
	uint32_t        loop = 0;
	uint32_t        reg = 4;
	uint32_t        val = 0;
	uint32_t       *base;

	base = (uint32_t *) ioc->pci_info.pci_bar_kva;

	while (loop) {
		bfa_hal_reg_write((base + (reg >> 2)), val);
		loop++;
		reg++;
		val++;
	}
}

/**
 * @brief
 * Query IOC memory requirement information.
 */
void
bfa_hal_ioc_meminfo(bfa_ioc_cfg_t *cfg, uint32_t *km_len, uint32_t *dm_len)
{
	bfa_hal_fw_cfg_sz(cfg, dm_len);
	bfa_hal_cqs_sz(cfg, dm_len);
}

/**
 * @brief
 * Query IOC memory requirement information.
 */
void
bfa_hal_ioc_attach(bfa_hal_t *hal, void *bfad, bfa_ioc_cfg_t *cfg,
		   bfa_hal_meminfo_t *meminfo, bfa_hal_pcidev_t *pcidev)
{
	int             i;

	bfa_hal_init_struct(hal, bfad, cfg, pcidev);
	bfa_hal_ioc_mem_claim(hal, cfg, meminfo);
	bfa_hal_timer_init(hal);

	bfa_q_init(&hal->comp_q);
	for (i = 0; i < BFA_IOC_MAX_CQS; i++)
		bfa_q_init(&hal->reqq_waitq[i]);

	hal->ioc.hwinit_count = 0;
	bfa_sm_set_state(&hal->ioc, bfa_hal_ioc_sm_reset);
}

/**
 * @brief
 * Query IOC memory requirement information.
 */
void
bfa_hal_ioc_detach(bfa_hal_t *hal)
{
	/*
	 * no-op
	 */
}

/**
 * @brief
 * Query IOC memory requirement information.
 */
void
bfa_hal_ioc_init(bfa_hal_t *hal)
{
	bfa_hal_ioc_t  *ioc = &hal->ioc;

	ioc->init_cb = BFA_TRUE;
	bfa_sm_send_event(&hal->ioc, IOC_E_INIT);
}

/**
 * @brief
 * IOC start called from bfa_hal_start(). Called to start IOC operations
 * at driver instantiation for this instance.
 */
void
bfa_hal_ioc_start(bfa_hal_t *hal)
{
	bfa_hal_ioc_t  *ioc = &hal->ioc;

	bfa_assert(ioc->ioc_started == BFA_FALSE);
	ioc->ioc_started = BFA_TRUE;
	bfa_sm_send_event(&hal->ioc, IOC_E_ENABLE);
}

/**
 * @brief
 * IOC stop called from bfa_hal_stop(). Called only when driver is unloaded
 * for this instance.
 */
void
bfa_hal_ioc_stop(bfa_hal_t *hal)
{
	bfa_hal_ioc_t  *ioc = &hal->ioc;

	bfa_assert(ioc->ioc_started == BFA_TRUE);

	ioc->ioc_started = BFA_FALSE;
	ioc->stop_cb = BFA_TRUE;
	bfa_sm_send_event(&hal->ioc, IOC_E_DISABLE);
}

/**
 * Enable IOC after it is disabled.
 */
bfa_status_t
bfa_hal_ioc_enable(bfa_hal_t *hal)
{
	bfa_hal_ioc_t  *ioc = &hal->ioc;

	if (!bfa_sm_cmp_state(ioc, bfa_hal_ioc_sm_disabled))
		return BFA_STATUS_IOC_ENABLED;

	ioc->ioc_started = BFA_TRUE;
	bfa_plog_str(hal->plog, BFA_PL_MID_HAL, BFA_PL_EID_MISC, 0,
		     "IOC Enable");
	bfa_sm_send_event(&hal->ioc, IOC_E_ENABLE);
	return BFA_STATUS_OK;
}

void
bfa_hal_ioc_disable(bfa_hal_t *hal)
{
	bfa_hal_ioc_t  *ioc = &hal->ioc;

	ioc->disable_cb = BFA_TRUE;
	bfa_plog_str(hal->plog, BFA_PL_MID_HAL, BFA_PL_EID_MISC, 0,
		     "IOC Disable");
	bfa_sm_send_event(&hal->ioc, IOC_E_DISABLE);
}

bfa_status_t
bfa_hal_intr_attr_set(bfa_hal_t *hal, bfa_ioc_intr_attr_t *attr)
{
	bfa_hal_ioc_t  	   *ioc = &hal->ioc;
	bfa_ioc_fwi_set_intr_req_t *m;

	ioc->cfginfo->intr_attr = *attr;
	if (!bfa_hal_ioc_is_operational(hal))
		return BFA_STATUS_OK;

	m = bfa_hal_reqq_next(hal, BFA_HAL_REQQ_IOC);
	if (!m)
		return BFA_STATUS_DEVBUSY;

	bfa_fwi_h2i_set(m->mh, BFA_MC_IOC, BFA_IOC_H2I_SET_INTR_REQ,
			bfa_hal_ioc_lpuid(hal));
	m->coalesce = attr->coalesce;
	m->delay    = bfa_os_htons(attr->delay);
	m->latency  = bfa_os_htons(attr->latency);

	bfa_trc(hal, attr->delay);
	bfa_trc(hal, attr->latency);

	bfa_hal_reqq_produce(hal, BFA_HAL_REQQ_IOC);
	return BFA_STATUS_OK;
}

/**
 * @brief
 */
bfa_boolean_t
bfa_hal_ioc_is_operational(bfa_hal_t *hal)
{
	return bfa_sm_cmp_state(&hal->ioc, bfa_hal_ioc_sm_op);
}

#define bfa_hal_ioc_state_disabled(__sm)		\
	(((__sm) == BFA_FWI_IOC_UNINIT) ||		\
	 ((__sm) == BFA_FWI_IOC_INITING) ||		\
	 ((__sm) == BFA_FWI_IOC_HWINIT) ||		\
	 ((__sm) == BFA_FWI_IOC_DISABLED) ||		\
	 ((__sm) == BFA_FWI_IOC_HBFAIL) ||		\
	 ((__sm) == BFA_FWI_IOC_CFG_DISABLED))

/**
 * Check if adapter is disabled -- both IOCs should be in a disabled
 * state.
 */
bfa_boolean_t
bfa_hal_adapter_is_disabled(bfa_hal_t *hal)
{
	bfa_hal_ioc_t  *ioc = &hal->ioc;
	uint32_t        ioc_state;
	bfa_os_addr_t   rb = ioc->pci_info.pci_bar_kva;

	if (!bfa_sm_cmp_state(ioc, bfa_hal_ioc_sm_disabled))
		return BFA_FALSE;

	ioc_state = bfa_hal_reg_read(rb + BFA_IOC0_STATE_REG);
	if (!bfa_hal_ioc_state_disabled(ioc_state))
		return BFA_FALSE;

	ioc_state = bfa_hal_reg_read(rb + BFA_IOC1_STATE_REG);
	if (!bfa_hal_ioc_state_disabled(ioc_state))
		return BFA_FALSE;

	return BFA_TRUE;
}

void
bfa_hal_ioc_isr(bfa_hal_ioc_t *ioc, bfa_ioc_i2h_msg_t *msg)
{
	bfa_trc(ioc->hal, msg->mh.msg_id);

	switch (msg->mh.msg_id) {
	case BFA_IOC_I2H_READY_EVENT:
		bfa_sm_send_event(ioc, IOC_E_FWREADY);
		break;

	case BFA_IOC_I2H_CFG_REPLY:
		ioc->event_args.cfg_reply = &msg->cfg_reply;
		bfa_hal_ioc_cfg_update(ioc);
		bfa_sm_send_event(ioc, IOC_E_FWRSP_CFG);
		break;

	case BFA_IOC_I2H_ENABLE_REPLY:
		bfa_sm_send_event(ioc, IOC_E_FWRSP_ENABLE);
		break;

	case BFA_IOC_I2H_DISABLE_REPLY:
		bfa_sm_send_event(ioc, IOC_E_FWRSP_DISABLE);
		break;

	case BFA_IOC_I2H_GET_STATS_RSP:
		if (ioc->stats_busy == BFA_FALSE
		    || ioc->stats_status == BFA_STATUS_ETIMER)
			break;

		bfa_hal_timer_stop(&ioc->stats_timer);
		ioc->stats_status = BFA_STATUS_OK;
		bfa_hcb_queue(ioc->hal, &ioc->stats_hcb_qe, __bfa_hcb_ioc_stats,
			      ioc);
		break;
	case BFA_IOC_I2H_CLEAR_STATS_RSP:
		/*
		 * check for timer pop before processing the rsp
		 */
		if (ioc->stats_busy == BFA_FALSE
		    || ioc->stats_status == BFA_STATUS_ETIMER)
			break;

		bfa_hal_timer_stop(&ioc->stats_timer);
		ioc->stats_status = BFA_STATUS_OK;
		bfa_hcb_queue(ioc->hal, &ioc->stats_hcb_qe,
			      __bfa_hcb_ioc_stats_clr, ioc);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @}
 */
