/*
 * Copyright (c) 2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

#include <hal/bfa_hal.h>
#include <cs/bfa_debug.h>
#include <cs/bfa_stats.h>
#include <bfa_hcb_ioim_macros.h>

BFA_TRC_FILE(HAL, IOIM);

/*
 * forward declarations.
 */
static bfa_boolean_t	bfa_hal_ioim_send_ioreq(bfa_hal_ioim_t *ioim);
static bfa_boolean_t	bfa_hal_ioim_sge_setup(bfa_hal_ioim_t *ioim);
static void		bfa_hal_ioim_sgpg_setup(bfa_hal_ioim_t *ioim);
static bfa_boolean_t	bfa_hal_ioim_send_abort(bfa_hal_ioim_t *ioim);
static void		bfa_hal_ioim_notify_cleanup(bfa_hal_ioim_t *ioim);
static void		__bfa_hcb_ioim_good_comp(void *cbarg,
							bfa_boolean_t complete);
static void		__bfa_hcb_ioim_comp(void *cbarg,
							bfa_boolean_t complete);
static void		__bfa_hcb_ioim_abort(void *cbarg,
							bfa_boolean_t complete);
static void		__bfa_hcb_ioim_failed(void *cbarg,
							bfa_boolean_t complete);
static void		__bfa_hcb_ioim_timedout(void *cbarg,
							bfa_boolean_t complete);

/**
 * @ig hal_ioim_sm
 * @{
 */

/**
 * IO state machine events
 */
typedef enum {
	BFA_HAL_IOIM_SM_START	   = 1,	/* !< io start request from host */
	BFA_HAL_IOIM_SM_COMP_GOOD  = 2,	/* !< io good comp, resource free */
	BFA_HAL_IOIM_SM_COMP	   = 3,	/* !< io comp, resource is free */
	BFA_HAL_IOIM_SM_COMP_UTAG  = 4,	/* !< io comp, resource is free */
	BFA_HAL_IOIM_SM_DONE	   = 5,	/* !< io comp, resource not free */
	BFA_HAL_IOIM_SM_FREE	   = 6,	/* !< io resource is freed */
	BFA_HAL_IOIM_SM_ABORT	   = 7,	/* !< abort request from scsi stack */
	BFA_HAL_IOIM_SM_ABORT_COMP = 8,	/* !< abort from f/w */
	BFA_HAL_IOIM_SM_ABORT_DONE = 9,	/* !< abort completion from f/w */
	BFA_HAL_IOIM_SM_QRESUME	   = 10,/* !< CQ space available to queue IO */
	BFA_HAL_IOIM_SM_SGALLOCED  = 11,/* !< SG page allocation successful */
	BFA_HAL_IOIM_SM_SQRETRY	   = 12,/* !< sequence recovery retry */
	BFA_HAL_IOIM_SM_HCB	   = 13,/* !< hal callback complete */
	BFA_HAL_IOIM_SM_CLEANUP	   = 14,/* !< IO cleanup from itnim */
	BFA_HAL_IOIM_SM_TMSTART	   = 15,/* !< IO cleanup from tskim */
	BFA_HAL_IOIM_SM_TMDONE	   = 16,/* !< IO cleanup from tskim */
	BFA_HAL_IOIM_SM_HWFAIL	   = 17,/* !< IOC h/w failure event */
	BFA_HAL_IOIM_SM_IOTOV	   = 18,/* !< ITN offline TOV       */
} bfa_hal_ioim_event_t;

/*
 * forward declaration of IO state machine
 */
static void     bfa_hal_ioim_sm_uninit(bfa_hal_ioim_t *ioim,
				       bfa_hal_ioim_event_t event);
static void     bfa_hal_ioim_sm_sgalloc(bfa_hal_ioim_t *ioim,
					bfa_hal_ioim_event_t event);
static void     bfa_hal_ioim_sm_active(bfa_hal_ioim_t *ioim,
				       bfa_hal_ioim_event_t event);
static void     bfa_hal_ioim_sm_abort(bfa_hal_ioim_t *ioim,
				      bfa_hal_ioim_event_t event);
static void     bfa_hal_ioim_sm_cleanup(bfa_hal_ioim_t *ioim,
					bfa_hal_ioim_event_t event);
static void     bfa_hal_ioim_sm_qfull(bfa_hal_ioim_t *ioim,
				      bfa_hal_ioim_event_t event);
static void     bfa_hal_ioim_sm_abort_qfull(bfa_hal_ioim_t *ioim,
					    bfa_hal_ioim_event_t event);
static void     bfa_hal_ioim_sm_cleanup_qfull(bfa_hal_ioim_t *ioim,
					      bfa_hal_ioim_event_t event);
static void     bfa_hal_ioim_sm_hcb(bfa_hal_ioim_t *ioim,
				    bfa_hal_ioim_event_t event);
static void     bfa_hal_ioim_sm_hcb_free(bfa_hal_ioim_t *ioim,
					 bfa_hal_ioim_event_t event);
static void     bfa_hal_ioim_sm_resfree(bfa_hal_ioim_t *ioim,
					bfa_hal_ioim_event_t event);

/**
 * @brief
 * 		IO is not started (unallocated).
 */
static void
bfa_hal_ioim_sm_uninit(bfa_hal_ioim_t *ioim, bfa_hal_ioim_event_t event)
{
	bfa_trc_fp(ioim->hal, ioim->iotag);
	bfa_trc_fp(ioim->hal, event);

	switch (event) {
	case BFA_HAL_IOIM_SM_START:
		if (!bfa_hal_itnim_is_online(ioim->itnim)) {
			if (!bfa_hal_itnim_is_offline(ioim->itnim)) {
				bfa_sm_set_state(ioim, bfa_hal_ioim_sm_hcb);
				bfa_q_qe_deq(ioim);
				bfa_q_enq(&ioim->fcpim->ioim_comp_q, ioim);
				bfa_hcb_queue(ioim->hal, &ioim->hcb_qe,
				      	__bfa_hcb_ioim_failed, ioim);
			} else {
				bfa_q_qe_deq(ioim);
				bfa_q_enq(&ioim->itnim->pending_q, ioim);
			}
			break;
		}

		if (ioim->nsges > BFA_SGE_DATA_MAX_INLINE) {
			if (!bfa_hal_ioim_sge_setup(ioim)) {
				bfa_sm_set_state(ioim, bfa_hal_ioim_sm_sgalloc);
				return;
			}
		}

		if (!bfa_hal_ioim_send_ioreq(ioim)) {
			bfa_sm_set_state(ioim, bfa_hal_ioim_sm_qfull);
			break;
		}

		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_active);
		break;

	case BFA_HAL_IOIM_SM_IOTOV:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_hcb);
		bfa_hcb_queue(ioim->hal, &ioim->hcb_qe,
			      __bfa_hcb_ioim_timedout, ioim);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * 		IO is waiting for SG pages.
 */
static void
bfa_hal_ioim_sm_sgalloc(bfa_hal_ioim_t *ioim, bfa_hal_ioim_event_t event)
{
	bfa_trc(ioim->hal, ioim->iotag);
	bfa_trc(ioim->hal, event);

	switch (event) {
	case BFA_HAL_IOIM_SM_SGALLOCED:
		if (!bfa_hal_ioim_send_ioreq(ioim)) {
			bfa_sm_set_state(ioim, bfa_hal_ioim_sm_qfull);
			break;
		}
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_active);
		break;

	case BFA_HAL_IOIM_SM_CLEANUP:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_hcb);
		bfa_hal_sgpg_wcancel(ioim->hal, &ioim->iosp->sgpg_wqe);
		bfa_hcb_queue(ioim->hal, &ioim->hcb_qe, __bfa_hcb_ioim_failed,
			      ioim);
		bfa_hal_ioim_notify_cleanup(ioim);
		break;

	case BFA_HAL_IOIM_SM_ABORT:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_hcb);
		bfa_hal_sgpg_wcancel(ioim->hal, &ioim->iosp->sgpg_wqe);
		bfa_hcb_queue(ioim->hal, &ioim->hcb_qe, __bfa_hcb_ioim_abort,
			      ioim);
		break;

	case BFA_HAL_IOIM_SM_HWFAIL:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_hcb);
		bfa_hal_sgpg_wcancel(ioim->hal, &ioim->iosp->sgpg_wqe);
		bfa_hcb_queue(ioim->hal, &ioim->hcb_qe, __bfa_hcb_ioim_failed,
			      ioim);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * 		IO is active.
 */
static void
bfa_hal_ioim_sm_active(bfa_hal_ioim_t *ioim, bfa_hal_ioim_event_t event)
{
	bfa_trc_fp(ioim->hal, ioim->iotag);
	bfa_trc_fp(ioim->hal, event);

	switch (event) {
	case BFA_HAL_IOIM_SM_COMP_GOOD:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_hcb);
		bfa_hcb_queue(ioim->hal, &ioim->hcb_qe,
			      __bfa_hcb_ioim_good_comp, ioim);
		break;

	case BFA_HAL_IOIM_SM_COMP:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_hcb);
		bfa_hcb_queue(ioim->hal, &ioim->hcb_qe, __bfa_hcb_ioim_comp,
			      ioim);
		break;

	case BFA_HAL_IOIM_SM_DONE:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_hcb_free);
		bfa_hcb_queue(ioim->hal, &ioim->hcb_qe, __bfa_hcb_ioim_comp,
			      ioim);
		break;

	case BFA_HAL_IOIM_SM_ABORT:
		ioim->iosp->abort_explicit = BFA_TRUE;
		ioim->io_cbfn = __bfa_hcb_ioim_abort;

		if (bfa_hal_ioim_send_abort(ioim))
			bfa_sm_set_state(ioim, bfa_hal_ioim_sm_abort);
		else {
			bfa_sm_set_state(ioim, bfa_hal_ioim_sm_abort_qfull);
			bfa_hal_reqq_wait(ioim->hal, ioim->itnim->reqq,
					  &ioim->iosp->reqq_wait);
		}
		break;

	case BFA_HAL_IOIM_SM_CLEANUP:
		ioim->iosp->abort_explicit = BFA_FALSE;
		ioim->io_cbfn = __bfa_hcb_ioim_failed;

		if (bfa_hal_ioim_send_abort(ioim))
			bfa_sm_set_state(ioim, bfa_hal_ioim_sm_cleanup);
		else {
			bfa_sm_set_state(ioim, bfa_hal_ioim_sm_cleanup_qfull);
			bfa_hal_reqq_wait(ioim->hal, ioim->itnim->reqq,
					  &ioim->iosp->reqq_wait);
		}
		break;

	case BFA_HAL_IOIM_SM_HWFAIL:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_hcb);
		bfa_hcb_queue(ioim->hal, &ioim->hcb_qe, __bfa_hcb_ioim_failed,
			      ioim);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * 		IO is being aborted, waiting for completion from firmware.
 */
static void
bfa_hal_ioim_sm_abort(bfa_hal_ioim_t *ioim, bfa_hal_ioim_event_t event)
{
	bfa_trc(ioim->hal, ioim->iotag);
	bfa_trc(ioim->hal, event);

	switch (event) {
	case BFA_HAL_IOIM_SM_COMP_GOOD:
	case BFA_HAL_IOIM_SM_COMP:
	case BFA_HAL_IOIM_SM_DONE:
	case BFA_HAL_IOIM_SM_FREE:
		break;

	case BFA_HAL_IOIM_SM_ABORT_DONE:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_hcb_free);
		bfa_hcb_queue(ioim->hal, &ioim->hcb_qe, __bfa_hcb_ioim_abort,
			      ioim);
		break;

	case BFA_HAL_IOIM_SM_ABORT_COMP:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_hcb);
		bfa_hcb_queue(ioim->hal, &ioim->hcb_qe, __bfa_hcb_ioim_abort,
			      ioim);
		break;

	case BFA_HAL_IOIM_SM_COMP_UTAG:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_hcb);
		bfa_hcb_queue(ioim->hal, &ioim->hcb_qe, __bfa_hcb_ioim_abort,
			      ioim);
		break;

	case BFA_HAL_IOIM_SM_CLEANUP:
		bfa_assert(ioim->iosp->abort_explicit == BFA_TRUE);
		ioim->iosp->abort_explicit = BFA_FALSE;

		if (bfa_hal_ioim_send_abort(ioim))
			bfa_sm_set_state(ioim, bfa_hal_ioim_sm_cleanup);
		else {
			bfa_sm_set_state(ioim, bfa_hal_ioim_sm_cleanup_qfull);
			bfa_hal_reqq_wait(ioim->hal, ioim->itnim->reqq,
					  &ioim->iosp->reqq_wait);
		}
		break;

	case BFA_HAL_IOIM_SM_HWFAIL:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_hcb);
		bfa_hcb_queue(ioim->hal, &ioim->hcb_qe, __bfa_hcb_ioim_failed,
			      ioim);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * IO is being cleaned up (implicit abort), waiting for completion from
 * firmware.
 */
static void
bfa_hal_ioim_sm_cleanup(bfa_hal_ioim_t *ioim, bfa_hal_ioim_event_t event)
{
	bfa_trc(ioim->hal, ioim->iotag);
	bfa_trc(ioim->hal, event);

	switch (event) {
	case BFA_HAL_IOIM_SM_COMP_GOOD:
	case BFA_HAL_IOIM_SM_COMP:
	case BFA_HAL_IOIM_SM_DONE:
	case BFA_HAL_IOIM_SM_FREE:
		break;

	case BFA_HAL_IOIM_SM_ABORT:
		/**
		 * IO is already being aborted implicitly
		 */
		ioim->io_cbfn = __bfa_hcb_ioim_abort;
		break;

	case BFA_HAL_IOIM_SM_ABORT_DONE:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_hcb_free);
		bfa_hcb_queue(ioim->hal, &ioim->hcb_qe, ioim->io_cbfn, ioim);
		bfa_hal_ioim_notify_cleanup(ioim);
		break;

	case BFA_HAL_IOIM_SM_ABORT_COMP:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_hcb);
		bfa_hcb_queue(ioim->hal, &ioim->hcb_qe, ioim->io_cbfn, ioim);
		bfa_hal_ioim_notify_cleanup(ioim);
		break;

	case BFA_HAL_IOIM_SM_COMP_UTAG:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_hcb);
		bfa_hcb_queue(ioim->hal, &ioim->hcb_qe, ioim->io_cbfn, ioim);
		bfa_hal_ioim_notify_cleanup(ioim);
		break;

	case BFA_HAL_IOIM_SM_HWFAIL:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_hcb);
		bfa_hcb_queue(ioim->hal, &ioim->hcb_qe, __bfa_hcb_ioim_failed,
			      ioim);
		break;

	case BFA_HAL_IOIM_SM_CLEANUP:
		/**
		 * IO can be in cleanup state already due to TM command. 2nd cleanup
		 * request comes from ITN offline event.
		 */
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * 		IO is waiting for room in request CQ
 */
static void
bfa_hal_ioim_sm_qfull(bfa_hal_ioim_t *ioim, bfa_hal_ioim_event_t event)
{
	bfa_trc(ioim->hal, ioim->iotag);
	bfa_trc(ioim->hal, event);

	switch (event) {
	case BFA_HAL_IOIM_SM_QRESUME:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_active);
		bfa_hal_ioim_send_ioreq(ioim);
		break;

	case BFA_HAL_IOIM_SM_ABORT:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_hcb);
		bfa_hal_reqq_wcancel(&ioim->iosp->reqq_wait);
		bfa_hcb_queue(ioim->hal, &ioim->hcb_qe, __bfa_hcb_ioim_abort,
			      ioim);
		break;

	case BFA_HAL_IOIM_SM_CLEANUP:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_hcb);
		bfa_hal_reqq_wcancel(&ioim->iosp->reqq_wait);
		bfa_hcb_queue(ioim->hal, &ioim->hcb_qe, __bfa_hcb_ioim_failed,
			      ioim);
		bfa_hal_ioim_notify_cleanup(ioim);
		break;

	case BFA_HAL_IOIM_SM_HWFAIL:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_hcb);
		bfa_hal_reqq_wcancel(&ioim->iosp->reqq_wait);
		bfa_hcb_queue(ioim->hal, &ioim->hcb_qe, __bfa_hcb_ioim_failed,
			      ioim);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * 		Active IO is being aborted, waiting for room in request CQ.
 */
static void
bfa_hal_ioim_sm_abort_qfull(bfa_hal_ioim_t *ioim, bfa_hal_ioim_event_t event)
{
	bfa_trc(ioim->hal, ioim->iotag);
	bfa_trc(ioim->hal, event);

	switch (event) {
	case BFA_HAL_IOIM_SM_QRESUME:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_abort);
		bfa_hal_ioim_send_abort(ioim);
		break;

	case BFA_HAL_IOIM_SM_CLEANUP:
		bfa_assert(ioim->iosp->abort_explicit == BFA_TRUE);
		ioim->iosp->abort_explicit = BFA_FALSE;
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_cleanup_qfull);
		break;

	case BFA_HAL_IOIM_SM_COMP_GOOD:
	case BFA_HAL_IOIM_SM_COMP:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_hcb);
		bfa_hal_reqq_wcancel(&ioim->iosp->reqq_wait);
		bfa_hcb_queue(ioim->hal, &ioim->hcb_qe, __bfa_hcb_ioim_abort,
			      ioim);
		break;

	case BFA_HAL_IOIM_SM_DONE:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_hcb_free);
		bfa_hal_reqq_wcancel(&ioim->iosp->reqq_wait);
		bfa_hcb_queue(ioim->hal, &ioim->hcb_qe, __bfa_hcb_ioim_abort,
			      ioim);
		break;

	case BFA_HAL_IOIM_SM_HWFAIL:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_hcb);
		bfa_hcb_queue(ioim->hal, &ioim->hcb_qe, __bfa_hcb_ioim_failed,
			      ioim);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * 		Active IO is being cleaned up, waiting for room in request CQ.
 */
static void
bfa_hal_ioim_sm_cleanup_qfull(bfa_hal_ioim_t *ioim, bfa_hal_ioim_event_t event)
{
	bfa_trc(ioim->hal, ioim->iotag);
	bfa_trc(ioim->hal, event);

	switch (event) {
	case BFA_HAL_IOIM_SM_QRESUME:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_abort);
		bfa_hal_ioim_send_abort(ioim);
		break;

	case BFA_HAL_IOIM_SM_ABORT:
		/**
		 * IO is alraedy being cleaned up implicitly
		 */
		ioim->io_cbfn = __bfa_hcb_ioim_abort;
		break;

	case BFA_HAL_IOIM_SM_COMP_GOOD:
	case BFA_HAL_IOIM_SM_COMP:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_hcb);
		bfa_hal_reqq_wcancel(&ioim->iosp->reqq_wait);
		bfa_hcb_queue(ioim->hal, &ioim->hcb_qe, ioim->io_cbfn, ioim);
		bfa_hal_ioim_notify_cleanup(ioim);
		break;

	case BFA_HAL_IOIM_SM_DONE:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_hcb_free);
		bfa_hal_reqq_wcancel(&ioim->iosp->reqq_wait);
		bfa_hcb_queue(ioim->hal, &ioim->hcb_qe, ioim->io_cbfn, ioim);
		bfa_hal_ioim_notify_cleanup(ioim);
		break;

	case BFA_HAL_IOIM_SM_HWFAIL:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_hcb);
		bfa_hal_reqq_wcancel(&ioim->iosp->reqq_wait);
		bfa_hcb_queue(ioim->hal, &ioim->hcb_qe, __bfa_hcb_ioim_failed,
			      ioim);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * IO hal callback is pending.
 */
static void
bfa_hal_ioim_sm_hcb(bfa_hal_ioim_t *ioim, bfa_hal_ioim_event_t event)
{
	bfa_trc_fp(ioim->hal, ioim->iotag);
	bfa_trc_fp(ioim->hal, event);

	switch (event) {
	case BFA_HAL_IOIM_SM_HCB:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_uninit);
		bfa_hal_ioim_free(ioim);
		break;

	case BFA_HAL_IOIM_SM_CLEANUP:
		bfa_hal_ioim_notify_cleanup(ioim);
		break;

	case BFA_HAL_IOIM_SM_HWFAIL:
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * IO hal callback is pending. IO resource cannot be freed.
 */
static void
bfa_hal_ioim_sm_hcb_free(bfa_hal_ioim_t *ioim, bfa_hal_ioim_event_t event)
{
	bfa_trc(ioim->hal, ioim->iotag);
	bfa_trc(ioim->hal, event);

	switch (event) {
	case BFA_HAL_IOIM_SM_HCB:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_resfree);
		bfa_q_qe_deq(ioim);
		bfa_q_enq(&ioim->fcpim->ioim_resfree_q, ioim);
		break;

	case BFA_HAL_IOIM_SM_FREE:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_hcb);
		break;

	case BFA_HAL_IOIM_SM_CLEANUP:
		bfa_hal_ioim_notify_cleanup(ioim);
		break;

	case BFA_HAL_IOIM_SM_HWFAIL:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_hcb);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * IO is completed, waiting resource free from firmware.
 */
static void
bfa_hal_ioim_sm_resfree(bfa_hal_ioim_t *ioim, bfa_hal_ioim_event_t event)
{
	bfa_trc(ioim->hal, ioim->iotag);
	bfa_trc(ioim->hal, event);

	switch (event) {
	case BFA_HAL_IOIM_SM_FREE:
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_uninit);
		bfa_hal_ioim_free(ioim);
		break;

	case BFA_HAL_IOIM_SM_CLEANUP:
		bfa_hal_ioim_notify_cleanup(ioim);
		break;

	case BFA_HAL_IOIM_SM_HWFAIL:
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @}
 */

/**
 * @ig hal_ioim_private
 * @{
 */

static void
__bfa_hcb_ioim_good_comp(void *cbarg, bfa_boolean_t complete)
{
	bfa_hal_ioim_t *ioim = cbarg;

	if (!complete) {
		bfa_sm_send_event(ioim, BFA_HAL_IOIM_SM_HCB);
		return;
	}

	bfa_hcb_ioim_good_comp(ioim->hal->bfad, ioim->dio);
}

static void
__bfa_hcb_ioim_comp(void *cbarg, bfa_boolean_t complete)
{
	bfa_hal_ioim_t	*ioim = cbarg;
	bfa_fwi_ioim_rsp_t *m;
	uint8_t		*snsinfo = NULL;
	uint8_t         sns_len = 0;
	int32_t         residue = 0;

	if (!complete) {
		bfa_sm_send_event(ioim, BFA_HAL_IOIM_SM_HCB);
		return;
	}

	m = (bfa_fwi_ioim_rsp_t *) &ioim->iosp->comp_rspmsg;
	if (m->rsp_info.io_status == BFA_IOIM_STS_OK) {
		/**
		 * setup sense information, if present
		 */
		if ((m->fcp_rsp.scsi_status == SCSI_STATUS_CHECK_CONDITION)
		    && fcp_snslen(&m->fcp_rsp)) {
			m->fcp_rsp.sns_len = bfa_os_ntohl(m->fcp_rsp.sns_len);
			m->fcp_rsp.rsp_len = bfa_os_ntohl(m->fcp_rsp.rsp_len);
			sns_len = fcp_snslen(&m->fcp_rsp);
			snsinfo = ioim->iosp->snsinfo;
		}

		/**
		 * setup residue value correctly for normal completions
		 */
		if (m->fcp_rsp.resid_flags == FCP_RESID_UNDER)
			residue = bfa_os_ntohl(m->fcp_rsp.residue);
		if (m->fcp_rsp.resid_flags == FCP_RESID_OVER) {
			residue = bfa_os_ntohl(m->fcp_rsp.residue);
			residue = -residue;
		}
	}

	bfa_hcb_ioim_done(ioim->hal->bfad, ioim->dio, m->rsp_info.io_status,
			  m->fcp_rsp.scsi_status, sns_len, snsinfo, residue);
}

static void
__bfa_hcb_ioim_failed(void *cbarg, bfa_boolean_t complete)
{
	bfa_hal_ioim_t *ioim = cbarg;

	if (!complete) {
		bfa_sm_send_event(ioim, BFA_HAL_IOIM_SM_HCB);
		return;
	}

	bfa_hcb_ioim_done(ioim->hal->bfad, ioim->dio, BFA_IOIM_STS_IO_ABORTED,
			  0, 0, NULL, 0);
}

static void
__bfa_hcb_ioim_timedout(void *cbarg, bfa_boolean_t complete)
{
	bfa_hal_ioim_t *ioim = cbarg;

	if (!complete) {
		bfa_sm_send_event(ioim, BFA_HAL_IOIM_SM_HCB);
		return;
	}

	bfa_hcb_ioim_done(ioim->hal->bfad, ioim->dio, BFA_IOIM_STS_IO_ABORTED,
			  0, 0, NULL, 0);
}

static void
__bfa_hcb_ioim_abort(void *cbarg, bfa_boolean_t complete)
{
	bfa_hal_ioim_t *ioim = cbarg;

	if (!complete) {
		bfa_sm_send_event(ioim, BFA_HAL_IOIM_SM_HCB);
		return;
	}

	bfa_hcb_ioim_abort(ioim->hal->bfad, ioim->dio);
}

static void
bfa_hal_ioim_sgpg_alloced(void *cbarg)
{
	bfa_hal_ioim_t *ioim = cbarg;

	ioim->nsgpgs = BFA_SGPG_NPAGE(ioim->nsges);
	bfa_q_mv(&ioim->iosp->sgpg_wqe.sgpg_q, &ioim->sgpg_q);
	bfa_hal_ioim_sgpg_setup(ioim);
	bfa_sm_send_event(ioim, BFA_HAL_IOIM_SM_SGALLOCED);
}

/**
 * @brief
 * Send I/O request to firmware.
 */
static          bfa_boolean_t
bfa_hal_ioim_send_ioreq(bfa_hal_ioim_t *ioim)
{
	bfa_hal_itnim_t *itnim = ioim->itnim;
	bfa_fwi_ioim_req_t *m;
	static fcp_cmnd_t cmnd_z0 = { 0 };
	bfa_sge_t      *sge;
	uint32_t        pgdlen = 0;

	/**
	 * check for room in queue to send request now
	 */
	m = bfa_hal_reqq_next(ioim->hal, itnim->reqq);
	if (!m) {
		bfa_hal_reqq_wait(ioim->hal, ioim->itnim->reqq,
				  &ioim->iosp->reqq_wait);
		return BFA_FALSE;
	}

	/**
	 * build i/o request message next
	 */
	m->io_tag = bfa_os_htons(ioim->iotag);
	m->rport_hdl = ioim->itnim->rport->fw_handle;
	m->io_timeout = bfa_hcb_ioim_get_timeout(ioim->dio);

	/**
	 * build inline IO SG element here
	 */
	sge = &m->sges[0];
	if (ioim->nsges) {
		sge->sga = bfa_hcb_ioim_get_sgaddr(ioim->dio, 0);
		pgdlen = bfa_hcb_ioim_get_sglen(ioim->dio, 0);
		sge->sg_len = pgdlen;
		sge->flags = (ioim->nsges > BFA_SGE_DATA_MAX_INLINE) ?
					BFA_SGE_DATA_CPL : BFA_SGE_DATA_LAST;
		bfa_sge_to_be(sge);
		sge++;
	}

	if (ioim->nsges > BFA_SGE_DATA_MAX_INLINE) {
		sge->sga = ioim->sgpg->sgpg_pa;
	} else {
		sge->sga.a32.addr_lo = 0;
		sge->sga.a32.addr_hi = 0;
	}
	sge->sg_len = pgdlen;
	sge->flags = BFA_SGE_PGDLEN;
	bfa_sge_to_be(sge);

	/**
	 * set up I/O command parameters
	 */
	m->cmnd = cmnd_z0;
	m->cmnd.lun = bfa_hcb_ioim_get_lun(ioim->dio);
	m->cmnd.iodir = bfa_hcb_ioim_get_iodir(ioim->dio);
	m->cmnd.cdb = *(scsi_cdb_t *) bfa_hcb_ioim_get_cdb(ioim->dio);
	m->cmnd.fcp_dl = bfa_os_htonl(bfa_hcb_ioim_get_size(ioim->dio));

	/**
	 * set up I/O message header
	 */
	switch (m->cmnd.iodir) {
	case FCP_IODIR_READ:
		bfa_fwi_h2i_set(m->mh, BFA_MC_IOIM_READ, 0,
				bfa_hal_ioc_lpuid(ioim->hal));
		break;
	case FCP_IODIR_WRITE:
		bfa_fwi_h2i_set(m->mh, BFA_MC_IOIM_WRITE, 0,
				bfa_hal_ioc_lpuid(ioim->hal));
		break;
	default:
		bfa_fwi_h2i_set(m->mh, BFA_MC_IOIM_IO, 0,
				bfa_hal_ioc_lpuid(ioim->hal));
	}
	if (itnim->seq_rec) {
		bfa_fwi_h2i_set(m->mh, BFA_MC_IOIM_IO, 0,
				bfa_hal_ioc_lpuid(ioim->hal));
	}

#ifdef IOIM_ADVANCED
	m->cmnd.crn = bfa_hcb_ioim_get_crn(ioim->dio);
	m->cmnd.priority = bfa_hcb_ioim_get_priority(ioim->dio);
	m->cmnd.taskattr = bfa_hcb_ioim_get_taskattr(ioim->dio);

	/**
	 * Handle large CDB (>16 bytes).
	 */
	m->cmnd.addl_cdb_len = (bfa_hcb_ioim_get_cdblen(ioim->dio) -
					FCP_CMND_CDB_LEN) / sizeof(uint32_t);
	if (m->cmnd.addl_cdb_len) {
		memcpy(&m->cmnd.cdb + 1, (scsi_cdb_t *)
				bfa_hcb_ioim_get_cdb(ioim->dio) + 1,
				m->cmnd.addl_cdb_len * sizeof(uint32_t));
		fcp_cmnd_fcpdl(&m->cmnd) =
				bfa_os_htonl(bfa_hcb_ioim_get_size(ioim->dio));
	}
#endif

	/**
	 * queue I/O message to firmware
	 */
	bfa_hal_reqq_produce(ioim->hal, itnim->reqq);
	return BFA_TRUE;
}

/**
 * @brief
 * Setup any additional SG pages needed.Inline SG element is setup
 * at queuing time.
 */
static bfa_boolean_t
bfa_hal_ioim_sge_setup(bfa_hal_ioim_t *ioim)
{
	uint16_t        nsgpgs;

	bfa_assert(ioim->nsges > BFA_SGE_DATA_MAX_INLINE);

	/**
	 * allocate SG pages needed
	 */
	nsgpgs = BFA_SGPG_NPAGE(ioim->nsges);
	if (!nsgpgs)
		return BFA_TRUE;

	if (bfa_hal_sgpg_malloc(ioim->hal, &ioim->sgpg_q, nsgpgs)
	    != BFA_STATUS_OK) {
		bfa_hal_sgpg_wait(ioim->hal, &ioim->iosp->sgpg_wqe, nsgpgs);
		return BFA_FALSE;
	}

	ioim->nsgpgs = nsgpgs;
	bfa_hal_ioim_sgpg_setup(ioim);

	return BFA_TRUE;
}

static void
bfa_hal_ioim_sgpg_setup(bfa_hal_ioim_t *ioim)
{
	int             sgeid, nsges, i;
	bfa_sge_t      *sge;
	bfa_hal_sgpg_t *sgpg;
	uint32_t        pgcumsz;

	sgeid = BFA_SGE_DATA_MAX_INLINE;
	ioim->sgpg = sgpg = bfa_q_first(&ioim->sgpg_q);

	do {
		sge = sgpg->sgpg->sges;
		nsges = ioim->nsges - sgeid;
		if (nsges > BFA_SGPG_NUM_DATA_SGES)
			nsges = BFA_SGPG_NUM_DATA_SGES;

		pgcumsz = 0;
		for (i = 0; i < nsges; i++, sge++, sgeid++) {
			sge->sga = bfa_hcb_ioim_get_sgaddr(ioim->dio, sgeid);
			sge->sg_len = bfa_hcb_ioim_get_sglen(ioim->dio, sgeid);
			pgcumsz += sge->sg_len;

			/**
			 * set flags
			 */
			if (i < (nsges - 1))
				sge->flags = BFA_SGE_DATA;
			else if (sgeid < (ioim->nsges - 1))
				sge->flags = BFA_SGE_DATA_CPL;
			else
				sge->flags = BFA_SGE_DATA_LAST;
		}

		sgpg = (bfa_hal_sgpg_t *) bfa_q_next(sgpg);

		/**
		 * set the link element of each page
		 */
		if (sgeid == ioim->nsges) {
			sge->flags = BFA_SGE_PGDLEN;
			sge->sga.a32.addr_lo = 0;
			sge->sga.a32.addr_hi = 0;
		} else {
			sge->flags = BFA_SGE_LINK;
			sge->sga = sgpg->sgpg_pa;
		}
		sge->sg_len = pgcumsz;
	} while (sgeid < ioim->nsges);
}

/**
 * @brief
 * Send I/O abort request to firmware.
 */
static          bfa_boolean_t
bfa_hal_ioim_send_abort(bfa_hal_ioim_t *ioim)
{
	bfa_hal_itnim_t          *itnim = ioim->itnim;
	bfa_fwi_ioim_abort_req_t *m;
	bfa_fwi_ioim_h2i_t       msgop;

	/**
	 * check for room in queue to send request now
	 */
	m = bfa_hal_reqq_next(ioim->hal, itnim->reqq);
	if (!m)
		return BFA_FALSE;

	/**
	 * build i/o request message next
	 */
	if (ioim->iosp->abort_explicit)
		msgop = BFA_FWI_IOIM_H2I_IOABORT_REQ;
	else
		msgop = BFA_FWI_IOIM_H2I_IOCLEANUP_REQ;

	bfa_fwi_h2i_set(m->mh, BFA_MC_IOIM, msgop,
			bfa_hal_ioc_lpuid(ioim->hal));
	m->io_tag    = bfa_os_htons(ioim->iotag);
	m->abort_tag = ++ ioim->abort_tag;

	/**
	 * queue I/O message to firmware
	 */
	bfa_hal_reqq_produce(ioim->hal, itnim->reqq);
	return BFA_TRUE;
}

/**
 * @brief
 * Call to resume any I/O requests waiting for room in request queue.
 */
static void
bfa_hal_ioim_qresume(void *cbarg)
{
	bfa_hal_ioim_t *ioim = cbarg;

	bfa_hal_fcpim_stats(ioim->fcpim, qresumes);
	bfa_sm_send_event(ioim, BFA_HAL_IOIM_SM_QRESUME);
}

/**
 * @brief Used to notify itn of IO completion as part of itn cleanup.
 */
static void
bfa_hal_ioim_notify_cleanup(bfa_hal_ioim_t *ioim)
{
	/**
	 * Move IO from itnim queue to fcpim global queue since itnim will be
	 * freed.
	 */
	bfa_q_qe_deq(ioim);
	bfa_q_enq(&ioim->fcpim->ioim_comp_q, ioim);

	if (!ioim->iosp->tskim)
		bfa_hal_itnim_iodone(ioim->itnim);
	else
		bfa_hal_tskim_iodone(ioim->iosp->tskim);
}

/**
 * @}
 */

/**
 * @ig hal_ioim_friend
 * @{
 */

/**
 * @brief
 * Memory allocation and initialization.
 */
void
bfa_hal_ioim_attach(bfa_hal_fcpim_mod_t *fcpim, bfa_hal_meminfo_t *minfo)
{
	bfa_hal_ioim_t		*ioim;
	bfa_hal_ioim_sp_t	*iosp;
	uint16_t		i;
	uint8_t			*snsinfo;
	uint32_t		snsbufsz;

	/**
	 * claim memory first
	 */
	ioim = (bfa_hal_ioim_t *) bfa_hal_meminfo_kva(minfo);
	fcpim->ioim_arr = ioim;
	bfa_hal_meminfo_kva(minfo) = (uint8_t *) (ioim + fcpim->num_ioim_reqs);

	iosp = (bfa_hal_ioim_sp_t *) bfa_hal_meminfo_kva(minfo);
	fcpim->ioim_sp_arr = iosp;
	bfa_hal_meminfo_kva(minfo) = (uint8_t *) (iosp + fcpim->num_ioim_reqs);

	/**
	 * Claim DMA memory for per IO sense data.
	 */
	snsbufsz = fcpim->num_ioim_reqs * BFA_FWI_IOIM_SNSLEN;
	fcpim->snsbase.pa  = bfa_hal_meminfo_dma_phys(minfo);
	bfa_hal_meminfo_dma_phys(minfo) += snsbufsz;

	fcpim->snsbase.kva = bfa_hal_meminfo_dma_virt(minfo);
	bfa_hal_meminfo_dma_virt(minfo) += snsbufsz;
	snsinfo = fcpim->snsbase.kva;
	bfa_hal_ioc_set_snsbase(fcpim->hal, fcpim->snsbase.pa);

	/**
	 * Initialize ioim free queues
	 */
	bfa_q_init(&fcpim->ioim_free_q);
	bfa_q_init(&fcpim->ioim_resfree_q);
	bfa_q_init(&fcpim->ioim_comp_q);

	for (i = 0; i < fcpim->num_ioim_reqs;
	     i++, ioim++, iosp++, snsinfo += BFA_FWI_IOIM_SNSLEN) {
		/*
		 * initialize IOIM
		 */
		bfa_os_memset(ioim, 0, sizeof(bfa_hal_ioim_t));
		ioim->iotag   = i;
		ioim->hal     = fcpim->hal;
		ioim->fcpim   = fcpim;
		ioim->iosp    = iosp;
		iosp->snsinfo = snsinfo;
		bfa_q_init(&ioim->sgpg_q);
		bfa_hal_reqq_winit(&ioim->iosp->reqq_wait,
				   bfa_hal_ioim_qresume, ioim);
		bfa_hal_sgpg_winit(&ioim->iosp->sgpg_wqe,
				   bfa_hal_ioim_sgpg_alloced, ioim);
		bfa_sm_set_state(ioim, bfa_hal_ioim_sm_uninit);

		bfa_q_enq(&fcpim->ioim_free_q, ioim);
	}
}

/**
 * @brief
 * Driver detach time call.
 */
void
bfa_hal_ioim_detach(bfa_hal_fcpim_mod_t *fcpim)
{
}

void
bfa_hal_ioim_isr(bfa_hal_t *hal, bfa_msg_t *m)
{
	bfa_hal_fcpim_mod_t *fcpim = BFA_HAL_FCPIM_MOD(hal);
	bfa_fwi_ioim_rsp_t *rsp = (bfa_fwi_ioim_rsp_t *) m;
	bfa_hal_ioim_t *ioim;
	uint16_t        iotag;
	bfa_hal_ioim_event_t evt = BFA_HAL_IOIM_SM_COMP;

	iotag = bfa_os_ntohs(rsp->rsp_info.io_tag);

	ioim = BFA_HAL_IOIM_FROM_TAG(fcpim, iotag);
	bfa_assert(ioim->iotag == iotag);

	bfa_trc(ioim->hal, ioim->iotag);
	bfa_trc(ioim->hal, rsp->rsp_info.io_status);
	bfa_trc(ioim->hal, rsp->rsp_info.reuse_io_tag);

	ioim->iosp->comp_rspmsg = *m;

	switch (rsp->rsp_info.io_status) {
	case BFA_IOIM_STS_OK:
		bfa_hal_fcpim_stats(fcpim, iocomp_ok);
		bfa_assert(rsp->rsp_info.reuse_io_tag);
		evt = BFA_HAL_IOIM_SM_COMP;
		break;

	case BFA_IOIM_STS_IO_TIMEDOUT:
	case BFA_IOIM_STS_IO_ABORTED:
		rsp->rsp_info.io_status = BFA_IOIM_STS_IO_ABORTED;
		bfa_hal_fcpim_stats(fcpim, iocomp_aborted);
		if (rsp->rsp_info.reuse_io_tag == 0)
			evt = BFA_HAL_IOIM_SM_DONE;
		else
			evt = BFA_HAL_IOIM_SM_COMP;
		break;

	case BFA_IOIM_STS_IO_PROTO_ERR:
		bfa_hal_fcpim_stats(fcpim, iocom_proto_err);
		bfa_assert(rsp->rsp_info.reuse_io_tag);
		evt = BFA_HAL_IOIM_SM_COMP;
		break;

	case BFA_IOIM_STS_IO_DIF_ERR:
		bfa_hal_fcpim_stats(fcpim, iocom_dif_err);
		evt = BFA_HAL_IOIM_SM_COMP;
		break;

	case BFA_IOIM_STS_IO_SQER_NEEDED:
		bfa_hal_fcpim_stats(fcpim, iocom_sqer_needed);
		bfa_assert(rsp->rsp_info.reuse_io_tag == 0);
		evt = BFA_HAL_IOIM_SM_SQRETRY;
		break;

	case BFA_IOIM_STS_IO_RES_FREE:
		bfa_hal_fcpim_stats(fcpim, iocom_res_free);
		evt = BFA_HAL_IOIM_SM_FREE;
		break;

	case BFA_IOIM_STS_IO_HOST_ABORTED:
		bfa_hal_fcpim_stats(fcpim, iocom_hostabrts);
		if (rsp->rsp_info.abort_tag != ioim->abort_tag) {
			bfa_trc(ioim->hal, rsp->rsp_info.abort_tag);
			bfa_trc(ioim->hal, ioim->abort_tag);
			return;
		}

		if (rsp->rsp_info.reuse_io_tag)
			evt = BFA_HAL_IOIM_SM_ABORT_COMP;
		else
			evt = BFA_HAL_IOIM_SM_ABORT_DONE;
		break;

	case BFA_IOIM_STS_UNKNOWN_TAG:
		bfa_hal_fcpim_stats(fcpim, iocom_utags);
		evt = BFA_HAL_IOIM_SM_COMP_UTAG;
		break;

	default:
		bfa_assert(0);
	}

	bfa_sm_send_event(ioim, evt);
}

void
bfa_hal_ioim_good_comp_isr(bfa_hal_t *hal, bfa_msg_t *m)
{
	bfa_hal_fcpim_mod_t *fcpim = BFA_HAL_FCPIM_MOD(hal);
	bfa_fwi_ioim_rsp_t *rsp = (bfa_fwi_ioim_rsp_t *) m;
	bfa_hal_ioim_t *ioim;
	uint16_t        iotag;

	iotag = bfa_os_ntohs(rsp->rsp_info.io_tag);

	ioim = BFA_HAL_IOIM_FROM_TAG(fcpim, iotag);
	bfa_assert(ioim->iotag == iotag);

	bfa_trc_fp(ioim->hal, ioim->iotag);
	bfa_sm_send_event(ioim, BFA_HAL_IOIM_SM_COMP_GOOD);
}

/**
 * @brief
 * Called by itnim to clean up IO while going offline.
 */
void
bfa_hal_ioim_cleanup(bfa_hal_ioim_t *ioim)
{
	bfa_trc(ioim->hal, ioim->iotag);
	bfa_hal_fcpim_stats(ioim->fcpim, io_cleanups);

	ioim->iosp->tskim = NULL;
	bfa_sm_send_event(ioim, BFA_HAL_IOIM_SM_CLEANUP);
}

void
bfa_hal_ioim_cleanup_tm(bfa_hal_ioim_t *ioim, bfa_hal_tskim_t *tskim)
{
	bfa_trc(ioim->hal, ioim->iotag);
	bfa_hal_fcpim_stats(ioim->fcpim, io_tmaborts);

	ioim->iosp->tskim = tskim;
	bfa_sm_send_event(ioim, BFA_HAL_IOIM_SM_CLEANUP);
}

/**
 * @brief
 * IOC failure handling.
 */
void
bfa_hal_ioim_iocdisable(bfa_hal_ioim_t *ioim)
{
	bfa_sm_send_event(ioim, BFA_HAL_IOIM_SM_HWFAIL);
}

/**
 * @brief
 * IO offline TOV popped. Fail the pending IO.
 */
void
bfa_hal_ioim_tov(bfa_hal_ioim_t *ioim)
{
	bfa_sm_send_event(ioim, BFA_HAL_IOIM_SM_IOTOV);
}

/**
 * @}
 */

/**
 * @ig hal_ioim_api
 * @{
 */

/**
 * @brief
 * Allocate IOIM resource for initiator mode I/O request.
 */
bfa_hal_ioim_t *
bfa_hal_ioim_alloc(bfa_hal_t *hal, bfad_ioim_t *dio, bfa_hal_itnim_t *itnim,
		   uint16_t nsges)
{
	bfa_hal_fcpim_mod_t *fcpim = BFA_HAL_FCPIM_MOD(hal);
	bfa_hal_ioim_t *ioim;

	/**
	 * alocate IOIM resource
	 */
	bfa_q_deq(&fcpim->ioim_free_q, &ioim);
	if (!ioim) {
		bfa_hal_fcpim_stats(fcpim, no_iotags);
		return NULL;
	}

	ioim->dio = dio;
	ioim->itnim = itnim;
	ioim->nsges = nsges;
	ioim->nsgpgs = 0;

	bfa_stats(fcpim, total_ios);
	bfa_stats(itnim, ios);
	fcpim->ios_active ++;

	bfa_q_enq(&itnim->io_q, ioim);
	bfa_trc_fp(ioim->hal, ioim->iotag);

	return ioim;
}

void
bfa_hal_ioim_free(bfa_hal_ioim_t *ioim)
{
	bfa_hal_fcpim_mod_t *fcpim = ioim->fcpim;

	bfa_trc_fp(ioim->hal, ioim->iotag);
	bfa_assert_fp(bfa_sm_cmp_state(ioim, bfa_hal_ioim_sm_uninit));

	bfa_assert_fp(bfa_q_is_empty(&ioim->sgpg_q)
		   || (ioim->nsges > BFA_SGE_DATA_MAX_INLINE));

	if (ioim->nsgpgs > 0)
		bfa_hal_sgpg_mfree(ioim->hal, &ioim->sgpg_q, ioim->nsgpgs);

	bfa_stats(ioim->itnim, io_comps);
	fcpim->ios_active --;

	bfa_q_qe_deq(ioim);
	bfa_q_enq(&fcpim->ioim_free_q, ioim);
}

void
bfa_hal_ioim_start(bfa_hal_ioim_t *ioim)
{
	bfa_trc_fp(ioim->hal, ioim->iotag);
	bfa_sm_send_event(ioim, BFA_HAL_IOIM_SM_START);
}

/**
 * @brief
 * Driver I/O abort request.
 */
void
bfa_hal_ioim_abort(bfa_hal_ioim_t *ioim)
{
	bfa_trc(ioim->hal, ioim->iotag);
	bfa_hal_fcpim_stats(ioim->fcpim, io_aborts);
	bfa_sm_send_event(ioim, BFA_HAL_IOIM_SM_ABORT);
}

/**
 * @}
 */
