/*
 * Copyright (c) 2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

#include <hal/bfa_hal_pport.h>
#include <hal/bfa_hal_diag.h>
#include <hal/bfa_hal_wwn.h>
#include <fwi/bfa_fwi_pport.h>
#include <cs/bfa_debug.h>
#include <aen/bfa_aen.h>
#include <cs/bfa_plog.h>
#include <aen/bfal_aen_port.h>

BFA_TRC_FILE(HAL, PPORT);
BFA_HAL_MODULE(pport);

#define bfa_hal_pport_callback(__pport, __event) do {			\
	if ((__pport)->hal->fcs) {					\
		(__pport)->event_cbfn((__pport)->event_cbarg, (__event)); \
	} else {							\
		(__pport)->hcb_event = (__event);			\
		bfa_hcb_queue((__pport)->hal, &(__pport)->hcb_qe,	\
		__bfa_hcb_port_event, (__pport));			\
	}								\
} while (0)

/*
 * forward declarations
 */
static bfa_boolean_t bfa_hal_pport_send_enable(bfa_hal_pport_t *port);
static bfa_boolean_t bfa_hal_pport_send_disable(bfa_hal_pport_t *port);
static void	bfa_hal_pport_update_linkinfo(bfa_hal_pport_t *pport);
static void	bfa_hal_pport_reset_linkinfo(bfa_hal_pport_t *pport);
static void	bfa_hal_pport_set_wwns(bfa_hal_pport_t *port);
static void	__bfa_hcb_port_event(void *cbarg, bfa_boolean_t complete);
static void	__bfa_hcb_port_stats(void *cbarg, bfa_boolean_t complete);
static void	__bfa_hcb_port_stats_clr(void *cbarg, bfa_boolean_t complete);
static void	bfa_hal_port_stats_timeout(void *cbarg);
static void	bfa_hal_port_stats_clr_timeout(void *cbarg);

/**
 * @dg hal_pport_private
 * @{
 */

/**
 * HAL port state machine events
 */
typedef enum {
	BFA_HAL_PPORT_SM_START	= 1,	/* !< start port state machine	*/
	BFA_HAL_PPORT_SM_STOP	= 2,	/* !< stop port state machine	*/
	BFA_HAL_PPORT_SM_ENABLE	= 3,	/* !< enable port		*/
	BFA_HAL_PPORT_SM_DISABLE = 4,	/* !< disable port state machine */
	BFA_HAL_PPORT_SM_FWRSP	= 5,	/* !< firmware enable/disable rsp */
	BFA_HAL_PPORT_SM_LINKUP	= 6,	/* !< firmware linkup event	*/
	BFA_HAL_PPORT_SM_LINKDOWN = 7,	/* !< firmware linkup down	*/
	BFA_HAL_PPORT_SM_QRESUME = 8,	/* !< CQ space available	*/
	BFA_HAL_PPORT_SM_HWFAIL = 9,	/* !< IOC h/w failure		*/
} bfa_hal_pport_sm_event_t;

static void	bfa_hal_pport_sm_uninit(bfa_hal_pport_t *pport,
					bfa_hal_pport_sm_event_t event);
static void	bfa_hal_pport_sm_enabling_qwait(bfa_hal_pport_t *pport,
						bfa_hal_pport_sm_event_t event);
static void	bfa_hal_pport_sm_enabling(bfa_hal_pport_t *pport,
						bfa_hal_pport_sm_event_t event);
static void	bfa_hal_pport_sm_linkdown(bfa_hal_pport_t *pport,
						bfa_hal_pport_sm_event_t event);
static void	bfa_hal_pport_sm_linkup(bfa_hal_pport_t *pport,
						bfa_hal_pport_sm_event_t event);
static void	bfa_hal_pport_sm_disabling(bfa_hal_pport_t *pport,
						bfa_hal_pport_sm_event_t event);
static void	bfa_hal_pport_sm_disabling_qwait(bfa_hal_pport_t *pport,
						 bfa_hal_pport_sm_event_t
						 event);
static void	bfa_hal_pport_sm_disabled(bfa_hal_pport_t *pport,
						bfa_hal_pport_sm_event_t event);
static void	bfa_hal_pport_sm_stopped(bfa_hal_pport_t *pport,
					 bfa_hal_pport_sm_event_t event);
static void	bfa_hal_pport_sm_iocdown(bfa_hal_pport_t *pport,
					 bfa_hal_pport_sm_event_t event);
static void	bfa_hal_pport_sm_iocfail(bfa_hal_pport_t *pport,
					 bfa_hal_pport_sm_event_t event);

static bfa_sm_table_t hal_pport_sm_table[] = {
	{BFA_SM(bfa_hal_pport_sm_uninit), BFA_HAL_PPORT_UNINIT},
	{BFA_SM(bfa_hal_pport_sm_enabling_qwait), BFA_HAL_PPORT_ENABLING_QWAIT},
	{BFA_SM(bfa_hal_pport_sm_enabling), BFA_HAL_PPORT_ENABLING},
	{BFA_SM(bfa_hal_pport_sm_linkdown), BFA_HAL_PPORT_LINKDOWN},
	{BFA_SM(bfa_hal_pport_sm_linkup), BFA_HAL_PPORT_LINKUP},
	{BFA_SM(bfa_hal_pport_sm_disabling_qwait),
						BFA_HAL_PPORT_DISABLING_QWAIT},
	{BFA_SM(bfa_hal_pport_sm_disabling), BFA_HAL_PPORT_DISABLING},
	{BFA_SM(bfa_hal_pport_sm_disabled), BFA_HAL_PPORT_DISABLED},
	{BFA_SM(bfa_hal_pport_sm_stopped), BFA_HAL_PPORT_STOPPED},
	{BFA_SM(bfa_hal_pport_sm_iocdown), BFA_HAL_PPORT_IOCDOWN},
	{BFA_SM(bfa_hal_pport_sm_iocfail), BFA_HAL_PPORT_IOCDOWN},
};

static void
bfa_hal_pport_aen_post(bfa_hal_pport_t *pport, bfa_port_aen_event_t event)
{
	bfa_aen_data_t	aen_data;

	bfa_log_mod_t	*logmod = pport->hal->logm;
	wwn_t		pwwn = pport->pwwn;
	char		pwwn_buf[BFA_STRING_32];
	char		*pwwn_ptr;

        pwwn_ptr = wwn2str(pwwn_buf, sizeof(pwwn_buf), pwwn);

	switch (event) {
	case BFA_PORT_AEN_ONLINE:
		bfa_log(logmod, BFAL_AEN_PORT_ONLINE, pwwn_ptr);
		break;
	case BFA_PORT_AEN_OFFLINE:
		bfa_log(logmod, BFAL_AEN_PORT_OFFLINE, pwwn_ptr);
		break;
	case BFA_PORT_AEN_ENABLE:
		bfa_log(logmod, BFAL_AEN_PORT_ENABLE, pwwn_ptr);
		break;
	case BFA_PORT_AEN_DISABLE:
		bfa_log(logmod, BFAL_AEN_PORT_DISABLE, pwwn_ptr);
		break;
	case BFA_PORT_AEN_DISCONNECT:
		bfa_log(logmod, BFAL_AEN_PORT_DISCONNECT, pwwn_ptr);
		break;
	default: 
		break;
	}

	aen_data.port.pwwn = pwwn;
	bfa_aen_post(pport->hal->aen, BFA_AEN_CAT_PORT, event, &aen_data);
}

static void
bfa_hal_pport_sm_uninit(bfa_hal_pport_t *pport, bfa_hal_pport_sm_event_t event)
{
	bfa_trc(pport->hal, event);

	switch (event) {
	case BFA_HAL_PPORT_SM_START:
		/**
		 * Start event after IOC is configured and HAL is started.
		 */
		if (bfa_hal_pport_send_enable(pport))
			bfa_sm_set_state(pport, bfa_hal_pport_sm_enabling);
		else
			bfa_sm_set_state(pport,
					 bfa_hal_pport_sm_enabling_qwait);
		break;

	case BFA_HAL_PPORT_SM_ENABLE:
		/**
		 * Port is persistently configured to be in enabled state. Do
		 * not change state. Port enabling is done when START event is
		 * received.
		 */
		break;

	case BFA_HAL_PPORT_SM_DISABLE:
		/**
		 * If a port is persistently configured to be disabled, the
		 * first event will a port disable request.
		 */
		bfa_sm_set_state(pport, bfa_hal_pport_sm_disabled);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_hal_pport_sm_enabling_qwait(bfa_hal_pport_t *pport,
				bfa_hal_pport_sm_event_t event)
{
	bfa_trc(pport->hal, event);

	switch (event) {
	case BFA_HAL_PPORT_SM_QRESUME:
		bfa_sm_set_state(pport, bfa_hal_pport_sm_enabling);
		bfa_hal_pport_send_enable(pport);
		break;

	case BFA_HAL_PPORT_SM_STOP:
		bfa_hal_reqq_wcancel(&pport->reqq_wait);
		bfa_sm_set_state(pport, bfa_hal_pport_sm_stopped);
		break;

	case BFA_HAL_PPORT_SM_ENABLE:
		/**
		 * Already enable is in progress.
		 */
		break;

	case BFA_HAL_PPORT_SM_DISABLE:
		/**
		 * Just send disable request to firmware when room becomes
		 * available in request queue.
		 */
		bfa_sm_set_state(pport, bfa_hal_pport_sm_disabling_qwait);
		bfa_plog_str(pport->hal->plog, BFA_PL_MID_HAL,
				BFA_PL_EID_PORT_DISABLE, 0, "Port Disable");
		bfa_hal_pport_aen_post(pport, BFA_PORT_AEN_DISABLE);
		break;

	case BFA_HAL_PPORT_SM_LINKUP:
	case BFA_HAL_PPORT_SM_LINKDOWN:
		/**
		 * Possible to get link events when doing back-to-back
		 * enable/disables.
		 */
		break;

	case BFA_HAL_PPORT_SM_HWFAIL:
		bfa_hal_reqq_wcancel(&pport->reqq_wait);
		bfa_sm_set_state(pport, bfa_hal_pport_sm_iocdown);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_hal_pport_sm_enabling(bfa_hal_pport_t *pport,
						bfa_hal_pport_sm_event_t event)
{
	bfa_trc(pport->hal, event);

	switch (event) {
	case BFA_HAL_PPORT_SM_FWRSP:
	case BFA_HAL_PPORT_SM_LINKDOWN:
		bfa_sm_set_state(pport, bfa_hal_pport_sm_linkdown);
		break;

	case BFA_HAL_PPORT_SM_LINKUP:
		bfa_hal_pport_update_linkinfo(pport);
		bfa_sm_set_state(pport, bfa_hal_pport_sm_linkup);

		bfa_assert(pport->event_cbfn);
		bfa_hal_pport_callback(pport, BFA_PPORT_LINKUP);
		break;

	case BFA_HAL_PPORT_SM_ENABLE:
		/**
		 * Already being enabled.
		 */
		break;

	case BFA_HAL_PPORT_SM_DISABLE:
		if (bfa_hal_pport_send_disable(pport))
			bfa_sm_set_state(pport, bfa_hal_pport_sm_disabling);
		else
			bfa_sm_set_state(pport,
					 bfa_hal_pport_sm_disabling_qwait);

		bfa_plog_str(pport->hal->plog, BFA_PL_MID_HAL,
				BFA_PL_EID_PORT_DISABLE, 0, "Port Disable");
		bfa_hal_pport_aen_post(pport, BFA_PORT_AEN_DISABLE);
		break;

	case BFA_HAL_PPORT_SM_STOP:
		bfa_sm_set_state(pport, bfa_hal_pport_sm_stopped);
		break;

	case BFA_HAL_PPORT_SM_HWFAIL:
		bfa_sm_set_state(pport, bfa_hal_pport_sm_iocdown);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_hal_pport_sm_linkdown(bfa_hal_pport_t *pport,
						bfa_hal_pport_sm_event_t event)
{
	bfa_trc(pport->hal, event);

	switch (event) {
	case BFA_HAL_PPORT_SM_LINKUP:
		bfa_hal_pport_update_linkinfo(pport);
		bfa_sm_set_state(pport, bfa_hal_pport_sm_linkup);
		bfa_assert(pport->event_cbfn);
		bfa_plog_str(pport->hal->plog, BFA_PL_MID_HAL,
				BFA_PL_EID_PORT_ST_CHANGE, 0, "Port Linkup");
		bfa_hal_pport_callback(pport, BFA_PPORT_LINKUP);
		bfa_hal_pport_aen_post(pport, BFA_PORT_AEN_ONLINE);
		break;

	case BFA_HAL_PPORT_SM_ENABLE:
		/**
		 * Already enabled.
		 */
		break;

	case BFA_HAL_PPORT_SM_DISABLE:
		if (bfa_hal_pport_send_disable(pport))
			bfa_sm_set_state(pport, bfa_hal_pport_sm_disabling);
		else
			bfa_sm_set_state(pport,
					 bfa_hal_pport_sm_disabling_qwait);

		bfa_plog_str(pport->hal->plog, BFA_PL_MID_HAL,
				BFA_PL_EID_PORT_DISABLE, 0, "Port Disable");
		bfa_hal_pport_aen_post(pport, BFA_PORT_AEN_DISABLE);
		break;

	case BFA_HAL_PPORT_SM_STOP:
		bfa_sm_set_state(pport, bfa_hal_pport_sm_stopped);
		break;

	case BFA_HAL_PPORT_SM_HWFAIL:
		bfa_sm_set_state(pport, bfa_hal_pport_sm_iocdown);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_hal_pport_sm_linkup(bfa_hal_pport_t *pport, bfa_hal_pport_sm_event_t event)
{
	bfa_trc(pport->hal, event);

	switch (event) {
	case BFA_HAL_PPORT_SM_ENABLE:
		/**
		 * Already enabled.
		 */
		break;

	case BFA_HAL_PPORT_SM_DISABLE:
		if (bfa_hal_pport_send_disable(pport))
			bfa_sm_set_state(pport, bfa_hal_pport_sm_disabling);
		else
			bfa_sm_set_state(pport,
					 bfa_hal_pport_sm_disabling_qwait);

		bfa_hal_pport_reset_linkinfo(pport);
		bfa_hal_pport_callback(pport, BFA_PPORT_LINKDOWN);
		bfa_plog_str(pport->hal->plog, BFA_PL_MID_HAL,
				BFA_PL_EID_PORT_DISABLE, 0, "Port Disable");
		bfa_hal_pport_aen_post(pport, BFA_PORT_AEN_DISABLE);
		break;

	case BFA_HAL_PPORT_SM_LINKDOWN:
		bfa_sm_set_state(pport, bfa_hal_pport_sm_linkdown);
		bfa_hal_pport_reset_linkinfo(pport);
		bfa_hal_pport_callback(pport, BFA_PPORT_LINKDOWN);
		bfa_plog_str(pport->hal->plog, BFA_PL_MID_HAL,
				BFA_PL_EID_PORT_ST_CHANGE, 0, "Port Linkdown");
		if (bfa_hal_pport_is_disabled(pport->hal) == BFA_TRUE) {
			bfa_hal_pport_aen_post(pport, BFA_PORT_AEN_OFFLINE);
		} else {
			bfa_hal_pport_aen_post(pport, BFA_PORT_AEN_DISCONNECT);
		}
		break;

	case BFA_HAL_PPORT_SM_STOP:
		bfa_sm_set_state(pport, bfa_hal_pport_sm_stopped);
		bfa_hal_pport_reset_linkinfo(pport);
		if (bfa_hal_pport_is_disabled(pport->hal) == BFA_TRUE) {
			bfa_hal_pport_aen_post(pport, BFA_PORT_AEN_OFFLINE);
		} else {
			bfa_hal_pport_aen_post(pport, BFA_PORT_AEN_DISCONNECT);
		}
		break;

	case BFA_HAL_PPORT_SM_HWFAIL:
		bfa_sm_set_state(pport, bfa_hal_pport_sm_iocdown);
		bfa_hal_pport_reset_linkinfo(pport);
		bfa_hal_pport_callback(pport, BFA_PPORT_LINKDOWN);
		if (bfa_hal_pport_is_disabled(pport->hal) == BFA_TRUE) {
			bfa_hal_pport_aen_post(pport, BFA_PORT_AEN_OFFLINE);
		} else {
			bfa_hal_pport_aen_post(pport, BFA_PORT_AEN_DISCONNECT);
		}
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_hal_pport_sm_disabling_qwait(bfa_hal_pport_t *pport,
				 bfa_hal_pport_sm_event_t event)
{
	bfa_trc(pport->hal, event);

	switch (event) {
	case BFA_HAL_PPORT_SM_QRESUME:
		bfa_sm_set_state(pport, bfa_hal_pport_sm_disabling);
		bfa_hal_pport_send_disable(pport);
		break;

	case BFA_HAL_PPORT_SM_STOP:
		bfa_sm_set_state(pport, bfa_hal_pport_sm_stopped);
		bfa_hal_reqq_wcancel(&pport->reqq_wait);
		break;

	case BFA_HAL_PPORT_SM_ENABLE:
		/**
		 * Just send enable request to firmware when room becomes
		 * available in request queue.
		 */
		bfa_sm_set_state(pport, bfa_hal_pport_sm_enabling_qwait);
		bfa_plog_str(pport->hal->plog, BFA_PL_MID_HAL,
				BFA_PL_EID_PORT_ENABLE, 0, "Port Enable");
		bfa_hal_pport_aen_post(pport, BFA_PORT_AEN_ENABLE);
		break;

	case BFA_HAL_PPORT_SM_DISABLE:
		/**
		 * Already being disabled.
		 */
		bfa_sm_set_state(pport, bfa_hal_pport_sm_disabling_qwait);
		break;

	case BFA_HAL_PPORT_SM_LINKUP:
	case BFA_HAL_PPORT_SM_LINKDOWN:
		/**
		 * Possible to get link events when doing back-to-back
		 * enable/disables.
		 */
		break;

	case BFA_HAL_PPORT_SM_HWFAIL:
		bfa_sm_set_state(pport, bfa_hal_pport_sm_iocfail);
		bfa_hal_reqq_wcancel(&pport->reqq_wait);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_hal_pport_sm_disabling(bfa_hal_pport_t *pport,
						bfa_hal_pport_sm_event_t event)
{
	bfa_trc(pport->hal, event);

	switch (event) {
	case BFA_HAL_PPORT_SM_FWRSP:
		bfa_sm_set_state(pport, bfa_hal_pport_sm_disabled);
		break;

	case BFA_HAL_PPORT_SM_DISABLE:
		/**
		 * Already being disabled.
		 */
		break;

	case BFA_HAL_PPORT_SM_ENABLE:
		if (bfa_hal_pport_send_enable(pport))
			bfa_sm_set_state(pport, bfa_hal_pport_sm_enabling);
		else
			bfa_sm_set_state(pport,
					 bfa_hal_pport_sm_enabling_qwait);

		bfa_plog_str(pport->hal->plog, BFA_PL_MID_HAL,
				BFA_PL_EID_PORT_ENABLE, 0, "Port Enable");
		bfa_hal_pport_aen_post(pport, BFA_PORT_AEN_ENABLE);
		break;

	case BFA_HAL_PPORT_SM_STOP:
		bfa_sm_set_state(pport, bfa_hal_pport_sm_stopped);
		break;

	case BFA_HAL_PPORT_SM_LINKUP:
	case BFA_HAL_PPORT_SM_LINKDOWN:
		/**
		 * Possible to get link events when doing back-to-back
		 * enable/disables.
		 */
		break;

	case BFA_HAL_PPORT_SM_HWFAIL:
		bfa_sm_set_state(pport, bfa_hal_pport_sm_iocfail);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_hal_pport_sm_disabled(bfa_hal_pport_t *pport,
						bfa_hal_pport_sm_event_t event)
{
	bfa_trc(pport->hal, event);

	switch (event) {
	case BFA_HAL_PPORT_SM_START:
		/**
		 * Ignore start event for a port that is disabled.
		 */
		break;

	case BFA_HAL_PPORT_SM_STOP:
		bfa_sm_set_state(pport, bfa_hal_pport_sm_stopped);
		break;

	case BFA_HAL_PPORT_SM_ENABLE:
		if (bfa_hal_pport_send_enable(pport))
			bfa_sm_set_state(pport, bfa_hal_pport_sm_enabling);
		else
			bfa_sm_set_state(pport,
					 bfa_hal_pport_sm_enabling_qwait);

		bfa_plog_str(pport->hal->plog, BFA_PL_MID_HAL,
				BFA_PL_EID_PORT_ENABLE, 0, "Port Enable");
		bfa_hal_pport_aen_post(pport, BFA_PORT_AEN_ENABLE);
		break;

	case BFA_HAL_PPORT_SM_DISABLE:
		/**
		 * Already disabled.
		 */
		break;

	case BFA_HAL_PPORT_SM_HWFAIL:
		bfa_sm_set_state(pport, bfa_hal_pport_sm_iocfail);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_hal_pport_sm_stopped(bfa_hal_pport_t *pport,
			 bfa_hal_pport_sm_event_t event)
{
	bfa_trc(pport->hal, event);

	switch (event) {
	case BFA_HAL_PPORT_SM_START:
		if (bfa_hal_pport_send_enable(pport))
			bfa_sm_set_state(pport, bfa_hal_pport_sm_enabling);
		else
			bfa_sm_set_state(pport,
					 bfa_hal_pport_sm_enabling_qwait);
		break;

	default:
		/**
		 * Ignore all other events.
		 */
		;
	}
}

/**
 * @brief
 * Port is enabled. IOC is down/failed.
 */
static void
bfa_hal_pport_sm_iocdown(bfa_hal_pport_t *pport,
			 bfa_hal_pport_sm_event_t event)
{
	bfa_trc(pport->hal, event);

	switch (event) {
	case BFA_HAL_PPORT_SM_START:
		if (bfa_hal_pport_send_enable(pport))
			bfa_sm_set_state(pport, bfa_hal_pport_sm_enabling);
		else
			bfa_sm_set_state(pport,
					 bfa_hal_pport_sm_enabling_qwait);
		break;

	default:
		/**
		 * Ignore all events.
		 */
		;
	}
}

/**
 * @brief
 * Port is disabled. IOC is down/failed.
 */
static void
bfa_hal_pport_sm_iocfail(bfa_hal_pport_t *pport,
			 bfa_hal_pport_sm_event_t event)
{
	bfa_trc(pport->hal, event);

	switch (event) {
	case BFA_HAL_PPORT_SM_START:
		bfa_sm_set_state(pport, bfa_hal_pport_sm_disabled);
		break;

	case BFA_HAL_PPORT_SM_ENABLE:
		bfa_sm_set_state(pport, bfa_hal_pport_sm_iocdown);
		break;

	default:
		/**
		 * Ignore all events.
		 */
		;
	}
}

/**
 * @}
 */

/**
 * @dg hal_pport_private
 * @{
 */

static void
__bfa_hcb_port_event(void *cbarg, bfa_boolean_t complete)
{
	bfa_hal_pport_t *pport = cbarg;

	if (complete)
		pport->event_cbfn(pport->event_cbarg, pport->hcb_event);
}

#define PPORT_STATS_DMA_SZ (BFA_ROUNDUP(sizeof(bfa_pport_stats_t), 	\
							BFA_HAL_CACHELINE_SZ))

static void
bfa_hal_pport_meminfo(bfa_ioc_cfg_t *cfg,
				uint32_t *ndm_len,
				uint32_t *dm_len)
{
	*dm_len += PPORT_STATS_DMA_SZ;
}

static void
bfa_hal_pport_qresume(void *cbarg)
{
	bfa_hal_pport_t *port = cbarg;

	bfa_sm_send_event(port, BFA_HAL_PPORT_SM_QRESUME);
}

static void
bfa_hal_pport_mem_claim(bfa_hal_pport_t *pport, bfa_hal_meminfo_t *meminfo)
{
	uint8_t		*dm_kva;
	uint64_t	dm_pa;

	dm_kva = bfa_hal_meminfo_dma_virt(meminfo);
	dm_pa = bfa_hal_meminfo_dma_phys(meminfo);

	pport->stats_kva = dm_kva;
	pport->stats_pa = dm_pa;
	pport->stats = (bfa_pport_stats_t *) dm_kva;

	dm_kva += PPORT_STATS_DMA_SZ;
	dm_pa += PPORT_STATS_DMA_SZ;

	bfa_hal_meminfo_dma_virt(meminfo) = dm_kva;
	bfa_hal_meminfo_dma_phys(meminfo) = dm_pa;
}

/**
 * @brief
 * Memory initialization.
 */
static void
bfa_hal_pport_attach(bfa_hal_t *hal, void *bfad, bfa_ioc_cfg_t *cfg,
			bfa_hal_meminfo_t *meminfo, bfa_hal_pcidev_t *pcidev)
{
	bfa_hal_pport_t *pport = BFA_HAL_PORT_MOD(hal);
	bfa_pport_cfg_t *port_cfg = &pport->cfg;

	bfa_os_memset(pport, 0, sizeof(bfa_hal_pport_t));
	pport->hal = hal;

	bfa_hal_pport_mem_claim(pport, meminfo);

	bfa_sm_set_state(pport, bfa_hal_pport_sm_uninit);

	/**
	 * initialize and set default configuration
	 */
	port_cfg->topology = BFA_PPORT_TOPOLOGY_P2P;
	port_cfg->speed = BFA_PPORT_SPEED_AUTO;
	port_cfg->trunked = BFA_FALSE;

	bfa_hal_reqq_winit(&pport->reqq_wait, bfa_hal_pport_qresume, pport);
}

static void
bfa_hal_pport_initdone(bfa_hal_t *hal)
{
	bfa_hal_pport_t *pport = BFA_HAL_PORT_MOD(hal);

	/**
	 * Initialize port attributes from IOC hardware data.
	 */
	bfa_hal_pport_set_wwns(pport);
	pport->cfg.maxfrsize = bfa_hal_ioc_maxfrsize(hal);
	pport->cfg.rx_bbcredit = bfa_hal_ioc_rx_bbcredit(hal);
	pport->speed_sup = bfa_hal_ioc_speed_sup(hal);

	bfa_assert(pport->cfg.maxfrsize);
	bfa_assert(pport->cfg.rx_bbcredit);
	bfa_assert(pport->speed_sup);
}

static void
bfa_hal_pport_detach(bfa_hal_t *hal)
{
}

/**
 * @brief
 * Called when IOC is ready.
 */
static void
bfa_hal_pport_start(bfa_hal_t *hal)
{
	bfa_sm_send_event(BFA_HAL_PORT_MOD(hal), BFA_HAL_PPORT_SM_START);
}

/**
 * @brief
 * Called before IOC is stopped.
 */
static void
bfa_hal_pport_stop(bfa_hal_t *hal)
{
	bfa_sm_send_event(BFA_HAL_PORT_MOD(hal), BFA_HAL_PPORT_SM_STOP);
}

/**
 * @brief
 * Called when IOC failure is detected.
 */
static void
bfa_hal_pport_iocdisable(bfa_hal_t *hal)
{
	bfa_sm_send_event(BFA_HAL_PORT_MOD(hal), BFA_HAL_PPORT_SM_HWFAIL);
}

static void
bfa_hal_pport_update_linkinfo(bfa_hal_pport_t *pport)
{
	bfa_fwi_pport_event_t *pevent = pport->event_arg.i2hmsg.event;

	pport->speed = pevent->port_event.link_state.speed;
	pport->topology = pevent->port_event.link_state.topology;

	if (pport->topology == BFA_PPORT_TOPOLOGY_LOOP)
		pport->myalpa =
			pevent->port_event.link_state.tl.loop_info.myalpa;

	bfa_trc(pport->hal, pport->speed);
	bfa_trc(pport->hal, pport->topology);
}

static void
bfa_hal_pport_reset_linkinfo(bfa_hal_pport_t *pport)
{
	pport->speed = BFA_PPORT_SPEED_UNKNOWN;
	pport->topology = BFA_PPORT_TOPOLOGY_NONE;
}

/**
 * @brief
 * Send port enable message to firmware.
 */
static bfa_boolean_t
bfa_hal_pport_send_enable(bfa_hal_pport_t *port)
{
	bfa_fwi_pport_enable_req_t *m;

	/**
	 * Increment message tag before queue check, so that responses to old
	 * requests are discarded.
	 */
	port->msgtag++;

	/**
	 * check for room in queue to send request now
	 */
	m = bfa_hal_reqq_next(port->hal, BFA_HAL_REQQ_PORT);
	if (!m) {
		bfa_hal_reqq_wait(port->hal, BFA_HAL_REQQ_PORT,
							&port->reqq_wait);
		return BFA_FALSE;
	}

	bfa_fwi_h2i_set(m->mh, BFA_MC_PORT, BFA_FWI_PPORT_H2I_ENABLE_REQ,
			bfa_hal_ioc_lpuid(port->hal));
	m->nwwn = port->nwwn;
	m->pwwn = port->pwwn;
	m->port_cfg = port->cfg;
	m->msgtag = port->msgtag;
	m->port_cfg.maxfrsize = bfa_os_htons(port->cfg.maxfrsize);
	bfa_hal_dmabe_addr_set(m->stats_dma_addr, port->stats_pa);
	bfa_trc(port->hal, m->stats_dma_addr.a32.addr_lo);
	bfa_trc(port->hal, m->stats_dma_addr.a32.addr_hi);

	/**
	 * queue I/O message to firmware
	 */
	bfa_hal_reqq_produce(port->hal, BFA_HAL_REQQ_PORT);
	return BFA_TRUE;
}

/**
 * @brief
 * Send port disable message to firmware.
 */
static	bfa_boolean_t
bfa_hal_pport_send_disable(bfa_hal_pport_t *port)
{
	bfa_fwi_pport_disable_req_t *m;

	/**
	 * Increment message tag before queue check, so that responses to old
	 * requests are discarded.
	 */
	port->msgtag++;

	/**
	 * check for room in queue to send request now
	 */
	m = bfa_hal_reqq_next(port->hal, BFA_HAL_REQQ_PORT);
	if (!m) {
		bfa_hal_reqq_wait(port->hal, BFA_HAL_REQQ_PORT,
							&port->reqq_wait);
		return BFA_FALSE;
	}

	bfa_fwi_h2i_set(m->mh, BFA_MC_PORT, BFA_FWI_PPORT_H2I_DISABLE_REQ,
			bfa_hal_ioc_lpuid(port->hal));
	m->msgtag = port->msgtag;

	/**
	 * queue I/O message to firmware
	 */
	bfa_hal_reqq_produce(port->hal, BFA_HAL_REQQ_PORT);

	return BFA_TRUE;
}

static void
bfa_hal_pport_set_wwns(bfa_hal_pport_t *port)
{
	port->pwwn = bfa_hal_wwn_get_naa1(port->hal);
	port->nwwn = bfa_hal_wwn_get_naa2(port->hal);

	bfa_trc(port->hal, port->pwwn);
	bfa_trc(port->hal, port->nwwn);
}

static void
bfa_hal_ppport_send_txcredit(void *port_cbarg)
{

	bfa_hal_pport_t *port = port_cbarg;
	bfa_fwi_pport_set_svc_params_req_t *m;

	/**
	 * check for room in queue to send request now
	 */
	m = bfa_hal_reqq_next(port->hal, BFA_HAL_REQQ_PORT);
	if (!m) {
		bfa_hal_reqq_winit(&port->svcreq_wait,
					bfa_hal_ppport_send_txcredit, port);
		bfa_hal_reqq_wait(port->hal, BFA_HAL_REQQ_PORT,
							&port->svcreq_wait);
		return;
	}

	bfa_fwi_h2i_set(m->mh, BFA_MC_PORT,
			BFA_FWI_PPORT_H2I_SET_SVC_PARAMS_REQ,
			bfa_hal_ioc_lpuid(port->hal));
	m->tx_bbcredit = bfa_os_htons((uint16_t)port->cfg.tx_bbcredit);

	/**
	 * queue I/O message to firmware
	 */
	bfa_hal_reqq_produce(port->hal, BFA_HAL_REQQ_PORT);
}

/**
 * @}
 */

/**
 * @ig hal_pport_public
 * @{
 */

/**
 * @brief
 * Firmware message handler.
 */
void
bfa_hal_pport_isr(bfa_hal_t *hal, bfa_msg_t *msg)
{
	bfa_hal_pport_t *pport = BFA_HAL_PORT_MOD(hal);
	bfa_fwi_pport_i2h_msg_t i2hmsg;

	i2hmsg.msg = msg;
	pport->event_arg.i2hmsg = i2hmsg;

	switch (msg->mhdr.msg_id) {
	case BFA_FWI_PPORT_I2H_ENABLE_RSP:
		if (pport->msgtag == i2hmsg.enable_rsp->msgtag)
			bfa_sm_send_event(pport, BFA_HAL_PPORT_SM_FWRSP);
		break;

	case BFA_FWI_PPORT_I2H_DISABLE_RSP:
		if (pport->msgtag == i2hmsg.enable_rsp->msgtag)
			bfa_sm_send_event(pport, BFA_HAL_PPORT_SM_FWRSP);
		break;

	case BFA_FWI_PPORT_I2H_EVENT:
		switch (i2hmsg.event->port_event.link_state.linkstate) {
		case BFA_PPORT_LINKUP:
			bfa_sm_send_event(pport, BFA_HAL_PPORT_SM_LINKUP);
			break;
		case BFA_PPORT_LINKDOWN:
			bfa_sm_send_event(pport, BFA_HAL_PPORT_SM_LINKDOWN);
			break;
		case BFA_PPORT_TRUNK_LINKDOWN:
			/**@todo: event notification */
			break;
		}
		break;

	case BFA_FWI_PPORT_I2H_GET_STATS_RSP:
		/*
		 * check for timer pop before processing the rsp
		 */
		if (pport->stats_busy == BFA_FALSE
				|| pport->stats_status == BFA_STATUS_ETIMER)
			break;

		bfa_hal_timer_stop(&pport->timer);
		pport->stats_status = BFA_STATUS_OK;
		bfa_hcb_queue(pport->hal, &pport->hcb_qe, __bfa_hcb_port_stats,
									pport);
		break;
	case BFA_FWI_PPORT_I2H_CLEAR_STATS_RSP:
		/*
		 * check for timer pop before processing the rsp
		 */
		if (pport->stats_busy == BFA_FALSE
				|| pport->stats_status == BFA_STATUS_ETIMER)
			break;

		bfa_hal_timer_stop(&pport->timer);
		pport->stats_status = BFA_STATUS_OK;
		bfa_hcb_queue(pport->hal, &pport->hcb_qe,
					__bfa_hcb_port_stats_clr, pport);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @}
 */

/**
 * @ig hal_pport_api
 * @{
 */

/**
 * @brief
 * Registered callback for port events.
 */
void
bfa_hal_pport_event_register(bfa_hal_t *hal,
				void (*cbfn) (void *cbarg,
				bfa_pport_event_t event),
				void *cbarg)
{
	bfa_hal_pport_t *pport = BFA_HAL_PORT_MOD(hal);

	pport->event_cbfn = cbfn;
	pport->event_cbarg = cbarg;
}

bfa_status_t
bfa_hal_pport_enable(bfa_hal_t *hal)
{
	if (bfa_hal_diag_is_busy(hal))
		return (BFA_STATUS_DIAG_BUSY);

	bfa_sm_send_event(BFA_HAL_PORT_MOD(hal), BFA_HAL_PPORT_SM_ENABLE);
	return BFA_STATUS_OK;
}

bfa_status_t
bfa_hal_pport_disable(bfa_hal_t *hal)
{
	bfa_sm_send_event(BFA_HAL_PORT_MOD(hal), BFA_HAL_PPORT_SM_DISABLE);
	return BFA_STATUS_OK;
}

/**
 * @brief
 * Configure port speed.
 */
bfa_status_t
bfa_hal_pport_cfg_speed(bfa_hal_t *hal, bfa_pport_speed_t speed)
{
	bfa_hal_pport_t *pport = BFA_HAL_PORT_MOD(hal);

	bfa_trc(hal, speed);

	if ((speed != BFA_PPORT_SPEED_AUTO) && (speed > pport->speed_sup)) {
		bfa_trc(hal, pport->speed_sup);
		return BFA_STATUS_UNSUPP_SPEED;
	}

	pport->cfg.speed = speed;

	return (BFA_STATUS_OK);
}

/**
 * @brief
 * Get current speed.
 */
bfa_pport_speed_t
bfa_hal_pport_get_speed(bfa_hal_t *hal)
{
	bfa_hal_pport_t *port = BFA_HAL_PORT_MOD(hal);

	return port->speed;
}

/**
 * @brief
 * Configure port topology.
 */
bfa_status_t
bfa_hal_pport_cfg_topology(bfa_hal_t *hal, bfa_pport_topology_t topology)
{
	bfa_hal_pport_t *pport = BFA_HAL_PORT_MOD(hal);

	bfa_trc(hal, topology);
	bfa_trc(hal, pport->cfg.topology);

	switch (topology) {
	case BFA_PPORT_TOPOLOGY_P2P:
	case BFA_PPORT_TOPOLOGY_LOOP:
	case BFA_PPORT_TOPOLOGY_AUTO:
		break;

	default:
		return BFA_STATUS_EINVAL;
	}

	pport->cfg.topology = topology;
	return (BFA_STATUS_OK);
}

/**
 * @brief
 * Get current topology.
 */
bfa_pport_topology_t
bfa_hal_pport_get_topology(bfa_hal_t *hal)
{
	bfa_hal_pport_t *port = BFA_HAL_PORT_MOD(hal);

	return port->topology;
}

bfa_status_t
bfa_hal_pport_cfg_hardalpa(bfa_hal_t *hal, uint8_t alpa)
{
	bfa_hal_pport_t *pport = BFA_HAL_PORT_MOD(hal);

	bfa_trc(hal, alpa);
	bfa_trc(hal, pport->cfg.cfg_hardalpa);
	bfa_trc(hal, pport->cfg.hardalpa);

	pport->cfg.cfg_hardalpa = BFA_TRUE;
	pport->cfg.hardalpa = alpa;

	return (BFA_STATUS_OK);
}

bfa_status_t
bfa_hal_pport_clr_hardalpa(bfa_hal_t *hal)
{
	bfa_hal_pport_t *pport = BFA_HAL_PORT_MOD(hal);

	bfa_trc(hal, pport->cfg.cfg_hardalpa);
	bfa_trc(hal, pport->cfg.hardalpa);

	pport->cfg.cfg_hardalpa = BFA_FALSE;
	return (BFA_STATUS_OK);
}

bfa_boolean_t
bfa_hal_pport_get_hardalpa(bfa_hal_t *hal, uint8_t *alpa)
{
	bfa_hal_pport_t *port = BFA_HAL_PORT_MOD(hal);

	*alpa = port->cfg.hardalpa;
	return port->cfg.cfg_hardalpa;
}

uint8_t
bfa_hal_pport_get_myalpa(bfa_hal_t *hal)
{
	bfa_hal_pport_t *port = BFA_HAL_PORT_MOD(hal);

	return port->myalpa;
}

bfa_status_t
bfa_hal_pport_cfg_maxfrsize(bfa_hal_t *hal, uint16_t maxfrsize)
{
	bfa_hal_pport_t *pport = BFA_HAL_PORT_MOD(hal);

	bfa_trc(hal, maxfrsize);
	bfa_trc(hal, pport->cfg.maxfrsize);

	/* with in range */
	if ((maxfrsize > FC_MAX_PDUSZ) || (maxfrsize < FC_MIN_PDUSZ))
		return (BFA_STATUS_INVLD_DFSZ);

	/* power of 2, if not the max frame size of 2112 */
	if ((maxfrsize != FC_MAX_PDUSZ) && (maxfrsize & (maxfrsize - 1)))
		return (BFA_STATUS_INVLD_DFSZ);

	pport->cfg.maxfrsize = maxfrsize;
	return (BFA_STATUS_OK);
}

uint16_t
bfa_hal_pport_get_maxfrsize(bfa_hal_t *hal)
{
	bfa_hal_pport_t *port = BFA_HAL_PORT_MOD(hal);

	return port->cfg.maxfrsize;
}

uint8_t
bfa_hal_pport_get_rx_bbcredit(bfa_hal_t *hal)
{
	bfa_hal_pport_t *port = BFA_HAL_PORT_MOD(hal);

	return port->cfg.rx_bbcredit;
}

void
bfa_hal_pport_set_tx_bbcredit(bfa_hal_t *hal, uint16_t tx_bbcredit)
{
	bfa_hal_pport_t *port = BFA_HAL_PORT_MOD(hal);

	bfa_assert((tx_bbcredit > 0)
				|| (port->topology == BFA_PPORT_TOPOLOGY_LOOP));
	port->cfg.tx_bbcredit = (uint8_t)tx_bbcredit;
	bfa_hal_ppport_send_txcredit(port);
}

/**
 * @brief
 * Get port attributes.
 */
void
bfa_hal_pport_get_attr(bfa_hal_t *hal, bfa_pport_attr_t *attr)
{
	bfa_hal_pport_t *pport = BFA_HAL_PORT_MOD(hal);

	bfa_os_memset(attr, 0, sizeof(bfa_pport_attr_t));

	attr->nwwn = pport->nwwn;
	attr->pwwn = pport->pwwn;

	bfa_os_memcpy(&attr->pport_cfg, &pport->cfg, sizeof(bfa_pport_cfg_t));
	/* speed attributes */
	attr->pport_cfg.speed = pport->cfg.speed;
	attr->speed_supported = pport->speed_sup;
	attr->speed = pport->speed;
	attr->cos_supported = FC_CLASS_3;

	/* topology attributes */
	attr->pport_cfg.topology = pport->cfg.topology;
	attr->topology = pport->topology;


	bfa_hal_diag_beacon_status(hal, &attr->beacon, &attr->link_e2e_beacon);
	attr->plog_enabled = bfa_plog_get_setting(pport->hal->plog);

	attr->pport_cfg.mpio_mode = bfa_hal_fcpim_mpio_mode_get(hal);
	attr->pport_cfg.path_tov  = bfa_hal_fcpim_path_tov_get(hal);
	attr->pport_cfg.q_depth  = bfa_hal_fcpim_qdepth_get(hal);
	attr->port_state = bfa_sm_to_state(hal_pport_sm_table, pport->sm);
}

static void
bfa_hal_port_stats_query(void *cbarg)
{
	bfa_hal_pport_t *port = (bfa_hal_pport_t *) cbarg;
	bfa_fwi_pport_get_stats_req_t *msg;

	msg = bfa_hal_reqq_next(port->hal, BFA_HAL_REQQ_PORT);

	if (!msg) {
		bfa_hal_reqq_winit(&port->stats_reqq_wait,
						bfa_hal_port_stats_query, port);
		bfa_hal_reqq_wait(port->hal, BFA_HAL_REQQ_PORT,
						&port->stats_reqq_wait);
		return;
	}

	bfa_os_memset(msg, 0, sizeof(bfa_fwi_pport_get_stats_req_t));
	bfa_fwi_h2i_set(msg->mh, BFA_MC_PORT, BFA_FWI_PPORT_H2I_GET_STATS_REQ,
			bfa_hal_ioc_lpuid(port->hal));
	bfa_hal_reqq_produce(port->hal, BFA_HAL_REQQ_PORT);

	return;
}

static void
bfa_hal_port_stats_clear(void *cbarg)
{
	bfa_hal_pport_t *port = (bfa_hal_pport_t *) cbarg;
	bfa_fwi_pport_clear_stats_req_t *msg;

	msg = bfa_hal_reqq_next(port->hal, BFA_HAL_REQQ_PORT);

	if (!msg) {
		bfa_hal_reqq_winit(&port->stats_reqq_wait,
						bfa_hal_port_stats_clear, port);
		bfa_hal_reqq_wait(port->hal, BFA_HAL_REQQ_PORT,
						&port->stats_reqq_wait);
		return;
	}

	bfa_os_memset(msg, 0, sizeof(bfa_fwi_pport_clear_stats_req_t));
	bfa_fwi_h2i_set(msg->mh, BFA_MC_PORT, BFA_FWI_PPORT_H2I_CLEAR_STATS_REQ,
			bfa_hal_ioc_lpuid(port->hal));
	bfa_hal_reqq_produce(port->hal, BFA_HAL_REQQ_PORT);
	return;
}

static void
bfa_hal_pport_stats_swap(bfa_pport_stats_t *d, bfa_pport_stats_t *s)
{
	uint32_t	*dip = (uint32_t *) d;
	uint32_t	*sip = (uint32_t *) s;
	int		i;

	for (i = 0; i < (sizeof(bfa_pport_stats_t) / sizeof(uint32_t));
								i = i + 2) {
		dip[i] = bfa_os_ntohl(sip[i + 1]);
		dip[i + 1] = bfa_os_ntohl(sip[i]);
	}
}

static void
__bfa_hcb_port_stats_clr(void *cbarg, bfa_boolean_t complete)
{
	bfa_hal_pport_t *port = cbarg;

	if (complete) {
		port->stats_cbfn(port->stats_cbarg, port->stats_status);
	} else {
		port->stats_busy = BFA_FALSE;
		port->stats_status = BFA_STATUS_OK;
	}
}

static void
bfa_hal_port_stats_clr_timeout(void *cbarg)
{
	bfa_hal_pport_t *port = (bfa_hal_pport_t *) cbarg;

	bfa_trc(port->hal, 0);

	port->stats_status = BFA_STATUS_ETIMER;
	bfa_hcb_queue(port->hal, &port->hcb_qe, __bfa_hcb_port_stats_clr, port);
}

static void
__bfa_hcb_port_stats(void *cbarg, bfa_boolean_t complete)
{
	bfa_hal_pport_t *port = cbarg;

	if (complete) {
		if (port->stats_status == BFA_STATUS_OK)
			bfa_hal_pport_stats_swap(port->stats_ret, port->stats);
		port->stats_cbfn(port->stats_cbarg, port->stats_status);
	} else {
		port->stats_busy = BFA_FALSE;
		port->stats_status = BFA_STATUS_OK;
	}
}

static void
bfa_hal_port_stats_timeout(void *cbarg)
{
	bfa_hal_pport_t *port = (bfa_hal_pport_t *) cbarg;

	bfa_trc(port->hal, 0);

	port->stats_status = BFA_STATUS_ETIMER;
	bfa_hcb_queue(port->hal, &port->hcb_qe, __bfa_hcb_port_stats, port);
}

#define BFA_HAL_PORT_STATS_TOV	1000

/**
 * @brief
 * Fetch port attributes.
 */
bfa_status_t
bfa_hal_pport_get_stats(bfa_hal_t *hal, bfa_pport_stats_t *stats,
			bfa_hcb_pport_t cbfn, void *cbarg)
{
	bfa_hal_pport_t *port = BFA_HAL_PORT_MOD(hal);

	if (port->stats_busy) {
		bfa_trc(hal, port->stats_busy);
		return (BFA_STATUS_DEVBUSY);
	}

	port->stats_busy = BFA_TRUE;
	port->stats_ret = stats;
	port->stats_cbfn = cbfn;
	port->stats_cbarg = cbarg;

	bfa_hal_port_stats_query(port);

	bfa_hal_timer_start(hal, &port->timer, bfa_hal_port_stats_timeout, port,
						BFA_HAL_PORT_STATS_TOV);
	return (BFA_STATUS_OK);
}

bfa_status_t
bfa_hal_pport_clear_stats(bfa_hal_t *hal, bfa_hcb_pport_t cbfn, void *cbarg)
{
	bfa_hal_pport_t *port = BFA_HAL_PORT_MOD(hal);

	if (port->stats_busy) {
		bfa_trc(hal, port->stats_busy);
		return (BFA_STATUS_DEVBUSY);
	}

	port->stats_busy = BFA_TRUE;
	port->stats_cbfn = cbfn;
	port->stats_cbarg = cbarg;

	bfa_hal_port_stats_clear(port);

	bfa_hal_timer_start(hal, &port->timer, bfa_hal_port_stats_clr_timeout,
					port, BFA_HAL_PORT_STATS_TOV);
	return (BFA_STATUS_OK);
}

bfa_status_t
bfa_hal_pport_trunk_enable(bfa_hal_t *hal, uint8_t bitmap)
{
	bfa_hal_pport_t *pport = BFA_HAL_PORT_MOD(hal);

	bfa_trc(hal, bitmap);
	bfa_trc(hal, pport->cfg.trunked);
	bfa_trc(hal, pport->cfg.trunk_ports);

	if (!bitmap || (bitmap & (bitmap - 1)))
		return BFA_STATUS_EINVAL;

	pport->cfg.trunked = BFA_TRUE;
	pport->cfg.trunk_ports = bitmap;

	return BFA_STATUS_OK;
}

/**
 * @brief
 * Fetch port attributes.
 */
bfa_status_t
bfa_hal_pport_trunk_disable(bfa_hal_t *hal)
{
	return (BFA_STATUS_OK);
}

bfa_boolean_t
bfa_hal_pport_trunk_query(bfa_hal_t *hal, uint32_t *bitmap)
{
	bfa_hal_pport_t *port = BFA_HAL_PORT_MOD(hal);

	*bitmap = port->cfg.trunk_ports;
	return port->cfg.trunked;
}

bfa_boolean_t
bfa_hal_pport_is_disabled(bfa_hal_t *hal)
{
	bfa_hal_pport_t *port = BFA_HAL_PORT_MOD(hal);

	return (bfa_sm_to_state(hal_pport_sm_table, port->sm) ==
		BFA_HAL_PPORT_DISABLED);

}

void
bfa_hal_pport_cfg_qos(bfa_hal_t *hal, bfa_boolean_t on_off)
{
	bfa_hal_pport_t *pport = BFA_HAL_PORT_MOD(hal);

	bfa_trc(hal, on_off);
	bfa_trc(hal, pport->cfg.qos_enabled);

	pport->cfg.qos_enabled = on_off;
}

void
bfa_hal_pport_cfg_ratelim(bfa_hal_t *hal, bfa_boolean_t on_off)
{
	bfa_hal_pport_t *pport = BFA_HAL_PORT_MOD(hal);

	bfa_trc(hal, on_off);
	bfa_trc(hal, pport->cfg.ratelimit);

	pport->cfg.ratelimit = on_off;
}

/**
 * @}
 */
