/*
 * Copyright (c) 2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

#include <hal/bfa_hal.h>
#include <hal/bfa_hal_rport.h>
#include <hcb/bfa_hcb_rport.h>
#include <cs/bfa_debug.h>
#include <cs/bfa_stats.h>
#include <fwi/bfa_fwi_rport.h>
#include "hal_intr.h"

BFA_TRC_FILE(HAL, RPORT);
BFA_HAL_MODULE(rport);

#define bfa_hal_rport_offline_cb(__rp) do {				\
	if ((__rp)->hal->fcs)						\
		bfa_hcb_rport_offline((__rp)->rport_drv);		\
	else {								\
		bfa_hcb_queue((__rp)->hal, &(__rp)->hcb_qe,		\
				__bfa_hcb_rport_offline, (__rp));	\
	}								\
} while (0)

#define bfa_hal_rport_online_cb(__rp) do {				\
	if ((__rp)->hal->fcs)						\
		bfa_hcb_rport_online((__rp)->rport_drv);		\
	else {								\
		bfa_hcb_queue((__rp)->hal, &(__rp)->hcb_qe,		\
				  __bfa_hcb_rport_online, (__rp));	\
		}							\
} while (0)

/*
 * forward declarations
 */
static bfa_hal_rport_t	*bfa_hal_rport_alloc(bfa_hal_rport_mod_t *rp_mod);
static void		bfa_hal_rport_free(bfa_hal_rport_t *rport);
static bfa_boolean_t 	bfa_hal_rport_send_fwcreate(bfa_hal_rport_t *rp);
static bfa_boolean_t	bfa_hal_rport_send_fwdelete(bfa_hal_rport_t *rp);
static void		__bfa_hcb_rport_online(void *cbarg,
							bfa_boolean_t complete);
static void		__bfa_hcb_rport_offline(void *cbarg,
							bfa_boolean_t complete);

/**
 * @dg hal_rport_sm HAL rport state machine
 * @img hal_rport_sm.jpg
 * @{
 */

/**
 * @brief HAL rport state machine events
 */
typedef enum {
	BFA_HAL_RPORT_SM_CREATE	= 1,	/* !< rport create event	*/
	BFA_HAL_RPORT_SM_DELETE = 2,	/* !< deleting an existing rport */
	BFA_HAL_RPORT_SM_ONLINE = 3,	/* !< rport is online		*/
	BFA_HAL_RPORT_SM_OFFLINE = 4,	/* !< rport is offline		*/
	BFA_HAL_RPORT_SM_FWRSP	= 5,	/* !< firmware response		*/
	BFA_HAL_RPORT_SM_HWFAIL	= 6,	/* !< IOC h/w failure		*/
} bfa_hal_rport_event_t;

static void	bfa_hal_rport_sm_uninit(bfa_hal_rport_t *rp,
					bfa_hal_rport_event_t event);
static void	bfa_hal_rport_sm_created(bfa_hal_rport_t *rp,
					 bfa_hal_rport_event_t event);
static void	bfa_hal_rport_sm_fwcreate(bfa_hal_rport_t *rp,
					  bfa_hal_rport_event_t event);
static void	bfa_hal_rport_sm_online(bfa_hal_rport_t *rp,
					bfa_hal_rport_event_t event);
static void	bfa_hal_rport_sm_fwdelete(bfa_hal_rport_t *rp,
					  bfa_hal_rport_event_t event);
static void	bfa_hal_rport_sm_offline(bfa_hal_rport_t *rp,
					 bfa_hal_rport_event_t event);
static void	bfa_hal_rport_sm_deleting(bfa_hal_rport_t *rp,
					  bfa_hal_rport_event_t event);
static void	bfa_hal_rport_sm_offline_pending(bfa_hal_rport_t *rp,
					  bfa_hal_rport_event_t event);
static void	bfa_hal_rport_sm_delete_pending(bfa_hal_rport_t *rp,
					  bfa_hal_rport_event_t event);
static void	bfa_hal_rport_sm_iocdisable(bfa_hal_rport_t *rp,
					    bfa_hal_rport_event_t event);

/**
 * @brief
 * Beginning state, only online event expected.
 */
static void
bfa_hal_rport_sm_uninit(bfa_hal_rport_t *rp, bfa_hal_rport_event_t event)
{
	bfa_trc(rp->hal, rp->rport_tag);
	bfa_trc(rp->hal, event);

	switch (event) {
	case BFA_HAL_RPORT_SM_CREATE:
		bfa_stats(rp, sm_un_cr);
		bfa_sm_set_state(rp, bfa_hal_rport_sm_created);
		break;

	default:
		bfa_stats(rp, sm_un_unexp);
		bfa_assert(0);
	}
}

static void
bfa_hal_rport_sm_created(bfa_hal_rport_t *rp, bfa_hal_rport_event_t event)
{
	bfa_trc(rp->hal, rp->rport_tag);
	bfa_trc(rp->hal, event);

	switch (event) {
	case BFA_HAL_RPORT_SM_ONLINE:
		bfa_stats(rp, sm_cr_on);
		bfa_sm_set_state(rp, bfa_hal_rport_sm_fwcreate);
		bfa_hal_rport_send_fwcreate(rp);
		break;

	case BFA_HAL_RPORT_SM_DELETE:
		bfa_stats(rp, sm_cr_del);
		bfa_sm_set_state(rp, bfa_hal_rport_sm_uninit);
		bfa_hal_rport_free(rp);
		break;

	case BFA_HAL_RPORT_SM_HWFAIL:
		bfa_stats(rp, sm_cr_hwf);
		bfa_sm_set_state(rp, bfa_hal_rport_sm_iocdisable);
		break;

	default:
		bfa_stats(rp, sm_cr_unexp);
		bfa_assert(0);
	}
}

/**
 * @brief
 * Waiting for rport create response from firmware.
 */
static void
bfa_hal_rport_sm_fwcreate(bfa_hal_rport_t *rp, bfa_hal_rport_event_t event)
{
	bfa_trc(rp->hal, rp->rport_tag);
	bfa_trc(rp->hal, event);

	switch (event) {
	case BFA_HAL_RPORT_SM_FWRSP:
		bfa_stats(rp, sm_fwc_rsp);
		bfa_sm_set_state(rp, bfa_hal_rport_sm_online);
		bfa_hal_rport_online_cb(rp);
		break;

	case BFA_HAL_RPORT_SM_DELETE:
		bfa_stats(rp, sm_fwc_del);
		bfa_sm_set_state(rp, bfa_hal_rport_sm_delete_pending);
		break;

	case BFA_HAL_RPORT_SM_OFFLINE:
		bfa_stats(rp, sm_fwc_off);
		bfa_sm_set_state(rp, bfa_hal_rport_sm_offline_pending);
		break;

	case BFA_HAL_RPORT_SM_HWFAIL:
		bfa_stats(rp, sm_fwc_hwf);
		bfa_sm_set_state(rp, bfa_hal_rport_sm_iocdisable);
		break;

	default:
		bfa_stats(rp, sm_fwc_unexp);
		bfa_assert(0);
	}
}

/**
 * @brief
 * Online state - normal parking state.
 */
static void
bfa_hal_rport_sm_online(bfa_hal_rport_t *rp, bfa_hal_rport_event_t event)
{
	bfa_trc(rp->hal, rp->rport_tag);
	bfa_trc(rp->hal, event);

	switch (event) {
	case BFA_HAL_RPORT_SM_OFFLINE:
		bfa_stats(rp, sm_on_off);
		bfa_sm_set_state(rp, bfa_hal_rport_sm_fwdelete);
		bfa_hal_rport_send_fwdelete(rp);
		break;

	case BFA_HAL_RPORT_SM_DELETE:
		bfa_stats(rp, sm_on_del);
		bfa_sm_set_state(rp, bfa_hal_rport_sm_deleting);
		bfa_hal_rport_send_fwdelete(rp);
		break;

	case BFA_HAL_RPORT_SM_HWFAIL:
		bfa_stats(rp, sm_on_hwf);
		bfa_sm_set_state(rp, bfa_hal_rport_sm_iocdisable);
		break;

	default:
		bfa_stats(rp, sm_on_unexp);
		bfa_assert(0);
	}
}

/**
 * @brief
 * Firmware rport is being deleted - awaiting f/w response.
 */
static void
bfa_hal_rport_sm_fwdelete(bfa_hal_rport_t *rp, bfa_hal_rport_event_t event)
{
	bfa_trc(rp->hal, rp->rport_tag);
	bfa_trc(rp->hal, event);

	switch (event) {
	case BFA_HAL_RPORT_SM_FWRSP:
		bfa_stats(rp, sm_fwd_rsp);
		bfa_sm_set_state(rp, bfa_hal_rport_sm_offline);
		bfa_hal_rport_offline_cb(rp);
		break;

	case BFA_HAL_RPORT_SM_DELETE:
		bfa_stats(rp, sm_fwd_del);
		bfa_sm_set_state(rp, bfa_hal_rport_sm_deleting);
		break;

	case BFA_HAL_RPORT_SM_HWFAIL:
		bfa_stats(rp, sm_fwd_hwf);
		bfa_sm_set_state(rp, bfa_hal_rport_sm_iocdisable);
		bfa_hal_rport_offline_cb(rp);
		break;

	default:
		bfa_stats(rp, sm_fwd_unexp);
		bfa_assert(0);
	}
}

/**
 * @brief
 * Offline state.
 */
static void
bfa_hal_rport_sm_offline(bfa_hal_rport_t *rp, bfa_hal_rport_event_t event)
{
	bfa_trc(rp->hal, rp->rport_tag);
	bfa_trc(rp->hal, event);

	switch (event) {
	case BFA_HAL_RPORT_SM_DELETE:
		bfa_stats(rp, sm_off_del);
		bfa_sm_set_state(rp, bfa_hal_rport_sm_uninit);
		bfa_hal_rport_free(rp);
		break;

	case BFA_HAL_RPORT_SM_ONLINE:
		bfa_stats(rp, sm_off_on);
		bfa_sm_set_state(rp, bfa_hal_rport_sm_fwcreate);
		bfa_hal_rport_send_fwcreate(rp);
		break;

	case BFA_HAL_RPORT_SM_HWFAIL:
		bfa_stats(rp, sm_off_hwf);
		bfa_sm_set_state(rp, bfa_hal_rport_sm_iocdisable);
		break;

	default:
		bfa_stats(rp, sm_off_unexp);
		bfa_assert(0);
	}
}

/**
 * @brief
 * Rport is deleted, waiting for firmware response to delete.
 */
static void
bfa_hal_rport_sm_deleting(bfa_hal_rport_t *rp, bfa_hal_rport_event_t event)
{
	bfa_trc(rp->hal, rp->rport_tag);
	bfa_trc(rp->hal, event);

	switch (event) {
	case BFA_HAL_RPORT_SM_FWRSP:
		bfa_stats(rp, sm_del_fwrsp);
		bfa_sm_set_state(rp, bfa_hal_rport_sm_uninit);
		bfa_hal_rport_free(rp);
		break;

	case BFA_HAL_RPORT_SM_HWFAIL:
		bfa_stats(rp, sm_del_hwf);
		bfa_sm_set_state(rp, bfa_hal_rport_sm_uninit);
		bfa_hal_rport_free(rp);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 * Waiting for rport create response from firmware. A delete is pending.
 */
static void
bfa_hal_rport_sm_delete_pending(bfa_hal_rport_t *rp,
				bfa_hal_rport_event_t event)
{
	bfa_trc(rp->hal, rp->rport_tag);
	bfa_trc(rp->hal, event);

	switch (event) {
	case BFA_HAL_RPORT_SM_FWRSP:
		bfa_stats(rp, sm_delp_fwrsp);
		bfa_sm_set_state(rp, bfa_hal_rport_sm_deleting);
		bfa_hal_rport_send_fwdelete(rp);
		break;

	case BFA_HAL_RPORT_SM_HWFAIL:
		bfa_stats(rp, sm_delp_hwf);
		bfa_sm_set_state(rp, bfa_hal_rport_sm_uninit);
		bfa_hal_rport_free(rp);
		break;

	default:
		bfa_stats(rp, sm_delp_unexp);
		bfa_assert(0);
	}
}

/**
 * @brief
 * Waiting for rport create response from firmware. Rport offline is pending.
 */
static void
bfa_hal_rport_sm_offline_pending(bfa_hal_rport_t *rp,
				 bfa_hal_rport_event_t event)
{
	bfa_trc(rp->hal, rp->rport_tag);
	bfa_trc(rp->hal, event);

	switch (event) {
	case BFA_HAL_RPORT_SM_FWRSP:
		bfa_stats(rp, sm_offp_fwrsp);
		bfa_sm_set_state(rp, bfa_hal_rport_sm_fwdelete);
		bfa_hal_rport_send_fwdelete(rp);
		break;

	case BFA_HAL_RPORT_SM_DELETE:
		bfa_stats(rp, sm_offp_del);
		bfa_sm_set_state(rp, bfa_hal_rport_sm_delete_pending);
		break;

	case BFA_HAL_RPORT_SM_HWFAIL:
		bfa_stats(rp, sm_offp_hwf);
		bfa_sm_set_state(rp, bfa_hal_rport_sm_iocdisable);
		break;

	default:
		bfa_stats(rp, sm_offp_unexp);
		bfa_assert(0);
	}
}

/**
 * @brief
 * IOC h/w failed.
 */
static void
bfa_hal_rport_sm_iocdisable(bfa_hal_rport_t *rp, bfa_hal_rport_event_t event)
{
	bfa_trc(rp->hal, rp->rport_tag);
	bfa_trc(rp->hal, event);

	switch (event) {
	case BFA_HAL_RPORT_SM_OFFLINE:
		bfa_stats(rp, sm_iocd_off);
		bfa_hal_rport_offline_cb(rp);
		break;

	case BFA_HAL_RPORT_SM_DELETE:
		bfa_stats(rp, sm_iocd_del);
		bfa_sm_set_state(rp, bfa_hal_rport_sm_uninit);
		bfa_hal_rport_free(rp);
		break;

	case BFA_HAL_RPORT_SM_ONLINE:
		bfa_stats(rp, sm_iocd_on);
		bfa_sm_set_state(rp, bfa_hal_rport_sm_fwcreate);
		bfa_hal_rport_send_fwcreate(rp);
		break;

	default:
		bfa_stats(rp, sm_iocd_unexp);
		bfa_assert(0);
	}
}

/**
 * @}
 */

/**
 * @dg hal_rport_private HAL rport private functions
 * @{
 */

static void
__bfa_hcb_rport_online(void *cbarg, bfa_boolean_t complete)
{
	bfa_hal_rport_t *rp = cbarg;

	if (complete)
		bfa_hcb_rport_online(rp->rport_drv);
}

static void
__bfa_hcb_rport_offline(void *cbarg, bfa_boolean_t complete)
{
	bfa_hal_rport_t *rp = cbarg;

	if (complete)
		bfa_hcb_rport_offline(rp->rport_drv);
}

static void
bfa_hal_rport_qresume(void *cbarg)
{
	bfa_assert(0);
}

static void
bfa_hal_rport_meminfo(bfa_ioc_cfg_t *cfg, uint32_t *km_len, uint32_t *dm_len)
{
	if (cfg->fwcfg.num_rports < BFA_HAL_RPORT_MIN)
		cfg->fwcfg.num_rports = BFA_HAL_RPORT_MIN;

	*km_len += cfg->fwcfg.num_rports * sizeof(bfa_hal_rport_t);
}

static void
bfa_hal_rport_attach(bfa_hal_t *hal, void *bfad, bfa_ioc_cfg_t *cfg,
		     bfa_hal_meminfo_t *meminfo, bfa_hal_pcidev_t *pcidev)
{
	bfa_hal_rport_mod_t *mod = BFA_HAL_RPORT_MOD(hal);
	bfa_hal_rport_t *rp;
	uint16_t        i;

	bfa_q_init(&mod->rp_free_q);
	bfa_q_init(&mod->rp_active_q);

	rp = (bfa_hal_rport_t *) bfa_hal_meminfo_kva(meminfo);
	mod->rps_list = rp;
	mod->num_rports = cfg->fwcfg.num_rports;

	bfa_assert(mod->num_rports
		   && !(mod->num_rports & (mod->num_rports - 1)));

	for (i = 0; i < mod->num_rports; i++, rp++) {
		bfa_os_memset(rp, 0, sizeof(bfa_hal_rport_t));
		rp->hal = hal;
		rp->rport_tag = i;
		bfa_sm_set_state(rp, bfa_hal_rport_sm_uninit);

		/**
		 *  - is unused
		 */
		if (i)
			bfa_q_enq(&mod->rp_free_q, rp);

		bfa_hal_reqq_winit(&rp->reqq_wait, bfa_hal_rport_qresume, rp);
	}

	/**
	 * consume memory
	 */
	bfa_hal_meminfo_kva(meminfo) = (uint8_t *) rp;
}

static void
bfa_hal_rport_initdone(bfa_hal_t *hal)
{
}

static void
bfa_hal_rport_detach(bfa_hal_t *hal)
{
}

static void
bfa_hal_rport_start(bfa_hal_t *hal)
{
}

static void
bfa_hal_rport_stop(bfa_hal_t *hal)
{
}

static void
bfa_hal_rport_iocdisable(bfa_hal_t *hal)
{
	bfa_hal_rport_mod_t *mod = BFA_HAL_RPORT_MOD(hal);
	bfa_hal_rport_t *rport;
	bfa_q_t        *qe, *qen;

	bfa_q_iter_safe(&mod->rp_active_q, qe, qen) {
		rport = (bfa_hal_rport_t *) qe;
		bfa_sm_send_event(rport, BFA_HAL_RPORT_SM_HWFAIL);
	}
}

static bfa_hal_rport_t *
bfa_hal_rport_alloc(bfa_hal_rport_mod_t *mod)
{
	bfa_hal_rport_t *rport;

	bfa_q_deq(&mod->rp_free_q, &rport);
	if (rport)
		bfa_q_enq(&mod->rp_active_q, rport);

	return (rport);
}

static void
bfa_hal_rport_free(bfa_hal_rport_t *rport)
{
	bfa_hal_rport_mod_t *mod = BFA_HAL_RPORT_MOD(rport->hal);

	bfa_assert(bfa_q_is_on_q(&mod->rp_active_q, rport));
	bfa_q_qe_deq(rport);
	bfa_q_enq(&mod->rp_free_q, rport);
}

static bfa_boolean_t
bfa_hal_rport_send_fwcreate(bfa_hal_rport_t *rp)
{
	bfa_fwi_rport_create_req_t *m;

	/**
	 * check for room in queue to send request now
	 */
	m = bfa_hal_reqq_next(rp->hal, BFA_HAL_REQQ_RPORT);
	if (!m) {
		bfa_hal_reqq_wait(rp->hal, BFA_HAL_REQQ_RPORT, &rp->reqq_wait);
		return BFA_FALSE;
	}

	bfa_fwi_h2i_set(m->mh, BFA_MC_RPORT, BFA_FWI_RPORT_H2I_CREATE_REQ,
			bfa_hal_ioc_lpuid(rp->hal));
	m->bfa_handle = rp->rport_tag;
	m->max_frmsz = bfa_os_htons(rp->rport_info.max_frmsz);
	m->pid = rp->rport_info.pid;
	m->local_pid = rp->rport_info.local_pid;
	m->fc_class = rp->rport_info.fc_class;
	m->vf_en = rp->rport_info.vf_en;
	m->vf_id = rp->rport_info.vf_id;
	m->cisc = rp->rport_info.cisc;

	/**
	 * queue I/O message to firmware
	 */
	bfa_hal_reqq_produce(rp->hal, BFA_HAL_REQQ_RPORT);
	return BFA_TRUE;
}

static bfa_boolean_t
bfa_hal_rport_send_fwdelete(bfa_hal_rport_t *rp)
{
	bfa_fwi_rport_delete_req_t *m;

	/**
	 * check for room in queue to send request now
	 */
	m = bfa_hal_reqq_next(rp->hal, BFA_HAL_REQQ_RPORT);
	if (!m) {
		bfa_hal_reqq_wait(rp->hal, BFA_HAL_REQQ_RPORT, &rp->reqq_wait);
		return BFA_FALSE;
	}

	bfa_fwi_h2i_set(m->mh, BFA_MC_RPORT, BFA_FWI_RPORT_H2I_DELETE_REQ,
			bfa_hal_ioc_lpuid(rp->hal));
	m->fw_handle = rp->fw_handle;

	/**
	 * queue I/O message to firmware
	 */
	bfa_hal_reqq_produce(rp->hal, BFA_HAL_REQQ_RPORT);
	return BFA_TRUE;
}

/**
 * @}
 */

/**
 * @ig hal_rport_public
 * @{
 */

/**
 * @brief
 * 		Rport interrupt processing.
 */
void
bfa_hal_rport_isr(bfa_hal_t *hal, bfa_msg_t *m)
{
	bfa_fwi_rport_i2h_msg_t msg;
	bfa_hal_rport_t *rp;

	bfa_trc(hal, m->mhdr.msg_id);

	msg.msg = m;

	switch (m->mhdr.msg_id) {
	case BFA_FWI_RPORT_I2H_CREATE_RSP:
		rp = BFA_HAL_RPORT_FROM_TAG(hal, msg.create_rsp->bfa_handle);
		rp->fw_handle = msg.create_rsp->fw_handle;
		bfa_assert(msg.create_rsp->status == BFA_STATUS_OK);
		bfa_sm_send_event(rp, BFA_HAL_RPORT_SM_FWRSP);
		break;

	case BFA_FWI_RPORT_I2H_DELETE_RSP:
		rp = BFA_HAL_RPORT_FROM_TAG(hal, msg.delete_rsp->bfa_handle);
		bfa_assert(msg.delete_rsp->status == BFA_STATUS_OK);
		bfa_sm_send_event(rp, BFA_HAL_RPORT_SM_FWRSP);
		break;

	default:
		bfa_trc(hal, m->mhdr.msg_id);
		bfa_assert(0);
	}
}

/**
 * @}
 */

/**
 * @ig hal_rport_api
 * @{
 */

bfa_hal_rport_t *
bfa_hal_rport_create(bfa_hal_t *hal, void *rport_drv)
{
	bfa_hal_rport_t *rp;

	rp = bfa_hal_rport_alloc(BFA_HAL_RPORT_MOD(hal));

	if (rp == NULL)
		return (NULL);

	rp->hal = hal;
	rp->rport_drv = rport_drv;
	bfa_hal_rport_clear_stats(rp);

	bfa_assert(bfa_sm_cmp_state(rp, bfa_hal_rport_sm_uninit));
	bfa_sm_send_event(rp, BFA_HAL_RPORT_SM_CREATE);

	return (rp);
}

void
bfa_hal_rport_delete(bfa_hal_rport_t *rport)
{
	bfa_sm_send_event(rport, BFA_HAL_RPORT_SM_DELETE);
}

void
bfa_hal_rport_online(bfa_hal_rport_t *rport, bfa_hal_rport_info_t *rport_info)
{
	bfa_assert(rport_info->max_frmsz != 0);

	rport->rport_info = *rport_info;
	bfa_sm_send_event(rport, BFA_HAL_RPORT_SM_ONLINE);
}

void
bfa_hal_rport_offline(bfa_hal_rport_t *rport)
{
	bfa_sm_send_event(rport, BFA_HAL_RPORT_SM_OFFLINE);
}

void
bfa_hal_rport_get_stats(bfa_hal_rport_t *rport, bfa_rport_hal_stats_t *stats)
{
	*stats = rport->stats;
}

void
bfa_hal_rport_clear_stats(bfa_hal_rport_t *rport)
{
	bfa_os_memset(&rport->stats, 0, sizeof(rport->stats));
}

/**
 * @}
 */
