/*
 * Copyright (c) 2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

#include <hal/bfa_hal.h>

BFA_TRC_FILE(HAL, SGPG);
BFA_HAL_MODULE(sgpg);

/**
 * @dg hal_sgpg_mod HAL SGPG Mode module
 * @{
 */

/**
 * @brief
 * Compute and return memory needed by FCP(im) module.
 */
static void
bfa_hal_sgpg_meminfo(bfa_ioc_cfg_t *cfg, uint32_t *km_len, uint32_t *dm_len)
{
	if (cfg->num_sgpgs < BFA_HAL_SGPG_MIN)
		cfg->num_sgpgs = BFA_HAL_SGPG_MIN;

	*km_len += cfg->num_sgpgs * sizeof(bfa_hal_sgpg_t);
	*dm_len += cfg->num_sgpgs * sizeof(bfa_sgpg_t);
}

/**
 * @brief
 */
static void
bfa_hal_sgpg_attach(bfa_hal_t *hal, void *bfad, bfa_ioc_cfg_t *cfg,
		    bfa_hal_meminfo_t *minfo, bfa_hal_pcidev_t *pcidev)
{
	bfa_hal_sgpg_mod_t	*mod = BFA_HAL_SGPG_MOD(hal);
	int			i;
	bfa_hal_sgpg_t		*hsgpg;
	bfa_sgpg_t		*sgpg;

	union {
		uint64_t        pa;
		bfa_addr_t      addr;
	} sgpg_pa;

	bfa_q_init(&mod->sgpg_q);
	bfa_q_init(&mod->sgpg_wait_q);

	bfa_trc(hal, cfg->num_sgpgs);

	mod->num_sgpgs = cfg->num_sgpgs;
	mod->hsgpg_arr = (bfa_hal_sgpg_t *) bfa_hal_meminfo_kva(minfo);
	mod->sgpg_arr = (bfa_sgpg_t *) bfa_hal_meminfo_dma_virt(minfo);
	mod->sgpg_arr_pa = bfa_hal_meminfo_dma_phys(minfo);

	hsgpg = mod->hsgpg_arr;
	sgpg = mod->sgpg_arr;
	sgpg_pa.pa = mod->sgpg_arr_pa;
	mod->free_sgpgs = mod->num_sgpgs;

	bfa_assert(!(sgpg_pa.pa & (sizeof(bfa_sgpg_t) - 1)));

	for (i = 0; i < mod->num_sgpgs; i++) {
		bfa_os_memset(hsgpg, 0, sizeof(*hsgpg));
		bfa_os_memset(sgpg, 0, sizeof(*sgpg));

		hsgpg->sgpg = sgpg;
		hsgpg->sgpg_pa = sgpg_pa.addr;
		bfa_q_enq(&mod->sgpg_q, hsgpg);

		hsgpg++;
		sgpg++;
		sgpg_pa.pa += sizeof(bfa_sgpg_t);
	}

	bfa_hal_meminfo_kva(minfo) = (uint8_t *) hsgpg;
	bfa_hal_meminfo_dma_virt(minfo) = (uint8_t *) sgpg;
	bfa_hal_meminfo_dma_phys(minfo) = sgpg_pa.pa;
}

static void
bfa_hal_sgpg_initdone(bfa_hal_t *hal)
{
}

static void
bfa_hal_sgpg_detach(bfa_hal_t *hal)
{
}

static void
bfa_hal_sgpg_start(bfa_hal_t *hal)
{
}

static void
bfa_hal_sgpg_stop(bfa_hal_t *hal)
{
}

static void
bfa_hal_sgpg_iocdisable(bfa_hal_t *hal)
{
}

/**
 * @}
 */

/**
 * @dg hal_sgpg_public HAL SGPG public functions
 * @{
 */

bfa_status_t
bfa_hal_sgpg_malloc(bfa_hal_t *hal, bfa_q_t *sgpg_q, int nsgpgs)
{
	bfa_hal_sgpg_mod_t *mod = BFA_HAL_SGPG_MOD(hal);
	bfa_hal_sgpg_t *hsgpg;
	int             i;

	bfa_trc_fp(hal, nsgpgs);

	if (mod->free_sgpgs < nsgpgs)
		return BFA_STATUS_ENOMEM;

	for (i = 0; i < nsgpgs; i++) {
		bfa_q_deq(&mod->sgpg_q, &hsgpg);
		bfa_assert(hsgpg);
		bfa_q_enq(sgpg_q, hsgpg);
	}

	mod->free_sgpgs -= nsgpgs;
	return BFA_STATUS_OK;
}

void
bfa_hal_sgpg_mfree(bfa_hal_t *hal, bfa_q_t *sgpg_q, int nsgpg)
{
	bfa_hal_sgpg_mod_t *mod = BFA_HAL_SGPG_MOD(hal);
	bfa_hal_sgpg_wqe_t *wqe;

	bfa_trc_fp(hal, nsgpg);

	mod->free_sgpgs += nsgpg;
	bfa_assert(mod->free_sgpgs <= mod->num_sgpgs);

	bfa_q_enq_q(&mod->sgpg_q, sgpg_q);

	if (bfa_q_is_empty(&mod->sgpg_wait_q))
		return;

	/**
	 * satisfy as many waiting requests as possible
	 */
	do {
		wqe = bfa_q_first(&mod->sgpg_wait_q);
		if (mod->free_sgpgs < wqe->nsgpg)
			nsgpg = mod->free_sgpgs;
		else
			nsgpg = wqe->nsgpg;
		bfa_hal_sgpg_malloc(hal, &wqe->sgpg_q, nsgpg);
		wqe->nsgpg -= nsgpg;
		if (wqe->nsgpg == 0) {
			bfa_q_qe_deq(wqe);
			wqe->cbfn(wqe->cbarg);
		}
	} while (mod->free_sgpgs && !bfa_q_is_empty(&mod->sgpg_wait_q));
}

void
bfa_hal_sgpg_wait(bfa_hal_t *hal, bfa_hal_sgpg_wqe_t *wqe, int nsgpg)
{
	bfa_hal_sgpg_mod_t *mod = BFA_HAL_SGPG_MOD(hal);

	bfa_assert(nsgpg > 0);
	bfa_assert(nsgpg > mod->free_sgpgs);

	wqe->nsgpg_total = wqe->nsgpg = nsgpg;

	/**
	 * allocate any left to this one first
	 */
	if (mod->free_sgpgs) {
		/**
		 * no one else is waiting for SGPG
		 */
		bfa_assert(bfa_q_is_empty(&mod->sgpg_wait_q));
		bfa_q_enq_q(&wqe->sgpg_q, &mod->sgpg_q);
		wqe->nsgpg -= mod->free_sgpgs;
		mod->free_sgpgs = 0;
	}

	bfa_q_enq(&mod->sgpg_wait_q, wqe);
}

void
bfa_hal_sgpg_wcancel(bfa_hal_t *hal, bfa_hal_sgpg_wqe_t *wqe)
{
	bfa_hal_sgpg_mod_t *mod = BFA_HAL_SGPG_MOD(hal);

	bfa_assert(bfa_q_is_on_q(&mod->sgpg_wait_q, wqe));
	bfa_q_qe_deq(wqe);

	if (wqe->nsgpg_total != wqe->nsgpg)
		bfa_hal_sgpg_mfree(hal, &wqe->sgpg_q,
				   wqe->nsgpg_total - wqe->nsgpg);
}

void
bfa_hal_sgpg_winit(bfa_hal_sgpg_wqe_t *wqe, void (*cbfn) (void *cbarg),
		   void *cbarg)
{
	bfa_q_init(&wqe->sgpg_q);
	wqe->cbfn = cbfn;
	wqe->cbarg = cbarg;
}

/**
 * @}
 */
