/*
 * Copyright (c) 2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

#include <hal/bfa_hal.h>
#include <cs/bfa_stats.h>
#include <bfa_hcb_ioim_macros.h>

BFA_TRC_FILE(HAL, TSKIM);

/**
 * task management completion handling
 */
#define bfa_hal_tskim_qcomp(__tskim, __cbfn) do {			     \
	bfa_hcb_queue((__tskim)->hal, &(__tskim)->hcb_qe, __cbfn, (__tskim));\
	bfa_hal_tskim_notify_comp(__tskim);				     \
} while (0)

#define bfa_hal_tskim_notify_comp(__tskim) do {				     \
	if ((__tskim)->notify)					     	     \
		bfa_hal_itnim_tskdone((__tskim)->itnim);		     \
} while (0)

/*
 * forward declarations
 */
static void     __bfa_hcb_tskim_done(void *cbarg, bfa_boolean_t complete);
static void     __bfa_hcb_tskim_failed(void *cbarg, bfa_boolean_t complete);
static bfa_boolean_t bfa_hal_tskim_match_scope(bfa_hal_tskim_t *tskim,
					       lun_t lun);
static void     bfa_hal_tskim_gather_ios(bfa_hal_tskim_t *tskim);
static void     bfa_hal_tskim_cleanp_comp(void *tskim_cbarg);
static void     bfa_hal_tskim_cleanup_ios(bfa_hal_tskim_t *tskim);
static bfa_boolean_t bfa_hal_tskim_send(bfa_hal_tskim_t *tskim);
static bfa_boolean_t bfa_hal_tskim_send_abort(bfa_hal_tskim_t *tskim);
static void     bfa_hal_tskim_iocdisable_ios(bfa_hal_tskim_t *tskim);

/**
 * @ig hal_tskim_sm
 * @{
 */

typedef enum {
	BFA_HAL_TSKIM_SM_START        = 1,  /*!< TM command start            */
	BFA_HAL_TSKIM_SM_DONE         = 2,  /*!< TM completion               */
	BFA_HAL_TSKIM_SM_QRESUME      = 3,  /*!< resume after qfull          */
	BFA_HAL_TSKIM_SM_HWFAIL       = 5,  /*!< IOC h/w failure event       */
	BFA_HAL_TSKIM_SM_HCB          = 6,  /*!< HAL callback completion     */
	BFA_HAL_TSKIM_SM_IOS_DONE     = 7,  /*!< IO and sub TM completions   */
	BFA_HAL_TSKIM_SM_CLEANUP      = 8,  /*!< TM cleanup on ITN offline   */
	BFA_HAL_TSKIM_SM_CLEANUP_DONE = 9,  /*!< TM abort completion         */
} bfa_hal_tskim_event_t;

static void     bfa_hal_tskim_sm_uninit(bfa_hal_tskim_t *tskim,
					bfa_hal_tskim_event_t event);
static void     bfa_hal_tskim_sm_active(bfa_hal_tskim_t *tskim,
					bfa_hal_tskim_event_t event);
static void     bfa_hal_tskim_sm_cleanup(bfa_hal_tskim_t *tskim,
					 bfa_hal_tskim_event_t event);
static void     bfa_hal_tskim_sm_iocleanup(bfa_hal_tskim_t *tskim,
					 bfa_hal_tskim_event_t event);
static void     bfa_hal_tskim_sm_qfull(bfa_hal_tskim_t *tskim,
				       bfa_hal_tskim_event_t event);
static void     bfa_hal_tskim_sm_cleanup_qfull(bfa_hal_tskim_t *tskim,
				       bfa_hal_tskim_event_t event);
static void     bfa_hal_tskim_sm_hcb(bfa_hal_tskim_t *tskim,
				     bfa_hal_tskim_event_t event);

/**
 * @brief
 *      Task management command beginning state.
 */
static void
bfa_hal_tskim_sm_uninit(bfa_hal_tskim_t *tskim, bfa_hal_tskim_event_t event)
{
	bfa_trc(tskim->hal, event);

	switch (event) {
	case BFA_HAL_TSKIM_SM_START:
		bfa_sm_set_state(tskim, bfa_hal_tskim_sm_active);
		bfa_hal_tskim_gather_ios(tskim);

		/**
		 * If device is offline, do not send TM on wire. Just cleanup
		 * any pending IO requests and complete TM request.
		 */
		if (!bfa_hal_itnim_is_online(tskim->itnim)) {
			bfa_sm_set_state(tskim, bfa_hal_tskim_sm_iocleanup);
			tskim->tsk_status = BFA_TSKIM_STS_OK;
			bfa_hal_tskim_cleanup_ios(tskim);
			return;
		}

		if (!bfa_hal_tskim_send(tskim)) {
			bfa_sm_set_state(tskim, bfa_hal_tskim_sm_qfull);
			bfa_hal_reqq_wait(tskim->hal, tskim->itnim->reqq,
					  &tskim->reqq_wait);
		}
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * brief
 *	TM command is active, awaiting completion from firmware to
 *	cleanup IO requests in TM scope.
 */
static void
bfa_hal_tskim_sm_active(bfa_hal_tskim_t *tskim, bfa_hal_tskim_event_t event)
{
	bfa_trc(tskim->hal, event);

	switch (event) {
	case BFA_HAL_TSKIM_SM_DONE:
		bfa_sm_set_state(tskim, bfa_hal_tskim_sm_iocleanup);
		bfa_hal_tskim_cleanup_ios(tskim);
		break;

	case BFA_HAL_TSKIM_SM_CLEANUP:
		bfa_sm_set_state(tskim, bfa_hal_tskim_sm_cleanup);
		if (!bfa_hal_tskim_send_abort(tskim)) {
			bfa_sm_set_state(tskim, bfa_hal_tskim_sm_cleanup_qfull);
			bfa_hal_reqq_wait(tskim->hal, tskim->itnim->reqq,
				&tskim->reqq_wait);
		}
		break;

	case BFA_HAL_TSKIM_SM_HWFAIL:
		bfa_sm_set_state(tskim, bfa_hal_tskim_sm_hcb);
		bfa_hal_tskim_iocdisable_ios(tskim);
		bfa_hal_tskim_qcomp(tskim, __bfa_hcb_tskim_failed);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 *	An active TM is being cleaned up since ITN is offline. Awaiting cleanup
 *	completion event from firmware.
 */
static void
bfa_hal_tskim_sm_cleanup(bfa_hal_tskim_t *tskim, bfa_hal_tskim_event_t event)
{
	bfa_trc(tskim->hal, event);

	switch (event) {
	case BFA_HAL_TSKIM_SM_DONE:
		/**
		 * Ignore and wait for ABORT completion from firmware.
		 */
		break;

	case BFA_HAL_TSKIM_SM_CLEANUP_DONE:
		bfa_sm_set_state(tskim, bfa_hal_tskim_sm_iocleanup);
		bfa_hal_tskim_cleanup_ios(tskim);
		break;

	case BFA_HAL_TSKIM_SM_HWFAIL:
		bfa_sm_set_state(tskim, bfa_hal_tskim_sm_hcb);
		bfa_hal_tskim_iocdisable_ios(tskim);
		bfa_hal_tskim_qcomp(tskim, __bfa_hcb_tskim_failed);
		break;

	default:
		bfa_assert(0);
	}
}

static void
bfa_hal_tskim_sm_iocleanup(bfa_hal_tskim_t *tskim, bfa_hal_tskim_event_t event)
{
	bfa_trc(tskim->hal, event);

	switch (event) {
	case BFA_HAL_TSKIM_SM_IOS_DONE:
		bfa_sm_set_state(tskim, bfa_hal_tskim_sm_hcb);
		bfa_hal_tskim_qcomp(tskim, __bfa_hcb_tskim_done);
		break;

	case BFA_HAL_TSKIM_SM_CLEANUP:
		/**
		 * Ignore, TM command completed on wire.
		 * Notify TM conmpletion on IO cleanup completion.
		 */
		break;

	case BFA_HAL_TSKIM_SM_HWFAIL:
		bfa_sm_set_state(tskim, bfa_hal_tskim_sm_hcb);
		bfa_hal_tskim_iocdisable_ios(tskim);
		bfa_hal_tskim_qcomp(tskim, __bfa_hcb_tskim_failed);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 *      Task management command is waiting for room in request CQ
 */
static void
bfa_hal_tskim_sm_qfull(bfa_hal_tskim_t *tskim, bfa_hal_tskim_event_t event)
{
	bfa_trc(tskim->hal, event);

	switch (event) {
	case BFA_HAL_TSKIM_SM_QRESUME:
		bfa_sm_set_state(tskim, bfa_hal_tskim_sm_active);
		bfa_hal_tskim_send(tskim);
		break;

	case BFA_HAL_TSKIM_SM_CLEANUP:
		/**
		 * No need to send TM on wire since ITN is offline.
		 */
		bfa_sm_set_state(tskim, bfa_hal_tskim_sm_iocleanup);
		bfa_hal_reqq_wcancel(&tskim->itnim->reqq);
		bfa_hal_tskim_cleanup_ios(tskim);
		break;

	case BFA_HAL_TSKIM_SM_HWFAIL:
		bfa_sm_set_state(tskim, bfa_hal_tskim_sm_hcb);
		bfa_hal_reqq_wcancel(&tskim->itnim->reqq);
		bfa_hal_tskim_iocdisable_ios(tskim);
		bfa_hal_tskim_qcomp(tskim, __bfa_hcb_tskim_failed);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 *      Task management command is active, awaiting for room in request CQ
 *	to send clean up request.
 */
static void
bfa_hal_tskim_sm_cleanup_qfull(bfa_hal_tskim_t *tskim, bfa_hal_tskim_event_t event)
{
	bfa_trc(tskim->hal, event);

	switch (event) {
	case BFA_HAL_TSKIM_SM_DONE:
		bfa_hal_reqq_wcancel(&tskim->itnim->reqq);
		/**
		 *
		 * Fall through !!!
		 */

	case BFA_HAL_TSKIM_SM_QRESUME:
		bfa_sm_set_state(tskim, bfa_hal_tskim_sm_cleanup);
		bfa_hal_tskim_send_abort(tskim);
		break;

	case BFA_HAL_TSKIM_SM_HWFAIL:
		bfa_sm_set_state(tskim, bfa_hal_tskim_sm_hcb);
		bfa_hal_reqq_wcancel(&tskim->itnim->reqq);
		bfa_hal_tskim_iocdisable_ios(tskim);
		bfa_hal_tskim_qcomp(tskim, __bfa_hcb_tskim_failed);
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @brief
 *      HAL callback is pending
 */
static void
bfa_hal_tskim_sm_hcb(bfa_hal_tskim_t *tskim, bfa_hal_tskim_event_t event)
{
	bfa_trc(tskim->hal, event);

	switch (event) {
	case BFA_HAL_TSKIM_SM_HCB:
		bfa_sm_set_state(tskim, bfa_hal_tskim_sm_uninit);
		bfa_hal_tskim_free(tskim);
		break;

	case BFA_HAL_TSKIM_SM_CLEANUP:
		bfa_hal_tskim_notify_comp(tskim);
		break;

	case BFA_HAL_TSKIM_SM_HWFAIL:
		break;

	default:
		bfa_assert(0);
	}
}

/**
 * @}
 */

/**
 * @ig hal_tskim_private
 * @{
 */

static void
__bfa_hcb_tskim_done(void *cbarg, bfa_boolean_t complete)
{
	bfa_hal_tskim_t *tskim = cbarg;

	if (!complete) {
		bfa_sm_send_event(tskim, BFA_HAL_TSKIM_SM_HCB);
		return;
	}

	bfa_stats(tskim->itnim, tm_success);
	bfa_hcb_tskim_done(tskim->hal->bfad, tskim->dtsk, tskim->tsk_status);
}

static void
__bfa_hcb_tskim_failed(void *cbarg, bfa_boolean_t complete)
{
	bfa_hal_tskim_t *tskim = cbarg;

	if (!complete) {
		bfa_sm_send_event(tskim, BFA_HAL_TSKIM_SM_HCB);
		return;
	}

	bfa_stats(tskim->itnim, tm_failures);
	bfa_hcb_tskim_done(tskim->hal->bfad, tskim->dtsk,
			   BFA_TSKIM_STS_TM_FAILED);
}

static          bfa_boolean_t
bfa_hal_tskim_match_scope(bfa_hal_tskim_t *tskim, lun_t lun)
{
	switch (tskim->tm_cmnd) {
	case FCP_TM_TARGET_RESET:
		return BFA_TRUE;

	case FCP_TM_ABORT_TASK_SET:
	case FCP_TM_CLEAR_TASK_SET:
	case FCP_TM_LUN_RESET:
	case FCP_TM_CLEAR_ACA:
		return (tskim->lun == lun);

	default:
		bfa_assert(0);
	}

	return BFA_FALSE;
}

/**
 * @brief
 *      Gather affected IO requests and task management commands.
 */
static void
bfa_hal_tskim_gather_ios(bfa_hal_tskim_t *tskim)
{
	bfa_hal_itnim_t *itnim = tskim->itnim;
	bfa_hal_ioim_t *ioim;
	bfa_q_t        *qe, *qen;

	bfa_q_init(&tskim->io_q);

	/**
	 * Gather any active IO requests first.
	 */
	bfa_q_iter_safe(&itnim->io_q, qe, qen) {
		ioim = (bfa_hal_ioim_t *) qe;
		if (bfa_hal_tskim_match_scope
		    (tskim, bfa_hcb_ioim_get_lun(ioim->dio))) {
			bfa_q_qe_deq(ioim);
			bfa_q_enq(&tskim->io_q, ioim);
		}
	}

	/**
	 * Failback any pending IO requests immediately.
	 */
	bfa_q_iter_safe(&itnim->pending_q, qe, qen) {
		ioim = (bfa_hal_ioim_t *) qe;
		if (bfa_hal_tskim_match_scope
		    (tskim, bfa_hcb_ioim_get_lun(ioim->dio))) {
			bfa_q_qe_deq(ioim);
			bfa_q_enq(&ioim->fcpim->ioim_comp_q, ioim);
			bfa_hal_ioim_tov(ioim);
		}
	}
}

/**
 * @brief
 * 		IO cleanup completion
 */
static void
bfa_hal_tskim_cleanp_comp(void *tskim_cbarg)
{
	bfa_hal_tskim_t *tskim = tskim_cbarg;

	bfa_stats(tskim->itnim, tm_io_comps);
	bfa_sm_send_event(tskim, BFA_HAL_TSKIM_SM_IOS_DONE);
}

/**
 * @brief
 *      Gather affected IO requests and task management commands.
 */
static void
bfa_hal_tskim_cleanup_ios(bfa_hal_tskim_t *tskim)
{
	bfa_hal_ioim_t *ioim;
	bfa_q_t        *qe, *qen;

	bfa_wc_init(&tskim->wc, bfa_hal_tskim_cleanp_comp, tskim);

	bfa_q_iter_safe(&tskim->io_q, qe, qen) {
		ioim = (bfa_hal_ioim_t *) qe;
		bfa_wc_up(&tskim->wc);
		bfa_hal_ioim_cleanup_tm(ioim, tskim);
	}

	bfa_wc_wait(&tskim->wc);
}

/**
 * @brief
 *      Send task management request to firmware.
 */
static bfa_boolean_t
bfa_hal_tskim_send(bfa_hal_tskim_t *tskim)
{
	bfa_hal_itnim_t *itnim = tskim->itnim;
	bfa_fwi_tskim_req_t *m;

	/**
	 * check for room in queue to send request now
	 */
	m = bfa_hal_reqq_next(tskim->hal, itnim->reqq);
	if (!m)
		return BFA_FALSE;

	/**
	 * build i/o request message next
	 */
	bfa_fwi_h2i_set(m->mh, BFA_MC_TSKIM, BFA_FWI_TSKIM_H2I_TM_REQ,
			bfa_hal_ioc_lpuid(tskim->hal));

	m->tsk_tag = bfa_os_htons(tskim->tsk_tag);
	m->itn_fhdl = tskim->itnim->rport->fw_handle;
	m->t_secs = tskim->tsecs;
	m->lun = tskim->lun;
	m->tm_flags = tskim->tm_cmnd;

	/**
	 * queue I/O message to firmware
	 */
	bfa_hal_reqq_produce(tskim->hal, itnim->reqq);
	return BFA_TRUE;
}

/**
 * @brief
 *      Send abort request to cleanup an active TM to firmware.
 */
static bfa_boolean_t
bfa_hal_tskim_send_abort(bfa_hal_tskim_t *tskim)
{
	bfa_hal_itnim_t             *itnim = tskim->itnim;
	bfa_fwi_tskim_abortreq_t    *m;

	/**
	 * check for room in queue to send request now
	 */
	m = bfa_hal_reqq_next(tskim->hal, itnim->reqq);
	if (!m)
		return BFA_FALSE;

	/**
	 * build i/o request message next
	 */
	bfa_fwi_h2i_set(m->mh, BFA_MC_TSKIM, BFA_FWI_TSKIM_H2I_ABORT_REQ,
			bfa_hal_ioc_lpuid(tskim->hal));

	m->tsk_tag  = bfa_os_htons(tskim->tsk_tag);

	/**
	 * queue I/O message to firmware
	 */
	bfa_hal_reqq_produce(tskim->hal, itnim->reqq);
	return BFA_TRUE;
}

/**
 * @brief
 *      Call to resume task management cmnd waiting for room in request queue.
 */
static void
bfa_hal_tskim_qresume(void *cbarg)
{
	bfa_hal_tskim_t *tskim = cbarg;

	bfa_hal_fcpim_stats(tskim->fcpim, qresumes);
	bfa_stats(tskim->itnim, tm_qresumes);
	bfa_sm_send_event(tskim, BFA_HAL_TSKIM_SM_QRESUME);
}

/**
 * Cleanup IOs associated with a task mangement command on IOC failures.
 */
static void
bfa_hal_tskim_iocdisable_ios(bfa_hal_tskim_t *tskim)
{
	bfa_hal_ioim_t *ioim;
	bfa_q_t        *qe, *qen;

	bfa_q_iter_safe(&tskim->io_q, qe, qen) {
		ioim = (bfa_hal_ioim_t *) qe;
		bfa_hal_ioim_iocdisable(ioim);
	}
}

/**
 * @}
 */

/**
 * @ig hal_tskim_friend
 * @{
 */

/**
 * Notification on completions from related ioim.
 */
void
bfa_hal_tskim_iodone(bfa_hal_tskim_t *tskim)
{
	bfa_wc_down(&tskim->wc);
}

/**
 * Handle IOC h/w failure notification from itnim.
 */
void
bfa_hal_tskim_iocdisable(bfa_hal_tskim_t *tskim)
{
	tskim->notify = BFA_FALSE;
	bfa_stats(tskim->itnim, tm_iocdowns);
	bfa_sm_send_event(tskim, BFA_HAL_TSKIM_SM_HWFAIL);
}

/**
 * Cleanup TM command and associated IOs as part of ITNIM offline.
 */
void
bfa_hal_tskim_cleanup(bfa_hal_tskim_t *tskim)
{
	tskim->notify = BFA_TRUE;
	bfa_stats(tskim->itnim, tm_cleanups);
	bfa_sm_send_event(tskim, BFA_HAL_TSKIM_SM_CLEANUP);
}

/**
 * @brief
 *      Memory allocation and initialization.
 */
void
bfa_hal_tskim_attach(bfa_hal_fcpim_mod_t *fcpim, bfa_hal_meminfo_t *minfo)
{
	bfa_hal_tskim_t *tskim;
	uint16_t        i;

	bfa_q_init(&fcpim->tskim_free_q);

	tskim = (bfa_hal_tskim_t *) bfa_hal_meminfo_kva(minfo);
	fcpim->tskim_arr = tskim;

	for (i = 0; i < fcpim->num_tskim_reqs; i++, tskim++) {
		/*
		 * initialize TSKIM
		 */
		bfa_os_memset(tskim, 0, sizeof(bfa_hal_tskim_t));
		tskim->tsk_tag = i;
		tskim->hal     = fcpim->hal;
		tskim->fcpim   = fcpim;
		tskim->notify  = BFA_FALSE;
		bfa_hal_reqq_winit(&tskim->reqq_wait, bfa_hal_tskim_qresume,
				   tskim);
		bfa_sm_set_state(tskim, bfa_hal_tskim_sm_uninit);

		bfa_q_enq(&fcpim->tskim_free_q, tskim);
	}

	bfa_hal_meminfo_kva(minfo) = (uint8_t *) tskim;
}

void
bfa_hal_tskim_detach(bfa_hal_fcpim_mod_t *fcpim)
{
    /**
     * @todo
     */
}

void
bfa_hal_tskim_isr(bfa_hal_t *hal, bfa_msg_t *m)
{
	bfa_hal_fcpim_mod_t *fcpim = BFA_HAL_FCPIM_MOD(hal);
	bfa_fwi_tskim_rsp_t *rsp = (bfa_fwi_tskim_rsp_t *) m;
	bfa_hal_tskim_t *tskim;
	uint16_t        tsk_tag = bfa_os_ntohs(rsp->tsk_tag);

	tskim = BFA_HAL_TSKIM_FROM_TAG(fcpim, tsk_tag);
	bfa_assert(tskim->tsk_tag == tsk_tag);

	tskim->tsk_status = rsp->tsk_status;

	/**
	 * Firmware sends BFA_TSKIM_STS_TM_ABORTED status for abort
	 * requests. All other statuses are for normal completions.
	 */
	if (rsp->tsk_status == BFA_TSKIM_STS_TM_ABORTED) {
		bfa_stats(tskim->itnim, tm_cleanup_comps);
		bfa_sm_send_event(tskim, BFA_HAL_TSKIM_SM_CLEANUP_DONE);
	} else {
		bfa_stats(tskim->itnim, tm_fw_rsps);
		bfa_sm_send_event(tskim, BFA_HAL_TSKIM_SM_DONE);
	}
}

/**
 * @}
 */

/**
 * @ig hal_tskim_api
 * @{
 */

/**
 * @brief
 *
 */
bfa_hal_tskim_t *
bfa_hal_tskim_alloc(bfa_hal_t *hal, bfad_tskim_t *dtsk)
{
	bfa_hal_fcpim_mod_t *fcpim = BFA_HAL_FCPIM_MOD(hal);
	bfa_hal_tskim_t *tskim;

	bfa_q_deq(&fcpim->tskim_free_q, &tskim);

	if (!tskim)
		bfa_hal_fcpim_stats(tskim->fcpim, no_tskims);
	else
		tskim->dtsk = dtsk;

	return tskim;
}

void
bfa_hal_tskim_free(bfa_hal_tskim_t *tskim)
{
	bfa_assert(bfa_q_is_on_q_func(&tskim->itnim->tsk_q, &tskim->qe));
	bfa_q_qe_deq(tskim);
	bfa_q_enq(&tskim->fcpim->tskim_free_q, tskim);
}

/**
 * @brief
 *      Start a task management command.
 *
 * @param[in]       tskim       HAL task management command instance
 * @param[in]       itnim       i-t nexus for the task management command
 * @param[in]       lun         lun, if applicable
 * @param[in]       tm_cmnd     Task management command code.
 * @param[in]       t_secs      Timeout in seconds
 *
 * @return None.
 */
void
bfa_hal_tskim_start(bfa_hal_tskim_t *tskim, bfa_hal_itnim_t *itnim, lun_t lun,
		    fcp_tm_cmnd_t tm_cmnd, uint8_t tsecs)
{
	tskim->itnim   = itnim;
	tskim->lun     = lun;
	tskim->tm_cmnd = tm_cmnd;
	tskim->tsecs   = tsecs;
	tskim->notify  = BFA_FALSE;
	bfa_stats(itnim, tm_cmnds);

	bfa_q_enq(&itnim->tsk_q, tskim);
	bfa_sm_send_event(tskim, BFA_HAL_TSKIM_SM_START);
}

/**
 * @}
 */
