/*
 * Copyright (c) 2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

#ifndef __HAL_FCXP_H__
#define __HAL_FCXP_H__

#include <cs/bfa_sm.h>
#include <protocol/fc.h>
#include <hal/bfa_hal_fcxp.h>
#include <fwi/bfa_fwi_fcxp.h>

#define BFA_FCXP_MAX_IBUF_SZ (2112 + sizeof(fc_vft_t) + sizeof(fchs_t))

/**
 * @todo hal private
 */
typedef struct {
	uint8_t         d[BFA_FCXP_MAX_IBUF_SZ];
} bfa_hal_fcxp_buf_t;

typedef struct bfa_hal_fcxp_mod {
	bfa_hal_t      *hal;	/* !< backpointer to HAL */
	bfa_hal_fcxp_t *fcxp_list;	/* !< array of FCXPs */
	uint16_t        num_fcxps;	/* !< max num FCXP requests */
	bfa_q_t         fcxp_free_q;	/* !< free FCXPs */
	bfa_q_t         fcxp_active_q;	/* !< active FCXPs */
	bfa_hal_fcxp_buf_t *req_pld_list_kva;	/* !< list of FCXP req pld */
	uint64_t        req_pld_list_pa;	/* !< list of FCXP req pld */
	bfa_hal_fcxp_buf_t *rsp_pld_list_kva;	/* !< list of FCXP resp pld */
	uint64_t        rsp_pld_list_pa;	/* !< list of FCXP resp pld */
	bfa_q_t         wait_q;	/* !< wait queue for free fcxp */
} bfa_hal_fcxp_mod_t;

#define BFA_HAL_FCXP_MOD(__hal)			(&(__hal)->modules.fcxp_mod)
#define BFA_HAL_FCXP_FROM_TAG(__mod, __tag)	(&(__mod)->fcxp_list[__tag])

typedef void    (*fcxp_send_cb_t) (bfa_hal_t *ioc, bfa_hal_fcxp_t *fcxp,
				   void *cb_arg, bfa_status_t req_status,
				   uint32_t rsp_len, uint32_t resid_len,
				   fchs_t *rsp_fchs);

/**
 * Information needed for a FCXP request
 */
typedef struct {
	bfa_hal_rport_t *hal_rport;	/* !< Pointer to the hal rport that was
					 *returned from bfa_hal_rport_create().
					 *This could be left NULL for WKA or for
					 *FCXP interactions before the rport
					 *nexus is established
					 */
	uint8_t         cts;	/* !< continous sequence */
	uint8_t         class;	/* !< FC class for the request/response */
	uint16_t        max_frmsz;	/* !< max send frame size */
	uint16_t        vf_id;	/* !< vsan tag if applicable */
	fchs_t          fchs;	/* !< request FC header structure */
	uint32_t        req_tot_len;	/* !< request payload total length */
} bfa_fcxp_req_info_t;

typedef struct bfa_fcxp_rsp_info {
	fchs_t          rsp_fchs;	/* !< Response frame's FC header will
					 *be *sent back in this field */
	uint8_t         rsp_timeout;	/* !< timeout in seconds, 0-no response
					 */
	uint8_t         rsvd2[3];
	uint32_t        rsp_maxlen;	/*!< max response length expected */
} bfa_fcxp_rsp_info_t;

struct bfa_hal_fcxp {
	bfa_q_t		qe;		/*!< fcxp queue element */
	bfa_sm_t        sm;		/*!< state machine */
	void           	*caller;	/*!< driver or fcs */
	bfa_hal_fcxp_mod_t *fcxp_mod;
					/*!< back pointer to fcxp mod */
	uint16_t        fcxp_tag;	/*!< internal tag */
	bfa_fcxp_req_info_t req_info;
					/*!< request info */
	bfa_fcxp_rsp_info_t rsp_info;
					/*!< response info */
	uint8_t 	use_ireqbuf;	/*!< use internal req buf */
	uint8_t         use_irspbuf;	/*!< use internal rsp buf */
	uint32_t        nreq_sgles;	/*!< num request SGLEs */
	uint32_t        nrsp_sgles;	/*!< num response SGLEs */
	bfa_q_t         req_sgpg_q;	/*!< SG pages for request buf */
	bfa_q_t         req_sgpg_wqe;	/*!< wait queue for req SG page */
	bfa_q_t         rsp_sgpg_q;	/*!< SG pages for response buf */
	bfa_q_t         rsp_sgpg_wqe;	/*!< wait queue for rsp SG page */

	bfa_fcxp_get_sgaddr_t	req_sga_cbfn;
					/*!< SG elem addr user function */
	bfa_fcxp_get_sglen_t	req_sglen_cbfn;
					/*!< SG elem len user function */
	bfa_fcxp_get_sgaddr_t	rsp_sga_cbfn;
					/*!< SG elem addr user function */
	bfa_fcxp_get_sglen_t	rsp_sglen_cbfn;
					/*!< SG elem len user function */
	bfa_hcb_fcxp_send_t	send_cbfn;
					/*!< send completion callback */

	void		*send_cbarg;	/*!< callback arg */
	bfa_hal_sge_t   req_sge[BFA_FCXP_MAX_SGES];
					/*!< req SG elems */
	bfa_hal_sge_t   rsp_sge[BFA_FCXP_MAX_SGES];
					/*!< rsp SG elems */
	uint8_t         rsp_status;	/*!< comp: rsp status */
	uint32_t        rsp_len;	/*!< comp: actual response len */
	uint32_t        residue_len;	/*!< comp: residual rsp length */
	fchs_t          rsp_fchs;	/*!< comp: response fchs */
	bfa_hcb_qe_t    hcb_qe;		/*!< comp: callback qelem */
};

#define BFA_FCXP_REQ_PLD(_fcxp) 	(bfa_hal_fcxp_get_reqbuf(_fcxp))

#define BFA_FCXP_RSP_FCHS(_fcxp) 	(&((_fcxp)->rsp_info.fchs))
#define BFA_FCXP_RSP_PLD(_fcxp) 	(bfa_hal_fcxp_get_rspbuf(_fcxp))

#define BFA_FCXP_REQ_PLD_PA(_fcxp)					\
	((_fcxp)->fcxp_mod->req_pld_list_pa +				\
		(sizeof(bfa_hal_fcxp_buf_t)  * (_fcxp)->fcxp_tag))

#define BFA_FCXP_RSP_PLD_PA(_fcxp) 					\
	((_fcxp)->fcxp_mod->rsp_pld_list_pa +				\
		(sizeof(bfa_hal_fcxp_buf_t) * (_fcxp)->fcxp_tag))

void	bfa_hal_fcxp_isr(bfa_hal_t *hal, bfa_msg_t *msg);
#endif /* __HAL_FCXP_H__ */
