/*
 * Copyright (c) 2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

/**
 * @file bfa_q.h Circular queue definitions.
 */

#ifndef __BFA_Q_H__
#define __BFA_Q_H__

typedef struct bfa_q {
	struct bfa_q	*next;
	struct bfa_q	*prev;
} bfa_q_t;

#define bfa_q_first(_q) ((void *)(((bfa_q_t *) (_q))->next))
#define bfa_q_next(_qe)	(((bfa_q_t *) (_qe))->next)
#define bfa_q_prev(_qe) (((bfa_q_t *) (_qe))->prev)

/*
 * bfa_q_init - to initialize a queue
 */
#define bfa_q_init(_q) {						\
	bfa_q_next(_q) = (bfa_q_t *) (_q);				\
	bfa_q_prev(_q) = (bfa_q_t *) (_q);				\
}

/*
 * bfa_q_qe_init - to initialize a queue element
 */
#define bfa_q_qe_init(_qe) {						\
	bfa_q_next(_qe) = (bfa_q_t *) NULL;				\
	bfa_q_prev(_qe) = (bfa_q_t *) NULL;				\
}

/*
 * bfa_q_enq - enqueue an element at the end of the list
 */
#define bfa_q_enq(_q, _qe) {						\
	bfa_assert_fp((bfa_q_next(_qe) == NULL) && 			\
					(bfa_q_prev(_qe) == NULL));	\
	bfa_q_next(_qe) = (bfa_q_t *) (_q);				\
	bfa_q_prev(_qe) = bfa_q_prev(_q);				\
	bfa_q_next(bfa_q_prev(_q)) = (bfa_q_t *) (_qe);			\
	bfa_q_prev(_q) = (bfa_q_t *) (_qe);				\
}

/*
 * bfa_q_enq_head - enqueue an element at the head of queue
 */
#define bfa_q_enq_head(_q, _qe) {					\
	bfa_assert_fp((bfa_q_next(_qe) == NULL) && 			\
					(bfa_q_prev(_qe) == NULL));	\
	bfa_q_next(_qe) = bfa_q_next(_q);				\
	bfa_q_prev(_qe) = (bfa_q_t *) (_q);				\
	bfa_q_prev(bfa_q_next(_q)) = (bfa_q_t *) (_qe);			\
	bfa_q_next(_q) = (bfa_q_t *) (_qe);				\
}

/*
 * bfa_q_is_empty - returns TRUE if queue is empty
 */
#define bfa_q_is_empty(_q)						\
	(bfa_q_next(_q) == ((bfa_q_t *) (_q)))

/*
 * bfa_q_enq_q - enqueue another queue at the tail
 */
#define bfa_q_enq_q(_dstq, _srcq) {					\
	if (!bfa_q_is_empty(_srcq)) {					\
		bfa_q_next(bfa_q_prev(_srcq)) = (bfa_q_t *) (_dstq);	\
		bfa_q_prev(bfa_q_next(_srcq)) = bfa_q_prev(_dstq);	\
		bfa_q_next(bfa_q_prev(_dstq)) = bfa_q_next(_srcq);	\
		bfa_q_prev(_dstq) = bfa_q_prev(_srcq);			\
		bfa_q_init(_srcq);					\
	}								\
}

/*
 * bfa_q_enq_q_head - enqueue another queue at the head
 */
#define bfa_q_enq_q_head(_dstq, _srcq) {				\
	if (!bfa_q_is_empty(_srcq)) {					\
		bfa_q_next(bfa_q_prev(_srcq)) = bfa_q_next(_dstq);	\
		bfa_q_prev(bfa_q_next(_srcq)) = (bfa_q_t *) (_dstq);	\
		bfa_q_prev(bfa_q_next(_dstq)) = bfa_q_prev(_srcq);	\
		bfa_q_next(_dstq) = bfa_q_next(_srcq);			\
		bfa_q_init(_srcq);					\
	}								\
}

/*
 * bfa_q_qe_deq - dequeue a queue element from a queue
 */
#define bfa_q_qe_deq(_qe) {						\
	bfa_assert_fp(bfa_q_next(_qe) && bfa_q_prev(_qe));		\
	bfa_q_next(bfa_q_prev(_qe)) = bfa_q_next(_qe);			\
	bfa_q_prev(bfa_q_next(_qe)) = bfa_q_prev(_qe);			\
	BFA_Q_DBG_INIT(_qe);						\
}

/*
 * bfa_q_deq - dequeue an element from head of the queue
 */
#define bfa_q_deq(_q, _qe) {						\
	if (!bfa_q_is_empty(_q)) {					\
		(*((bfa_q_t **) (_qe))) = bfa_q_next(_q);		\
		bfa_q_prev(bfa_q_next(*((bfa_q_t **) _qe))) =		\
						(bfa_q_t *) (_q);	\
		bfa_q_next(_q) = bfa_q_next(*((bfa_q_t **) _qe));	\
		BFA_Q_DBG_INIT(*((bfa_q_t **) _qe));			\
	} else {							\
		*((bfa_q_t **) (_qe)) = (bfa_q_t *) NULL;		\
	}								\
}

/*
 * bfa_q_deq_tail - dequeue an element from tail of the queue
 */
#define bfa_q_deq_tail(_q, _qe) {					\
	if (!bfa_q_is_empty(_q)) {					\
		*((bfa_q_t **) (_qe)) = bfa_q_prev(_q);			\
		bfa_q_next(bfa_q_prev(*((bfa_q_t **) _qe))) = 		\
						(bfa_q_t *) (_q);	\
		bfa_q_prev(_q) = bfa_q_prev(*(bfa_q_t **) _qe);		\
		BFA_Q_DBG_INIT(*((bfa_q_t **) _qe));			\
	} else {							\
		*((bfa_q_t **) (_qe)) = (bfa_q_t *) NULL;		\
	}								\
}

/*
 * bfa_q_mv - move all queue elements from one to another queue
 */
#define bfa_q_mv(_srcq, _dstq) {					\
	bfa_assert_fp(bfa_q_is_empty(_dstq));				\
									\
	if (!bfa_q_is_empty(_srcq)) {					\
		bfa_q_next(_dstq) = bfa_q_next(_srcq);			\
		bfa_q_prev(_dstq) = bfa_q_prev(_srcq);			\
		bfa_q_prev(bfa_q_next(_srcq)) = (bfa_q_t *) _dstq;	\
		bfa_q_next(bfa_q_prev(_srcq)) = (bfa_q_t *) _dstq;	\
		bfa_q_init(_srcq);					\
	}								\
}

/*
 * bfa_q_iter - iterator macros, use only if it is known that list
 * does not change.
 */
#define bfa_q_iter(_q, _qe)						\
	for (_qe = bfa_q_first(_q); _qe != (_q); _qe = bfa_q_next(_qe))

#define bfa_q_iter_safe(_q, _qe, _qe_next) 				\
	for (_qe = bfa_q_first(_q), _qe_next = bfa_q_next(_qe); 	\
		 _qe != (_q); _qe = _qe_next, _qe_next = bfa_q_next(_qe))

/*
 * #ifdef BFA_DEBUG (Using bfa_assert to check for debug_build is not
 * consistent across modules)
 */
#ifndef BFA_PERF_BUILD
#define BFA_Q_DBG_INIT(_qe)	bfa_q_qe_init(_qe)
#else
#define BFA_Q_DBG_INIT(_qe)
#endif

#define bfa_q_is_on_q(_q, _qe)	bfa_q_is_on_q_func(_q, (bfa_q_t *)(_qe))
extern int bfa_q_is_on_q_func(bfa_q_t *q, bfa_q_t *qe);

#endif
