/*
 * Copyright (c) 2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */
#ifndef __BFA_TRC_H__
#define __BFA_TRC_H__

#include <bfa_os_inc.h>

#ifndef BFA_TRC_MAX
#define BFA_TRC_MAX	(4 * 1024)
#endif

#ifndef BFA_TRC_TS
#define BFA_TRC_TS(_trcm)	((_trcm)->ticks ++)
#endif

typedef struct {
#ifdef __BIGENDIAN
	uint16_t	fileno;
	uint16_t	line;
#else
	uint16_t	line;
	uint16_t	fileno;
#endif
	uint32_t	timestamp;
	union {
		struct {
			uint32_t	rsvd;
			uint32_t	u32;
		} u32;
		uint64_t	u64;
	} data;
} bfa_trc_t;

/**
 * @brief Embed this in main driver structure
 */
typedef struct {
	uint32_t	head;
	uint32_t	tail;
	uint32_t	ntrc;
	uint32_t	stopped;
	uint32_t	ticks;
	uint32_t	rsvd[3];
	bfa_trc_t	trc[BFA_TRC_MAX];
} bfa_trc_mod_t;

/**
 * @brief separate modules that are traced
 */
enum {
	BFA_TRC_FW   = 1,	/* !< firmware modules */
	BFA_TRC_HAL  = 2,	/* !< BFA HAL modules */
	BFA_TRC_FCS  = 3,	/* !< BFA FCS modules */
	BFA_TRC_LDRV = 4,	/* !< Linux driver modules */
	BFA_TRC_SDRV = 5,	/* !< Solaris driver modules */
	BFA_TRC_VDRV = 6,	/* !< vmware driver modules */
	BFA_TRC_WDRV = 7,	/* !< windows driver modules */
	BFA_TRC_AEN  = 8,	/* !< AEN module */
	BFA_TRC_BIOS = 9,	/* !< bios driver modules */
	BFA_TRC_EFI  = 10,	/* !< EFI driver modules */
};
#define BFA_TRC_MOD_SH	12
#define BFA_TRC_MOD(__mod)	((BFA_TRC_ ## __mod) << BFA_TRC_MOD_SH)

/**
 * @brief
 * Define a new tracing file (module). Module should match one defined above.
 */
#define BFA_TRC_FILE(__mod, __submod)					\
	static int __trc_fileno = ((BFA_TRC_ ## __mod ## _ ## __submod) | \
						 BFA_TRC_MOD(__mod))

/**
 * @brief Trace a word
 */
#define bfa_trc32(_trcp, _data)	\
	__bfa_trc((_trcp)->trcmod, __trc_fileno, __LINE__, (uint32_t)_data)

/**
 * @brief Trace a long word (64 bits)
 */
#ifndef BFA_BOOT_BUILD
#define bfa_trc(_trcp, _data)	\
	__bfa_trc((_trcp)->trcmod, __trc_fileno, __LINE__, (uint64_t)_data)
#else
#define bfa_trc(_trcp, _data)	\
	bfa_boot_trc((_trcp)->trcmod, __trc_fileno, __LINE__, (uint32_t)_data)
#endif

/**
 * @brief Tracing initialize or restart.
 */
static inline void
bfa_trc_init(bfa_trc_mod_t *trcm)
{
	trcm->head = trcm->tail = trcm->stopped = 0;
	trcm->ntrc = BFA_TRC_MAX;
}

/**
 * @brief Stop tracing.
 */
static inline void
bfa_trc_stop(bfa_trc_mod_t *trcm)
{
	trcm->stopped = 1;
}

#ifdef CBFW
extern void dc_flush(void *data);
#else
#define dc_flush(data)
#endif

/**
 * @brief Internal - use bfa_trc() or bfa_trcl()
 */
static inline void
__bfa_trc(bfa_trc_mod_t *trcm, int fileno, int line, uint64_t data)
{
	int		tail = trcm->tail;
	bfa_trc_t 	*trc = &trcm->trc[tail];

	if (trcm->stopped)
		return;

	trc->fileno = (uint16_t) fileno;
	trc->line = (uint16_t) line;
	trc->data.u64 = data;
	trc->timestamp = BFA_TRC_TS(trcm);
	dc_flush(trc);

	trcm->tail = (trcm->tail + 1) & (BFA_TRC_MAX - 1);
	if (trcm->tail == trcm->head)
		trcm->head = (trcm->head + 1) & (BFA_TRC_MAX - 1);
	dc_flush(trcm);
}

/**
 * @brief Internal - use bfa_trc() or bfa_trcl()
 */
static inline void
__bfa_trc32(bfa_trc_mod_t *trcm, int fileno, int line, uint32_t data)
{
	int		tail = trcm->tail;
	bfa_trc_t	*trc = &trcm->trc[tail];

	if (trcm->stopped)
		return;

	trc->fileno = (uint16_t) fileno;
	trc->line = (uint16_t) line;
	trc->data.u32.u32 = data;
	trc->timestamp = BFA_TRC_TS(trcm);
	dc_flush(trc);

	trcm->tail = (trcm->tail + 1) & (BFA_TRC_MAX - 1);
	if (trcm->tail == trcm->head)
		trcm->head = (trcm->head + 1) & (BFA_TRC_MAX - 1);
	dc_flush(trcm);
}

#ifndef BFA_PERF_BUILD
#define bfa_trc_fp(_trcp, _data)	bfa_trc(_trcp, _data)
#else
#define bfa_trc_fp(_trcp, _data)
#endif

#endif /* __BFA_TRC_H__ */

