/*
 * Copyright (c)  2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

#ifndef __BFA_DEFS_IOC_H__
#define __BFA_DEFS_IOC_H__

#include <protocol/types.h>
#include <defs/bfa_defs_types.h>
#include <defs/bfa_defs_version.h>
#include <defs/bfa_defs_adapter.h>
#include <defs/bfa_defs_pm.h>

/**
 * IOC driver attributes
 */
enum {
	BFA_IOC_DRIVER_LEN = 16,
};

/**
 * Driver and firmware versions.
 */
typedef struct {
	char            driver[BFA_IOC_DRIVER_LEN];	/*!< driver name */
	char            driver_ver[BFA_VERSION_LEN];	/*!< driver version */
	char            fw_ver[BFA_VERSION_LEN];	/*!< firmware version*/
	char            bios_ver[BFA_VERSION_LEN];	/*!< bios version */
	char            efi_ver[BFA_VERSION_LEN];	/*!< EFI version */
	char            ob_ver[BFA_VERSION_LEN];	/*!< openboot version*/
} bfa_ioc_driver_attr_t;

/**
 * IOC chip attributes
 */
enum {
	BFA_IOC_CHIP_NAME_LEN 	= 16,
	BFA_IOC_CHIP_REV_LEN 	= 8,
};

#define BFA_CHIP_NAME		"Crossbow"

typedef struct {
	uint8_t         chip_name[BFA_IOC_CHIP_NAME_LEN];/*!< chip name */
	uint8_t         chip_rev[BFA_IOC_CHIP_REV_LEN];	 /*!< chip revision */
} bfa_ioc_chip_attr_t;

/**
 * IOC PCI device attributes
 */
typedef struct {
	uint16_t        vendor_id;	/*!< PCI vendor ID */
	uint16_t        device_id;	/*!< PCI device ID */
	uint16_t        ssid;		/*!< subsystem ID */
	uint16_t        ssvid;		/*!< subsystem vendor ID */
	uint32_t        pcifn;		/*!< PCI device function */
	uint32_t        rsvd;		/* padding */
} bfa_ioc_pci_attr_t;

#define BFA_IOC_INTR_DELAY	1125
#define BFA_IOC_INTR_LATENCY	225
/**
 * Interrupt coalescing configuration.
 */
typedef struct {
	bfa_boolean_t	coalesce;	/*!< enable/disable coalescing */
	uint16_t	latency;	/*!< latency in microseconds   */
	uint16_t	delay;		/*!< delay in microseconds     */
} bfa_ioc_intr_attr_t;

/**
 * IOC firmware configuraton
 */
typedef struct {
	uint8_t         num_cqs;	/*!< number of req/rsp queues */
	uint8_t         rsvd;		/*!< padding */
	uint16_t        num_itnims;	/*!< number of i-t-nexus initiator
					 *    mode */
	uint16_t        num_ioim_reqs;	/*!< number of outstanding IO reqs */
	uint16_t        num_tskim_reqs;	/*!< active task management requests */
	uint16_t        num_fcxp_reqs;	/*!< number of active FC transport
					 *    reqs */
	uint16_t        num_uf_bufs;	/*!< number of unsolicited bufs posted
					 */
	uint16_t        num_rports;	/*!< number of unsolicited bufs posted
					 */
	uint16_t        num_itntms;	/*!< number of target mode I-T nexuses
					 */
} bfa_ioc_fwcfg_t;

/**
 * IOC configuration
 */
typedef struct {
	bfa_ioc_fwcfg_t fwcfg;		/*!< firmware side configuration */
	uint16_t        num_reqq_elems;	/*!< number of req queue elements */
	uint16_t        num_rspq_elems;	/*!< number of rsp queue elements */
	uint16_t        num_sgpgs;	/*!< number of total SG pages */
	uint16_t        num_sboot_tgts;	/*!< number of SAN boot targets */
	uint16_t        num_sboot_luns;	/*!< number of SAN boot luns */
	uint16_t        rsvd[2];
	uint16_t        path_tov;	/*!< path timeout in non-mpio mode */
	bfa_boolean_t   mpio_mode;	/*!< Enable fast failover */
	bfa_boolean_t   ioc_recover;	/*!< IOC recovery mode */
} bfa_ioc_cfg_t;

/**
 * IOC states
 */
typedef enum {
	BFA_IOC_RESET       = 1,  /*!< IOC is in reset state */
	BFA_IOC_SEMWAIT     = 2,  /*!< Waiting for IOC hardware semaphore */
	BFA_IOC_HWINIT 	    = 3,  /*!< IOC hardware is being initialized */
	BFA_IOC_CFGINIT     = 4,  /*!< IOC is being configured */
	BFA_IOC_CFGDONE     = 5,  /*!< IOC is configured */
	BFA_IOC_OPERATIONAL = 6,  /*!< IOC is operational */
	BFA_IOC_HWFAIL      = 7,  /*!< IOC hardware failure */
	BFA_IOC_HBFAIL      = 8,  /*!< IOC heart-beat failure */
	BFA_IOC_RECOVER     = 9,  /*!< IOC recovery in progress */
	BFA_IOC_DISABLING   = 10, /*!< IOC is being disabled */
	BFA_IOC_DISABLED    = 11, /*!< IOC is disabled */
} bfa_ioc_state_t;

/**
 * IOC statistics
 */
typedef struct {
	uint32_t        hb_count;

	uint32_t        disable_reqs;
	uint32_t        enable_reqs;

	uint32_t        disable_replies;
	uint32_t        enable_replies;

	/*
	 * TODO - remove ....
	 */
	uint32_t        nsuspends;	/*!< number of IOC suspends */
	uint32_t        rsvd;		/* padding */
} bfa_ioc_stats_t;

/**
 * IOC attributes returned in queries
 */
typedef struct {
	bfa_ioc_cfg_t   	fwcfg;		/*!< IOC firmware config */
	bfa_ioc_driver_attr_t 	driver_attr; 	/*!< driver attributes */
	bfa_ioc_state_t 	state;		/*!< IOC current state */
	bfa_pm_ds_t     	power_attr;	/*!< IOC power attribute */
	bfa_adapter_attr_t 	adapter_attr;	/*!< HBA attributes */
	bfa_ioc_chip_attr_t 	chip_attr;	/*!< chip level attributes */
	bfa_ioc_pci_attr_t 	pci_attr;	/*!< PCI device attributes */
	bfa_ioc_intr_attr_t 	intr_attr;	/*!< IOC interrupt attributes*/
} bfa_ioc_attr_t;

/**
 * BFA IOC level events
 */
typedef enum {
	BFA_IOC_AEN_HBGOOD  = 1,	/*!< Heart Beat restore event */
	BFA_IOC_AEN_HBFAIL  = 2,	/*!< Heart Beat failure event */
	BFA_IOC_AEN_ENABLE  = 3,	/*!< IOC enabled event */
	BFA_IOC_AEN_DISABLE = 4,	/*!< IOC disabled event */
} bfa_ioc_aen_event_t;

/**
 * BFA IOC level event data, now just a place holder
 */
typedef struct {
	wwn_t           pwwn;
} bfa_ioc_aen_data_t;

#define BFA_IOC_MAX_PATH_FAILOVER_TOV	60
#define BFA_IOC_MAX_Q_DEPTH		2000

#endif /* __BFA_DEFS_IOC_H__ */
