/*
 * Copyright (c)  2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

#ifndef __BFA_DEFS_PORT_H__
#define __BFA_DEFS_PORT_H__

#include <bfa_os_inc.h>
#include <protocol/types.h>
#include <defs/bfa_defs_pport.h>

/**
 * symbolic names for base port/virtual port
 */
#define BFA_SYMNAME_MAXLEN	128	/* vmware/windows uses 128 bytes */
typedef struct {
	char            symname[BFA_SYMNAME_MAXLEN];
} bfa_port_symname_t;

/**
* Roles of FCS port:
 *     - FCP IM and FCP TM roles cannot be enabled together for a FCS port
 *     - Create multiple ports if both IM and TM functions required.
 *     - Atleast one role must be specified.
 */
typedef enum {
	BFA_PORT_ROLE_FCP_IM 	= 0x01,	/*!< FCP initiator role */
	BFA_PORT_ROLE_FCP_TM 	= 0x02,	/*!< FCP target role */
	BFA_PORT_ROLE_FCP_IPFC 	= 0x04,	/*!< IP over FC role */
	BFA_PORT_ROLE_FCP_MAX 	= BFA_PORT_ROLE_FCP_IPFC
} bfa_port_role_t;

/**
 * FCS port configuration.
 */
typedef struct {
	wwn_t           	pwwn;		/*!< port wwn */
	wwn_t           	nwwn;		/*!< node wwn */
	bfa_port_symname_t 	sym_name;	/*!< vm port symbolic name */
	bfa_port_role_t 	roles;		/*!< FCS port roles */
	uint32_t        	rsvd;
} bfa_port_cfg_t;

/**
 * FCS port states
 */
typedef enum {
	BFA_PORT_UNINIT  = 0,	/*!< PORT is not yet initialized */
	BFA_PORT_FDISC   = 1,	/*!< FDISC is in progress */
	BFA_PORT_ONLINE  = 2,	/*!< login to fabric is complete */
	BFA_PORT_OFFLINE = 3,	/*!< No login to fabric */
} bfa_port_state_t;

/**
 * FCS port type. Required for VmWare.
 */
typedef enum {
	BFA_PORT_TYPE_PHYSICAL = 0,
	BFA_PORT_TYPE_VIRTUAL,
} bfa_port_type_t;

/**
 * FCS port offline reason. Required for VmWare.
 */
typedef enum {
	BFA_PORT_OFFLINE_UNKNOWN = 0,
	BFA_PORT_OFFLINE_LINKDOWN,
	BFA_PORT_OFFLINE_FAB_UNSUPPORTED,	/*!< NPIV not supported by the
						 *    fabric */
	BFA_PORT_OFFLINE_FAB_NORESOURCES,
	BFA_PORT_OFFLINE_FAB_LOGOUT,
} bfa_port_offline_reason_t;

/**
 * FCS lport info. Required for VmWare.
 */
typedef struct {
	uint8_t         port_type;	/* bfa_port_type_t : physical or
					 * virtual */
	uint8_t         port_state;	/* one of bfa_port_state_t values */
	uint8_t         offline_reason;	/* one of bfa_port_offline_reason_t
					 * values */
	wwn_t           port_wwn;
	wwn_t           node_wwn;

	/*
	 * following 4 feilds are valid for Physical Ports only
	 */
	uint32_t        max_vports_supp;	/* Max supported vports */
	uint32_t        num_vports_inuse;	/* Num of in use vports */
	uint32_t        max_rports_supp;	/* Max supported rports */
	uint32_t        num_rports_inuse;	/* Num of doscovered rports */

} bfa_port_info_t;

/**
 * FCS port statistics
 */
typedef struct {
	uint32_t        ns_plogi_sent;
	uint32_t        ns_plogi_rsp_err;
	uint32_t        ns_plogi_acc_err;
	uint32_t        ns_plogi_accepts;
	uint32_t        ns_rejects;	/* NS command rejects */
	uint32_t        ns_plogi_unknown_rsp;
	uint32_t        ns_plogi_alloc_wait;

	uint32_t        ns_retries;	/* NS command retries */
	uint32_t        ns_timeouts;	/* NS command timeouts */

	uint32_t        ns_rspnid_sent;
	uint32_t        ns_rspnid_accepts;
	uint32_t        ns_rspnid_rsp_err;
	uint32_t        ns_rspnid_rejects;
	uint32_t        ns_rspnid_alloc_wait;

	uint32_t        ns_rftid_sent;
	uint32_t        ns_rftid_accepts;
	uint32_t        ns_rftid_rsp_err;
	uint32_t        ns_rftid_rejects;
	uint32_t        ns_rftid_alloc_wait;

	uint32_t	ns_rffid_sent;
	uint32_t	ns_rffid_accepts;
	uint32_t	ns_rffid_rsp_err;
	uint32_t	ns_rffid_rejects;
	uint32_t	ns_rffid_alloc_wait;

	uint32_t        ns_gidft_sent;
	uint32_t        ns_gidft_accepts;
	uint32_t        ns_gidft_rsp_err;
	uint32_t        ns_gidft_rejects;
	uint32_t        ns_gidft_unknown_rsp;
	uint32_t        ns_gidft_alloc_wait;

	/*
	 * Mgmt Server stats
	 */
	uint32_t        ms_retries;	/* MS command retries */
	uint32_t        ms_timeouts;	/* MS command timeouts */
	uint32_t        ms_plogi_sent;
	uint32_t        ms_plogi_rsp_err;
	uint32_t        ms_plogi_acc_err;
	uint32_t        ms_plogi_accepts;
	uint32_t        ms_rejects;	/* NS command rejects */
	uint32_t        ms_plogi_unknown_rsp;
	uint32_t        ms_plogi_alloc_wait;

	uint32_t        num_rscn;	/* Num of RSCN received */
	uint32_t        num_portid_rscn;/* Num portid format RSCN
					 * received */

	uint32_t	uf_recvs; 	/* unsolicited recv frames      */
	uint32_t	uf_recv_drops; 	/* dropped received frames	*/

	uint32_t	rsvd; 		/* padding for 64 bit alignment */
} bfa_port_stats_t;

/**
 * BFA port attribute returned in queries
 */
typedef struct {
	bfa_port_state_t state;		/*!< port state */
	uint32_t         pid;		/*!< port ID */
	bfa_port_cfg_t   port_cfg;	/*!< port configuration */
	bfa_pport_type_t port_type;	/*!< current topology */
	uint32_t         loopback;	/*!< cable is externally looped back */
} bfa_port_attr_t;

/**
 * BFA physical port Level events
 * Arguments below are in BFAL context from Mgmt
 * BFA_PORT_AEN_ONLINE:     [in]: pwwn	[out]: pwwn
 * BFA_PORT_AEN_OFFLINE:    [in]: pwwn	[out]: pwwn
 * BFA_PORT_AEN_RLIR:       [in]: None	[out]: pwwn, rlir_data, rlir_len
 * BFA_PORT_AEN_SFP_INSERT: [in]: pwwn	[out]: pwwn
 * BFA_PORT_AEN_SFP_REMOVE: [in]: pwwn	[out]: pwwn
 * BFA_PORT_AEN_SFP_POM:    [in]: pwwn	[out]: pwwn, level
 * BFA_PORT_AEN_ENABLE:     [in]: pwwn	[out]: pwwn
 * BFA_PORT_AEN_DISABLE:    [in]: pwwn	[out]: pwwn
 * BFA_PORT_AEN_AUTH_ON:    [in]: pwwn	[out]: pwwn
 * BFA_PORT_AEN_AUTH_OFF:   [in]: pwwn	[out]: pwwn
 * BFA_PORT_AEN_DISCONNECT: [in]: pwwn	[out]: pwwn
 *
 */
typedef enum {
	BFA_PORT_AEN_ONLINE     = 1,	/*!< Physical Port online event */
	BFA_PORT_AEN_OFFLINE    = 2,	/*!< Physical Port offline event */
	BFA_PORT_AEN_RLIR       = 3,	/*!< RLIR event, not supported */
	BFA_PORT_AEN_SFP_INSERT = 4,	/*!< SFP inserted event */
	BFA_PORT_AEN_SFP_REMOVE = 5,	/*!< SFP removed event */
	BFA_PORT_AEN_SFP_POM    = 6,	/*!< SFP POM event */
	BFA_PORT_AEN_ENABLE     = 7,	/*!< Physical Port enable event */
	BFA_PORT_AEN_DISABLE    = 8,	/*!< Physical Port disable event */
	BFA_PORT_AEN_AUTH_ON    = 9,	/*!< Physical Port auth success event */
	BFA_PORT_AEN_AUTH_OFF   = 10,	/*!< Physical Port auth fail event */
	BFA_PORT_AEN_DISCONNECT = 11,	/*!< Physical Port disconnect event */
} bfa_port_aen_event_t;

typedef enum {
	BFA_PORT_AEN_SFP_POM_GREEN = 1,	/*!< Normal */
	BFA_PORT_AEN_SFP_POM_AMBER = 2,	/*!< Warning */
	BFA_PORT_AEN_SFP_POM_RED   = 3	/*!< Critical */
} bfa_port_aen_sfp_pom_t;

typedef struct {
	wwn_t           pwwn;	      /*!< WWN of the physical port */
	int             phy_port_num; /*! For SFP related events */
	bfa_port_aen_sfp_pom_t level; /*!< Only transitions will be informed */
} bfa_port_aen_data_t;

#endif /* __BFA_DEFS_PORT_H__ */
