/*
 * Copyright (c)  2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

/**
 * @page bfa_fwi_info Brocade Firmmare Interface
 *
* The transport between the host and the firmware consitutes of:
 *
 * - <b> Interface Registers </b>
 * 	- PSS Control and Status register
 * 		+ Check FLASH status and take LPUs out of RESET
 * 	- PLL Programming Register
 * 	- Interrupt Status/Control Register
 * 	- @b TODO - VRK
 *
 * - <b> Host Mailbox </b>
 * 	- One-deep FIFO used to communicate with the ASIC
 * 	- Primary use is during start-of-day initialization and for Async
 * 	  event delivery
 * 	- One mailbox for host-to-IOC communication and one mailbox for
 * 	  IOC-to-host communication
 *
 * 	- <b> Request/Response Queues </b> (resides in Host memory)
 * 	- Request/Response Queue elements can either be 64 bytes or 128 bytes
 * 	  (selected during start-of-day)
 * 	- Request/Response Queue Base address and number of elements to be
 * 	  programmed by the host driver
 * 	- Each Request/Response elements are used to send/recv msgs. Msgs are
 * 	  interpretted by the Firmware.
 *
 * @note
 * 	<b> All msg have to be formatted in Big-endian byte ordering. The SGPG
 * 	information alone needs to be in Little-endian byte ordering. </b>
 *
 *
 * 	@img fwi_transport.jpg
 *
 *
 * @section hmbox Host Mailbox details
 *
 * @note Host Mailboxes are meant to be used during initialization/rare
 * eventing. Only one msg can be outstanding in the mailbox registers.
 *
 * Mechanism for the host to send a msg to LPU through the mailbox:
 * - Write upto 8 Word long msg by writing into the Host to LPU
 *   mailbox registers
 * - Set bit 0 in the Host to LPU mailbox CMD_STAT register to indicate that a
 *   mailbox msg is ready for the LPU
 * - Either:
 * 	- poll for the bit 0 of Host to LPU mailbox CMD_STAT register to be
 * 	  cleared (by the LPU)
 * 	- @b TODO wait for interrupt (???)
 *
* Mechanism for the LPU to send a msg to host through the mailbox:
 * - Write upto 8 Word long msg by writing into the LPU to Host
 *   mailbox registers
 * - Set bit 0 in the LPU to Host mailbox CMD_STAT register to indicate that a
 *   mailbox msg is ready for the Host
* - Either:
 * 	- poll for the bit 0 of LPU to Host mailbox CMD_STAT register to be
 * 	  cleared (by the Host)
 * 	- @b TODO wait for interrupt (???)
 *
 * @section bfa_msg_classes Message Classes
 * 	- <b> IO Controller </b>
 * 		@li @ref bfa_ioc_msgs.h
 * 	- <b> Port </b>
 * 		@li @ref bfa_port_msgs.h
 * 	- <b> FC Transport </b>
 * 		@li @ref bfa_fcxp_msgs.h
 * 	- <b> FCP (Initiator) </b>
 * 		@li @ref bfa_fcp_ini_msgs.h
 * 	- <b> FCP Target Mode </b>
 * 		@li @ref bfa_fcp_tgt_msgs.h
 * 	- <b> IP over FC </b>
 * 		@li @b TBD
 * 	- <b> Diagnostics </b>
 * 		@li @b TBD
 ******************************************************************************
 */

#ifndef __BFA_FWI_H__
#define __BFA_FWI_H__

#include <defs/bfa_defs_status.h>

#pragma pack(1)

/**
 * Msg header common to all msgs
 */
typedef struct bfa_mhdr {
uint8_t         msg_class;	/*!< Msg classes: @ref bfa_mclass_t */
	uint8_t         msg_id;	/*!< Msg ID/opcode with in the class */
	union {
		struct {
			uint8_t         rsvd;
			uint8_t         lpu_id;	/*!< LPU this msg is intended
						 *for */
		} h2i;
		uint16_t        i2htok;	/*!< unique token for msgs for host */
	} mtag;
} bfa_mhdr_t;

#define bfa_fwi_h2i_set(_mh, _mc, _op, _lpuid) do {		\
	(_mh).msg_class 		= (_mc);		\
	(_mh).msg_id			= (_op);		\
	(_mh).mtag.h2i.lpu_id	= (_lpuid);			\
} while (0)

#define bfa_fwi_i2h_set(_mh, _mc, _op, _i2htok) do {		\
	(_mh).msg_class 		= (_mc);		\
	(_mh).msg_id			= (_op);		\
	(_mh).mtag.i2htok		= (_i2htok);		\
} while (0)

#define BFA_MSG_I2H_MSG_BASE	128	/*!< the msg_id value ranges: *- Host
					 *to IOC - 0-127 *- IOC to Host -
					 *128-255 */
#define BFA_I2HM(_x) 			((_x) + BFA_MSG_I2H_MSG_BASE)

/**
 ****************************************************************************
 *
 * Scatter Gather Element and Page definition
 *
 ****************************************************************************
 */

#define BFA_SGE_DATA_MAX_INLINE	1
#define BFA_SGE_MAX_INLINE		(BFA_SGE_DATA_MAX_INLINE + 1)

/**
 * SG Flags
 */
enum {
	BFA_SGE_DATA = 0,	/*!< data address, not last */
	BFA_SGE_DATA_CPL = 1,	/*!< data address, last in current page */
	BFA_SGE_DATA_LAST = 3,	/*!< data address, last */
	BFA_SGE_LINK = 2,	/*!< link address */
	BFA_SGE_PGDLEN = 2,	/*!< cumulative data length for page */
};

/**
 * DMA addresses
 */
typedef union {
	struct {
		uint32_t        addr_lo;
		uint32_t        addr_hi;
	} a32;
} bfa_addr_t;

/**
 * Scatter Gather Element
 */
typedef struct bfa_sge {
#ifdef __BIGENDIAN
uint32_t        flags:2, rsvd:2, sg_len:28;
#else
uint32_t        sg_len:28, rsvd:2, flags:2;
#endif
	bfa_addr_t      sga;
} bfa_sge_t;

/**
 * Scatter Gather Page
 */
#define BFA_SGPG_NUM_DATA_SGES	7
#define BFA_SGPG_NUM_SGES		(BFA_SGPG_NUM_DATA_SGES + 1)
#define BFA_SGPG_RSVD_WD_LEN	8

typedef struct bfa_sgpg {
	bfa_sge_t       sges[BFA_SGPG_NUM_SGES];
	union {
		uint32_t        rsvd[BFA_SGPG_RSVD_WD_LEN];
		struct {
			uint32_t        pg_num;	/*! used to lookup SGPGs on
						 *the driver side */
			uint32_t        rsvd1[BFA_SGPG_RSVD_WD_LEN - 1];
		} sw_handle;
	} pad;
} bfa_sgpg_t;

#define bfa_sge_to_be(__sge) do {				\
	((uint32_t *)(__sge))[0] = bfa_os_htonl(((uint32_t *)(__sge))[0]); \
	((uint32_t *)(__sge))[1] = bfa_os_htonl(((uint32_t *)(__sge))[1]); \
	((uint32_t *)(__sge))[2] = bfa_os_htonl(((uint32_t *)(__sge))[2]); \
} while (0)

/**
 * Given number of SG elements, BFA_SGPG_NPAGE() returns the number of
 * SG pages required.
 *
* Here is the structure of a SG element list:
 * o Inline SG entries (upto 2) of following types
 *   - BFA_SGE_DATA_LAST, BFA_SGE_PGDLEN or
 *   - BFA_SGE_LINK
 * o Following SG page
 */
#define BFA_SGPG_NPAGE(_nsges)	(((_nsges) / BFA_SGPG_NUM_DATA_SGES) + 1)

/*
 * Large Message structure - 128 Bytes size Msgs
 */
#define BFA_LMSG_SZ		128
#define BFA_LMSG_PL_WSZ ((BFA_LMSG_SZ - sizeof (bfa_mhdr_t)) / 4)

typedef struct bfa_msg {
	bfa_mhdr_t      mhdr;
	union {
		uint32_t        pl[BFA_LMSG_PL_WSZ];
	} mpl;
} bfa_msg_t;

/**
 * Message Classes
 */
typedef enum {
	BFA_MC_IOC = 1,		/*!< IO Controller (IOC) */
	BFA_MC_PORT = 2,	/*!< physical port */
	BFA_MC_FLASH = 3,	/*!< Flash message class */
	BFA_MC_UF = 4,		/*!< Unsolicited frame receive */
	BFA_MC_FCXP = 5,	/*!< FC Transport */
	BFA_MC_RPORT = 6,	/*!< Remote port */
	BFA_MC_ITNIM = 7,	/*!< I-T nexus (Initiator mode) */
	BFA_MC_IOIM_READ = 8,	/*!< read IO (Initiator mode) */
	BFA_MC_IOIM_WRITE = 9,	/*!< write IO (Initiator mode) */
	BFA_MC_IOIM_IO = 10,	/*!< IO (Initiator mode) */
	BFA_MC_IOIM = 11,	/*!< IO (Initiator mode) */
	BFA_MC_IOIM_IOCOM = 12,	/*!< good IO completion */
	BFA_MC_TSKIM = 16,	/*!< Task management (initiator mode) */
	BFA_MC_DIAG = 17,	/*!< Diagnostic Msgs */
	BFA_MC_IP = 18,		/*!< IP over FC Msgs */
	BFA_MC_SBOOT = 19,	/*!< SAN boot services */
	BFA_MC_SB = 31,		/*!< Single Byte Command Code Set(FICON) */
	BFA_MC_MAX
} bfa_mclass_t;

#define BFA_IOC_VER1		1
#define BFA_IOC_MAX_CQS		4

typedef enum {
	BFA_NUM_CQS_1 = 1,
	BFA_NUM_CQS_2 = 2,
	BFA_NUM_CQS_4 = 4,
} bfa_num_cqs_t;

/*
 *      Mailbox register definition
 *
 * TODO
 */
#define BFA_MBOX_REG_LEN 32	/* 32 bytes */

/*
 *      Interrupt status register
 *
 * TODO
 */

/*
 *      Interrupt control register
 *
 * TODO
 */

 /*
  * Interrupt coeleasing control
  *
  * TODO
  */

#pragma pack()

#endif /* __BFA_FWI_H__ */
