/*
 * Copyright (c)  2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

#ifndef __BFA_FWI_IOIM_H__
#define __BFA_FWI_IOIM_H__

#include "bfa_fwi.h"
#include <protocol/fcp.h>

#pragma pack(1)

typedef enum {
	BFA_FWI_IOIM_H2I_IOABORT_REQ = 1,	/*!< IO abort request */
	BFA_FWI_IOIM_H2I_IOCLEANUP_REQ = 2,	/*!< IO cleanup request */

} bfa_fwi_ioim_h2i_t;

typedef enum {
	BFA_FWI_IOIM_I2H_IO_RSP = BFA_I2HM(1),	/*!< non-fp IO response */
	BFA_FWI_IOIM_I2H_IOABORT_RSP = BFA_I2HM(2),	/*!< io abort response
							 */
} bfa_fwi_ioim_i2h_t;

/**
 * IO command DIF info
 */
typedef struct {

	uint32_t        dif_crc:16,	/* h/w calculated CRC */
			dif_atg:16;	/* logical block application tag */
	uint32_t        dif_rtg;	/* logical block reference tag */

#ifdef __BIGENDIAN
			/*
			 * byte 0
			 */
	uint32_t        rsvd_a:8,
			/*
			 * byte 1
			 */
			rx_last_frm:1,	/* status valid when set */
			rx_wdd_cnt_low:7, /* f/w init to 0; h/w use -
					   *lower 7 bits */
			/*
			 * byte 2
			 */
			rx_wdd_cnt_high:3,	/* f/w init to 0; h/w use -
						 *upper 3 bits */
			dif_state:2,	/* h/w state */
			blk_cnt_en:1,	/* increment ref tag when enabled */
			blk_size_sel:2,	/* block size encoding */
			/*
			 * byte 3
			 */
			dif_action:2,	/* DIF action */
			verify_msk:3,	/* for verify actions */
			replace_msk:3;	/* for verify actions */
#else
			/*
			 * byte 0
			 */
			uint32_t        rsvd_a:8,
			/*
			 * byte 1
			 */
			rx_wdd_cnt_low:7, /* f/w init to 0; h/w use -
					   *lower 7 bits */
			rx_last_frm:1,	/* status valid when set */
			/*
			 * byte 2
			 */
			blk_size_sel:2,	/* block size encoding */
			blk_cnt_en:1,	/* increment ref tag when enabled */
			dif_state:2,	/* h/w state */
			rx_wdd_cnt_high:3,	/* f/w init to 0; h/w use -
						 *upper 3 bits */
			/*
			 * byte 3
			 */
			replace_msk:3,	/* for verify actions */
			verify_msk:3,	/* for verify actions */
			dif_action:2;	/* DIF action */
#endif

#ifdef __BIGENDIAN
	uint32_t        blk_num_u:24,	/* block count */
			blk_num_l:2,
			rsvd_b:3,	/* reserved - 3 bites for h/w
					 *use */
			dif_field_err:3;	/* DIF status */
#else
	uint32_t        blk_num_u:24, dif_field_err:3,	/* DIF status */
			rsvd_b:3,	/* reserved - 3 bites for h/w use */
			blk_num_l:2;
#endif
} bfa_ioim_dif_t;

/**
 * FCP IO messages overview
 *
 * @note
 * - Max CDB length supported is 64 bytes.
 * - SCSI Linked commands and SCSI bi-directional Commands not
 * 	 supported.
 *
 */
typedef struct {
	bfa_mhdr_t      mh;	/*!< Common msg header */
	uint16_t        io_tag;	/*!< I/O tag */
	uint16_t        rport_hdl;	/*!< itnim/rport firmware handle */
	fcp_cmnd_t      cmnd;	/*!< FCP command info for I/O request */

	/**
	 * SG elements array within the IO request must be double word
	 * aligned. This aligment is required to optimize SGM setup for the IO.
	 */
	bfa_sge_t       sges[BFA_SGE_MAX_INLINE];
	uint8_t         io_timeout;
	uint8_t         dif_en;
	uint8_t         rsvd_a[2];
	bfa_ioim_dif_t  dif;
} bfa_fwi_ioim_req_t;

/**
 *  @brief
 * 	This table shows various IO status codes from firmware and their
 *	meaning. Host driver can use these status codes to further process
 *	IO completions.
 *
 * 	BFA_IOIM_STS_OK			: IO completed with error free SCSI &
 *					   transport status.
 * 					   - io-tag can be reused.
 *
 * 	BFA_IOIM_STS_SCSI_ERR		: IO completed with scsi error.
 * 					  - io-tag can be reused.
 *
 * 	BFA_IOIM_STS_IO_HOST_ABORTED	: IO was aborted successfully due to
 *					   host request.
 * 					   - io-tag cannot be reused yet.
 *
 * 	BFA_IOIM_STS_IO_ABORTED		: IO was aborted successfully
 *					   internally by f/w.
 * 					   - io-tag cannot be reused yet.
 *
 * 	BFA_IOIM_STS_IO_TIMEDOUT	: IO timedout and ABTS/RRQ is happening
*					   in the firmware and
 * 					   - io-tag cannot be reused yet.
 *
 * 	BFA_IOIM_STS_IO_SQER_NEEDED	: Firmware could not recover the IO
 *					  with sequence level error
 * 					  logic and hence host needs to retry
 *					  this IO with a different IO tag
 * 					  - io-tag cannot be used yet.
 *
 * 	BFA_IOIM_STS_IO_NEXUS_ABORT	: Second Level Error Recovery from host
 *					  is required because 2 consecutive ABTS
 *					  timedout and host needs logout and
 *					  re-login with the target
 * 					  - io-tag cannot be used yet.
 *
 * 	BFA_IOIM_STS_IO_UNDERRUN	: IO completed with SCSI status good,
 *					  but the data tranferred is less than
 *					  the fcp data length in the command.
 *					  ex. SCSI INQUIRY where transferred
 *					  data length and residue count in FCP
 *					  response accounts for total fcp-dl
 * 					  - io-tag can be reused.
 *
 * 	BFA_IOIM_STS_IO_OVERRUN		: IO completed with SCSI status good,
 *					  but the data transerred is more than
 *					  fcp data length in the command. ex.
 *					  TAPE IOs where blocks can of unequal
 *					  lengths.
 * 					  - io-tag can be reused.
 *
 * 	BFA_IOIM_STS_IO_RES_FREE	: Firmware has completed using io-tag
 *					  during abort process
 * 					  - io-tag can be reused.
 *
 * 	BFA_IOIM_STS_IO_PROTO_ERR	: Firmware detected a protocol error.
 *					  ex target sent more data than
 *					  requested, or there was data frame
 *					  loss and other reasons
 * 					  - io-tag cannot be used yet.
 *
 * 	BFA_IOIM_STS_IO_DIF_ERR		: Firwmare detected DIF error. ex: DIF
 * 					  CRC err or Ref Tag err or App tag err.
 * 					  - io-tag can be reused.
 *
 * 	BFA_IOIM_STS_TSK_MGT_ABORT	: IO was aborted because of Task
 *					  Management command from the host
 * 					  - io-tag can be reused.
 *
 * 	BFA_IOIM_STS_UNKNOWN_TAG	: Firmware does not know about this
 *					  io_tag.
 * 					  - io-tag can be reused.
 */
typedef enum {
	BFA_IOIM_STS_OK = 0,
	BFA_IOIM_STS_IO_HOST_ABORTED = 1,
	BFA_IOIM_STS_IO_ABORTED = 2,
	BFA_IOIM_STS_IO_TIMEDOUT = 3,
	BFA_IOIM_STS_IO_RES_FREE = 4,
	BFA_IOIM_STS_IO_SQER_NEEDED = 5,
	BFA_IOIM_STS_IO_PROTO_ERR = 7,
	BFA_IOIM_STS_IO_DIF_ERR = 8,
	BFA_IOIM_STS_UNKNOWN_TAG = 9,
} bfa_ioim_status_t;

/**
 * I/O abort reason codes
 */
typedef enum {
	BFA_IOIM_ABORT_RC_IO_TIMEOUT = 1,
	BFA_IOIM_ABORT_RC_LOGO_IMP = 2,
	BFA_IOIM_ABORT_RC_LOGO = 3,
	BFA_IOIM_ABORT_RC_SEQ_FAILURE = 4,
} bfa_ioim_abort_rc_t;

/**
 * I/O protocol error reason codes
 */
typedef enum {
	BFA_IOIM_PROTO_ERR_BURST_LEN_ERR = 1,
	BFA_IOIM_PROTO_ERR_CMND_ERR = 2,
	BFA_IOIM_PROTO_ERR_DATA_RO_ERR = 3,
} bfa_ioim_proto_err_rc_t;

#define BFA_IOIM_SNSLEN_MAX	64

typedef struct {
	uint16_t        io_tag;
	uint16_t        bfa_rport_hndl;
	uint8_t         io_status;	/*!< I/O status */
	uint8_t         io_err_rc;	/*!< io error reason code @ref
					 *bfa_fcpim_io_abort_rc_t @ref
					 *bfa_fcpim_io_proto_err_rc_t */
	uint16_t        sns_info_usf_idx;	/*!< Unsolicitied frame index
						 *used to store the sense
						 *info when sense info length
						 *is > 96 bytes */
	uint16_t	abort_tag;	/*!< host abort request tag           */
	uint8_t         reuse_io_tag;	/*!< IO tag can be reused             */
	uint8_t         rsvd;
} bfa_fwi_ioim_rsp_info_t;

#define BFA_FWI_IOIM_RSP_PAD					\
	((sizeof(fchs_t) - sizeof(bfa_mhdr_t) -		\
	  sizeof(bfa_fwi_ioim_rsp_info_t))/			\
	 sizeof(uint32_t))

#define BFA_FWI_IOIM_SNSLEN	(256)

/**
 * I/O response message
 */
typedef struct {
	bfa_mhdr_t      mh;	/*!< common msg header */
	bfa_fwi_ioim_rsp_info_t rsp_info;	/*!< io completion info */
	uint32_t        pad[BFA_FWI_IOIM_RSP_PAD];
	fcp_resp_t      fcp_rsp;	/*!< FCP_RSP contents */
} bfa_fwi_ioim_rsp_t;

typedef struct {
	bfa_mhdr_t      mh;             /*!< Common msg header  */
	uint16_t        io_tag;         /*!< I/O tag            */
	uint16_t        abort_tag;      /*!< unique request tag */
} bfa_fwi_ioim_abort_req_t;

#pragma pack()

#endif /* __BFA_FWI_IOIM_H__ */
