/*
 * Copyright (c)  2005-2008 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * See LICENSE.bfa for copyright and licensing details.
 */

#include <bfa_os_inc.h>
#include <bfa_os.h>
#include <cs/bfa_plog.h>
#include <cs/bfa_debug.h>

static int
plkd_validate_logrec(bfa_plog_rec_t *pl_rec)
{
	if ((pl_rec->log_type != BFA_PL_LOG_TYPE_INT)
	    && (pl_rec->log_type != BFA_PL_LOG_TYPE_STRING))
		return 1;

	if ((pl_rec->log_type != BFA_PL_LOG_TYPE_INT)
	    && (pl_rec->log_num_ints > BFA_PL_INT_LOG_SZ))
		return 1;

	return 0;
}

static void
bfa_plog_add(bfa_plog_t *plog, bfa_plog_rec_t *pl_rec)
{
	uint16_t        tail;
	bfa_plog_rec_t *pl_recp;

	if (plog->plog_enabled == 0)
		return;

	if (plkd_validate_logrec(pl_rec)) {
		bfa_assert(0);
		return;
	}

	tail = plog->tail;

	pl_recp = &(plog->plog_recs[tail]);

	bfa_os_memcpy(pl_recp, pl_rec, sizeof(bfa_plog_rec_t));

	pl_recp->tv = BFA_TRC_TS(plog);
	BFA_PL_LOG_REC_INCR(plog->tail);

	if (plog->head == plog->tail)
		BFA_PL_LOG_REC_INCR(plog->head);
}

void
bfa_plog_init(bfa_plog_t *plog)
{
	bfa_os_memset((char *)plog, 0, sizeof(bfa_plog_t));

	bfa_os_memcpy(plog->plog_sig, BFA_PL_SIG_STR, BFA_PL_SIG_LEN);
	plog->head = plog->tail = 0;
	plog->plog_enabled = 1;
}

void
bfa_plog_str(bfa_plog_t *plog, bfa_plog_mid_t mid, bfa_plog_eid_t event,
	     uint16_t misc, char *log_str)
{
	bfa_plog_rec_t  lp;

	if (plog->plog_enabled) {
		bfa_os_memset(&lp, 0, sizeof(bfa_plog_rec_t));
		lp.mid = mid;
		lp.eid = event;
		lp.log_type = BFA_PL_LOG_TYPE_STRING;
		lp.misc = misc;
		strncpy(lp.log_entry.string_log, log_str,
			BFA_PL_STRING_LOG_SZ - 1);
		lp.log_entry.string_log[BFA_PL_STRING_LOG_SZ - 1] = '\0';
		bfa_plog_add(plog, &lp);
	}
}

void
bfa_plog_intarr(bfa_plog_t *plog, bfa_plog_mid_t mid, bfa_plog_eid_t event,
		uint16_t misc, uint32_t *intarr, uint32_t num_ints)
{
	bfa_plog_rec_t  lp;
	uint32_t        i;

	if (num_ints > BFA_PL_INT_LOG_SZ)
		num_ints = BFA_PL_INT_LOG_SZ;

	if (plog->plog_enabled) {
		bfa_os_memset(&lp, 0, sizeof(bfa_plog_rec_t));
		lp.mid = mid;
		lp.eid = event;
		lp.log_type = BFA_PL_LOG_TYPE_INT;
		lp.misc = misc;

		for (i = 0; i < num_ints; i++)
			lp.log_entry.int_log[i] = intarr[i];

		lp.log_num_ints = (uint8_t) num_ints;

		bfa_plog_add(plog, &lp);
	}
}

void
bfa_plog_fchdr(bfa_plog_t *plog, bfa_plog_mid_t mid, bfa_plog_eid_t event,
	       uint16_t misc, fchs_t *fchdr)
{
	bfa_plog_rec_t  lp;
	uint32_t       *tmp_int = (uint32_t *) fchdr;
	uint32_t        ints[BFA_PL_INT_LOG_SZ];

	if (plog->plog_enabled) {
		bfa_os_memset(&lp, 0, sizeof(bfa_plog_rec_t));

		ints[0] = tmp_int[2];
		ints[1] = tmp_int[3];
		ints[2] = tmp_int[6];

		bfa_plog_intarr(plog, mid, event, misc, ints, 3);
	}
}

void
bfa_plog_fchdr_and_pl(bfa_plog_t *plog, bfa_plog_mid_t mid,
		      bfa_plog_eid_t event, uint16_t misc, fchs_t *fchdr,
		      uint32_t pld_w0)
{
	bfa_plog_rec_t  lp;
	uint32_t       *tmp_int = (uint32_t *) fchdr;
	uint32_t        ints[BFA_PL_INT_LOG_SZ];

	if (plog->plog_enabled) {
		bfa_os_memset(&lp, 0, sizeof(bfa_plog_rec_t));

		ints[0] = tmp_int[2];
		ints[1] = tmp_int[3];
		ints[2] = tmp_int[6];
		ints[3] = pld_w0;

		bfa_plog_intarr(plog, mid, event, misc, ints, 4);
	}
}

void
bfa_plog_clear(bfa_plog_t *plog)
{
	plog->head = plog->tail = 0;
}

void
bfa_plog_enable(bfa_plog_t *plog)
{
	plog->plog_enabled = 1;
}

void
bfa_plog_disable(bfa_plog_t *plog)
{
	plog->plog_enabled = 0;
}

bfa_boolean_t
bfa_plog_get_setting(bfa_plog_t *plog)
{
	return((bfa_boolean_t)plog->plog_enabled);
}
